extjs_lcss(hq, "static/extjs/resources/css/ext-all-notheme.css");
extjs_lcss(hq, "redir/theme.debug.css");
extjs_lcss(hq, "static/livegrid/resources/css/ext-ux-livegrid.css");
 extjs_lcss(hq, "static/extjs/examples/ux/gridfilters/css/GridFilters.css");
 extjs_lcss(hq, "static/extjs/examples/ux/gridfilters/css/RangeMenu.css");
 extjs_lcss(hq, "static/extjs/examples/ux/css/Spinner.css");
 extjs_lcss(hq, "static/xcheckbox/xcheckbox.css");
 extjs_lcss(hq, "static/app/ext.css");
extjs_lcss(hq, "redir/theme.app.debug.css");
extjs_load(hq, "redir/locale.js");
extjs_load(hq, "static/extjs/adapter/ext/ext-base-debug.js");
 extjs_load(hq, "static/extjs/ext-all-debug.js");
 extjs_load(hq, "static/app/i18n.js");
 extjs_load(hq, "static/app/extensions.js");
 extjs_load(hq, "static/app/marked.js");
 extjs_load(hq, "static/livegrid/livegrid-all.js");
 extjs_load(hq, "static/lovcombo/lovcombo-all.js");
 extjs_load(hq, "static/multiselect/multiselect.js");
 extjs_load(hq, "static/multiselect/ddview.js");
 extjs_load(hq, "static/xcheckbox/xcheckbox.js");
 extjs_load(hq, "static/checkcolumn/CheckColumn.js");
 extjs_load(hq, "static/extjs/examples/ux/gridfilters/GridFilters.js");
 extjs_load(hq, "static/extjs/examples/ux/gridfilters/filter/Filter.js");
 extjs_load(hq, "static/extjs/examples/ux/gridfilters/filter/BooleanFilter.js");
 extjs_load(hq, "static/extjs/examples/ux/gridfilters/filter/DateFilter.js");
 extjs_load(hq, "static/extjs/examples/ux/gridfilters/filter/ListFilter.js");
 extjs_load(hq, "static/extjs/examples/ux/gridfilters/filter/NumericFilter.js");
 extjs_load(hq, "static/extjs/examples/ux/gridfilters/filter/StringFilter.js");
 extjs_load(hq, "static/extjs/examples/ux/gridfilters/menu/ListMenu.js");
 extjs_load(hq, "static/extjs/examples/ux/gridfilters/menu/RangeMenu.js");
 extjs_load(hq, "static/extjs/examples/ux/Spinner.js");
 extjs_load(hq, "static/extjs/examples/ux/SpinnerField.js");
 extjs_load(hq, "static/app/i18n-post.js");
 extjs_load(hq, "static/app/namespace.js");
 extjs_load(hq, "static/smoothie.js");
 extjs_load(hq, "static/app/comet.js");
 extjs_load(hq, "static/app/cteditor.js");
 extjs_load(hq, "static/app/acleditor.js");
 extjs_load(hq, "static/app/caclient.js");
 extjs_load(hq, "static/app/tvadapters.js");
 extjs_load(hq, "static/app/idnode.js");
 extjs_load(hq, "static/app/esfilter.js");
 extjs_load(hq, "static/app/mpegts.js");
 extjs_load(hq, "static/app/timeshift.js");
 extjs_load(hq, "static/app/chconf.js");
 extjs_load(hq, "static/app/epg.js");
 extjs_load(hq, "static/app/dvr.js");
 extjs_load(hq, "static/app/epggrab.js");
 extjs_load(hq, "static/app/config.js");
 extjs_load(hq, "static/app/tvhlog.js");
 extjs_load(hq, "static/app/status.js");
 extjs_load(hq, "static/app/wizard.js");
 extjs_load(hq, "static/tv.js");
 extjs_load(hq, "static/app/servicemapper.js");
 extjs_load(hq, "static/app/tvheadend.js");
