// Automatically generated by configure - DO NOT EDIT!

#include "build.h"

const char* build_config_str =
  "Configure arguments:\n"
  "  --prefix=/usr --sysconfdir=/etc --mandir=/usr/share/man --infodir=/usr/share/info --localstatedir=/var --disable-libav --disable-ffmpeg_static --disable-libx264_static --disable-libx265_static --disable-libvpx_static --disable-libtheora_static --disable-libvorbis_static --disable-libfdkaac_static --disable-libmfx_static --python=python3 --nowerror" "\n"
  "" "\n"
  "Compiler:" "\n"
  "  Using C compiler:                        cc" "\n"
  "  Using C flags:                           -march=i686 -mtune=generic -O2 -pipe -fno-plt -fexceptions         -Wp,-D_FORTIFY_SOURCE=3 -Wformat -Werror=format-security         -fstack-clash-protection          -fno-omit-frame-pointer -mno-omit-leaf-frame-pointer -g -ffile-prefix-map=/build/tvheadend/src=/usr/src/debug/tvheadend -flto=auto -std=gnu17" "\n"
  "  Using LD flags:                          -Wl,-O1 -Wl,--sort-common -Wl,--as-needed -Wl,-z,relro -Wl,-z,now          -Wl,-z,pack-relative-relocs -flto=auto" "\n"
  "  Build for arch:                          i686" "\n"
  "" "\n"
  "Binaries:" "\n"
  "  Using PYTHON:                            python3" "\n"
  "  Using GZIP:                              gzip" "\n"
  "  Using BZIP2:                             bzip2" "\n"
  "" "\n"
  "Options:" "\n"
  "  pie                                      yes" "\n"
  "  ccdebug                                  no" "\n"
  "  cwc                                      yes" "\n"
  "  capmt                                    yes" "\n"
  "  constcw                                  yes" "\n"
  "  linuxdvb                                 yes" "\n"
  "  satip_server                             yes" "\n"
  "  satip_client                             yes" "\n"
  "  hdhomerun_client                         yes" "\n"
  "  hdhomerun_static                         yes" "\n"
  "  iptv                                     yes" "\n"
  "  tsfile                                   yes" "\n"
  "  dvbscan                                  yes" "\n"
  "  timeshift                                yes" "\n"
  "  trace                                    yes" "\n"
  "  imagecache                               yes" "\n"
  "  avahi                                    yes" "\n"
  "  zlib                                     yes" "\n"
  "  libav                                    no" "\n"
  "  ffmpeg_static                            no" "\n"
  "  libx264                                  yes" "\n"
  "  libx264_static                           no" "\n"
  "  libx265                                  yes" "\n"
  "  libx265_static                           no" "\n"
  "  libvpx                                   yes" "\n"
  "  libvpx_static                            no" "\n"
  "  libtheora                                yes" "\n"
  "  libtheora_static                         no" "\n"
  "  libvorbis                                yes" "\n"
  "  libvorbis_static                         no" "\n"
  "  libfdkaac                                yes" "\n"
  "  libfdkaac_static                         no" "\n"
  "  nvenc                                    no" "\n"
  "  qsv                                      no" "\n"
  "  libmfx_static                            no" "\n"
  "  inotify                                  yes" "\n"
  "  epoll                                    yes" "\n"
  "  uriparser                                yes" "\n"
  "  ccache                                   no" "\n"
  "  tvhcsa                                   yes" "\n"
  "  bundle                                   no" "\n"
  "  pngquant                                 no" "\n"
  "  dvbcsa                                   no" "\n"
  "  dvben50221                               no" "\n"
  "  kqueue                                   no" "\n"
  "  dbus_1                                   yes" "\n"
  "  android                                  no" "\n"
  "  tsdebug                                  no" "\n"
  "  gtimer_check                             no" "\n"
  "  slow_memoryinfo                          no" "\n"
  "  libsystemd_daemon                        no" "\n"
  "  pcloud_cache                             yes" "\n"
  "  execinfo                                 yes" "\n"
  "  mmx                                      yes" "\n"
  "  sse2                                     yes" "\n"
  "  W_unused_result                          yes" "\n"
  "  strlcat                                  yes" "\n"
  "  strlcpy                                  yes" "\n"
  "  fdatasync                                yes" "\n"
  "  getloadavg                               yes" "\n"
  "  atomic32                                 yes" "\n"
  "  atomic64                                 yes" "\n"
  "  atomic_time_t                            yes" "\n"
  "  bitops64                                 yes" "\n"
  "  lockowner                                yes" "\n"
  "  gmtoff                                   yes" "\n"
  "  recvmmsg                                 yes" "\n"
  "  sendmmsg                                 yes" "\n"
  "  ifnames                                  yes" "\n"
  "  py_gzip                                  yes" "\n"
  "  bin_pkg_config                           yes" "\n"
  "  bin_xgettext                             yes" "\n"
  "  bin_msgmerge                             yes" "\n"
  "  bin_gzip                                 yes" "\n"
  "  bin_bzip2                                yes" "\n"
  "  ssl                                      yes" "\n"
  "  linuxdvbapi                              yes" "\n"
  "  upnp                                     yes" "\n"
  "  inotify_h                                yes" "\n"
  "  inotify_init1                            yes" "\n"
  "  epoll_create1                            yes" "\n"
  "  mpegts                                   yes" "\n"
  "  mpegts_dvb                               yes" "\n"
  "" "\n"
  "Packages:" "\n"
  "  openssl                                  3.5.1" "\n"
  "  zlib                                     1.3.2" "\n"
  "  liburiparser                             1.0.0" "\n"
  "  avahi-client                             0.8" "\n"
  "  dbus-1                                   1.14.10" "\n"
  "" "\n"
  "Installation paths:" "\n"
  "  Prefix:                                  /usr" "\n"
  "  Binaries:                                ${prefix}/bin" "\n"
  "  Libraries:                               ${prefix}/lib" "\n"
  "  Data files:                              ${prefix}/share" "\n"
  "  Man pages:                               /usr/share/man" "\n"
  "" "\n"
  "";
