#define PLATFORM_LINUX 1
#define ENABLE_PIE 1
#define CONFIG_PIE 1
#define ENABLE_CCDEBUG 0
#define ENABLE_CWC 1
#define CONFIG_CWC 1
#define ENABLE_CAPMT 1
#define CONFIG_CAPMT 1
#define ENABLE_CONSTCW 1
#define CONFIG_CONSTCW 1
#define ENABLE_LINUXDVB 1
#define CONFIG_LINUXDVB 1
#define ENABLE_SATIP_SERVER 1
#define CONFIG_SATIP_SERVER 1
#define ENABLE_SATIP_CLIENT 1
#define CONFIG_SATIP_CLIENT 1
#define ENABLE_HDHOMERUN_CLIENT 1
#define CONFIG_HDHOMERUN_CLIENT 1
#define ENABLE_HDHOMERUN_STATIC 1
#define CONFIG_HDHOMERUN_STATIC 1
#define ENABLE_IPTV 1
#define CONFIG_IPTV 1
#define ENABLE_TSFILE 1
#define CONFIG_TSFILE 1
#define ENABLE_DVBSCAN 1
#define CONFIG_DVBSCAN 1
#define ENABLE_TIMESHIFT 1
#define CONFIG_TIMESHIFT 1
#define ENABLE_TRACE 1
#define CONFIG_TRACE 1
#define ENABLE_IMAGECACHE 1
#define CONFIG_IMAGECACHE 1
#define ENABLE_AVAHI 1
#define CONFIG_AVAHI 1
#define ENABLE_ZLIB 1
#define CONFIG_ZLIB 1
#define ENABLE_LIBAV 0
#define ENABLE_FFMPEG_STATIC 0
#define ENABLE_LIBX264 1
#define CONFIG_LIBX264 1
#define ENABLE_LIBX264_STATIC 0
#define ENABLE_LIBX265 1
#define CONFIG_LIBX265 1
#define ENABLE_LIBX265_STATIC 0
#define ENABLE_LIBVPX 1
#define CONFIG_LIBVPX 1
#define ENABLE_LIBVPX_STATIC 0
#define ENABLE_LIBTHEORA 1
#define CONFIG_LIBTHEORA 1
#define ENABLE_LIBTHEORA_STATIC 0
#define ENABLE_LIBVORBIS 1
#define CONFIG_LIBVORBIS 1
#define ENABLE_LIBVORBIS_STATIC 0
#define ENABLE_LIBFDKAAC 1
#define CONFIG_LIBFDKAAC 1
#define ENABLE_LIBFDKAAC_STATIC 0
#define ENABLE_NVENC 0
#define ENABLE_QSV 0
#define ENABLE_LIBMFX_STATIC 0
#define ENABLE_INOTIFY 1
#define CONFIG_INOTIFY 1
#define ENABLE_EPOLL 1
#define CONFIG_EPOLL 1
#define ENABLE_URIPARSER 1
#define CONFIG_URIPARSER 1
#define ENABLE_CCACHE 0
#define ENABLE_TVHCSA 1
#define CONFIG_TVHCSA 1
#define ENABLE_BUNDLE 0
#define ENABLE_PNGQUANT 0
#define ENABLE_DVBCSA 0
#define ENABLE_DVBEN50221 0
#define ENABLE_KQUEUE 0
#define ENABLE_DBUS_1 1
#define CONFIG_DBUS_1 1
#define ENABLE_ANDROID 0
#define ENABLE_TSDEBUG 0
#define ENABLE_GTIMER_CHECK 0
#define ENABLE_SLOW_MEMORYINFO 0
#define ENABLE_LIBSYSTEMD_DAEMON 0
#define ENABLE_PCLOUD_CACHE 1
#define CONFIG_PCLOUD_CACHE 1
#define ENABLE_EXECINFO 1
#define CONFIG_EXECINFO 1
#define ENABLE_MMX 1
#define CONFIG_MMX 1
#define ENABLE_SSE2 1
#define CONFIG_SSE2 1
#define ENABLE_W_UNUSED_RESULT 1
#define CONFIG_W_UNUSED_RESULT 1
#define ENABLE_STRLCAT 1
#define CONFIG_STRLCAT 1
#define ENABLE_STRLCPY 1
#define CONFIG_STRLCPY 1
#define ENABLE_FDATASYNC 1
#define CONFIG_FDATASYNC 1
#define ENABLE_GETLOADAVG 1
#define CONFIG_GETLOADAVG 1
#define ENABLE_ATOMIC32 1
#define CONFIG_ATOMIC32 1
#define ENABLE_ATOMIC64 1
#define CONFIG_ATOMIC64 1
#define ENABLE_ATOMIC_TIME_T 1
#define CONFIG_ATOMIC_TIME_T 1
#define ENABLE_BITOPS64 1
#define CONFIG_BITOPS64 1
#define ENABLE_LOCKOWNER 1
#define CONFIG_LOCKOWNER 1
#define ENABLE_GMTOFF 1
#define CONFIG_GMTOFF 1
#define ENABLE_RECVMMSG 1
#define CONFIG_RECVMMSG 1
#define ENABLE_SENDMMSG 1
#define CONFIG_SENDMMSG 1
#define ENABLE_IFNAMES 1
#define CONFIG_IFNAMES 1
#define ENABLE_PY_GZIP 1
#define CONFIG_PY_GZIP 1
#define ENABLE_BIN_PKG_CONFIG 1
#define CONFIG_BIN_PKG_CONFIG 1
#define ENABLE_BIN_XGETTEXT 1
#define CONFIG_BIN_XGETTEXT 1
#define ENABLE_BIN_MSGMERGE 1
#define CONFIG_BIN_MSGMERGE 1
#define ENABLE_BIN_GZIP 1
#define CONFIG_BIN_GZIP 1
#define ENABLE_BIN_BZIP2 1
#define CONFIG_BIN_BZIP2 1
#define ENABLE_SSL 1
#define CONFIG_SSL 1
#define ENABLE_LINUXDVBAPI 1
#define CONFIG_LINUXDVBAPI 1
#define ENABLE_UPNP 1
#define CONFIG_UPNP 1
#define ENABLE_INOTIFY_H 1
#define CONFIG_INOTIFY_H 1
#define ENABLE_INOTIFY_INIT1 1
#define CONFIG_INOTIFY_INIT1 1
#define ENABLE_EPOLL_CREATE1 1
#define CONFIG_EPOLL_CREATE1 1
#define ENABLE_MPEGTS 1
#define CONFIG_MPEGTS 1
#define ENABLE_MPEGTS_DVB 1
#define CONFIG_MPEGTS_DVB 1
extern const char* build_config_str;
extern const char* build_timestamp;
#define TVHEADEND_DATADIR "/usr/share/tvheadend"
