static const char *tvh_locale_1[] = {
NULL, NULL
};

static const char *tvh_locale_2[] = {
NULL, NULL
};

static const char *tvh_locale_3[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nلمزيد من المعلومات يرجى زيارة الموقع الالكتروني https://tvheadend.org\n",
"# Mapped channels", "# قنوات مدرجة",
"# Muxes", "# الترددات",
"# Services", "# الخوادم",
"(Read-only)", "(قرائة فقط)",
"1 day", "1 يوم",
"1 month", "1 شهر",
"1 week", "1 أسبوع",
"1 year", "1 سنة",
"2 months", "2 شهران",
"2 weeks", "2 أسبوعان",
"2 years", "2 سنتان",
"2-Port switch (universal LNB)", "محول إل إن بي ذو مخرجين (عالمي)",
"3 days", "3 أيام",
"3 months", "3 شهور",
"3 weeks", "3 أسابيع",
"3 years", "3 سنوات",
"4-Port switch (universal LNB)", "محول إل إن بي ذو أربع مخارج (عالمي)",
"5 days", "5 أيام",
"6 months", "6 أشهر",
"A", "أ",
"AA", "أأ",
"AAC audio", "صوت AAC",
"AB", "أب",
"AC-3 detection", "إكتشاف AC-3",
"AC3 audio", "صوت AC3",
"ACTIVE", "فعال",
"ANY", "أي",
"AUTO", "تلقائي",
"Aborted by user", "تم الإلغاء بواسطة المستخدم",
"Accept zero value for TSID", "قبول قيمة صفر لل TSID",
"Access", "منفذ",
"Access (ACL)", "منفذ (ACL)",
"Active", "فعال",
"Adapter in use by another subscription", "المحول قيد الإستخدام",
"Admin", "المسؤول",
"Admin password", "كلمة السر للمسؤول",
"Admin username", "أسم المستخدم للمسئول",
"Administrator login", "دخول المسؤول",
"Adult movie", "افلام إباحية",
"Advanced", "متقدم",
NULL, NULL
};

static const char *tvh_locale_4[] = {
"# Services", "# Xidmətlər",
"(Read-only)", "(Yalnız-oxuma)",
"1 day", "1 gün",
"1 month", "1 ay",
"1 week", "1 həftə",
"1 year", "1 il",
"2 months", "2 ay",
"2 weeks", "2 həftə",
"2 years", "2 il",
"2-Port switch (universal LNB)", "2-Portlu Switch (Universal LNB)",
"3 days", "3 gün",
"3 months", "3 ay",
"ACTIVE", "AKTİV",
"Aborted by user", "İstifadəçi tərəfindən dayandırıldı",
"Admin password", "Admin şifrəsi",
"Admin username", "Admin adı",
"Administrator login", "Administrator girişi",
"Adult movie", "18+ Videol",
"Adventure", "Macəra",
"Advertisement / Shopping", "Reklam / Alış-veriş",
"Allow/Disallow access to the 'Configuration' tab.", "\"Konfiqurasiya\" bölməsinə Giriş/Qadağa",
"Allowed network", "İcazə verilən şəbəkə",
"Allowed networks", "İcazə verilən şəbəkələr",
"Animals", "Heyvanlar",
"Any", "Hər hansı",
"Auto record", "Avto yazılış",
"Basic", "Əsas",
"Blue", "Mavi",
"Channel", "Kanal",
"Channel icon", "Kanal işarəsi",
"Channel icon path", "Kanal işarəsinin yerləşdiyi bölmə",
"Channel name", "Kanal adı",
"Channel number", "Kanal nömrəsi",
"Channels", "Kanallar",
"Cinema", "Kino",
"Comedy", "Komediya",
"Comment", "Rəy",
"Configuration", "Konfiqurasiya",
"Configuration name", "Konfiqurasiya adı",
"Satellite position.", "Sputnikin mövqeyi",
"Satellite positions", "Sputnik mövqeləri",
"Scan after creation", "Yaratdıqdan dərhal sonra Axtar",
"Scan progress", "Axtarır",
"Scan result", "Axtarış nəticələri",
"Server details.", "Server barədə",
"Service", "Xidmət",
NULL, NULL
};

static const char *tvh_locale_5[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nЗа повече информация посетете сайта на Tvheadend\nhttps://tvheadend.org\n",
"# Mapped channels", "# Свързани канали",
"# Muxes", "# Муксове",
"# Services", "# Услуги",
"# tune repeats", "Повтаряне на търсенето",
"(Default profile)", "(Профил по подразбиране)",
"(Read-only)", "(Само за четене)",
"1 day", "1 ден",
"1 month", "1 месец",
"1 week", "1 седмица",
"1 year", "1 година",
"2 months", "2 месеца",
"2 weeks", "2 седмици",
"2 years", "2 години",
"2-Port switch (universal LNB)", "2-портов ключ (универсален конвертор)",
"3 days", "3 дни",
"3 months", "3 месеца",
"3 weeks", "3 седмици",
"3 years", "3 години",
"4-Port switch (universal LNB)", "4-портов ключ (универсален конвертор)",
"5 days", "5 дни",
"6 months", "6 месеца",
"AC-3 = descriptor 6", "AC-3 = дескриптор 6",
"AC-3 detection", "АС-3 търсене",
"ACTIVE", "АКТИВНО",
"AF41", "AF41 /препоръчително/",
"ANY", "всички",
"ATSC source ID", "ATSC ID на източника",
"AUTO", "Автоматично",
"Aborted by user", "Прекратено от потребителя",
"Accept remote subscription weight", "Преиемане значението на дистанционния абонамент",
"Accept the remote subscription weight (from the SAT>IP client).", "Преиемане значението на дистанционния абонамент (от SAT>IP клиент)",
"Access", "Достъп",
"Action", "Действие",
"Adapter in use by another subscription", "Адаптера е зает от друг абонамент",
"Add file and line numbers to debug", "Добавяне на имена на файлове и номера на редове към информацията за дебъгване",
"Add the thread ID to debug", "Добавяне на ид. на нишките към информацията за дебъгване",
"Admin", "Администратор",
"Admin password", "Парола на администратор",
"Admin username", "Име за администратор",
"Administrator login", "Вписване като администратор",
"Adult movie", "Филм за възрастни",
"Advanced", "Разширено",
"Advanced (non-universal LNBs, rotors, etc.)", "Разширено (неуниверсални конвертори, мотор и др.)",
"Adventure", "Приключения",
"Advertisement / Shopping", "Реклама/пазаруване",
"All (Streaming plus DVR)", "Всички(стриим плюс DVR",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Всички знаци, които може да представляват проблем, ще бъдат заменени с долна черта. Вижте помощната информация за подробности.",
"All proceeds are used to support server infrastructure and buy test equipment.", "Всички постъпления се използват за поддръжка на сървърната инфраструктура, тестове и оборудване",
"Allow high bitrate mode (CI+ CAMs only).", "Позволяване на режим на висока побитова скорост (само за модули CI+ CAM).",
"Allowed network", "Разрешена мрежа",
"Allowed networks", "Разрешени мрежи",
"Alternate PID path", "Алтернативен PID път",
"Alternate configuration path", "Път до алтернативна конфигурация",
"Altitude (in meters).", "Надморска височина (в м.)",
"Altitude (meters)", "Височина (метри)",
"Animals", "Животни",
"Any", "Всички",
"Arts", "Изкуства",
"Athletics", "Атлетика",
"Audio bitrate (kb/s) (0=auto)", "Побитова скорост на звука (кб/сек) (0=автоматично) ",
"Audio bitrate to use for transcoding.", "Побитова скорост на звука при прекодиране.",
"Audio codec", "Аудио кодек",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Звуков кодек, който да се използва при прекодиране. „Без“ ще изключи извеждането на звук.",
"Auto", "Авто",
"Auto check disabled", "Авт. проверка е изключена",
"Auto check enabled", "Авт. проверка е включена",
"Auto record", "Автоматичен запис",
"Auto record caption", "Автоматичен запис на надписите",
"Auto recording%s%s", "Автоматично записване%s%s",
"Auto time record", "Автоматочно време за запис",
"Auto-Map to channels", "Автоматично мапване към каналите ",
"Automatic checking", "Автоматично проверяване",
"Automatic schedule entry not found", "Елементът за авт. планиране не е намерен",
"Automatically map EPG source", "Авт. свързване на източник на справочник",
"Automatically name from network", "Автоматично именоване от мрежата",
"Autorec maximum count (0=unlimited)", "Максимален брой авт. записи (0=неограничен)",
"Autorec maximum schedules limit (0=unlimited)", "Максимален брой планирани авт. записи (0=неограничен)",
"Bad", "Лош",
"Bandwidth", "Пропускателната способност",
"Based on software from", "Базиран на софтуер от",
"Basic", "Основно",
"Bind to specific local IP address.", "Обвържи към специфичен локален IP адрес",
"Bouquet", "Букет",
"Bouquet (auto)", "Букет (авто)",
"Bouquets", "Букети",
"Brand", "Марка",
"Broadcast", "Предаване",
"Broadcasting", "Излъчване на предаване",
"Build", "Създаване",
"CA client %i", "CA клиент %i",
"CA identification", "CA идентифициране",
"CA provider", "Доставчик на условен достъп",
"CAPMT interval (ms)", "CAPMT интервал (ms)",
"CAPMT query interval (ms)", "CAPMT интервал за запитване (ms)",
"CRID authority", "CRID пълномощно",
"Cache scheme", "Схема на кеша",
"Camd.socket filename / IP Address (TCP mode)", "Име на файл Camd.socket / IP адрес (режим TCP)",
"Cartoons", "Детски",
"Channel", "Канал",
"Channel does not exist", "Канала не съществува",
"Channel icon", "Иконка на канала",
"Channel icon URL.", "Адрес за иконката на канала.",
"Channel icon as defined in EPG data.", "Иконка на канала, определена от данните в справочника.",
"Channel icon name scheme", "Схема за именуване на иконките на каналите",
"Channel icon path", "Път до иконката на канала",
"Channel name", "Име на канала",
"Channel number", "Номера на канала",
"Channel number offset", "Отместване номера на канала",
"Channel numbers from", "Номера на канала от",
"Channel tag", "Етикет на канала",
"Channel tag reference", "Описание на етикета",
"Channel tags", "Етикети на каналите",
"Channels", "Канали",
"Character set", "Енкодинг",
"Check availability", "Проверка на наличноста",
"Children's / Youth programs", "Детски / Младежки ",
"Cinema", "Кино",
"Class", "Клас",
"Classical", "Класически",
"Classical music", "Класическа музика",
"Clean effects", "Изчистване на ефектите",
"Client name", "Име на клиента",
"Clone scheduled entry on error", "Клониране на планирания елемент при грешка",
"Comedy", "Комедия",
"Command delay time (ms) (10-200)", "Времезакъснение на командата (ms) (10-200)",
"Command time (ms) (10-100)", "Време за командата (ms) (10-100)",
"Comment", "Коментар",
"Commercial break", "Рекламна пауза",
"Committed", "Ангажиран",
"Completed OK", "Завършено без проблем",
"Configuration", "Конфигурация",
"Configuration ID", "ИД на конфигурация",
"Configuration name", "Име на конфигурацията",
"Configuration version", "Версия на конфигурацията",
"Connection limit type", "Тип на ограничение на връзката",
"Constellation", "Съзвездие",
"Container", "Контейнер",
"Content character set", "Определяне знаците за съдържанието",
"Content reference identifier authority.", "Бележка за съдържанието на пълномощното за идентификация",
"Content type", "Съдържание Тип",
"Contest", "Състезание",
"Continue recording for x minutes after scheduled stop time", "Продължаване на записа за Х минути след планирането време за край",
"Continue recording for x minutes after scheduled stop time.", "Продължаване на записа за Х минути след планирането време за край.",
"Cookie expiration (days)", "Запазване на бисквитките (days)",
"Cooking", "Готварство",
"Copy codec type", "Копиране на типа кодек",
"Copy layout", "Копиране на макета",
"Count of objects", "Брой на обектите",
"Create bouquet", "Създаване на букет",
"Create network name tags", "Създаване на етикети с имената на мрежите",
"Create network name tags (set by provider).", "Създаване на етикети с имената на мрежите (зададено от доставчика).",
"Create network tags", "Създаване на етикети за мрежите",
"Create provider name tags", "Създаване на етикети с имената на доставчиците",
"Create provider tags", "Създаване на етикети за доставчиците",
"Create tags", "Създаване на етикети",
"Created", "Създаден",
"Creator", "Създател",
"Cron", "Планирани задачи",
"Cron multi-line", "мултилиния на Cron",
"Culture (without music)", "Култура",
"Current RTSP port.", "Текущ RTSP порт",
"Current affairs", "Актуални въпроси",
"Current tuner configuration.", "Конфигурация на избрания тунер.",
"Custom HTTP headers", "Персонални HTTP хедъри",
"DBus - use the session message bus instead of the system one", "DBus – използване на съобщението на сесията вместо системното",
"DES key", "DES ключ",
"DSCP/TOS for streaming", "DSCP/TOS за стрииминг",
"DVB EPG ID", "DVB – ИД на справочник",
"DVB scan files", "DVB – сканиране за файлове",
"DVB scan files path", "DVB – път за сканиране на файлове",
"DVR", "Дигитален видео рекордер",
"DVR - Auto-recording (Autorecs)", "DVR – Автоматично записване (авт. запис)",
"DVR behavior", "DVR режим на работа",
"DVR configuration", "DVR – конфигурация",
"DVR configuration profiles", "Профили на конфигурация на DVR",
"DVR file retention period", "DVR – период за съхранение на файловете",
"DVR log retention", "DVR – запазване на журнала",
"DVR log retention period", "DVR – период за запазване на журнала",
"DVR override: high", "DVR презапис: висок приоритет",
"DVR override: important", "DVR презапис: важен приоритет",
"DVR override: low", "DVR презапис: нисък приоритет",
"DVR override: normal", "DVR презапис: нормален приоритет",
"DVR override: unimportant", "DVR презапис: неважен приоритет",
"Dance", "Танци",
"Data errors", "Грешка в данните",
"Days of Week", "Ден от седмицата",
"Days of week", "Дни от седмицата",
"Debate", "Дебати",
"Debug libav log", "Журнал за дебъгване на „libav“",
"Debug options", "Опции за откриване грешки",
"Debug subsystems", "откриване грешки в субсистемите",
"Debug to syslog", "Системен журнал за дебъгване",
"Debug trace (low-level)", "Трасиране на грешките (ниско ниво)",
"Debugging", "Дебъгване",
"Default", "По подразбиране",
"Default language(s)", "Език(езици) по подразбиране",
"Default language.", "Език по подразбиране.",
"Default priority", "Приоритет По подразбиране",
"Delivery system", "Система за доставка",
"Demux path", "Път за разделяне",
"Descramble services (limit per mux)", "Декодиращи услуги (ограничение по mux)",
"Descrambler", "Декодиране",
"Descrambler buffer (TS packets)", "Буфер за декодиране (TS пакети)",
"Description", "Описание",
"Detective", "Детективски",
"Device ID", "ИД на устройството",
"Device model", "Модел на устройството",
"Device path", "Път до устройството",
"Directory", "Папка",
"Directory permissions (octal, e.g. 0775)", "Разрешения за директориите  (octal, e.g. 0775)",
"Disable", "Изключване",
"Disable DVB bouquets", "Изключване на букетите на DVB",
"Disable SAT>IP client", "Изключване на клиента за SAT>IP",
"Disable all access control checks", "Изключване на всички проверки за достъп",
"Disable debug on stderr", "Изключване на извеждането на информацията за дебъгване на стандартния изход за грешки",
"Disable device/firmware-specific workarounds", "Изключване на специфичните за устройства/фърмуери поправки",
"Disable syslog (all messages)", "Изключване на системните журнали (всички съобщения)",
"Disabled", "Изключено",
"Discussion", "Дискусии",
"DiseqC repeats", "DiSEqC повторения",
"Do not use", "Без",
"Documentary", "Документални",
"Don't backup configuration tree at upgrade", "Да се се прави резервно копи на конфигурацията при надграждане",
"Don't include title in filename", "Заглавието да не се включва в името на файла",
"Don't keep", "Без запазване",
"Don't re-record", "Без презаписване",
"Don't reschedule", "Да не се планира отново",
"Drama", "Драма",
"Duplicate handling", "Дублиране на потдръжката",
"EIT - skip TSID check", "EIT скрии TSID проверката",
"EPG name", "Име на справочника",
"EPG scan", "Сканиране на справочника",
"EPG source", "Източник на справочник",
"EPG update window", "Прозорец на обновяване на справочника",
"Economics", "Икономични",
"Education", "Образование",
"Educational", "Образователни",
"Elementary stream filter", "Филтър на елементарния стриим",
"Enable", "Включване",
"Enable (auto)", "Включване (авт.)",
"Enable DBus", "Включване на DBus",
"Enable NTP driver", "Включване на драйвера за NTP",
"Enable coredumps for daemon", "Включване на „coredump“ за демона",
"Enable debug on stderr", "Извеждане на информацията за дебъгване на стандартния изход за грешки",
"Enable debug subsystems", "Включване на подсистемите за дебъгване",
"Enable debug to file", "Извеждане на информацията за дебъгване във файл",
"Enable debug to syslog", "Извеждане на информацията за дебъгване в системния журнал",
"Enable trace subsystems", "Включване на подсистемите за проследяване",
"Enabled", "Включено",
"Encrypted", "Кодирано",
"Environment", "Среда",
"Environment (pipe)", "Среда (pipe)",
"Episode", "Епизод",
"Episode %d", "Епизод %d",
"Errors", "Грешки",
"Every day", "Всеки ден",
"Expeditions", "Експедиция",
"Experimental film", "Експериментален филм",
"Expert", "Експерт",
"External", "Външен",
"External URL", "Външен адрес",
"FAIL", "ПРОВАЛ",
"Factual topics", "Фактически теми",
"Fantasy", "Фантастични",
"File missing", "Файлът липсва",
"File not created", "Файлът не е създаден",
"Filename", "Име на файла",
"Film", "Фимл",
"Fine arts", "Изящни изкуства",
"Fitness and health", "Фитнес и здраве",
"Folk", "Фолк",
"Folkloric", "Фолклорни",
"Football", "Футбол",
"Force (auto)", "Принудително (автоматично)",
"Force CA ID (e.g. 0x2600)", "Принудително CA ID (e.g. 0x2600)",
"Force pilot for DVB-S2", "Принудително пилот за DVB-S2",
"Force priority", "Принудителен приоритет",
"Force teardown delay", "Принудително закъснение на прекъсването",
"Foreign countries", "Чужди страни",
"Fork and run as daemon", "Отделяне и стартиране като демон",
"Frequency", "Честота",
"Frequency (Hz)", "Честота (Hz)",
"Frequency (MHz)", "Честота (MHz)",
"Frequency (kHz)", "Честота (kHz)",
"Fri", "Пет",
"Friendly name", "Кратко име",
"Full DiseqC", "Пълен DiseqC",
"Full-text", "Пълен текст",
"Further education", "Бъдещо образование",
"Game show", "Шоу",
"Gardening", "Градинарство",
"General", "Основни",
"General configuration", "Основна конфигурация",
"Generic", "Общ",
"Generic options", "Основни настройки",
"HD: high definition", "HD: висока резолюция",
"HTSP Default Stream Settings", "Основни настройки за поточно предаване чрез HTSP по подразбиране",
"Handicraft", "Занаят",
"Hearing impaired", "Увреден слух",
"Hierarchy", "Йерархия",
"High", "Високо",
"High bitrate mode (CI+ CAMs only)", "Режим на висока побитова скорост (само за модули CI+ CAM)",
"Historical movie", "Исторически филми",
"Horror", "Ужаси",
"Hostname/IP", "Име на хоста/IP",
"IDLE", "БЕЗДЕЙСТВИЕ",
"IP address", "IP Адрес",
"IP blocking", "ИП блокиране",
"Icon (full URL)", "Иконка (пълен адрес)",
"Icon URL", "Адрес на иконката",
"Icon has title", "Иконката има заглавие",
"Icons from", "Иконки от",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Ако няма съществуващ акаунт, тогава създайте такъв\nбез потребител и парола. Използвайте го внимателно, тъй като\nтой ще даде административен достъп от всяко точка на света\nдо вашата инсталация на Tvheadend, докато не създадете или\nредактирате правата за достъп чрез уеб интерфейса.",
"If you'd like to support the project, please consider a donation.", "Ако искате да помогнете за проекта, направете дарение. ",
"Ignore EPG (EIT)", "Пренебрегване на справочника",
"Ignore descriptor 5", "Пренебрегване на дескриптор 5",
"Ignore invalid SSL certificate", "Пренебрегване на неправилния SSL сертификат",
"Immediately abort", "Незабавно прекрати",
"Important", "Важно",
"Index", "Индекс",
"Informational", "Информационни",
"Initial scan", "Начално сканиране",
"Input path", "Входящ път",
"Interface", "Интерфейс",
"Internal", "Вътрешен",
"Interview", "Интервю",
"Invalid", "Невалидно",
"Invalid adapter number '%s'\n", "Невалиден номер на адаптер '%s'\n",
"Invalid target", "Невалиден обект",
"Jazz", "Джаз",
"LNB", "Конвертор",
"Language", "Език",
"Language settings", "Езикови настройки",
"Languages", "Езици",
"Last seen", "Последно видян",
"Leisure hobbies", "услуги за свободното време хобита",
"Limit connections", "Ограничение на връзките",
"Listen on IPv6", "Очакване на IPv6",
"Literature", "Литература",
"Local bind IP address", "Локално привързани IP адреси",
"Local channel number", "Локален номер на канал",
"Location", "Местонахождение",
"Log", "Журнал",
"Login password.", "Парола за вписване",
"Login username.", "Потребителско име за вписване.",
"Login/Logout", "Вписване/отписване",
"Low", "Ниско",
"Magazines", "Списания",
"Manufacturer", "Производител",
"Manufacturer URL", " URL Производител",
"Map all services", "Свързване на всички услуги",
"Maximum PIDs", "Максимален брой ид. на процеси",
"Maximum bandwidth (Kbps)", "Максимална честотна лента(Kbps)",
"Maximum length of PIDs", "Максимална дължина на PIDs",
"Maximum timeout (seconds)", "Макс. време на изчакване (секунди)",
"Medicine", "Медицина",
"Melodrama", "Мелодрама",
"Missing In PAT/SDT", "Липсва в PAT/SDT",
"Mode", "Режим",
"Model description", "Описание на Модела",
"Model name", "Име на Модела",
"Model number", "Номер на Модела",
"Modulation", "Модулация",
"Mon", "Понеделник",
"Mono", "Моно",
"More verbose libav log", "По-подробен журнал на „libav“",
"Motor sport", "Моторни спортове",
"Motoring", "Мониторинг",
"Movie", "Филми",
"Music", "Музика",
"Musical", "Музикални",
"Mux", "Мукс",
"Mux UUID", "Мукс UUID",
"Mux not enabled", "Муксът не е включен",
"NONE", "ПРАЗНО",
"Name", "Име",
"Natural sciences", "Природни науки",
"Nature", "Природа",
"Network", "Мрежа",
"Network A", "Мрежа А",
"Network B", "Мрежа Б",
"Network ID (limit scanning)", "ИД на мрежата(ограничено сканиране",
"Network UUID", "UUID Мрежа",
"Network discovery", "Мрежово сканиране",
"Network name", "Име на мрежата",
"Network prefix", "Мрежов префикс",
"Network type", "Вид на мрежата",
"Networks", "Мрежи",
"News", "Новини",
"News magazine", "Новини от списания",
"Next tune delay in ms (0-2000)", "Макс. закъснение при следваща настройка (мсек: 0-2000)",
"No", "Не",
"No PIN", "Без ПИН",
"No access", "Без достъп",
"No adapters specified!\n", "Няма определен адаптер\n",
"No days", "Няма дни",
"No descrambler", "Няма декодиране",
"No free adapter", "Няма свободен адаптер",
"No input detected", "Няма намерен вход",
"No service assigned to channel", "Няма услуга към канала",
"No service enabled", "Няма включени услуги",
"No source available", "Не е наличен източник",
"None", "Празно",
"Normal", "Нормален",
"Not set", "Не зададен",
"Not set (none or channel configuration)", "Без (няма или да се ползва конфигурацията на каналите)",
"Not set (use channel or DVR configuration)", "Няма (ще се използва конфигурацията на каналите или DVR)",
"Number", "Номер",
"Number of tsfile tuners", "Номер на тунерите от tsfile",
"OK", "ДОБРЕ",
"OK (partial)", "ДОБРЕ (частично)",
"OSCam net protocol (rev >= 10389)", "OSCam net протокол (rev >= 10389)",
"Off", "Изключен",
"Older OSCam", "По-стар OSCam",
"On", "Включен",
"Only Bulsatcom 39E", "Само Булсатком 39И",
"Only EIT", "Само EIT",
"Only OpenTV Sky Ausat", "Само OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Само OpenTV Sky Italia",
"Only OpenTV Sky UK", "Само OpenTV Sky UK",
"Only UK Freesat", "Само UK Freesat",
"Only UK Freeview", "Само UK Freeview",
"Only Viasat Baltic", "Само Viasat Baltic",
"Only admin can watch the realtime log.", "Само администратор може да разглежда журнала в реално време.",
"Only allow users with this tag (or those with no tags at all) set in access configuration to use the tag.", "Само потребителите с този етикет (или тези без никакъв етикет), зададен в настройките за достъпа, да могат да използват този етикет.",
"Only digits", "Само цифри",
"Opera", "Опера",
"Output directory for tsdebug", "Изходна директория за TS дебъг",
"Over-the-air", "Ефирен",
"Over-the-air EPG", "Ефирен справочник",
"Over-the-air EPG grabber", "Извличане на ефирен справочник",
"Owner", "Притежател",
"PEND", "ИЗЧАКВАНЕ",
"PIDs 21 in setup", "ИД на процеси 21 в установката",
"Password", "Парола",
"Password2", "Парола2",
"Performing arts", "Приложни изкуства",
"Physiology", "Физиология",
"Pilot", "Пилот",
"Political issues", "Политически въпроси",
"Pop", "Поп",
"Popular culture", "Популярна култура",
"Port", "Порт",
"Position", "Позиция",
"Power save", "Пестене на енергия ",
"Preferred CA PID", "Предпочитан CA PID",
"Presentation", "Презентация",
"Press", "Преса",
"Priority", "Приоритет",
"Priority (-10..10)", "Приоритет (-10..10)",
"Private", "Лично",
"Profile name", "Име на профила",
"Provider", "Доставчик",
"Provider ID", "ИД на доставчик",
"Psychology", "Психология",
"Puppets", "Куклени детски",
"Quiz", "Викторина",
"Radio", "Радио",
"Record all", "Запиши всичко",
"Religion", "Религия",
"Religious", "Религиозни",
"Remarkable people", "Забележителни хора",
"Reports", "Рапорт",
"Rerun of", "Върни от",
"Reserved", "Резервирано",
"Respawn (pipe)", "Respawn",
"Restart on error", "Рестартиране при грешка",
"Rewrite EIT", "Презапиши EIT",
"Rewrite PAT", "Презапиши PAT",
"Rewrite PMT", "Презапиши PMT",
"Rewrite SDT", "Презапиши SDT",
"Rock", "Рок",
"Rolloff", "Пропадане",
"Romance", "Романтика",
"Run as group", "Стартиране като група",
"Run as user", "Стартиране като потребител",
"Running", "Стартиран",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "Номер на порт за RTSP на сървъра за SAT>IP\n(по подразбиране: -1 = изключено, 0 = webconfig, стандартният порт е 554)",
"SD: standard definition", "SD: стандартна резолюция",
"Sat", "Съб",
"Satconf", "Сателитна конфигурация",
"Satellite config", "Сателитна конфигурация",
"Scan progress", "Прогрес на сканитане",
"Scan status", "Статус на сканиране",
"Schedule status", "Състояние на планиране",
"Schedule status.", "Състояние на планиране",
"Scheduled Duration", "Планирана продължителност",
"Scheduled for recording", "Планирано за запис",
"Scheduled start time", "Планирано начално време",
"Scheduled stop time", "Планирано време за край",
"Science", "Наука",
"Science fiction", "Научна фантастика",
"Season", "Сезон",
"Season %d", "Сезон %d",
"Send full PLAY cmd", "Изпрати full Play команда",
"Serious", "Сериозни",
"Serious music", "Сериозна музика",
"Server", "Сървър",
"Service", "Услуги",
"Service ID", "ИД на услугата",
"Services", "Услуги",
"Services associated with the channel.", "Услуги, свързани с канала",
"Settings", "Настройки",
"Show", "Покажи",
"Show this page", "Покажи тази страница",
"Show version information", "Покажи информация за версията",
"Signal quality too poor", "Качеството на сигнала е много лишо",
"Signal scale (240 or 100)", "Скала на сигнала (240 или 100)",
"Size", "Размер",
"Soap", "Сапунки",
"Soccer", "Футбол",
"Social", "Социални",
"Social advisory", "Социални съвети",
"Source", "Източник",
"Source deleted", "Изтечника е изтрит",
"Source quality is bad", "Качеството на източника е лошо",
"Source reconfigured", "Източника е реконфигуриран",
"Special events (Olympic Games, World Cup, etc.)", "Специални събития(Оломпиади, Световни купи и т.н.",
"Specify User-Agent header for the http client", "Определяне на User-Agent заглавие за http клиент",
"Specify alternative htsp port", "Определяне на алтернативен https порт",
"Specify alternative http port", "Определяне на алтернативен http порт",
"Specify alternative http webroot", "Определяне на алтернативен http руут",
"Specify bind address", "Определяне на свързан адрес",
"Specify extra htsp port", "Определяне на extra htsp порт",
"Spiritual sciences", "Духовна наука",
"Sports", "Спорт",
"Sports magazines", "Спопртни списания",
"Squash", "Скуош",
"Standard", "Стандарт",
"Start", "Старт",
"Start after", "Старт след",
"Start before", "Старт преди",
"Start time", "Време за старт",
"Status", "Статус",
"Stereo", "Стерео",
"Stop", "Спри",
"Stop time", "Време за спиране",
"Storage space", "Място за съхранение",
"Stream type", "Тип на стриима",
"Streaming", "Стрииминг",
"Subscribe to a service permanently", "Абониране за услугата за постоянно",
"Subscription overridden", "Презаписване на абонамента",
"Subtitle", "Резюме",
"Sun", "Неделя",
"Switch: %s", "Ключ:  %s",
"Sync", "Синхронизиране",
"Sync + Don't keep", "Синхронизиране + Без запазване",
"System", "Система",
"Tags", "Етикети",
"Talk show", "Интервюта",
"Team sports (excluding football)", "Отборни спортове (без футбол)",
"Technology", "Технологии",
"Tennis", "Тенис",
"Testing options", "Тестови опции",
"Thriller", "Трилър",
"Thu", "Четвъртък",
"Time", "Време",
"Time missed", "Липсващо време",
"Time recording%s%s", "Време на записа%s%s",
"Timeout (sec) (0=infinite)", "Време на изчакване (секунди) (0=неограничено)",
"Timeout (seconds)", "Време на изчакване (секунди)",
"Timeshift", "Отместване във времето",
"Title", "Заглавие",
"Toggle details", "Превключване на подробностите",
"Tourism / Travel", "Туризъм/Пътуване",
"Traditional arts", "Традиционни изкуства",
"Traditional music", "Традиционна музика",
"Transcode", "Прекодиране",
"Transcode/av-lib", "Прекодиране/av-lib",
"Transcoding", "Прекодиране",
"Transponder ID", "Транспондер ID",
"Tue", "Вторник",
"Tuner", "Тунер",
"Tuner bind IP address", "Свързан с тунера IP адрес",
"Tuning failed", "Нстройката беше неуспешна",
"Turn off LNB when idle", "Изключване на конвертора при бездействане",
"Tvheadend server name", "Име на сървъра",
"Type", "Вид",
"URL with the SAT>IP server XML location", "URL със SAT>IP сървър XML локация",
"Unable to open file", "Файлът не може да бъде отворен",
"Uncommitted", "Необвързани",
"Unimportant", "Маловажни",
"Universal LNB only", "Само универсален конвертор",
"Unknown", "Непознат",
"Unknown reason (%i)", "Неизвестна причина (%i)",
"Unset (default)", "Ненастроено(по подразбиране)",
"Update channel icon", "Обновяване на иконката на канала",
"Update channel name", "Обновяване на името на канала",
"Update channel number", "Обновяване на номера на канала",
"Update disabled", "Обновяването е изключено",
"Update time", "Обновяване на часа",
"Updated", "Обновено",
"Usage: %s [OPTIONS]\n", "Натоварване: %s [OPTIONS]\n",
"Use Windows-compatible filenames", "Използвай на файлови имена съвместими с Уиндоус",
"Use XSPF playlist instead of M3U", "Използване на списък във формат XSPF вместо M3U",
"Use original", "Използване на оригинала",
"User access error", "Грешка в достъпа на потребителя",
"User does not have access", "Потребителят няма достъп",
"User does not have access to channel", "Потребителят няма достъп до канала",
"User icon", "Иконка на потребителя",
"User language", "Потребителски интерфейс",
"User limit reached", "Лимита на юзери е достигнат",
"User login", "Вписване като потребител",
"Username", "Име на потребител",
"Variety show", "Вариететно шоу",
"Video", "Видео",
"Video bitrate (kb/s) (0=auto)", "Побитова скорост на видеото (кб/сек) (0=автоматично) ",
"Video codec to use for the transcode. \"Do not use\" will disable video output.", "Видео кодек, който да се използва при прекодиране. „Без“ ще изключи извеждането на видео.",
"Video recorder", "Видео рекордер",
"View all", "Виж всички",
"Visually impaired commentary/audio description", "За хора с увредено зрение/аудио описания",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV профил H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV профил H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV профил VP8/Vorbis/WEBM",
"Waiting for program start", "Изчакване за започване на програмата",
"Waiting for stream", "Изчакване на стриим",
"War", "Война",
"Water sport", "Воден спорт",
"Weather report", "Прогнози за времето",
"Web interface", "Уеб интерфейс",
"Web interface language", "Език на уеб интерфейса",
"Web interface language.", "Език на уеб интерфейса.",
"Web interface theme.", "Тема на уеб интерфейса.",
"Web user interface", "Потребителски интерфейс",
"Wed", "Сряда",
"Western", "Уестърн",
"Wizard", "Помощник",
"Yes", "Да",
"addpids/delpids supported", "Потдръжка на изтриване/добавяне на пидове",
"hrs", "часа",
"invalid option specified [%s]", "посочена е неправилна стойност [%s]",
"min", "мин",
"mins", "минути",
"option %s requires a value", "опцията %s изисква стойност",
"tsfile input (mux file)", "вход от tsfile (файл mux)",
NULL, NULL
};

static const char *tvh_locale_6[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nPro získání více informací navštivte webovou stránku programu Tvheadend:\nhttps://tvheadend.org\n",
"# Mapped channels", "# Mapovaných programů",
"# Muxes", "# Muxů",
"# Services", "# Služeb",
"# tune repeats", "Opakování ladění",
"(Default profile)", "(Výchozí profil)",
"(Read-only)", "(Pouze pro čtení)",
"1 day", "1 den",
"1 month", "1 měsíc",
"1 week", "1 týden",
"1 year", "1 rok",
"1316 bytes", "1316 bytů",
"16356 bytes", "16356 bytů",
"2 months", "2 měsíce",
"2 weeks", "2 týdny",
"2 years", "2 roky",
"2-Port switch (universal LNB)", "2-portový přepínač (Univerzální LNB)",
"2632 bytes", "2632 bytů",
"3 days", "3 dny",
"3 months", "3 měsíce",
"3 weeks", "3 týdny",
"3 years", "3 roky",
"32712 bytes", "32712 bytů",
"4-Port switch (universal LNB)", "4-portový přepínač (Univerzální LNB)",
"5 days", "5 dnů",
"5264 bytes", "5264 bytů",
"6 months", "6 měsíců",
"65424 bytes", "65424 bytů",
"7896 bytes", "7896 bytů",
"A channel tag (e.g. a group of channels) to which this rule applies.", "Štítek programu (t.j. skupina programů) pro které bude aplikováno toto pravidlo.",
"AC-3 = descriptor 6", "AC-3 = deskriptor 6",
"AC-3 detection", "AC-3 detekce",
"ACTIVE", "AKTIVNÍ",
"ANY", "JAKÝKOLIV",
"ATSC SI Tables", "ATSC SI Tabulky",
"ATSC source ID", "ATSC zdrojové ID",
"ATSC-C Network", "Síť ATSC-C",
"ATSC-T Network", "Síť ATSC-T",
"Aborted by user", "Zrušeno uživatelem",
"Accept remote subscription weight", "Akceptovat vzdálenou váhu odběru",
"Accept the remote subscription weight (from the SAT>IP client).", "Akceptovat vzdálenou váhu odběru (od SAT>IP klienta)",
"Accept zero value for TSID", "Povolit nulové hodnoty pro TSID",
"Access", "Přístup",
"Access (ACL)", "Přístup (ACL)",
"Action", "Akce",
"Active", "Aktivní",
"Adapter in use by another subscription", "Adaptéry používané jiným odběratelem",
"Add file and line numbers to debug", "Přidá název souboru a řádek do ladícího výstupu",
"Add the thread ID to debug", "Přidá identifikaci vlákna do ladícího výstupu",
"Additional arguments to pass to the grabber.", "Dodatečné argumenty pro příkaz grabberu.",
"Additional service names found in EPG data.", "Dodatečné názvy služeb nalezené v EPG datech.",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "Dodatečný čas (v sekundách) nutný pro start tuneru pro vytvoření záznamu. Toto může být užitečné pro tunery, které potřebují více času pro aktivaci nebo posílají na začátku chybná data.",
"Admin", "Administrátor",
"Admin password", "Heslo administrátora",
"Admin username", "Uživatelské jméno administrátora",
"Administrator login", "Uživatelské jméno administrátora",
"Adult movie", "Film pro dospělé",
"Advanced", "Rozšířené",
"Advanced (non-universal LNBs, rotors, etc.)", "Pokročílé (neuniverzální LNB, rotory, apod.)",
"Adventure", "Dobrodružství",
"Advertisement / Shopping", "Reklama / Nakupování",
"After creating the network scan it for services.", "Po vytvoření sítě vyhledej služby.",
"All (Streaming plus DVR)", "Vše (streamování plus DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Všechny znakz, které mohou být problematické pro název souboru, budou nahrazeny podtržítky. Viz. nápověda pro více detailů.",
"All lower-case", "Vše malá písmena",
"All proceeds are used to support server infrastructure and buy test equipment.", "Všechny příspěvky jsou použity pro serverovou infrastrukturu a pro nákup testovacího vybavení.",
"All: Record if different description", "Vše: Zaznamenat pokud je jiný popis",
"All: Record if different episode number", "Vše: Zaznamenat pokud je jiné číslo epizody",
"All: Record if different subtitle", "Vše: Zaznamenat pokud je jiný podtitulek",
"All: Record once per day", "Vše: Zaznamenat jednou denně",
"All: Record once per month", "Vše: Zaznamenat jednou měsíčně",
"All: Record once per week", "Vše: Zaznamenat jednou týdně",
"Allow all PIDs", "Povolit všechny PID",
"Allow all PIDs (no filter) when the 'Maximum PIDs' limit is reached.", "Povolit všechny PID (bez filtru) když je dosažen 'Maximální limit PID'.",
"Allow high bitrate mode (CI+ CAMs only).", "Povolit vyšší datovou propustnost (pouze CI+ CAMs).",
"Allow idle scan tuning on this device.", "Povolit skenování pokud zařízení není využíváno.",
"Allow the combined size of all timeshift buffers to potentially grow unbounded until your storage media runs out of space.", "Povolit kombinovanou velikost všech vyrovnávacích pamětí časového posunu růst potenciálně bez omezení, dokud se nevyčerpá diskový prostor.",
"Allow the initial scan tuning on this device (scan when Tvheadend starts or when a new multiplex is added automatically). At least one tuner or input should have this settings turned on. See also 'Skip Startup Scan' in the network settings for further details.", "Povolit počáteční ladění na tomto zařízení (ladit po spuštění Tvheadend, nebo když je přidán nový multiplex). Minimálně jeden tuner, nebo vstup, by měl mít tuto volbu povolenu. Podívejte se na 'Přeskočit ladění po spuštění' v záložce nastavení sítě pro více informací.",
"Allow the timeshift buffer to grow unbounded until your storage media runs out of space. Warning, enabling this option may cause your system to slow down or crash completely!", "Povolit vyrovnávací pamětí časového posunu růst bez omezení, dokud se nevyčerpá diskový prostor. Pozor, tato volby může způsobit zpomalení, nebo celkový pád systému!",
"Allow/Disallow access to the 'Configuration' tab.", "Povolit/zakázat přístup k záložce 'Nastavení'.",
"Allow/Disallow web interface access (this  includes access to the EPG).", "Povolit/zakázat přístup k webovému rozhranní (včetně EPG).",
"Allowed DVR profiles. This limits the profiles the user has access to.", "Povolené profily DVR. Toto omezuje profily, ke kterým má užiatel přístup.",
"Allowed network", "Povolené síťě",
"Allowed networks", "Povolené síťě",
"Alternate PID path", "Alternativní cesta k PID",
"Alternate configuration path", "Alternativní cesta ke konfiguraci",
"Altitude (in meters).", "Nadmořská výška (metry)",
"Altitude (meters)", "Výška (metry)",
"Always send the whole DiseqC sequence including LNB setup (voltage, tone). If this is not checked, only changed settings are sent, which may cause issues with some drivers. If the tuning is not reliable, try activating this option.", "Vždy posílat celou sekvenci DiseqC včetně nastavení LNB (napětí, tón). Není-li zaškrtnuto, jsou odeslána pouze změněná nastavení a toto může způsobit problémy s některými ovladači. Pokud ladění není spolehlivé, zkuste aktivovat tuto volbu.",
"Always use the name defined by the network.", "Vždy použít jméno nastavené sítí.",
"Always use the old ioctls to read the linuxdvb status (signal strength, SNR, error counters). Some drivers are not mature enough to provide the correct values using the new v5 linuxdvb API.", "Vždy použijte staré ioctl pro čtení linuxdvb stavu (síla signálu, SNR, počítadla chyb). Některé ovladače nejsou natolik vyspělé, aby poskytovaly tyto informace s použitím nového linuxdvb API v5.",
"An event which starts between this \"start after\" and \"start before\" will be matched (including boundary values).", "Událost, která začíná mezi těmito \"Začátek po\" a \"Začátek před\" (včetně).",
"Animals", "Zvířata",
"Anonymize", "Anonymizovat",
"Anonymize HTSP access", "Anonymizace přístupu přes HTSP",
"Any", "Jakýkoliv",
"Arts", "Umění",
"Arts / Culture (without music)", "Umění / Kultura (bez hudby)",
"Arts magazines", "Umělecé magazíny",
"Assign predefined muxes to networks", "Přiřadit předdefinované muxy k sítím.",
"Associate this device with one or more networks.", "Přidružit toto zařízení k jedné nebo více sítím.",
"Athletics", "Atletika",
"Audio Stream Filter", "Filtr audio streamu",
"Audio bitrate (kb/s) (0=auto)", "Audio bitrate (kb/s) (0=Auto)",
"Audio bitrate to use for transcoding.", "Nastavení použité rychlosti pro audio pro transkódování.",
"Audio channel layout.", "Konfigurace audio kanálů.",
"Audio codec", "Audio kodek",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Audio kodek použitý pro transkódování. \"Nepoužívat\" zakáže zvukový výstup.",
"Audio type", "Typ audia",
"Audio-only stream", "Pouze audio proud",
"Authenticated user", "Přihlášený uživatel",
"Auto", "Automatické",
"Auto check disabled", "Zakázána automatická kontrola",
"Auto check enabled", "Povolena automatická kontrola",
"Auto record", "Automatický rekordér",
"Auto record caption", "Titulek pro automatický rekordér",
"Auto recording%s%s", "Automatický rekordér%s%s",
"Auto time record", "Automatický časový rekordér",
"Auto-Map to channels", "Automatické mapování do programů",
"Automatic checking", "Automatická kontrola",
"Automatic recording caption.", "Titulek pro automatický rekordér.",
"Automatic schedule entry not found", "Automatický záznam nebyl nalezen",
"Automatic time scheduler entry not found", "Automatický záznam časovače nebyl nalezen",
"Automatically link EPG data to the channel (using the channel name for matching). If you turn this option off, only the OTA EPG grabber will be used for this channel unless you've specifically set a different EPG Source.", "Automaticky propojit data EPG s kanálem (pomocí shodného názvu programu). Pokud tuto volbu vypnete, použije se pro tento program pouze OTA EPG grabber, pokud jste nenastavili jiný specifický zdroj EPG.",
"Automatically map EPG source", "Automatické mapování EPG zdroje",
"Automatically map all available services to channels.", "Automaticky namapovat všechny dostupné služby na programy.",
"Automatically name from network", "Automatický název ze sítě",
"Automatically record.", "Automatický záznam.",
"Automatically update channel icons using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel icons.", "Automaticky aktualizovat ikony programů pomocí informací poskytovaných poskytovateli služeb EPG. Toto může způsobit nežádoucí změny již nastavených ikon programů.",
"Automatically update channel names using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel names.", "Automaticky aktualizovat názvy programů pomocí informací poskytnutých poskytovateli služeb EPG. Toto může způsobit nežádoucí změny již nastavených názvů programů.",
"Automatically update channel numbers using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel numbers.", "Automaticky aktualizujte čísla programů pomocí informací poskytnutých poskytovateli služeb EPG. Toto může způsobit nežádoucí změny již nastavených čísel programů.",
"Autorec maximum count (0=unlimited)", "Autorekordér - maximální počet (0=neomezeno)",
"Autorec maximum schedules limit (0=unlimited)", "Autorekordér - maximálni počet časovačů (0=neomezeno)",
"Bad", "Špatný",
"Bad request", "Chybný požadavek",
"Ballet", "Balet",
"Bandwidth", "Šířka pásma",
"Based on software from", "Použitý software od",
"Basic", "Základní",
"Bind to specific local IP address.", "Naslouchej na specifikované lokální IP adrese.",
"Bitrate to use for the transcode. See Help for details.", "Bitrate pro transkódování. Přečtěte si nápovědu do více informací.",
"Blue", "Modrý",
"Bouquet", "Buket",
"Bouquet (auto)", "Buket (automatický)",
"Bouquet source.", "Zdroj pro buket.",
"Brand", "Značka",
"Branding information (if available).", "Informace o značení (pokud je dostupné).",
"Broadcast", "Vysílání",
"Broadcast type", "Typ vysílání",
"Broadcast.", "Vysílání.",
"Broadcasting", "Vysílání",
"Buffering limit (ms)", "Limit pro vyrovnávací paměť (ms)",
"Build", "Sestavení",
"By default, linuxdvb's status read period is 1000ms (one second). The accepted range is 250ms to 8000ms. Note that for some hardware or drivers (like USB), the status operations take too much time and CPU. In this case, increase the default value. For fast hardware, this value might be decreased to make the decision of the re-tune algorithm based on the signal status faster.", "Ve výchozím nastavení je doba čtení stavu linuxdvb 1000 ms (jedna vteřina). Podporovaný rozsah je 250ms až 8000ms. Mějte prosím na mysli, že některého hardwaru, nebo ovladačů (například USB), zabírají operace příliš mnoho času a CPU. V takovém případě zvětšete výchozí hodnotu. V případě rychlého hardwaru může být tato hodnota snížena, aby bylo možné rychlejší použít rychlejší přeladění signálu na základě rychlejšího signálu.",
"CA Stream Filter", "Filtr CA streamu",
"CA client %i", "CA klient %i",
"CA identification", "CA identifikace",
"CA provider", "CA poskytovatel",
"CAPMT CA Client", "CAPMT CA Klient",
"CAPMT interval (in ms).", "CAPMT interval (v ms).",
"CAPMT interval (ms)", "CAPMT Interval (ms)",
"CAPMT query interval (ms)", "CAPMT interval dotazů (ms)",
"CAPMT query interval (ms).", "CAPMT interval dotazu (ms).",
"CI Module", "CI Modul",
"CRID authority", "CRID autorita",
"Cache scheme", "Schéma mezipaměti",
"Camd.socket filename / IP Address (TCP mode)", "Camd.socket soubor / IP adresa (TCP mód)",
"Cartoons", "Animované pohádky",
"Change parameters", "Změnit parametry",
"Channel", "Program",
"Channel Tags", "Štítky programu",
"Channel does not exist", "Kanál neexistuje",
"Channel icon", "Ikona programu",
"Channel icon URL.", "URL ikony programu.",
"Channel icon as defined in EPG data.", "Ikona programu definovaná v EPG datech.",
"Channel icon path", "Cesta k ikonám programů",
"Channel mapping options", "Volby pro mapování programů",
"Channel name", "Název programu",
"Channel not found", "Program nebyl nalezen",
"Channel number", "Číslo programu",
"Channel number as defined in EPG data.", "Číslo programu definované v EPG datech.",
"Channel number offset", "Přidej k číslu programu",
"Channel numbers from", "Počáteční číslo programu",
"Channel tag", "Štítek programu",
"Channel tag does not exist", "Štítek programu neexistuje",
"Channel tag reference", "Reference štítku programu",
"Channel tags", "Štítky programů",
"Channels", "Programy",
"Channels EPG data is used by.", "Kanály využívající EPG data.",
"Character set", "Znaková sada",
"Characters not supported in Windows filenames (e.g. for an SMB/CIFS share) will be stripped out or converted.", "Znaky nepodporované v názvech souborů pro Windows (např. SMB/CIFS sdílená složka) budou přeskočeny nebo konvertovány.",
"Charset", "Kódová stránka",
"Check availability", "Kontrola dostupnosti",
"Check services for availability. If enabled, services that are not currently broadcasting (or can't be decrypted) will be ignored. Leave disabled if you want Tvheadend to also map offline services.", "Zkontrolovat dostupnost služby. Pokud je povoleno, služby, které yrovna nevysílají (nebo nemohou být rozkódovány) budou ignorovány. Ponechte zakázané. pokud chcete, aby Tvheadend namapoval i služby, které jsou offline.",
"Children's / Youth Programs", "Dětské / programy pro mladé",
"Children's / Youth programs", "Dětské / programy pro mladé",
"Cinema", "Kino",
"Class", "Třída",
"Classical", "Klasické",
"Classical music", "Klasická hudba",
"Clean effects", "Čisté efekty",
"Client name", "Název klienta",
"Clone scheduled entry on error", "Klonovat naplánovaný záznam při chybě",
"Comedy", "Komedie",
"Command delay time (ms) (10-200)", "Zpoždění příkazu (ms) (10-200)",
"Command time (ms) (10-100)", "Čas pro příkaz (ms) (10-100)",
"Comment", "Komentář",
"Commercial break", "Komerční přestávka",
"Commercials will be dropped from the recordings. Commercial detection works using EITp/f (EPG running state) and for the Swedish channel TV4 (using teletext info).", "Reklamy budou v záznamu vynechány.Detekce reklam funguje s EITp/f (běžící stav EPG) a pro švédský kanál TV4 (s použitím teletextových informací).",
"Completed OK", "Dokončeno v pořádku",
"Compress EPG database", "Komprimovat EPG databázi",
"Compress the EPG database to reduce disk I/O and space.", "Komprimovat EPG databázi pro šetření místa na disku a I/O.",
"Configuration", "Konfigurace",
"Configuration - Base", "Konfigurace - Základní",
"Configuration ID", "ID konfigurace",
"Configuration name", "Název konfigurace",
"Configuration version", "Verze konfigurace",
"Connection limit type", "Typ limitu pro spojení",
"Constellation", "Konstelace",
"Container", "Kontejner",
"Container to use for the transcoded stream.", "Kontejner pro transkódování sttreamu.",
"Content character set", "Znaková sada pro obsah",
"Content type", "Typ obsahu",
"Content type.", "Typ obsahu.",
"Contest", "Soutěž",
"Continue recording for x minutes after scheduled stop time", "Pokračovat v záznamu dalších x minut po naplánovaném konci",
"Continue recording for x minutes after scheduled stop time.", "Pokračovat v záznamu dalších x minut po naplánovaném konci.",
"Cookie expiration (days)", "Expirace cookie (dny)",
"Cooking", "Vaření",
"Copy codec type", "Kopíruj typ kodeku",
"Copy layout", "Kopíruj dispozici",
"Count of objects", "Počet objektů",
"Create SDTV/HDTV/Radio tags.", "Vytvořit SDTV/HDTV/Radio štítky.",
"Create a directory per channel when storing recordings. If both this and the 'directory per day' checkbox is enabled, the date-directory will be the parent of the per-channel directory.", "Při ukládání záznamu vytvoří adresář pro kanál. Pokud je zaškrtnuto toto i 'adresář pro den', bude adresář datumu nadřazený adresáři pro kanál.",
"Create a directory per title when storing recordings. If the day/channel directory checkboxes are also enabled, those directories will be parents of this directory.", "Při ukládání záznamu vytvoří adresář pro pořad. Pokud je zaškrtnuto toto i adresář pro den/kanál, tyto adresáře budou nadřazené tomuto adresáři.",
"Create a new directory per day in the recording system path. Folders will only be created when something is recorded. The format of the directory will be ISO standard YYYY-MM-DD.", "Vytvořit každý den nový adresář v systémové cestě pro nahrávání.  Adresář bude vytvořem. pouze pokud bude něco nahráno. Formát jměna adresáře je podle ISO standardu YYYY-MM-DD.",
"Create a provider name tag.", "Vytvořit štítek poskytovatele.",
"Create and associate a network tag to created channels.", "Vytvořit a přiřadit štítky sítě pro vytvořené kanály.",
"Create and associate a provider tag to created channels.", "Vytvořit a přiřadit štítky poskytovatele pro vytvořené kanály.",
"Create and link these tags to channels when mapping.", "Založit a přiřadit tyto štítky kanálům při mapování.",
"Create bouquet", "Vytvořit buket",
"Create bouquet tag", "Vytvořit štítek buketu",
"Create directories using these permissions.", "Vytvoří adresáře s použitím těchto práv.",
"Create files using these permissions.", "Vytvoří soubory s použitím těchto práv.",
"Create network name tags", "Vytvořit štítky pro názvy sítí",
"Create network name tags (set by provider).", "Vytvořit štítky pro názvy sítí (od poskytovatele obsahu).",
"Create network tags", "Vytvořit štítky pro sítě",
"Create provider name tags", "Vytvořit štítek poskytovatele",
"Create provider tags", "Vytvořit štítek poskytovatele",
"Create tags", "Vytvořit štítky",
"Created", "Vytvořeno",
"Creator", "Vytvořil",
"Cron multi-line", "Víceřádkový cron",
"Culture (without music)", "Kultura (bez hudby)",
"Culture magazines", "Kulturní magazíny",
"Current RTSP port.", "Současný RTSP port.",
"Current affairs", "Aktuální události",
"Current object size.", "Současná velikost objektu.",
"Custom HTTP headers", "Zvláštní HTTP hlavičky",
"DAB Network", "DAB Síť",
"DBus - use the session message bus instead of the system one", "DBUs - použije se 'session' sběrnice pro zprávy místo systémové",
"DES key", "Klíč DES",
"DSCP/TOS for streaming", "DSCP/TOS pro streamování",
"DVB Inputs - Networks", "DVB vstupy - Síťě",
"DVB scan files", "DVB scan soubory",
"DVB scan files path", "Cesta k souborům pro DVB scan",
"DVB-C Network", "DVB-C Síť",
"DVB-S Network", "DVB-S síť",
"DVB-T Network", "DVB-T síť",
"DVR - Auto-recording (Autorecs)", "DVR - Auto-nahrávání (Autorekordér)",
"DVR - Profiles", "Profily - DVR",
"DVR - Time-based Recording (Timers)", "DVR - Časem řízené nahrávání (Časovače)",
"DVR behavior", "Chování DVR",
"DVR configuration", "Konfigurace DVR",
"DVR configuration profiles", "Konfigurační profily DVR",
"DVR configurations", "DVR nastavení",
"DVR entry not found", "Položka DVR nebyla nalezena",
"DVR file retention period", "Čas vymazání DVR souborů",
"DVR log retention", "Čas držení DVR logů",
"DVR log retention period", "Čas vymazání DVR souborů",
"DVR override: high", "Vynucení vysoké DVR priority",
"DVR override: important", "Vynucení důležité DVR priority",
"DVR override: low", "Vynucení nízké DVR priority",
"DVR override: normal", "Vynucení standardní DVR priority",
"DVR override: unimportant", "Vynucení nedůležité DVR priority",
"DVR profile to use/used for the recording.", "DVR profil pro nahrávání.",
"DVR schedule not found", "Záznam nebyl nelezen",
"Dance", "Tanec",
"Data errors", "Chyby v datech",
"Data path (if applicable).", "Datová cesta (pokud je nutná).",
"Days of Week", "Dny v týdnu",
"Days of the week to which the rule should apply.", "Dny v týdnu, pro které bude pravidlo použito.",
"Days of week", "Dny v týdnu",
"Debate", "Debata",
"Debug libav log", "Ladící výstup z libav",
"Debug options", "Ladící možnosti",
"Debug subsystems", "Povolit ladící podsystémy",
"Debug to syslog", "Povolit ladící výstop pro syslog",
"Debug trace (low-level)", "Povol trasování (velmi podrobné)",
"Debugging", "Ladění",
"Default", "Standardní",
"Default language(s)", "Výchozí jazyk(y)",
"Default language.", "Výchozí jazyk.",
"Default login", "Defaultní přihlášení",
"Default priority", "Výchozí priorita",
"Default user interface level.", "Výchozí úroveň uživatelského rozhraní.",
"Define network group to limit network usage.", "Definovat síťovou skupinu pro limitování využití sítě.",
"Delivery system", "Systém doručení",
"Demux path", "Cesta k demuxu",
"Descramble services (limit per mux)", "Dekóduj služby (limit pro mux)",
"Descrambler buffer (TS packets)", "Vyrovnávací paměť dekódéru (TS pakety)",
"Description", "Popis",
"Detective", "Detektivka",
"Device ID", "ID zařízení",
"Device model", "Model zařízení",
"Device path", "Cesta k zařízení",
"Device path in sysfs", "Cesta k zařízení v sysfs",
"Digital Video Recorder", "Digitální Video Rekordér",
"Directory", "Adresář",
"Directory permissions (octal, e.g. 0775)", "Práva k adresáři (osmičkové, např. 0775)",
"Directory used by the entry.", "Adresář použitý touto položkou.",
"Disable", "Zakázat",
"Disable DVB bouquets", "Zakázat DVB bukety",
"Disable SAT>IP client", "Zakázat SAT>IP klienta",
"Disable all access control checks", "Zakázat veškeré přístupové kontroly",
"Disable debug on stderr", "Povolit ladící pro syslog",
"Disable device/firmware-specific workarounds", "Zakaž specifické workaroundy pro zařízení/firmware",
"Disable syslog (all messages)", "Zakázat veškerý výstup pro syslog",
"Disabled", "Zakázáno",
"Discussion", "Diskuze",
"DiseqC repeats", "DiseqC opakování",
"Do not set", "Nenastavovat",
"Do not use", "Nepoužívat",
"Documentary", "Dokumentární",
"Don't backup configuration tree at upgrade", "Nezálohovat konfigurační strom při aktualizaci",
"Don't include the title in the filename.", "Nevkládat název vysílání do názvu souboru",
"Don't include title in filename", "Nevkládat název vysílání do názvu souboru",
"Don't keep", "Nezanechávej",
"Don't re-record", "Nezaznamenávej znovu",
"Don't re-record if recording fails.", "Nezaznamenávej znovu, pokud záznam selže.",
"Don't re-schedule if recording fails.", "Nepřeplánuj další záznam, pokud záznam selže.",
"Don't reschedule", "Nepřeplánovávat",
"Don't use the provider's channel numbers.", "Nepoužívat čislo kanálu poskytovatele.",
"Duplicate handling", "Zacházení s duplicitami",
"Duplicate recording handling.", "Zacházení s duplicitami záznamu.",
"EIT - skip TSID check", "EIT - nekotrolovat TSID",
"EIT time offset", "EIT časový posun",
"EPG Database", "EPG databáze",
"EPG Grabber", "EPG graber",
"EPG Grabber Channel", "Program EPG graberu",
"EPG Grabber Configuration", "Konfigurace EPG graberu",
"EPG Language (priority order)", "Jazyk EPG (dle priority)",
"EPG data ID.", "Identifikátor EPG dat.",
"EPG name", "Název EPG",
"EPG source", "Zdroj EPG",
"EPG update window", "EPG časové okno aktualizace",
"Economics", "Ekonomika",
"Education", "Vzdělání",
"Education / Science / Factual topics", "Vzdělání / Věda / Fakta",
"Educational", "Vzdělávací",
"Electronic Program Guide", "Elektronický Programový Průvodce",
"Elementary stream filter", "Filtr elementárního streamu",
"Enable (auto)", "Povolit (auto)",
"Enable DBus", "Povolit DBus",
"Enable NTP driver", "Povolit NTP ovladač",
"Enable coredumps for daemon", "Povolit coredump pro démona",
"Enable debug on stderr", "Povolit ladící výstup na stderr",
"Enable debug subsystems", "Povolit ladící podsystémy",
"Enable debug to file", "Povolit ladící výstup do souboru",
"Enable debug to syslog", "Povolit ladící výstop pro syslog",
"Enable or disable fast input switching.", "Povolit nebo zakázat rychlé přepínání vstupů.",
"Enable or disable full mux mode.", "Povolit nebo zakázat příjem celého muxu.",
"Enable or disable this configuration.", "Povolit nebo zakázat tuto konfiguraci.",
"Enable syslog", "Povolit syslog",
"Enable the CAs (conditional accesses) tab in web user interface for the advanced level. By default, this tab is visible only in the expert level.", "Povolit záložku CAs (podmíněný přístup) ve webovém uživatelské rozhraní pro úroveň rozsířený. Standartně je tato záložka viditelná pouze v úrovni expert.",
"Enable this filter.", "Povolit tento filtr.",
"Enable trace subsystems", "Povolit trasovací subsystémy",
"Enable web UI debug (non-minified JS)", "Povolit ladění webUI (neminimalizované JS)",
"Enable/Disable timeshift.", "Povolit/zakázat časový posun.",
"Enable/disable LNA.", "Povolit/zakázat LNA.",
"Enable/disable auto-rec rule.", "Povolit/zakázat pravidlo autorekordéru.",
"Enable/disable pilot tone.", "Povolit/zakázat pilot.",
"Enable/disable profile.", "Povolit/zakázat profil.",
"Enable/disable the channel.", "Povolit/zakázat kanál.",
"Enable/disable the device.", "Povolit/zakázat zařízení.",
"Enable/disable the tag.", "Povolit/zakázat štítek.",
"Enable/disable tuner/adapter.", "Povolit/zakázat tuner/adaptér.",
"Enabled", "Povoleno",
"Encrypted", "Zakódováno",
"Encrypted service", "Zakódovaná služba",
"Enter an administrator password.", "Vložit heslo administrátora.",
"Enter an administrator username. Note: do not use the same username as the superuser backdoor account.", "Vložte uživatelské jméno administrátora. Pozn: nepoužívejte stejné uživatelské jméno jako pro super-administrátora / backdoor.",
"Enter comma-separated list of subsystems you want debugging output for (e.g +linuxdvb,+subscriptions,+mpegts).", "Zadejte čárkou odělený seznam sybsystémů, pro které chcete zobrazit ladicí záznam (např. +linuxdvb,+subscriptions,+mpegts).",
"Enter the filename (including path) where Tvheadend should write the log.", "Zadejte název souboru (včetně cesty), kam má Tvheadend ukládat logy.",
"Entertainment programs for 10 to 16", "Zábavné programy od 10 do 16",
"Entertainment programs for 6 to 14", "Zábavné programy od 6 do 14",
"Environment", "Životní prostředí",
"Environment (pipe)", "Prostředí (roura)",
"Episode", "Epizoda",
"Episode %d", "Epizoda %d",
"Episode number/ID.", "Číslo epizody/identifikátor",
"Equestrian", "Jezdectví",
"Error code", "Kód chyby",
"Error code of entry.", "Kód chyby pro položku.",
"Errors", "Chyby",
"Even key", "Sudý klíč",
"Event does not exist", "Událost neexistuje",
"Every day", "Každý den",
"Exclude channel tags", "Vyjmout štítky programů",
"Expeditions", "Expedice",
"Experimental film", "Experimentální film",
"Exported tuners", "Exportované tunery",
"External", "Externí",
"External EPG grabber", "Externí EPG graber",
"External IP (NAT)", "Externí IP (NAT)",
"External URL", "Externí URL",
"External URL of the bouquet.", "Externí URL pro buket.",
"Extra arguments", "Extra argumenty",
"Extra start time", "Extra začáteční čas",
"Extra stop time", "Extra koncový čas ",
"Extra warming up time (seconds)", "Extra zahřívací čas (sekundy)",
"FAIL", "NEÚSPĚŠNÝ",
"Factual topics", "Fakta",
"Failed", "Nezdařené",
"Failed to open image", "Nemohu otevřít soubor",
"Fashion", "Móda",
"Fast input switch", "Rychlé přepnutí vstupu",
"File missing", "Soubor chybí",
"File not created", "Soubor nebyl vytvořen",
"File permissions (octal, e.g. 0664)", "Práva souboru (osmičkové, např. 0664)",
"File size", "Velikost souboru",
"Filename", "Jméno souboru",
"Filename (including path)", "Název souboru (včetně cesty)",
"Filename character set", "Znaková sada jména souboru",
"Filename options", "Volby pro jméno souboru",
"Filename used by the entry.", "Jméno souboru použíté v této položce.",
"Fine arts", "Výtvarné umění",
"Finished", "Ukončený",
"First word", "První slovo",
"Fit to RAM (cut rewind)", "Pouze v RAM (omezit převíjení zpátky)",
"Fitness and health", "Fitnes a zdraví",
"Folk", "Lidový",
"Folkloric", "Folklór",
"Football", "Fotbal",
"Force (auto)", "Vynutit (auto)",
"Force CA ID (e.g. 0x2600)", "Vynutit CA ID (e.g. 0x2600)",
"Force initial EPG grab at start-up", "Vynutit počáteční EPG grabování po startu",
"Force old status", "Vynutit starý status",
"Force pilot for DVB-S2", "Vynutit pilot pro DVB-S2",
"Force priority", "Vynutit prioritu",
"Force service type to 1", "Vynutit typ služby jako 1",
"Force teardown delay", "Vynutit zpoždění teardown",
"Foreign countries", "Cizí země",
"Forever", "Navždy",
"Fork and run as daemon", "Démon mód",
"Format string", "Řetězec formátu",
"Found muxes", "Nalezeno multiplexů",
"Found services", "Nalezeno služeb",
"Free subscription weight", "Volná váha odběru",
"Free-form field, enter whatever you like here.", "Volné pole, zadejte cokoliv, co se vám líbí.",
"Free-form text field, enter whatever you like here.", "Volné textové pole, zadejte cokoliv, co se vám líbí.",
"Frequency", "Frekvence",
"Frequency (Hz)", "Frekvence (Hz)",
"Frequency (MHz)", "Frekvence (MHz)",
"Frequency (in MHz).", "Frekvence (v MHz).",
"Frequency (kHz)", "Frekvence (kHz)",
"Fri", "Pát",
"Friendly name", "Přátelské jméno",
"Frontend number", "Číslo frontendu",
"Frontend path", "Cesta k frontendu",
"Full DiseqC", "Celý DiseqC",
"Full mux RX mode supported", "Podporován plný mux rx mód",
"Full pathname specification", "Plná specifikace cesty",
"Further education", "Další vzdělání",
"GOTOX position", "GOTOX pozice",
"Game show", "Herní show",
"Gardening", "Zahradničení",
"General", "Základní",
"General configuration", "Základní konfigurace",
"Generic", "Obecný",
"Generic options", "Generické možnosti",
"Gray", "Šedivý",
"HD: high definition", "HD: vysoké rozlišení",
"HTSP Default Stream Settings", "HTSP standardní nastavení streamu",
"HTTP Client", "HTTP Klient",
"Handicraft", "Řemeslo",
"Hearing impaired", "Zhoršený sluch",
"Hierarchy", "Hierarchie",
"High", "Vysoká",
"High bitrate mode (CI+ CAMs only)", "High bitrate mód (CI+ CAMs Only)",
"Higher LNB voltage", "Vyšší LNB napětí",
"Highest channel number the user can access.", "Nejvyšší číslo kanálu přístupné pro uživatele.",
"Highest count of objects.", "Nejvyšší počet objektů.",
"Historical movie", "Historický film",
"Horror", "Horor",
"Hostname/IP", "Jméno /IP",
"ID of the network.", "ID sítě.",
"IDLE", "NEČINNÝ",
"IP address", "IP adresa",
"IP blocking", "Blokování IP",
"IPTV Automatic Network", "IPTV automatická síť",
"IPTV Network", "IPTV síť",
"IPTV input", "IPTV vstup",
"IPTV signal level", "IPTV úroveň signálu",
"ISDB-C Network", "ISDB-C síť",
"ISDB-S Network", "ISDB-S síť",
"ISDB-T Network", "ISDB-T síť",
"Icon", "Ikona",
"Icon (full URL)", "Ikona (celé URL)",
"Icon URL", "URL ikony",
"Icon base URL", "Základní URL pro ikony",
"Icon base URL.", "Základní URL pro ikony",
"Icon has title", "Ikona má název",
"Icons from", "Ikony jsou od",
"Idle", "Nečinný",
"Idle scan", "Vyhledávání při nečinnosti",
"Idle scan muxes", "Vyhledávání muxů při nečinnosti",
"If an error occurs clone the scheduled entry and try to record again (if possible).", "Když nastane chyba, zkopíruj naplánovanou položku a pokus se nahrát ji znovu (pokud je to možné).",
"If both a picon and a channel-specific (e.g. channelname.jpg) icon are defined, prefer the picon.", "Pokud jsou nastaveny picon i specifická ikony pro kanál (např. nazevkanalu.jpg), preferuj picon.",
"If more than x errors occur during a recording schedule a re-record (if possible).", "Když nastane během naplánovaného nahrávání více než x chyb, naplánuj nahrávání znovu (pokud je to možné).",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Pokud není vytvořený žádný uživatelský účet,\ntak se vytvoří účet bez uživatelského jména\na hesla. Je doporučené tento administrátorský\núčet okamžitě změnit (uživatelské jméno a heslo)",
"If you'd like to support the project, please consider a donation.", "Pokud se Vám tento projekt líbí, prosíme, zvažte příspěvek.",
"Ignore", "Ignorovat",
"Ignore EPG (EIT)", "Ignorovat EPG (EIT)",
"Ignore descriptor 5", "Igoruj deskritor 5",
"Ignore encryption flag, include encrypted services anyway.", "Ignorovat příznak kódování, vždy zahrnout i kódované programy.",
"Ignore invalid SSL certificate", "Ignorovat neplatné SSL certifikáty",
"Ignore invalid/unverifiable (expired, self-certified, etc.) certificates", "Ignorovat neplatné/neověřitlné (vypršená platnost, samo-podepsané, atd.) certifikáty",
"Ignore provider's channel numbers", "Ignorovat čísla programů od poskytovatele",
"Immediately abort", "Okamžitě ukonči činnost",
"Important", "Důležíté",
"Include channel name in filename", "Vložit název programu do jména souboru",
"Include date in filename", "Vložit datum do jména souboru",
"Include episode in filename", "Vložit epizodu do jména souboru",
"Include subtitle in filename", "Vložit podtitulek do jména souboru",
"Include teletext", "Zahrnout teletext",
"Include the date for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "Vložit datum záznamu do názvu události. Toto platí jak pro název uložený v souboru, tak pro samotný název souboru.",
"Include the episode subtitle in the title (if available).", "Vložit název epizody do názvu (pokud je dostupný).",
"Include the name of the channel in the event title. This applies to both the title stored in the file and to the filename itself.", "Zahrnout název kanálu do jména. Toto platí pro název uložený v souboru i pro jméno souboru.",
"Include the season and episode in the title (if available).", "Vložit sezónu a epizody do názvu (pokud jsou dostupné).",
"Include the time for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "Vložit čas záznamu do názvu. Toto platí jak pro název uložený v souboru, tak pro samotný název souboru.",
"Include time in filename", "Vložit čas do jména souboru",
"Information area", "Informační oblast",
"Informational", "Informační",
"Initial scan", "Počáteční vyhledávání",
"Input base", "Základní vstup",
"Input buffer (bytes)", "Vstupní paměť (byty)",
"Input instance", "Instance vstupu",
"Input path", "Cesta k zařízení",
"Interface", "Interfejs",
"Internal", "Interní",
"Internal EPG grabber", "Interní EPG graber",
"Internal error", "Interní chyba",
"Internal grabber", "Interní graber",
"Invalid", "Chybný",
"Invalid EPG object request", "Neplatný požadavek na EPG objekt",
"Invalid adapter number '%s'\n", "Chybné číslo adaptéru '%s'\n",
"Invalid arguments", "Chybné argumenty",
"Invalid file", "Chybný soubor",
"Invalid parameters", "Chybné parametry",
"Invalid service", "Chybná služba",
"Invalid target", "Chybný cíl",
"Items", "Položky",
"Keep", "Zachovat",
"Keep timeshift buffers in RAM only. With this option enabled, the amount of rewind time is limited by how much RAM TVHeadend is allowed.", "Držet dočasnou vyrovnávací pamět časového posunu pouze v paměti RAM. Pokud je tato volba zapnuta, je doba zpětného převíjení omezena podle toho, kolik RAM TVHeadend je povoleno.",
"Keep x amount of storage space free.", "Ponechat x volného místa na disku.",
"Kill signal (pipe)", "Signál pro kill (roura)",
"Kill timeout (pipe/secs)", "Timeout pro kill (roura/sek)",
"LNB type", "Typ LNB",
"Language", "Jazyk",
"Language 1", "Jazyk 1",
"Language 2", "Jazyk 2",
"Language 3", "Jazyk 3",
"Language settings", "Jazykové nastavení",
"Languages", "Jazyky",
"Last seen", "Naposledy viděn",
"Last updated from", "Naposledy aktualizováno z",
"Layer A: Constellation", "Vrstva A: Konstelace",
"Layer A: FEC", "Vrstva A: FEC",
"Layer A: Segment count", "Vrstva A: Počet segmentů",
"Layer A: Time interleaving", "Vrstva A: Časové prokládání",
"Layer B: Constellation", "Vrstva B: Konstelace",
"Layer B: FEC", "Vrstva B: FEC",
"Layer B: Segment count", "Vrstva B: Počet segmentů",
"Layer B: Time interleaving", "Vrstva B: Časové prokládání",
"Layer C: Constellation", "Vrstva C: Konstelace",
"Layer C: FEC", "Vrstva C: FEC",
"Layer C: Segment count", "Vrstva C: Počet segmentů",
"Layer C: Time interleaving", "Vrstva C: Časové prokládání",
"Leisure hobbies", "Volný čas / koníčky",
"Limit connections", "Limit spojení",
"Limited/limit scanning to this network ID only.", "Limitování prohlížení (skenu) pouze pro toto ID sítě.",
"Linked input", "Spojený vstup",
"LinuxDVB Input", "LinuxDVB vstup",
"LinuxDVB adapter", "LinuxDVB adaptér",
"LinuxDVB network", "LinuxDVB síť",
"List of allowed IPv4 or IPv6 hosts or networks (comma-separated).", "Seznam povolených IPv4 a IPv6 hostů nebo sítí (oddělených čárkou).",
"Listen / Connect port", "Port pro příjem/vysílání",
"Listen on IPv6", "Poslouchej na IPv6",
"Literature", "Literatura",
"Local (server) time", "Lokální čas (na serveru)",
"Local bind IP address", "Lokální IP adresa pro komunikaci",
"Local channel minor", "Lokální číslo programu - vedlejší",
"Local channel number", "Lokální číslo programu",
"Local discovery IP address", "Lokální IP adresa pro vyhledávání",
"Local: Record if different description", "Místní: Zaznamenat pokud je jiný popis",
"Local: Record if different episode number", "Místní: Zaznamenat pokud je jiné číslo epizody",
"Local: Record if different subtitle", "Místní: Zaznamenat pokud je jiný podtitulek",
"Local: Record if different title", "Místní: Zaznamenat pokud je jiný název",
"Local: Record once per day", "Místní: Zaznamenat jednou denně",
"Local: Record once per month", "Místní: Zaznamenat jednou měsíčně",
"Local: Record once per week", "Místní: Zaznamenat jednou týdně",
"Location", "Umístění",
"Lock preferred CA PID", "Pouze preferovaná CA PID",
"Login password.", "Přihlašovací heslo",
"Login username.", "Přihlašovací jméno.",
"Login/Logout", "Přihlášení/Odhlášení",
"Loglevel debug: disabled", "Úrovaň záznamu ladění: zakázán",
"Loglevel debug: enabled", "Úrovaň záznamu ladění: povolen",
"Logout", "Odhlásit",
"Low", "Nízká",
"Lowest channel number the user can access.", "Nejnižsí číslo kanálu přístupné pro uživatele.",
"Lowest starting channel number.", "Nejnižší počáteční číslo kanálu.",
"MPEG-TS File", "MPEG-TS Soubor",
"MPEG-TS Multiplex", "MPEG-TS multiplex",
"MPEG-TS Service", "MPEG-TS služba",
"MPEG-TS input", "MPEG-TS vstup",
"MPEG-TS raw service", "MPEG-TS holá služba",
"Magazines", "Magazíny",
"Maintain free storage space in MiB", "Udržovat na disku volné místo v MiB.",
"Maintain used storage space in MiB (0=disabled)", "Udržovat na disku volné místo v MiB (0=zakázáno)",
"Maintained space", "Udržovaný prostor",
"Make subdirectories per channel", "Vytvořit podadresáře pro každý programu",
"Make subdirectories per day", "Vytvořit podadresáře pro každý den",
"Make subdirectories per title", "Vytvořit podadresáře pro každý název",
"Manage all", "Spravovat vše",
"Manufacturer", "Výrobce",
"Manufacturer URL", "URL výrobce",
"Map all services", "Mapuj všechny služby",
"Map encrypted services", "Mapovat šifrované služby",
"Map radio channels", "Mapuj rádio programy",
"Map unnamed channels", "Mapuj nepojmenované programy",
"Map zero-numbered channels", "Mapování nulových čísel",
"Mapped", "Namapován",
"Martial sports", "Bojové sporty",
"Master tuner", "Hlavní tuner",
"Matroska specific", "Specifické pro matroska formát",
"Maximal channel number", "Maximální číslo programu",
"Maximum # input streams", "Maximální počet vstupních streamů",
"Maximum PIDs", "Max. počet PID",
"Maximum RAM size (MB)", "Maximální velikost RAM (MB)",
"Maximum bandwidth (Kbps)", "Maximální šířka pásma (Kbps)",
"Maximum count (0=default)", "Maximální počet (0=výchozí)",
"Maximum duration", "Maximální doba",
"Maximum length in characters for the command setting PIDs to the SAT>IP box.", "Maximální délka ve znacích pro příkaz nastavující PID pro SAT>IP server.",
"Maximum length of PIDs", "Maximální velikost PID",
"Maximum period (mins)", "Maximální perioda (minuty)",
"Maximum schedules limit (0=default)", "Maximálni počet časovačů (0=výchozí)",
"Maximum size (MB)", "Maximální velikost (MB)",
"Maximum timeout (seconds)", "Maximální čas odezvy (sekund)",
"Medicine", "Lékařství",
"Memory Information", "Informace o paměti",
"Merge same name", "Spoj stejný název",
"Merge services with the same name into one channel.", "Spojit služby se stejným jménem do jednoho kanálu.",
"Method not found", "Metoda nebyla nalezena",
"Minimal channel number", "Minimální číslo programu",
"Minimum duration", "Minimální doba",
"Minimum rotor time (seconds)", "Min. čas rotoru (sekundy)",
"Missing In PAT/SDT", "Chybí v PAT/SDT",
"Mode", "Mód",
"Model description", "Popis modelu",
"Model name", "Název modelu",
"Model number", "Číslo modelu",
"Modulation", "Modulace",
"Module", "Modul",
"Module ID", "ID modulu",
"Mon", "Pon",
"More verbose libav log", "Více upovídaný log pro libav",
"Motor rate (in milliseconds/deg).", "Rychlost motoru (v milisekundách/stupeň).",
"Motor rate (milliseconds/deg)", "Rychlost motoru (ms/stupeň)",
"Motor sport", "Motorové sporty",
"Motoring", "Automobilismus",
"Movie", "Film",
"Movie / drama", "Film / drama",
"Music", "Hudba",
"Music / Ballet / Dance", "Hudba / Balet / Tanec",
"Musical", "Muzikál",
"Mux UUID", "UUID muxu",
"Mux handling", "Zacházení s muxy",
"Mux name", "Jméno muxu",
"Mux not enabled", "Mux není povolen",
"NONE", "NIC",
"Name", "Název",
"Name (or date) of program the entry is a rerun of.", "Název (nebo čas) opakování události.",
"Name of object.", "Název objektu.",
"Name of the entry.", "Název položky.",
"Name of the network.", "Název sítě.",
"Name of the profile.", "Název profilu.",
"Name of the tag.", "Název štítku.",
"Name of the tuner.", "Název tuneru.",
"Name of the tuner/adapter.", "Název tuneru/adaptéru.",
"Names", "Jména",
"Natural sciences", "Přírodní vědy",
"Nature", "Příroda",
"Network", "Síť",
"Network ", "Síť",
"Network A", "Síť A",
"Network B", "Síť B",
"Network C", "Síť C",
"Network D", "Síť D",
"Network ID (limit scanning)", "ID síťě (limit pro vyhledávání)",
"Network UUID", "UUID sítě",
"Network access", "Přístup k síti",
"Network discovery", "Vyhledávání sítí",
"Network for port A.", "Síť pro port A.",
"Network for port AA.", "Síť pro port AA.",
"Network for port AB.", "Síť pro port AB.",
"Network for port B.", "Síť pro port B.",
"Network for port BA.", "Síť pro port BA.",
"Network for port BB.", "Síť pro port BB.",
"Network for port C.", "Síť pro port C.",
"Network for port D.", "Síť pro port D.",
"Network group", "Skupina sítě",
"Network limit per position", "Limit sítě na pozici",
"Network name", "Název sítě",
"Network prefix", "Prefix sítě",
"Network settings", "Nastavení sítě",
"Network type", "Typ sítě",
"Networks", "Sítě",
"New / premiere / unknown", "Nové / premiéra / neznámé",
"New login", "Nové přihlášení",
"New media", "Nová média",
"New muxes + changed muxes", "Nové + změněné muxy",
"New muxes only", "Pouze nové muxy",
"News", "Zprávy",
"News / Current Affairs", "Zprávy / Fakta",
"News magazine", "Magazín se zprávami",
"Next tune delay in ms (0-2000)", "Zpoždění pro další ladění v ms (0-2000)",
"No", "Ne",
"No PIN", "Žádný PIN",
"No access", "Přístup nepovolen",
"No adapters specified!\n", "Žádné adaptéry nebyly specifikovány!\n",
"No assigned adapters", "Žádné adaptéry nejsou přiřazeny",
"No days", "Žádné dny",
"No descrambler", "Žádný dešifrátor",
"No free adapter", "Žádný adaptér není volný",
"No input detected", "Žádný vstup nebyl detekován",
"No scheme", "Žádné schéma",
"No service assigned to channel", "Žádná služba není přiřazena programu",
"No service enabled", "Žádná služba není povolená",
"No source available", "Žádný zdroj není dostupný",
"None", "Žádný",
"Normal", "Standardní",
"Not enough disk space", "Nedostatek místa na disku",
"Not enough memory", "Nedostatek paměti",
"Not linked", "Nespojeno",
"Not set", "Nenastaveno",
"Not set (none or channel configuration)", "Nenastaveno (žádný nebo z konfigurace programu)",
"Not set (use channel or DVR configuration)", "Nenastaveno (použij konfiguraci programu nebo DVR)",
"Number", "Číslo",
"Number of days to keep the file.", "Počet dní pro uchování souboru.",
"Number of days to keep the recorded file.", "Počet dní pro uchování nahrávky.",
"Number of days to keep the recorded files.", "Počet dní pro uchovávání nahrávek.",
"Number of days to retain information about recording.", "Počet dní pro uchování informací o nahrávce.",
"Number of days to retain information about recordings. Once this period is exceeded, duplicate detection will not be possible anymore.", "Počet dní pro uchovávání informací o nahrávkách. Když tato doba uplyne, nebude už možná detekce duplicitních nahrávek.",
"Number of muxes found.", "Počet nalezených muxů.",
"Number of services.", "Počet služeb.",
"Number of tsfile tuners", "Počet tsfile tunerů",
"OK (partial)", "OK (částečný)",
"OSCam net protocol (rev >= 10389)", "OSCam net protokol (rev >= 10389)",
"Odd key", "Lichý klíč",
"Off", "Vypnuto",
"Older OSCam", "Starší OSCam",
"On", "Zapnuto",
"On file removal", "Vymazání souborů",
"On-demand (no first rewind)", "Dle požadavku (bez prvního převíjení zpátky)",
"Once per auto channel", "Jednou pro automatický program",
"Only Bulsatcom 39E", "Pouze Bulsatcom 39E",
"Only EIT", "Pouze EIT",
"Only OpenTV Sky Ausat", "Pouze OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Pouze OpenTV Sky Italia",
"Only OpenTV Sky UK", "Pouze OpenTV Sky UK",
"Only PSIP (ATSC)", "Pouze PSIP (ATSC)",
"Only UK Freesat", "Pouze UK Freesat",
"Only UK Freeview", "Pouze UK Freeview",
"Only Viasat Baltic", "Pouze Viasat Baltic",
"Only admin can watch the realtime log.", "Pouze administrátor může sledovat záznamy ladění v reálném čase.",
"Only digits", "Pouze čislovky",
"Only preferred CA PID", "Pouze preferovaná CA PID",
"Only use specified DVB adapters (comma-separated, -1 = none)", "Použij pouze určené adaptéry DVB (oddělené čárkami, -1 = žádné)",
"OpenTV channel number", "OpenTV číslo programu",
"Options to use/used when mapping - see Help for details.", "Volby pro mapování - podrobnosti naleznete v nápovědě.",
"Options used when updating channels.", "Volby pro aktualizaci programů.",
"Orbital position", "Orbitální pozice",
"Orbital positions", "Orbitální pozice",
"Orbital positions.", "Orbitální pozice.",
"Original network ID", "Originální ID sítě",
"Output directory for tsdebug", "Výstupní adresář pro tsdebug",
"Over-the-air Cron multi-line", "Over-the-air víceřádkový cron",
"Over-the-air EPG grabber", "Over-the-air EPG graber",
"Over-the-air grabbers", "Over-the-air grabery",
"Override disabled", "Zakázané vynucení",
"Override the frontend type.", "Vynutit typ frontendu.",
"Owner", "Majitel",
"Owner of the entry.", "Majitel položky.",
"Owner of the rule.", "Majitel pravidla.",
"PEND", "NEDOKONČENÝ",
"PIDs 21 in setup", "PID číslo 21 v setup příkazu",
"PIN inquiry match string", "PIN pro dotaz z karty",
"PIN inquiry match string.", "PIN pro dotaz z karty.",
"Parent entry", "Rodičovská položka",
"Parent entry.", "Rodičovská položka.",
"Password", "Heslo",
"Password for the entry.", "Heslo pro položku.",
"Password2", "Heslo2",
"Path", "Cesta",
"Path to a directory (folder) containing your picon collection. See Help for more detailed information.", "Cesta k adresáři (složce) obsahujícímu sbírku picon. Přečtěte si nápovědu pro podrobnější popis.",
"Path to an icon for this channel. This can be named however you wish, as either a local (file://) or remote (http://) image. See Help for more infomation.", "Cesta k ikoně pro tento kanál. Tato může být pojmenována buď jako místní (file://) nebo vzdálené (http://) obrázky. Přečtěte si nápovědu pro více informací.",
"Path to the frontend used by the device.", "Cesta k frontendu použitá zařízením.",
"Path to where the timeshift data will be stored. If nothing is specified this will default to CONF_DIR/timeshift/buffer.", "Cesta, kde budou ukládána data časového posunu. Pokud není nastavena, bude standartně v CONF_DIR/timeshift/buffer.",
"Path used by the device.", "Cesta použitá zařízením.",
"Path where the recordings are stored. If components of the path do not exist, Tvheadend will try to create them.", "Cesta, kde budou ukládány nahrávky. Pokud cesta neexistuje, Tvhedend se pokusí ji založit.",
"Peak count of objects", "Největší počet objektů",
"Peak size", "Největší velikost",
"Performing arts", "Múzických umění",
"Periodically save EPG to disk (hours)", "Periodické ukládáni EPG na disk (hod)",
"Persistent user interface level", "Neměnná úroveň uživatelského rozhraní",
"Physiology", "Fyziologie",
"Pick the stream with given audio type only.", "Zvolit pouze stream se zadaným typem audia.",
"Picon name scheme", "Schéma pro názvy picon",
"Picon path", "Cesta k picon",
"Please, follow %s link and cancel the next authorization to correctly clear the cached browser credentals (login and password cache). Then click to the 'Default login' (anonymous access) or 'New login' link in the error page to reauthenticate.", "Následujte prosím odkaz %s a zrušte další autorizaci pro vymazání přihlašovacích údajů uložených v mezipaměti prohlížeče (mezipaměť přihlašovacích údajů a hesel). Poté klikněte na odkaz \"Defaultní přihlášení\" (anonymní přístup) nebo \"Nové přihlášení\" na stránce pro opětovné přihlášení.",
"Polarization", "Polarizace",
"Political issues", "Politická témata",
"Popular culture", "Populární kultura",
"Position", "Pozice",
"Position of the input.", "Pozice vstupu.",
"Post-processor command", "Příkaz pro další zpracování",
"Post-recording padding", "Přidat čas na konci",
"Post-remove command", "Příkaz po vymazání souboru",
"Power save", "Úspora energie",
"Power-up time (ms) (15-200)", "Startovací čas (ms) (15-200)",
"Pre-defined muxes", "Předdefinované muxy",
"Pre-recording padding", "Přidat čas na začátku",
"Pre-school children's programs", "Programy pro předškolní děti",
"Prefer picons over channel name", "Preferuj picony před názvy programů",
"Preferred CA PID", "Preferovaná CA PID",
"Preferred audio language.", "Preferovaný jazyk zvukové stopy.",
"Preferred service video type", "Preferovaný typ videa pro službu",
"Presentation", "Prezentace",
"Presentation details.", "Detaily prezentace.",
"Press", "Tisk",
"Prevent the user from overriding the default user interface level setting and removes the view level drop-dowm from the interface.", "Zakáže uživatelům měnit nastavení úrovně uživatelského rozhraní a odstraní volbu úrovně zobrazení z rozbalovacího menu.",
"Prevent users from overriding the user interface level setting, and remove the view level drop-dowm from the interface.", "Zakáže uživatelům měnit nastavení úrovně uživatelského rozhraní a odstraní volbu úrovně zobrazení z rozbalovacího menu.",
"Priority", "Priorita",
"Priority (-10..10)", "Priorita (-10..10)",
"Priority of the entry, higher-priority entries will take precedence and cancel lower-priority events.", "Priorita položky, položka s vyšší prioritou má přednost a ukončí položky s nižší prioritou.",
"Priority of the recording. Higher priority entries will take precedence and cancel lower-priority events.", "Priorita záznamu. Položka s vyšší prioritou má přednost a ukončí položky s nižší prioritou.",
"Private", "Privátní",
"Profile name", "Název profilu",
"Program synopsis.", "Synonymum programu.",
"Provider", "Poskytovatel",
"Provider ID", "ID poskytovatele",
"Provider network name", "Název sítě poskytovatele",
"Provider's network name.", "Název sítě poskytovatele",
"Psychology", "Psychologie",
"Puppets", "Batolata",
"Quiz", "Kvíz",
"RAM only", "Pouze RAM",
"RTP over TCP payload", "Vytížení RTP přes TCP",
"RTSP Protocol", "RTSP protokol",
"RTSP port", "RTSP Port",
"RTSP port (554 or 9983, 0 = disable)", "RTSP port (554 or 9983), 0 = zakázat",
"Radio", "Rádio",
"Raw Audio Stream", "Holý zvukový stream",
"Re-fetch period (hours)", "Perioda opětovního načtení (v hodinách)",
"Re-fetch period (mins)", "Perioda pro nový pokus (v minutách)",
"Re-record%s%s", "Opětovný rekordér%s%s",
"Re-try period (hours)", "Perioda pro nový pokus (v hodinách)",
"Read error", "Chyba čtení",
"Real Time Streaming Protocol (RTSP) port the server should listen on (554 or 9983, 0 = disable).", "Real Time Streaming Protocol (RTSP) port, na kterém server naslouchá (554 or 9983, 0 = zakázat).",
"Record all", "Zaznamenat vše",
"Record on these days only.", "Zaznamenat pouze v těchto dnech.",
"Recording file options", "Volby souboru pro záznam",
"Recording file size.", "Velikost souboru pro záznam.",
"Recording play count", "Počet přehrání záznamu",
"Recording system path", "Systémový adresář pro záznamy",
"Reject", "Odmítnout",
"Religion", "Náboženství",
"Religious", "Náboženský",
"Remarkable people", "Pozoruhodní lidé",
"Remove HTTP arguments", "Vyjmout HTTP argumenty",
"Remove all unsafe characters from filename", "Odebrat všechny nebezpečné znaky z jména souboru",
"Repeated", "Opakovaný",
"Replace whitespace in title with '-'", "Nahradit mezery v názvu znakem '-'",
"Replaces all whitespace in the title with '-'.", "Nahradit všechny mezery v názvu znakem '-'",
"Reply to CAM PIN inquiries", "Odpovědět na CAM PIN dotazy",
"Reply to PIN inquiries.", "Odpovědět na CAM PIN dotazy",
"Reports", "Zprávy",
"Rerun of", "Opakovaní",
"Reserved", "Rezervováno",
"Resolution (height)", "Rozlišení (výška)",
"Respawn (pipe)", "Znovu spustit (roura)",
"Restart on error", "Restartuj při chybě",
"Restart streaming on error. This is useful for DVR so a recording isn't aborted if an error occurs.", "Restartuj stream po chybě. Toto je užitečné pro DVR, aby se nahrávání nepřerušilo, pokud nastane nějaká chyba.",
"Restrict connections to this type.", "Omezit spojení na tento typ.",
"Reuse EPG from", "Použij EPG z",
"Reuse the EPG from another channel.", "Použij EPG z jiného kanálu.",
"Rewrite EIT", "Přepisovat EIT",
"Rewrite MPEG-TS SI tables", "Přepsat MPEG-TS SI tabulky",
"Rewrite PAT", "Přepisovat PAT",
"Rewrite PMT", "Přepisovat PMT",
"Rewrite SDT", "Přepisovat SDT",
"Rights", "Práva",
"Rotor initialization time (seconds)", "Počateční čas rotoru (sekundy)",
"Rotor type", "Typ rotoru",
"Run as group", "Skupina pro běh",
"Run as user", "Uživatel pro běh",
"Running", "Běží",
"SAT->IP frontend number.", "Číslo SAT->IP frontendu.",
"SAT>IP Client", "SAT>IP Klient",
"SAT>IP DVB-C frequency (Hz)", "SAT>IP DVB-C frekvence (Hz)",
"SAT>IP DVB-S frequency (kHz)", "SAT>IP DVB-S frekvence (kHz)",
"SAT>IP DVB-T frequency (Hz)", "SAT>IP DVB-T frekvence (Hz)",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "SAT>IP RTSP číslo portu pro server\n(defaultní: -1 = disable, 0 = webová konfigurace, standardní port je 554)",
"SAT>IP Satellite Configuration", "SAT>IP Satellite Konfigurace",
"SAT>IP Server SI Tables", "SAT>IP Server SI Tabulky",
"SAT>IP source number", "SAT>IP číslo zdroje",
"SD: standard definition", "SD: standardní rozlišení",
"SSL verify peer", "Verifikovat SSL",
"Sat", "Sob",
"Satconf", "Sat.konf.",
"Satellite config", "Konfigurace satelitu",
"Satellite longitude", "Zeměpisná délka satelitu",
"Satellite longitude.", "Zeměpisná délka satelitu.",
"Satellite position.", "Satelitní pozice.",
"Satellite positions", "Satelitní pozice",
"Save EPG to disk after xmltv import", "Po importu xmltv ulož EPG na disk",
"Scan after creation", "Skenovat po vytvoření",
"Scan progress", "Postup skenu",
"Scan queue length", "Velikost fronty skenu",
"Scan result", "Výsledek scanu",
"Scan status", "Stav scanu",
"Schedule a re-recording if more errors than (0=off)", "Zaznamenat znovu, pokud je počet chyb vyšší než (0=vypnout)",
"Schedule status", "Stav plánovače",
"Schedule status.", "Stav plánovače.",
"Scheduled Duration", "Naplánovaná doba",
"Scheduled for recording", "Naplánované k záznamu",
"Scheduled start time", "Naplánovaný čas začátku",
"Scheduled stop time", "Naplánovaný čas konce",
"School programs", "Školní programy",
"Science", "Věda",
"Science fiction", "Sci-fi",
"Script/program to run when a recording completes.", "Skript/program ke spuštění po dokončení nahrávky.",
"Script/program to run when a recording gets removed.", "Skript/program ke spuštění po smazání nahrávky.",
"Script/program to run when a recording starts (service is subscribed but no filename available).", "Skript/program ke spuštění po začátku nahrávky (služba běží, ale jméno souboru není dostupné).",
"Season", "Sezóna",
"Season %d", "Sezóna %d",
"Season information (if available).", "Informace o sezóně (pokud je dostupná).",
"Seek error", "Chyba nastavení pozice",
"Select an available network type for this tuner.", "Zvolte dostupný typ sítě pro tento tuner.",
"Select high priority (default) EPG language.", "Vyberte jazyk EPG pro vysokou prioritu (defaultní).",
"Select low priority EPG language.", "Vyberte jazyk EPG pro nízkou prioritu.",
"Select medium priority EPG language.", "Vyberte jazyk EPG pro střední prioritu.",
"Select subtitle codec to use for transcoding.", "Zvolte kodek titulků pro transkódování",
"Select the default user interface language. This can be overridden later in \"Access Entries\" on a per-user basis.", "Volba defaultního jazyka uživatelského rozhraní. Může být přepsáno v \"Seznam uživatelů\" pro každého uživatele.",
"Select the list of languages (in order of priority) to be used for supplying EPG information to clients that don't provide their own configuration.", "Vyberte seznam jazyků (v pořadí priority), který bude použit pro poskytování informací EPG klientům, kteří neposkytují vlastní konfiguraci.",
"Select the path to use for DVB scan configuration files. Typically dvb-apps stores these in /usr/share/dvb/. Leave blank to use the internal file set.", "Vyberte cestu pro konfigurační soubory skenování DVB. Aplikace dvb-apps obvykle ukládají do /usr/share/dvb/. Ponechte prázdné pro použití interní sady souborů.",
"Select the time offset for EIT events.", "Nastavení časového posunu pro EIT události.",
"Select type of broadcast.", "Zvolte typ vysílání.",
"Send CAPMT query", "Poslat CAPMT dotaz",
"Send full PLAY cmd", "Poslat plný PLAY příkaz",
"Serial number", "Sériové číslo",
"Series link", "Odkaz série",
"Series link ID.", "Identifikátor spojení pro série.",
"Serious", "Vážný",
"Serious music", "Vážná hudba",
"Server UUID", "UUID serveru",
"Server connectivity", "Konektivita serveru",
"Server details.", "Detaily serveru.",
"Service", "Služba",
"Service ID", "ID služby",
"Service configuration", "Konfigurace služeb",
"Service count %zi", "Počet služeb %zi",
"Service mapping", "Mapování služeb",
"Service name", "Název služby",
"Service name found in EPG data.", "Název služby byl nalezen v EPG datech.",
"Service name picons", "Picony s názvy služeb",
"Service raw", "Základní služba",
"Service type", "Typ služby",
"Services", "Služby",
"Set as default profile.", "Nastavit jako defaultní profil.",
"Set the display name.", "Nastaví název pro zobrazení.",
"Set the name of the server so you can distinguish multiple instances apart.", "Nastavte název serveru tak, abyste od sebe mohli odlišit více instancí.",
"Sets the default interface view level (next to the Help button).", "Nastaví výchozí úroveň zobrazení rozhraní (vedle tlačítka Nápovědy).",
"Settings", "Nastavení",
"Show / Game show", "Show / Herní show",
"Show only information for sessions which are initiated from an IP address of the requester.", "Zobrazit pouze informace pro relace, které jsou iniciovány z IP adresy žadatele.",
"Show this page", "Ukaž tuto stránku",
"Show version information", "Ukaž informace o verzi",
"Show, hide and sort the various details that appear on the interface next to the About tab.", "Zobrazit, skrýt a třídit detaily, které se objeví na rozhraní vedle karty O programu.",
"Signal level for IPTV sources (0-240).", "Úroveň signálu IPTV zdrojů (0-240).",
"Signal quality too poor", "Kvalita signálu je velmi nízká",
"Signal scale (240 or 100)", "Škála signálu (240 nebo 100)",
"Site latitude", "Zeměpisná šířka umistění",
"Site latitude.", "Zeměpisná šířka umistění.",
"Site longitude", "Zeměpisná délka umístění",
"Site longitude.", "Zeměpisná délka umístění.",
"Size", "Velikost",
"Skip TS packets (0-200)", "Přeskočit TS pakety (0-200)",
"Skip TSID checking. For when providers use invalid Transport Stream IDs.", "Přeskočit kontrolu TSID. Použijte, pokud poskytovatel používá chybné Transport Stream IDs.",
"Skip commercials", "Přeskočit reklamy",
"Skip initial bytes", "Přeskoč počáteční byty",
"Skip scanning known muxes when Tvheadend starts. If \"startup scan\" is allowed and new muxes are found then they will still be scanned. See Help for more details.", "Přeskočit prohledávání známých muxů při spuštění Tvheadend. Je-li povoleno \"Ladění po spuštění\" a objeví se nové muxy, budou stále prohledány. Podrobnější informace naleznete v nápovědě.",
"Skip startup scan", "Přeskočit ladění po spuštění",
"Slave entry", "Závislá položka",
"Slave entry.", "Závislá položka",
"Slot state", "Stav slotu",
"Soap", "Telenovely",
"Soccer", "Fotbal",
"Social", "Sociální",
"Social / Political issues / Economics", "Sociální / Politická témata / Ekonomika",
"Social advisory", "Sociální poradenství",
"Socket or IP Address (when in TCP mode).", "Socket, nebo IP adresa (v TCP módu).",
"Some DVB devices have an optional ioctl that allows changing between normal voltage for LNB (13V/18V) to a higher voltage mode (usually, 14V/19V), meant to compensate for voltage loss on long cabling. Without that, it is not possible to properly switch the polarization.", "Některá zařízení DVB mají volitelný ioctl, který umožňuje přepínat mezi normálním napětím pro LNB (13V/18V) a vyšším napětím (obvykle 14V/19V), který má kompenzovat ztráty napětí při dlouhé kabeláži. Bez tohoto není možné správně přepínat polarizaci.",
"Sort index", "Index třídění",
"Sort index.", "Index pro třídění.",
"Source", "Zdroj",
"Source deleted", "Zdroj smazán",
"Source quality is bad", "Kvalita zdroje je špatná",
"Source reconfigured", "Zdroj byl rekonfigurován",
"Source video codec", "Zdrojový video kodek",
"Southern hemisphere (latitude direction)", "Jižní polokoule (směr zem. šířky)",
"Southern hemisphere (latitude direction).", "Jižní polokoule (směr zem. šířky).",
"Special events (Olympic Games, World Cup, etc.)", "Speciální události (Olympijské hry, Světový pohár, apod.)",
"Specify User-Agent header for the http client", "Specifikuj hlavičku User-Agent pro http klienta",
"Specify alternative htsp port", "Specifikuj alternativní htsp port",
"Specify alternative http port", "Specifikuj alternativní http port",
"Specify alternative http webroot", "Specifikuj alternativní http webroot",
"Specify bind address", "Specifikuj IP adresu pro příjem",
"Specify extra htsp port", "Specifikuj extra htsp port",
"Spiritual sciences", "Spirituální vědy",
"Sports", "Sporty",
"Sports magazines", "Sportovní magazíny",
"Standard", "Standardní",
"Start", "Začátek",
"Start after", "Začátek po",
"Start before", "Začátek před",
"Start recording earlier than the EPG/timer defined start time by x minutes, for example if a program is to start at 13:00 and you set a padding of 5 minutes it will start recording at 12:54:30 (including a warming-up time of 30 seconds). If this isn't set the pre-recording padding if set in the DVR entry or DVR profile will be used.", "Začne nahrávat o x minut dříve, než v čas nastavený EPG/časovačem: například pokud má program začít v 13:00 a nastavíte padding 5 minut, začne nahrávat v 12:54:30 ( včetně zahřívací doby 30 sekund). Není-li nastaveno, použije se předběžné nahrávání nastavené v kanálu, nebo DVR.",
"Start recording earlier than the EPG/timer-defined start time by x minutes.", "Začne nahrávat o x minut dříve, než v čas nastavený EPG/časovačem.",
"Start recording earlier than the defined Start recording earlier than the defined start time by x minutes: for example, if a program is to start at 13:00 and you set a padding of 5 minutes it will start recording at 12:54:30 (including a warm-up time of 30 seconds). If this isn't specified, any pre-recording padding as set in the channel or DVR entry will be used.", "Začne nahrávat o x minut dříve, než v nastavený Začátek nahrávání: například pokud má program začít v 13:00 a nastavíte padding 5 minut, začne nahrávat v 12:54:30 ( včetně zahřívací doby 30 sekund). Není-li nastaveno, použije se předběžné nahrávání nastavené v kanálu, nebo DVR.",
"Start recording earlier than the defined start time by x minutes.", "Začne nahrávat o x minut dříve, než v nastavený Začátek nahrávání.",
"Start time", "Začátek",
"Status", "Stav",
"Status period (ms)", "Perioda čtení statusu (ms)",
"Stop", "Zastavit",
"Stop time", "Konec",
"Storage path", "Cesta k úložišti",
"Storage space", "Místo na úložišti",
"Stream ID", "Identifikace streamu",
"Stream Profile", "Profil streamu",
"Stream index", "Index streamu",
"Stream index (starts with zero).", "Index streamu (začíná nulou).",
"Stream profile", "Profil streamu",
"Stream setup error", "Chyba v založení streamu",
"Stream type", "Typ streamu",
"Streaming", "Streamování",
"Streaming Profile", "Streamovací profil",
"Streaming flags, allow/disallow HTTP streaming, advanced HTTP streaming (e.g, direct service or mux links), HTSP protocol streaming (e.g, Kodi (via pvr.hts) or Movian.", "Volby streamování, povolení/zakázání HTTP streamování, pokročilé HTTP streamování (např. přímá služba, nebo multiplex linka), protokol HTSP (např. Kodi (přes pvr.hts), nebo Movian.",
"Streaming priority", "Priorita streamu",
"Streaming profiles", "Profily streamů",
"Subdirectory options", "Volby podadresáře",
"Subscribe to a service permanently", "Odebírej tuto službu permanentně",
"Subscription", "Odběry",
"Subscription does not exist", "Odběr neexistuje",
"Subscription overridden", "Odběr přehlasován jiným požadavkem",
"Subscription weight", "Váha odběru",
"Substitute formatters", "Nahraď formátovací znaky",
"Subtitle", "Podtitulek",
"Subtitle Stream Filter", "Filtr streamu titulků",
"Subtitle codec", "Kodek podtitulků",
"Subtitle of the program (if any) (display only).", "Titulky programu (pokud jsou dostupné) (pouze pro zobrazení).",
"Subtitle of the program (if any).", "Titulky programu (pokud jsou dostupné)",
"Sun", "Ned",
"Surround (2 front, rear mono)", "Surround (2 přední, zadní mono)",
"Switch before rotor", "Přepínač pak rotor",
"Switch to another service", "Přepnout na jinou službu",
"Switch type", "Typ přepínače",
"Switch: %s", "Přepínač: %s",
"Sync", "Synchronizovat",
"Sync + Don't keep", "Synch. ₊ nezanechávej",
"System", "Systém",
"TCP Protocol", "TCP Protokol",
"TV Adapters - Linux ATSC-C Frontend", "TV Adaptéry - Linux ATSC-C Frontend",
"TV Adapters - Linux ATSC-T Frontend", "TV Adaptéry - Linux ATSC-T Frontend",
"TV Adapters - Linux DAB Frontend", "TV Adaptéry - Linux DAB Frontend",
"TV Adapters - Linux DVB-C Frontend", "TV Adaptéry - Linux DVB-C Frontend",
"TV Adapters - Linux DVB-S Frontend (Master)", "TV Adaptéry - Linux DVB-S Frontend (Master)",
"TV Adapters - Linux DVB-S Slave Frontend", "TV Adaptéry - Linux DVB-S Slave Frontend",
"TV Adapters - Linux DVB-T Frontend", "TV Adaptéry - Linux DVB-T Frontend",
"TV Adapters - Linux ISDB-C Frontend", "TV Adaptéry - Linux ISDB-C Frontend",
"TV Adapters - Linux ISDB-S Frontend (Master)", "TV Adaptéry - Linux ISDB-S Frontend (Master)",
"TV Adapters - Linux ISDB-T Frontend", "TV Adaptéry - Linux ISDB-T Frontend",
"TV Adapters - SatConfig - 4-Port", "TV Adaptéry - SatConfig - 4-Port",
"TV Adapters - SatConfig - Advanced", "TV Adaptéry - SatConfig - Rozšířené",
"TV Adapters - SatConfig - DiseqC Rotor", "TV Adaptéry - SatConfig - DiseqC Rotor",
"TV Adapters - SatConfig - DiseqC Switch", "TV Adaptéry - SatConfig - DiseqC Přepínač",
"TV Adapters - SatConfig - EN50494/UniCable I", "TV Adaptéry - SatConfig - EN50494/UniCable I",
"TV Adapters - SatConfig - EN50607/UniCable II", "TV Adaptéry - SatConfig - EN50607/UniCable II",
"TV Adapters - SatConfig - GOTOX Rotor", "TV Adaptéry - SatConfig - GOTOX Rotor",
"TV Adapters - SatConfig - Tone Burst/2 Port", "TV Adaptéry - SatConfig - Tone Burst/2 Port",
"TV Adapters - SatConfig - USALS Rotor", "TV Adaptéry - SatConfig - USALS Rotor",
"TV Adapters - SatConfig - Universal LNB (Simple)", "TV Adaptéry - SatConfig - Universal LNB (Simple)",
"TVHDHomeRun Client", "TVHDHomeRun Klient",
"Tag files with metadata", "Vlož metadata do souborů",
"Tags", "Štítky",
"Team sports (excluding football)", "Týmové sporty (vyjma fotbalu)",
"Technology", "Technologie",
"Tennis", "Tenis",
"Testing", "Testování",
"Testing options", "Testovací volby",
"The CAM slot status.", "Stav CAM vstupu.",
"The DVR profile to be used/used by the recording.", "DVR profil použitý pro nahrávku.",
"The DVR profile to be used/used by this rule.", "DVR profil použitý pro toto pravidlo.",
"The PIN to use.", "Použít toto PIN.",
"The SAT>IP server's name.", "Název SAT>IP serveru.",
"The SAT>IP server's universally unique identifier.", "Unikátní identifikátor pro SAT>IP server.",
"The SAT>IP source number.", "Číslo SAT>IP zdroje.",
"The SAT>IP's discovered IP address.", "Detekovaná IP adresa pro SAT>IP zařízení.",
"The URL to the icon to use/used for the channel. The local files are referred using file:/// URLs.", "Cesta k ikoně kanálu. Místní soubory se nastaví pomocí URL file:///",
"The cache scheme to use/used to store recordings. Leave as \"system\" unless you have a special use case for one of the others. See Help for details.", "Schéma mezipaměti pro ukládání záznamů. Ponechte jako \"systém\", pokud nemáte zvláštní důvod pro jiné nastavení. Podrobné informace naleznete v nápovědě.",
"The channel on which this rule applies, i.e. the channel you're aiming to record. You can leave this field blank to apply the rule to all channels.", "Kanál, na který se toto pravidlo vztahuje, tj. kanál, který chcete zaznamenat. Toto pole můžete ponechat prázdné, aby bylo pravidlo použito pro všechny kanály.",
"The channel this service is mapped to.", "Kanál, který je mapován pro tuto službu.",
"The character encoding for this network (e.g. UTF-8).", "Znaková sada pro tuto síť (např. UTF-8).",
"The character encoding for this service (e.g. UTF-8).", "Znaková sada pro tuto službu (např. UTF-8).",
"The content type (Movie/Drama, Sports, etc.) to be used to filter matching events/programs.", "Typ obsahu (Film/Drama, Sport, atd.) použitý pro filtrování pořadů.",
"The current boot ID.", "Současné boot ID.",
"The current configuration ID.", "Současné konfigurační ID.",
"The current configuration version.", "Současná verze konfigurace.",
"The default language to use if the user  language isn't set in the Access Entries tab.", "Defaultní jazyk, pokud není nastaven v záložce Seznam uživatelů.",
"The default web interface to use if the user  theme isn't set in the Access Entries tab.", "Defaultní téma uživatelského rozhraní, pokud není nastaveno v záložce Seznam uživatelů.",
"The device ID.", "ID zařízení.",
"The device's serial number.", "Sériové číslo zařízení.",
"The input path used by the device.", "Vstupní cesta používaná zařízením.",
"The language to which the filter should apply.", "Jazyk, pro který aplikovat filtr.",
"The layer A constellation.", "Konstelace vrstvy A.",
"The layer A forward error correction.", "Forward error correction (FEC) vrstvy A.",
"The layer A segment count.", "Počet segmentů vrstvy A.",
"The layer A time interleaving.", "Časové prokládání vrstvy A.",
"The layer B constellation.", "Počet segmentů vrstvy A.",
"The layer B forward error correction.", "Forward error correction (FEC) vrstvy B.",
"The layer B segment count.", "Počet segmentů vrstvy B.",
"The layer B time interleaving.", "Časové prokládání vrstvy B.",
"The layer C constellation.", "Konstelace vrstvy C.",
"The layer C forward error correction.", "Forward error correction (FEC) vrstvy C.",
"The layer C segment count.", "Počet segmentů vrstvy C.",
"The layer C time interleaving.", "Časové prokládání vrstvy C.",
"The logical stream index to compare. Note that this index is computed using all filters.Example: If filter is set to AC3 audio type and the language to 'eng' and there are two AC3 'eng' streams in the service, the first stream could be identified using number 1 and the second using number 2.", "Index logického toku pro porovnání. Všimněte si, že tento index je vypočítán pomocí všech filtrů.Příklad: Pokud je filtr nastaven na typ audia AC3 a jazyk na 'eng' a ve službě jsou dva proudy AC3 'eng', první proud by mohl být identifikován pomocí čísla 1 a druhá používá číslo 2.",
"The maximal duration of a matching event - in other words, only match programmes that are no longer than this duration.", "Maximální doba trvání vybraného pořadu - zjednodušeně, vybrat pouze pořady, které nejsou delší, než tento čas.",
"The maximum RAM (system memory) size for timeshift buffers. When free RAM buffers are available they are used for timeshift data in preference to using storage.", "Maximální velikost RAM (operační paměti) pro paměť časového posunu. Pokud je dostupná volná RAM, je přednostně použita namísto ukládání na disk.",
"The maximum combined size of all timeshift buffers. If you specify an unlimited period it's highly recommended you specify a value here.", "Maximální celková velikost všech pamětí časového posunu. Pokud jste nastavili neomezenou velikost, je vysoce doporučováno nastavit hodnotu limitu zde.",
"The maximum number of entries that can be matched.", "Maximální počet položek, které mohou být úspěšně porovnány.",
"The maximum number of input streams allowed on this network.", "Maximální počet vstupních proudů povolených na této síti.",
"The maximum number of recording entries this rule can create.", "Maximální počet nahrávek vytvořených tímto pravidlem.",
"The maximum number of recordings that can be scheduled.", "Maximální počet nahrávek, které mohou být současně naplánovány.",
"The maximum number of times this rule can be triggered.", "Maximální počet opakování tohoto pravidla.",
"The minimal duration of a matching event - in other words, only match programs that are no shorter than this duration.", "Minimální doba trvání vybraného pořadu - zjednodušeně, vybrat pouze pořady, které nejsou kratší, než tento čas.",
"The modulation used on the mux.", "Modulace použitá pro mux.",
"The mux the service is on.", "Mux, na kterém je služba.",
"The name of the profile.", "Název profilu.",
"The name of the the rule.", "Název pravidla.",
"The network prefix(es) to block, e.g.192.168.2.0/24 (comma-separated list).", "Síťové prefixy k blokování, např: 192.168.2.0/24 (seznam oddělený čárkou).",
"The network the mux is on.", "Síť na které je tento mux.",
"The network the service is on.", "Síť na které je tato služba.",
"The networks using this configuration.", "Sítě používající tuto konfiguraci.",
"The networks' universally unique identifier (UUID).", "Unikátní identifikátor pro síť (UUID).",
"The number of ATSC-C (Cable/AnnexB) tuners to export.", "Počet ATSC-C (Cable/AnnexB) tunerů pro export.",
"The number of ATSC-T (Terresterial) tuners to export.", "Počet ATSC-T (Terresterial) tunerů pro export.",
"The number of DVB-C (Cable) tuners to export.", "Počet DVBC-C (Cable) tunerů pro export.",
"The number of DVB-C2 (Cable) tuners to export.", "Počet DVBC-C2 (Cable) tunerů pro export.",
"The number of DVB-S (Satellite) tuners to export.", "Počet DVB-S (Satellite) tunerů pro export.",
"The number of DVB-S2 (Satellite) tuners to export.", "Počet DVB-S2 (Satellite) tunerů pro export.",
"The number of DVB-T (Terresterial) tuners to export.", "Počet DVB-T (Terresterial) tunerů pro export.",
"The number of DVB-T2 (Terresterial) tuners to export.", "Počet DVB-T2 (Terresterial) tunerů pro export.",
"The number of bytes to buffer. By default, linuxdvb's input buffer is 18800 bytes long. The accepted range is 18800-1880000 bytes.", "Počet bajtů pro vyrovnávací paměť. Ve výchozím nastavení je vstupní vyrovnávací paměť linuxdvb dlouhá 18800 bajtů. Povolený rozsah je 18800-1880000 bajtů.",
"The number of muxes left to scan on this network.", "Celkový počet neprohlídnutých muxů v této síti.",
"The number of services currently mapped to channels.", "Počet služeb právě namapovaných na kanál.",
"The provider's ID.", "ID poskytovatele.",
"The provider's name.", "Název poskytovatele.",
"The provider's network ID.", "ID sítě poskytovatele",
"The provider's network name.", "Název sítě poskytovatele.",
"The quadrature amplitude modulation (QAM) used by the mux. If you're not sure of the value leave as AUTO.", "Kvadraturní amplitudová modulace (QAM) používaná mux. Pokud si nejste jisti, ponechte hodnotu na AUTO.",
"The recorded file was removed intentionally", "Nahraný soubor byl úmyslně smazán.",
"The recording/entry status.", "Stav záznamu/položky.",
"The scheduled start time, including any padding.", "Naplánovaný čas začátku, včetně času přidaného na začátku.",
"The scheduled stop time, including any padding.", "Naplánovaný čas konce, včetně času přidaného na konci.",
"The service ID", "ID služby",
"The service ID as set by the provider.", "ID služby nastavené poskytovatelem obsahu.",
"The service ID.", "ID služby.",
"The service name as set by the provider.", "Název služby nastavené poskytovatelem obsahu.",
"The service type flag as defined by the DVB specifications (e.g. 0x02 = radio, 0x11 = MPEG2 HD TV, 0x19 = H.264 HD TV)", "Příznak typu služby, jak je definovaný DVB specifikací (např. 0x02 = rádio, 0x11 = MPEG2 HD TV, 0x19 = HDTV H.264)",
"The service's channel minor as set by the provider.", "Čislo kanálu (vedlejší) nastavené poskytovatelem obsahu.",
"The service's channel number as set by the provider.", "Čislo kanálu (hlavní) nastavené poskytovatelem obsahu.",
"The service's encryption status.", "Stav šifrování služby.",
"The start time of the recording.", "Časový začátek záznamu.",
"The stream profile the DVR profile will use for recordings.", "Profil DVR streamu použitý pro nahrávky.",
"The string allows you to manually specify the full path generation using predefined modifiers. See Help for full details.", "Řetězec umožňuje ručně zadat generování úplné cesty pomocí předdefinovaných modifikátorů. Podrobné informace naleznete v nápovědě.",
"The symbol rate used on the mux.", "Přenosová rychlost použitá na tomto muxu.",
"The symbol rate.", "Přenosová rychlost.",
"The total number of services found.", "Celkový počet nalezených služeb.",
"The total scheduled duration.", "Celková doba plánovaného záznamu.",
"The transport stream ID of the mux within the network.", "ID transportního streamu pro mux v síti.",
"The video stream types the filter should apply to.", "Typu video streamů pro které má být filtr aplikován.",
"Theme", "Téma",
"This tuner", "Tento tuner",
"Thu", "Čtv",
"Time", "Čas",
"Time missed", "Čas byl zmeškán",
"Time record caption", "Titulek pro automatický časový rekordér",
"Time recording%s%s", "Časový rekordér%s%s",
"Time to start the recording/time the recording started.", "Čas pro začátek nahrávání/čas, kdy nahrávání začalo.",
"Time to stop recording/time the recording stopped.", "Čas pro konec nahrávání/čas, kdy nahrávání skončilo.",
"Time update", "Aktualizace času",
"Timeout (sec) (0=infinite)", "Časový limit (sek.) (0=bez limitu)",
"Timeout (seconds)", "Časový limit (sekundy)",
"Timeout (secs)", "Časový limit (sek)",
"Timer-based automatic recording.", "Automatický záznam na základě časovačů.",
"Title", "Název",
"Title (regexp)", "Název (regexp)",
"Title of the program (display only).", "Název vysílání (pouze pro zobrazení).",
"Title of the program.", "Název vysílání.",
"Title of the recording - this is used to generate the filename.", "Název záznamu - toto je použito pro generaci názvu souboru.",
"Toggle details", "Přepnout detaily",
"Tone burst", "Tone Burst",
"Too many data errors", "Příliš mnoho chyb v datech",
"Total number of mapped channels on this network.", "Celkový počet namapovaných programů v této síti.",
"Total number of muxes found on this network.", "Celkový počet nalezených muxů v této síti.",
"Total number of services found on this network.", "Celkový počet nalezených služeb v této síti.",
"Total number of services found.", "Celkový počet nalezených služeb.",
"Total number of services seen.", "Celkový počet viděných služeb.",
"Total number of services.", "Celkový počet služeb.",
"Tourism / Travel", "Turismus / Cestování",
"Trace subsystems", "Subsystémy pro trasování",
"Traditional arts", "Tradiční umění",
"Traditional music", "Tradiční hudba",
"Transcode", "Transkódování",
"Transcode video only if source video codec matches. \"Any\" ignores source video codec checking and always transcodes. If no codec match is found, transcoding is done using the \"copy\" codec. if a match is found, transcode with the parameters in this profile. Separate codec names with comma.", "Transkódvat video pouze pokud zdrojový videokodek odpovídá. \"Jakýkoliv\" ignoruje kontrolu zdrojového kodeku videa a vždy transkóduje. Pokud není nalezena shoda kodeku, transkódování se provádí pomocí kodeku \"Kopie\". Pokud je nalezena shoda, transkódování probíhá s parametry v tomto profilu. Názvy kodeků oddělujte čárkou.",
"Transcoding", "Transkódování",
"Transmission mode", "Mód vysílání",
"Transponder ID", "ID transpodéru",
"Transport stream ID", "ID streamu transpondéru",
"Tue", "Úte",
"Tune before DiseqC", "Ladit před DiseqC",
"Tuner bind IP address", "IP adresa pro komunikaci s tunerem",
"Tuner configuration", "Konfigurace tuneru",
"Tuner configuration.", "Konfigurace tuneru",
"Tuning failed", "Ladění selhalo",
"Turn off LNB when idle", "Vypni LNB při nečinnosti",
"Tvheadend server name", "Jméno Tvheadend serveru",
"Type", "Typ",
"Type override", "Vynutit typ služby",
"UDP RTP port number (2 ports)", "UDP RTP číslo portu (2 porty)",
"UHD: ultra high definition", "UHD: ultra vysoké rozlišení",
"UPnP Protocol", "UPnP protokol",
"URL of the M3U playlist.", "URL pro M3U playlist.",
"URL with the SAT>IP server XML location", "URL s umístěním XML souboru na SAT>IP serveru",
"Unable to get system UTC time", "Nemohu získat systémový UTC čas",
"Unable to get system local time", "Nemohu získat systémový lokální čas",
"Unable to get system time", "Nemohu získat systémový čas",
"Unable to open file", "Nemohu otevřít soubor",
"Unable to stat path", "Nemohu zjistit informace o cestě",
"Uncommitted first", "Preferuj uncommited",
"Unicable I switch (universal LNB)", "Unicable I přepínač (universal LNB)",
"Unicable II switch (universal LNB)", "Unicable II přepínač (universal LNB)",
"Unicable type", "Typ unicable",
"Unimportant", "Nedůležité",
"Universal LNB only", "Pouze univerzální LNB",
"Universally unique identifier. Read only.", "Univerzální unikátní identifikátor. Pouze pro čtení.",
"Unknown", "Neznámé",
"Unknown file", "Neznámý soubor",
"Unknown reason (%i)", "Neznámá příčina (%i)",
"Unlimited size", "Nelimitovaná velikost",
"Unlimited time", "Nelimitovaná doba",
"Unset (default)", "Nenastaveno (standardní)",
"Update card (EMM)", "Aktualizovat kartu (EMM)",
"Update channel icon", "Aktualizovat ikonu programu",
"Update channel name", "Aktualizovat název programu",
"Update channel number", "Aktualizovat číslo programu",
"Update disabled", "Aktualizace zakázána",
"Update time", "Aktualizovat čas",
"Update tolerance (ms)", "Tolerance aktualizace (ms)",
"Updated", "Aktualizováno",
"Updates from one mux (EMM)", "Aktualizace pouze z jednoho muxu (EMM)",
"Upon start, Tvheadend doesn't know the last rotor position. This value defines the initial rotor movement. TVHeadend waits the specified time when the first movement is requested.", "Při startu, Tvheadend nezná poslední pozici rotoru. Tato hodnota definuje počáteční pohyb rotoru. Tvheadend čeká specifikovanou dobu po požadavku na nastavení první pozice.",
"Usage: %s [OPTIONS]\n", "Použití: %s [VOLBY]\n",
"Use EITp/f to decide event start/stop. This is also known as \"Accurate Recording\". See Help for details.", "Použít EITp/f pro určení začátku a konce pořadu. Toto je také známo jako \"Přesné nahrávání\". Přečtěte si nápovědu pro prodrobnosti.",
"Use EPG running state", "Použít stav vysílání z EPG ",
"Use HTTP digest authentication", "Použít 'HTTP digest' autorizaci",
"Use PROXY protocol & X-Forwarded-For", "Použít PROXY protokol & X-Forwarded-For",
"Use WEBM format.", "Použij formát WEBM.",
"Use Windows-compatible filenames", "Použij názvy souborů kompatibilní s Windows",
"Use XSPF playlist instead of M3U", "Použít XSPF playlist místo M3U",
"Use a pre-defined list of ATSC-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Použít předdefinovaný seznam ATSC-C muxů. Pozn: tyto seznamy mohou být zastaralé a mohou zapříčinit, že hledání bude trvat déle než obvykle.",
"Use a pre-defined list of ATSC-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Použít předdefinovaný seznam ATSC-T muxů. Pozn: tyto seznamy mohou být zastaralé a mohou zapříčinit, že hledání bude trvat déle než obvykle.",
"Use a pre-defined list of DAB muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Použít předdefinovaný seznam DABmuxů. Pozn: tyto seznamy mohou být zastaralé a mohou zapříčinit, že hledání bude trvat déle než obvykle.",
"Use a pre-defined list of DVB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Použít předdefinovaný seznam DVB-C muxů. Pozn: tyto seznamy mohou být zastaralé a mohou zapříčinit, že hledání bude trvat déle než obvykle.",
"Use a pre-defined list of DVB-S/S2 muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Použít předdefinovaný seznam DVB-S/S2 muxů. Pozn: tyto seznamy mohou být zastaralé a mohou zapříčinit, že hledání bude trvat déle než obvykle.",
"Use a pre-defined list of DVB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Použít předdefinovaný seznam DVB-T muxů. Pozn: tyto seznamy mohou být zastaralé a mohou zapříčinit, že hledání bude trvat déle než obvykle.",
"Use a pre-defined list of ISDB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Použít předdefinovaný seznam ISDB-C muxů. Pozn: tyto seznamy mohou být zastaralé a mohou zapříčinit, že hledání bude trvat déle než obvykle.",
"Use a pre-defined list of ISDB-S muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Použít předdefinovaný seznam ISDB-S muxů. Pozn: tyto seznamy mohou být zastaralé a mohou zapříčinit, že hledání bude trvat déle než obvykle.",
"Use a pre-defined list of ISDB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Použít předdefinovaný seznam ISDB-T muxů. Pozn: tyto seznamy mohou být zastaralé a mohou zapříčinit, že hledání bude trvat déle než obvykle.",
"Use only this service ID, filter out others.", "Použít pouze toto ID služby, ignorovat ostatní.",
"Use original", "Použít originál",
"Use packet backlog", "Používat neúplné snímky",
"Use service IDs as channel numbers", "ID služby jako čísla programů",
"Use tag internally (don't expose to clients).", "Používat štítek interně (není viditelný pro klienty).",
"Use the provider's service IDs as channel numbers.", "Použít IDs poskytovatele služby jako čísla kanálů.",
"Use this character set when setting filenames.", "Použij tuto znakovou sadu při vytváření souborů.",
"Use x amount of storage space.", "Použít x volného místa na úložišti.",
"User access error", "Chyba přístupu uživatele",
"User does not have access", "Uživatel nemá přístup",
"User does not have access to channel", "Uživatel nemá přístup k programu",
"User icon", "Uživatelská ikona",
"User interface level", "Úroveň uživatelského rozhraní",
"User interface quick tips (tooltips)", "Rychlá nápověda (tooltipy) uživatelského rozhraní",
"User language", "Jazyk uživatele",
"User limit reached", "Dosažen limit uživatele",
"User login", "Uživatelské jméno",
"User request", "Požadavek od uživatele",
"Username", "Uživatelské jméno",
"Username for the entry (login username).", "Uživatelské jméno pro položku (přihlašovací uživatelské jméno).",
"Username of the entry (this should match a username from within the \"Access Entries\" tab).", "Uživatelské jméno (musí souhlasit s uživatelským jménem ze záložky \"Uživatelé - Seznam uživatelů\")",
"Users - Access Entries", "Uživatelé - Seznam uživatelů",
"Users - IP Blocking", "Uživatelé - Blokování IP",
"Users - Passwords", "Uživatelé - Hesla",
"Variety show", "Varietní show",
"Verify the SSL certificate.", "Ověřit SSL certifikát.",
"Video bitrate (kb/s) (0=auto)", "Šířka pásma pro video (kb/s) (0=Auto)",
"Video codec", "Video kodek",
"Video codec preset", "Nastavení pro video kodek",
"Video codec preset to use for transcoding.", "Nastavení pro video kodek použité pro transkódování.",
"Video codec to use for the transcode. \"Do not use\" will disable video output.", "Video kodek použítý pro transkódování. \"Nepoužívat\" zakáže video výstup.",
"Video recorder", "Video rekordér",
"Video recorder flags, allow/disallow access to video recorder functionality (including Autorecs), allow/disallow users to view other DVR entries, allow/disallow users to work with DVR entries of other users (remove, edit) etc.", "Štítky videorekordéru, povolit/zakázat přístup k funkcím videorekordéru (včetně Autorekordéru), povolit/zakázat uživatelům prohlížení ostatních záznamů DVR, povolit/zakázat uživatelům práci se záznamy DVR ostatních uživatelů (odstranění, úpravy) atd.",
"View all", "Zobrazit vše",
"Visually impaired commentary/audio description", "Komentář pro zrakově postižené",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV profil H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV profil H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV profil VP8/Vorbis/WEBM",
"Waiting for EPG running flag", "Čekám na stav vysílání z EPG",
"Waiting for program start", "Čekám na start programu",
"Waiting for stream", "Čekám na stream",
"War", "Válka",
"Water sport", "Vodní sporty",
"Weak stream", "Slabý proud dat",
"Weather report", "Zprávy o počasí",
"Web User Interface", "Webové uživatelské rozhraní",
"Web interface", "Webové rozhraní",
"Web interface language", "Jazyk pro webové rozhraní",
"Web interface language.", "Jazyk webového rozhraní.",
"Web interface theme.", "Téma webového rozhraní.",
"Web theme", "Webové téma",
"Web user interface", "Uživatelské webové rozhraní",
"Wed", "Stř",
"Welcome - Tvheadend - your TV streaming server and video recorder", "Vítejte - Tvheadend - Váš TV server pro streamování a video rekordér",
"Western hemisphere (latitude direction)", "Západní polokoule (směr zem. délky)",
"Western hemisphere (latitude direction).", "Západní polokoule (směr zem. délky).",
"When specified, this setting overrides the subdirectory rules (except the base directory) defined in the DVR configuration and puts all recordings done by this entry into the subdirectory named here. See Help for more info.", "Pokud je povoleno, toto nastavení přepíše pravidla podadresářů (kromě základního adresáře) definovaných v konfiguraci DVR a umístí všechny záznamy provedené touto položkou do zde nastaveného podadresáře. Další informace naleznete v nápovědě.",
"When the fulltext is checked, the title pattern is matched against title, subtitle, summary and description.", "Pokud je fulltext zaškrtnutý, šablona vložena v názvu je porovnána s názvem, podtitulkem, souhrnem a popisem.",
"When the service was first identified and recorded.", "Čas, kdy byla služba poprvé nalezena a uložena.",
"When the service was last seen during a mux scan.", "Čas, kdy byla služba naposledy viděna při skenu muxu.",
"Winter sports", "Zimní sporty",
"With \"RAM only\" enabled, and when \"Maximum RAM size\" is reached, remove the oldest segment in the buffer instead of replacing it completely. Note, this may reduce the amount of rewind time.", "Pokud je zapnuto \"Pouze RAM\" a je dosažena \"Maximální velikost RAM\", odstraní nejstarší segment ve vyrovnávací paměti namísto jejího úplného nahrazení. Poznámka: může se tím snížit doba převíjení zpátky.",
"Wizard", "Pomocník",
"Write a short message to log identifying the matched parameters. It is useful for debugging your setup or structure of incoming streams.", "Zapíše krátkou zprávu do logu identifikující porovnávané parametry. Toto je užitečné pro ladění nastavení nebo pro získání struktury streamů.",
"Writes the current in-memory EPG database to disk every x hours, so should a crash/unexpected shutdown occur EPG data is saved periodically to the database (re-read on next startup). Set to 0 to disable.", "Zapíše současnou databázi EPG z paměti na disk každých x hodin, takže pokud dojde k pádu/neočekávanému vypnutí, data EPG jsou periodicky uložena do databáze (a znovu načtena při dalším spuštění). Nastavte 0 pro zakázání.",
"Writes the current in-memory EPG database to disk shortly after an xmltv import has completed, so should a crash/unexpected shutdown occur EPG data is saved (re-read on next startup).", "Zapíše současnou databázi EPG z paměti na disk krátce po dokončení importu xmltv, takže pokud dojde k pádu/neočekávanému vypnutí, data EPG jsou uložena (a znovu načtena při dalším spuštění).",
"XMLTV EPG Import", "XMLTV EPG import",
"Yes", "Ano",
"addpids/delpids supported", "podpora addpids/delpids",
"hrs", "hod",
"invalid option specified [%s]", "specifikována chybná volba [%s]",
"mins", "min",
"option %s requires a value", "volba %s vyžaduje argument",
"tsfile input (mux file)", "tsfile vstup (mux soubor)",
"tvhdhomerun client", "Klient tvhdhomerun",
NULL, NULL
};

static const char *tvh_locale_7[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nFor mere information besøg Tvheadend web-siden:\nhttps://tvheadend.org\n",
"# Mapped channels", "Kortlagte Kanaler",
"# Muxes", "# Mux'er",
"# Services", "# Tjenester",
"# tune repeats", "# indstilling, gentagelse",
"(Default profile)", "(Standard-profil)",
"1 day", "1 dag",
"1 month", "1 måned",
"1 week", "1 uge",
"1 year", "1 år",
"2 months", "2 måneder",
"2 weeks", "2 uger",
"2 years", "2 år",
"3 days", "3 dage",
"3 months", "3 måneder",
"3 weeks", "3 uger",
"3 years", "3 år",
"5 days", "5 dage",
"6 months", "6 måneder",
"AC-3 detection", "AC-3 Registrering",
"ACTIVE", "AKTIV",
"ANY", "ENHVER",
"ATSC source ID", "ATSC kilde ID",
"Aborted by user", "Afbrudt af bruger",
"Accept remote subscription weight", "tillad fjern-abonnementens prioritet ",
"Accept the remote subscription weight (from the SAT>IP client).", "Tillad fjern-abonnementens prioritet (fra SAT>IP klient)",
"Access", "Adgang",
"Action", "Aktion",
"Adapter in use by another subscription", "Adapter i brug af anden abonnement",
"Add file and line numbers to debug", "Tilføj fil og linje numer til fejlsøgning",
"Add the thread ID to debug", "Tilføj tråd-ID til fejlsøgning",
"Additional arguments to pass to the grabber.", "Yderligere kommandoer, der sendes videre til grabber.",
"Additional service names found in EPG data.", "Yderligere tjeneste navne fundet i EPG data.",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "Yderligere tid (i sekunder) tuner skal bruge for at være klar til at optage. Dette bruges hvis din tuner tager længere tid end normalt.",
"Admin password", "Admin kodeord",
"Admin username", "Admin brugernavn",
"Adult movie", "Voksen film",
"Advanced", "Avanceret",
"Advanced (non-universal LNBs, rotors, etc.)", "Avanceret (ikke-universelle LNB'ere, rotorer, mm.)",
"Adventure", "Eventyr",
"Advertisement / Shopping", "Reklame / Indkøb",
"All (Streaming plus DVR)", "Alt (Streaming + DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Alle karakterer som kunne give problemer ved navngivning af filen, vil blive erstattet af en _. Se hjælp for detaljer",
"All lower-case", "Kun små bogstaver",
"All proceeds are used to support server infrastructure and buy test equipment.", "Alle tilskud anvendes til at forbedre server infrastruktur og købe testudstyr.",
"Allowed DVR profiles. This limits the profiles the user has access to.", "Tillad DVR profiler. Dette begrænser de profiler brugeren har adgang til.",
"Allowed network", "Tilladte netværk",
"Allowed networks", "Tilladte netværk",
"Alternate PID path", "Alternativ PID sti",
"Alternate configuration path", "Alternativ konfigurationssti",
"Altitude (in meters).", "Højde (i meter).",
"Altitude (meters)", "Højde (Meter)",
"Animals", "Dyr",
"Anonymize HTSP access", "Anonymiser HTSP adgang",
"Any", "Enhver",
"Arts", "Kunst",
"Assign predefined muxes to networks", "Tilføj forudindtastede mux til netværk",
"Associate this device with one or more networks.", "Forbind denne enhed med mere et eller flere netværk.",
"Athletics", "Atletik",
"Audio bitrate (kb/s) (0=auto)", "Lyd bitrate (kb/s) (0=auto)",
"Audio channel layout.", "Lyd kanal layout.",
"Audio codec", "Lyd codec",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Lyd codec der bruges til transcoding. \"Brug ikke\" vil slå lyden fra.",
"Auto check disabled", "Autocheck Deaktiveret",
"Auto check enabled", "Autocheck Aktiveret",
"Auto record", "Auto-optagelse",
"Auto record caption", "Auto-optag undertekster",
"Auto recording%s%s", "Auto-optagelse%s%s",
"Auto time record", "Automatisk tidsoptagelse",
"Auto-Map to channels", "Kortlæg Kanaler Automatisk ",
"Automatic checking", "Automatisk kontrol",
"Automatic recording caption.", "Automatisk optage navngiving.",
"Automatic schedule entry not found", "Automatisk tidsplan for denne post blev ikke fundet",
"Automatically map EPG source", "Kortlæg EPG-kilde Automatisk ",
"Automatically map all available services to channels.", "Automatisk tilføj alle tilgængelige tjenester til kanaler.",
"Automatically name from network", "Navngiv automatisk fra netværk",
"Automatically record.", "Automatisk optagelse.",
"Autorec maximum count (0=unlimited)", "Antal automatiske optagelser (0=ubegrænset)",
"Autorec maximum schedules limit (0=unlimited)", "Maximum antal planlagte automatiske optagelser (0=ubegrænset)",
"Bad", "Dårlig",
"Bad request", "Dårlig anmodning.",
"Bandwidth", "Båndbredde",
"Based on software from", "Baseret på software fra",
"Basic", "Grundlæggende",
"Bind to specific local IP address.", "Bind til specifik lokal IP-adresse.",
"Brand", "Mærke",
"Broadcast", "Udsende",
"Broadcasting", "Udsendelse",
"Build", "Version",
"CA client %i", "CA klient %i",
"CA identification", "CA-identifikation",
"CA provider", "CA-udbyder",
"CAPMT interval (ms)", "CAPMT-interval (ms)",
"CAPMT query interval (ms)", "CAPMT Forespørgselsinterval (ms)",
"CRID authority", "CRID-autoritet",
"Cache scheme", "Cache skema",
"Camd.socket filename / IP Address (TCP mode)", "Camd.socket filnavn / IP Addresse (TCP mode)",
"Cartoons", "Tegnefilm",
"Channel", "Kanal",
"Channel does not exist", "Kanalen eksistere ikke",
"Channel icon", "Kanal-ikon",
"Channel icon URL.", "Kanal ikon URL",
"Channel icon as defined in EPG data.", "Kanal ikon er defineret i EPG data.",
"Channel icon name scheme", "Kanal ikon navne skema",
"Channel mapping options", "Kanal kortlægnings indstillinger",
"Channel name", "Kanalnavn",
"Channel not found", "Kanal ikke fundet",
"Channel number", "Kanalnummer",
"Channel number as defined in EPG data.", "Kanal nummer er defineret i EPG data.",
"Channel number offset", "Kanalnummer-forskydning",
"Channel numbers (heuristic)", "Kanalnumre (heuristiske)",
"Channel numbers from", "Kanalnumre fra",
"Channel tag", "Kanal gruppe",
"Channel tag does not exist", "Kanal gruppen findes ikke",
"Channel tag reference", "Kanal gruppe reference",
"Channel tags", "Kanal etiketter",
"Channel tags the user is allowed access to/excluded from.", "Kanal etiketter, som brugeren har/ikke har adgang til.",
"Channel to use/used for the recording.", "Kanal der bruges/brugt til optagelse.",
"Channels", "Kanaler",
"Channels EPG data is used by.", "Kanal EPG data bliver brugt af.",
"Character set", "Karaktersæt",
"Check availability", "Kontroller tilgængelighed",
"Children's / Youth programs", "Børn og Unge-programmer",
"Cinema", "Biograf",
"Class", "Klasse",
"Classical", "Klassisk",
"Classical music", "Klassisk musik",
"Clean effects", "Ryd effekter",
"Client name", "Klient navn",
"Clone scheduled entry on error", "Kopiér Planlagt Post ved Fejl",
"Comedy", "Komedie",
"Command delay time (ms) (10-200)", "Kommando forsinkelsestid (ms) (10-200)",
"Command time (ms) (10-100)", "Kommandotid (ms) (10-100)",
"Comment", "Kommentar",
"Commercial break", "Reklamepause",
"Committed", "Tilføjet",
"Completed OK", "Færdiggjort OK",
"Compress EPG database", "Komprimer EPG database",
"Configuration", "Konfiguration",
"Configuration ID", "Konfigurations-ID",
"Configuration name", "Konfigurations navn",
"Configuration version", "Konfigurations-version",
"Connection limit type", "Forbindelses begrænsnings type",
"Constellation", "Opstilling",
"Container", "Beholder",
"Content character set", "Indhold, karaktersæt",
"Content reference identifier authority.", "Indholds reference identifikations autoritet",
"Content type", "Indholdstype",
"Content type.", "Indholdstype.",
"Contest", "Konkurrence",
"Cookie expiration (days)", "Cookie udløbstid (dage)",
"Cooking", "Madlavning",
"Copy codec type", "kopi-codec-type",
"Copy layout", "Kopiér Layout",
"Create SDTV/HDTV/Radio tags.", "Opret SDTV/HDTV/Radio etiketter.",
"Create a directory per channel when storing recordings. If both this and the 'directory per day' checkbox is enabled, the date-directory will be the parent of the per-channel directory.", "Opret en mappe pr. kanal, når optagelsen gemmes. Hvis både denne og \"mappe pr. dag\" tjekboks er valgt, vil \"mappe pr. dag\" være hoved mappen.",
"Create a provider name tag.", "Opret en gruppe for udbydere",
"Create and associate a network tag to created channels.", "Opret og knyt et netværk til gruppen.",
"Create and associate a provider tag to created channels.", "Opret og knyt en udbyder til de oprettede kanaler.",
"Create bouquet", "Opret Buket",
"Create network name tags", "Lav en etikette til netværk",
"Create network name tags (set by provider).", "Opret etiketter til netværk (bestemt af udbyder).",
"Create network tags", "Opret netværks etiketter",
"Create provider name tags", "Lav en etikette til udbydere",
"Create provider tags", "Opret udbyder etiketter",
"Create tags", "Opret etiketter",
"Created", "Oprettet",
"Creator", "Opretter",
"Cron multi-line", "Cron, flerliniet",
"Culture (without music)", "Kultur (eksl. Musik)",
"Current RTSP port.", "Aktuelle RTSP port.",
"Current affairs", "Aktuelt",
"Current tuner configuration.", "Aktuelle tuner konfiguration",
"Custom HTTP headers", "Brugerdefinerede HTTP headere",
"DAB multiplex", "DAB Multiplex",
"DBus - use the session message bus instead of the system one", "DBus - brug 'session message bus' istedet for systemets",
"DES key", "DES nøgle",
"DSCP/TOS for streaming", "DSCP/TOS til streaming",
"DVB scan files", "DVB Skan Filer",
"DVB scan files path", "DVB Skan fil-stier",
"DVR behavior", "DVR-politik",
"DVR configuration", "DVR Konfiguration",
"DVR configuration profiles", "DVR Konfigurationsprofil",
"DVR entry not found", "DVR optagelse ikke fundet",
"DVR file retention period", "DVR filens levetids periode",
"DVR log retention", "DVR loggens levetid",
"DVR log retention period", "DVR loggens levetids periode",
"DVR override: high", "DVR override: høj",
"DVR override: important", "DVR override: vigtig",
"DVR override: low", "DVR override: lav",
"DVR override: unimportant", "DVR override: ligegyldig",
"DVR profile to use/used for the recording.", "DVR profil der skal bruges til denne optagelse.",
"DVR schedule does not exist", "DVR plan findes ikke.",
"DVR schedule not found", "DVR plan ikke fundet",
"Dance", "Dans",
"Data errors", "Datafejl",
"Data path (if applicable).", "Data sti (hvis relevant).",
"Days of Week", "Ugedage",
"Days of week", "Ugedage",
"Debate", "Debat",
"Debug libav log", "Fejlfind libav log",
"Debug options", "Fejlfindings-indstillinger",
"Debug subsystems", "Fejlfind undersystemer",
"Debug to syslog", "Fejlfind til syslog",
"Debug trace (low-level)", "Fejlfind, trace (low-level)",
"Debugging", "Debugge",
"Default", "Standard",
"Default language(s)", "Standardsprog",
"Default language.", "Standard sprog.",
"Default priority", "standards-prioritering",
"Default user interface level.", "Standard brugergrænseflade niveau.",
"Delivery system", "Leveringssytem",
"Demux path", "Demux-sti",
"Descramble services (limit per mux)", "Afkodnings tjeneste (begræns per mux)",
"Descrambler buffer (TS packets)", "Descramble-buffer (TS-pakker)",
"Description", "Beskrivelse",
"Detective", "Detektiv",
"Device ID", "Enheds-ID",
"Device model", "Enheds-model",
"Device path", "Enheds-sti",
"Differentiated Services Code Point / Type of Service: Set the service class Tvheadend sends with each packet. Depending on the option selected this tells your router the prority in which to give packets sent from Tvheadend, this option does not usually need changing. See https://en.wikipedia.org/wiki/Differentiated_services for more information. ", "Differentiated Services Code Point / Type of Service : Indstil tjenesten klassen Tvheadend sender med hver pakke . Afhængigt af den valgte indstilling, fortæller dette din router den PRIORITERET som den skal give pakker, der sendes fra Tvheadend, denne indstilling behøver man normalt ikke at ændre. Se https://en.wikipedia.org/wiki/Differentiated_services for yderligere information.",
"Directory", "Mappe",
"Directory permissions (octal, e.g. 0775)", "Mappe-tilladelser (octal, f.eks. 0775)",
"Disable", "Deaktiver",
"Disable DVB bouquets", "Deaktiver DVB bouquets",
"Disable SAT>IP client", "Deaktiver SAT>IP-klient",
"Disable all access control checks", "Deaktiver alt adganskontrol",
"Disable debug on stderr", "Deaktiver fejlsøgning på stderr",
"Disable device/firmware-specific workarounds", "Deaktiver enheds/firmware-specifikke løsninger",
"Disable syslog (all messages)", "Deaktiver syslog (alle meddelelser)",
"Disabled", "Deaktiveret",
"Discussion", "Diskussion",
"DiseqC repeats", "DiseqC gentagelser",
"Do not use", "Brug ikke",
"Documentary", "Dokumentar",
"Don't backup configuration tree at upgrade", "Lav ikke backup af configurations-træet ved opgradering",
"Don't include title in filename", "Inkludér ikke titel i filnavn",
"Don't keep", "Behold ikke",
"Don't re-record", "Genoptag ikke",
"Don't re-schedule if recording fails.", "Planlæg ikke optagelsen igen, hvis den fejler.",
"Don't reschedule", "planlæg ikke igen",
"Don't use the provider's channel numbers.", "Brug ikke udbyderens kanal numre.",
"Duplicate handling", "Duplikér handling",
"EIT - skip TSID check", "EIT - skip TSID-check",
"EPG name", "EPG-navn",
"EPG scan", "EPG-scan",
"EPG source", "EPG-kilde",
"EPG update window", "EPG-opdateringsvindue",
"Economics", "Økonomi",
"Education", "Uddannelse",
"Educational", "Uddannelses",
"Elementary stream filter", "Elementær streamfilter",
"Enable (auto)", "Slå til (auto)",
"Enable DBus", "Slå DBus til",
"Enable NTP driver", "Slå NTP driver til",
"Enable coredumps for daemon", "Slå coredumps for daemon til",
"Enable debug on stderr", "Slå debug ved stderr til",
"Enable debug subsystems", "Slå debug af undersystemer til",
"Enable debug to file", "Slå debug til fil til",
"Enable debug to syslog", "Slå debug til syslog til",
"Enable if the SAT>IP box requests plts=on parameter in the SETUP RTSP command for DVB-S2 muxes.", "Aktiver hvis SAT>IP boksen efterspørger plts=on parameteren i SETUP RTSP kommandoen til DVB-S2 mux'er.",
"Enable if the SAT>IP box requires pids=21 parameter in the SETUP RTSP command", "Aktiver hvis SAT>IP boksen efterspørger pids=21 parameteren i SETUP RTSP kommandoen.",
"Enable if the SAT>IP box supports the addpids/delpids commands.", "Aktiver hvis SAT>IP boksen understøtter addpids/delpids kommandoer.",
"Enable or disable fast input switching.", "Aktiver eller deaktiver hurtig skift af indgang.",
"Enable or disable full mux mode.", "Aktiver eller deaktiver full mux tilstand.",
"Enable or disable ignoring of Event Information Table (EIT) data on this mux.", "Aktiver eller deaktiver ignorere \"Event Information Tabel\" (EIT) data på denne mux.",
"Enable or disable this configuration.", "Aktiver eller deaktiver denne konfiguration.",
"Enable syslog", "Aktiver syslog",
"Enable system time updates. This will only work if the user running Tvheadend has rights to update the system clock (normally only root).", "Aktiver system tids opdateringer. Dette vil kun virke hvis brugeren, som kører Tvheadend har rettighederne til at opdatere systemets tid (normalt kun \"root\").",
"Enable this filter.", "Aktiver dette filter.",
"Enable trace subsystems", "Slå trace undersystemer til",
"Enable web UI debug (non-minified JS)", "Slå web UI debug (non-minified JS) til",
"Enable/disable LNA.", "Aktiver/deaktiver LNA.",
"Enable/disable auto-rec rule.", "Aktiver/deaktiver auto-optagelses reglen.",
"Enable/disable logging to syslog.", "Aktiver/deaktiver logning til syslog.",
"Enable/disable profile.", "Aktiver/deaktiver profilen.",
"Enable/disable the channel.", "Aktiver/deaktiver kanalen.",
"Enable/disable the device.", "Aktiver/deaktiver enheden.",
"Enable/disable the tag.", "Aktiver/deaktiver gruppen.",
"Enable/disable tuner/adapter.", "Aktiver/deaktiver tuner/adapter.",
"Enabled", "Slået til",
"Encrypted", "Krypteret",
"Encrypted service", "Krypteret tjeneste",
"Entertainment programs for 10 to 16", "Underholdningsprogram (10 - 16)",
"Entertainment programs for 6 to 14", "Underholdningsprogram (6 - 14)",
"Environment", "Miljø",
"Environment (pipe)", "Miljø (pipe)",
"Error code", "Fejlkode",
"Errors", "Fejl",
"Even key", "Lige nøgle",
"Every day", "Hver dag",
"Exclude channel tags", "Udeluk Kanal etiketter",
"Expeditions", "Ekspeditioner",
"Experimental film", "Eksperimental-film",
"Expert", "Ekspert",
"Exported tuners", "Eksporterede tunere",
"External", "Ekstern",
"External EPG grabber", "Ekstern EPG-grabber",
"External IP (NAT)", "Ekstern IP (NAT)",
"External URL", "Ekstern URL",
"Extra arguments", "Ekstra agumenter",
"Extra start time", "Ekstra starttid",
"Extra stop time", "Ekstra stoptid",
"Extra warming up time (seconds)", "Extra opvarmnings tid (sekunder)",
"FEC high", "FEC høj",
"FEC low", "FEC lav",
"Factual topics", "Faktulle emner",
"Failed", "Fejlede",
"Fast input switch", "Hurtig-indgangsskift",
"File missing", "Manglende fejl",
"File not created", "Fil ikke oprettet",
"File permissions (octal, e.g. 0664)", "Filtilladelser (octal, f.eks. 0664)",
"File size", "Fil størrelse",
"Filename", "Filnavn",
"Filename character set", "Filnavn karaktersæt",
"Filename options", "Filnavns muligheder",
"Fine arts", " Æstetik",
"Finished", "Slut",
"Fitness and health", "Fitness og sundhed",
"Folkloric", "Folklore",
"Football", "Fodbold",
"Force (auto)", "Tving (auto)",
"Force CA ID (e.g. 0x2600)", "Tving CA ID (f.eks. 0x2600)",
"Force old status", "Fremtving gammel status",
"Force pilot for DVB-S2", "Fremtving pilot for DVB-S2",
"Force priority", "Fremtving prioritet",
"Force service type to 1", "Tving tjeneste type til 1",
"Force teardown delay", "Fremtving teardown-forsinkelse",
"Force usage of entered CA ID on this service.", "Tving CA ID til at bruges den indtastede værdi, for denne tjeneste.",
"Foreign countries", "Andre lande",
"Forever", "For altid",
"Fork and run as daemon", "Opsplit og kør som daemon",
"Format string", "Formaterings streng",
"Found muxes", "Fundet mux'ere",
"Found services", "Fundet tjenester",
"Free subscription weight", "Ledig abonnements prioritet",
"Free-format text field. Enter whatever you like here.", "Frit formaterings tekst felt. Indtast lige hvad du har lyst til",
"Frequency", "Frekvens",
"Frequency (Hz)", "Frekvens (Hz)",
"Frequency (MHz)", "Frekvens (MHz)",
"Frequency (kHz)", "Frekvens (kHz)",
"Fri", "Fre",
"Friendly name", "Nemt navn",
"Frontend number", "Frontend nummer",
"Frontend path", "Frontend sti",
"Full mux RX mode supported", "Fuld mux RX mode understøttet",
"Full pathname specification", "Fuld stinavns-specifikation",
"Full-text", "Fuldtekst",
"Further education", "Videre uddannelse",
"Game show", "Gameshow",
"Gardening", "Havebrug",
"General", "Generelt",
"General configuration", "Generel konfiguration",
"Generic", "Generisk",
"Generic options", "Generisk valg",
"Guard interval", "Vagtinterval",
"HTSP Default Stream Settings", "HTSP Standardindstillinger",
"HTTP CORS origin", "HTTP CORS oprindelse",
"Handicraft", "Håndværk",
"Hearing impaired", "Hørehæmmede",
"Hierarchy", "Hierarki",
"High", "Høj",
"High bitrate mode (CI+ CAMs only)", "Høj bitrate mode (kun CI+ CAM)",
"Highest channel number the user can access.", "Højeste kanal nummer brugeren kan få adgang til.",
"Historical movie", "Historisk Film",
"Horror", "Horro",
"Hostname/IP", "Hostnavn/IP",
"IP address", "IP-adresse",
"IP blocking", "IP-blokering",
"IPTV input", "IPTV indgang",
"Icon", "Ikon",
"Icon (full URL)", "Ikon (fuld URL)",
"Icon URL", "Ikon URL",
"Icon base URL", "Ikon Basis-URL",
"Icon has title", "Ikon har Titel",
"Icons from", "Ikoner fra ",
"Idle scan", "Idle-skan",
"Idle scan muxes", "Idle skan muxes",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Hvis ingen brugerkonto eksisterer, opret een uden\nbrugernavn og password. Brug med omtanke, idet\nden vil tillade fulde brugerrettigheder til din \nTvheadend-installation indtil du opretter eller retter\nadgangskontrollen fra Tvheadend webinterfacet.",
"If you'd like to support the project, please consider a donation.", "Hvis du gerne vil støtte projektet, så overvej venligst en donation",
"Ignore EPG (EIT)", "Ignorer EPG (EIT)",
"Ignore descriptor 5", "Ignorér descriptor 5",
"Ignore encryption flag, include encrypted services anyway.", "Ignorere krypterede flag, inkl. krypterede tjeneste alligevel.",
"Ignore invalid SSL certificate", "Ignorér ugyldigt SSL-certifikat",
"Ignore provider's channel numbers", "Ignorér udbyders kanalnumre",
"Immediately abort", "Stop øjeblikkeligt",
"Important", "Vigtigt",
"Include channel name in filename", "Inkudér kanalnavn i filnavn",
"Include date in filename", "Medtag dato i filnavn  ",
"Include episode in filename", "Inkludér episode i filnavn",
"Include subtitle in filename", "Inkludér undertekster i filnavn",
"Include time in filename", "Medtag klokkeslæt i filnavn",
"Index", "Indeks",
"Information area", "Informationsområde",
"Informational", "Informerende",
"Initial scan", "Først skanning",
"Input base", "Indgangs bund",
"Input buffer (bytes)", "Indgangs buffer (bytes)",
"Input instance", "Indgangs instans",
"Input path", "Indgangs sti",
"Internal", "Internt",
"Internal EPG grabber", "intern EPG-grabber",
"Internal grabber", "Intern grabber",
"Invalid", "Ugyldig",
"Invalid adapter number '%s'\n", "Ugyldig adapternummer '%s'\n",
"Invalid target", "Ugyldigt mål",
"Keep", "Behold",
"Kill signal (pipe)", "Dræb signal (pipe)",
"Kill timeout (pipe/secs)", "Dræb timeout (pipe/sek)",
"LNA (low noise amplifier)", "LNA (støjsvag forstærker)",
"Language", "Sprog",
"Language settings", "Sprogindstillinger",
"Languages", "Sprog",
"Last seen", "Sidst set",
"Last updated from", "Sidst opdateret fra",
"Leisure hobbies", "Hobbyer",
"Limit connections", "Begræns forbindelser",
"Linked input", "Forbundet Indgang",
"LinuxDVB adapter", "Linux DVB-adapter",
"LinuxDVB network", "Linux DVB-netværk",
"Listen / Connect port", "Listen / Connect-port",
"Listen on IPv6", "Lyt på IPv6",
"Literature", "Literatur",
"Local bind IP address", "Local bind IP-adresse",
"Local channel minor", "Lokal mindre kanal",
"Local channel number", "Lokalt kanalnummer",
"Local discovery IP address", "Lokal discovery IP-adresse",
"Location", "Lokation",
"Location details of the SAT>IP Server.", "Placerings detaljer fra SAT>IP serveren.",
"Lock preferred CA PID", "Lås foretrukket CA PID",
"Login/Logout", "Login/logout",
"Low", "Lav",
"Lowest channel number the user can access.", "Laveste kanal nummer brugeren kan få adgang til.",
"MPEG-TS Pass-thru/built-in", "MPEG-TS Pass-thru/indbygget",
"MPEG-TS input", "MPEG-TS indgang",
"MPEG-TS raw service", "MPEG-TS rå tjeneste",
"Magazines", "Magasiner",
"Maintain free storage space in MiB", "Vedligehold ledig diskplads i MiB",
"Maintain used storage space in MiB (0=disabled)", "Vedligehold brugt diskplads i MiB (0=deaktiveret)",
"Maintained space", "Vedligeholdt plads",
"Make subdirectories per channel", "Opret underfoldere per kanal",
"Make subdirectories per day", "Opret underfoldere per dag",
"Make subdirectories per title", "Opret underfoldere per titel",
"Manufacturer", "Producent",
"Manufacturer URL", "Producent-URL",
"Manufacturer's model number.", "Fabrikantens modelnummer.",
"Manufacturer's product description.", "Fabrikantens produkt beskrivelse.",
"Manufacturer's product information page for the device.", "Fabrikantens produktinformations side for denne enhed.",
"Manufacturer's product name.", "Fabrikantens produkt navn.",
"Map all services", "Tilføj alle tjenester",
"Map encrypted services", "Tilføj krypterede tjenester",
"Map radio channels", "Kortlæg radiokanaler",
"Map unnamed channels", "Kortlæg unavngivne kanaler",
"Map zero-numbered channels", "Kortlæg nul-numererede kanaler",
"Master tuner", "Hovedindstiller",
"Matroska (mkv)/built-in", "Matroska (mkv)/indbygget",
"Matroska specific", "Matroska specifik",
"Maximal channel number", "Maksimalt kanalnummer",
"Maximum # input streams", "Maksimalt # indgangsstreams",
"Maximum PIDs", "Maximal PIDs",
"Maximum RAM size (MB)", "Maximal RAM-størrelse (MB)",
"Maximum bandwidth (Kbps)", "Maximal båndbrede (Kbps)",
"Maximum count (0=default)", "Maximum nummer (0=ubegrænset)",
"Maximum duration", "Maximal varighed",
"Maximum length of PIDs", "Maximal længde af PID'er",
"Maximum period (mins)", "Maximal periode (min)",
"Maximum schedules limit (0=default)", "Maximal planlægning begrænsning (0=standard)",
"Maximum size (MB)", "Maksimal størrelse (MB)",
"Maximum timeout (seconds)", "Maksimal timeout (sekunder)",
"Medicine", "Medicin",
"Merge same name", "Flet ved samme navn",
"Minimal channel number", "Minimal kanalnummer",
"Minimum duration", "Minimal varighed",
"Minimum rotor time (seconds)", "Minimal rotortid (sekunder)",
"Missing In PAT/SDT", "Mangler i PAT/SDT",
"Model description", "Modelbeskrivelse",
"Model name", "Modelnavn",
"Model number", "Modelnummer",
"Module", "Modul",
"Module ID", "Modul-ID",
"Mon", "Man",
"More verbose libav log", "Mere dybdegående libav-log",
"Motor rate (milliseconds/deg)", "Motorrate (millisekunder/grader)",
"Motoring", "Motorstyring",
"Movie", "Film",
"Music", "Musik",
"Mux handling", "Mux-styring",
"Mux name", "MUX-navn",
"Mux not enabled", "Mux slået fra",
"NONE", "INGEN",
"Name", "Navn",
"Name of the tag.", "Navn på gruppen.",
"Names", "Navne",
"Natural sciences", "Naturvidenskab",
"Nature", "Natur",
"Network", "Netværk",
"Network A", "Netværk A",
"Network B", "Netværk B",
"Network ID (limit scanning)", "Netværks-ID (Begrænset Skanning)",
"Network UUID", "Netværks-UUID",
"Network access", "Netværks adgang",
"Network discovery", "Netværks-søgning",
"Network name", "Netværks-navn",
"Network prefix", "Netværks-prefix",
"Network settings", "Netværks indstillinger",
"Network type", "Netværks-type",
"Networks", "Netværk",
"News", "Nyheder",
"News magazine", "Nyhedsmagasin",
"Next tune delay in ms (0-2000)", "Næste indstillings-forsinkelse i ms (0-2000)",
"No", "Nej",
"No PIN", "Ingen PIN",
"No access", "Ingen Adgang",
"No adapters specified!\n", "Ingen Adapter valgt\n",
"No assigned adapters", "Ingen adapter tilføjet",
"No days", "Ingen dage",
"No descrambler", "In descambler",
"No free adapter", "Ingen tilgængelig Adapter",
"No input detected", "Ingen indgang opdaget",
"No service assigned to channel", "Ingen tjeneste tildelt kanal",
"No service enabled", "Ingen tjeneste er aktiveret.",
"No source available", "Ingen Kilde til rådighed",
"None", "Ingen",
"Not linked", "ikke tilknyttet",
"Not set", "Ikke sat",
"Not set (none or channel configuration)", "Ikke sat (none eller kanal-konfiguration)",
"Not set (use channel or DVR configuration)", "Ikke sat (brug kanal eller DVR-konfiguration)",
"Number", "Nummer",
"Number of seconds to wait before timing out.", "Antal sekunder der ventes, før timeout",
"Number of tsfile tuners", "Antal ts-fil-tunere",
"OK (partial)", "OK (delvist)",
"OSCam net protocol (rev >= 10389)", "OSCam netprotokol (rev >= 10389)",
"OSCam new pc-nodmx (rev >= 10389)", "OSCam ny pc-nodmx (rev >= 10389)",
"Odd key", "Ulige Nøgle",
"Off", "Slukket",
"Older OSCam", "Ældre OSCAM",
"On", "Til",
"On file removal", "Ved fjernelse af fil",
"Only Bulsatcom 39E", "Kun Bulsatcom 39E",
"Only EIT", "Kun EIT",
"Only OpenTV Sky Ausat", "Kun OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Kun OpenTV Sky Italia ",
"Only OpenTV Sky UK", "Kun OpenTV Sky UK",
"Only PSIP (ATSC)", "Kun PSIP (ATSC)",
"Only UK Freesat", "Kun UK Freesat",
"Only UK Freeview", "Kun UK Freeview",
"Only Viasat Baltic", "Kun Viasat Baltic",
"Only preferred CA PID", "Kun foretrukket CA PID",
"OpenTV channel number", "OpenTV-kanalnummer",
"Orbital position", "Baneposition",
"Orbital positions", "Bbanepositioner",
"Original network ID", "Oprindeligt NetværksID",
"Output directory for tsdebug", "Destiantionsmappe for tsdebug",
"Over-the-air Cron multi-line", "Over-the-air Cron multi-linie",
"Over-the-air EPG grabber", "Over-the-air EPG-grabber",
"Over-the-air grabbers", "Over-the-air Grabbere",
"Owner", "Ejer",
"PIDs 21 in setup", "PID 21 i opsætning",
"PIN inquiry match string", "PIN forespørgsels-matchstreng",
"PLS code", "PLS-kode",
"PLS mode", "PLS-mode",
"Parent entry", "Foregående Post",
"Pass subscription weight", "Passer abonnement-prioritet",
"Password", "Adgangskode",
"Password2", "Password 2",
"Path", "Sti",
"Performing arts", "Scenekunst",
"Periodically save EPG to disk (hours)", "Gem EPG periodisk til dsik (Timer)",
"Physiology", "Fysiologi",
"Polarization", "Polarisering",
"Political issues", "Politisk Emne",
"Popular culture", "Populær-kultur",
"Position of the input.", "Position af indgang",
"Post-processor command", "Post-processorkommando",
"Post-recording padding", "Post-processor padding",
"Power save", "Strømsparer",
"Power-up time (ms) (15-200)", "Opvågningstid (ms) (15-200)",
"Pre-defined muxes", "Præ-definerede muxer",
"Pre-recording padding", "Fyld før Optagelse",
"Pre-school children's programs", "Førskole børneprogram",
"Prefer picons over channel name", "Foretræk picon'er frem for kanalnavne",
"Preferred CA PID", "Foretræk CA PID",
"Preferred service video type", "Foretrukken tjenestes videotype",
"Presentation", "Præsentation",
"Press", "Presse",
"Priority", "Prioritet",
"Priority (-10..10)", "Prioritet (-10..10)",
"Private", "Privat",
"Profile name", "Profilnavn",
"Provider", "Udbyder",
"Provider ID", "UdbyderID",
"Provider network name", "Udbyderens netværks navn",
"Psychology", "Psykologi",
"Puppets", "Dukker",
"RAM only", "Kun RAM",
"RTSP port (554 or 9983, 0 = disable)", "RTSP Port (554 or 9983, 0=deaktiver)",
"Re-fetch period (hours)", "Gen-hentnings-periode (timer)",
"Re-fetch period (mins)", "Gen-hentnings-periode (Minutter)",
"Re-record%s%s", "Gen-optag%s%s",
"Re-try period (hours)", "Prøv-igen-periode (timer)",
"Real Time Streaming Protocol (RTSP) port the server should listen on (554 or 9983, 0 = disable).", "Real Time Streaming Protocol (RTSP) porten som serveren skal lytte til er (554 eller 9983, 0=deaktiveret).",
"Record all", "Optag alt",
"Recording file options", "Optagelse, fil-muligheder",
"Recording system path", "Optagelses-systemsti",
"Reject", "Afslå",
"Religious", "Reilgiøst",
"Remarkable people", "Bemærkelsesværdige mennesker",
"Remove HTTP arguments", "Fjern HTTp-argumenter",
"Remove all unsafe characters from filename", "Fjern alle farlige karaktere fra filnavn",
"Replace whitespace in title with '-'", "Erstat mellemrum i titel med '-'",
"Reply to CAM PIN inquiries", "Svar på CAM PIN-forespørgsler",
"Reports", "Rapporter",
"Rerun of", "Genudsendelse af",
"Reserved", "Reserveret",
"Resolution (height)", "Opløsning (højde)",
"Restart on error", "Genstart ved fejl",
"Reuse EPG from", "Genbrug EPG fra",
"Rewrite EIT", "Genskriv EIT",
"Rewrite PAT", "Genskriv PAT",
"Rewrite PMT", "Genskriv PMT",
"Rewrite SDT", "Genskriv SDT",
"Romance", "Romantik",
"Rotor initialization time (seconds)", "Rotor opstartstid (seconds)",
"Run as group", "Kør som gruppe",
"Run as user", "Kør som user",
"Running", "Kører",
"SAT->IP frontend number.", "SAT->IP frontend nummer.",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "SAT>IP RTSP portnummer for server\n(default: -1=deaktiveret 0=webconfig, standard port er 554)",
"SAT>IP source number", "SAT>IP kildenummer",
"SSL verify peer", "SSL godkend peer",
"Sat", "Lør",
"Satellite longitude", "Satellit længdegrad",
"Satellite positions", "Satellitpositioner",
"Scan after creation", "Skan efter oprettelse ",
"Scan queue length", "Skannings-kø, længde",
"Scan result", "Skannings-resultat",
"Scan status", "Skanningsstatus",
"Schedule a re-recording if more errors than (0=off)", "Planlæg genoptagelse hvis flere fejl end (0=fra)",
"Schedule status", "Planlægnings status",
"Schedule status.", "Planlægnings status.",
"Scheduled for recording", "Planlagt optagelse",
"Scheduled start time", "Planlagt start",
"Scheduled stop time", "Planlagt sluttid",
"School programs", "Skoleprogrammer",
"Science", "Videnskab",
"Science fiction", "Science Fiction",
"Season", "Sæson",
"Season %d", "Sæson %d",
"Send CAPMT query", "Send CAPMT-forespørgsel",
"Send full PLAY cmd", "Send fuld PLAY cmd",
"Serial number", "Serielnummer",
"Series link", "Serie-link",
"Serious", "Alvorlig",
"Serious music", "Alvorlig musik",
"Server connectivity", "Server-forbindelse",
"Server details.", "Server detaljer.",
"Service", "Tjeneste",
"Service ID", "Tjeneste ID",
"Service configuration", "Tjeneste konfiguration",
"Service count %zi", "Tjeneste tæller %zi",
"Service mapping", "Tjeneste kortlægning",
"Service name", "Tjeneste navn",
"Service name found in EPG data.", "Tjeneste navn fundet i EPG data.",
"Service name picons", "Tjeneste navn picons",
"Service raw", "tjeneste, rå",
"Service type", "Tjeneste type",
"Services", "Tjenester",
"Set the display name.", "indtast vist navn.",
"Settings", "Indstillinger",
"Show this page", "Vis denne side",
"Show version information", "Vis versions-information",
"Signal quality too poor", "Signalkvalitet for lav",
"Signal scale (240 or 100)", "Signalskala (240 eller 100)",
"Site latitude", "Lokations-breddegrad",
"Site longitude", "Lokations-længdegrad",
"Skip TS packets (0-200)", "Skip TS-pakker (0-200)",
"Skip commercials", "Skip reklamer",
"Skip initial bytes", "Skip første bytes",
"Slave entry", "Slave-entry",
"Slot state", "Slot-status",
"Soccer", "Fodbold",
"Social advisory", "Socialrådgivning",
"Sort index", "Sortér indeks",
"Source", "Kilde",
"Source deleted", "Kilde fjernet",
"Source quality is bad", "Kilde-signal er dårligt",
"Source reconfigured", "Kilde genkonfigureret",
"Southern hemisphere (latitude direction)", "Sydlige halvkugle (breddegrads-retning)",
"Special events (Olympic Games, World Cup, etc.)", "Specielle begivenheder (Olympiske Lege, WM, osv.)",
"Specify User-Agent header for the http client", "Specificér UserAgent-header for http-klienten",
"Specify alternative htsp port", "Specificér alternativ HTSP-port",
"Specify alternative http port", "Specificér alternativ HTTP-port",
"Specify alternative http webroot", "Specificér alternativ HTTP-webroot",
"Specify bind address", "Specificér bind-adresse",
"Specify extra htsp port", "Specificér ekstra HTSP-port",
"Spiritual sciences", "Sprituel videnskab",
"Sports", "Sport",
"Sports magazines", "Sportsmagasiner",
"Start after", "Start efter",
"Start before", "Start før",
"Start time", "Starttid",
"Status period (ms)", "Statusperiode (ms)",
"Stop time", "Sluttid",
"Storage path", "Lagrings-sti",
"Storage space", "Diskplads",
"Stream index", "Stream-index",
"Stream profile", "Stream-profile",
"Stream type", "Stream-type",
"Streaming priority", "Streamings-prioritet",
"Streaming profiles", "Streaming-profiler",
"Subdirectory options", "Undermappe-indstillinger",
"Subscribe to a service permanently", "Abonner på en tjeneste permanent",
"Subscription overridden", "Abonnement overskrevet",
"Subscription weight", "Abonnement-prioritet",
"Substitute formatters", "Substitutions-formatteringer",
"Subtitle", "Undertekst",
"Subtitle codec", "Undertekst-codec",
"Sun", "Søn",
"Switch before rotor", "Skift før rotor",
"Switch type", "Switch-type",
"Symbol rate (Sym/s)", "Symbolrate (Sym/s)",
"Sync + Don't keep", "Sync + Behold ikke",
"Tag files with metadata", "Tilføj metadata til filer",
"Tags", "etiketter",
"Talk show", "Talkshow",
"Team sports (excluding football)", "Teamsport (eksklusiv fodbold)",
"Technology", "Teknologi",
"Testing", "Test",
"Testing options", "Test-indstillinger",
"The OpenTV channel number as set by the provider.", "OpenTV kanal nummer, fastsat af udbyderen.",
"The current boot ID.", "Nuværende boot ID.",
"The current configuration ID.", "Nuværende konfigurations ID.",
"The current configuration version.", "Nuværende konfigurations version.",
"The default language to use if the user  language isn't set in the Access Entries tab.", "Standard sprog, som bruges hvis brugerens sprog ikke er sat i Adgangsposteringer fanebladet",
"The default subscription weight for each subscription.", "Standard abonnements prioritet for hver abonnement.",
"The device ID.", "Enhedens ID.",
"The device's serial number.", "Enhedens serienummer.",
"The frequency of the mux (in Hertz).", "Mux'ens frekvens (i Hertz).",
"The maximum number of recordings that can be scheduled.", "Maximale antal optagelser som kan planlægges.",
"The mux the service is on.", "Tjenesten er placeret på mux.",
"The mux's universally unique identifier.", "Mux'ens universelle entydige ID.",
"The network the service is on.", "Netværket som kanalen er på.",
"The networks using this configuration.", "Netværket bruger denne konfiguration.",
"The number of ATSC-C (Cable/AnnexB) tuners to export.", "Antal af ATSC-C (Kabel/AnnexB) tunere til eksport.",
"The number of ATSC-T (Terresterial) tuners to export.", "Antal af ASTC-T (Tag-antenne) tunere til eksport.",
"The number of DVB-C (Cable) tuners to export.", "Antal af DVB-C (Kabel-TV) tunere til eksport.",
"The number of DVB-C2 (Cable) tuners to export.", "Antal af DVB-C2 (Kabel-TV) tunere til eksport.",
"The number of DVB-S (Satellite) tuners to export.", "Antal af DVB-S (Satellit) tunere til eksport.",
"The number of DVB-S2 (Satellite) tuners to export.", "Antal af DVB-S2 (Satellit) tunere til eksport.",
"The number of DVB-T (Terresterial) tuners to export.", "Antal af DVB-T (Tag-antenne) tunere til eksport.",
"The number of DVB-T2 (Terresterial) tuners to export.", "Antal af DVB-T2 (Tag-antenne) tunere til eksport.",
"The number of days cookies set by Tvheadend should expire.", "Antallet af dage, før \"cookies\" lavet af Tvheadend, udløber",
"The provider's name.", "Udbyderens navn.",
"The provider's network ID.", "Udbyderens netværks ID.",
"The provider's network name.", "Udbyderens netværks navn.",
"The service ID as set by the provider.", "Tjeneste ID, som er fastsat af udbyderen.",
"The service name as set by the provider.", "Tjenestens navn, som er fastsat af udbyderen.",
"The service's channel number as set by the provider.", "Tjenestens kanal nummer, fastsat af udbyderen.",
"The symbol rate.", "Symbol rate",
"The transport stream ID of the mux within the network.", "Transportstream ID fra mux'en i netværket.",
"The video stream types the filter should apply to.", "Videostream typer som filteret skal bruges på.",
"This tuner", "Denne tuner",
"Thu", "Tor",
"Time", "Tid",
"Time missed", "Tid overskredet",
"Time record caption", "Tids-optag undertekst",
"Time recording%s%s", "Tids-optagelse%s%s",
"Time update", "Opdater tid",
"Timeout (sec) (0=infinite)", "Timeout (sek) (0=uendelig)",
"Timeout (seconds)", "Timeout (sekunder)",
"Timeshift", "Tidsforskydning",
"Title", "Titel",
"Title (regexp)", "Titel (regexp)",
"Toggle details", "Slå detaljer til/fra",
"Total number of mapped channels on this network.", "Samlede antal kortlagte kanaler i dette netværk.",
"Total number of services found on this network.", "Det samlede antal tjenester fundet på dette netværk.",
"Total number of services found.", "Det samlede antal tjenester fundet.",
"Tourism / Travel", "Turisme / Rejser",
"Trace subsystems", "Trace undersystemer",
"Traditional arts", "Traditionel kunst",
"Traditional music", "Traditionel musik",
"Transcoding", "Omkodning",
"Transmission mode", "Transmissions-mode",
"Transponder ID", "TransponderID",
"Transport stream ID", "Transport-streamID",
"Tue", "Tir",
"Tune before DiseqC", "Indstil før DiseqC",
"Tuner bind IP address", "Tuner bind IP-adresse",
"Tuner configuration", "Tuner-konfiguration",
"Tuning failed", "Indstilling fejlede",
"Turn off LNB when idle", "Slå LNB fra ved idle",
"Tvheadend server name", "Tvheadend servernavn",
"UDP RTP port number (2 ports)", "UDP RTP portnumre (2 porte)",
"URL with the SAT>IP server XML location", "URL med SAT>IP server XML-lokation",
"Uncommitted", "Ikke commited",
"Uncommitted first", "Ikke commited først",
"Unicable type", "Unicable-type",
"Unimportant", "Ikke vigtigt",
"Universal LNB only", "Kun universal LNB",
"Universally unique identifier. Read only.", "Universal unik ID. Kun læsning.",
"Unknown", "Ukendt",
"Unknown reason (%i)", "Ukendt grund (%i)",
"Unlimited size", "Uendelig størrelse",
"Unlimited time", "Uendelig tid",
"Update card (EMM)", "Opdatér kort (EMM)",
"Update channel icon", "Opdatér kanalikon",
"Update channel name", "Opdatér kanalnavn",
"Update channel number", "Opdater kanalnummer",
"Update disabled", "Opdatering deaktiveret",
"Update time", "Opdatér tid",
"Update tolerance (ms)", "Opdateringstolerence (ms)",
"Updated", "Opdateret",
"Use EPG running state", "Brug EPG kørende tilstand",
"Use Windows-compatible filenames", "Brug Windows-kompatible filnavne",
"Use XSPF playlist instead of M3U", "Brug XSPF playlist i stedet for M3U",
"Use a pre-defined list of DVB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Brug en foruddefineret liste over DVB-C mux'er. Bemærk: disse lister kan nogle gange være forældet og kan gøre at scanningen tager længere tid end normalt.",
"Use original", "Brug original",
"Use service IDs as channel numbers", "Brug tjeneste ID som kanalnumre",
"Use tag internally (don't expose to clients).", "Brug gruppen internt (vis ikke til klienter).",
"Use the provider's service IDs as channel numbers.", "Brug udbyderens tjeneste ID, som kanal nummer.",
"User access error", "Brugeradgangs-fejl",
"User does not have access", "Brugeren har ikke adgang",
"User does not have access to channel", "Brugeren har ikke adgang til kanalen",
"User icon", "Bruger ikon",
"User interface level", "Bruger niveau",
"User language", "Bruger sprog",
"User limit reached", "Brugergrænse nået",
"User login", "Bruger login",
"Username", "Brugernavn",
"Variety show", "Varietyshow",
"Vertical resolution (height) of the output video stream. Horizontal resolution is adjusted automatically to preserve aspect ratio. When set to 0, the input resolution is used.", "Lodret opløsning (højde) af udgangssignalets videostream. Vandret opløsning justeres automatisk for at bevare billedformatet. Når den er indstillet til 0, bruges indgangens opløsning.",
"Video recorder", "Videooptager",
"Visually impaired commentary/audio description", "Kommentar for svatgseende / audio-beskrivelse",
"WEBM/built-in", "WEBM/indbygget",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV H264/AAC/MPEG-TS-profil",
"WEBTV profile H264/AAC/Matroska", "WEBTV H264/AAC/Matroska-profil",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV VP8/Vorbis/WEBM-profil",
"Waiting for EPG running flag", "Venter på EPG \"running\"-flag",
"Waiting for program start", "Venter på at program starter",
"Waiting for stream", "Venter på stream",
"War", "Krig",
"Water sport", "Vandsport",
"Weak stream", "Svag stream",
"Weather report", "Vejrudsigt",
"Web interface", "Webinterface",
"Web interface language", "Webinterface, Sprog",
"Web user interface", "Webbrugerinterface",
"Wed", "Ons",
"Welcome - Tvheadend - your TV streaming server and video recorder", "Velkommen - TVHeadend - Din TV streaming server og video optager",
"Western", "Vestlig",
"Western hemisphere (latitude direction)", "Vestlig Halvkugle",
"Wizard", "Vejledning",
"Yes", "Ja",
"addpids/delpids supported", "addpids/delpids understøttet",
"hrs", "timer",
"invalid option specified [%s]", "ugyldig paramter angivet [%s]",
"mins", "minutter",
"option %s requires a value", "paramter %s kræver en værdi",
"tsfile input (mux file)", "tsfile indgang (mux fil)",
"tvhdhomerun client", "tvhdhomerun-klient",
NULL, NULL
};

static const char *tvh_locale_8[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nFür mehr Informationen besuchen Sie bitte die Tvheadend Webseite:\nhttps://tvheadend.org\n",
"# Mapped channels", "# Zugeordnete Kanäle",
"# Muxes", "# Muxe",
"# tune repeats", "# Tune Wiederholungen",
"(Default profile)", "(Standard-Profil)",
"(Read-only)", "(Schreibgeschützt)",
"1 day", "1 Tag",
"1 month", "1 Monat",
"1 week", "1 Woche",
"1 year", "1 Jahr",
"2 months", "2 Monate",
"2 weeks", "2 Wochen",
"2 years", "2 Jahre",
"2-Port switch (universal LNB)", "2-Port-Switch (Universal LNB)",
"3 days", "3 Tage",
"3 months", "3 Monate",
"3 weeks", "3 Wochen",
"3 years", "3 Jahre",
"4-Port switch (universal LNB)", "4-Port-Switch (Universal LNB)",
"5 days", "5 Tage",
"6 months", "6 Monate",
"A channel tag (e.g. a group of channels) to which this rule applies.", "Ein Kanal-Tag (z.B. eine Gruppe von Kanälen) auf welche die Regel wirkt.",
"AAC audio", "AAC Audio",
"AC-3 = descriptor 6", "AC-3 = Deskriptor 6",
"AC-3 detection", "AC-3 Erkennung",
"AC3 audio", "AC3 Audio",
"ACTIVE", "AKTIV",
"ANY", "BELIEBIG",
"ATSC SI Tables", "ATSC SI Tabellen",
"ATSC source ID", "ATSC Quelle-ID",
"ATSC-C Network", "ATSC-C Netzwerk",
"ATSC-C multiplex", "ATSC-C Multiplex",
"ATSC-T Network", "ATSC-T Netzwerk",
"ATSC-T multiplex", "ATSC-T Multiplex",
"Aborted by user", "Durch Benutzer abgebrochen",
"Accept remote subscription weight", "Remote Abonnementgewicht zulassen",
"Accept the remote subscription weight (from the SAT>IP client).", "Remote Abonnementgewicht zulassen (von SAT>IP Client)",
"Accept zero value for TSID", "0 für TSID akzeptieren",
"Access", "Zugang",
"Access (ACL)", "Zugang (ACL)",
"Action", "Aktion",
"Active", "Aktiv",
"Adapter in use by another subscription", "Adapter von anderem Abonnement in Verwendung",
"Add file and line numbers to debug", "Füge Datei und Zeilennummern zu Debugmeldungen hinzu",
"Add the thread ID to debug", "Füge Thread-ID zu Debugmeldungen hinzu",
"Add this value to PTS for the teletext subtitles. The time value is in milliseconds and may be negative.", "Diesen Wert zum PTS für die Teletext-Untertitel hinzufügen. Der Zeitwert ist in Millisekunden und kann negativ sein.",
"Additional arguments to pass to the grabber.", "Zusätzliche Argumente zur Übergabe an den Grabber.",
"Additional service names found in EPG data.", "Zusätzliche Service Namen in den EPG Daten.",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "Zusätzliche Zeit (in Sekunden) für den Tuner um bereit für die Aufnahme zu sein. Das ist hilfreich für Tuner, die einige Zeit zum tunen benötigen und/oder am Anfang Mülldaten senden.",
"Admin password", "Admin Passwort",
"Admin username", "Admin Benutzername",
"Administrator login", "Administratorzugang",
"Adult movie", "Erwachsenenfilm",
"Advanced", "Erweitert",
"Advanced (non-universal LNBs, rotors, etc.)", "Erweitert (Nicht-universale LNBs, Rotoren, etc.)",
"Adventure", "Abenteuer",
"Advertisement / Shopping", "Werbung / Shopping",
"After creating the network scan it for services.", "Nach dem Erstellen des Netzwerks die Suche nach Services starten.",
"All (Streaming plus DVR)", "Alle (Streaming + DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Alle Zeichen, die Probleme in Dateinamen machen könnten, werden durch Unterstriche ersetzt. Siehe Hilfe für Details.",
"All lower-case", "Alles wird kleingeschrieben",
"All proceeds are used to support server infrastructure and buy test equipment.", "Alle Erlöse werden dafür genutzt, die Serverinfrastruktur zu unterhalten und Testgeräte zu kaufen.",
"All: Record if different description", "Alle: Aufnehmen, wenn abweichende Beschreibung",
"All: Record if different episode number", "Alle: Aufnehmen, wenn abweichende Episiodennummer",
"All: Record if different subtitle", "Alle: Aufnehmen, wenn abweichender Untertitel",
"All: Record once per day", "Alle: Einmal pro Tag aufnehmen ",
"All: Record once per month", "Alle: Einmal pro Monat aufnehmen",
"All: Record once per week", "Alle: Einmal pro Woche aufnehmen",
"Allow all PIDs", "Erlaube alle PIDs",
"Allow all PIDs (no filter) when the 'Maximum PIDs' limit is reached.", "Erlaube alle PIDs (keine Filter) wenn das \"Maximum PIDs\" Limit erreicht ist.",
"Allow high bitrate mode (CI+ CAMs only).", "Erlaube hohe Bitrate Modus (Nur CI+ CAMs)",
"Allow idle scan tuning on this device.", "Erlaube Idle Scan Tuning auf diesem Gerät.",
"Allow the combined size of all timeshift buffers to potentially grow unbounded until your storage media runs out of space.", "Erlaube der kombinierten Größe aller Timeshift Puffer potentiell unbegrenzt zu wachsen bis das Speichermedium voll ist.",
"Allow the timeshift buffer to grow unbounded until your storage media runs out of space. Warning, enabling this option may cause your system to slow down or crash completely!", "Erlaubt dem Timeshift-Puffer unbegrenzt zu wachsen bis das Speichermedium voll ist. Achtung, das Aktivieren dieser Option kann das System verlangsamen oder komplett abstürzen lassen.",
"Allow/Disallow access to the 'Configuration' tab.", "Erlaube/Verbiete Zugriff auf den \"Konfiguration\" Tab.",
"Allow/Disallow web interface access (this  includes access to the EPG).", "Erlaube/Verbiete Web-Interface Zugriff (inklusive Zugriff auf den EPG)",
"Allowed DVR profiles. This limits the profiles the user has access to.", "Erlaubte DVR Profile. Das limitiert die Profile, auf die der Nutzer Zugriff hat.",
"Allowed network", "Zugelassenes Netzwerk",
"Allowed networks", "Erlaubte Netzwerke",
"Alternate PID path", "Alternativer PID-Pfad",
"Alternate configuration path", "Alternativer Konfigurationspfad",
"Altitude (in meters).", "Höhe (in Meter)",
"Altitude (meters)", "Höhe (Meter)",
"Always send the whole DiseqC sequence including LNB setup (voltage, tone). If this is not checked, only changed settings are sent, which may cause issues with some drivers. If the tuning is not reliable, try activating this option.", "Sende immer die ganze DiseqC Sequenz inklusive LNB Setup (Spannung, Ton). Wenn dies nicht aktiviert ist, werden nur geänderte Einstellungen gesendet, was Probleme mit einigen Treibern verursachen kann. Wenn das Tuning nicht zuverlässig ist, sollte diese Option ausprobiert werden.",
"Always use the name defined by the network.", "Immer den vom Netzwerk vorgegebenen Namen verwenden.",
"Always use the old ioctls to read the linuxdvb status (signal strength, SNR, error counters). Some drivers are not mature enough to provide the correct values using the new v5 linuxdvb API.", "Benutze immer die alten ioctls zum Lesen des Linux DVB Status (Signalstärke, SNR, Fehlerzähler). Einige Treiber sind nicht ausgereift genug um die korrekten Werte über die neue V5 Linux DVB API zurückzugeben.",
"An event which starts between this \"start after\" and \"start before\" will be matched (including boundary values).", "Ein Ereignis, das zwischen \"start nach\" und \" starten, bevor\" startet, stimmt (einschließlich der Grenzwerte ) überein.",
"Animals", "Tiere",
"Anonymize", "Anonymisieren",
"Anonymize HTSP access", "HTSP-Zugriff anonymisieren",
"Any", "Beliebig",
"Arts", "Kunst",
"Arts / Culture (without music)", "Kunst/Kultur (ohne Musik)",
"Arts magazines", "Kunstmagazine",
"Assign predefined muxes to networks", "Den Netzwerken vordefinierte Muxe zuweisen",
"Associate this device with one or more networks.", "Verbinde das Gerät mit einem oder mehreren Netzwerken.",
"Athletics", "Athletik",
"Audio Stream Filter", "Audio Streamfilter",
"Audio bitrate (kb/s) (0=auto)", "Audio-Bitrate (kb/s) (0=Auto)",
"Audio bitrate to use for transcoding.", "Für das Transkodieren zu verwendende Audiobitrate.",
"Audio channel layout.", "Audio-Kanal Layout.",
"Audio codec", "Audio-Codec",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Audio-Codec, der für das Transkodieren genutzt wird. \"Nicht benutzen\" wird Audio-Ausgang deaktivieren.",
"Audio stream", "Audio Stream",
"Audio type", "Audio Type",
"Audio-only stream", "Audio-Only Stream",
"Audioes muxer", "Audio Muxer",
"Authenticated user", "Autorisierter Benutzer",
"Auto check disabled", "Automatische Überprüfung deaktiviert",
"Auto check enabled", "Automatische Prüfung aktiviert",
"Auto record", "Auto-Aufnahme",
"Auto record caption", "Auto-Aufnahme Beschriftung",
"Auto recording%s%s", "Auto-Aufnahme%s%s",
"Auto time record", "Automatische Zeitaufnahme",
"Auto-Map to channels", "Automatische Zuordnung zu Kanälen",
"Automatic checking", "Automatische Überprüfung",
"Automatic recording caption.", "Automatische Aufnahme-Beschriftung",
"Automatic schedule entry not found", "Automatischen Aufnahmeeintrag nicht gefunden",
"Automatic time scheduler entry not found", "Automatische Zeitplaner-Eintrag nicht gefunden",
"Automatically link EPG data to the channel (using the channel name for matching). If you turn this option off, only the OTA EPG grabber will be used for this channel unless you've specifically set a different EPG Source.", "Automatisch die EPG Daten mit dem Kanal verbinden (basierend auf dem Kanalnamen). Wenn Sie diese Option deaktivieren, wird für diesen Kanal ausschließlich der OTA EPG Graber verwendent, solange sie nicht explizit eine andere EPG Quelle angegeben haben.",
"Automatically map EPG source", "EPG Quelle automatisch zuordnen",
"Automatically map all available services to channels.", "Automatisch alle verfügbaren Services zu Kanälen zuordnen.",
"Automatically map channels defined within the bouquet.", "Kanäle gemäß dem Bouqet automatisch zuordnen.",
"Automatically name from network", "Automatisch vom Netzwerk benennen",
"Automatically record.", "Automatisch aufnehmen.",
"Automatically update channel icons using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel icons.", "Senderlogos automatisch mit Hilfe der von den aktivierten EPG-Anbietern bereitgestellten Informationen aktualisieren. Beachte, dass dies ungewünschte Änderungen an bereits festgelegten Senderlogos zur Folge haben kann.",
"Automatically update channel names using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel names.", "Sendernamen automatisch mit Hilfe der von den aktivierten EPG-Anbietern bereitgestellten Informationen aktualisieren. Beachte, dass dies ungewünschte Änderungen an bereits festgelegten Sendernamen zur Folge haben kann.",
"Automatically update channel numbers using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel numbers.", "Sendernummern automatisch mit Hilfe der von den aktivierten EPG-Anbietern bereitgestellten Informationen aktualisieren. Beachte, dass dies ungewünschte Änderungen an bereits festgelegten Sendernummern zur Folge haben kann.",
"Autorec maximum count (0=unlimited)", "Maximale Anzahl Auto-Aufnahmen (0=unlimitiert)",
"Autorec maximum schedules limit (0=unlimited)", "Maximale Auto-Aufnahme-Pläne (0=unlimitiert)",
"Bad", "Schlecht",
"Bad request", "Fehlerhafte Anfrage",
"Ballet", "Ballett",
"Bandwidth", "Bandbreite",
"Base DVB SI Tables (PAT,CAT,PMT,SDT etc.)", "Basis-DVB-SI-Tabellen (PAT,CAT,PMT,SDT etc.)",
"Based on software from", "Basiert auf Software von",
"Basic", "Einfach",
"Bind to specific local IP address.", "An spezifische lokale IP Adresse binden",
"Bitrate to use for the transcode. See Help for details.", "Zu verwendende Bitrate für das Transkodieren. Siehe Hilfe für Details.",
"Blue", "Blau",
"Boot ID", "Boot-ID",
"Bouquet source.", "Bouquet Quelle",
"Brand", "Marke",
"Branding information (if available).", "Markeninformationen (wenn vorhanden)",
"Broadcast", "Ausstrahlung",
"Broadcast type", "Broadcast Typ",
"Broadcast.", "Ausstrahlung",
"Broadcasting", "Rundfunk",
"Buffering limit (ms)", "Limit für Pufferung (ms)",
"By default, linuxdvb's status read period is 1000ms (one second). The accepted range is 250ms to 8000ms. Note that for some hardware or drivers (like USB), the status operations take too much time and CPU. In this case, increase the default value. For fast hardware, this value might be decreased to make the decision of the re-tune algorithm based on the signal status faster.", "Standardmäßig wird der Linuxdvb-Status alle 1000ms (jede Sekunde) gelesen. Mögliche Werte sind 250ms bis 8000ms. Einige Hardware oder Treiber (wie USB) benötigen zu viel Zeit oder CPU. In diesem Fall sollte der Wert erhöht werden. Für schnelle Hardware kann der Wert verringert werden, um die Entscheidung des Retune-Algorithmus auf der Signalzustand zu beschleunigen.",
"CA (descrambling) Client", "CA-(Entschlüsselungs)-Client",
"CA ID", "CA-ID",
"CA Stream Filter", "CA-Streamfilter",
"CA client %i", "CA-Client %i",
"CA identification", "CA-Identifikation",
"CA provider", "CA-Anbieter",
"CAPMT interval (in ms).", "CAPMT Intervall (in ms)",
"CAPMT interval (ms)", "CAPMT Intervall (ms)",
"CAPMT query interval (ms)", "CAPMT Abfrageintervall (ms)",
"CAPMT query interval (ms).", "CAPMT Abfrageintervall (in ms)",
"CI Module", "CI Modul",
"CRID authority", "CRID Authority",
"CSA (descrambling)", "CSA (Entschlüsselung)",
"Cache scheme", "Puffer-Schema",
"Camd.socket filename / IP Address (TCP mode)", "Camd.socket Dateiname / IP-Adresse (TCP-Modus)",
"Cartoons", "Zeichentrickfilme",
"Change parameters", "Ändere Parameter",
"Channel", "Kanal",
"Channel Tags", "Kanal-Tags",
"Channel does not exist", "Kanal existiert nicht",
"Channel icon", "Kanallogo",
"Channel icon URL.", "Kanallogo URL",
"Channel icon as defined in EPG data.", "Kanallogo wie im EPG definiert",
"Channel icon name scheme", "Kanallogo Namensschema",
"Channel icon path", "Kanallogo Pfad",
"Channel mapping options", "Kanalzuordnung-Optionen",
"Channel name", "Kanalname",
"Channel not found", "Kanal nicht gefunden",
"Channel number", "Kanalnummer",
"Channel number as defined in EPG data.", "Kanalnummer wie im EPG definiert.",
"Channel number offset", "Kanalnummer Versatz",
"Channel number range", "Kanalnummer-Bereich",
"Channel numbers (heuristic)", "Kanalnummern (heuristisch)",
"Channel numbers from", "Kanalnummern von",
"Channel tag", "Kanal-Tag",
"Channel tag does not exist", "Kanal-Tag existiert nicht",
"Channel tag reference", "Kanal-Tag Referenz",
"Channel tags", "Kanal-Tags",
"Channel tags the user is allowed access to/excluded from.", "Kanal-Tags auf die der Nutzer zugreifen darf bzw. von denen er ausgeschlossen ist.",
"Channel to use/used for the recording.", "Kanal der für die Aufnahme genutzt wird/wurde.",
"Channel update options", "Kanal-Update-Optionen",
"Channels", "Kanäle",
"Channels EPG data is used by.", "Kanal EPG Daten werden genutzt von.",
"Character set", "Zeichensatz",
"Characters not supported in Windows filenames (e.g. for an SMB/CIFS share) will be stripped out or converted.", "Zeichen, welche von Windows Dateinamen nicht unterstützt werden (z.B. für eine SMB/CIFS Freigabe) werden entfernt oder konvertiert.",
"Charset", "Zeichensatz",
"Check availability", "Verfügbarkeit testen",
"Check for the services' presence. If the service is no longer broadcast this field will change to Missing In PAT/SDT.", "Prüfe die Verfügbarkeit des Services. Sollte der Service nicht länger ausgestrahlt werden, wird sich dieses Feld auf \"Fehlt in PAT/SDT\" ändern.",
"Check services for availability. If enabled, services that are not currently broadcasting (or can't be decrypted) will be ignored. Leave disabled if you want Tvheadend to also map offline services.", "Services auf ihre Verfügbarkeit prüfen. Wenn aktiviert, werden Services welche aktuell nicht senden (oder nicht entschlüsselt werden können) ignoriert. Lass es deaktiviert, wenn auch Services gemappt werden sollen, die nicht senden.",
"Check the descrambling status after this timeout.", "Überprüfe den Entschlüsselungsstatus nach dieser Zeit.",
"Children's / Youth Programs", "Kinder-/Jugendprogramm",
"Children's / Youth programs", "Kinder-/Jugendprogramm",
"Cinema", "Kino",
"Class", "Klasse",
"Classical", "Klassisch",
"Classical music", "Klassische Musik",
"Clean effects", "Nur Effekte",
"Client name", "Client-Name",
"Clone scheduled entry on error", "Kopiere geplanten Eintrag bei Fehlern",
"Comedy", "Komödie",
"Command delay time (ms) (10-200)", "Befehlsverzögerung (ms) (10-200)",
"Command time (ms) (10-100)", "Befehlszeit (ms) (10-100)",
"Comment", "Kommentar",
"Commercial break", "Werbeunterbrechung",
"Commercials will be dropped from the recordings. Commercial detection works using EITp/f (EPG running state) and for the Swedish channel TV4 (using teletext info).", "Werbung wird aus der Aufnahme geschnitten. Werbeerkennung  benutzt EITp/f (wenn EPG aktiv ist) und  für den schwedischen Kanal TV4 (Teletext Info).",
"Completed OK", "Erfolgreich abgeschlossen",
"Compress EPG database", "Komprimiere EPG Datenbank",
"Compress the EPG database to reduce disk I/O and space.", "Komprimiere die EPG Datenbank um Festplattenzugriffe und Speicherplatz zu sparen.",
"Conditional Access Identification.", "Conditional Access Identifikation",
"Configuration", "Konfiguration",
"Configuration - Base", "Konfiguration - Basis",
"Configuration - Image Cache", "Konfiguration - Bild-Cache",
"Configuration - SAT>IP Server", "Konfiguration - SAT>IP Server ",
"Configuration ID", "Konfiguration-ID",
"Configuration name", "Konfigurationsname",
"Configuration version", "Konfigurationsversion",
"Connection limit type", "Art des Verbindungslimits",
"Connection limits", "Verbindungslimits",
"Container to use for the transcoded stream.", "Zu nutzender Container für den transkodierten Stream.",
"Content character set", "Inhalte Zeichensatz",
"Content reference identifier authority.", "Content Reference Identifier Authority.",
"Content type", "Inhaltsart",
"Content type.", "Inhaltsart.",
"Contest", "Wettkampf",
"Continue if descrambling fails", "Fortfahren, wenn Entschlüsselung fehlschlägt",
"Continue recording for x minutes after scheduled stop time", "Aufnahme nach der geplanten Stop Zeit für x Minuten fortsetzen.",
"Continue recording for x minutes after scheduled stop time.", "Aufnahme nach der geplanten Stop Zeit für x Minuten fortsetzen.",
"Cookie expiration (days)", "Cookieablauf (Tage)",
"Cooking", "Kochen",
"Copy codec type", "Codectyp kopieren",
"Copy layout", "Layout kopieren",
"Count of objects", "Anzahl der Objekte",
"Create SDTV/HDTV/Radio tags.", "SDTV/HDTV/Radio Tags anlegen.",
"Create a bouquet from the playlist.", "Bouquet aus der Playlist erzeugen.",
"Create a directory per channel when storing recordings. If both this and the 'directory per day' checkbox is enabled, the date-directory will be the parent of the per-channel directory.", "Erstelle ein Verzeichnis pro Kanal beim Speichern von Aufnahmen. Wenn diese und \"Verzeichnis pro Tag\" beide aktiviert sind, wird das Datumsverzeichnis oberhalb des Kanalverzeichnisses erstellt.",
"Create a directory per title when storing recordings. If the day/channel directory checkboxes are also enabled, those directories will be parents of this directory.", "Erstelle ein Verzeichnis pro Titel beim Speichern von Aufnahmen. Wenn auch Datum-/Kanalverzeichnisse aktiviert sind, werden die Verzeichnisse oberhalb von diesem sein.",
"Create a new directory per day in the recording system path. Folders will only be created when something is recorded. The format of the directory will be ISO standard YYYY-MM-DD.", "Erstelle ein neues Verzeichnis pro Tag im Aufnahmensystempfad. Ordner werden nur erstellt wenn etwas aufgenommen wird. Das Format des Verzeichnisses ist der ISO Standard JJJJ-MM-DD.",
"Create a provider name tag.", "Anbietername-Tags anlegen.",
"Create and associate a network tag to created channels.", "Ein Netzwerk-Tag anlegen und den erstellten Kanälen zuordnen.",
"Create and associate a provider tag to created channels.", "Anbieter-Tag anlegen und den erstellten Kanälen zuordnen.",
"Create and link these tags to channels when mapping.", "Erstelle und verknüpfe diese Tags mit dem Kanal beim Mappen.",
"Create bouquet", "Bouquet erstellen",
"Create bouquet tag", "Bouquet erstellen",
"Create directories using these permissions.", "Erstelle Verzeichnisse mit diesen Berechtigungen.",
"Create files using these permissions.", "Erstelle Dateien mit diesen Berechtigungen.",
"Create network name tags", "Netzwerknamen-Tags anlegen",
"Create network name tags (set by provider).", "Erstelle Netzwerknamen-Tags (vom Anbieter gesetzt).",
"Create network tags", "Netzwerk-Tags anlegen",
"Create provider name tags", "Anbietername-Tags anlegen",
"Create provider tags", "Anbieter-Tags anlegen",
"Create tags", "Tags erstellen",
"Create tags in recordings using media containers that support metadata (if possible).", "Erstelle Tags in Aufnahmen unter Nutzung von Mediencontainern die Metadaten unterstützen (wenn möglich).",
"Create type-based tags", "Typbasierte Tags anlegen",
"Created", "Erstellt",
"Creator", "Ersteller",
"Cron multi-line", "Cron mehrzeilig",
"Culture (without music)", "Kultur (ohne Musik)",
"Culture magazines", "Kulturmagazine",
"Current RTSP port.", "Aktueller RTSP Port.",
"Current affairs", "Zeitgeschehen",
"Current number of objects.", "Aktuelle Zahl der Objekte.",
"Current object size.", "Aktuelle Objektgröße.",
"Current tuner configuration.", "Aktuelle Tuner-Konfiguration.",
"Custom HTTP headers", "Benutzerdefinierte HTTP Header",
"DAB Network", "DAB-Netzwerk",
"DAB multiplex", "DAB-Multiplex",
"DBus - use the session message bus instead of the system one", "DBus - Benutzen des Session- anstatt des System-Bus",
"DES Constant Code Word", "DES-Kennwort",
"DES Key.", "DES-Schlüssel.",
"DES key", "DES-Schlüssel",
"DSCP/TOS for streaming", "DSCP/TOS für Streaming",
"DVB CSA (descrambling) Tables", "DVB CSA (Entschlüsselung) Tabellen",
"DVB EPG ID", "DVB EPG-ID",
"DVB EPG Tables", "DVB EPG Tabelle",
"DVB Inputs - Networks", "DVB Eingang - Netzwerke",
"DVB SI Tables", "DVB SI Tabelle",
"DVB Time Tables", "DVB Zeit Tabellen",
"DVB multiplex", "DVB Multiplex",
"DVB scan files", "DVB Scandateien",
"DVB scan files path", "DVB Scandateien Pfad",
"DVB-C Network", "DVB-C Netzwerk",
"DVB-C multiplex", "DVB-C Multiplex",
"DVB-S Network", "DVB-S Netzwerk",
"DVB-S Satellite Configuration", "DVB-S Satellitenkonfiguration",
"DVB-S multiplex", "DVB-S Multiplex",
"DVB-T Network", "DVB-T Netzwerk",
"DVB-T multiplex", "DVB-T Multiplex",
"DVR - Auto-recording (Autorecs)", "DVR - Auto-Aufnahme (Autorecs)",
"DVR - Profiles", "DVR - Profile",
"DVR - Time-based Recording (Timers)", "DVR - Zeitbasierte Aufnahmen (Timer)",
"DVR behavior", "DVR Verhalten",
"DVR configuration", "DVR Konfiguration",
"DVR configuration profiles", "DVR Konfigurationsprofile",
"DVR configurations", "DVR Konfiguration",
"DVR entry not found", "DVR Eintrag nicht gefunden",
"DVR file retention period", "DVR Datei Speicherzeit",
"DVR log retention", "DVR Log Aufbewahrung",
"DVR log retention period", "DVR Log Speicherzeit",
"DVR override: high", "DVR Überschreibung Hoch",
"DVR override: important", "DVR Überschreibung Wichtig",
"DVR override: low", "DVR Überschreibung Niedrig",
"DVR override: normal", "DVR Überschreibung Normal",
"DVR override: unimportant", "DVR Überschreibung Unwichtig",
"DVR profile to use/used for the recording.", "DVR Profil, das für die Aufnahme genutzt wird/wurde.",
"DVR schedule does not exist", "DVR Planung existiert nicht.",
"DVR schedule does not have a file yet", "DVR Planung hat noch keine Datei.",
"DVR schedule not found", "DVR Planung nicht gefunden",
"Dance", "Tanz",
"Data errors", "Datenfehler",
"Data path (if applicable).", "Datenpfad (wenn anwendbar).",
"Date the EPG data was last updated (not set for OTA grabbers).", "Datum, wenn die EPG Daten das letzte mal aktualisiert wurden (wird nicht für OTA Grabber gesetzt).",
"Days of Week", "Wochentage",
"Days of the week to which the rule should apply.", "Wochentage an welchen die Regel wirken soll.",
"Days of week", "Wochentage",
"Debate", "Debatte",
"Debug libav log", "Debug libav-Log",
"Debug options", "Debug-Optionen",
"Debug subsystems", "Debugsubsysteme",
"Debug to syslog", "Debug nach Syslog",
"Debug trace (low-level)", "Debug Trace (systemnah)",
"Default", "Standard",
"Default language(s)", "Standardsprache(n)",
"Default language.", "Standardsprache.",
"Default login", "Standard Login",
"Default priority", "Standardpriorität",
"Default user interface level.", "Standard Ansichtsmodus.",
"Define network group to limit network usage.", "Definiere Netzwerkgruppen, um die Netzwerknutzung zu begrenzen",
"Delivery system", "Übertragungssystem",
"Demux path", "Demux Pfad",
"Descramble services (limit per mux)", "Entschlüssle Services (Limit pro Mux)",
"Descrambler", "Dekodierer",
"Descrambler EMM", "Entschlüsselungs-EMM",
"Descrambler buffer (TS packets)", "Entschlüsselungspuffer (TS Pakete)",
"Descrambling timeout (ms)", "Entschlüsselungs-Timeout (ms)",
"Description", "Beschreibung",
"Detective", "Krimi",
"Device ID", "Geräte-ID",
"Device model", "Gerätemodell",
"Device path", "Gerätepfad",
"Device path in sysfs", "Gerätepfad im sysfs",
"Differentiated Services Code Point / Type of Service: Set the service class Tvheadend sends with each packet. Depending on the option selected this tells your router the prority in which to give packets sent from Tvheadend, this option does not usually need changing. See https://en.wikipedia.org/wiki/Differentiated_services for more information. ", "DiffServ (Differentiated Services Code Point / Type of Service): Setzt die Service-Klasse, die Tvheadend mit jedem Paket mitsendet. Abhängig von der ausgewählten Option, ermöglicht es die Priorisierung ausgehend von deinem Router, mit welcher Tvheadend-Pakete bevorzugt werden. Normalerweise muss diese Option nicht geändert werden. Siehe https://en.wikipedia.org/wiki/Differentiated_services und https://de.wikipedia.org/wiki/DiffServ für mehr Informationen.",
"Digital Video Recorder", "Digitaler Videorekorder",
"Directory", "Verzeichnis",
"Directory override. Override the subdirectory rules specified by the DVR configuration and put all recordings done by this entry into the specified subdirectory", "Verzeichnis außer kraft setzen. Setzt die Unterverzeichnis-Regel in der DVR-konfiguration ausser Kraft und speichert alle Aufnahmen dieses Eintrages im angegebenen Unterverzeichnis",
"Directory permissions (octal, e.g. 0775)", "Verzeichnisrechte (oktal, z.B. 0775)",
"Directory used by the entry.", "Verzeichnis des Eintrags",
"Disable", "Deaktivieren",
"Disable DVB bouquets", "DVB-Bouquets deaktivieren",
"Disable RTP/AVP/TCP support", "Deaktiviere RTP/AVP/TCP-Unterstützung",
"Disable SAT>IP client", "Deaktiviere SAT>IP Client",
"Disable UPnP", "Deaktiviere UPnP",
"Disable X_SATIPM3U tag", "X_SATIPM3U-Tag deaktivieren",
"Disable all access control checks", "All Zugangsprüfungen deaktivieren",
"Disable debug on stderr", "Debug auf stderr deaktivieren",
"Disable device/firmware-specific workarounds", "Geräte-/Firmware-spezifische Umgehungslösungen deaktivieren",
"Disable syslog (all messages)", "Syslog deaktivieren (alle Meldungen)",
"Disabled", "Deaktiviert",
"Discover more muxes using the Network Information Table (if available).", "Suche weitere Muxe unter Nutzung der Network Information Table (wenn verfügbar).",
"Discussion", "Diskussion",
"DiseqC repeats", "DiseqC Wiederholungen",
"Do not send X_SATIPM3U information in the XML description to clients.", "Keine X_SATIPM3U-Informationen in der XML-Beschreibung an die Clients senden.",
"Do not send any stream specific information to the HTSP client like signal strength, input source etc.", "Sende keine streamspezifischen Informationen zum HTSP Client, wie Signalstärke, Eingangsquelle etc.",
"Do not use", "Nicht benutzen",
"Documentary", "Dokumentation",
"Don't abort streaming when an encrypted stream can't be decrypted by a CA client that normally should be able to decrypt the stream.", "Streaming nicht abbrechen, wenn ein verschlüsselter Stream von einem CA-Client nicht entschlüsselt werden kann, der normalerweise den Stream entschlüsseln können sollte.",
"Don't backup configuration tree at upgrade", "Konfigurationsbaum beim Upgrade nicht sichern",
"Don't include the title in the filename.", "Titel nicht in den Dateinamen einfügen.",
"Don't include title in filename", "Titel nicht in den Dateinamen einfügen",
"Don't keep", "Nicht behalten",
"Don't re-record", "Nicht neu aufnehmen",
"Don't re-record if recording fails.", "Nicht Neuaufnehmen, wenn die Aufnahme fehlschlägt.",
"Don't re-schedule if recording fails.", "Nicht Neuplanen, wenn die Aufnahme fehlschlägt.",
"Don't reschedule", "Nicht neu ansetzen",
"Don't use the provider's channel numbers.", "Kanalnummern des Anbieters nicht benutzen.",
"Duplicate handling", "Dublettenbehandlung",
"Duplicate recording handling.", "Dublettenaufnahmebehandlung.",
"EIT - skip TSID check", "EIT - überspringe TSID Überprüfung",
"EIT time offset", "EIT Zeitversatz",
"EPG Database", "EPG Datenbank",
"EPG Grabber", "EPG-Grabber",
"EPG Grabber Channel", "EPG-Grabber-Kanal",
"EPG Grabber Configuration", "EPG-Grabber Konfiguration",
"EPG Language (priority order)", "EPG Sprache (nach Priorität)",
"EPG data ID.", "EPG Daten-ID",
"EPG name", "EPG-Name",
"EPG overlap cut", "EPG Überlagerung",
"EPG scan", "EPG-Abfrage",
"EPG scan time-out in seconds (30-7200)", "EPG-Scan-Timeout in Sekunden (30-7200)",
"EPG source", "EPG-Quelle",
"EPG update window", "EPG Änderungszeitfenster",
"Economics", "Wirtschaft",
"Education", "Bildung",
"Education / Science / Factual topics", "Bildung / Wissenschaft / Sachthemen",
"Educational", "Lehrreich",
"Elementary Stream Filter", "Streamkomponentenfilter",
"Elementary stream filter", "Streamkomponentenfilter",
"Enable", "Aktiviert",
"Enable (auto)", "Aktivieren (auto)",
"Enable DBus", "DBus aktivieren",
"Enable NTP driver", "Aktiviere NTP-Treiber",
"Enable coredumps for daemon", "Coredumps für Daemon aktivieren",
"Enable debug on stderr", "Debug auf stderr aktivieren",
"Enable debug subsystems", "Debug-Subsysteme aktivieren",
"Enable debug to file", "Debug in Datei aktivieren",
"Enable debug to syslog", "Debug nach Syslog aktivieren",
"Enable if the SAT>IP box requests plts=on parameter in the SETUP RTSP command for DVB-S2 muxes.", "Aktiviere, wenn die SAT>IP-Box den \"plts=on\" Parameter im \"SETUP RTSP\"-Befehl für DVB-S2-Muxes anfordert.",
"Enable if the SAT>IP box requires pids=21 parameter in the SETUP RTSP command", "Aktivieren, wenn die SAT>IP Box den pids=21 Parameter im SETUP RTSP Kommando benötigt.",
"Enable if the SAT>IP box supports the addpids/delpids commands.", "Aktivieren, wenn die SAT>IP Box addpid/delpid Kommandos unterstützt.",
"Enable or disable RTP/AVP/TCP transfer mode (embedded data in the RTSP session) support.", "Aktiviere oder deaktiviere Unterstützung für RTP/AVP/TCP Übertragungsmodus (In die RTSP-Session eingebettete Daten)",
"Enable or disable fast input switching.", "Aktiviere oder Deaktiviere schnellen Eingangswechsel.",
"Enable or disable full mux mode.", "Full Mux Mode aktivieren oder deaktivieren.",
"Enable or disable ignoring of Event Information Table (EIT) data on this mux.", "Berücksichtige oder ignoriere Event Information Table (EIT) Daten bei diesem Mux.",
"Enable or disable this configuration.", "Diese Einstellungen aktivieren oder deaktivieren.",
"Enable over-the-air program guide (EPG) scanning on this input device.", "Aktiviere over-the-air EPG Scan auf diesem Eingangsgerät.",
"Enable syslog", "Syslog aktivieren",
"Enable system time updates. This will only work if the user running Tvheadend has rights to update the system clock (normally only root).", "Aktiviere Aktualisierung der Systemzeit. Das funktioniert nur wenn der Nutzer unter dem Tvheadend läuft die Berechtigung zum aktualisieren der Systemzeit hat (normalerweise nur root)",
"Enable the CAs (conditional accesses) tab in web user interface for the advanced level. By default, this tab is visible only in the expert level.", "Aktivere den CA (Conditional Acesses) Tab in der Benutzeroberfläche für den Ansichtmodus \"Erweitert\". Standardmäßig ist dieser Tab nur im Modus \"Experte\" sichtbar.",
"Enable this filter.", "Aktiviere den Filter.",
"Enable trace subsystems", "Trace-Subsysteme aktivieren",
"Enable web UI debug (non-minified JS)", "WebUI Debug aktivieren (nicht-minified JS)",
"Enable, disable or ignore the mux. When the mux is marked as ignore, all discovered services are removed.", "Aktiviere, deaktiviere oder ignoriere den Mux. Wenn \"Ignorieren\" markiert ist, werden alle gefundenen Dienste entfernt.",
"Enable/Disable CA client.", "Aktiviere/Deaktiviere CA Client",
"Enable/Disable service.", "Aktiviere/deaktiviere Service.",
"Enable/Disable the entry.", "Eintrag aktivieren/deaktivieren.",
"Enable/Disable timeshift.", "Timeshift ein/ausschalten",
"Enable/disable EPG data for the entry.", "Aktiviere/deaktiviere EPG Daten für diesen Eintrag.",
"Enable/disable LNA.", "Aktiviere/deaktiviere LNA.",
"Enable/disable auto-rec rule.", "Aktiviere/deaktiviere Auto-Aufnahme-Regel.",
"Enable/disable debugging output to syslog.", "Aktiviere/deaktiviere Debugausgabe ins Syslog.",
"Enable/disable inclusion of low-level debug traces.", "Aktiviere/deaktiviere die Einbeziehung des systemnahen Debug Trace",
"Enable/disable interface quick tips.", "Aktiviere/deaktiviere Benutzeroberflächen Quick-Tips",
"Enable/disable libav log output.", "Aktiviere/deaktiviere libav Logausgabe.",
"Enable/disable logging to syslog.", "Aktiviere/deaktiviere Protokollierung ins Syslog.",
"Enable/disable offering of Entitlement Management Message updates.", "Aktiviere/deaktiviere Entitlement Management Message updates.",
"Enable/disable pilot tone.", "Aktiviere/deaktiviere Pilotton.",
"Enable/disable power save mode (if supported by the device).", "Aktiviere/deaktiviere Stromsparmodus (wenn vom Gerät unterstützt).",
"Enable/disable profile.", "Aktiviere/deaktiviere Profil.",
"Enable/disable the bouquet.", "Aktiviere/deaktiviere Bouquet",
"Enable/disable the channel.", "Aktiviere/deaktiviere den Kanal.",
"Enable/disable the device.", "Aktiviere/deaktiviere das Gerät.",
"Enable/disable the entry.", "Aktiviere/deaktiviere den Eintrag.",
"Enable/disable the grabber.", "Aktiviere/deaktiviere den Grabber.",
"Enable/disable the tag.", "Aktiviere/deaktiviere den Tag.",
"Enable/disable tuner/adapter.", "Aktiviere/deaktiviere den Tuner/Adapter.",
"Enabled", "Aktiviert",
"Encrypted", "Verschlüsselt",
"Encrypted service", "Verschlüsselter Service",
"Enter a non-admin user password.", "Gib ein Nicht-Admin Benutzerpasswort ein.",
"Enter a non-admin user username.", "Gib einen Nicht-Admin Benutzernamen ein.",
"Enter allowed network prefix(es). You can enter a comma-seperated list of prefixes here.", "Gib die erlaubten Netzwerkprefixe ein. Sie können hier eine kommagetrennte Liste von Prefixen eingeben. ",
"Enter an administrator password.", "Gib ein Administratorpasswort ein.",
"Enter an administrator username. Note: do not use the same username as the superuser backdoor account.", "Gib einen Administratorbenutzernamen ein. Hinweis: Benutze nicht den gleichen Benutzernamen wie beim Superbenutzer-Hintertür-Konto. ",
"Enter comma-separated list of subsystems you want debugging output for (e.g +linuxdvb,+subscriptions,+mpegts).", "Gib eine mit Komma getrennte Liste an subsystems an, für die die Debugging-Ausgabe aktiviert werden soll (z.B. +linuxdvb,+subscriptions,+mpegts).",
"Enter comma-separated list of subsystems you want to get traces for (e.g +linuxdvb,+subscriptions,+mpegts).", "Gib eine mit Komma getrennte Liste an subsystems an, die du  verfolgen willst (z.B. +linuxdvb,+subscriptions,+mpegts).",
"Enter external IP if behind Network address translation (NAT). Asterisk (*) means accept all IP addresses.", "Gib die externe IP-Addresse an, wenn du hinter einem NAT (Network address translation) bist. Ein Sternchen (*) bedeutet, alle IP-Addressen zu akzeptieren.",
"Enter the filename (including path) where Tvheadend should write the log.", "Gib einen Dateinamen (mit Dateipfad) an, in den Tvheadend Logdateien schreiben soll.",
"Entertainment programs for 10 to 16", "Unterhaltungsprogramme für 10 bis 16",
"Entertainment programs for 6 to 14", "Unterhaltungsprogramme für 6 bis 14",
"Environment", "Umwelt",
"Environment (pipe)", "Umgebungsvariablen (pipe)",
"Episode number/ID.", "Episodennummer/ID.",
"Equestrian", "Pferdesport",
"Error code", "Fehlercode",
"Error code of entry.", "Errorcode des Eintrags",
"Errors", "Fehler",
"Even key", "Gerader Schlüssel",
"Even key.", "Gerader Schlüssel",
"Event does not exist", "Ereignis existiert nicht.",
"Every day", "Jeden Tag",
"Exclude channel tags", "Kanal-Tags ausschließen",
"Expeditions", "Expeditionen",
"Experimental film", "Experimenteller Film",
"Expert", "Experte",
"Exported tuners", "Exportierte Tuner",
"External", "Extern",
"External EPG grabber", "Externer EPG-Grabber",
"External IP (NAT)", "Externe IP (NAT)",
"External PyEPG Grabber", "Externer PyEPG-Grabber",
"External URL", "Externe URL",
"External URL of the bouquet.", "Externe URL des Bouquets.",
"External XMLTV EPG Grabber", "Externer XMLTV EPG-Grabber",
"Extra arguments", "Extra Parameter",
"Extra start time", "Extra Startzeit",
"Extra stop time", "Extra Endzeit",
"Extra warming up time (seconds)", "Extra Aufwärmzeit (Sekunden)",
"FAIL", "FEHLER",
"FEC high", "FEC Hoch",
"FEC low", "FEC Niedrig",
"Factual topics", "Sachliche Themen",
"Failed", "Fehlgeschlagen",
"Failed to open image", "Öffnen des Bildes fehlgeschlagen",
"Failed view", "Fehgeschlagene Wiedergabe",
"Fashion", "Mode",
"Fast input switch", "Schneller Eingangswechsel",
"Fastscan DVB", "DVB Schnellscan",
"File missing", "Datei fehlt",
"File not created", "Datei nicht erstellt",
"File permissions (octal, e.g. 0664)", "Dateirechte (oktal, z.B. 0664)",
"File removed", "Datei gelöscht",
"File size", "Dateigröße",
"Filename", "Dateiname",
"Filename (including path)", "Dateiname (inkl. Pfadangabe)",
"Filename character set", "Dateiname-Zeichensatz",
"Filename options", "Dateiname-Optionen",
"Filename used by the entry.", "Dateiname des Eintrags",
"Filesystem monitor", "Dateisystem Monitor",
"Fine arts", "Bildende Kunst",
"Finished", "Abgeschlossen",
"First word", "Erstes Wort",
"Fit to RAM (cut rewind)", "An RAM anpassen (Rückspulen beschneiden)",
"Fitness and health", "Fitness und Gesundheit",
"Folk", "Leute",
"Folkloric", "Folkloristisch",
"Force (auto)", "Erzwingen (auto)",
"Force CA ID (e.g. 0x2600)", "Erzwungene CA-ID (z.B. 0x2600)",
"Force all network connections to this tuner to be made over the specified IP address, similar to the setting for the SAT>IP device itself. Setting this overrides the device-specific setting.", "Erzwinge, dass alle Netzwerkverbindungen zu diesem Tuner über eine bestimmte IP-Adresse aufgebaut werden, ähnlich der Einstellung für das SAT>IP Gerät selbst. Diese Einstellung überschreibt die gerätespezifische Einstellung.",
"Force an initial EPG grab at start-up.", "Erzwinge eine initiale EPG-Suche beim Start.",
"Force initial EPG grab at start-up", "Erzwinge initiale EPG-Suche beim Start",
"Force old status", "Alten Status erzwingen",
"Force pilot for DVB-S2", "Erzwinge Pilot für DVB-S2",
"Force priority", "Priorität erzwingen",
"Force profile to use this priority.", "Erzwingen der Verwendung dieser Priorität für das Profil.",
"Force service type to 1", "Setze den Typ des Dienstes auf 1",
"Force teardown delay", "Erzwinge Abbauverzögerung",
"Force the delay between RTSP TEARDOWN and RTSP SETUP command (value from 'Next tune delay in ms' is used). Some devices are not able to handle quick continuous tuning.", "Erzwinge das Delay zwischen dem RTSP TEARDOWN- und RTSP SETUP-Befehl (der Wert von \"Nächste Tuneverzögerung in ms\" wird verwendet). Einige Geräte sind nicht in der Lage, schnell durchgehend zu tunen.",
"Force the grace period for which SAT>IP client waits for the data from server. After this grace period, the tuner is handled as dead. The default value is 5 seconds (for DVB-S/S2: 10 seconds).", "Erzwinge eine Wartezeit, die der SAT>IP Client wartet, bis er Daten vom Server erhalten hat. Nach dieser Wartezeit wird der Tuner als nicht verbunden behandelt. Der Standardwert ist 5 Sekunden (für DVB-S/S2: 10 Sekunden).",
"Force the local UDP Port number here. The number should be even (RTP port). The next odd number (+1) will be used as the RTCP port.", "Erzwingt die lokale UDP-Port-Nummer. Die Nummer sollte gerade sein (RTP-Port). Die darauffolgende ungerade Nummer (+1) wird als RTCP-Port benutzt.",
"Force usage of entered CA ID on this service.", "Erzwinge die Benutzung der eingegeben CA ID für diesen Service.",
"Foreign countries", "Fremde Länder",
"Forever", "Ewig",
"Fork and run as daemon", "Als Hintergrundprozess ausführen",
"Format string", "Formatzeichenkette",
"Found muxes", "Muxe gefunden",
"Found services", "Services gefunden",
"Free subscription weight", "Freies Abonnementgewicht",
"Free-form field, enter whatever you like here.", "Frei-Form-Feld. Es kann alles eingegeben werden.",
"Free-form text field, enter whatever you like here.", "Frei-Form-Textfeld. Es kann alles eingegeben werden.",
"Free-form text field, enter whatever you like.", "Frei-Form-Textfeld. Es kann alles eingegeben werden.",
"Free-form text field. You can enter whatever you like here.", "Frei-Form-Textfeld. Es kann alles eingegeben werden.",
"Free-format text field. Enter whatever you like here.", "Frei-Format-Textfeld. Es kann alles eingegeben werden.",
"Frequency", "Frequenz",
"Frequency (Hz)", "Frequenz (Hz)",
"Frequency (MHz)", "Frequenz (MHz)",
"Frequency (in MHz).", "Frequenz (in MHz)",
"Frequency (kHz)", "Frequenz (kHz)",
"Fri", "Fr",
"Friendly name", "Benutzer-lesbarer Name",
"Frontend number", "Frontend-Nummer",
"Frontend path", "Frontend-Pfad",
"Full DiseqC", "Volles DiSEqC",
"Full mux RX mode supported", "Full Mux Rx mode unterstützt",
"Full path to an icon used to depict the tag. This can be a TV network logotype, etc.", "Voller Pfad zu einem Logo, das den Tag abbildet. Dies kann ein Fernsehnetzwerk-Logotyp, etc. sein",
"Full pathname specification", "Volle Pfadnamenspezifikation",
"Full-text", "Volltext",
"Further education", "Weiterbildung",
"GOTOX", "GotoX",
"GOTOX position", "GotoX Position",
"Game show", "Spielshow",
"Gardening", "Gartenarbeit",
"General", "Allgemein",
"General configuration", "Allgemeine Konfiguration",
"Generic", "Allgemein",
"Generic options", "Allgemeine Optionen",
"Global Headers", "Globale Kopfzeilen",
"Global timer", "Globaler Timer",
"Grace period", "Wartezeit",
"Gray", "Grau",
"Guard interval", "Guard Interval",
"HD: high definition", "HD: Hohe Auflösung",
"HDHomeRun ATSC-C frontend", "HDHomeRun ATSC-C Frontend",
"HDHomeRun ATSC-T frontend", "HDHomeRun ATSC-T Frontend",
"HDHomeRun DVB frontend", "HDHomeRun DVB Frontend",
"HDHomeRun DVB-C frontend", "HDHomeRun DVB-C Frontend",
"HDHomeRun DVB-T frontend", "HDHomeRun DVB-T Frontend",
"HTSP Answer", "HTSP Antwort",
"HTSP Default Stream Settings", "HTSP Standard Streameinstellungen",
"HTSP Request", "HTSP Anfrage",
"HTSP Stream Profile", "HTSP Streamprofil",
"HTSP Subscription", "HTSP Abonnement",
"HTTP CORS (cross-origin resource sharing) origin. This option is usually set when Tvheadend is behind a proxy. Enter a domain (or IP) to allow cross-domain requests.", "HTTP CORS (cross-origin resource sharing) Herkunft. Diese Option wird normalerweise gesetzt, wenn Tvheadend sich hinter einem Proxy befindet. Gib eine Domain (oder ein IP-Adresse) ein, um cross-domain-Zugriffe zu erlauben.",
"HTTP CORS origin", "HTTP CORS Origin",
"Handicraft", "Handwerk",
"Hearing impaired", "Hörgeschädigt",
"Hierarchy", "Hierarchie",
"High", "Hoch",
"High bitrate mode (CI+ CAMs only)", "Hohe-Bitrate Modus (nur CI+ CAMs)",
"High frequency offset", "Hochfrequenzversatz",
"Highest channel number the user can access.", "Höchste vom Benutzer erreichbare Kanalnummer",
"Highest count of objects.", "Höchste Anzahl von Objekten",
"Historical movie", "Historischer Film",
"Hostname (or IP) of the server.", "Name (oder IP-Adresse) des Servers",
"How frequently it will re-try fetching an image that has failed to be fetched.", "Anzahl der Versuche ein Bild zu laden welches nicht geladen werden konnte.",
"How frequently the upstream provider is checked for changes.", "Wie oft der Upstream-Provider auf Änderungen überprüft wird.",
"ID of the network.", "ID des Netzwerks",
"IDLE", "INAKTIV",
"IGNORE", "IGNORIEREN",
"IP address", "IP-Adresse",
"IP blocking", "IP-Sperre",
"IPTV Automatic Network", "IPTV Automatisches Netzwerk",
"IPTV Multiplex", "IPTV-Multiplex",
"IPTV Network", "IPTV-Netzwerk",
"IPTV PCR", "IPTV-PCR",
"IPTV Subcription", "IPTV Abonnement",
"IPTV input", "IPTV-Input",
"ISDB-C Network", "ISDB-C Netzwerk",
"ISDB-C multiplex", "ISDB-C Multiplex",
"ISDB-S Network", "ISDB-S Netzwerk",
"ISDB-S multiplex", "ISDB-S Multiplex",
"ISDB-T Network", "ISDB-T Netzwerk",
"ISDB-T multiplex", "ISDB-T Multiplex",
"ISI (Stream ID)", "ISI (Stream-ID)",
"Icon", "Logo",
"Icon (full URL)", "Logo (komplette URL)",
"Icon URL", "Logo-URL",
"Icon base URL", "Logo-Basis-URL",
"Icon base URL.", "Logo-Basis-URL.",
"Icon has title", "Logo hat Titel",
"Icons from", "Logos von",
"Idle", "Leerlauf",
"Idle scan", "Leerlaufsuche",
"Idle scan muxes", "Mux Leerlaufsuche",
"If an error occurs clone the scheduled entry and try to record again (if possible).", "Im Fehlerfall dupliziere den geplanten Eintrag und versuche die Aufzeichnung nochmals (wenn möglich)",
"If both a picon and a channel-specific (e.g. channelname.jpg) icon are defined, prefer the picon.", "Wenn sowohl ein Picon als auch ein Kanal-spezifisches Logo (z.B. kanalname.jpg) definiert sind, bevorzuge das Picon.",
"If more than x errors occur during a recording schedule a re-record (if possible).", "Wenn mehr als x Fehler während der Aufzeichnung auftreten, plane eine Wiederholung der Aufnahme ",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Erstellen eines Kontos ohne Benutzername\nund Passwort, wenn kein Benutzerkonto existiert.\nBeachten Sie, dass es weltweiten administrativen\nZugriff zu Ihrer Tvheadend Installation erlaubt, bis\nSie die Zugangskontrolle in der Tvheadend Weboberfläche \nanpassen.",
"If something fails, try to switch to a different service on another network. Do not try to iterate through all inputs/tuners which are capable to receive the service.", "Wenn irgendetwas schiefgeht, versuche auf einen anderen Service auf einem anderen Netzwerk zu schalten. Versuche nicht, alle Eingänge/Tuner, die diesen Service empfangen können, auszuprobieren.",
"If the DiseqC switch is located before the rotor (i.e. tuner - switch - rotor), enable this.", "Aktiviere dies, wenn der DiseqC-Schalter sich vor dem Rotor befindet. (z.B. Tuner - Schalter - Rotor) ",
"If you'd like to support the project, please consider a donation.", "Wenn Sie das Projekt unterstützen möchten, erwägen Sie bitte eine Spende.",
"Ignore", "Ignorieren",
"Ignore EPG (EIT)", "EPG (EIT) ignorieren",
"Ignore descriptor 5", "Deskriptor 5 ignorieren",
"Ignore encryption flag, include encrypted services anyway.", "Ignoriere ",
"Ignore invalid SSL certificate", "Ungültiges SSL-Zertifikat ignorieren",
"Ignore invalid/unverifiable (expired, self-certified, etc.) certificates", "Ungültige oder nicht überprüfbare (abgelaufene, selbstsignierte, etc.) Zertifikate ignorieren",
"Ignore provider's channel numbers", "Kanalnummern des Anbieters ignorieren",
"Image Cache", "Bild-Cache",
"Immediately abort", "Direkt abbrechen",
"Important", "Wichtig",
"Include channel name in filename", "Kanalname in Dateiname einfügen",
"Include date in filename", "Datum in Dateiname einfügen",
"Include episode in filename", "Episode in Dateiname einfügen",
"Include subtitle in filename", "Untertitel in Dateiname einfügen",
"Include teletext", "Teletext einschließen",
"Include teletext in the timeshift buffer. Enabling this may cause issues with some services where the teletext DTS is invalid.", "Teletext in den Timeshift-Puffer schreiben. Das Aktivieren kann Probleme mit einigen Services bereiten, bei denen der Telext DTS ungültige ist.",
"Include the date for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "Das Datum der Aufnahme in den Ereignistitel einbinden. Dies trifft sowohl auf den in der Datei gespeicherten Titel als auch auf den eigentlichen Dateinamen zu.",
"Include the episode subtitle in the title (if available).", "Untertitel der Folge in den Titel aufnehmen (Falls vorhanden)",
"Include the name of the channel in the event title. This applies to both the title stored in the file and to the filename itself.", "Den Namen des Kanals in den Ereignistitel einbinden. Dies trifft sowohl auf den in der Datei gespeicherten Titel als auch auf den eigentlichen Dateinamen zu.",
"Include the season and episode in the title (if available).", "Staffel und Episode in den Titel einbinden (wenn verfügbar).",
"Include the time for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "Zeit der Aufnahme in den Ereignistitel einbinden. Dies trifft sowohl auf den in der Datei gespeicherten Titel als auch auf den eigentlichen Dateinamen zu.",
"Include time in filename", "Zeit in Dateiname einfügen",
"Information area", "Informationsbereich",
"Informational", "Informativ",
"Initial scan", "Initiale Suche",
"Input base", "Input Basis",
"Input buffer (bytes)", "Input-Puffer (Byte)",
"Input instance", "Eingangsinstanz",
"Input path", "Eingangspfad",
"Interface", "Schnittstelle",
"Internal", "Intern",
"Internal EPG grabber", "Interner EPG-Grabber",
"Internal PyEPG Grabber", "Interner PyEPG Grabber",
"Internal XMLTV EPG Grabber", "Interner XMLTV EPG-Grabber",
"Internal error", "Interner Fehler",
"Internal grabber", "Interner Grabber",
"Invalid", "Ungültig",
"Invalid EPG object request", "Ungültige EPG-Objektanfrage",
"Invalid adapter number '%s'\n", "Ungültige Adapternummer '%s'\n",
"Invalid arguments", "Ungültige Argumente",
"Invalid file", "Ungültige Datei",
"Invalid parameters", "Ungültige Parameter",
"Invalid service", "Ungültiger Dienst",
"Invalid target", "Ungültiges Ziel",
"Items", "Elemente",
"Keep", "Beibehalten",
"Keep timeshift buffers in RAM only. With this option enabled, the amount of rewind time is limited by how much RAM TVHeadend is allowed.", "Den Timeshift-Puffer nur im RAM behalten. Die Länge der Rückspulzeit wird dann durch die an TVHeadend zugewiesene Menge an RAM begrenzt.",
"Keep x amount of storage space free.", "Halte x Menge des Speicherplatzes frei.",
"Keepalive interval", "Keepalive Intervall",
"Keepalive interval in seconds", "Keepalive Intervall in Sekunden",
"Key and value pairs to remove from the query string in the URL.", "Schlüssel-Wert-Paare, die vom Aufrufstring in der URL entfernt werden.",
"LNA (low noise amplifier)", "LNA (Low Noise Amplifier)",
"LNB type", "LNB Typ",
"Language", "Sprache",
"Language 1", "Sprache 1",
"Language 2", "Sprache 2",
"Language 3", "Sprache 3",
"Language settings", "Spracheinstellungen",
"Languages", "Sprachen",
"Largest size the object has reached.", "Maximale Größe des Objektes ist erreicht.",
"Last played position", "Zuletzt gespielte Position",
"Last played position when the recording isn't fully watched yet.", "Zuletzt gespielte Position, wenn die Aufnahme noch nicht vollständig betrachtet wurde.",
"Last seen", "Zuletzt gefunden",
"Last updated from", "Zuletzt aktualisiert durch",
"Layer A: Constellation", "Layer A. Konstellation",
"Layer A: Segment count", "Layer A: Segmentzählung",
"Layer A: Time interleaving", "Layer A: Zeitverschachtelung",
"Layer B: Constellation", "Layer B: Konstellation",
"Layer B: Segment count", "Layer B: Segmentzählung",
"Layer B: Time interleaving", "Layer B: Zeitverschachtelung",
"Layer C: Constellation", "Layer C: Konstellation",
"Layer C: Segment count", "Layer C: Segmentzählung",
"Layer C: Time interleaving", "Layer C: Zeitverschachtelung",
"Leisure hobbies", "Freizeitbeschäftigungen",
"Limit connections", "Verbindungsbegrenzung",
"Linked input", "Verknüpfter Input",
"Linux DVB frontend", "Linux DVB Frontend",
"LinuxDVB adapter", "LinuxDVB Adapter",
"LinuxDVB network", "LinuxDVB Netzwerk",
"List of allowed IPv4 or IPv6 hosts or networks (comma-separated).", "Liste der erlaubten IPv4 oder IPv6 Hosts oder Netzwerke (kommagetrennt).",
"List subsystems", "Liste Subsysteme",
"Listen / Connect port", "Empfangs-/Verbindungsport",
"Listen on IPv6", "Auf IPv6 hören",
"Literature", "Literatur",
"Local (server) time", "Lokale (Server-)Zeit",
"Local bind IP address", "IP-Adresse für lokale Bindung",
"Local channel minor", "Lokale untergeordnete Kanalnummer",
"Local channel number", "Lokale Kanalnummer",
"Local discovery IP address", "IP-Adresse für lokale Suche",
"Local: Record if different description", "Lokal: Aufnehmen, wenn abweichende Beschreibung",
"Local: Record if different episode number", "Lokal: Aufnehmen, wenn abweichende Episiodennummer",
"Local: Record if different subtitle", "Lokal: Aufnehmen, wenn abweichender Untertitel",
"Local: Record if different title", "Lokal: Aufnehmen, wenn abweichender Titel",
"Local: Record once per day", "Lokal: Einmal pro Tag aufnehmen ",
"Local: Record once per month", "Lokal: einmal pro Monat aufnehmen",
"Local: Record once per week", "Lokal: Einmal pro Woche aufnehmen",
"Location", "Ort",
"Location details of the SAT>IP Server.", "Standortdetails des SAT>IP Servers.",
"Lock preferred CA PID", "Bevorzugte CA PID verriegeln",
"Locking", "Verriegeln",
"Login password.", "Passwort",
"Login username.", "Benutzername",
"Login/Logout", "An-/Abmelden",
"Loglevel debug: disabled", "Debugloglevel: deaktiviert",
"Loglevel debug: enabled", "Debugloglevel: aktiviert",
"Low", "Niedrig",
"Low frequency offset", "Niederfrequenzversatz",
"Lowest channel number the user can access.", "Niedrigste vom Benutzer erreichbare Kanalnummer",
"Lowest starting channel number.", "Niedrigste startende Kanalnummer.",
"MPEG-TS File", "MPEG-TS Datei",
"MPEG-TS Parser", "MPEG-TS Analysierer",
"MPEG-TS Pass-thru", "MPEG-TS Durchreichen",
"MPEG-TS Pass-thru/built-in", "MPEG-TS Durchreichen/eingebaut",
"MPEG-TS input", "MPEG-TS-Input",
"MPEG-TS multiplex PHY", "MPEG-TS Multiplex PHY",
"MPEG-TS raw service", "MPEG-TS Unbearbeiteter Service",
"Magazines", "Magazine",
"Main", "Haupt",
"Maintain free storage space in MiB", "Verwalte freien Speicher in MiB",
"Maintain used storage space in MiB (0=disabled)", "Verwalte genutzten Speicher in MiB (0=deaktiviert)",
"Maintained space", "Verwalteter Speicher",
"Make subdirectories per channel", "Erstelle Unterverzeichnisse je Kanal",
"Make subdirectories per day", "Erstelle Unterverzeichnisse je Tag",
"Make subdirectories per title", "Erstelle Unterverzeichnisse je Titel",
"Manage all", "Alles verwalten",
"Manufacturer", "Hersteller",
"Manufacturer URL", "Hersteller URL",
"Manufacturer's model number.", "Hersteller Modellnummer",
"Manufacturer's product description.", "Hersteller Produktbeschreibung",
"Manufacturer's product information page for the device.", "Hersteller Produktinformationsseite für das Gerät",
"Manufacturer's product name.", "Herstellter Produktname",
"Map all services", "Alle Services zuordnen",
"Map encrypted services", "Verschlüsselte Services einschließen",
"Map radio channels", "Radiokanäle zuordnen",
"Map unnamed channels", "Kanäle ohne Namen zuordnen",
"Map zero-numbered channels", "Null-genummerte Kanäle zuordnen",
"Mapped", "Zugeordnet",
"Martial sports", "Kampfsportarten",
"Master tuner", "Haupttuner",
"Matroska (mkv)/built-in", "Matroska (mkv)/eingebaut",
"Matroska specific", "Matroskaspezifisch",
"Maximal channel number", "Maximale Kanalnummer",
"Maximum # input streams", "Maximale # von Input-Streams",
"Maximum PIDs", "Maximale PIDs",
"Maximum RAM size (MB)", "Maximale RAM Größe (MB)",
"Maximum allowed difference between event start time when the EPG event is changed.", "Größte erlaubte Zeitdifferenz zwischen der Startzeit des Ereignisses und dem Zeitpunkt, in dem das EPG-Ereignis geändert wird.",
"Maximum bandwidth (Kbps)", "Maximale Bandbreite (Kb/s)",
"Maximum count (0=default)", "Maximale Anzahl (0=Standard)",
"Maximum duration", "Maximale Länge",
"Maximum input bandwidth.", "Maximale Eingangsbandbreite.",
"Maximum length of PIDs", "Maximale Länge von PIDs",
"Maximum period (mins)", "Maximaler Zeitraum (Min.)",
"Maximum schedules limit (0=default)", "Maximale Pläne (0=Standard)",
"Maximum size (MB)", "Maximale Größe (MB)",
"Maximum timeout (seconds)", "Maximale Zeitüberschreitung (Sekunden)",
"Medicine", "Medizin",
"Melodrama", "Melodram",
"Memory Information", "Speicher Information",
"Merge same name", "Gleiche Namen zusammenführen",
"Merge services with the same name into one channel.", "Services mit dem gleichen Namen in einem Sender zusammenführen.",
"Method not found", "Methode nicht gefunden",
"Minimal channel number", "Minimale Kanalnummer",
"Minimum duration", "Minimale Länge",
"Minimum rotor time (seconds)", "Minimale Rotor Zeit (Sekunden)",
"Missing In PAT/SDT", "Fehlt in PAT/SDT",
"Mode", "Modus",
"Model description", "Modellbeschreibung",
"Model name", "Modellname",
"Model number", "Modellnummer",
"Module", "Modul",
"Module ID", "Modul-ID",
"Module ID used to grab EPG data.", "ID des für das Grabben der EPG-Daten verwendeten Moduls.",
"Mon", "Mo",
"More verbose libav log", "Ausführlicheres libav-Log",
"Motor rate (in milliseconds/deg).", "Geschwindigkeit des Motors (Millisekunden/Grad)",
"Motor rate (milliseconds/deg)", "Motor Rate (Millisekunden/Grad)",
"Motor sport", "Motorsport",
"Motoring", "Autofahren",
"Movie", "Film",
"Movie / drama", "Film / Drama",
"Music", "Musik",
"Music / Ballet / Dance", "Musik / Ballett / Tanz",
"Mux Scheduler", "Mux Planer",
"Mux handling", "Muxbehandlung",
"Mux name", "Mux-Name",
"Mux not enabled", "Mux nicht aktiviert",
"NONE", "Kein",
"Name (or date) of program the entry is a rerun of.", "Name (oder Datum) des Programms das eine Wiederholung ist von.",
"Name of channel the entry recorded from.", "Name des Kanals der für die Aufnahme genutzt wird.",
"Name of object.", "Objektname",
"Name of the bouquet.", "Name des Bouquets.",
"Name of the client.", "Name des Clients.",
"Name of the entry.", "Name des Eintrags.",
"Name of the module used to grab EPG data.", "Name des für das Grabben der EPG-Daten verwendeten Moduls.",
"Name of the network.", "Name des Netzwerks.",
"Name of the profile.", "Name des Profils.",
"Name of the tag.", "Name des Tags.",
"Name of the tuner.", "Name des Tuners.",
"Name of the tuner/adapter.", "Name des Tuners/Adapters.",
"Names", "Namen",
"Natural sciences", "Naturwissenschaften",
"Nature", "Natur",
"Network", "Netzwerk",
"Network ", "Netzwerk",
"Network A", "Netzwerk A",
"Network B", "Netzwerk B",
"Network C", "Netzwerk C",
"Network D", "Netzwerk D",
"Network ID (limit scanning)", "Netzwerk-ID (Suchbeschränkung)",
"Network UUID", "Netzwerk UUID",
"Network access", "Netzwerkzugang",
"Network discovery", "Netzwerksuche",
"Network for port A.", "Netzwerk auf Port A.",
"Network for port AA.", "Netzwerk auf Port AA.",
"Network for port AB.", "Netzwerk auf Port AB.",
"Network for port B.", "Netzwerk auf Port B.",
"Network for port BA.", "Netzwerk auf Port BA.",
"Network for port BB.", "Netzwerk auf Port BB.",
"Network for port C.", "Netzwerk auf Port C.",
"Network for port D.", "Netzwerk auf Port D.",
"Network group", "Netzwerkgruppe",
"Network limit per position", "Netzwerklimitierung pro Position",
"Network name", "Netzwerkname",
"Network prefix", "Netzwerkpräfix",
"Network settings", "Netzwerkeinstellungen",
"Network type", "Netzwerktyp",
"Networks", "Netzwerke",
"New / premiere / unknown", "Neu / Premiere / Unbekannt",
"New login", "Neuer Login",
"New muxes + changed muxes", "Neue und geänderte Muxes",
"New muxes only", "Nur neue Muxes",
"News", "Nachrichten",
"News / Current Affairs", "Nachrichten / Aktuelles",
"News magazine", "Nachrichtenmagazin",
"Next tune delay in ms (0-2000)", "Nächste Tuneverzögerung in ms (0-2000)",
"No", "Nein",
"No PIN", "Keine PIN",
"No access", "Kein Zugriff",
"No adapters specified!\n", "Keine Adapter spezifiziert!\n",
"No assigned adapters", "Kein zugeordneter Adapter",
"No days", "Kein Tag",
"No descrambler", "Kein Descrambler",
"No free adapter", "Kein freier Adapter",
"No input detected", "Keine Daten erkannt",
"No scheme", "Kein Schema",
"No service assigned to channel", "Dem Kanal ist kein Service zugeordnet",
"No service enabled", "Kein Service aktiviert",
"No source available", "Keine Quelle verfügbar",
"None", "Keine",
"Not all SAT>IP servers use the same signal scaling. Change this setting if the signal level displayed within Tvheadend looks too low.", "Nicht alle SAT>IP Server nutzen die gleiche Signalauflösung. Ändere diese Einstellung wenn die in Tvheadend angezeigte Signalstärke zu gering erscheint.",
"Not enough disk space", "Nicht genügend Speicherplatz",
"Not enough memory", "Nicht genug Speicher",
"Not linked", "Nicht verbunden",
"Not set", "Nicht gesetzt",
"Not set (none or channel configuration)", "Nicht gesetzt (keine oder Kanalkonfiguration)",
"Not set (use channel or DVR configuration)", "Nicht gesetzt (benutze Kanal- oder DVR-Konfiguration)",
"Number", "Nummer",
"Number of days to keep the file.", "Anzahl der Tage die Datei zu behalten.",
"Number of days to keep the recorded file.", "Anzahl der Tage die aufgenommene Datei zu behalten.",
"Number of days to keep the recorded files.", "Anzahl der Tage die aufgenommenen Dateien zu behalten.",
"Number of days to retain entry information.", "Anzahl der Tage die Informationen zum Eintrag zu behalten.",
"Number of errors that occurred during recording (Transport errors).", "Anzahl der Fehler, die bei der Aufnahme aufgetreten sind (Übertragungsfehler).",
"Number of errors that occurred during recording.", "Anzahl der Fehler, die bei der Aufnahme aufgetreten sind.",
"Number of muxes found.", "Anzahl der gefundenen Muxe.",
"Number of seconds to wait before timing out.", "Sekunden",
"Number of services.", "Anzahl der Services",
"Number of times this recording was played.", "Anzahl der Wiedergaben dieser Aufnahme.",
"Number of tsfile tuners", "Anzahl von tsfile-Tuner",
"OK (partial)", "OK (teilweise)",
"Odd key", "Ungerader Schlüssel",
"Odd key.", "Ungerader Schlüssel",
"Off", "Aus",
"Older OSCam", "Älteres OSCam",
"On", "Ein",
"On file removal", "Dateientfernung (Tage)",
"On-demand (no first rewind)", "On-demand (kein sofortiges zurückspulen)",
"Once per auto channel", "Einmal pro Auto-Kanal",
"Only Bulsatcom 39E", "Nur Bulsatcom 39E",
"Only EIT", "Nur EIT",
"Only OpenTV Sky Ausat", "Nur OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Nur OpenTV Sky Italia",
"Only OpenTV Sky UK", "Nur OpenTV Sky UK",
"Only PSIP (ATSC)", "Nur PSIP (ATSC)",
"Only UK Freesat", "Nur UK Freesat",
"Only UK Freeview", "Nur UK Freeview",
"Only Viasat Baltic", "Nur Viasat Baltic",
"Only admin can watch the realtime log.", "Nur der Administrator kann Logdateien in Echtzeit betrachten.",
"Only digits", "Nur Ziffern",
"Only preferred CA PID", "Nur bevorzugte CA PID",
"Only use specified DVB adapters (comma-separated, -1 = none)", "Benutze nur angegebene DVB-Adapter (Komma getrennt, -1 = keiner)",
"OpenTV channel number", "OpenTV Kanalnummer",
"Opera", "Oper",
"Options to use/used when mapping - see Help for details.", "Genutze Kanalzuordnung-Optionen - Siehe Hilfe für details",
"Options used when updating channels.", "Genutzte Kanal-Update-Optionen",
"Orbital position", "Orbitale Position",
"Orbital positions", "Orbitale Positionen",
"Orbital positions.", "Orbitale Positionen",
"Original network ID", "Original Netzwerk-ID",
"Oscam mode.", "Oscam Modus",
"Other Stream Filter", "Sonstiger Streamfilter",
"Output directory for tsdebug", "Zielverzeichnis für tsdebug",
"Over-the-air Cron multi-line", "Over-the-air Cron mehrzeilig",
"Over-the-air EPG grabber", "Over-the-air EPG-Grabber",
"Over-the-air grabbers", "Over-the-air Grabber",
"Override disabled", "Überschreiben deaktiviert",
"Override the frontend type.", "Überschreibe den Frontend-Typ.",
"Owner", "Besitzer",
"Owner of the entry.", "Eigentümer des Eintrags.",
"Owner of the rule.", "Eigentümer der Regel.",
"PEND", "AUSSTEHEND",
"PIDs 21 in setup", "PIDs 21 im Setup",
"PIN inquiry match string", "PIN-Abfrage Suchstring",
"PIN inquiry match string.", "PIN-Abfrage Suchstring.",
"PLS code", "PLS Code",
"PLS mode", "PLS Modus",
"Parent entry", "Übergeordneter Eintrag",
"Parent entry.", "Übergeordneter Eintrag.",
"Pass subscription weight", "Abonnementgewicht durchreichen",
"Password", "Passwort",
"Password for the entry.", "Passwort für den Eintrag.",
"Password2", "Passwort2",
"Path", "Pfad",
"Path to the frontend used by the device.", "Pfad zum Frontend, das vom Gerät benutzt wird.",
"Path to the grabber executable.", "Grabber Programmpfad",
"Path to the socket Tvheadend will read data from.", "Dateipfad des Sockets von dem Tvheadend Daten liest.",
"Path used by the device.", "Gerätepfad",
"Peak count of objects", "Höchste Anzahl von Objekten",
"Peak size", "Spitzenwert",
"Performing arts", "Darstellende Künste",
"Periodically save EPG to disk (hours)", "Speichere EPG periodisch auf Festplatte (Stunden)",
"Persistent user interface level", "Ansichtsmodus fixieren",
"Physiology", "Physiologie",
"Picon name scheme", "Picon Namensschema",
"Picon path", "Picon-Pfad",
"Polarization", "Polarisierung",
"Political issues", "Politik",
"Popular culture", "Popkultur",
"Port to connect to.", "Zu verbindender Prot",
"Port to listen on or to connect to.", "Port, auf dem gehorcht oder zu dem verbunden werden soll.",
"Position of the input.", "Position der Eingabe.",
"Post-processor command", "Kommando zur Nachbearbeitung",
"Post-recording padding", "Nach-Aufname-Polster",
"Post-remove command", "Kommando nach Entfernen",
"Power save", "Energie sparen",
"Power-up time (ms) (15-200)", "Hochfahrzeit (ms) (15-200)",
"Pre-defined muxes", "Vordefinierte Muxe",
"Pre-processor command", "Preprozessor-Kommando",
"Pre-recording padding", "Vor-Aufname-Polster",
"Pre-school children's programs", "Vorschulkinderprogramme",
"Prefer picons over channel name", "Picons dem Kanalnamen vorziehen",
"Preferred CA PID", "Bevorzugte CA PID",
"Preferred audio language.", "Bevorzugte Sprache der Tonspur",
"Preferred service video type", "Bevorzugter Service-Videotyp",
"Presentation", "Präsentation",
"Presentation details.", "Präsentationsdetails.",
"Press", "Presse",
"Priority", "Priorität",
"Priority (-10..10)", "Priorität (-10..10)",
"Private", "Privat",
"Profile name", "Profilname",
"Program synopsis (display only).", "Inhaltsangabe (nur anzeigen).",
"Program synopsis.", "Inhaltsangabe",
"Provider", "Anbieter",
"Provider ID", "Anbieter-ID",
"Provider network name", "Anbieter Netzwerkname",
"Provider's network name.", "Netzwerkname des Providers",
"Psychology", "Psychologie",
"Puppets", "Marionetten",
"RAM only", "Nur RAM",
"RTP/AVP/TCP (embedded data)", "RTP/AVP/TCP (eingebettete Daten)",
"RTSP Protocol", "RTSP-Protokoll",
"RTSP port", "RTSP-Port",
"RTSP port (554 or 9983, 0 = disable)", "RTSP Port (554 oder 9983, 0 = deaktiviert)",
"Re-fetch period (hours)", "Abrufintervall (Stunden)",
"Re-fetch period (mins)", "Abrufintervall (Minuten)",
"Re-record%s%s", "Neu-Aufnahme%s%s",
"Re-try period (hours)", "Wiederholungsintervall (Stunden)",
"Read error", "Lesefehler",
"Record all", "Alle aufnehmen",
"Record on these days only.", "Nur an diesen Tagen aufnehmen.",
"Recording file options", "Aufnahmedatei-Optionen",
"Recording file size.", "Dateigröße der Aufnahme.",
"Recording play count", "Aufnamewiedergabezähler",
"Recording system path", "Aufnahmedateipfad",
"Reject", "Ablehnen",
"Reject exact match", "Exakte Übereinstimmung ablehnen",
"Relative path to the imagecache copy of the icon.", "Relativer Pfad zu der Kopie des Logos im Bildcache.",
"Religious", "Religiös",
"Remarkable people", "Bemerkenswerten Menschen",
"Remove HTTP arguments", "Entferne HTTP-Argumente",
"Remove all unsafe characters from filename", "Alle unsicheren Zeichen aus Dateiname entfernen",
"Remove scrambled bits", "verschlüsselte Teile entfernen",
"Repeated", "Wiederholt",
"Replace whitespace in title with '-'", "Leerzeichen im Titel durch '-' ersetzen",
"Replaces all whitespace in the title with '-'.", "Ersetze alle Leerzeichen im Titel mit '-'.",
"Reply to CAM PIN inquiries", "Auf CAM PIN-Abfragen antworten",
"Reply to PIN inquiries.", "Auf PIN-Abfragen antworten.",
"Reports", "Berichte",
"Rerun of", "Wiederholung von",
"Reserved", "Reserviert",
"Resolution (height)", "Auflösung (Höhe)",
"Respawn (pipe)", "Neustart (pipe)",
"Restart", "Neustart",
"Restart on error", "Neustart nach Fehler",
"Restrict connections to this type.", "Verbindungen auf diesen Typ beschränken",
"Reuse EPG from", "Benutze EPG von",
"Reuse the EPG from another channel.", "Wiederverwendung des EPG von einem anderen Kanal.",
"Rewrite EIT", "EIT umschreiben",
"Rewrite MPEG-TS SI tables", "MPEG-TS SI Tabellen erneuern",
"Rewrite PAT", "PAT umschreiben",
"Rewrite PMT", "PMT umschreiben",
"Rewrite SDT", "SDT umschreiben",
"Rights", "Rechte",
"Romance", "Romantik",
"Rotor initialization time (seconds)", "Rotor Initialisierungszeit (Sekunden)",
"Rotor type", "Rotor Typ",
"Run as group", "Als Gruppe ausführen",
"Run as user", "Als Benutzer ausführen",
"Running", "Läuft",
"SAT->IP frontend number.", "SAT>IP Frontend Nummer",
"SAT>IP DVB-T frequency (Hz)", "SAT>IP DVB-T Frequenz (Hz)",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "SAT>IP RTSP Portnummer für den Server\n(Standard: -1 = deaktiviert, 0 = webconfig, Standardport ist 554)",
"SAT>IP Satellite Configuration", "SAT>IP Satelliten Konfiguration",
"SAT>IP Server SI Tables", "SAT>IP-Server-SI-Tabellen",
"SAT>IP source number", "SAT>IP Quellennummer",
"SD: standard definition", "SD: Standardauflösung",
"SSL verify peer", "Überprüfe SSL-Peer",
"Satellite config", "Satellitenkonfiguration",
"Satellite longitude", "Satellit Longitude",
"Satellite longitude.", "Längengrad des Satelliten",
"Satellite position.", "Satellitenposition",
"Satellite positions", "Satellitenpositionen",
"Scan after creation", "Scan nach Erstellen",
"Scan progress", "Fortschritt",
"Scan queue length", "Suchwarteschlangenlänge",
"Scan result", "Suchergebnis",
"Scan status", "Suchstatus",
"Schedule a re-recording if more errors than (0=off)", "Plane eine Neuaufnahme bei mehr Fehlern als (0=aus)",
"Schedule frequency (in cron format).", "Zeitplan (im cron Format).",
"Schedule status", "Programmierstatus",
"Schedule status.", "Programmierstatus",
"Scheduled Duration", "Geplante Dauer",
"Scheduled for recording", "Für Aufnahme programmiert",
"Scheduled start time", "Programmierte Startzeit",
"Scheduled stop time", "Programmierte Stopzeit",
"School programs", "Schulprogramme",
"Science", "Wissenschaft",
"Science fiction", "Science-Fiction",
"Season", "Staffel",
"Season %d", "Staffel %d",
"Season information (if available).", "Staffel Information ( wenn vorhanden)",
"Seek error", "Suchfehler",
"Select an available network type for this tuner.", "Einen vorhandenen Netzwerktyp für diesen Tuner auswählen.",
"Select type of broadcast.", "Boradcast Typ auswählen",
"Send CAPMT query", "Sende CAPMT Anfrage",
"Send full PLAY cmd", "Sende vollen PLAY Befehl",
"Serial number", "Seriennummer",
"Series link", "Serienverknüpfung",
"Series link ID.", "Serienverknüpfung-ID",
"Serious", "Seriös",
"Serious music", "Ernste Musik",
"Server connectivity", "Server Verbindung",
"Server details.", "Serverdetails.",
"Service ID", "Service-ID",
"Service Mapping (Map services to channels)", "Servicezuordnung (Services zu Kanälen zuordnen)",
"Service configuration", "Servicekonfiguration",
"Service count %zi", "Anzahl Services %zi",
"Service mapping", "Servicezuordnung",
"Service name", "Servicename",
"Service name found in EPG data.", "Servicename in den EPG-Daten gefunden.",
"Service name picons", "Servicename Picon",
"Service raw", "Unbearbeiteter Service",
"Service type", "Servicetyp",
"Service type override. This value will override the service type provided by the stream.", "Servicetyp-Überschreibung. Der vom Stream empfangene Servicetyp wird mit diesem Wert überschrieben. ",
"Services associated with the channel.", "dem Kanal zugeordnete Services.",
"Services seen", "Gefundene Services",
"Set as default profile.", "Als Standard-Profil festlegen.",
"Set the display name.", "Anzeigenamen festlegen",
"Set the priority of this configuration.", "Setze die Priorität dieser Einstellungen.",
"Sets the default interface view level (next to the Help button).", "Setzt den Standard Ansichtsmodus (neben dem Hilfe-Knopf)",
"Settings", "Einstellungen",
"Show / Game show", "Unterhaltung / Spielsendung",
"Show this page", "Zeige diese Seite",
"Show version information", "Versionsinformationen anzeigen",
"Signal quality too poor", "Signalqualität zu schlecht",
"Signal scale (240 or 100)", "Signalauflösung (240 oder 100)",
"Site latitude", "Ort Latitude",
"Site latitude.", "Ort Latitude.",
"Site longitude", "Ort Longitude",
"Site longitude.", "Ort Longitude.",
"Size", "Größe",
"Skip TS packets (0-200)", "Überspringe TS Pakete (0-200)",
"Skip commercials", "Überspringe Werbung",
"Skip initial bytes", "Überspringe erste Daten",
"Skip startup scan", "Scan beim Starten überspringen",
"Skip x number of transport packets.", "x Transportpakete überspringen",
"Slave entry", "Slave Eintrag",
"Slave entry.", "Slave Eintrag.",
"Slot state", "Slotstatus",
"Soccer", "Fußball",
"Social", "Sozial",
"Social / Political issues / Economics", "Soziale / politische Fragen / Wirtschaft",
"Social advisory", "Sozialberatung",
"Socket or IP Address (when in TCP mode).", "Socket oder IP-Adresse (wenn TCP Modus)",
"Sort index", "Sortier-Index",
"Sort index.", "Index sortieren",
"Source", "Quelle",
"Source deleted", "Quelle gelöscht",
"Source quality is bad", "Quellenqualität ist schlecht",
"Source reconfigured", "Quelle neu konfiguriert",
"Source video codec", "Videocodec der Quelle",
"Southern hemisphere (latitude direction)", "Südliche Hemisphäre (Latitude Richtung)",
"Southern hemisphere (latitude direction).", "Südliche Hemisphäre (Latitude Richtung)",
"Special events (Olympic Games, World Cup, etc.)", "Spezielle Ereignisse (Olympische Spiele, Weltmeisterschaften, usw.)",
"Specify User-Agent header for the http client", "Spezifiziere User-Agent Header für den HTTP Client",
"Specify alternative htsp port", "Spezifiziere alternativen HTSP Port",
"Specify alternative http port", "Spezifiziere alternativen HTTP Port",
"Specify alternative http webroot", "Spezifiziere alternativen HTTP Webroot",
"Specify bind address", "Spezifiziere Binde-Adresse",
"Specify bind address for SAT>IP server", "Spezifiziere Binde-Adresse für SAT>IP Server",
"Specify extra htsp port", "Spezifiziere extra HTSP Port",
"Spiritual sciences", "Geisteswissenschaften",
"Sports", "Sport",
"Sports magazines", "Sportmagazine",
"Start after", "Start nach",
"Start before", "Start vor",
"Start recording earlier than the EPG/timer-defined start time by x minutes.", "Die Aufnahme um x Minuten vor den Start laut EPG-Daten beginnen.",
"Start recording earlier than the defined Start recording earlier than the defined start time by x minutes: for example, if a program is to start at 13:00 and you set a padding of 5 minutes it will start recording at 12:54:30 (including a warm-up time of 30 seconds). If this isn't specified, any pre-recording padding as set in the channel or DVR entry will be used.", "Die Aufnahme um x Minuten vor den Start laut EPG-Daten beginnen. Zum Beispiel: Ein Programm startet um 13:00 Uhr und du hast 5 Minuten Polster angegeben, so wird die Aufnahme 12:54:30 gestartet (30 Sekunden werden als Aufwärmzeit abgezogen)\nWenn hier nichts angegeben wird, wird der angegebene Puffer vom jeweiligen Kanal genommen.",
"Start recording earlier than the defined start time by x minutes.", "Aufnahme x Minuten vor der festgelegten Startzeit beginnen",
"Start time", "Start Zeit",
"Status period (ms)", "Statuszeitraum (ms)",
"Stop time", "Stop Zeit",
"Storage path", "Speicherpfad",
"Storage space", "Speicherplatz",
"Stream ID", "Stream-ID",
"Stream Profile", "Streamprofil",
"Stream index", "Streamindex",
"Stream index (starts with zero).", "Streamindex (beginnt mit null)",
"Stream profile", "Streamprofil",
"Stream type", "Streamyp",
"Streaming Profile", "Streaming-Profil",
"Streaming priority", "Streamingpriorität",
"Streaming profiles", "Streamingprofil",
"Subdirectory options", "Unterverzeichnis-Optionen",
"Subscribe to a service permanently", "Abonniere einen Service permanent",
"Subscription", "Abonnement",
"Subscription does not exist", "Abonnement existiert nicht",
"Subscription overridden", "Abonnement überschrieben",
"Subscription weight", "Abonnementgewicht",
"Subtitle", "Untertitel",
"Subtitle Stream Filter", "Untertitel Streamfilter",
"Subtitle codec", "Untertitel-Codec",
"Subtitle of the program (if any) (display only).", "Untertitel des Programms (wenn vorhanden) (nur anzeigen).",
"Subtitle of the program (if any).", "Untertitel des Programms (wenn vorhanden).",
"Sun", "So",
"Surround (2 front, rear mono)", "Surround (2 vorne, hinten Mono)",
"Switch before rotor", "Switch dann Rotor",
"Switch frequency offset", "Frequenzversatz umstellen",
"Switch to another service", "Zu einem anderen Service schalten",
"Switch type", "Switch Typ",
"Symbol rate (Sym/s)", "Symbolrate (Sym/s)",
"Sync + Don't keep", "Sync + Nicht behalten",
"TCP Protocol", "TCP-Protokoll",
"TV Adapters - Linux ATSC-C Frontend", "TV Adapter - Linux ATSC-C Frontend",
"TV Adapters - Linux ATSC-T Frontend", "TV Adapter - Linux ATSC-T Frontend",
"TV Adapters - Linux DAB Frontend", "TV Adapter - Linux DAB Frontend",
"TV Adapters - Linux DVB-C Frontend", "TV Adapter - Linux DVB-C Frontend",
"TV Adapters - Linux DVB-S Frontend (Master)", "TV Adapter - Linux DVB-S Frontend (Master)",
"TV Adapters - Linux DVB-S Slave Frontend", "TV Adapter - Linux DVB-S Frontend (Slave)",
"TV Adapters - Linux DVB-T Frontend", "TV Adapter - Linux DVB-T Frontend",
"TV Adapters - Linux ISDB-C Frontend", "TV Adapter - Linux ISDB-C Frontend",
"TV Adapters - Linux ISDB-S Frontend (Master)", "TV Adapter - Linux ISDB-S Frontend (Master)",
"TV Adapters - Linux ISDB-T Frontend", "TV Adapter - Linux ISDB-T Frontend",
"TV Adapters - SatConfig - 4-Port", "TV Adapter - SatConfig - 4-Port",
"TV Adapters - SatConfig - Advanced", "TV Adapter - SatConfig - Erweitert",
"TV Adapters - SatConfig - DiseqC Rotor", "TV Adapter - SatConfig - DiseqCRotor",
"TV Adapters - SatConfig - DiseqC Switch", "TV Adapter - SatConfig - DiseqCSwitch",
"TV Adapters - SatConfig - GOTOX Rotor", "TV Adapter - SatConfig - GOTOX Rotor",
"TV Adapters - SatConfig - Tone Burst/2 Port", "TV Adapter - SatConfig - Tone Burst/2 Port",
"TV Adapters - SatConfig - USALS Rotor", "TV Adapter - SatConfig - USALS Rotor",
"TV Adapters - SatConfig - Universal LNB (Simple)", "TV Adapter - SatConfig - Universal LNB (Einfach)",
"TVHDHomeRun Client", "TVHDHomeRun-Client",
"Tag files with metadata", "Dateien mit Metadaten taggen",
"Talk show", "Talk-Show",
"Team sports (excluding football)", "Mannschaftssport  (außer Football)",
"Technology", "Technologie",
"Teletext Stream Filter", "Videotext Streamfilter",
"Teletext stream type is only available for this filter.", "Teletext Stream Typ ist nur für diesen Filter verfügbar.",
"Testing", "Testen",
"Testing options", "Test-Optionen",
"The CA stream type is only available for this filter.", "Der CA Stream Typ ist nur für diesen Filter verfügbar.",
"The CAM slot status.", "Status des CAM Slots",
"The DVR profile to be used/used by the recording.", "DVR Profil, das für die Aufnahme genutzt wird/wurde.",
"The DVR profile to be used/used by this rule.", "DVR Profil, das für die Regel genutzt wird/wurde.",
"The EPG ID used by the entry.", "Die EPG-Kennung dieses Eintrags",
"The EPG grabber name.", "EPG-Grabbername",
"The EPG grabber to use on the mux. Enable (auto) is the recommended value.", "Der für den Mux genutzte EPG-Grabber. Aktivieren (auto) ist der empfohlene Wert.",
"The EPG grabber type.", "EPG-Grabber Typ",
"The MPEGTS stream type is only available for this filter.", "Der MPEGTS Stream Typ ist nur für diesen Filter verfügbar.",
"The OpenTV channel number as set by the provider.", "Die OpenTV-Kanalnummer, wie sie vom Provider gesetzt wurde.",
"The PIN to use.", "Zu benutzender PIN",
"The Physical Layer Scrambling (PLS) code used on the mux.", "Der für diesen MUX genutzte Physical Layer Scrambling (PLS) Code.",
"The Physical Layer Scrambling (PLS) mode used on the mux.", "Der für diesen MUX genutzte Physical Layer Scrambling (PLS) Modus.",
"The SAT>IP server's name.", "Der SAT>IP Servername.",
"The SAT>IP server's universally unique identifier.", "Die SAT> IP Server universell eindeutige Kennung.",
"The SAT>IP source number.", "Die SAT>IP Quellennummer.",
"The SAT>IP's discovered IP address.", "Die SAT>IP entdeckte IP-Adresse.",
"The URL to the playlist.", "Die URL zu der Playlist",
"The channel name the entry will record from.", "Name des Kanals, der für die Aufnahme genutzt wird.",
"The character encoding for this network (e.g. UTF-8).", "Zeichenkodierung dieses Netzwerkes (z.B. UTF-8).",
"The character encoding for this service (e.g. UTF-8).", "Zeichenkodierung für diesen Service (z.B. UTF-8).",
"The current boot ID.", "Die aktuelle Boot-ID.",
"The current configuration ID.", "Die ID der derzeitigen Konfiguration.",
"The current configuration version.", "Tuner Konfiguration",
"The device ID.", "Die Geräte-ID.",
"The device's serial number.", "Die Geräte-Seriennummer.",
"The frequency of the mux (in Hertz).", "Die Frequenz des Mux (in Hertz)",
"The frequency of the mux/transponder in Hertz.", "Die Frequenz des Mux/Transponders in Hertz.",
"The frontend number given to the device.", "Die Frontend-Nummer, die dem Gerät gegeben wird.",
"The input path used by the device.", "Der Eingabe-Pfad des Gerätes",
"The language to which the filter should apply.", "Die Sprache, auf die der Filter angewendet werden soll.",
"The layer A constellation.", "Layer A Konstellation",
"The layer A forward error correction.", "Die layer-A-forward-error Berichtigung.",
"The layer A segment count.", "Die layer-A-Segment-Anzahl.",
"The layer A time interleaving.", "Der layer-A-Zeitversatz.",
"The layer B constellation.", "Layer B Konstellation",
"The layer B forward error correction.", "Die layer-B-forward-error Berichtigung.",
"The layer B segment count.", "Die layer-B-Segment-Anzahl.",
"The layer B time interleaving.", "Der layer-B-Zeitversatz.",
"The layer C constellation.", "Layer C Konstellation",
"The layer C forward error correction.", "Die layer-C-forward-error Berichtigung.",
"The layer C segment count.", "Die layer-C-Segment-Anzahl.",
"The layer C time interleaving.", "Der layer-C-Zeitversatz.",
"The length of time (in seconds) to play the mux (1 hour = 3600).", "Die Länge der Zeit (in Sekunden), die der Mux abgespielt werden soll (1 Stunde = 3600).",
"The limit for the PID filter (driver or hardware).", "Das Limit für den PID-Filter (Treiber oder Hardware).",
"The manufacturer of the SAT>IP server.", "Der Hersteller des SAT>IP-Servers.",
"The maximum number of services to decrypt per mux.", "Die maximale Anzahl von zu entschlüsselnden Services pro Mux.",
"The modulation used on the mux.", "Die für diesen Mux benutzte Modulation.",
"The mux the service is on.", "Der Mux in dem sich der Service befindet.",
"The mux to play when the entry is triggered.", "Abzuspielender Mux, sobald der Eintrag getriggert wird.",
"The mux's universally unique identifier.", "Die UUID des Mux.",
"The name (or frequency) the mux is on.", "Name (oder Frequenz) auf dem sich der Mux befindet.",
"The name given to/of the channel (This is how it'll appear in your EPG.)", "Der vorgegebene oder gewählte Name des Kanals (So wird er im EPG erscheinen).",
"The name of the profile.", "Name des Profils.",
"The name of the the rule.", "Der Name der Regel",
"The network prefix(es) to block, e.g.192.168.2.0/24 (comma-separated list).", "Das zu blockierende Netzwerk-Prefix(e), z.B. 192.168.2.0/24 (getrennt mit Kommas).",
"The network the mux is on.", "Das Netzwerk in dem sich der Mux befindet.",
"The network the service is on.", "Das Netzwerk in dem sich der Service befindet.",
"The network's priority. The network with the highest priority value will be used out of preference if available. See Help for details.", "Die Priorität des Netzwerks. Dem Netzwerk mit der höchsten Priorität wird der Vorzug gegeben, sobald verfügbar. Siehe Hilfe für Details.",
"The network's service ID", "Netzwerk-Service-ID",
"The networks assigned to the device.", "Das Netzwerk das diesem Gerät zugeordnet ist",
"The networks using this configuration.", "Die Netzwerke, die diese Konfiguration nutzen.",
"The networks' universally unique identifier (UUID).", "Die UUID des Netzwerkes",
"The number of ATSC-C (Cable/AnnexB) tuners to export.", "Die Anzahl zu exportierender ATSC-C (Kabel/AnnexB) Tuner.",
"The number of ATSC-T (Terresterial) tuners to export.", "Die Anzahl zu exportierender ATSC-T (Terrestrisch) Tuner.",
"The number of DVB-C (Cable) tuners to export.", "Die Anzahl zu exportierender DVB-C (Kabel) Tuner.",
"The number of DVB-C2 (Cable) tuners to export.", "Die Anzahl zu exportierender DVB-C2 (Kabel) Tuner.",
"The number of DVB-S (Satellite) tuners to export.", "Die Anzahl zu exportierender DVB-S (Satellit) Tuner.",
"The number of DVB-S2 (Satellite) tuners to export.", "Die Anzahl zu exportierender DVB-S2 (Satellit) Tuner.",
"The number of DVB-T (Terresterial) tuners to export.", "Die Anzahl zu exportierender DVB-T (Terrestrisch) Tuner.",
"The number of DVB-T2 (Terresterial) tuners to export.", "Die Anzahl zu exportierender DVB-T2 (Terrestrisch) Tuner.",
"The number of MPEG-TS packets Tvheadend buffers in case there is a delay receiving CA keys. ", "Die Anzahl an MPEG-TS-Paketen, die Tvheadend im Falle einer Verzögerung beim Empfangen von CA-Schlüsseln zwischenspeichert.",
"The number of allowed connections this user can make to the server.", "Die Anzahl an erlaubten Verbindungen, die dieser Benutzer zum Server tätigen kann.",
"The number of bytes to buffer. By default, linuxdvb's input buffer is 18800 bytes long. The accepted range is 18800-1880000 bytes.", "Die Anzahl an Bytes, die zwischengespeichert werden sollen. Standardmäßig ist der linuxdvb-Zwischenspeicher 18800 Bytes groß. Angenommen werden Werte zwischen 18800 und 1880000 Bytes.",
"The number of days cookies set by Tvheadend should expire.", "Nach wie vielen Tagen Cookies, die von Tvheadend gesetzt werden, ablaufen sollen.",
"The number of muxes left to scan on this network.", "Die Anzahl der noch zu durchsuchenden Muxe in deisem Netzwerk.",
"The number of seconds to wait for a stream to start.", "Die Anzahl an Sekunden, die gewartet werden sollen, bevor ein Stream gestartet wird.",
"The number of services currently mapped to channels.", "Die Anzahl an Services, die aktuell Kanälen zugeordnet sind.",
"The orbital position of the satellite the mux is on.", "Die orbitale Position des Satelliten, auf dem sich der Mux befindet.",
"The orbital position of the satellite your dish is pointing at.", "Die orbitale Position, auf die deine Satellitenschüssel zeigt.",
"The outcome of the last scan performed.", "Das Ergebnis des letzten durchgeführten Scans.",
"The physical layer pipe ID. Most people will not need to change this setting.", "Die physical-layer-pipe-ID. Die meisten Leute werden diese Einstellung nicht verändern müssen.",
"The playlist's character set.", "Die Zeichenkodierung der Playlist.",
"The polarization used on the mux.", "Die für diesen Mux benutzte Polarisation",
"The provider's ID.", "Die ID des Anbieters.",
"The provider's name.", "Der Name des Anbieters.",
"The provider's network ID.", "Die Netzwerk-ID des Anbieters.",
"The provider's network name.", "Der Netzwerkname des Anbieters.",
"The quadrature amplitude modulation (QAM) used by the mux. If you're not sure of the value leave as AUTO.", "Die vom Mux genutzte Quadratur-Amplituden-Modulation (QAM). Bei Unsicherheit AUTO wählen. ",
"The recorded file was removed intentionally", "Die aufgenommene Datei wurde absichtlich gelöscht",
"The recording/entry status.", "Der Aufnahme/Eintrag-Status.",
"The scheduled start time, including any padding.", "Die geplante Startzeit einschließlich aller Zugaben",
"The scheduled stop time, including any padding.", "Die geplante Endezeit einschließlich aller Zugaben",
"The selected video type should be preferred when multiple services are available for a channel.", "Der ausgewählte Videotyp sollte bevorzugt werden, wenn mehrere Services für einen Kanal verfügbar sind.",
"The service ID", "Die Service-ID",
"The service ID as set by the provider.", "Die vom Anbieter gesetzte Service-ID.",
"The service ID.", "Die Service-ID.",
"The service name as set by the provider.", "Der vom Anbieter gesetzte Servicename.",
"The service the filter should apply to. Leave blank to apply the filter to all services.", "Der Service, auf den der Filter angewendet werden soll. Leerlassen, um den Filter auf alle Services anzuwenden.",
"The service's channel minor as set by the provider.", "Die vom Anbieter gesetzte untergeordnete Kanalnummer.",
"The service's channel number as set by the provider.", "Die vom Anbieter gesetzte Kanalnummer.",
"The service's encryption status.", "Verschlüsselungsstatus dieses Service",
"The start time of the recording.", "Die Startzeit der Aufnahme.",
"The stream ID used for the mux.", "Die für diesen Mux benutzte Stream-ID.",
"The stream profile the DVR profile will use for recordings.", "Das Stream-Profil, das das DVR-Profil für die Aufnahmen verwenden soll.",
"The streaming profile to use/used. If not set, the default will be used.", "Das zu verwendende Streaming-Profil. Falls nicht gesetzt, wird die Standardeinstellung verwendet.",
"The subtitle stream types the filter should apply to.", "Die Untertitel-Stream-Typen, auf die der Filter angewendet werden soll.",
"The symbol rate used on the mux.", "Die für diesen Mux benutzte Symbolrate.",
"The symbol rate used on the mux/transponder.", "Die für diesen Mux/Transponder benutzte Symbolrate.",
"The symbol rate.", "Die Symbolrate.",
"The time the entry stops/stopped being recorded.", "Die Zeit, wann der Eintrag stoppt/die Aufnahme gestoppt wurde.",
"The title of the program to look for. Note that this accepts case-insensitive regular expressions.", "Der Titel der Sendung nach der gesucht werden soll. Beachte, dass reguläre Ausdrücke akzeptiert werden. Groß-/Kleinschreibung wird nicht beachtet.",
"The total number of services found.", "Die Gesamtanzahl an gefundenen Services.",
"The total scheduled duration.", "Die insgesamt geplante Dauer.",
"The transponder ID.", "Die Transponder-ID.",
"The transport stream ID of the mux within the network.", "Die Transport Stream-ID des Mux innerhalb des Netzwerkes.",
"The user who created the recording, or the auto-recording source and IP address if scheduled by a matching rule.", "Der Benutzer, der die Aufnahme gemacht hat oder die Auto-Aufnahme-Quelle und IP-Adresse, wenn durch eine treffende Regel aufgenommen.",
"The version of Tvheadend that last updated the config.", "Die Version von Tvheadend, die zuletzt die Konfiguration modifiziert hat.",
"The video stream types the filter should apply to.", "Die Videostream-Typen, auf die der Filter angewendet werden soll.",
"Theme", "Thema",
"This tuner", "Dieser Tuner",
"Thu", "Do",
"Time", "Zeit",
"Time (in milliseconds) for a command to complete.", "Zeit (in Millisekunden) für den Abschluss eines Befehls.",
"Time (in milliseconds) for the rotor to power up.", "Zeit (in Millisekunden) für den Start des Rotors.",
"Time (in minutes) to re-fetch the playlist.", "Zeit (in Minuten) um die Playlist erneut abzurufen.",
"Time Stamp Fix", "Zeitstempel-Anpassung",
"Time missed", "Zeit verpasst",
"Time record caption", "Zeitaufnahmebeschriftung",
"Time recording%s%s", "Zeit Aufnahme%s%s",
"Time to start the recording/time the recording started.", "Zeit, um die Aufnahme zu beginnen/Zeit, wo die Aufnahme startete",
"Time to stop recording/time the recording stopped.", "Zeit, um die Aufnahme zu stoppen/Zeit, wo die Aufnahme stoppte",
"Time update", "Zeitaktualisierung",
"Timeout (sec) (0=infinite)", "Zeitüberschreitung (Sek) (0=unendlich)",
"Timeout (seconds)", "Zeitüberschreitung (Sekunden)",
"Timeout (secs)", "Timeout (Sekunden)",
"Timer-based automatic record caption.", "Titel der zeitbasierten Auto-Aufnahme",
"Timer-based automatic recording.", "Zeitbasierte Auto-Aufnahme",
"Title", "Titel",
"Title (regexp)", "Titel (Regexp)",
"Title of the program (display only).", "Titel des Programms (nur anzeigen).",
"Title of the program.", "Titel des Programms",
"Toggle details", "Details umschalten",
"Tone burst", "Tonburst",
"Too many data errors", "Zu viele Dateifehler",
"Total number of mapped channels on this network.", "Gesamtzahl der gemappten Kanäle dieses Netzwerks",
"Total number of muxes found on this network.", "Anzahl der gefundenen Services auf diesem Netzwerk",
"Total number of services found on this network.", "Anzahl der gefundenen Services auf diesem Netzwerk",
"Total number of services found.", "Anzahl der gefundenen Services",
"Total number of services seen.", "Gesamtzahl der gesehenen Services.",
"Total number of services.", "Gesamtzahl der Services.",
"Tourism / Travel", "Tourismus / Reisen",
"Trace subsystems", "Tracesubsysteme",
"Traditional arts", "Traditionelle Kunst",
"Traditional music", "Traditionelle Musik",
"Transcode", "Transkodieren",
"Transcode/av-lib", "Transkodierung/av-lib",
"Transcoding", "Transkodierung",
"Transmission mode", "Transmission Modus",
"Transponder ID", "Transponder-ID",
"Transport stream ID", "Transport Stream-ID",
"Tue", "Di",
"Tune before DiseqC", "Tune vor DiseqC",
"Tuner bind IP address", "Tuner Bind IP-Adresse",
"Tuner configuration", "Tuner Konfiguration",
"Tuner configuration.", "Tuner Konfiguration.",
"Tuning failed", "Tuning fehlgeschlagen",
"Turn off LNB when idle", "LNB abschalten, wenn ungenutzt",
"Tvheadend server name", "Tvheadend Servername",
"Type", "Typ",
"Type override", "Typ überschreiben",
"UDP RTP port number (2 ports)", "UDP RTP Port Nummer (2 ports)",
"UHD: ultra high definition", "UHD: Ultra-High-Definition",
"UPnP Protocol", "UPnP-Protokoll",
"URL of the M3U playlist.", "URL der M3U Playlist.",
"URL with the SAT>IP server XML location", "URL der SAT>IP Server XML",
"Unable to get system UTC time", "Die UTC Systemzeit konnte nicht ermittelt werden",
"Unable to get system local time", "Die lokale Systemzeit konnte nicht ermittelt werden",
"Unable to get system time", "Die Systemzeit konnte nicht ermittelt werden",
"Unable to open file", "Datei konnte nicht geöffnet werden",
"Unable to stat path", "Pfad kann nicht gefunden werden",
"Uncommitted first", "Uncommited zuerst",
"Unicable I switch (universal LNB)", "Unicable I Switch (Universal LNB)",
"Unicable II switch (universal LNB)", "Unicable II Switch (Universal LNB)",
"Unicable type", "Unicable Typ",
"Unimportant", "Unwichtig",
"Universal LNB only", "Nur Universal LNB",
"Universally unique identifier. Read only.", "Universally Unique Identifier. Schreibgeschützt.",
"Unknown", "Unbekannt",
"Unknown file", "Unbekannte Datei",
"Unknown reason (%i)", "Unbekannter Grund (%i)",
"Unlimited size", "Unbegrenzte Größe",
"Unlimited time", "Unbegrenzte Zeit",
"Unset (default)", "Zurücksetzen (Standard)",
"Update Entitlement Management Messages from one mux only.", "Aktualisiere EMMs nur von einem Mux.",
"Update card (EMM)", "Karte aktualisieren (EMM)",
"Update channel icon", "Aktualisiere Kanallogo",
"Update channel name", "Aktualisiere Kanalname",
"Update channel number", "Aktualisiere Kanalnummer",
"Update disabled", "Aktualisierung deaktiviert",
"Update time", "Aktualisiere Zeit",
"Update tolerance (ms)", "Updatetoleranz (ms)",
"Updated", "Aktualisiert",
"Updates from one mux (EMM)", "Aktualisierung von einem Mux (EMM)",
"Upon start, Tvheadend doesn't know the last rotor position. This value defines the initial rotor movement. TVHeadend waits the specified time when the first movement is requested.", "Vor dem Start weiß Tvheadend nicht die letzte Rotorposition. Dieser Wert bestimmt die anfängliche Rotorbewegung. Tvheadend wartet die angegebene Zeit, wenn die erste Bewegung angefordert wird.",
"Usage: %s [OPTIONS]\n", "Benutzung: %s [OPTIONS]\n",
"Use AC-3 detection.", "AC-3-Erkennung verwenden.",
"Use EPG running state", "Benutze EPG Laufend Status",
"Use HTTP digest authentication", "HTTP digest Authentifizierung benutzen",
"Use PROXY protocol & X-Forwarded-For", "Verwende das PROXY-Protokoll & X-Forwarded-For",
"Use WEBM format.", "WEBM-Format nutzen",
"Use Windows-compatible filenames", "Benutze Windows-kompatible Dateinamen",
"Use XSPF playlist instead of M3U", "Benutze XSPF-Playliste statt M3U",
"Use a pre-defined list of ATSC-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Eine vordefinierte Liste von ATSC-C-Muxe nutzen. Beachte: Diese Liste kann veraltet sein und das Scannen verlangsamen.",
"Use a pre-defined list of ATSC-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Eine vordefinierte Liste von ATSC-T-Muxe nutzen. Beachte: Diese Liste kann veraltet sein und das Scannen verlangsamen.",
"Use a pre-defined list of DAB muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Eine vordefinierte Liste von DAB-Muxe nutzen. Beachte: Diese Liste kann veraltet sein und das Scannen verlangsamen.",
"Use a pre-defined list of DVB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Eine vordefinierte Liste von DVB-C-Muxe nutzen. Beachte: Diese Liste kann veraltet sein und das Scannen verlangsamen.",
"Use a pre-defined list of DVB-S/S2 muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Eine vordefinierte Liste von DVB-S/S2-Muxe nutzen. Beachte: Diese Liste kann veraltet sein und das Scannen verlangsamen.",
"Use a pre-defined list of DVB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Eine vordefinierte Liste von DVB-T-Muxe nutzen. Beachte: Diese Liste kann veraltet sein und das Scannen verlangsamen.",
"Use a pre-defined list of ISDB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Eine vordefinierte Liste von ISDB-C-Muxe nutzen. Beachte: Diese Liste kann veraltet sein und das Scannen verlangsamen.",
"Use a pre-defined list of ISDB-S muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Eine vordefinierte Liste von ISDB-S-Muxe nutzen. Beachte: Diese Liste kann veraltet sein und das Scannen verlangsamen.",
"Use a pre-defined list of ISDB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Eine vordefinierte Liste von ISDB-T-Muxe nutzen. Beachte: Diese Liste kann veraltet sein und das Scannen verlangsamen.",
"Use only this service ID, filter out others.", "Nur diese Service-ID verwenden, andere herausfiltern.",
"Use original", "Benutze Original",
"Use packet backlog", "Benutze Paket-Backlog",
"Use service IDs as channel numbers", "Service-IDs als Kanalnummern",
"Use tag internally (don't expose to clients).", "Tag intern verwenden (nicht an Clients weitergeben)",
"Use the provider's service IDs as channel numbers.", "Die Servicekennung des Providers als Kanalnummer benutzen",
"Use this character set when setting filenames.", "Diesen Zeichensatz für Dateinamen benutzen",
"Use x amount of storage space.", "Benutze x Menge des Speicherplatzes",
"User access error", "Benutzerzugriffsfehler",
"User does not have access", "Benutzer hat keinen Zugang",
"User does not have access to channel", "Benutzer hat keinen Zugang zu diesem Kanal",
"User icon", "Benutzerlogo",
"User interface level", "Ansichtsmodus wählen",
"User interface quick tips (tooltips)", "Benutzeroberflächen Quick-Tips (Kurzinfo)",
"User language", "Benutzersprache",
"User limit reached", "Benutzerlimit erreicht",
"User login", "Benutzerlogin",
"User request", "Benutzeranfrage",
"Username", "Benutzername",
"Username for the entry (login username).", "Benutzername des Eintrags (Loginname)",
"Username of the entry (this should match a username from within the \"Access Entries\" tab).", "Benutzername des Eintrags (Sollte einem Benutzernamen im \"Zugriffseinträge\"-TAB entsprechen)",
"Users - Access Entries", "Benutzer - Zugriffseinträge",
"Users - IP Blocking", "Benutzer - IP-Sperrung",
"Users - Passwords", "Benutzer - Passwörter",
"Variety show", "Varieteshow",
"Verify the SSL certificate.", "Verifiziere das SSL-Zertifikat.",
"Verify the peer's SSL.", "Überprüfe das SSL-Zertifikat der Gegenstelle.",
"Vertical resolution (height) of the output video stream. Horizontal resolution is adjusted automatically to preserve aspect ratio. When set to 0, the input resolution is used.", "Vertikale Auflösung (höhe) des Ausgabe-Video-Streams. Die horizontale Auflösung wird unter beachtung des Seitenverhältnisses automatisch bestimmt. Wenn 0 ausgwählt wird, wird die Originalauflösung beibehalten.",
"Video Stream Filter", "Video Streamfilter",
"Video bitrate (kb/s) (0=auto)", "Video-Bitrate (kb/s) (0=Auto)",
"Video codec", "Video-Codec",
"Video codec preset", "Video-Codec Voreinstellung",
"Video codec preset to use for transcoding.", "Video-Codec Voreinstellung, die für das Transkodieren genutzt wird.",
"Video codec to use for the transcode. \"Do not use\" will disable video output.", "Video-Codec, der für das Transkodieren genutzt wird. \"Nicht benutzen\" wird Video-Ausgang deaktivieren.",
"Video recorder", "Video Rekorder",
"View all", "Alle anzeigen",
"Visually impaired commentary/audio description", "Kommentar/Beschreibung für Sehbehinderte",
"WEBM/built-in", "WEBM/eingebaut",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV Profil H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV Profil H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV Profil VP8/Vorbis/WEBM",
"Waiting for EPG running flag", "Warte auf EPG Laufend Kennzeichen",
"Waiting for program start", "Warte auf Sendungsstart",
"Waiting for stream", "Warte auf Stream",
"War", "Krieg",
"Water sport", "Wassersport",
"Weak stream", "Schwacher Stream",
"Weather report", "Wetterbericht",
"Web User Interface", "Webbenutzeroberfläche",
"Web interface", "Weboberfläche",
"Web interface language", "Sprache der Weboberfläche",
"Web interface language.", "Sprache der Weboberfläche",
"Web interface theme.", "Aussehen der Weboberfläche",
"Web theme", "Web Theme",
"Web user interface", "Webbenutzeroberfläche",
"Wed", "Mi",
"Welcome - Tvheadend - your TV streaming server and video recorder", "Willkommen - Tvheadend - Dein TV Streaming Server und Videorekorder",
"Western hemisphere (latitude direction)", "Westliche Hemisphäre (Latitude Richtung)",
"Western hemisphere (latitude direction).", "Westliche Hemisphäre (Latitude Richtung).",
"When specified, this setting overrides the subdirectory rules (except the base directory) defined in the DVR configuration and puts all recordings done by this entry into the subdirectory named here. See Help for more info.", "Falls angegeben, überschreibt diese Einstellung die Unterordnerregeln (Basisordner ausgenommen), die in der DVR-Konfiguration definiert wurden und stellt alle Aufnahmen, die von diesem Eintrag getätigt werden, in den hier genannten Unterordner. Siehe Hilfe für mehr Informationen.",
"When streaming a service (via http or htsp) Tvheadend will use the network with the highest streaming priority set here. See Help for details.", "Beim Streamen eines Services (über HTTP oder HTSP) wird Tvheadend das Netzwerk mit der höchsten Priorität verwenden, die hier eingestellt ist. Siehe Hilfe für Details.",
"When the service was first identified and recorded.", "Wann der Service das erste mal identifiziert und aufgezeichnet wurde.",
"When the service was last seen during a mux scan.", "Wann der Service das letzte mal beim Mux Scan gefunden wurde.",
"Winter sports", "Wintersport",
"Wizard", "Assistent",
"Write a short message to log identifying the matched parameters. It is useful for debugging your setup or structure of incoming streams.", "Schreibt eine kurze Nachricht in die Logdateien, um die zutreffenden Parameter zu identifizieren. Dies kann nützlich zum Debuggen deines Setups oder des Aufbaus eingehender Streams sein.",
"Writes the current in-memory EPG database to disk every x hours, so should a crash/unexpected shutdown occur EPG data is saved periodically to the database (re-read on next startup). Set to 0 to disable.", "Schreibt die aktuelle, im RAM speicherte EPG-Datenbank auf die Festplatte alle x Stunden, sodass, falls ein Absturz passiert oder der Server unterwartet heruntergefahren wird, die EPG-Daten periodisch in die Datenbank geschrieben werden (werden beim nächsten Start wieder ausgelesen). Zum Deaktivieren auf 0 setzen.",
"Yes", "Ja",
"addpids/delpids supported", "addpids/delpids unterstützt",
"default: nvenc(h264 / h265)", "Standard: nvenc(h264 / h265)",
"en50494", "EN50494",
"en50607", "EN50607",
"fast: h264 / h265 / qsv(h264 / h265)", "schnell: h264 / h265 / qsv(h264 / h265)",
"faster: h264 / h265 / qsv(h264)", "schneller: h264 / h265 / qsv(h264)",
"hrs", "Std.",
"invalid option specified [%s]", "ungültige Option spezifiziert [%s]",
"logout", "Abmelden",
"medium: h264 / h265 / qsv(h264 / h265)", "mittel: h264 / h265 / qsv(h264 / h265)",
"min", "Min.",
"mins", "Min.",
"option %s requires a value", "Option %s benötigt einen Wert",
"return", "Zurück",
"slow: h264 / h265 / qsv(h264 / h265)", "langsam: h264 / h265 / qsv(h264 / h265)",
"slower: h264 / h265 / qsv(h264)", "langsamer: h264 / h265 / qsv(h264)",
"superfast: h264 / h265", "superschnell: h264 / h265",
"tsfile input (mux file)", "tsfile-Input (Mux-Datei)",
"tvhdhomerun client", "TV HDHomeRun Client",
"ultrafast: h264 / h265", "ultraschnell: h264 / h265",
"veryfast: h264 / h265 / qsv(h264)", "sehr schnell: h264 / h265 / qsv(h264)",
"veryslow: h264 / h265 / qsv(h264)", "sehr langsam: h264 / h265 / qsv(h264)",
NULL, NULL
};

static const char *tvh_locale_9[] = {
"Audioes muxer", "Audio muxer",
"Autorec maximum schedules limit (0=unlimited)", "Autorec maximum schedule limit (0=unlimited)",
"Concurrent limit per network position (src=) for satellite SAT>IP tuners. The first limit number is for src=1 (AA), second for src=2 (AB) etc.", "Concurrent limit per network position (src=) for satellite SAT>IP tuners. The first limit number is for src=1 (AA), second for src=2 (AB), etc.",
"Differentiated Services Code Point / Type of Service: Set the service class Tvheadend sends with each packet. Depending on the option selected this tells your router the prority in which to give packets sent from Tvheadend, this option does not usually need changing. See https://en.wikipedia.org/wiki/Differentiated_services for more information. ", "Differentiated Services Code Point / Type of Service: Set the service class Tvheadend sends with each packet. Depending on the option selected, this tells your router what priority to assign to packets sent from Tvheadend. This option does not usually need changing. See https://en.wikipedia.org/wiki/Differentiated_services for more information. ",
"Do not send any stream specific information to the HTSP client like signal strength, input source etc.", "Do not send any stream-specific information to the HTSP client, like signal strength, input source, etc.",
"Enable or disable ignoring of Event Information Table (EIT) data on this mux.", "Enable or disable the use of Event Information Table (EIT) data on this mux.",
"Enter allowed network prefix(es). You can enter a comma-seperated list of prefixes here.", "Enter allowed network prefix(es). You can enter a comma-separated list of prefixes here.",
"Force usage of entered CA ID on this service.", "Force usage of the entered CA ID on this service.",
"How frequently it will re-try fetching an image that has failed to be fetched.", "How frequently a failed image fetch is retried.",
"If something fails, try to switch to a different service on another network. Do not try to iterate through all inputs/tuners which are capable to receive the service.", "If something fails, try and switch to a different service on another network. Do not try to iterate through all inputs/tuners which are able to receive the service.",
"Maximum schedules limit (0=default)", "Maximum schedule limit (0=default)",
"Only activate timeshift when the client makes the first rewind, fast-forward or pause request. Note, because there is no buffer on the first request rewinding is not possible at that point.", "Only activate timeshift when the client makes the first rewind, fast-forward or pause request. Note that, because there is no buffer on the first request, rewinding is not possible at that point.",
"Only admin can watch the realtime log.", "Only admin users can watch the realtime log.",
"Only update the system clock (doesn't affect NTP driver) if the delta between the system clock and DVB time is greater than this. This can help stop excessive oscillations on the system clock.", "Only update the system clock (doesn't affect NTP driver) if the difference between the system clock and DVB time is greater than this. This can help stop excessive oscillations on the system clock.",
"Pass subscription weight to the SAT>IP server (Tvheadend specific extension).", "Pass subscription weight to the SAT>IP server (Tvheadend-specific extension).",
"Passthrough Muxer SI Tables", "Passthru Muxer SI Tables",
"Please, follow %s link and cancel the next authorization to correctly clear the cached browser credentals (login and password cache). Then click to the 'Default login' (anonymous access) or 'New login' link in the error page to reauthenticate.", "Please, follow %s link and cancel the next authorization to correctly clear the cached browser credentials (login and password cache). Then click to the 'Default login' (anonymous access) or 'New login' link in the error page to re-authenticate.",
"Program identification (PID) number to compare. Zero means any. This comparison is processed only when service comparison is active and for the Conditional Access filter.", "Program identification (PID) number to compare. Zero means any. This comparison is processed only when service comparison is active for the Conditional Access filter.",
"Real Time Streaming Protocol (RTSP) port the server should listen on (554 or 9983, 0 = disable).", "Real Time Streaming Protocol (RTSP) port on which the server should listen (554 or 9983, 0 = disable).",
"Restart when the subscription is overriden (in then timeout time window).", "Restart when the subscription is overridden (in the timeout time window).",
"Schedule a re-recording if more errors than (0=off)", "Schedule a re-recording if more errors than... (0=off)",
"Select the list of languages (in order of priority) to be used for supplying EPG information to clients that don't provide their own configuration.", "Select the list of languages (in order of priority) to be used when supplying EPG information to clients that don't provide their own configuration.",
"Select the master tuner.The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, the polarization and low-high band settings must be equal.if you set other tuner as master, then this tuner will act like a slave one and tvheadend will assure that this tuner will not use incompatible parameters (position, polarization, lo-hi).", "Select the master tuner. The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, polarization and low-high band settings must be equal. If you set another tuner as master, then this tuner will act as a slave and Tvheadend will insure that this tuner won't use incompatible parameters (position, polarization, lo-hi).",
"Send the full RTSP PLAY command after full RTSP SETUP command. Some devices firmware require this to get an MPEG-TS stream.", "Send the full RTSP PLAY command after full RTSP SETUP command. Some devices' firmware require this to get an MPEG-TS stream.",
"Set the maxiumum packet identifiers your SAT>IP server supports.", "Set the maximum number of packet identifiers your SAT>IP server supports.",
"Skip TSID checking. For when providers use invalid Transport Stream IDs.", "Skip TSID checking. Use when providers use invalid Transport Stream IDs.",
"Skip x number of transport packets.", "Skip 'n' transport packets.",
"Start recording earlier than the defined Start recording earlier than the defined start time by x minutes: for example, if a program is to start at 13:00 and you set a padding of 5 minutes it will start recording at 12:54:30 (including a warm-up time of 30 seconds). If this isn't specified, any pre-recording padding as set in the channel or DVR entry will be used.", "Start recording earlier than the defined start time by x minutes: for example, if a program is to start at 13:00 and you set a padding of 5 minutes it will start recording at 12:54:30 (including a warm-up time of 30 seconds). If this isn't specified, any pre-recording padding as set in the channel or DVR entry will be used.",
"The audio stream types the filter should apply to.", "The audio stream types to which the filter should apply.",
"The locking mechanism selection for The Preferred Conditional Access Packet Identifier. See Help for more information.", "The locking mechanism selection for the preferred Conditional Access Packet Identifier. See Help for more information.",
"The logical stream index to compare. Note that this index is computed using all filters.Example: If filter is set to AC3 audio type and the language to 'eng' and there are two AC3 'eng' streams in the service, the first stream could be identified using number 1 and the second using number 2.", "The logical stream index to compare. Note that this index is computed using all filters. For example, if you're filtering on AC3 audio with a language of 'eng', then two AC3 'eng' streams in the service can be simply identified as indices 1 and 2.",
"The maximal duration of a matching event - in other words, only match programmes that are no longer than this duration.", "The maximal duration of a matching event - in other words, only match programs that are no longer than this duration.",
"The name given to/of the channel (This is how it'll appear in your EPG.)", "The name given to/of the channel (This is how it'll appear in your EPG).",
"The number of days cookies set by Tvheadend should expire.", "The number of days in which any cookies set by Tvheadend should expire.",
"The service the filter should apply to. Leave blank to apply the filter to all services.", "The service to which the filter should apply. Leave blank to apply the filter to all services.",
"The subtitle stream types the filter should apply to.", "The subtitle stream types to which the filter should apply.",
"The time window to cut the stop time from the overlapped event in seconds.", "The time window to cut the stop time from an overlapped event in seconds.",
"The video stream types the filter should apply to.", "The video stream types to which the filter should apply.",
"This will create an NTP driver (using shmem interface) that you can feed into ntpd. This can be run without root privileges, but generally the performance is not that great.", "This will create an NTP driver (using a shmem interface) that you can feed into ntpd. This can be run without root privileges, but generally the performance is not that great.",
"Try to obtain channel numbers from the display-name xml tag. If the first word is number, it is used as the channel number.", "Try to obtain channel numbers from the display-name XML tag. If the first word is number, it is used as the channel number.",
"Use a pre-defined list of DVB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Use a pre-defined list of DVB-T/T2 muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.",
"Writes the current in-memory EPG database to disk every x hours, so should a crash/unexpected shutdown occur EPG data is saved periodically to the database (re-read on next startup). Set to 0 to disable.", "Writes the current in-memory EPG database to disk every x hours, so it can be re-read on the next startup (e.g. in case an unexpected shutdown occurs). Set to 0 to disable.",
NULL, NULL
};

static const char *tvh_locale_10[] = {
"Add this value to PTS for the teletext subtitles. The time value is in milliseconds and may be negative.", "Add this value to PTS for the Teletext subtitles. The time value is in milliseconds and may be negative.",
"Altitude (meters)", "Altitude (metres)",
"Anonymize", "Anonymise",
"Audioes muxer", "Audio muxer",
"Auto-Map to channels", "Auto-map to channels",
"Autorec maximum schedules limit (0=unlimited)", "Autorec maximum schedule limit (0=unlimited)",
"Children's / Youth Programs", "Children's / Youth Programmes",
"Children's / Youth programs", "Children's / Youth programmes",
"DVR behavior", "DVR behaviour",
"Differentiated Services Code Point / Type of Service: Set the service class Tvheadend sends with each packet. Depending on the option selected this tells your router the prority in which to give packets sent from Tvheadend, this option does not usually need changing. See https://en.wikipedia.org/wiki/Differentiated_services for more information. ", "Differentiated Services Code Point / Type of Service: Set the service class Tvheadend sends with each packet. Depending on the option selected, this tells your router the prority to assign to packets sent from Tvheadend. This option does not usually need changing. See https://en.wikipedia.org/wiki/Differentiated_services for more information. ",
"Discard the frontend parameter in RTSP requests, as some clients incorretly use it.", "Discard the front-end parameter in RTSP requests, as some clients incorrectly use it.",
"Electronic Program Guide", "Electronic Programme Guide",
"Enable if the SAT>IP box supports the frontend identifier. This allows the auto-tuner allocation, but it might cause trouble for boxes with different tuner reception connections like satellite inputs.", "Enable if the SAT>IP box supports the front-end identifier. This allows automatic tuner allocation, but it might cause trouble for boxes with different tuner reception connections like satellite inputs.",
"Enable or disable ignoring of Event Information Table (EIT) data on this mux.", "Enable or disable the use of Event Information Table (EIT) data on this mux.",
"Enable over-the-air program guide (EPG) scanning on this input device.", "Enable over-the-air programme guide (EPG) scanning on this input device.",
"Enter allowed network prefix(es). You can enter a comma-seperated list of prefixes here.", "Enter allowed network prefix(es). You can enter a comma-separated list of prefixes here.",
"Enter external PORT if behind Forwarding redirection.(0 = use the same local port).", "Enter external PORT if behind forwarding redirection.(0 = use the same local port).",
"Entertainment programs for 10 to 16", "Entertainment programmes for 10 to 16",
"Entertainment programs for 6 to 14", "Entertainment programmes for 6 to 14",
"Gray", "Grey",
"How frequently it will re-try fetching an image that has failed to be fetched.", "How frequently a failed image fetch is retried.",
"Include teletext", "Include Teletext",
"Include teletext in the timeshift buffer. Enabling this may cause issues with some services where the teletext DTS is invalid.", "Include Teletext in the timeshift buffer. Enabling this may cause issues with some services where the Teletext DTS is invalid.",
"Keep timeshift buffers in RAM only. With this option enabled, the amount of rewind time is limited by how much RAM TVHeadend is allowed.", "Keep timeshift buffers in RAM only. With this option enabled, the amount of rewind time is limited by how much RAM Tvheadend is allowed.",
"MPEG-TS Pass-thru", "MPEG-TS Pass-through",
"MPEG-TS Pass-thru/built-in", "MPEG-TS Pass-through/built-in",
"Maximum schedules limit (0=default)", "Maximum schedule limit (0=default)",
"Name (or date) of program the entry is a rerun of.", "Name (or date) of programme the entry is a rerun of.",
"Only activate timeshift when the client makes the first rewind, fast-forward or pause request. Note, because there is no buffer on the first request rewinding is not possible at that point.", "Only activate timeshift when the client makes the first rewind, fast-forward or pause request. Note that, because there is no buffer on the first request, rewinding is not possible at that point.",
"Only update the system clock (doesn't affect NTP driver) if the delta between the system clock and DVB time is greater than this. This can help stop excessive oscillations on the system clock.", "Only update the system clock (doesn't affect NTP driver) if the difference between the system clock and DVB time is greater than this. This can help stop excessive oscillations on the system clock.",
"PIN inquiry match string", "PIN enquiry match string",
"PROXY protocol is an extension for support incoming TCP connections from a remote server (like a firewall) sending the original IP address of the client. The HTTP header 'X-Forwarded-For' do the same with HTTP connections. Both enable tunneled connections.This option should be disabled for standard usage.", "PROXY protocol is an extension for support incoming TCP connections from a remote server (like a firewall) sending the original IP address of the client. The HTTP header 'X-Forwarded-For' do the same with HTTP connections. Both enable tunnelled connections. This option should be disabled for standard usage.",
"Pass subscription weight to the SAT>IP server (Tvheadend specific extension).", "Pass subscription weight to the SAT>IP server (Tvheadend-specific extension).",
"Pass-thru muxer", "Pass-through muxer",
"Please, follow %s link and cancel the next authorization to correctly clear the cached browser credentals (login and password cache). Then click to the 'Default login' (anonymous access) or 'New login' link in the error page to reauthenticate.", "Please, follow %s link and cancel the next authorization to correctly clear the cached browser credentials (login and password cache). Then click to the 'Default login' (anonymous access) or 'New login' link in the error page to re-authenticate.",
"Polarization", "Polarisation",
"Pre-school children's programs", "Pre-school children's programmes",
"Program identification (PID) number to compare. Zero means any. This comparison is processed only when service comparison is active and for the Conditional Access filter.", "Program identification (PID) number to compare. Zero means any. This comparison is processed only when service comparison is active for the Conditional Access filter.",
"Program synopsis (display only).", "Programme synopsis (display only).",
"Program synopsis.", "Programme synopsis.",
"Real Time Streaming Protocol (RTSP) port the server should listen on (554 or 9983, 0 = disable).", "Real Time Streaming Protocol (RTSP) port on which the server should listen (554 or 9983, 0 = disable).",
"Schedule a re-recording if more errors than (0=off)", "Schedule a re-recording if more errors than... (0=off)",
"School programs", "School programmes",
"Select the list of languages (in order of priority) to be used for supplying EPG information to clients that don't provide their own configuration.", "Select the list of languages (in order of priority) to be used when supplying EPG information to clients that don't provide their own configuration.",
"Select the master tuner.The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, the polarization and low-high band settings must be equal.if you set other tuner as master, then this tuner will act like a slave one and tvheadend will assure that this tuner will not use incompatible parameters (position, polarization, lo-hi).", "Select the master tuner. The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, polarization and low-high band settings must be equal. If you set another tuner as master, then this tuner will act as a slave and Tvheadend will ensure that this tuner won't use incompatible parameters (position, polarization, lo-hi).",
"Send the full RTSP PLAY command after full RTSP SETUP command. Some devices firmware require this to get an MPEG-TS stream.", "Send the full RTSP PLAY command after full RTSP SETUP command. Some devices' firmware require this to get an MPEG-TS stream.",
"Set the maxiumum packet identifiers your SAT>IP server supports.", "Set the maximum number of packet identifiers your SAT>IP server supports.",
"Skip TSID checking. For when providers use invalid Transport Stream IDs.", "Skip TSID checking. Use when providers use invalid Transport Stream IDs.",
"Skip x number of transport packets.", "Skip 'n' transport packets.",
"Some DVB devices have an optional ioctl that allows changing between normal voltage for LNB (13V/18V) to a higher voltage mode (usually, 14V/19V), meant to compensate for voltage loss on long cabling. Without that, it is not possible to properly switch the polarization.", "Some DVB devices have an optional ioctl that allows a change between normal voltage for LNB (13V/18V) to a higher voltage mode (usually, 14V/19V). This is meant to compensate for voltage loss on long cabling. Without this, it is not possible to properly switch polarisation.",
"Start recording earlier than the defined Start recording earlier than the defined start time by x minutes: for example, if a program is to start at 13:00 and you set a padding of 5 minutes it will start recording at 12:54:30 (including a warm-up time of 30 seconds). If this isn't specified, any pre-recording padding as set in the channel or DVR entry will be used.", "Start recording earlier than the defined start time by x minutes: for example, if a programme is to start at 13:00 and you set a padding of 5 minutes, it will start recording at 12:54:30 (including a warm-up time of 30 seconds). If this isn't specified, any pre-recording padding as set in the channel or DVR entry will be used.",
"Subtitle of the program (if any) (display only).", "Subtitle of the programme (if any) (display only).",
"Subtitle of the program (if any).", "Subtitle of the programme (if any).",
"The audio stream types the filter should apply to.", "The audio stream types to which the filter should apply.",
"The content type (Movie/Drama, Sports, etc.) to be used to filter matching events/programs.", "The content type (Movie/Drama, Sports, etc.) to be used to filter matching events/programmes.",
"The default web interface to use if the user  theme isn't set in the Access Entries tab.", "The default web interface to use if the user theme isn't set in the Access Entries tab.",
"The locking mechanism selection for The Preferred Conditional Access Packet Identifier. See Help for more information.", "The locking mechanism selection for the preferred Conditional Access Packet Identifier. See Help for more information.",
"The logical stream index to compare. Note that this index is computed using all filters.Example: If filter is set to AC3 audio type and the language to 'eng' and there are two AC3 'eng' streams in the service, the first stream could be identified using number 1 and the second using number 2.", "The logical stream index to compare. Note that this index is computed using all filters. For example, if you're filtering on AC3 audio with a language of 'eng', then two AC3 'eng' streams in the service can be simply identified as indices 1 and 2.",
"The minimal duration of a matching event - in other words, only match programs that are no shorter than this duration.", "The minimal duration of a matching event - in other words, only match programmes that are no shorter than this duration.",
"The number of days cookies set by Tvheadend should expire.", "The number of days in which any cookies set by Tvheadend should expire.",
"The service the filter should apply to. Leave blank to apply the filter to all services.", "The service to which the filter should apply. Leave blank to apply the filter to all services.",
"The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, the polarization and low-high band settings must be equal. If you set another tuner as master, then this tuner will act as a slave and Tvheadend will assure that this tuner will not use incompatible parameters (position, polarization, lo-hi).", "The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, the polarisation and low-high band settings must be equal. If you set another tuner as master, then this tuner will act as a slave and Tvheadend will assure that this tuner will not use incompatible parameters (position, polarisation, lo-hi).",
"The subtitle stream types the filter should apply to.", "The subtitle stream types to which the filter should apply.",
"The title of the program to look for. Note that this accepts case-insensitive regular expressions.", "The title of the programme to look for. Note that this accepts case-insensitive regular expressions.",
"The video stream types the filter should apply to.", "The video stream types to which the filter should apply.",
"This will create an NTP driver (using shmem interface) that you can feed into ntpd. This can be run without root privileges, but generally the performance is not that great.", "This will create an NTP driver (using a shmem interface) that you can feed into ntpd. This can be run without root privileges, but generally the performance is not that great.",
"Title of the program (display only).", "Title of the programme (display only).",
"Title of the program.", "Title of the programme.",
"Use a pre-defined list of DVB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Use a pre-defined list of DVB-T/T2 muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.",
"Waiting for program start", "Waiting for programme start",
"Writes the current in-memory EPG database to disk every x hours, so should a crash/unexpected shutdown occur EPG data is saved periodically to the database (re-read on next startup). Set to 0 to disable.", "Writes the current in-memory EPG database to disk every x hours, so it can be re-read on the next startup (e.g. in case an unexpected shutdown occurs). Set to 0 to disable.",
NULL, NULL
};

static const char *tvh_locale_11[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nPara más información, por favor visita el sitio web de Tvheadend:\nhttps://tvheadend.org\n",
"# Mapped channels", "# Canales mapeados",
"# Services", "# Servicios",
"# tune repeats", "# Repitir sintonización",
"(Default profile)", "(Perfil por defecto)",
"(Read-only)", "(Sólo lectura)",
"1 day", "1 día",
"1 month", "1 mes",
"1 week", "1 semana",
"1 year", "1 año",
"2 months", "2 meses",
"2 weeks", "2 semanas",
"2 years", "2 años",
"2-Port switch (universal LNB)", "Switch de 2 puertos (LNB universal)",
"3 days", "3 días",
"3 months", "3 meses",
"3 weeks", "3 semanas",
"3 years", "3 años",
"4-Port switch (universal LNB)", "Switch de 4 puertos (LNB universal)",
"5 days", "5 días",
"6 months", "6 meses",
"A channel tag (e.g. a group of channels) to which this rule applies.", "Una categoría de canal (p.e. un grupo de canales) a los cuales se aplica esta regla.",
"AAC audio", "Audio AAC",
"AC-3 = descriptor 6", "AC-3 = Descriptor 6",
"AC-3 detection", "Detección AC-3",
"AC3 audio", "Audio AC3",
"ACTIVE", "ACTIVO",
"AES Constant Code Word", "AES Constante Código Mundial",
"ANY", "CUALQUIERA",
"ATSC SI Tables", "Tablas ATSC SI",
"ATSC source ID", "Fuente ID ATSC",
"ATSC-C Network", "Red ATSC-C",
"ATSC-C multiplex", "Multiplexo ATSC-C",
"ATSC-T Network", "Red ATSC-T",
"ATSC-T multiplex", "Multiplexo ATSC-T",
"Aborted by user", "Cancelado por el usuario",
"Accept remote subscription weight", "Acepta el peso de suscripción remoto",
"Accept the remote subscription weight (from the SAT>IP client).", "Acepte la fuerza de suscripción a distancia (desde el cliente SAT>IP).",
"Accept zero value for TSID", "Aceptar valor cero para TSID",
"Access", "Acceso",
"Access (ACL)", "Acceso (ACL)",
"Action", "Acción",
"Active", "Activo",
"Adapter in use by another subscription", "Adaptador en uso por otra suscripción",
"Add file and line numbers to debug", "Añade archivo y número de línea para depurar",
"Add the thread ID to debug", "Añade el ID de la tarea para depurar",
"Add this value to PTS for the teletext subtitles. The time value is in milliseconds and may be negative.", "Agregue este valor a PTS para los subtítulos de teletexto. El valor de tiempo es en milisegundos y puede ser negativo.",
"Additional arguments to pass to the grabber.", "Argumentos adicionales para pasar al capturador.",
"Additional service names found in EPG data.", "Nombres de servicio adicionales encontrados en los datos EPG.",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "Tiempo adicional (en segundos) de espera para que el sintonizador esté listo para grabar. Útil para aquellos sintonizadores que necesitan un tiempo para sintonizar y/o envían datos basura al principio.",
"Admin", "Administrador",
"Admin password", "Contraseña de administrador",
"Admin username", "Nombre de usuario de administrador",
"Administrator login", "Login de administrador",
"Adult movie", "Película para Adultos",
"Advanced", "Avanzado",
"Advanced (non-universal LNBs, rotors, etc.)", "Avanzado (LNBs no-universales, rotores, etc.)",
"Adventure", "Aventuras",
"Advertisement / Shopping", "Publicidad / Compras",
"After creating the network scan it for services.", "Tras crear la red escanearla para servicios.",
"All (Streaming plus DVR)", "Todos (Retransmisión más DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Todos los caracteres que pueda causar problemas para el nombrado de archivos serán reemplazados con un guión bajo. Ver Ayuda para más detalles.",
"All lower-case", "Todo en minúsculas",
"All proceeds are used to support server infrastructure and buy test equipment.", "Todos los fondos recaudados se invertirán en la infraestructura de servidores y comprar equipos de prueba.",
"All: Record if different description", "Todos: Grabar si la descripción es diferente",
"All: Record if different episode number", "Todos: Grabar si el número de episodio es diferente",
"All: Record if different subtitle", "Todos: Grabar si los subtítulos son diferentes",
"All: Record once per day", "Todos: Grabar una vez al día",
"All: Record once per month", "Todos: Grabar una vez al mes",
"All: Record once per week", "Todos: Grabar una vez a la semana",
"Allow all PIDs", "Permitir todos los PIDs",
"Allow all PIDs (no filter) when the 'Maximum PIDs' limit is reached.", "Permitir todos los PIDs (sin filtrar) cuando \"Máximos PIDs\" es alcanzado.",
"Allow high bitrate mode (CI+ CAMs only).", "Habilitar modo de alto bitrate (sólo CAMs CI+).",
"Allow idle scan tuning on this device.", "Habilitar el escaneo de sintonización cuando está inactivo en este dispositivo.",
"Allow the combined size of all timeshift buffers to potentially grow unbounded until your storage media runs out of space.", "Permitir que el tamaño combinado de todos los búferes timeshift crezca sin límites hasta que el dispositivo de almacenamiento se quede sin espacio.",
"Allow the initial scan tuning on this device (scan when Tvheadend starts or when a new multiplex is added automatically). At least one tuner or input should have this settings turned on. See also 'Skip Startup Scan' in the network settings for further details.", "Permitir la exploración inicial de en este dispositivo (escanear cuando Tvheadend se inicia o cuando un nuevo multiplex se agrega automáticamente). Al menos un sintonizador o entrada debe tener esta configuración activada. Consulte también \"Omitir la exploración de inicio\" en la configuración de red para obtener más detalles.",
"Allow the timeshift buffer to grow unbounded until your storage media runs out of space. Warning, enabling this option may cause your system to slow down or crash completely!", "Permitir que el tamaño del búfer timeshift crezca sin límites hasta que el dispositivo de almacenamiento se quede sin espacio. ¡Atención, activar esta opción puede hacer que su sistema se ralentice o sea completamente inestable!",
"Allow/Disallow access to the 'Configuration' tab.", "Permitir/denegar el acceso a la pestaña \"Configuración\".",
"Allow/Disallow web interface access (this  includes access to the EPG).", "Permitir/Denegar acceso a la interfaz web (esto incluye acceso a la EPG).",
"Allowed DVR profiles. This limits the profiles the user has access to.", "Perfiles DVR permitidos. Esto limita los perfiles a los que tiene acceso el usuario.",
"Allowed network", "Red permitida",
"Allowed networks", "Redes permitidas",
"Alternate PID path", "Ruta alternativa del PID",
"Alternate configuration path", "Ruta alternativa de la configuracion",
"Altitude (in meters).", "Altitud (en metros)",
"Altitude (meters)", "Altitud (metros)",
"Always send the whole DiseqC sequence including LNB setup (voltage, tone). If this is not checked, only changed settings are sent, which may cause issues with some drivers. If the tuning is not reliable, try activating this option.", "Siempre enviar la secuencia DiseqC completa incluyendo la configuración del LNB (voltaje, tono). Si no está activado, sólo se envían los parámetros modificados, lo cual puede causar problemas con algunos drivers. Si la sintonización no es correcta, pruebe a activar esta opción.",
"Always use the name defined by the network.", "Siempre utilizar el nombre definido por la red.",
"Always use the old ioctls to read the linuxdvb status (signal strength, SNR, error counters). Some drivers are not mature enough to provide the correct values using the new v5 linuxdvb API.", "Siempre utilizar el antiguo ioctls para obtener el estado de linuxdvb (fuerza de señal, SNR, contadores de error). Algunos drivers no son suficientemente estables para facilitar los valores correctos con la nueva API v5 de linuxdvb.",
"An event which starts between this \"start after\" and \"start before\" will be matched (including boundary values).", "Un evento que se inicia entre \"se inicia después\" y \"empezar antes de\" será igualado (incluidos los valores límites).",
"Animals", "Animales",
"Anonymize", "Anonimizar",
"Anonymize HTSP access", "Anonimizar acceso HTSP",
"Any", "Cualquiera",
"Arts", "Artes",
"Arts / Culture (without music)", "Arte / Cultura (sin música)",
"Arts magazines", "Revistas de artes",
"Assign predefined muxes to networks", "Asignar muxes predefinidos a las redes",
"Associate this device with one or more networks.", "Asociar este dispositivos con una o mas redes.",
"Athletics", "Atletismo",
"Audio Stream Filter", "Filtro de Retransmisión de Audio",
"Audio bitrate (kb/s) (0=auto)", "Tasa de audio (kb/s) (0=Auto)",
"Audio bitrate to use for transcoding.", "Tasa de audio para la transcodificación.",
"Audio channel layout.", "Distribución de canales de audio.",
"Audio codec", "Códec de audio",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Codec de Audio usado para transcodificar. \"No usar\" deshabilitará la salida de audio.",
"Audio stream", "Retransmisión de audio",
"Audio type", "Tipo de audio",
"Audio-only stream", "Retransmisión de solo audio",
"Audioes muxer", "Mux Audioes",
"Authenticated user", "Usuario autenticado",
"Auto", "Automático",
"Auto check disabled", "Autocomprobación deshabilitada",
"Auto check enabled", "Autocomprobación habilitada",
"Auto record", "Grabar automáticamente",
"Auto record caption", "Etiqueta de grabación automática",
"Auto recording%s%s", "Grabación Automática%s%s",
"Auto time record", "Tiempo de grabación automático",
"Auto-Map to channels", "Auto-Mapear a los canales",
"Automatic checking", "Comprobar automáticamente",
"Automatic recording caption.", "Etiquetado automático de grabación.",
"Automatic schedule entry not found", "Entrada de la programación automática no encontrada",
"Automatic time scheduler entry not found", "Entrada del planificador de tiempo automático no encontrado",
"Automatically link EPG data to the channel (using the channel name for matching). If you turn this option off, only the OTA EPG grabber will be used for this channel unless you've specifically set a different EPG Source.", "Enlace automáticamente los datos EPG al canal (utilizando el nombre del canal para hacer coincidir). Si desactiva esta opción, sólo se utilizará el capturador EPG OTA para este canal a menos que haya definido específicamente una Fuente de EPG diferente.",
"Automatically map EPG source", "Mapear automáticamente la fuente EPG.",
"Automatically map all available services to channels.", "Mapear automáticamente todos los servicios disponibles a canales.",
"Automatically map channels defined within the bouquet.", "Mapear automáticamente los canales definidos en el paquete.",
"Automatically name from network", "Nombrar automáticamente desde la red",
"Automatically record.", "Grabación automática.",
"Automatically update channel icons using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel icons.", "Actualizar automáticamente los iconos de los canales utilizando la información proporcionada por los proveedores de EPG habilitados. Nota: esto puede causar cambios no deseados en iconos ya definidos en un canal.",
"Automatically update channel names using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel names.", "Actualizar automáticamente los nombres de los canales utilizando la información proporcionada por los proveedores de EPG habilitados. Nota: esto puede causar cambios no deseados en el nombre de canales ya definidos.",
"Automatically update channel numbers using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel numbers.", "Actualizar automáticamente los números de canal utilizando la información proporcionada por los proveedores de EPG habilitados. Tenga en cuenta que esto puede causar cambios no deseados en los números de canal ya definidos.",
"Autorec maximum count (0=unlimited)", "Recuento máximo de auto grabación (0=ilimitado)",
"Autorec maximum schedules limit (0=unlimited)", "Límite de cola máximo de auto grabación (0=ilimitado)",
"Bad", "Mal",
"Bad request", "Petición fallida",
"Ballet", "Balet",
"Bandwidth", "Ancho de Banda",
"Base DVB SI Tables (PAT,CAT,PMT,SDT etc.)", "Tablas Base DVB SI (PAT,CAT,PMT,SDT etc.)",
"Based on software from", "Basado en el software de",
"Basic", "Básico",
"Bind to specific local IP address.", "Enlazar a la dirección IP local específica.",
"Bitrate to use for the transcode. See Help for details.", "Bitrate a usar en transcodificación. Mira la Ayuda para mas detalles.",
"Blue", "Azul",
"Boot ID", "ID de Arranque",
"Bouquet", "Paquete",
"Bouquet (auto)", "Paquete (automático)",
"Bouquet source.", "Fuente de paquete.",
"Bouquets", "Paquetes",
"Brand", "Marca",
"Branding information (if available).", "Información de personalización (si está disponible).",
"Broadcast", "Emisión",
"Broadcast type", "Tipo de emisión",
"Broadcast.", "Emisión.",
"Broadcasting", "Radiodifusión",
"Buffering limit (ms)", "Limite de buffer (ms)",
"Build", "Compilación",
"By default, linuxdvb's status read period is 1000ms (one second). The accepted range is 250ms to 8000ms. Note that for some hardware or drivers (like USB), the status operations take too much time and CPU. In this case, increase the default value. For fast hardware, this value might be decreased to make the decision of the re-tune algorithm based on the signal status faster.", "De manera predeterminada, el período de leer el estado de linuxdvb es 1000ms (un segundo). El rango aceptado es de 250ms a 8000ms. Tenga en cuenta que en algún hardware o controladores (como USB), las operaciones de estado toman demasiado tiempo y CPU. En este caso, aumente el valor predeterminado. Para el hardware rápido, este valor podría ser disminuido para tomar la decisión de de re-sintonizar basándose en el estado de la señal más rápidamente.",
"CA (descrambling) Client", "CA (desencriptando) Cliente",
"CA ID", "ID CA",
"CA Stream Filter", "Filtro de Retransmisión CA",
"CA client %i", "Cliente %i CA",
"CA identification", "Identificación CA",
"CA provider", "Proveedor CA",
"CAPMT (Linux Network DVBAPI)", "CAPMT (Red Linux DVBAPI)",
"CAPMT CA Client", "Cliente CAPMT CA",
"CAPMT interval (in ms).", "Intervalo CAPMT (en ms)",
"CAPMT interval (ms)", "Intervalo CAPMT (ms)",
"CAPMT query interval (ms)", "Intervalo de consulta (ms) CAPMT",
"CAPMT query interval (ms).", "CAPMT Intervalo de consulta (ms) ",
"CI Module", "Módulo CI",
"CRASH", "CRASHEO",
"CRID authority", "Autoridad CRID",
"CSA (descrambling)", "CSA (descifrando)",
"CWC CA Client", "Cliente CWC CA",
"Cache scheme", "Esquema de caché",
"Camd.socket filename / IP Address (TCP mode)", "Nombre del archivo camd.socket / Dirección IP (modo TCP)",
"Cartoons", "Dibujos Animados",
"Change parameters", "Cambiar parámetros",
"Channel", "Canal",
"Channel Tags", "Categorías del Canal",
"Channel does not exist", "El canal no existe",
"Channel icon", "Icono del canal",
"Channel icon URL.", "URL de icono del Canal.",
"Channel icon as defined in EPG data.", "Icono de canal tal y como está definido en los datos EPG.",
"Channel icon name scheme", "Esquema de nombrado de icono de canal",
"Channel icon path", "Ruta del Icono del Canal",
"Channel mapping options", "Opciones de mapeo de canal",
"Channel name", "Nombre del canal",
"Channel not found", "Canal no encontrado",
"Channel number", "Número de Canal",
"Channel number as defined in EPG data.", "Número de canal tal y como esta definido en el EPG.",
"Channel number offset", "Compensación de número de canal",
"Channel number range", "Rango de números de canal",
"Channel numbers (heuristic)", "Números de Canal (heurístico).",
"Channel numbers from", "Números de canales desde",
"Channel tag", "Categoría del canal",
"Channel tag does not exist", "La categoría del Canal no existe.",
"Channel tag reference", "Referencia de etiqueta del canal",
"Channel tags", "Categorías del canal",
"Channel tags the user is allowed access to/excluded from.", "Etiquetas de canal a las cuales el usuario está autorizado/desautorizado a acceder.",
"Channel to use/used for the recording.", "Canal a utilizar/utilizado para la grabación.",
"Channel update options", "Opciones de actualización de canal",
"Channels", "Canales",
"Channels EPG data is used by.", "Los datos de EPG de los Canales son usados por ",
"Character set", "Juego de carácteres",
"Characters not supported in Windows filenames (e.g. for an SMB/CIFS share) will be stripped out or converted.", "Los caracteres no soportados en nombres de archivo de Windows (p.e. para un recurso compartido SMB/CIFS) serán eliminados o sustituidos.",
"Charset", "Juego de carácteres",
"Check availability", "Comprobar disponibilidad",
"Check for the services' presence. If the service is no longer broadcast this field will change to Missing In PAT/SDT.", "Compruebe la presencia de los servicios. Si el servicio ya no se transmite, este campo cambiará a Ausente en PAT/SDT.",
"Check services for availability. If enabled, services that are not currently broadcasting (or can't be decrypted) will be ignored. Leave disabled if you want Tvheadend to also map offline services.", "Compruebe la disponibilidad de servicios. Si está activado, los servicios que no están emitiendo actualmente (o no se pueden descifrar) se ignorarán. Déjelo desactivado si quieres que Tvheadend también coloque los servicios offline.",
"Check the descrambling status after this timeout.", "Comprobar el estado de descifrado tras el tiempo agotado.",
"Children's / Youth Programs", "Programas para Niños / Jovenes",
"Children's / Youth programs", "Programas para Niños / Jovenes",
"Cinema", "Cine",
"Class", "Clase",
"Classical", "Clásico",
"Classical music", "Música Clásica",
"Clean effects", "Efectos Limpios",
"Client name", "Nombre del cliente",
"Clone scheduled entry on error", "Clonar entrada programada en error",
"Code Word Client (newcamd)", "Código de Palabra del Cliente (newcamd)",
"Comedy", "Comedia",
"Command delay time (ms) (10-200)", "Tiempo de retraso (ms) (10-200)",
"Command time (ms) (10-100)", "Tiempo de comando (ms) (10-100)",
"Comment", "Comentario",
"Commercial break", "Anuncios",
"Commercials will be dropped from the recordings. Commercial detection works using EITp/f (EPG running state) and for the Swedish channel TV4 (using teletext info).", "La publicidad comercial será eliminada de las grabaciones. Por el momento, la detección de publicidad comercial sólo funciona para el canal sueco TV4.",
"Completed OK", "Completado OK",
"Compress EPG database", "Comprimir la base de datos de EPG",
"Compress the EPG database to reduce disk I/O and space.", "Comprimir la base de datos de EPG para reducir la E/S del disco y el espacio utilizado.",
"Concurrent limit per network position (src=) for satellite SAT>IP tuners. The first limit number is for src=1 (AA), second for src=2 (AB) etc.", "Límite concurrente por posición de la red (src=) para sintonizadores satélite SAT>IP. El primero es el número límite para src=1 (AA), en segundo lugar para src=2 (AB), etc.",
"Conditional Access", "Acceso Condicional",
"Conditional Access Client", "Acceso Condicional de Cliente",
"Conditional Access Identification.", "Identificación de Acceso Condicional",
"Configuration", "Configuración",
"Configuration - Base", "Configuración - Base",
"Configuration - Image Cache", "Configuración - Caché de Imagen",
"Configuration - SAT>IP Server", "Configuración - Servidor SAT>IP",
"Configuration ID", "ID de Configuración",
"Configuration name", "Nombre de configuración",
"Configuration version", "Versión de Configuración",
"Connection limit type", "Tipo de límite de conexión",
"Connection limits", "Límite de conexión",
"Constellation", "Constelación",
"Container", "Contenedor",
"Container to use for the transcoded stream.", "Contenedor a utilizar para la retransmisión transcodificada.",
"Content character set", "Contenido de carácteres seleccionado",
"Content reference identifier authority.", "Contenido autorizado para un identificador de referencia.",
"Content type", "Tipo de contenido",
"Content type.", "Tipo de contenido.",
"Contest", "Concurso",
"Continue if descrambling fails", "Continuar si el descifrado falla",
"Continue recording for x minutes after scheduled stop time", "Continuar grabando durante x minutos tras la hora de finalización prevista.",
"Continue recording for x minutes after scheduled stop time.", "Continuar grabando durante x minutos tras la hora de finalización prevista.",
"Cookie expiration (days)", "Expiración de cookies (días)",
"Cooking", "Cocina",
"Copy codec type", "Copiar Tipo de Codec",
"Copy layout", "Composicion de la Copia",
"Count of objects", "Recuento de objetos",
"Create SDTV/HDTV/Radio tags.", "Crear categorías SDTV/HDTV/Radio.",
"Create a bouquet from the playlist.", "Crear un nuevo paquete desde la lista.",
"Create a directory per channel when storing recordings. If both this and the 'directory per day' checkbox is enabled, the date-directory will be the parent of the per-channel directory.", "Crear un directorio por canal al almacenar las grabaciones. Si tanto esta, como la opción \"directorio por día\" estan activadas, el directorio por día será superior al directorio por canal.",
"Create a directory per title when storing recordings. If the day/channel directory checkboxes are also enabled, those directories will be parents of this directory.", "Crear un directorio por título al almacenar las grabaciones. Si las opciones de directorio por día y/o directorio por canal están activadas, dichos directorios serán superiores a este.",
"Create a new directory per day in the recording system path. Folders will only be created when something is recorded. The format of the directory will be ISO standard YYYY-MM-DD.", "Crear un nuevo directorio por día en la ruta del sistema de grabación. El directorio sólo será creado cuando algo se graba. El formato del directorio será el estándar ISO AAAA-MM-DD.",
"Create a provider name tag.", "Crear etiqueta de nombre de proveedor.",
"Create and associate a network tag to created channels.", "Crear y asociar etiqueta de red a los canales creados.",
"Create and associate a provider tag to created channels.", "Crear y asociar etiqueta de proveedor a los canales creados.",
"Create and link these tags to channels when mapping.", "Crear y unir estas categorías a los canales a la hora de mapear.",
"Create bouquet", "Crear paquete",
"Create bouquet tag", "Crear etiqueta de paquete",
"Create directories using these permissions.", "Crear directorios utilizando estos permisos.",
"Create files using these permissions.", "Crear archivos utilizando estos permisos.",
"Create network name tags", "Crear categoría de nombres de red",
"Create network name tags (set by provider).", "Crear etiquetas de nombre de red (asignadas por el proveedor).",
"Create network tags", "Crear etiquetas red",
"Create provider name tags", "Crear etiquetas de nombre proveedor",
"Create provider tags", "Crear etiquetas de proveedor.",
"Create tags", "Crear etiqueta",
"Create tags in recordings using media containers that support metadata (if possible).", "Crear etiquetas en grabaciones utilizando contenedores de medios que soportan metadatos (si es posible).",
"Create type-based tags", "Crear etiquetas de tipo base",
"Created", "Creado",
"Creator", "Creador",
"Cron", "Tarea Programada",
"Cron multi-line", "Cronología multi-línea",
"Culture (without music)", "Cultura (sin musica)",
"Culture magazines", "Revistas de cultura",
"Current RTSP port.", "Puerto RTSP actual.",
"Current affairs", "Actualidad",
"Current number of objects.", "Número de objetos actual.",
"Current object size.", "Tamaño de objetos actual.",
"Current tuner configuration.", "Configuración de sintonización actual.",
"Custom HTTP headers", "Cabeceras HTTP personalizadas",
"DAB Network", "Red DAB",
"DAB multiplex", "Multiplexo DAB",
"DBus - use the session message bus instead of the system one", "DBus - usar el mensaje de sesión de bus en vez de el del sistema",
"DES Constant Code Word", "DES constante en código de palabra",
"DES Key.", "Clave DES",
"DES key", "Clave DES",
"DSCP/TOS for streaming", "DSCP/TOS para streaming",
"DVB CAM Client", "Cliente DVB CAM",
"DVB CSA (descrambling) Tables", "Tablas DVB CSA (desencriptado)",
"DVB EPG ID", "ID EPG DVB",
"DVB EPG Tables", "Tablas DVB EPG",
"DVB Inputs - Networks", "Entradas DVB - Redes",
"DVB SI Tables", "Tablas DVB SI",
"DVB Time Tables", "Tiempo de Tablas DVB",
"DVB multiplex", "Multiplexo DVB",
"DVB scan files", "Escanear archivos DVB",
"DVB scan files path", "Escanear rutas de archivo DVB",
"DVB-C Network", "Red DVB-C",
"DVB-C multiplex", "Multiplexo DVB-C",
"DVB-S Network", "Red DVB-S",
"DVB-S Satellite Configuration", "Configuración de Satélite DVB-S",
"DVB-S multiplex", "Multiplexo DVB-S",
"DVB-T Network", "Red DVB-T",
"DVB-T multiplex", "Multiplexo DVB-T",
"DVR - Auto-recording (Autorecs)", "DVR - Auto-grabación (Autograbaciones)",
"DVR - Profiles", "DVR - Perfiles",
"DVR - Time-based Recording (Timers)", "DVR - Grabaciones basadas en tiempo (Temporizadores)",
"DVR Inotify", "Inotificar DVR",
"DVR behavior", "Comportamiento de DVR",
"DVR configuration", "Configuración DVR",
"DVR configuration profiles", "Perfiles de configuración DVR",
"DVR configurations", "Configuraciones DVR",
"DVR entry not found", "Entrada DVR no encontrada",
"DVR file retention period", "Período de retención de archivos DVR",
"DVR log retention", "Retención del registro de DVR",
"DVR log retention period", "Período de retención de registro DVR",
"DVR override: high", "Anular DVR: alto",
"DVR override: important", "Anular DVR: importante",
"DVR override: low", "Anular DVR: bajo",
"DVR override: normal", "Anular DVR: normal",
"DVR override: unimportant", "Anular DVR: importante",
"DVR profile to use/used for the recording.", "Perfil DVR a utilizar/utilizado para la grabación.",
"DVR schedule does not exist", "La programación DVR no existe",
"DVR schedule does not have a file yet", "La programación DVR no tiene un archivo todavía",
"DVR schedule not found", "Programación DVR no encontrada",
"Dance", "Baile",
"Data errors", "Errores de datos",
"Data path (if applicable).", "Ruta de datos (si es aplicable).",
"Date the EPG data was last updated (not set for OTA grabbers).", "Fecha en la que los datos EPG fueron actualizados por última vez (no aplicable a capturadores OTA).",
"Days of Week", "Días de la Semana",
"Days of the week to which the rule should apply.", "Días de la semana en los que será aplicada la regla.",
"Days of week", "Días de la semana",
"Debug libav log", "Depurar registro libav",
"Debug options", "Opciones de depuración",
"Debug subsystems", "Depurar subsistemas",
"Debug to syslog", "Depurar registro del sistema",
"Debug trace (low-level)", "Depurar trazado (bajo nivel)",
"Debugging", "Depuración",
"Default", "Por defecto",
"Default language(s)", "Idioma(s) por defecto",
"Default language.", "Idioma por defecto.",
"Default login", "Login por defecto",
"Default priority", "Prioridad por defecto",
"Default user interface level.", "Nivel de interfaz de usuario por defecto.",
"Define network group to limit network usage.", "Definir grupo de red para limitad el uso de red.",
"Delivery system", "Sistema de entrega",
"Demux path", "Ruta de demux",
"Descramble services (limit per mux)", "Decodificar servicio (límite por Mux)",
"Descrambler", "Desencriptado",
"Descrambler EMM", "Desencriptado de EMM",
"Descrambler buffer (TS packets)", "Decodificar buffer (Paquetes TS)",
"Descrambling timeout (ms)", "Tiempo agotado en descifrado (ms)",
"Description", "Descripción",
"Detective", "Detectable",
"Device ID", "ID Dispositivo",
"Device model", "Modelo de dispositivo",
"Device path", "Ruta de dispositivo",
"Device path in sysfs", "Ruta del dispositivo en sysfs",
"Differentiated Services Code Point / Type of Service: Set the service class Tvheadend sends with each packet. Depending on the option selected this tells your router the prority in which to give packets sent from Tvheadend, this option does not usually need changing. See https://en.wikipedia.org/wiki/Differentiated_services for more information. ", "Código de servicios diferenciados Punto / Tipo de Servicio: Establecer la clase de servicio Tvheadend  para envíar con cada paquete. Dependiendo de la opción seleccionada esto dice al router la prioridad en el cual dar los paquetes enviados desde Tvheadend, esta opción no suele necesitar ser cambiada. Ver https://es.wikipedia.org/wiki/Servicios_Diferenciados (o https://en.wikipedia.org/wiki/Differentiated_services en inglés) para más información.",
"Digest access authentication is intended as a security trade-off. It is intended to replace unencrypted HTTP basic access authentication. This option should be enabled for standard usage.", "La autenticación de acceso implícita está pensada como una seguridad de intercambio. Se tiene la intención de reemplazar la autenticación de acceso básico no cifrada de HTTP. Esta opción debe ser habilitada para uso estándar.",
"Digital Video Recorder", "Grabador de Vídeo Digital",
"Directory", "Directorio",
"Directory override. Override the subdirectory rules specified by the DVR configuration and put all recordings done by this entry into the specified subdirectory", "Sobreescribir directorio. Sustituir las reglas del subdirectorio especificado por la configuración de DVR y poner todas las grabaciones realizadas por esta entrada en el subdirectorio especificado",
"Directory permissions (octal, e.g. 0775)", "Permisos de directorio (ej.: 0775)",
"Directory used by the entry.", "Directorio utilizado por la entrada.",
"Disable", "Deshabilitar",
"Disable DVB bouquets", "Deshabilitar paquetes DVB",
"Disable RTP/AVP/TCP support", "Desactivar soporte RTP/AVP/TCP",
"Disable SAT>IP client", "Deshabilitar cliente SAT>IP",
"Disable UPnP", "Deshabilitar UPnP",
"Disable UPnP discovery.", "Deshabilitar búsqueda UPnP",
"Disable X_SATIPM3U tag", "Deshabilitar tag X_SATIPM3U",
"Disable all access control checks", "Deshabilitar los Check de Control de Accesos",
"Disable debug on stderr", "Deshabilitar depuración en stderr",
"Disable device/firmware-specific workarounds", "Deshabilitar Dispositivo/Soluciones Especificas del Firmware",
"Disable syslog (all messages)", "Deshabilitar syslog (Todos los Mensajes)",
"Disabled", "Deshabilitado",
"Discard the frontend parameter in RTSP requests, as some clients incorretly use it.", "Descartar el parámetro de interfaz en las peticiones RTSP, ya que algunos clientes no lo utilizan.",
"Discover more muxes using the Network Information Table (if available).", "Descubrir nuevos muxes utilizando la Tabla de Información de Red (NIT) si está disponible.",
"Discussion", "Discusión",
"DiseqC repeats", "Repetir DiseqC",
"Do not send X_SATIPM3U information in the XML description to clients.", "No enviar la información X_SATIPM3U en el XML a la descripción de los clientes.",
"Do not send any stream specific information to the HTSP client like signal strength, input source etc.", "No enviar ninguna información específica de retransmisión al cliente HTSP como fuerza de la señal, fuente de entrada, etc.",
"Do not set", "No configurar",
"Do not use", "No Usar",
"Documentary", "Documental",
"Don't abort streaming when an encrypted stream can't be decrypted by a CA client that normally should be able to decrypt the stream.", "No abortar la retransmisión cuando una retransmisión no pueda ser descodificada por un cliente CA que normalmente debería ser capaz de descodificarlo.",
"Don't backup configuration tree at upgrade", "No guardar arbol de configuración al actualizar",
"Don't include the title in the filename.", "No incluir el título en el nombre de archivo.",
"Don't include title in filename", "No incluir el título en el nombre de archivo",
"Don't keep", "No mantener",
"Don't re-record", "No re-grabar",
"Don't re-record if recording fails.", "No re-grabar si la grabación falla.",
"Don't re-schedule if recording fails.", "No re-programar si la grabación falla.",
"Don't reschedule", "No re-planificar",
"Don't use the provider's channel numbers.", "No utilizar los números de canal del proveedor.",
"Drop \"fe=\" parameter", "No utilizar parámetro \"fe=\"",
"Duplicate handling", "Duplicar manipulación",
"Duplicate recording handling.", "Manejo de grabaciones duplicadas.",
"EIT - skip TSID check", "EIT- pasar chequeo de TSID",
"EIT time offset", "Desfase de tiempo EIT",
"EPG Database", "Base de datos de EPG",
"EPG Grabber", "Capturador de EPG",
"EPG Grabber Channel", "Canal del capturador de EPG",
"EPG Grabber Configuration", "Configuración del capturador de EPG",
"EPG Language (priority order)", "Idioma de EPG (orden de prioridad)",
"EPG data ID.", "ID de datos EPG.",
"EPG name", "Nombre de guía",
"EPG overlap cut", "Recorte de superposición de EPG",
"EPG scan", "Escanear guía",
"EPG scan time-out in seconds (30-7200)", "Tiempo de agote del escaneo de guía en segundos (30-7200)",
"EPG source", "Fuente de guía",
"EPG update window", "Ventana de actualización de guía",
"Economics", "Económicos",
"Education", "Educación",
"Education / Science / Factual topics", "Educación / Ciencia / Temas de hechos",
"Educational", "Educativo",
"Electronic Program Guide", "Guía",
"Elementary Stream Filter", "Filtros de Retransmisión Esenciales",
"Elementary stream filter", "Filtro de retransmisión elemental",
"Enable", "Habilitado",
"Enable (auto)", "Habilitar (automatico)",
"Enable DBus", "Habilitar DBus",
"Enable NTP driver", "Habilitar controlador NTP",
"Enable coredumps for daemon", "Habilitar volcados (coredump) para el demonio",
"Enable debug on stderr", "Habilitar depuración en stderr",
"Enable debug subsystems", "Habilitar subsistemas de depuración",
"Enable debug to file", "Habilitar depuración a fichero",
"Enable debug to syslog", "Habilitar debug a syslog",
"Enable exclusion of user-config defined channel tags. This will prevent the user from accessing channels associated with the tags selected (below).", "Activar la exclusión de las categorías de canal definidas por la configuración del usuario. Esto impedirá al usuario acceder a los canales asociados a las categorías seleccionadas (debajo).",
"Enable if the SAT>IP box requests plts=on parameter in the SETUP RTSP command for DVB-S2 muxes.", "Activar si el cuadro SAT>IP solicita PLTS = el parámetro en el comando de instalación RTSP para muxes DVB-S2.",
"Enable if the SAT>IP box requires pids=21 parameter in the SETUP RTSP command", "Activar si el SAT>IP requiere el parámetro PID = 21 en el comando CONFIGURACIÓN RTSP",
"Enable if the SAT>IP box supports the addpids/delpids commands.", "Activar si el cuadro IP>SAT soporta los comandos addpids / delpids.",
"Enable if the SAT>IP box supports the frontend identifier. This allows the auto-tuner allocation, but it might cause trouble for boxes with different tuner reception connections like satellite inputs.", "Habilitar si el dispositivo SAT>IP soporta identificadores de interfaz. Esto permite la auto-asignación de adaptadores, pero puedo causar problemas con dispositivos con diferentes adaptadores de recepción, como varias entradas diferentes de satélite.",
"Enable or disable RTP/AVP/TCP transfer mode (embedded data in the RTSP session) support.", "Habilitar o deshabilitar el soporte de transferencia RTP/AVP/TCP (datos no incrustados en la sesión RTSP).",
"Enable or disable fast input switching.", "Activar o desactivar la conmutación de entrada rápida.",
"Enable or disable full mux mode.", "Activar o desactivar el modo de mux completo.",
"Enable or disable ignoring of Event Information Table (EIT) data on this mux.", "Activar o desactivar haciendo caso omiso de la tabla de eventos de Información (EIT) datada en este mux.",
"Enable or disable this configuration.", "Activar o desactivar esta configuración.",
"Enable over-the-air program guide (EPG) scanning on this input device.", "Activar escaneo EPG a-través-del-aire (OTA) en este dispositivo de entrada.",
"Enable syslog", "Habilitar registro del sistema",
"Enable system time updates. This will only work if the user running Tvheadend has rights to update the system clock (normally only root).", "Habilitar actualizaciones en tiempo del sistema. Esto sólo funcionará si el usuario que ejecuta Tvheadend tiene derechos para actualizar el reloj del sistema (normalmente solo root).",
"Enable the CAs (conditional accesses) tab in web user interface for the advanced level. By default, this tab is visible only in the expert level.", "Habilitar las CAd (acceso condicional) en la interfaz de usuario web para el nivel avanzado. Por defecto, esta pestaña es visible sólo en el nivel de expertos.",
"Enable this filter.", "Activar este filtro.",
"Enable trace subsystems", "Habilitar subsistemas de traceo",
"Enable web UI debug (non-minified JS)", "Habilitar depuración webUI (JS non-minified)",
"Enable, disable or ignore the mux. When the mux is marked as ignore, all discovered services are removed.", "Habilitar, deshabilitar o ignorar el mux. Cuando el mux está marcado como ignorado, todos los servicios descubiertos se eliminan.",
"Enable/Disable CA client.", "Habilitar/Deshabilitar Cliente CA.",
"Enable/Disable service.", "Activar/Desactivar servicio",
"Enable/Disable the entry.", "Activar/Desactivar la entrada.",
"Enable/Disable timeshift.", "Activar/desactivar timeshift.",
"Enable/disable EPG data for the entry.", "Activar/desactivar datos EPG para la entrada.",
"Enable/disable LNA.", "Activar/desactivar LNA.",
"Enable/disable auto-rec rule.", "Activar/desactivar regla de autograbación.",
"Enable/disable debugging output to syslog.", "Activar/desactivar la salida de depuración al syslog.",
"Enable/disable inclusion of low-level debug traces.", "Activar/desactivar inclusión de trazas de depuración de bajo nivel.",
"Enable/disable interface quick tips.", "Habilitar/deshabilitar consejos rápidos en la interfaz de usuario.",
"Enable/disable libav log output.", "Activar/desactivar salida de log para libav.",
"Enable/disable logging to syslog.", "Activar/desactivar al registro a syslog.",
"Enable/disable offering of Entitlement Management Message updates.", "Activar/desactivar el ofrecimiento de cambios en Titularidad de gestión de mensajes.",
"Enable/disable pilot tone.", "Activar/desactivar el tono piloto.",
"Enable/disable power save mode (if supported by the device).", "Activar/desactivar modo de bajo consumo (si es soportado por el dispositivo).",
"Enable/disable profile.", "Activar/desactivar perfil.",
"Enable/disable the bouquet.", "Activar/desactivar el paquete.",
"Enable/disable the channel.", "Activar/desactivar el canal.",
"Enable/disable the device.", "Activar/desactivar el dispositivo.",
"Enable/disable the entry.", "Activar/desactivar la entrada.",
"Enable/disable the grabber.", "Activar/desactivar el capturador.",
"Enable/disable the tag.", "Activar/desactivar la etiqueta.",
"Enable/disable tuner/adapter.", "Activar/desactivar sintonizador/adaptador.",
"Enabled", "Habilitado",
"Encrypted", "Encriptado",
"Encrypted service", "Servicio encriptado",
"Enter a non-admin user password.", "Introduzca una contraseña de usuario no-administrador.",
"Enter a non-admin user username.", "Introduzca un nombre de usuario de un usuario no-administrador.",
"Enter allowed network prefix(es). You can enter a comma-seperated list of prefixes here.", "Introduzca los prefijos de red permitidos. Puede introducir una lista separada por comas de los prefijos aquí.",
"Enter an administrator password.", "Introduzca una contraseña de administrador.",
"Enter an administrator username. Note: do not use the same username as the superuser backdoor account.", "Introduzca un nombre de usuario administrador. Nota: no utilice el mismo nombre de usuario que la cuenta de superusuario, puerta trasera.",
"Enter comma-separated list of subsystems you want debugging output for (e.g +linuxdvb,+subscriptions,+mpegts).", "Introducir en la lista, separando con comas, los subsistemas en los que desea una salida de depuración para Linux (por ejemplo +linuxdvb,+subscriptions,+mpegts).",
"Enter comma-separated list of subsystems you want to get traces for (e.g +linuxdvb,+subscriptions,+mpegts).", "Introducir en la lista, separando con comas, los subsistemas de los que quiera obtener trazados (por ejemplo +linuxdvb,+subscriptions,+mpegts).",
"Enter external IP if behind Network address translation (NAT). Asterisk (*) means accept all IP addresses.", "Introduzca la IP externa si está detrás de una Traducción de direcciones de red (NAT). Asterisco (*) significa aceptar todas las direcciones IP.",
"Enter external PORT if behind Forwarding redirection.(0 = use the same local port).", "Introduce el PUERTO externo si se encuentra detrás de una redirección. (0 = usar el mismo puerto que el local)",
"Enter the filename (including path) where Tvheadend should write the log.", "Introduzca el nombre de archivo (incluyendo la ruta) donde Tvheadend debería escribir el log.",
"Entertainment programs for 10 to 16", "Programas de entretenimiento de 10 a 16",
"Entertainment programs for 6 to 14", "Programas de entretenimiento de 6 a 14",
"Environment", "Entorno",
"Environment (pipe)", "Entorno (pipe)",
"Episode", "Episidio",
"Episode %d", "Episodio %d",
"Episode number/ID.", "Número de episodio/ID.",
"Equestrian", "Ecuestre",
"Error code", "Código de error",
"Error code of entry.", "Código de error de la entrada.",
"Errors", "Errores",
"Even key", "Cada clave",
"Even key.", "Cada clave",
"Event does not exist", "El evento no existe",
"Every day", "Todos los días",
"Exclude channel tags", "Excluir categorías de canal",
"Expeditions", "Expediciones",
"Experimental film", "Pelicula experimental",
"Expert", "Experto",
"Exported tuners", "Sintonizadores exportados",
"External", "Externo",
"External EPG grabber", "Capturador de EPG externo",
"External IP (NAT)", "IP externa (NAT)",
"External PyEPG Grabber", "Capturador PyEPG externo",
"External RTSP port (NAT)", "Puerto externo RTSP (NAT)",
"External URL", "Enlace Externo",
"External URL of the bouquet.", "URL externa del paquete.",
"External XMLTV EPG Grabber", "Capturador XMLTV EPG externo",
"Extra arguments", "Información adicional",
"Extra start time", "Tiempo de inicio adicional",
"Extra stop time", "Tiempo de finalización adicional",
"Extra warming up time (seconds)", "Tiempo adicional de calentamiento (segundo)",
"FAIL", "FALLO",
"FE supported", "FE soportado",
"FEC high", "FEC alto",
"FEC low", "FEC bajo",
"Factual topics", "Temas de hechos",
"Failed", "Fallido",
"Failed to open image", "Fallo al abrir imagen",
"Failed view", "Fallo de visionado",
"Fantasy", "Fantasia",
"Fashion", "Moda",
"Fast input switch", "Cambiador de entrada rápida",
"Fastscan DVB", "Escaneo rápido DVB",
"File missing", "Fichero no encontrado",
"File not created", "Fichero no creado",
"File permissions (octal, e.g. 0664)", "Permisos de archivo (ej.: 0664)",
"File removed", "Archivo eliminado",
"File size", "Tamaño de archivo",
"Filename", "Nombre de fichero",
"Filename (including path)", "Nombre de archivo (incluyendo la ruta)",
"Filename character set", "Juego de carácteres de nombre de fichero",
"Filename options", "Opciones de nombre de archivo",
"Filename used by the entry.", "Nombre de archivo utilizado por la entrada.",
"Filesystem monitor", "Monitor de sistema de archivos",
"Film", "Pelicula",
"Fine arts", "Bellas artes",
"Finished", "Acabada",
"First word", "Primera palabra",
"Fit to RAM (cut rewind)", "Ajustar a RAM (cortar rebobinado)",
"Fitness and health", "Fitness y salud",
"Folk", "Folclore",
"Folkloric", "Folclórico",
"Football", "Futbol",
"For example: 12610500. This frequency is 12610.5Mhz or 12.6105Ghz.", "Por ejemplo: 12610500. Esta frecuencia es 12610.5MhzMhz o 12.6105Ghz.",
"For example: 312000000. This frequency is 312Mhz.", "Por ejemplo: 312000000. Esta frecuencia es 312Mhz.",
"For example: 658000000. This frequency is 658Mhz.", "Por ejemplo: 658000000. Esta frecuencia es 658Mhz.",
"Force (auto)", "Forzar (auto)",
"Force CA ID (e.g. 0x2600)", "Forzar CA ID (ej. 0x2600)",
"Force all network connections to this tuner to be made over the specified IP address, similar to the setting for the SAT>IP device itself. Setting this overrides the device-specific setting.", "Forzar todas las conexiones de red para este sintonizador que actúa sobre la dirección IP especificada, similar a la configuración de SAT>IP propia. Establecer este anula la configuración específica del dispositivo.",
"Force an initial EPG grab at start-up.", "Forzar una obtención inicial de EPG al iniciar.",
"Force initial EPG grab at start-up", "Forzar obtención inicial de EPG al iniciar.",
"Force old status", "Forzar estado anterior",
"Force pilot for DVB-S2", "Forzar piloto para DVB-S2",
"Force priority", "Forzar prioridad",
"Force profile to use this priority.", "Forzar al perfil a utilizar esta prioridad.",
"Force service type to 1", "Forzar tipo de servicio a 1",
"Force signal level", "Forzar nivel de señal",
"Force signal level for all streaming (1-240, 0=do not use).", "Forzar nivel de señal para todo streaming (1-240, 0=no utilizar).",
"Force teardown delay", "Forzar retraso de desmontaje",
"Force the delay between RTSP TEARDOWN and RTSP SETUP command (value from 'Next tune delay in ms' is used). Some devices are not able to handle quick continuous tuning.", "Forzar el retardo entre comandos RTSP TEARDOWN y RTSP SETUP (se utiliza el valor de la \"demora en sintonizar en ms\"). Algunos dispositivos no son capaces de manejar sintonización continua rápida.",
"Force the grace period for which SAT>IP client waits for the data from server. After this grace period, the tuner is handled as dead. The default value is 5 seconds (for DVB-S/S2: 10 seconds).", "Forzar el período de gracia para el cual el cliente SAT>IP espera los datos del servidor. Después de este período de gracia, el sintonizador se maneja como muerto. El valor predeterminado es 5 segundos (para DVB-S/S2: 10 segundos).",
"Force the local UDP Port number here. The number should be even (RTP port). The next odd number (+1) will be used as the RTCP port.", "Forzar el número de puerto UDP local aquí. El número debe ser uniforme (puerto RTP). El siguiente número impar (+1) se utilizará como el puerto RTCP.",
"Force usage of entered CA ID on this service.", "Forzar el uso del CA ID introducido en este servicio.",
"Forced OK", "OK Forzado",
"Foreign countries", "Países extranjeros",
"Forever", "Para siempre",
"Fork and run as daemon", "Bifurcar y ejecutar como demonio",
"Format string", "Cadena de formato",
"Found muxes", "Muxes encontrados",
"Found services", "Encontrar servicios",
"Free subscription weight", "Peso de la suscripción libre.",
"Free-form field, enter whatever you like here.", "Campo libre, introduzca cualquier cosa que desee.",
"Free-form text field, enter whatever you like here.", "Campo libre, introduzca cualquier cosa que desee.",
"Free-form text field, enter whatever you like.", "Campo libre, introduzca cualquier cosa que desee.",
"Free-form text field. You can enter whatever you like here.", "Campo libre, puede introducir cualquier cosa que desee.",
"Free-format text field. Enter whatever you like here.", "Campo de texto de formato libre. Introduzca lo que quiera aquí.",
"Frequency", "Frecuencia",
"Frequency (Hz)", "Frecuencia (Hz)",
"Frequency (MHz)", "Frecuencia (Mhz)",
"Frequency (in MHz).", "Frecuencia (en Mhz).",
"Frequency (kHz)", "Frecuencia (kHz)",
"Fri", "Vie",
"Friendly name", "Nombre amigable",
"Frontend number", "Numero de interfaz",
"Frontend path", "Ruta de interfaz",
"Full DiseqC", "DisqC Completo",
"Full mux RX mode supported", "Modo Mux RX lleno soportado",
"Full path to an icon used to depict the tag. This can be a TV network logotype, etc.", "Ruta completa al icono utilizado para representar la etiqueta. Puede ser un logotipo de una red de TV, etc.",
"Full pathname specification", "Especificación de ruta de archivo completa",
"Full-text", "Texto completo",
"Further education", "Futura educación",
"GOTOX position", "Posición de GOTOX",
"Game show", "Show de juegos",
"Gardening", "Jardinería",
"General configuration", "Configuración general",
"Generic", "Genérico",
"Generic options", "Opciones genéricas",
"Global Headers", "Cabeceras globales",
"Global timer", "Temporizador global",
"Grabber priority. This option lets you pick which EPG grabber's data get used first. Priority is given to the grabber with the highest value set here. See Help for more info.", "Prioridad de Grabber. Esta opción le permite seleccionar los datos del captador de EPG que se utilizarán primero. Se da prioridad al grabber con el valor más alto establecido aquí. Consulte la Ayuda para obtener más información.",
"Grace period", "Periodo de gracia",
"Gray", "Gris",
"Guard interval", "Intervalo de guarda",
"HD: high definition", "HD: alta definición",
"HDHomeRun ATSC-C frontend", "Índice HDHomeRun ATSC-C",
"HDHomeRun ATSC-T frontend", "índice HDHomeRun ATSC-T",
"HDHomeRun DVB frontend", "Índice HDHomeRun DVB",
"HDHomeRun DVB-C frontend", "Índice HDHomeRun DVB-C",
"HDHomeRun DVB-T frontend", "Índice HDHomeRun DVB-T",
"HTSP Answer", "Respuesta HTSP",
"HTSP Default Stream Settings", "Ajustes por defecto de Retransmisión HTSP",
"HTSP Request", "Petición HTSP",
"HTSP Server", "Servidor HTSP",
"HTSP Stream Profile", "Perfil de Retransmisión HTSP",
"HTSP Subscription", "Suscripción HTSP",
"HTTP CORS (cross-origin resource sharing) origin. This option is usually set when Tvheadend is behind a proxy. Enter a domain (or IP) to allow cross-domain requests.", "HTTP CORPS (cruz-origen compartido) originario. Esta opción se establece generalmente cuando Tvheadend está detrás de un proxy. Introduzca un dominio (o IP) para permitir peticiones entre dominios.",
"HTTP CORS origin", "Origen de CORS HTTP",
"HTTP Client", "Cliente HTTP",
"HTTP Server", "Servidor HTTP",
"Handicraft", "Artesanía",
"Hearing impaired", "Personas con discapacidad auditiva",
"Hierarchy", "Jerarquía",
"High", "Alto",
"High bitrate mode (CI+ CAMs only)", "Modo de alta tasa de datos (CI+ solo cámaras)",
"High frequency offset", "Offset en alta frecuencia",
"Higher", "Superior",
"Higher LNB voltage", "Voltaje Superior a LNB",
"Highest channel number the user can access.", "Número máximo de canal al que el usuario puede acceder.",
"Highest count of objects.", "El mayor recuento de objetos.",
"Historical movie", "Película histórica",
"Horror", "Terror",
"Hostname (or IP) of the server.", "Dirección de host (o IP) del servidor.",
"Hostname/IP", "Nombre del Host/IP",
"How frequently it will re-try fetching an image that has failed to be fetched.", "¿Con qué frecuencia se volverá a tratar de buscar una imagen que no ha podido ser encontrada?",
"How frequently the upstream provider is checked for changes.", "¿Con qué frecuencia se comprueba el proveedor para encontrar cambios?",
"ID of the network.", "ID de la red.",
"IDLE", "EN ESPERA",
"IGNORE", "IGNORAR",
"IP address", "Dirección IP",
"IP blocking", "Bloqueo de IP",
"IPTV Automatic Network", "Red automática IPTV",
"IPTV Multiplex", "Multiplexo IPTV",
"IPTV Network", "Red IPTV",
"IPTV Subcription", "Suscripción IPTV",
"IPTV input", "Entrada de IPTV",
"IPTV signal level", "Nivel de señal IPTV",
"ISDB-C Network", "Red ISDB-C",
"ISDB-C multiplex", "Multiplexo ISDB-C",
"ISDB-S Network", "Red ISDB-S",
"ISDB-S multiplex", "Multiplexo ISDB-S",
"ISDB-T Network", "Red ISDB-T",
"ISDB-T multiplex", "Multiplexo ISDB-T",
"ISI (Stream ID)", "ISI (Identificador de Retransmisión)",
"Icon", "Icono",
"Icon (full URL)", "Icono (URL completo)",
"Icon URL", "URL Icono",
"Icon base URL", "Icono basado en URL",
"Icon base URL.", "Icono basado en URL",
"Icon has title", "El icono tiene título",
"Icons from", "Iconos de",
"Idle", "En Espera",
"Idle scan", "Escaneo en Espera",
"Idle scan muxes", "Buscar muxes en espera",
"If an error occurs clone the scheduled entry and try to record again (if possible).", "Si se produce un error al clonar la entrada programada se tratará de grabar de nuevo (si es posible).",
"If both a picon and a channel-specific (e.g. channelname.jpg) icon are defined, prefer the picon.", "Si se definen ambas, un picon y un canal específico (por ejemplo channelname.jpg) el icono preferirá el picón.",
"If more than x errors occur during a recording schedule a re-record (if possible).", "Si se producen más de x errores durante la grabación de un programa de re-grabación (si es posible).",
"If no specific priority was requested. This gives certain users a higher priority by assigning a streaming profile with a higher priority.", "Si no se ha solicitado ninguna prioridad específica. Esto le da a ciertos usuarios una prioridad más alta mediante la asignación de un perfil de retransmisión con una prioridad más alta.",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Si no existe ninguna cuenta de usuario a continuación, crear una con\nningún nombre de usuario ni contraseña. Utilizar con cuidado\nésto permitirá un acceso administrativo a todo el mundo\na toda su instalación de Tvheadend hasta que crees o edites\nel control de acceso desde la interfaz web de Tvheadend.",
"If set, presentation of the tag icon will not superimpose the tag name on top of the icon.", "Si se establece, la presentación la etiqueta de icono no debe superponer el nombre de la etiqueta en la parte superior del icono.",
"If set, the first bytes from the MPEG-TS stream are discarded. It may be required for some drivers or hardware which do not flush the MPEG-TS buffers completely after a frequency/parameter change.", "Si se activa, los primeros bytes de retransmisión de MPEG-TS se descartan. Es posible que se requiera para algunos drivers o hardware que no vuelcan los búferes de MPEG-TS por completo después de un cambio de frecuencia/parámetro.",
"If something fails, try to switch to a different service on another network. Do not try to iterate through all inputs/tuners which are capable to receive the service.", "Si algo falla, intente cambiar a otro servicio en otra red. No trate de iteractuar a través de todas las entradas/sintonizadores que son capaces de recibir el servicio.",
"If the DiseqC switch is located before the rotor (i.e. tuner - switch - rotor), enable this.", "Si el interruptor DiseqC se encuentra antes que el rotor (es decir,  sintonizador - conmutador - rotor), activar esto.",
"If the subscription weight for the input is below the specified threshold, the tuner is handled as free (according the priority settings). Otherwise, the next tuner (without any subscriptions) is used. Set this value to 10, if you are willing to override scan and epggrab subscriptions.", "Si el peso de suscripción para la entrada está por debajo del umbral especificado, el sintonizador se maneja como libre (según la configuración de prioridad). De lo contrario, se utiliza el siguiente sintonizador (sin ninguna suscripción). Establecer este valor en 10, si usted está dispuesto a anular las suscripciones de exploración y epggrab.",
"If you'd like to support the project, please consider a donation.", "Si quiere apoyar el proyecto, por favor considere realizar una donación.",
"Ignore", "Ignorar",
"Ignore EPG (EIT)", "Ignorar EPG (EIT)",
"Ignore descriptor 5", "Ignorar descriptor 5",
"Ignore encryption flag, include encrypted services anyway.", "Ignorar el indicador de cifrado, incluir servicios codificados de todos modos.",
"Ignore invalid SSL certificate", "Ignorar certificado SSL inválido",
"Ignore invalid/unverifiable (expired, self-certified, etc.) certificates", "Ignorar certificados no válidos / no verificables (expirados, autocertificados, etc...)",
"Ignore provider's channel numbers", "Ignorar números de canales del proveedor",
"Image Cache", "Caché de imagen",
"Immediately abort", "Abortar inmediatamente",
"Important", "Importante",
"Include channel name in filename", "Incluir nombre de canal en el nombre del fichero",
"Include date in filename", "Incluir fecha en el nombre del archivo",
"Include episode in filename", "Incluir episodio en el nombre del archivo",
"Include subtitle in filename", "Incluir subtítulo en el nombre del archivo",
"Include teletext", "Incluir teletexto",
"Include teletext in the timeshift buffer. Enabling this may cause issues with some services where the teletext DTS is invalid.", "Incluir el teletexto en el buffer de tiempo. Habilitar esto puede causar problemas con algunos servicios donde el teletexto DTS no es válido.",
"Include the date for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "Incluir la fecha de la grabación en el título del evento. Esto se aplica tanto en el título almacenado como en el archivo y en el propio nombre de archivo.",
"Include the episode subtitle in the title (if available).", "Incluir los subtítulos del episodio en el título (si está disponible).",
"Include the name of the channel in the event title. This applies to both the title stored in the file and to the filename itself.", "Incluir el nombre del canal en el título del evento. Esto se aplica tanto en el título almacenado como en el archivo y en el propio nombre de archivo.",
"Include the season and episode in the title (if available).", "Incluir la temporada y el episodio en el título (si está disponible).",
"Include the time for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "Incluir el momento de la grabación en el título del evento. Esto se aplica tanto al título almacenado como en el archivo y en el propio nombre de archivo.",
"Include time in filename", "Incluir hora en el nombre del archivo",
"Index", "Índice",
"Information area", "Área de Información",
"Informational", "Informativo",
"Initial scan", "Escaneo inicial",
"Input base", "Entrada base",
"Input buffer (bytes)", "Entrada de buffer (bytes)",
"Input instance", "Instancia de entrada",
"Input path", "Ruta de entrada",
"Interface", "Interfaz",
"Internal", "Interno",
"Internal EPG grabber", "Capturador EPG interno",
"Internal PyEPG Grabber", "Capturador PyEPG interno",
"Internal XMLTV EPG Grabber", "Capturador XMLTV EPG interno",
"Internal error", "Error interno",
"Internal grabber", "Capturador interno",
"Interview", "Entrevista",
"Invalid", "Inválido",
"Invalid EPG object request", "Solicitud de EPG no válida",
"Invalid adapter number '%s'\n", "Número de adaptador inválido '%s'\n",
"Invalid arguments", "Argumentos inválidos",
"Invalid file", "Archivo incorrecto",
"Invalid parameters", "Parámetros inválidos",
"Invalid service", "Servicio inválido",
"Invalid target", "Selección inválida",
"Items", "Artículos",
"Jazz", "Música Jazz",
"Keep", "Conservar",
"Keep timeshift buffers in RAM only. With this option enabled, the amount of rewind time is limited by how much RAM TVHeadend is allowed.", "Mantener los búferes Timeshift en RAM solamente. Con esta opción habilitada, la cantidad de tiempo de rebobinado está limitada por la cantidad de RAM que TVHeadend tenga permitida.",
"Keep x amount of storage space free.", "Mantener x cantidad de espacio de almacenamiento libre.",
"Keepalive interval", "Intervalo de espera",
"Keepalive interval in seconds", "Intervalo de espera en segundos",
"Key and value pairs to remove from the query string in the URL.", "Clave y valor para eliminar la cadena de consulta en la URL.",
"Kill signal (pipe)", "Matar señal (tubería)",
"Kill timeout (pipe/secs)", "Matar tiempo agotado (tubería/segs)",
"LNA (low noise amplifier)", "LNA (amplificador de bajo sonido)",
"LNB type", "Tipo de LNB",
"Language", "Idiomas",
"Language 1", "Idioma 1",
"Language 2", "Idioma 2",
"Language 3", "Idioma 3",
"Language settings", "Ajustes de idioma",
"Languages", "Idiomas",
"Largest size the object has reached.", "Tamaño más grande que el objeto ha alcanzado.",
"Last played position", "Última posición reproducida",
"Last played position when the recording isn't fully watched yet.", "Última posición reproducida mientras la grabación no había sido completamente vista aún.",
"Last seen", "Última vez visto",
"Last updated from", "Última actualización",
"Layer A: Constellation", "Capa A: Constelación",
"Layer A: FEC", "Capa A: FEC",
"Layer A: Segment count", "Capa A: Recuento de segmentos",
"Layer A: Time interleaving", "Capa A: Entrelazado de tiempo",
"Layer B: Constellation", "Capa B: Constelación",
"Layer B: FEC", "Capa B: FEC",
"Layer B: Segment count", "Capa B: Recuento de segmentos",
"Layer B: Time interleaving", "Capa B: Entrelazado de tiempo",
"Layer C: Constellation", "Capa B: Constelación",
"Layer C: FEC", "Capa C: FEC",
"Layer C: Segment count", "Capa C: Recuento de segmentos",
"Layer C: Time interleaving", "Capa C: Entrelazado de tiempo",
"Leisure hobbies", "Aficiones de ocio",
"Limit connections", "Limitar conexiones",
"Limited/limit scanning to this network ID only.", "Exploración limitada/limitar sólo a este identificador de red.",
"Linked input", "Entrada adjunta",
"Linux DVB frontend", "Interfaz Linux DVB",
"LinuxDVB Input", "Entrada LinuxDVB",
"LinuxDVB adapter", "Adaptador LinuxDVB",
"LinuxDVB network", "Red LinuxDVB",
"List of allowed IPv4 or IPv6 hosts or networks (comma-separated).", "Lista de hosts y redes autorizadas IPv4 o IPv6 (separados por comas).",
"List subsystems", "Listar subsistemas",
"Listen / Connect port", "Escuchar / Conectar en puerto",
"Listen on IPv6", "Escuchar en IPv6",
"Literature", "Literatura",
"Local (server) time", "Hora local (servidor)",
"Local bind IP address", "Dirección IP de enlace local",
"Local channel minor", "Canal menor local",
"Local channel number", "Número local de canal",
"Local discovery IP address", "Descubrir IP local",
"Local: Record if different description", "Local: Grabar en caso de una descripción diferente",
"Local: Record if different episode number", "Local: Grabar en caso de un diferente número de episodio",
"Local: Record if different subtitle", "Local: Grabar en caso de subtítulos diferentes",
"Local: Record if different title", "Local: Grabar en caso de título diferente",
"Local: Record once per day", "Local: Grabar una vez al día",
"Local: Record once per month", "Local: Grabar una vez al mes",
"Local: Record once per week", "Local: Grabar una vez por semana",
"Location", "Localización",
"Location details of the SAT>IP Server.", "Detalles de la ubicación del servidor SAT>IP.",
"Lock preferred CA PID", "Bloquear PID CA preferido",
"Locking", "Bloqueando",
"Log", "Registro",
"Login password.", "Contraseña de acceso.",
"Login username.", "Usuario de acceso.",
"Login/Logout", "Entrar/Salir",
"Loglevel debug: disabled", "Registro de depuración: deshabilitado",
"Loglevel debug: enabled", "Registro de depuración: habilitado",
"Logout", "Salir",
"Low", "Bajo",
"Low frequency offset", "Offset en baja frecuencia",
"Lowest channel number the user can access.", "Número más pequeño de canal, al que el usuario puede acceder.",
"Lowest starting channel number.", "El menor número de canal de comienzo.",
"MP4 audio", "Audio MP4",
"MPEG-2 audio", "Audio MPEG-2",
"MPEG-TS File", "Archivo MPEG-TS",
"MPEG-TS Input Debug", "Depuración de MPEG-TS de entrada",
"MPEG-TS Multiplex", "Multiplexo MPEG-TS",
"MPEG-TS Parser", "Parser MPEG-TS",
"MPEG-TS Pass-thru", "Pasar a través de MPEG-TS",
"MPEG-TS Pass-thru/built-in", "Pasar a través de/construido en MPEG-TS",
"MPEG-TS Service", "Servicio MPEG-TS",
"MPEG-TS input", "Entrada MPEG-TS",
"MPEG-TS multiplex PHY", "Multiplexo MPEG-TS PHY",
"MPEG-TS raw service", "Servicio plano MPEG-TS",
"Magazines", "Revistas",
"Main", "Principal",
"Maintain free storage space in MiB", "Mantener el espacio de almacenamiento libre en MiB",
"Maintain used storage space in MiB (0=disabled)", "Mantener el espacio de almacenamiento utilizado en MiB (0 = desactivado)",
"Maintained space", "Espacio mantenido",
"Make subdirectories per channel", "Crear subdirectorios por Canal",
"Make subdirectories per day", "Crear subdirectorios por día",
"Make subdirectories per title", "Crear subdirectorios por título",
"Manage all", "Controlar todos",
"Manufacturer", "Manufacturador",
"Manufacturer URL", "URL de Manufacturador",
"Manufacturer's model number.", "Número de modelo del fabricante.",
"Manufacturer's product description.", "Descripción de producto del fabricante.",
"Manufacturer's product information page for the device.", "La página de información de producto del fabricante para el dispositivo.",
"Manufacturer's product name.", "Nombre de producto del fabricante.",
"Map all services", "Mapear todos los servicios",
"Map encrypted services", "Mapear canales encriptados",
"Map radio channels", "Mapear canales de radio",
"Map unnamed channels", "Mapear canales sin nombre",
"Map zero-numbered channels", "Mapear canales sin numerar",
"Mapped", "Mapeado",
"Martial sports", "Artes marciales",
"Master tuner", "Sintonizador maestro",
"Matroska muxer", "Mux Matroska",
"Matroska specific", "Especificidad matroska",
"Maximal channel number", "Número de canal máximo",
"Maximum # input streams", "Máximas # entradas de retransmisiones",
"Maximum PIDs", "Máximos PIDs",
"Maximum RAM size (MB)", "Máximo tamaño de RAM (MB)",
"Maximum allowed difference between event start time when the EPG event is changed.", "Diferencia máxima permitida entre la hora de inicio cuando se cambia el evento EPG.",
"Maximum bandwidth (Kbps)", "Máximo ancho de banda (Kbps)",
"Maximum count (0=default)", "Recuento máximo (0=por defecto)",
"Maximum duration", "Máxima duración",
"Maximum input bandwidth.", "Máximo ancho de banda de entrada.",
"Maximum length in characters for the command setting PIDs to the SAT>IP box.", "Longitud máxima de caracteres para el ajuste de comandos en los PID a la caja SAT>IP.",
"Maximum length of PIDs", "Máxima longitud de PIDs",
"Maximum period (mins)", "Periodo máximo (mins)",
"Maximum schedules limit (0=default)", "Límite máximo en espera (0=por defecto)",
"Maximum size (MB)", "Tamaño máximo (MB)",
"Maximum time to wait (in seconds) for a stream before a timeout.", "Tiempo máximo a esperar (en segundos) para una retransmisión antes del tiempo agotado.",
"Maximum timeout (seconds)", "Máxima tiempo agotado (segundos)",
"Medicine", "Medicina",
"Memory Information", "Información de la memoria",
"Merge same name", "Combinar mismo nombre",
"Merge services with the same name into one channel.", "Combinar servicios con el mismo nombre a un único canal.",
"Method not found", "Método no encontrado",
"Minimal channel number", "Número de canal mínimo",
"Minimum duration", "Duración mínima",
"Minimum rotor time (seconds)", "Velocidad de rotor mínima (segundos)",
"Missing In PAT/SDT", "Ausente en PAT/SDT",
"Mode", "Modo",
"Model description", "Descripción de modelo",
"Model name", "Nombre de modelo",
"Model number", "Número de modelo",
"Modulation", "Modulación",
"Module", "Modulo",
"Module ID", "ID de Módulo",
"Module ID used to grab EPG data.", "Módulo de Identificación ID utilizado para extraer los datos EPG.",
"Mon", "Lun",
"Monitonic timer", "Temporizador monótono",
"More verbose libav log", "Registro libav más detallado",
"Motor rate (in milliseconds/deg).", "tasa de Motor (en milisegundos/deg).",
"Motor rate (milliseconds/deg)", "Velocidad de rotor (milisegundos/grados)",
"Motor sport", "Deporte de moto",
"Motoring", "Automovilismo",
"Movie", "Película",
"Movie / drama", "Película / Drama",
"Multiple lines of the cron time specification. The default cron triggers the internal grabbers daily at 12:04 and 00:04. See Help on how to define your own.", "Varias líneas de la especificación de tiempo cron. El cron por defecto ejecuta los capturadores internos diariamente a las 12:04 y 00:04. Consulte la Ayuda sobre cómo definir el suyo propio.",
"Multiple lines of the cron time specification. The default cron triggers the over-the-air grabber daily at 02:04 and 14:04. See Help on how to define your own.", "Líneas múltiples de la especificación de tiempo cron. El cron predeterminado activa la guía a través del aire a las 02:04 y 14:04. Consulte Ayuda sobre cómo definir la configuración.",
"Music", "Música",
"Music / Ballet / Dance", "Musica / Ballet / Danza",
"Mux Scheduler", "Horario de Mux",
"Mux handling", "Manipulación de mux",
"Mux name", "Nombre de mux",
"Mux not enabled", "Mux no habilitado",
"NONE", "Ninguno",
"Name", "Nombre",
"Name (or date) of program the entry is a rerun of.", "Nombre (o fecha) del programa en el que la entrada es repetida.",
"Name of channel the entry recorded from.", "Nombre del canal de la entrada a partir de la grabación.",
"Name of object.", "Nombre del objeto.",
"Name of the bouquet.", "Nombre del paquete.",
"Name of the client.", "Nombre del cliente.",
"Name of the entry.", "Nombre de la entrada.",
"Name of the module used to grab EPG data.", "Nombre del módulo usado para extraer los datos de la guía.",
"Name of the module, grabber or channel that should be used to update this channels EPG info.", "Nombre del módulo, capturador o canal que debería ser utilizado para actualizar la información EPG de este canal.",
"Name of the network.", "Nombre de la red.",
"Name of the profile.", "Nombre del perfil.",
"Name of the tag.", "Nombre de la etiqueta.",
"Name of the tuner.", "Nombre del sintonizador.",
"Name of the tuner/adapter.", "Nombre del sintonizador/adaptador.",
"Names", "Nombres",
"Natural sciences", "Ciencias naturales",
"Nature", "Naturaleza",
"Network", "Red",
"Network ", "Red",
"Network A", "Red A",
"Network B", "Red B",
"Network C", "Red C",
"Network D", "Red D",
"Network ID (limit scanning)", "ID de red (límite escaneando)",
"Network UUID", "UUID de Red",
"Network access", "Acceso a la red",
"Network discovery", "Descubrimiento de red",
"Network for port A.", "Red para el puerto A.",
"Network for port AA.", "Red para el puerto AA.",
"Network for port AB.", "Red para el puerto AB.",
"Network for port B.", "Red para el puerto B.",
"Network for port BA.", "Red para el puerto BA.",
"Network for port BB.", "Red para el puerto BB.",
"Network for port C.", "Red para el puerto C.",
"Network for port D.", "Red para el puerto D.",
"Network group", "Grupo de red",
"Network limit per position", "Límite de red por posición",
"Network name", "Nombre de red",
"Network prefix", "Prefijo de Red",
"Network settings", "Ajustes de red",
"Network type", "Tipo de red",
"Networks", "Redes",
"New / premiere / unknown", "Nuevo / estreno / desconocido",
"New login", "Nuevo login",
"New media", "Nueva multimedia",
"New muxes + changed muxes", "Nuevos muxes + muxes cambiados",
"New muxes only", "Solo nuevos muxes",
"News", "Noticias",
"News / Current Affairs", "Noticias / Actualidad",
"News magazine", "Revista de noticias",
"Next tune delay in ms (0-2000)", "Retraso al siguiente sintonizador en ms (0-2000)",
"No PIN", "Sin PIN",
"No access", "Sin acceso",
"No adapters specified!\n", "No se han especificado adaptadores!\n",
"No assigned adapters", "Sin adaptadores asignados",
"No days", "No hay días",
"No descrambler", "Sin descodificador",
"No free adapter", "No hay ningún adaptador libre",
"No input detected", "No hay ninguna entrada detectada",
"No scheme", "Sin esquema",
"No service assigned to channel", "No existe un servicio asignado al canal",
"No service enabled", "No hay un servicio habilitado",
"No source available", "Fuente no disponible",
"Node subsystem", "Nodo de subsistema",
"None", "Ninguno",
"Not all SAT>IP servers use the same signal scaling. Change this setting if the signal level displayed within Tvheadend looks too low.", "No todos los servidores SAT>IP utilizan la misma escala de señal. Cambie este ajuste si el nivel de señal que se muestra dentro de Tvheadend parece demasiado bajo.",
"Not enough disk space", "No hay suficiente espacio en el disco",
"Not enough memory", "No hay suficiente memoria",
"Not linked", "Sin unir",
"Not set", "No configurado",
"Not set (none or channel configuration)", "No seleccionado (nada o configuración de canal)",
"Not set (use channel or DVR configuration)", "No seleccionado (usar canal o configuración DVR)",
"Number", "Número",
"Number of days to keep the file.", "Número de días a conservar en el archivo.",
"Number of days to keep the recorded file.", "Número de días a conservar el archivo grabado.",
"Number of days to keep the recorded files.", "Número de días a conservar los archivos grabados.",
"Number of days to retain entry information.", "Número de días para retener la información de la entrada.",
"Number of days to retain information about recording.", "Número de días para retener la información acerca de la grabación.",
"Number of days to retain information about recordings. Once this period is exceeded, duplicate detection will not be possible anymore.", "Número de días a retener información acerca de las grabaciones. Una vez este periodo se exceda, duplicar la detección no será posible.",
"Number of errors that occurred during recording (Transport errors).", "Número de errores producidos durante la grabación (Errores de transporte).",
"Number of errors that occurred during recording.", "Número de errores producidos durante la grabación.",
"Number of muxes found.", "Número de muxes encontrados",
"Number of repeats for the DiseqC commands (default is zero - no DiseqC repeats). Note: this represents the number of repeats, not the number of requests - so 0 means 'send once: don't repeat', 1 means 'send twice: send once, then send one repeat', etc.", "Número de repeticiones para los comandos DiSEqC (por defecto es cero - Sin repeticiones DiSEqC). Nota: esto representa el número de repeticiones, no el número de solicitudes - por lo que 0 significa \"enviar una vez; no repita\", 1 significa \"enviar dos veces; una vez a enviar y a continuación enviar una repetición\", etc.",
"Number of repeats for the tune requests (default is zero - no repeats). Note: this represents the number of repeats, not the number of requests - so 0 means 'send once: don't repeat', 1 means 'send twice: send once, then send one repeat', etc.", "Número de repeticiones para las solicitudes Tune (por defecto es cero - no se repite). Nota: esto representa el número de repeticiones, no el número de solicitudes - por lo que 0 significa \"enviar una vez: no repita \", 1 significa \"enviar dos veces: una vez a enviar y a continuación, enviar una repetición\", etc.",
"Number of seconds to wait before timing out.", "Número de segundos a esperar para llegar al tiempo agotado.",
"Number of services.", "Número de servicios.",
"Number of times this recording was played.", "Número de veces en que esta grabación fue reproducida.",
"Number of tsfile tuners", "Número de sintonizadores tsarchivos",
"OK (partial)", "OK (parcial)",
"OSCam (rev >= 9095)", "OSCam (versión >= 9095)",
"OSCam TCP (rev >= 9574)", "OSCam TCP (versión >= 9574)",
"OSCam net protocol (rev >= 10389)", "OSCam net protocol (versión >= 10389)",
"OSCam new pc-nodmx (rev >= 10389)", "OSCam nuevo pc-nodmx (versión >= 10389)",
"OSCam pc-nodmx (rev >= 9756)", "OSCam pc-nodmx (versión >= 9756)",
"Odd key", "Clave odd",
"Odd key.", "Clave odd",
"Off", "Apagado",
"Offset the mapped channel numbers by x (value here + channel number).", "Compensar los números de los canales asignados por x (valor aquí + número de canal).",
"Older OSCam", "OSCam antiguo",
"On", "Encendido",
"On file removal", "En eliminación del archivo",
"On-demand (no first rewind)", "Bajo demanda (sin rebobinado rápido)",
"Once per auto channel", "Una por cada canal automático",
"One tune request (setup) is sent before the DiseqC sequence (voltage, tone settings). Some linux drivers require this procedure.", "Una petición de sintonía (configuración) será enviada antes de la secuencia DiseqC (voltaje, ajustes de tono). Algunos controladores de Linux requieren este procedimiento.",
"Only Bulsatcom 39E", "Solo Bulsatcom 39E",
"Only EIT", "Solo EIT",
"Only OpenTV Sky Ausat", "Solo OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Solo OpenTV Sky Italia",
"Only OpenTV Sky UK", "Solo OpenTV Sky UK",
"Only PSIP (ATSC)", "Solo PSIP (ATSC)",
"Only UK Freesat", "Solo UK Freesat",
"Only UK Freeview", "Solo UK Freeview",
"Only Viasat Baltic", "Solo Viasat Baltic",
"Only activate timeshift when the client makes the first rewind, fast-forward or pause request. Note, because there is no buffer on the first request rewinding is not possible at that point.", "Sólo activar Timeshift cuando el cliente realiza el primer rebobinado, avance rápido, pausa o petición. Nota, porque no hay buffer en la primera solicitud de rebobinado no es posible en ese punto.",
"Only admin can watch the realtime log.", "Solo el administrador puede ver el registro en tiempo real.",
"Only allow users with this tag (or those with no tags at all) set in access configuration to use the tag.", "Sólo permitir a los usuarios con ésta etiqueta (o a aquellos sin ninguna etiqueta) fijada en la configuración de acceso usar la etiqueta.",
"Only digits", "Solo dígitos ",
"Only preferred CA PID", "Solo  CA PID preferido",
"Only update the system clock (doesn't affect NTP driver) if the delta between the system clock and DVB time is greater than this. This can help stop excessive oscillations on the system clock.", "Sólo actualizar el reloj del sistema (no afecta al controlador NTP) si el delta entre la hora del reloj del sistema y DVB es mayor que este. Esto puede ayudar a detener las oscilaciones excesivas en el reloj del sistema.",
"Only use specified DVB adapters (comma-separated, -1 = none)", "Solo utilizar los adaptadores DVB especificados (separada por comas, -1 = ninguno)",
"Only use this EPG data once when automatically determining what EPG data to set for a channel.", "Sólo utilice estos datos de EPG cuando se determinen automáticamente qué datos de EPG se deben definir a un canal.",
"OpenTV EPG", "EPG OpenTV",
"OpenTV channel number", "Número de canal OpenTV",
"Opera", "Ópera",
"Options to use/used when mapping - see Help for details.", "Opción a usar/usada durante el mapeo - ver Ayuda para más detalles.",
"Options used when updating channels.", "Opción usada durante la actualización de canales.",
"Orbital position", "Posición orbital",
"Orbital positions", "Posiciones orbitales",
"Orbital positions.", "Posiciones orbitales.",
"Original network ID", "ID de red original",
"Oscam mode.", "Modo de Oscam.",
"Other Stream Filter", "Otro Filtro de Retransmisión",
"Output directory for tsdebug", "Directorio de salida para tsdepuración",
"Over-the-air", "A través del aire",
"Over-the-air Cron multi-line", "Cronología multi-línea a través del aire",
"Over-the-air EPG", "Guía a través del aire",
"Over-the-air EPG grabber", "Capturador de EPG Over-the-air",
"Over-the-air grabbers", "Capturadores Over-the-air",
"Override disabled", "Invalidar deshabilitado",
"Override the frontend type.", "Sobrescribir el tipo de índice.",
"Owner", "Propietario",
"Owner of the entry.", "Propietario de la entrada.",
"Owner of the rule.", "Propietario de la regla.",
"PEND", "PENDIENTE",
"PIDs 21 in setup", "PIDs 21 en configuración",
"PIN inquiry match string", "Consulta de PIN de entrada",
"PIN inquiry match string.", "Información de PIN en la cadena de búsqueda.",
"PLP ID", "ID PLP",
"PLS code", "Código PLS",
"PLS mode", "Modo PLS",
"PROXY protocol is an extension for support incoming TCP connections from a remote server (like a firewall) sending the original IP address of the client. The HTTP header 'X-Forwarded-For' do the same with HTTP connections. Both enable tunneled connections.This option should be disabled for standard usage.", "El protocolo PROXY es una extensión para soportar las conexiones TCP entrantes desde un servidor remoto (como un firewall) que envía la dirección IP original del cliente. La cabecera HTTP 'X-Forwarded-For' hace lo mismo con las conexiones HTTP. Ambas activan conexiones túnelizadas. Esta opción debe desactivarse para el uso estándar.",
"Parent entry", "Entrada de causa",
"Parent entry.", "Entrada de causa",
"Pass Spectrum inversion to the SAT>IP server.", "Pasar inversión de espectro al servidor SAT>IP.",
"Pass specinv", "Pasar specinv",
"Pass subscription weight", "Pasar peso de suscripción",
"Pass subscription weight to the SAT>IP server (Tvheadend specific extension).", "Pase el peso de la suscripción al servidor SAT>iP (extensión específica de Tvheadend).",
"Pass-thru muxer", "Muxer pasarela",
"Passthrough Muxer SI Tables", "Tablas Muxer SI de paso",
"Password", "Contraseña",
"Password for the entry.", "Contraseña para la entrada.",
"Password2", "Contraseña2",
"Path", "Ruta",
"Path to a directory (folder) containing your picon collection. See Help for more detailed information.", "Camino al directorio (carpeta) que contiene su colección de picón. Consulte la Ayuda para obtener información más detallada.",
"Path to an icon for this channel. This can be named however you wish, as either a local (file://) or remote (http://) image. See Help for more infomation.", "Camino a un icono de este canal. Esto puede ser nombrado como desee, ya sea como local (file://) o remotamente (http://). Consulte la Ayuda para obtener más información.",
"Path to the frontend used by the device.", "Ruta de acceso a la interfaz utilizada por el dispositivo.",
"Path to the grabber executable.", "Ruta al ejecutable del capturador.",
"Path to the socket Tvheadend will read data from.", "Ruta de acceso al socket de Tvheadend del que se leerán datos.",
"Path to where the timeshift data will be stored. If nothing is specified this will default to CONF_DIR/timeshift/buffer.", "Ruta del lugar dónde se almacenarán los datos de Timeshift. Si no se especifica nada, esto será por defecto en CONF_DIR/timeshift/buffer..",
"Path used by the device.", "Ruta utilizada por el dispositivo.",
"Path where the recordings are stored. If components of the path do not exist, Tvheadend will try to create them.", "Ruta en la que se almacenan las grabaciones. Si no existen las componentes de la ruta, Tvheadend intentará crearlas.",
"Peak count of objects", "Recuento máximo de objetos",
"Peak size", "Tamaño pico",
"Performing arts", "Artes escénicas",
"Periodically save EPG to disk (hours)", "Guardado periódico de la guía en disco (horas)",
"Persistent user interface level", "Nivel de interfaz de usuario persistente",
"Physiology", "Fisiología",
"Pick the stream with given audio type only.", "Escoja la retransmisión con el tipo de audio dado solamente.",
"Picon name scheme", "Estructura de nombres Picon",
"Picon path", "Ruta de Picon",
"Pilot", "Piloto",
"Please, follow %s link and cancel the next authorization to correctly clear the cached browser credentals (login and password cache). Then click to the 'Default login' (anonymous access) or 'New login' link in the error page to reauthenticate.", "Por favor, sigue %s enlace y cancela la siguiente autorización para limpiar correctamente los credenciales en cache (usuario y contraseña en cache). Después haz click en 'Login por defecto' (acceso anónimo) o 'Nuevo login' junto a la página de error para re-autenticarte.",
"Polarization", "Polarización",
"Political issues", "Problemas Políticos",
"Poll multiplexer", "Encuesta multiplexor",
"Popular culture", "Cultura Popular",
"Port", "Puerto",
"Port to connect to.", "Puerto al que conectar.",
"Port to listen on or to connect to.", "Puerto en el que escuchar o conectarse.",
"Position", "Posición",
"Position of the input.", "Posición de la entrada",
"Post-processor command", "Comando de post-procesador",
"Post-recording padding", "Postgrabaciones",
"Post-remove command", "Comando de post-eliminación",
"Power save", "Ahorro de energía",
"Power-up time (ms) (15-200)", "Tiempo encendido (ms) (15-200)",
"Pre-defined muxes", "Muxes pre-definidos",
"Pre-processor command", "Comando Pre-Procesador",
"Pre-recording padding", "Pregrabaciones",
"Pre-school children's programs", "Programas para niños en edad preescolar",
"Prefer picons over channel name", "Preferir picons sobre el nombre del canal",
"Preferred CA PID", "CA PID Preferido",
"Preferred audio language.", "Idioma de audio preferido.",
"Preferred service video type", "Tipo de servicio de vídeo preferido",
"Presentation", "Presentación",
"Presentation details.", "Detalles de presentación",
"Press", "Prensa",
"Prevent the user from overriding the default user interface level setting and removes the view level drop-dowm from the interface.", "Evite que el usuario reemplace el valor predeterminado del nivel de interfaz de usuario y elimine el nivel de visualización drop-dowm de la interfaz.",
"Prevent users from overriding the user interface level setting, and remove the view level drop-dowm from the interface.", "Evite que los usuarios sobrescriban la configuración del nivel de la interfaz de usuario y elimine el drop-down del nivel de vista de la interfaz.",
"Priority", "Prioridad",
"Priority (-10..10)", "Prioridad(-10..10)",
"Priority of the entry, higher-priority entries will take precedence and cancel lower-priority events.", "La prioridad de la entrada, las entradas de mayor prioridad tendrán precedencia y cancelarán los eventos de menor prioridad.",
"Priority of the recording. Higher priority entries will take precedence and cancel lower-priority events.", "Prioridad de la grabación. Las entradas de mayor prioridad tendrán prioridad y cancelarán los eventos de menor prioridad.",
"Private", "Privado",
"Profile name", "Nombre del perfil",
"Program identification (PID) number to compare. Zero means any. This comparison is processed only when service comparison is active and for the Conditional Access filter.", "Identificación de programa (PID), número de comparar. Cero significa ninguna. Esta comparación se procesa sólo cuando la comparación del servicio está activa y para el filtro de acceso condicional.",
"Program synopsis (display only).", "Sinopsis de programa (mostrar únicamente).",
"Program synopsis.", "Sinopsis de programa.",
"Provider", "Proveedor",
"Provider ID", "ID de Proveedor",
"Provider network name", "Nombre de provedor de red",
"Provider's network name.", "Nombre del proveedor de red.",
"Psychology", "Psicología",
"Puppets", "Títeres",
"PyEPG Import", "Importación PyEPG",
"Quad (4.0)", "Cuádruple (4.0)",
"Quiz", "Pregunta",
"RAM only", "Solo RAM",
"RTP over TCP payload", "Carga útil RTP sobre TCP",
"RTP/AVP/TCP (embedded data)", "RTP/AVP/TCP (datos incrustados)",
"RTSP Protocol", "Protocolo RTSP",
"RTSP port", "Puerto RTSP",
"RTSP port (554 or 9983, 0 = disable)", "Puerto RTSP (554 o 9983, 0 = deshabilitado)",
"Raw Audio Stream", "Retransmisión de Audio Limpia",
"Re-fetch period (hours)", "Periodo de recuperación (horas)",
"Re-fetch period (mins)", "Periodo de Recuperación (mins)",
"Re-fetch the bouquet every x minutes.", "Re-descargar el paquetizado cada x mins.",
"Re-record%s%s", "Re-grabar%s%s",
"Re-try period (hours)", "Periodo de reintento (horas)",
"Read error", "Error de lectura",
"Real Time Streaming Protocol (RTSP) port the server should listen on (554 or 9983, 0 = disable).", "Protocolo de retransmisión en tiempo real Protocol (RTSP) el servidor debe escuchar en el puerto (554 o 9983, 0 = desactivar).",
"Record all", "Grabar todo",
"Record on these days only.", "Grabar en esos días unicamente.",
"Recording file options", "Grabando opciones de archivo",
"Recording file size.", "Tamaño del archivo de grabación.",
"Recording play count", "Recuento de grabaciones",
"Recording system path", "Grabando ruta del sistema",
"Reject", "Rechazar",
"Reject exact match", "Rechazar coincidencia exacta",
"Relative path to the imagecache copy of the icon.", "Ruta relativa a la copia en cache del icono.",
"Religion", "Religión",
"Religious", "Religioso",
"Remarkable people", "Gente destacable",
"Remove HTTP arguments", "Eliminar argumentos HTTP",
"Remove all unsafe characters from filename", "Eliminar todos los carácteres inseguros del nombre del archivo",
"Remove scrambled bits", "Eliminar los bits codificados",
"Remove server support for RTP/AVP/TCP transfer mode (embedded data in the RTSP session).", "Eliminar soporte del servidor para el modo de transferencia RTP/AVP/TCP (contenidos RTP usados en TCP en modo de datos incrustados).",
"Repeated", "Repetido",
"Replace the full Transport Stream with a range 0x00-0x02,0x10-0x1F,0x1FFB pids only.", "Reemplazar toda la trama de transporte por un rango de pids únicamente;  0x00-0x02,0x10-0x1F,0x1FFB.",
"Replace whitespace in title with '-'", "Reemplazar espacio en blanco con '-'",
"Replaces all whitespace in the title with '-'.", "Reemplazar todos los espacios en blanco en el título con '-'.",
"Reply to CAM PIN inquiries", "Responder a consultas CAM PIN",
"Reply to PIN inquiries.", "Contestar a las peticiones de PIN.",
"Reports", "Reportes",
"Rerun of", "Volver a ejecutar",
"Reserved", "Reservado",
"Resolution (height)", "Resolución (Altura)",
"Respawn (pipe)", "Reaparecer (tubería)",
"Restart", "Reiniciar",
"Restart on error", "Reiniciar en error",
"Restart streaming on error. This is useful for DVR so a recording isn't aborted if an error occurs.", "Reiniciar la transmisión en caso de error. Esto es útil para DVR por lo que una grabación no se interrumpe si se produce un error.",
"Restart when the subscription is overriden (in then timeout time window).", "Reiniciar cuando la suscripción sea sobreescrita (en la ventana de tiempo agotado).",
"Restrict \"pids=all\"", "Restringir \"pids=all\"",
"Restrict connections to this type.", "Restringir las conexiones a este tipo.",
"Restricted log mode (no administrator)", "Modo de registro restrictivo (no es administrador)",
"Reuse EPG from", "Reusar EPG desde",
"Reuse the EPG from another channel.", "Reutilizar EPG de otro canal.",
"Rewrite EIT", "Reescribir EIT",
"Rewrite EIT (Event Information Table) packets to only include information about the currently-streamed service.", "Reescribir EIT (Tabla de Información de Eventos), paquetes para incluir información sobre el servicio de streaming actual.",
"Rewrite MPEG-TS SI tables", "Reescribir MPEG-TS SI tablas",
"Rewrite PAT", "Reescribir PAT",
"Rewrite PAT (Program Association Table) packets to only include information about the currently-streamed service.", "Reescribir PAT (Tabla de Asociación de Programas), paquetes para incluir información sobre el servicio de streaming actual.",
"Rewrite PMT", "Reescribir PMT",
"Rewrite PMT (Program Map Table) packets to only include information about the currently-streamed service.", "Reescribir PMT (Tabla de Mapeo de Porgramas), paquetes para incluir información sobre el servicio de streaming actual.",
"Rewrite Program Map Table (PMT) packets to only include information about the currently streamed service.", "Reescriba los paquetes de la Tabla de mapas de programas (PMT) para incluir únicamente información sobre el servicio transmitido actualmente.",
"Rewrite SDT", "Reescribir SDT",
"Rewrite SDT (Service Description Table) packets to only include information about the currently-streamed service.", "Reescribir SDT (Tabla de Descripción de Servicios), paquetes para incluir información sobre el servicio de streaming actual.",
"Rights", "Derechos",
"Rotor initialization time (seconds)", "Tiempo de Iniciación del rotor (segundos)",
"Rotor type", "Tipo de rotor",
"Run as group", "Ejecutar como grupo",
"Run as user", "Ejecutar como usuario",
"Running", "Funcionando",
"SAT->IP frontend number.", "Número de índice SAT->IP",
"SAT>IP ATSC-C Frontend", "Índice SAT>IP ATSC-C",
"SAT>IP ATSC-T Frontend", "índice SAT>IP ATSC-T",
"SAT>IP Client", "Cliente SAT>IP",
"SAT>IP DVB Frontend", "Índice SAT>IP DVB",
"SAT>IP DVB-C Frontend", "Índice SAT>IP DVB-C",
"SAT>IP DVB-C frequency (Hz)", "Frecuencia (en Hz) para SAT>IP DVB-C",
"SAT>IP DVB-S Frontend", "Índice SAT>IP DVB-S",
"SAT>IP DVB-S Slave Frontend", "Interfaz esclava SAT>IP DVB-S",
"SAT>IP DVB-S frequency (kHz)", "Frecuencia (en kHz) para SAT>IP DVB-S",
"SAT>IP DVB-T Frontend", "Índice SAT>IP DVB-T",
"SAT>IP DVB-T frequency (Hz)", "SAT>IP DVB-T frecuencia (Hz)",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "Número de puerto para el servidor SAT>IP RTSP\n(por defecto: -1 = deshabilitado, 0 = configuración web, el puerto estándar es 554)",
"SAT>IP Satellite Configuration", "SAT>Configuración Satélite IP",
"SAT>IP Server", "Servidor SAT>IP",
"SAT>IP Server SI Tables", "Tablas SAT>IP SI de Servidor",
"SAT>IP source number", "Número de fuente SAT>IP",
"SCR (Satellite Channel Router) ID.", "SCR (Enrutador de Canal de Satélite) ID.",
"SD: standard definition", "SD: definición estándar",
"SSL verify peer", "Verificar pares SSL",
"START", "INICIAR",
"STOP", "PARAR",
"Sat", "Sáb",
"Satellite config", "Configuración de satélite",
"Satellite longitude", "Longitud de satélite",
"Satellite longitude.", "Longitud de satélite.",
"Satellite position.", "Posición de satélite.",
"Satellite positions", "Posiciones de satélite",
"Save EPG to disk after xmltv import", "Guardar EPG en el disco después de importación xmltv",
"Scan after creation", "Escanear después de crear",
"Scan progress", "Progreso de escaneo",
"Scan queue length", "Logitud de la lista de espera del escaneo",
"Scan result", "Resultado del escaneo",
"Scan status", "Estado del escaneo",
"Scanfile", "Archivo de escaneo",
"Schedule a re-recording if more errors than (0=off)", "Aplazar una re-grabación si se producen más errores que (0 = apagado)",
"Schedule frequency (in cron format).", "Frecuencia de planificación (en formato cron).",
"Schedule status", "Estado del escaneo",
"Schedule status.", "Estado de programación.",
"Scheduled Duration", "Duración de programación.",
"Scheduled for recording", "Aplazado para grabación",
"Scheduled start time", "Horario de comienzo programado",
"Scheduled stop time", "Horario de fin programado",
"Scheme to generate the channel icon names (all lower-case, service name picons etc.).", "Esquema para generar los nombres de los iconos de canal (todos minúsculas, picons del nombres de servicio, etc.).",
"School programs", "Programas de colegio",
"Science", "Ciencia",
"Science fiction", "Ciencia ficción",
"Script/program to run when a recording completes.", "Script/programa a ejecutarse cuando una grabación se completa.",
"Script/program to run when a recording gets removed.", "Script/programa a ejecutarse cuando una grabación se borre.",
"Script/program to run when a recording starts (service is subscribed but no filename available).", "Script/programa a ejecutar cuando se inicie una grabación (el servicio está suscrito pero no hay ningún nombre de archivo disponible).",
"Season", "Temporada",
"Season %d", "Temporada %d",
"Season information (if available).", "Información de la temporada (si está disponible).",
"Seek error", "Error de busqueda",
"Select an available network type for this tuner.", "Seleccione un tipo de red disponible para este sintonizador.",
"Select high priority (default) EPG language.", "Seleccionar el idioma de alta prioridad (por defecto) en EPG.",
"Select how Tvheadend should handle muxes. See Help for details.", "Especifica como Tvheadend debería manejar los muxes. Mira la Ayuda para más detalles.",
"Select low priority EPG language.", "Seleccionar el idioma de prioridad baja en EPG.",
"Select medium priority EPG language.", "Seleccionar el idioma de prioridad media en EPG.",
"Select scheme to generate the picon names (standard, force service type to 1)", "Elija un esquema para generar los nombres de picon (estándar, forzar tipo de servicio a 1)",
"Select subtitle codec to use for transcoding.", "Seleccionar códec de subtítulos a usar para la transcodificación.",
"Select the default user interface language. This can be overridden later in \"Access Entries\" on a per-user basis.", "Seleccione el idioma de la interfaz de usuario por defecto. Esto puede cambiarse más adelante en \"Entradas de acceso\" en función de cada usuario.",
"Select the list of languages (in order of priority) to be used for supplying EPG information to clients that don't provide their own configuration.", "Seleccione la lista de idiomas (en orden de prioridad) que se utilizará para el suministro de la información EPG a los clientes que no proporcionan su propia configuración.",
"Select the master tuner.The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, the polarization and low-high band settings must be equal.if you set other tuner as master, then this tuner will act like a slave one and tvheadend will assure that this tuner will not use incompatible parameters (position, polarization, lo-hi).", "Seleccione la señal tuner. La señal LNB universal estándar se puede dividir mediante un sencillo divisor coaxial (sin multiswitch) a varias salidas. En este caso, la posición, la configuración de polarización y bajo-alto de la banda debe ser igual. Si configura otro sintonizador como maestro, a continuación, este sintonizador actuará como un esclavo y un solo tvheadend asegurará que este sintonizador no utilizará los parámetros incompatibles (posición, polarización, lo-hi).",
"Select the number of satellite positions supported by the SAT>IP hardware and your coaxial cable wiring.", "Seleccione el número de posiciones de los satélites con el apoyo del SAT>IP y el cableado de su cable coaxial.",
"Select the path to use for DVB scan configuration files. Typically dvb-apps stores these in /usr/share/dvb/. Leave blank to use the internal file set.", "Seleccionar la rura para el escaneo DVB de los archivos. Típicamente dvb-apps lo guarda en /usr/share/dvb/. Dejar en blanco para usar los archivos internos.",
"Select the payload size for RTP contents used in the TCP embedded data mode. Some implementations like ffmpeg and VideoLAN have maximum limit 7896 bytes. The recommended value for tvheadend is 16356 or 32712 bytes.", "Seleccione el tamaño de carga útil para los contenidos RTP usados en TCP en modo de datos incrustados. Algunas implementaciones como ffmpeg o VideoLAN tienen un límite máximo de 7896 bytes. El valor recomendado para tvheadend es  16356 o 32712 bytes.",
"Select the time offset for EIT events.", "Seleccionar el tiempo de desplazamiento de eventos EIT.",
"Select type of broadcast.", "Seleccionar tipo de emisión.",
"Select whether or not to enable caching. Note, even with this disabled you can still specify local (file://) icons and these will be served by the built-in webserver.", "Seleccione si desea o no habilitar el almacenamiento en caché. Tenga en cuenta que, incluso con esta opción deshabilitada, todavía puede especificar iconos locales (file: //) y éstos serán servidos por el servidor web incorporado.",
"Select/Selected services to map.", "Seleccionar/Seleccionado servicios a mapear.",
"Send CAPMT OK query before descrambling.", "Enviar consulta OK a CAPMT después del descifrado.",
"Send CAPMT query", "Enviar consulta CAPMT",
"Send full PLAY cmd", "Enviar reproducción completa cmd",
"Send previous stream frames to upper layers (before frame start is signalled in the stream). It may cause issues with some clients / players.", "Enviar retransmisiones anteriores a las capas superiores (antes del inicio de retransmisión señalado). Esto puede causar problemas con algunos clientes / reproductores.",
"Send the full RTSP PLAY command after full RTSP SETUP command. Some devices firmware require this to get an MPEG-TS stream.", "Enviar el comando completo de RTSP PLAY y después el comando completo CONFIGURACIÓN RTSP. Algunos dispositivos de firmware requieren de esto para conseguir una retransmisión MPEG-TS.",
"Serial number", "Número de serie",
"Series link", "Enlace a series",
"Series link ID.", "Series de acoplamiento de ID.",
"Serious", "Serio/a",
"Serious music", "Música Seria",
"Server", "Servidor",
"Server UUID", "Servidor UUID",
"Server connectivity", "Conectividad del servidor",
"Server details.", "Detalles de servidor.",
"Service", "Servicio",
"Service ID", "ID del Servicio",
"Service Mapper", "Mapeador de servicio",
"Service Mapping (Map services to channels)", "Mapeo de Servicios (Mapear servicios a canales)",
"Service configuration", "Configuración del servicio",
"Service count %zi", "Recuento de servicio %zi",
"Service mapping", "Mapeado de servicio",
"Service name", "Nombre del servicio",
"Service name found in EPG data.", "Nombre de servicio encontrado en datos de EPG.",
"Service name picons", " Nombre de picons del servicio",
"Service priority. Enter a value between -10 and 10. A higher value indicates a higher preference. See Help for more info.", "Prioridad del servicio. Introduzca un valor entre -10 y 10. Un valor más alto indica una mayor prioridad. Consulte la Ayuda para obtener más información.",
"Service raw", "Texto plano del servicio",
"Service type", "Tipo de servicio",
"Service type override. This value will override the service type provided by the stream.", "Sobrescribir el tipo de servicio. Este valor prevalecerá sobre el tipo de servicio prestado por la retransmisión.",
"Services", "Servicios",
"Services associated with the channel.", "Servicios asociados con los canales.",
"Services seen", "Servicios vistos",
"Set as default profile.", "Elegir como perfil por defecto.",
"Set the display name.", "Establezca el nombre para mostrar.",
"Set the maxiumum packet identifiers your SAT>IP server supports.", "Establezca los identificadores de paquete máximos que su SAT> Servidor IP soporta.",
"Set the name of the server so you can distinguish multiple instances apart.", "Establezca el nombre del servidor para que pueda distinguir varias instancias aparte.",
"Set the priority of this configuration.", "Establecer la prioridad en esta configuración.",
"Sets the default interface view level (next to the Help button).", "Establece el nivel de vista de la interfaz por defecto (al lado del botón de ayuda).",
"Settings", "Ajustes",
"Shift PTS (ms)", "Retardo PTS (ms)",
"Show", "Mostrar",
"Show / Game show", "Show / Show de juegos",
"Show only information for sessions which are initiated from an IP address of the requester.", "Mostrar solo información para las sesiones que se inician desde una dirección IP del solicitante.",
"Show this page", "Mostrar esta página",
"Show version information", "Mostrar información de versión",
"Show, hide and sort the various details that appear on the interface next to the About tab.", "Mostrar, ocultar y ordenar los diversos detalles que aparecen en la interfaz de al lado de la pestaña Acerca de.",
"Signal level for IPTV sources (0-240).", "Nivel de señal para fuentes IPTV (0-240).",
"Signal quality too poor", "Calidad de señal muy pobre",
"Signal scale (240 or 100)", "Escala de señal (240 o 100)",
"Site latitude", "Latitud del sitio",
"Site latitude.", "Latitud del sitio.",
"Site longitude", "Longitud del sitio",
"Site longitude.", "Longitud del sitio.",
"Size", "Tamaño",
"Skip TS packets (0-200)", "Saltar paquetes TS (0-200)",
"Skip TSID checking. For when providers use invalid Transport Stream IDs.", "Saltar chequeos TSID. Para cuando los proveedores usan identificadores de transporte no válidos.",
"Skip commercials", "Saltar comerciales",
"Skip initial bytes", "Saltar bytes iniciales",
"Skip scanning known muxes when Tvheadend starts. If \"startup scan\" is allowed and new muxes are found then they will still be scanned. See Help for more details.", "Saltar muxes conocidos cuando se inicia Tvheadend. Si se permite la \"exploración inicial\" y se encuentran nuevos muxes, estos continuarán siendo escaneados. Consulte la Ayuda para obtener más detalles.",
"Skip startup scan", "Saltar escaneo en arranque",
"Skip x number of transport packets.", "Saltar x número de paquetes de transporte.",
"Slave entry", "Entrada esclava",
"Slave entry.", "Entrada esclava.",
"Slot state", "Estado del espacio",
"Soap", "Telenovela",
"Soccer", "Fútbol",
"Social / Political issues / Economics", "Social / Problemas Políticos / Economía",
"Social advisory", "Asesoramiento social",
"Socket or IP Address (when in TCP mode).", "Socket o Dirección IP (en caso de modo TCP).",
"Some DVB devices have an optional ioctl that allows changing between normal voltage for LNB (13V/18V) to a higher voltage mode (usually, 14V/19V), meant to compensate for voltage loss on long cabling. Without that, it is not possible to properly switch the polarization.", "Algunos dispositivos DVB tinen una llamada en el driver que permite cambiar entre el voltaje normal del LNB (13V/18V) a un modo de voltaje incrementado (generalmente, 14V/19V), con intención de compensar las perdidas de voltaje en un cableado largo. Sin esta opción puede ser dificil cambiar de polarización en determinadas instalaciones.",
"Sort index", "Ordenar índice",
"Sort index.", "Índice de clasificación.",
"Source", "Fuente",
"Source deleted", "Fuente eliminada",
"Source quality is bad", "La Fuente de calidad es mala",
"Source reconfigured", "Fuente reconfigurada",
"Source video codec", "Fuente de Códec de video",
"Southern hemisphere (latitude direction)", "Hemisferio sur (dirección latitud)",
"Southern hemisphere (latitude direction).", "Hemisferio sur (dirección de latitud).",
"Spawn", "Aparecer",
"Special events (Olympic Games, World Cup, etc.)", "Eventos especiales (Juegos Olímpicos, Copa del Mundo, etc.)",
"Specifies the incoming buffering limit in milliseconds (PCR based). If the PCR time difference from the system clock is higher than this, the incoming stream is paused.", "Especifica el límite de almacenamiento en búfer de entrada en milisegundos (basado en PCR). Si la diferencia de PCR en tiempo del reloj del sistema es mayor que esta, la retransmisión de entrada será pausada.",
"Specify User-Agent header for the http client", "Especifique cabecera User-Agent para el cliente http",
"Specify alternative htsp port", "Especifique alternativa al puerto htsp",
"Specify alternative http port", "Especifique alternativa al puerto http",
"Specify alternative http webroot", "Especifique alternativa al puerto http webroot",
"Specify bind address", "Especificar dirección unida",
"Specify bind address for SAT>IP server", "Especificar dirección unida para servidor SAT>IP",
"Specify extra htsp port", "Especificar puerto htsp extra",
"Specify the parameters to be changed. See Help for details.", "Especifica los parámetros a ser cambiados. Mira la Ayuda para más detalles.",
"Spiritual sciences", "Ciencias espirituales",
"Sports", "Deportes",
"Sports magazines", "Revistas de deportes",
"Standard", "Por defecto",
"Start", "Inicio",
"Start after", "Iniciar después de",
"Start before", "Iniciar antes de",
"Start recording earlier than the EPG/timer defined start time by x minutes, for example if a program is to start at 13:00 and you set a padding of 5 minutes it will start recording at 12:54:30 (including a warming-up time of 30 seconds). If this isn't set the pre-recording padding if set in the DVR entry or DVR profile will be used.", "Inicie la grabación antes de la hora de inicio EPG/cronómetro por x minutos, por ejemplo si un programa se inicia a las 13:00 y se establece un relleno de 5 minutos éste comenzará a grabar a las 12:54:30 (incluyendo un calentamiento-Tiempo de activación de 30 segundos). Si no se establece el relleno de pregrabación se usará el establecido en la entrada DVR o el perfil DVR.",
"Start recording earlier than the EPG/timer-defined start time by x minutes.", "Iniciar la grabación x minutos antes del EPG/temporizador.",
"Start recording earlier than the defined Start recording earlier than the defined start time by x minutes: for example, if a program is to start at 13:00 and you set a padding of 5 minutes it will start recording at 12:54:30 (including a warm-up time of 30 seconds). If this isn't specified, any pre-recording padding as set in the channel or DVR entry will be used.", "Iniciar la grabación antes de la hora de inicio definida por x minutos: por ejemplo, si un programa se va a iniciar a las 13:00 y se establece un margen de 5 minutos, este se comenzará a grabar a las 12:54:30 ( incluyendo un tiempo de calentamiento de 30 segundos). Si no se especifica, se utilizará cualquier margen antes de la grabación como se establece en la entrada del canal o DVR.",
"Start recording earlier than the defined start time by x minutes.", "Iniciar la grabación antes de la hora de inicio definida por x minutos.",
"Start time", "Tiempo de inicio",
"Status", "Estado",
"Status period (ms)", "Periodo de estado (ms)",
"Stop", "Parar",
"Stop time", "Parar tiempo",
"Storage path", "Ruta de almacenamiento",
"Storage space", "Espacio de almacenamiento",
"Stream ID", "ID de Retransmisión",
"Stream Profile", "Perfil de Retransmisión",
"Stream index", "Índice de retransmisión",
"Stream index (starts with zero).", "Índice de retransmisión (comienza en cero).",
"Stream profile", "Perfil de retransmisión",
"Stream setup error", "Error de ejecución de retransmisión",
"Stream type", "Tipo de retransmisión",
"Streaming", "Retransmisión",
"Streaming Profile", "Perfil de retransmisión",
"Streaming flags, allow/disallow HTTP streaming, advanced HTTP streaming (e.g, direct service or mux links), HTSP protocol streaming (e.g, Kodi (via pvr.hts) or Movian.", "Características de Streaming, permitir/prohibir la transmisión HTTP, streaming HTTP avanzado (por ejemplo, servicio directo o enlaces mux), transmisión de protocolo HTSP (por ejemplo, Kodi (a través de pvr.hts) o Movian.",
"Streaming priority", "Prioridad de retransmisón",
"Streaming profiles", "Perfiles de retransmisión",
"Subdirectory options", "Opciones de subdirectorio",
"Subscribe to a service permanently", "Suscribir a un servicio permanentemente",
"Subscription", "Suscripción",
"Subscription does not exist", "La suscripción no existe",
"Subscription overridden", "Suscripción anulada",
"Subscription weight", "Peso de suscripción",
"Substitute formatters", "Formatos de subtítulos",
"Subtitle", "Subtítulo",
"Subtitle Stream Filter", "Filtro de Subtítulos en Retransmisión",
"Subtitle codec", "Códec de subtítulo",
"Subtitle of the program (if any) (display only).", "Subtitulo del programa (si hay) (mostrar únicamente).",
"Subtitle of the program (if any).", "Subtítulo del programa (si existe).",
"Sun", "Dom",
"Surround (2 front, rear mono)", "Envolvente (2 frontales, posterior mono)",
"Switch before rotor", "Cambiar después del rotor",
"Switch frequency offset", "Cambiar el offset en frecuencia",
"Switch off the power to the LNB when idle. Note: this may cause interference with other devices when the LNB is powered back up.", "Desconectar la alimentación al LNB cuando está inactivo. Nota: esto puede causar interferencias con otros dispositivos cuando el LNB se vuelve a encender.",
"Switch to another service", "Cambiar a otro servicio",
"Switch type", "Cambiar tipo",
"Switch: %s", "Cambiar: %s",
"Symbol rate (Sym/s)", "Tasa de símbolos (Símb/s)",
"Sync", "Sincronizar",
"Sync + Don't keep", "Sincronizar + No recordar",
"System", "Sistema",
"TCP Protocol", "Protocolo TCP",
"TV Adapters - Linux ATSC-C Frontend", "Adaptadores de TV - Índice Linux ATSC-C",
"TV Adapters - Linux ATSC-T Frontend", "Adaptadores de TV - Índice Linux ATSC-T",
"TV Adapters - Linux DAB Frontend", "Adaptadores de TV - Índice Linux DAB",
"TV Adapters - Linux DVB-C Frontend", "Adaptadores de TV - Índice Linux DVB-C",
"TV Adapters - Linux DVB-S Frontend (Master)", "Adaptadores de TV - Índice Linux DVB-S (Maestro)",
"TV Adapters - Linux DVB-S Slave Frontend", "Adaptadores de TV - Índice Linux DVB-S Esclavo",
"TV Adapters - Linux DVB-T Frontend", "Adaptadores de TV - Índice Linux DVB-T",
"TV Adapters - Linux ISDB-C Frontend", "Adaptadores de TV - Índice Linux ISDB-C",
"TV Adapters - Linux ISDB-S Frontend (Master)", "Adaptadores de TV - Índice Linux ISDB-S (Maestro)",
"TV Adapters - Linux ISDB-T Frontend", "Adaptadores de TV - Índice Linux ISDB-T",
"TV Adapters - SatConfig - 4-Port", "TV Adapters - SatConfig - 4-Puertos",
"TV Adapters - SatConfig - Advanced", "TV Adapters - SatConfig - Avanzado",
"TV Adapters - SatConfig - DiseqC Rotor", "TV Adapters - SatConfig - Rotor DiseqC",
"TV Adapters - SatConfig - DiseqC Switch", "TV Adapters - SatConfig - Switch DiseqC",
"TV Adapters - SatConfig - GOTOX Rotor", "TV Adapters - SatConfig - Rotor GOTOX",
"TV Adapters - SatConfig - Tone Burst/2 Port", "TV Adapters - SatConfig - Tono Ráfaga/2 Puertos",
"TV Adapters - SatConfig - USALS Rotor", "TV Adapters - SatConfig - Rotor USALS ",
"TV Adapters - SatConfig - Universal LNB (Simple)", "TV Adapters - SatConfig - LNB Universal (Simple)",
"TVHDHomeRun Client", "Cliente TVHDHomeRun",
"Tag files with metadata", "Categoría de archivos con metadatos",
"Tags", "Categorías",
"Tags linked/to link to the channel.", "Etiquetas enlazadas/a para enlazar con el canal.",
"Talk show", "Programa de entrevistas",
"Team sports (excluding football)", "Deportes de quipo (excluyendo fútbol)",
"Technology", "Tecnología",
"Teletext Stream Filter", "Filtro de Retransmisión de Teletexto",
"Teletext stream type is only available for this filter.", "Éste tipo de retransmisión de teletexto sólo está disponible para este filtro.",
"Tennis", "Tenis",
"Testing", "Pruebas",
"Testing options", "Opciones de prueba",
"The ATSC source ID as set by the provider.", "La identificación de la fuente ATSC según lo establecido por el proveedor.",
"The CA provider to compare. Leave blank to apply to all providers.", "El proveedor de CA para comparar. Dejar en blanco para aplicar a todos los proveedores.",
"The CA stream type is only available for this filter.", "El tipo de corriente CA sólo está disponible para este filtro.",
"The CAID to compare. Leave blank to apply to all IDs.", "El CAID para comparar. Dejar en blanco para aplicar a todos los ID.",
"The CAM slot status.", "El estado de la ranura CAM.",
"The COFDM modulation used by the mux. If you're not sure of the value leave as AUTO.", "La modulación OFDM utilizada por el mux. Si no está seguro del valor déjelo como AUTO.",
"The Conditional Access ID used for the service.", "El Acceso Condicional ID utilizado por el servicio.",
"The Content reference identifier (CRID) authority.", "Identificador de la autoridad de contenido de referencia (CRID).",
"The DVR profile to be used/used by the recording.", "El perfil de DVR a ser utilizado/utilizado por la grabación.",
"The DVR profile to be used/used by this rule.", "El perfil de DVR que se utilizará/utilizado por esta regla.",
"The EPG ID used by the entry.", "La ID de EPG utilizada por la entrada.",
"The EPG grabber name.", "Nombre del capturador de EPG.",
"The EPG grabber to use on the mux. Enable (auto) is the recommended value.", "Capturador utilizado en el EPG de este mux. Habilitar (automático) es el valor recomendado.",
"The EPG grabber type.", "El tipo de capturador de EPG.",
"The MPEGTS stream type is only available for this filter.", "El tipo de retransmisión MPEGTS sólo está disponible para este filtro.",
"The OpenTV channel number as set by the provider.", "El número de canales de TV abierta según lo establecido por el proveedor.",
"The PIN to use.", "Pin para usar.",
"The Physical Layer Scrambling (PLS) code used on the mux.", "El código de capa física Scrambling (PLS) utilizado en el mux.",
"The Physical Layer Scrambling (PLS) mode used on the mux.", "El modo de capa física Scrambling (PLS) utilizado en el mux.",
"The Preferred Conditional Access Packet Identifier. Used for decrypting scrambled streams.", "Recomendado: Acceso Condicional de Paquete Identificador. Se utiliza para descifrar corrientes revueltas.",
"The SAT>IP server's name.", "El nombre del servidor de SAT> IP.",
"The SAT>IP server's universally unique identifier.", "El servidor SAT>IP universalmente identificador único.",
"The SAT>IP source number.", "El número de la fuente SAT> IP.",
"The SAT>IP's discovered IP address.", "La dirección IP descubierta por SAT>IP",
"The URL to the icon to use/used for the channel. The local files are referred using file:/// URLs.", "La URL para el icono a usar por el canal. Los archivos locales pueden ser usados con la estructura file:///.",
"The URL to the playlist.", "La URL a la lista de reproducción.",
"The audio stream types the filter should apply to.", "Los tipos de streaming de audio del filtro deben aplicarse.",
"The bandwidth the mux uses. If you're not sure of the value leave as AUTO but be aware that tuning may fail as some drivers do not like the AUTO setting.", "El ancho de banda utilizando por el MUX. Si no está seguro de el valor déjelo como AUTO pero tenga en cuenta que la sintonización puede fallar ya que a algunos drivers no les gusta el ajuste AUTO.",
"The bouquet the channel is associated with.", "El paquete al que está asociado el canal.",
"The cache scheme to use/used to store recordings. Leave as \"system\" unless you have a special use case for one of the others. See Help for details.", "El esquema de caché para usar/utilizado para almacenar las grabaciones. Dejar como \"sistema\" a menos que tengas un caso de uso especial para uno de los otros. Consulte la Ayuda para obtener más detalles.",
"The channel name the entry will record from.", "El nombre del canal de la entrada a grabar.",
"The channel on which this rule applies, i.e. the channel you're aiming to record. You can leave this field blank to apply the rule to all channels.", "El canal al que se aplica esta regla, por ejemplo, el canal que está apuntando a grabar. Puede dejar este campo en blanco para aplicar la regla a todos los canales.",
"The channel this service is mapped to.", "El canal al que este servicio esta mapeado.",
"The character encoding for this network (e.g. UTF-8).", "La codificación de caracteres para esta red (por ejemplo, UTF-8).",
"The character encoding for this service (e.g. UTF-8).", "La codificación de caracteres para este servicio (por ejemplo, UTF-8).",
"The character set to use/used. You should not have to change this unless channel names and EPG data appear garbled.", "El conjunto de caracteres a utilizar/utilizado. No debería tener que cambiar esto a menos que los nombres de canal y los datos de EPG aparezcan ilegibles.",
"The content type (Movie/Drama, Sports, etc.) to be used to filter matching events/programs.", "El tipo de contenido (Película/Drama, Deportes, etc.) que se utilizará para filtrar eventos/programas.",
"The current boot ID.", "El ID de inicio actual.",
"The current configuration ID.", "El ID de configuración actual.",
"The current configuration version.", "La versión actual de configuración.",
"The default language to use if the user  language isn't set in the Access Entries tab.", "El idioma por defecto a utilizar si el idioma del usuario no se encuentra en la pestaña de acceso.",
"The default subscription weight for each subscription.", "La fuerza de suscripción predeterminada para cada suscripción.",
"The default web interface to use if the user  theme isn't set in the Access Entries tab.", "La interfaz web por defecto a utilizar si el tema del usuario no se encuentra en la pestaña de las Entradas de Acceso.",
"The delivery system the mux uses. Make sure that your tuner supports the delivery system selected here.", "El sistema de suministro usado por el mux. Asegúrese de que el sintonizador es compatible con el sistema de suministro seleccionado aquí.",
"The delivery system used by the mux.", "Sistema de entrega utilizado por el mux.",
"The delivery system used by your cable provider.", "El sistema de entrega utilizado por su proveedor de cable.",
"The delivery system used by your provider.", "El sistema de entrega utilizado por su proveedor.",
"The demux path used by the device.", "La ruta demux utilizada por el dispositivo.",
"The device ID.", "La ID del dispositivo.",
"The device path in sysfs filesystem (/sys).", "Ruta del dispositivo en sistema de archivos (/sys).",
"The device's serial number.", "Número de serie del dispositivo.",
"The forward error correction high value. Most people will not need to change this setting.", "Corrección de errores de alto valor. La mayoría de la gente no tendrá que cambiar esta configuración.",
"The forward error correction low value. Most people will not need to change this setting.", "Corrección de errores de bajo valor . La mayoría de la gente no tendrá que cambiar esta configuración.",
"The forward error correction used on the mux.", "La corrección de errores utilizado en el mux.",
"The forward error correction. Most people will not need to change this setting.", "La corrección de errores hacia adelante. La mayoría de la gente no tendrá que cambiar esta configuración.",
"The frequency of the mux (in Hertz).", "La frecuencia del mux (en Hercios).",
"The frequency of the mux/transponder in Hertz.", "La frecuencia del mux/transpondedor en Hercios.",
"The frontend number given to the device.", "El número de índice dado al dispositivo.",
"The guard interval used by the mux. If you're not sure of the value leave as AUTO.", "El intervalo de guarda utilizado por el mux. Si no está seguro del valor a utilizar, déje el valor como AUTO.",
"The hierarchical modulation used by the mux. Most people will not need to change this setting.", "La modulación jerárquica utilizada por el mux. La mayoría de la gente no tendrá que cambiar esta configuración.",
"The imagecache path to the icon to use/used for the channel.", "La ruta imagecache al icono que se va a utilizar para el canal.",
"The input path used by the device.", "La ruta de entrada utilizada por el dispositivo.",
"The items have the following functions:", "Los artículos tienen las siguientes funciones:",
"The language to which the filter should apply.", "El idioma al que el filtro debe aplicarse.",
"The layer A constellation.", "Constelación de la capa A.",
"The layer A forward error correction.", "Saltarse errores de correción en la capa A.",
"The layer A segment count.", "Recuento de segmentos de la capa A.",
"The layer A time interleaving.", "Entrelazado de tiempo en la capa A.",
"The layer B constellation.", "Constelación de la capa B.",
"The layer B forward error correction.", "Saltarse errores de correción en la capa B.",
"The layer B segment count.", "Recuento de segmentos de la capa B.",
"The layer B time interleaving.", "Entrelazado de tiempo en la capa B.",
"The layer C constellation.", "Constelación de la capa C.",
"The layer C forward error correction.", "Saltarse errores de correción en la capa C.",
"The layer C segment count.", "Recuento de segmentos de la capa C.",
"The layer C time interleaving.", "Entrelazado de tiempo en la capa C.",
"The length of time (in seconds) to play the mux (1 hour = 3600).", "La longitud de tiempo (en segundos) para reproducir el mux (1 hora = 3600).",
"The limit for the PID filter (driver or hardware).", "El límite para el filtro PID (controlador o hardware).",
"The locking mechanism selection for The Preferred Conditional Access Packet Identifier. See Help for more information.", "La selección mecanismo de bloqueo para el Acceso Condicional  Preferido de Identificación de Paquetes. Consulte la Ayuda para obtener más información.",
"The logical stream index to compare. Note that this index is computed using all filters.Example: If filter is set to AC3 audio type and the language to 'eng' and there are two AC3 'eng' streams in the service, the first stream could be identified using number 1 and the second using number 2.", "El índice de retransmisión lógico a comparar. Tenga en cuenta que este índice se calcula utilizando todos los filtros. Ejemplo: Si el filtro se establece en tipo de audio AC3 y el idioma a 'eng' y hay dos corrientes AC3 'eng' en el servicio, la primera corriente se pudo identificar mediante el número 1 y la segundo utilizando el número 2.",
"The manufacturer of the SAT>IP server.", "El fabricante del servidor SAT>IP.",
"The maximal duration of a matching event - in other words, only match programmes that are no longer than this duration.", "La duración máxima de un evento - en otras palabras, sólo encontrar programas que no son más largos que esta duración.",
"The maximum RAM (system memory) size for timeshift buffers. When free RAM buffers are available they are used for timeshift data in preference to using storage.", "El tamaño máximo de memoria RAM (memoria del sistema) para buffers timeshift. Cuando hay buffers de memoria RAM libres, estos se utilizan para los datos de intercambio en lugar de utilizar el almacenamiento.",
"The maximum amount of time a grabber is allowed scan a mux for data (in seconds).", "El tiempo máximo que una fuente de guía puede escanear datos en un mux (en segundos).",
"The maximum combined size of all timeshift buffers. If you specify an unlimited period it's highly recommended you specify a value here.", "El tamaño combinado máximo de todas las memorias intermedias de buffer. Si se especifica un período de tiempo ilimitado es muy recomendable que especifique un valor aquí.",
"The maximum number of entries that can be matched.", "El número máximo de entradas que pueden ser emparejadas.",
"The maximum number of input streams allowed on this network.", "Número máximo de retransmisiones de entrada admitidos en esta red.",
"The maximum number of recording entries this rule can create.", "El número máximo de entradas de grabación que esta regla puede crear.",
"The maximum number of recordings that can be scheduled.", "Número máximo de grabaciones que se pueden programar.",
"The maximum number of services to decrypt per mux.", "El número máximo de servicios para descifrar por mux.",
"The maximum number of times this rule can be triggered.", "El número máximo de veces que esta regla puede desencadenarse.",
"The maximum time period that will be buffered for any given (client) subscription.", "El período máximo de tiempo que se van a almacenar para cualquier suscripción especificada (cliente)",
"The minimal duration of a matching event - in other words, only match programs that are no shorter than this duration.", "La duración mínima de un evento - en otras palabras, sólo encontrar programas que no son más cortos que esta duración.",
"The minimum delay after the rotor movement command is sent.", "El retardo mínimo después del comando de movimiento enviado al rotor.",
"The minimum delay before tuning in milliseconds after tuner stop. If the time between the previous and next start is greater than this value then the delay is not applied.", "El retardo mínimo antes de dar una vuelta en milisegundos después de la parada del sintonizador. Si el tiempo transcurrido entre el inicio anterior y siguiente es superior a este valor, entonces no se aplica la demora.",
"The modulation used on the mux.", "La modulación usada por el mux.",
"The mux the service is on.", "El servicio de mux está encendido.",
"The mux to play when the entry is triggered.", "El mux a reproducir cuando se activa la entrada.",
"The mux's universally unique identifier.", "Identificador único universal del mux.",
"The name (or frequency) the mux is on.", "El nombre (o frecuencia) del mux está encendido.",
"The name given to/of the channel (This is how it'll appear in your EPG.)", "El nombre dado a/del canal (Así es como aparecerá en tu EPG.)",
"The name of the profile.", "El nombre del perfil.",
"The name of the the rule.", "El nombre de la regla.",
"The network prefix(es) to block, e.g.192.168.2.0/24 (comma-separated list).", "Los prefijos de red(es) para bloquear, p.ej.: 192.168.2.0/24 (lista separada por comas).",
"The network the mux is on.", "El mux de la red esta encendido.",
"The network the service is on.", "La red de servicio está activada.",
"The network's priority. The network with the highest priority value will be used out of preference if available. See Help for details.", "Prioridad de red. Una prioridad de red mayor será usado preferentemente si está disponible. Ver la Ayuda para detalles.",
"The network's service ID", "La ID del servicio de red",
"The networks assigned to the device.", "La red asignada al dispositivo.",
"The networks using this configuration.", "La red está usando esta configuración.",
"The networks' universally unique identifier (UUID).", "Identificador único universal de las redes (UUID).",
"The number of ATSC-C (Cable/AnnexB) tuners to export.", "El número de ATSC-C (Cable/AnnexB) sintonizadores para exportar.",
"The number of ATSC-T (Terresterial) tuners to export.", "El número de ATSC-T (Terrestres) sintonizadores para exportar.",
"The number of DVB-C (Cable) tuners to export.", "El número de DVB-C (Cable) sintonizadores para exportar.",
"The number of DVB-C2 (Cable) tuners to export.", "El número de DVB-C2 (Cable) sintonizadores para exportar.",
"The number of DVB-S (Satellite) tuners to export.", "El número de DVB-S (Satélite) sintonizadores para exportar.",
"The number of DVB-S2 (Satellite) tuners to export.", "El número de DVB-S2 (Satélite) sintonizadores para exportar.",
"The number of DVB-T (Terresterial) tuners to export.", "El número de DVB-T (Terrestres) sintonizadores para exportar.",
"The number of DVB-T2 (Terresterial) tuners to export.", "El número de ATSC-T2 (Terrestres) sintonizadores para exportar.",
"The number of MPEG-TS packets Tvheadend buffers in case there is a delay receiving CA keys. ", "El número de paquetes MPEG-TS de buffer usados por Tvheadend en caso de que haya un retraso al recibir las claves CA.",
"The number of allowed connections this user can make to the server.", "El número de conexiones permitidas que este usuario puede hacer al servidor.",
"The number of bytes to buffer. By default, linuxdvb's input buffer is 18800 bytes long. The accepted range is 18800-1880000 bytes.", "El número de bytes de buffer. Por defecto, en Linux el buffer de entrada dvb-s es de 18800 bytes de longitud. El rango aceptado es 18800-1880000 bytes.",
"The number of days cookies set by Tvheadend should expire.", "El número de días de cookies creados por Tvheadend en los que debe expirar.",
"The number of muxes left to scan on this network.", "El número de multiplexores a la izquierda a escanear en esta red.",
"The number of seconds to wait for a stream to start.", "El número de segundos de espera para comenzar una restransmisión.",
"The number of services currently mapped to channels.", "El número de servicios actualmente mapeados a canales.",
"The orbital position of the satellite the mux is on.", "Posición orbital donde el satélite se encuentra.",
"The orbital position of the satellite your dish is pointing at.", "La posición orbital del satélite al que su plato está apuntando.",
"The outcome of the last scan performed.", "El resultado del último escaneo efectuado.",
"The physical layer pipe ID. Most people will not need to change this setting.", "ID del interior de la tubería de capa física. La mayoría de la gente no necesitará cambiar esta configuración.",
"The playlist's character set.", "Contenido de caracteres de la lista de reproducción.",
"The polarization used on the mux.", "Polarización utilizada en el mux.",
"The position the channel will appear on your EPG. This is not used by Tvheadend internally, but rather intended to be used by HTSP clients for mapping to remote control buttons, presentation order, etc.", "La posición en la que  el canal aparecerá en tu EPG. Esto no es utilizado por Tvheadend internamente, sino que está destinado a ser utilizado por clientes HTSP para mapeo a botones de control remoto, orden de presentación, etc.",
"The priority of any recordings set because of this rule will take precedence and cancel lower-priority events.", "La prioridad de cualquier grabación establecida, debido a esta regla tendrá prioridad y cancelará eventos de menor prioridad.",
"The provider's ID.", "La ID de provedor de red.",
"The provider's name.", "El nombre del proveedor.",
"The provider's network ID.", "ID de provedor de red.",
"The provider's network name.", "Nombre de provedor de red",
"The quadrature amplitude modulation (QAM) used by the mux. If you're not sure of the value leave as AUTO.", "Modulación de amplitud en cuadratura (QAM) utilizada por el mux. Si no está seguro del valor, déjelo en AUTO.",
"The recorded file was removed intentionally", "El archivo grabado fue eliminado intencionadamente",
"The recording/entry status.", "El estado de la grabación/entrada.",
"The rolloff used on the mux.", "El Roloff utilizado en el mux.",
"The rule action defines the operation when all comparisons succeed. See Help for more information on what the various rules do.", "La acción de la regla define la operación cuando todas las comparaciones tengan éxito. Consulte la Ayuda para obtener más información sobre lo que hacen las distintas reglas.",
"The satellite configuration to use.", "Configuración de satélite a utilizar.",
"The scan state. New muxes will automatically be changed to the PEND state. You can change this to ACTIVE to queue a scan of this mux.", "El estado de exploración. Nuevos muxes serán automáticamente cambiados al estado PENDIENTE. Usted puede cambiar esto a ACTIVO para hacer cola a una exploración de este mux.",
"The scheduled start time, including any padding.", "La hora de inicio programada, incluyendo cualquier relleno.",
"The scheduled stop time, including any padding.", "El tiempo de parada programada, incluyendo cualquier relleno.",
"The scrambled bits in MPEG-TS packets are always cleared. It is a workaround for the special streams which are descrambled, but these bits are not touched.", "Los bits codificados en paquetes MPEG-TS siempre se borrarán. Es una solución para las retransmisiones especiales que son descifradas, pero estos bits no son tocados.",
"The selected video type should be preferred when multiple services are available for a channel.", "El tipo de vídeo seleccionado se debe preferir cuando varios servicios están disponibles para un canal.",
"The service ID", "La ID del Servicio",
"The service ID as set by the provider.", "La ID del servicio puesta por el proveedor.",
"The service ID.", "La ID del Servicio.",
"The service name as set by the provider.", "El nombre del servicio según lo establecido por el proveedor.",
"The service the filter should apply to. Leave blank to apply the filter to all services.", "El servicio de filtro debe aplicarse. Dejar en blanco para aplicar el filtro a todos los servicios.",
"The service to which the filter should apply. Leave blank to apply the filter to all services.", "El servicio de filtro debe aplicarse. Dejar en blanco para aplicar el filtro a todos los servicios.",
"The service type flag as defined by the DVB specifications (e.g. 0x02 = radio, 0x11 = MPEG2 HD TV, 0x19 = H.264 HD TV)", "La bandera de tipo de servicio según la definición de las especificaciones DVB (por ejemplo, 0x02 = radio, 0x11 = HDTV MPEG2, 0x19 = H.264 HD TV)",
"The service's channel minor as set by the provider.", "Canal menor del servicio según lo establecido por el proveedor.",
"The service's channel number as set by the provider.", "Número de canal del servicio según lo establecido por el proveedor.",
"The service's encryption status.", "Estado de cifrado de servicios.",
"The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, the polarization and low-high band settings must be equal. If you set another tuner as master, then this tuner will act as a slave and Tvheadend will assure that this tuner will not use incompatible parameters (position, polarization, lo-hi).", "La señal procedente del LNB universal estándar se puede dividir mediante un simple divisor coaxial (no conmutador múltiple) a varias salidas. En este caso, la posición, la polarización y la configuración baja-alta de la banda deben ser iguales. Si se establece otro sintonizador como maestro, en este caso, este sintonizador actuará como un esclavo y Tvheadend asegurará que este sintonizador no utilizará los parámetros incompatibles (posición, polarización, alta-baja).",
"The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, the polarization and low-high band settings must be equal. If you set another tuner as master, then this tuner will act as a slave and tvheadend will assure that this tuner will not use incompatible parameters (position, polarization, lo-hi).", "La señal procedente del LNB universal estándar se puede dividir mediante un simple divisor coaxial (no conmutador múltiple) a varias salidas. En este caso, la posición, la polarización y la configuración baja-alta de la banda deben ser iguales. Si se establece otro sintonizador como maestro, en este caso, este sintonizador actuará como un esclavo y Tvheadend asegurará que este sintonizador no utilizará los parámetros incompatibles (posición, polarización, alta-baja).",
"The start time of the recording.", "Hora de inicio de la grabación.",
"The stream ID used for the mux.", "La ID de retransmisión utilizada por el mux.",
"The stream profile the DVR profile will use for recordings.", "El perfil de retransmisión para el perfil DVR utilizado en grabaciones.",
"The streaming profile to use/used. If not set, the default will be used.", "Perfil de retransmisión a usar/utilizado. Si no se establece, se utilizará el valor por defecto.",
"The string allows you to manually specify the full path generation using predefined modifiers. See Help for full details.", "La cadena le permite especificar manualmente la generación de ruta completa usando modificadores predefinidos. Consulte Ayuda para obtener más detalles.",
"The subtitle stream types the filter should apply to.", "Los tipos de retransmisión de subtitulo a los que el filtro que debe aplicarse.",
"The symbol rate used on the mux.", "La velocidad de símbolo utilizado en el mux.",
"The symbol rate used on the mux/transponder.", "La velocidad de símbolo utilizado en el mux/transpondedor.",
"The symbol rate.", "Ratio de símbolos.",
"The time the entry stops/stopped being recorded.", "El tiempo en el que se detiene la entrada/deja de ser grabada.",
"The time window to cut the stop time from the overlapped event in seconds.", "La ventana de tiempo para cortar el tiempo de parada del evento superpuesto en segundos.",
"The title of the program to look for. Note that this accepts case-insensitive regular expressions.", "El título del programa a buscar. Tenga en cuenta que este acepta expresiones regulares entre mayúsculas y minúsculas.",
"The total number of services found.", "El número total de servicios encontrados.",
"The total scheduled duration.", "Duración total programada.",
"The transmission/OFDM mode used by the mux. If you're not sure of the value leave as AUTO but be aware that tuning may fail as some drivers do not like the AUTO setting.", "El modo de transmisión/OFDM utilizados por el mux. Si no está seguro del valor, déjelo como AUTO pero tenga en cuenta que la sintonización puede fallar ya que a algunos drivers no les gusta el ajuste AUTO.",
"The transponder ID.", "La ID de Transpondedor.",
"The transport stream ID of the mux within the network.", "El ID de transporte de retransmisión del mux dentro de la red.",
"The tuner priority value (a higher value means to use this tuner out of preference). See Help for details.", "El valor de prioridad del sintonizador (un valor más alto significa utilizar este sintonizador preferentemente). Consulte la Ayuda para obtener más información.",
"The tuner priority value for streamed channels through HTTP or HTSP (a higher value means to use this tuner out of preference). If not set (zero), the standard priority value is used. See Help for details.", "El valor de prioridad del sintonizador para los canales transmitidos a través de HTTP o HTSP (un valor más alto significa usar este sintonizador de preferencia). Si no se establece (cero), se utiliza el valor de prioridad estándar. Consulte la Ayuda para obtener más información.",
"The user who created the recording, or the auto-recording source and IP address if scheduled by a matching rule.", "El usuario que creó la grabación o la fuente de grabación automática y la dirección IP programada por una regla de coincidencia.",
"The version of Tvheadend that last updated the config.", "La última versión de Tvheadend que actualizó la configuración.",
"The video stream types the filter should apply to.", "Los tipos de retransmisión de vídeo a los que el filtro debe aplicarse.",
"Theme", "Tema",
"This tuner", "Este sintonizador",
"This will create an NTP driver (using shmem interface) that you can feed into ntpd. This can be run without root privileges, but generally the performance is not that great.", "Esto creará un controlador de NTP (utilizando la interfaz shmem) que se puede alimentar en ntpd. Esto se puede ejecutar sin privilegios de root, pero en general el rendimiento no es tan grande.",
"Thread", "Hilo",
"Thriller", "Programa de suspense",
"Thu", "Jue",
"Time", "Tiempo",
"Time (in milliseconds) for a command to complete.", "Tiempo (en milisengudos) para que un comando se complete.",
"Time (in milliseconds) for the rotor to power up.", "Tiempo (en milisegundos) para que el rotor de encienda.",
"Time (in minutes) to re-fetch the playlist.", "Tiempo (en minutos) para recargar la lista de producción.",
"Time Stamp Fix", "Corrección de la hora",
"Time missed", "Tiempo olvidado",
"Time record caption", "Tiempo de captación para grabación",
"Time recording%s%s", "Tiempo de grabación%s%s",
"Time to start the recording/time the recording started.", "Hora a comenzar la grabación/tiempo en que la grabación ha comenzado.",
"Time to stop recording/time the recording stopped.", "Tiempo para detener la grabación/tiempo tener la grabación detenida.",
"Time update", "Actualización de tiempo",
"Timeout (sec) (0=infinite)", "Tiempo agotado (seg) (0=infinito)",
"Timeout (seconds)", "Tiempo agotado (segundos)",
"Timeout (secs)", "Tiempo agotado (segs)",
"Timer-based automatic record caption.", "Temporizador-basado en epígrafe de registro automático.",
"Timer-based automatic recording.", "Temporizador-basado en grabación automática.",
"Title", "Título",
"Title (regexp)", "Título (regexp)",
"Title of the program (display only).", "Título del programa (solo mostrar).",
"Title of the program.", "Título del programa.",
"Title of the recording - this is used to generate the filename.", "Título de la grabación - usado para generar el nombre del archivo.",
"Toggle details", "Ver detalles",
"Tone burst", "Tono de ráfaga",
"Too many data errors", "Demasiados errores de datos",
"Total number of mapped channels on this network.", "Número total de canales mapeados en esta red.",
"Total number of muxes found on this network.", "Número total de muxes que se encuentran en esta red.",
"Total number of services found on this network.", "Número total de servicios encontrados en esta red.",
"Total number of services found.", "Número total de servicios encontrados.",
"Total number of services seen.", "Número total de los servicios vistos.",
"Total number of services.", "Número total de servicios.",
"Tourism / Travel", "Turismo / Viajes",
"Trace subsystems", "Trazar subsistemas",
"Traditional arts", "Artes tradicionales",
"Traditional music", "Música tradicional",
"Transcode", "Transcodificado",
"Transcode video only if source video codec matches. \"Any\" ignores source video codec checking and always transcodes. If no codec match is found, transcoding is done using the \"copy\" codec. if a match is found, transcode with the parameters in this profile. Separate codec names with comma.", "Transcodificar vídeo sólo si coincide con el códec de vídeo de origen. \"Cualquiera\" ignora la codificación del códec de vídeo de origen y siempre transcodifica. Si no se encuentra ninguna coincidencia de codec, la transcodificación se realiza mediante el códec \"copy\". Si se encuentra una coincidencia, transcodifique con los parámetros de este perfil. Separa los nombres de códec con coma.",
"Transcoding", "Transcodificar",
"Transmission mode", "Modo de transmisión",
"Transponder ID", "ID de Transpondedor",
"Transport Stream", "Transporte de Retrasmisión",
"Transport stream ID", "ID de retransmisión de transporte",
"Try to obtain channel numbers from the display-name xml tag. If the first word is number, it is used as the channel number.", "Tratar de obtener los números de canal de la etiqueta XML display-name. Si la primera palabra es el número, este es usado como el número de canal.",
"Tue", "Mar",
"Tune before DiseqC", "Sintonizar depués de DiseqC",
"Tuner", "Sintonizador",
"Tuner bind IP address", "Sintonización de IP local",
"Tuner configuration", "Configuración de sintonización",
"Tuner configuration.", "Configuración de sintonizador.",
"Tuning failed", "Sintonización fallida",
"Turn off LNB when idle", "Apagar LNB en espera",
"Tvheadend server name", "Nombre de servidor Tvheadend",
"Type", "Tipo",
"Type override", "Anular tipo",
"UDP RTP port number (2 ports)", "Nombre de puerto UDP RTP (2 puertos)",
"UHD: ultra high definition", "UHD: ultra alta definición",
"UPnP Protocol", "Protocolo UPnP",
"URL of the M3U playlist.", "URL para la lista de reproducción M3U.",
"URL with the SAT>IP server XML location", "URL con la locación XML del servidor SAT>IP",
"Unable to get system UTC time", "Incapaz de obtener la hora del servidor UTC",
"Unable to get system local time", "Incapaz de obtener la hora del sistema local",
"Unable to get system time", "Incapaz de obtener la hora del sistema",
"Unable to open file", "Incapaz de abrir el archivo",
"Unable to stat path", "No se puede obtener ruta",
"Uncommitted", "No confirmado",
"Uncommitted first", "No confirmado primero",
"Unicable I switch (universal LNB)", "Unicable I switch (LNB universal)",
"Unicable II switch (universal LNB)", "Unicable II switch (LNB universal)",
"Unicable type", "Tipo de un sólo cable",
"Unimportant", "Sin importancia",
"Universal LNB only", "LNB Universal únicamente",
"Universally unique identifier. Read only.", "Identificador único universal. Solo lectura.",
"Unknown", "Desconocido",
"Unknown file", "Archivo desconocido",
"Unknown reason (%i)", "Razón desconocida (%i)",
"Unlimited size", "Tamaño ilimitado",
"Unlimited time", "Tiempo ilimitado",
"Unset (default)", "Sin definir (por defecto)",
"Update Entitlement Management Messages from one mux only.", "Actualizar los mensajes de gestión de derechos de un solo MUX.",
"Update card (EMM)", "Actualizar tarjeta (EMM)",
"Update channel icon", "Actualizar icono del canal",
"Update channel name", "Actualizar nombre del canal",
"Update channel number", "Actualizar número del canal",
"Update disabled", "Actualizar deshabilitado",
"Update time", "Actualizar hora",
"Update tolerance (ms)", "Tolerancia de Actualización (ms)",
"Updated", "Actualizado",
"Updates from one mux (EMM)", "Actualizaciones desde un mux (EMM)",
"Upon start, Tvheadend doesn't know the last rotor position. This value defines the initial rotor movement. TVHeadend waits the specified time when the first movement is requested.", "En el arranque, Tvheadend no sabe la última posición del rotor. Este valor define el movimiento inicial del rotor. TVHeadend espera el tiempo especificado cuando se solicita el primer movimiento.",
"Usage: %s [OPTIONS]\n", "Uso: %s [OPTIONS]\n",
"Use AC-3 detection.", "Usar detección AC-3",
"Use EITp/f to decide event start/stop. This is also known as \"Accurate Recording\". See Help for details.", "Utilice EITp /f para decidir el inicio/fin del evento. Esto también se conoce como \"Grabación exacta\". Consulte la Ayuda para obtener más información.",
"Use EPG running state", "Usar estado de la guía",
"Use HTTP digest authentication", "Usar autenticación implícita HTTP",
"Use PROXY protocol & X-Forwarded-For", "Usar protocolo PROXY & X-Forwarded-For",
"Use WEBM format.", "Usar formato WEBM.",
"Use Windows-compatible filenames", "Usar nombres de archivo compatibles con Windows",
"Use XSPF playlist instead of M3U", "Usar lista XSPF en vez de M3U",
"Use a pre-defined list of ATSC-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Utilice una lista predefinida de muxes ATSC. Nota: estas listas pueden estar a veces obsoletas y pueden causar digitalización más tardía de lo habitual.",
"Use a pre-defined list of ATSC-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Utilice una lista predefinida de muxes ATSC-T. Nota: estas listas pueden estar a veces obsoletas y pueden causar una digitalización tardía.",
"Use a pre-defined list of DAB muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Utilice una lista predefinida de muxes DAB. Nota: estas listas pueden ser a veces obsoletas y puede causar digitalización más tardía de lo habitual.",
"Use a pre-defined list of DVB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Utilice una lista predefinida de muxes DVB-C. Nota: estas listas pueden estar a veces obsoletas y pueden causar una digitalización tardía.",
"Use a pre-defined list of DVB-S/S2 muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Utilizar una lista predefinida de DVB-S / S2 muxes. Nota: estas listas pueden ser a veces obsoletas y pueden causar una digitalización más tardía de lo habitual.",
"Use a pre-defined list of DVB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Utilice una lista predefinida de muxes DVB-T. Nota: estas listas pueden ser a veces obsoletas y puede causar digitalización más tardía de lo habitual.",
"Use a pre-defined list of ISDB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Utilice una lista predefinida de muxes ISDB-C. Nota: estas listas pueden ser a veces obsoletas y puede causar digitalización más tardía de lo habitual.",
"Use a pre-defined list of ISDB-S muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Utilice una lista predefinida de muxes ISDB-S. Nota: estas listas pueden ser a veces obsoletas y puede causar digitalización más tardía de lo habitual.",
"Use a pre-defined list of ISDB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Utilice una lista predefinida de muxes ISDB-T. Nota: estas listas pueden ser a veces obsoletas y puede causar digitalización más tardía de lo habitual.",
"Use only this service ID, filter out others.", "Usar solo esta ID de servicio, filtrar las demás.",
"Use original", "Usar original",
"Use packet backlog", "Uso de retraso de paquetes",
"Use service IDs as channel numbers", "Usar ID del servicio como números de canal",
"Use tag internally (don't expose to clients).", "Usar categoría internamente (no exponer a los clientes).",
"Use the provider's service IDs as channel numbers.", "Utilice las IDs de servicio asignadas por el proveedor como los números de canal.",
"Use this character set when setting filenames.", "Utilizar este carácter al asignar nombres de archivo.",
"Use x amount of storage space.", "Utilizar x cantidad de espacio de almacenamiento.",
"User access error", "Error de usuario",
"User does not have access", "El usuario no tiene acceso",
"User does not have access to channel", "El usuario no tiene acceso al canal",
"User icon", "Icono de usuario",
"User interface level", "Nivel de interfaz de usuario",
"User interface quick tips (tooltips)", "Consejos rápidos en la interfaz de usuario (de ayuda)",
"User language", "Idioma de usuario",
"User limit reached", "Límite de usuarios alcanzado",
"User login", "Login de usuario",
"User request", "Petición de Usuario",
"Username", "Usuario",
"Username for the entry (login username).", "Usuario para la entrada (usuario de inicio de sesión)",
"Username of the entry (this should match a username from within the \"Access Entries\" tab).", "Nombre de usuario de la entrada (este debe coincidir con un nombre de usuario de la pestaña \"Entradas de acceso\").",
"Users - Access Entries", "Usuarios - Entradas de Acceso",
"Users - IP Blocking", "Usuarios - Bloqueo de IP",
"Users - Passwords", "Usuarios - Contraseñas",
"Variety show", "Show variado",
"Verify the SSL certificate.", "Verificar certificado SLL.",
"Verify the peer's SSL.", "Verificar certificado SSL de peers.",
"Vertical resolution (height) of the output video stream. Horizontal resolution is adjusted automatically to preserve aspect ratio. When set to 0, the input resolution is used.", "Resolución vertical (altura) del stream de video de salida. La resolución horizontal se autoajustará para mantener la relación de aspecto. Si está fijada a 0, se utilizará la resolución de entrada.",
"Video Stream Filter", "Filtro de Retransmisión de Vídeo",
"Video bitrate (kb/s) (0=auto)", "Tasa de muestreo de vídeo (kb/s) (0=Auto)",
"Video codec", "Códec de Video",
"Video codec preset", "Codec de vídeo definido",
"Video codec preset to use for transcoding.", "Codec de vídeo definido para transcodificado.",
"Video codec to use for the transcode. \"Do not use\" will disable video output.", "Códec de vídeo a utilizar para la transcodificación. \"No usar\" desactivará la salida de vídeo.",
"Video recorder", "Grabador de vídeo",
"Video recorder flags, allow/disallow access to video recorder functionality (including Autorecs), allow/disallow users to view other DVR entries, allow/disallow users to work with DVR entries of other users (remove, edit) etc.", "Las banderas de la grabadora de video, permitir/no permitir el acceso a la funcionalidad del grabador de video (incluyendo Autograbaciones), permitir/no permitir a los usuarios ver otras entradas de DVR, permitir/no permitir a los usuarios trabajar con entradas de DVR de otros usuarios (eliminar, edutar, etc...)",
"View all", "Ver todo",
"Visually impaired commentary/audio description", "Discapacidad visual comentario/audiodescripción",
"Vorbis audio", "Audio Vorbis",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV perfil H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV perfil H264/AAC/Matroska ",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV perfil VP8/Vorbis/WEBM ",
"Waiting for EPG running flag", "Esperando por Guía funcionando con la característica",
"Waiting for program start", "Esperando el inicio del programa",
"Waiting for stream", "Esperando para difundir",
"Wake up the linked input whenever this adapter is used. The subscriptions are named as \"keep\". Note that this isn't normally needed, and is here simply as a workaround to driver bugs in certain dual tuner cards that otherwise lock the second tuner.", "Despertar la entrada unida siempre que se utilice este adaptador. Las suscripciones se denominan como \"mantener\". Tenga en cuenta que esto no es normalmente necesario, y simplemente se utiliza como una solución a los errores del conductor en ciertas tarjetas de sintonizador dual que de otro modo bloquearían el segundo sintonizador.",
"War", "Guerra",
"Water sport", "Deporte de agua",
"Weak stream", "Retransmisión débil",
"Weather report", "El Tiempo",
"Web User Interface", "Interfaz Web de Usuario",
"Web interface", "Interfaz Web",
"Web interface language", "Lenguaje de la interfaz web",
"Web interface language.", "Lenguaje de la interfaz web.",
"Web interface theme.", "Tema de la interfaz web.",
"Web theme", "Tema Web",
"Web user interface", "Interfaz web de usuario",
"Wed", "Mié",
"Welcome - Tvheadend - your TV streaming server and video recorder", "Bienvenido - Tvheadend - tu servidor de televisión y grabación de vídeos",
"Western", "Oeste",
"Western hemisphere (latitude direction)", "Hemisferio occidental (dirección latitud)",
"Western hemisphere (latitude direction).", "Hemisferio occidental (dirección latitud).",
"When nothing else is happening Tvheadend will continuously rotate among all muxes and tune to them to verify that they are still working when the inputs are not used for streaming. If your adapters have problems with lots of (endless) tuning, disable this. Note that this option should be OFF for the normal operation. This type of mux probing is not required and it may cause issues for SAT>IP (limited number of PID filters).", "Cuando nada está sucediendo Tvheadend girará de forma continua entre todos los multiplexores y los sintonizará para verificar que todavía están trabajando cuando las entradas no se utilizan para la transmisión. Si los adaptadores tienen problemas con un montón (sin fin) de sintonizaciones desactive esto. Tenga en cuenta que esta opción debe estar en OFF para el funcionamiento normal. Este tipo de mux de sondeo no es necesario y puede causar problemas para el SAT> IP (número limitado de filtros PID).",
"When specified, this setting overrides the subdirectory rules (except the base directory) defined in the DVR configuration and puts all recordings done by this entry into the subdirectory named here. See Help for more info.", "Cuando se especifica, este parámetro tiene prioridad sobre las reglas de subdirectorios (excepto el directorio base) definidos en la configuración DVR y pone todas las grabaciones realizadas por esta entrada en el subdirectorio llamado aquí. Consulte la Ayuda para obtener más información.",
"When streaming a service (via http or htsp) Tvheadend will use the network with the highest streaming priority set here. See Help for details.", "Cuando se retransmite un servicio (via http o https) Tvheadend usará la red con mayor prioridad seleccionada aquí. Mire la Ayuda para más detalles.",
"When the fulltext is checked, the title pattern is matched against title, subtitle, summary and description.", "Cuando se marca el texto completo, el patrón del título es comparado con el título, subtítulo, resumen y descripción.",
"When the service was first identified and recorded.", "Cuando el servicio se identificó y registró por primera vez",
"When the service was last seen during a mux scan.", "Cuando el servicio fue visto por última vez durante una exploración mux.",
"Winter sports", "Deportes de invierno",
"With \"RAM only\" enabled, and when \"Maximum RAM size\" is reached, remove the oldest segment in the buffer instead of replacing it completely. Note, this may reduce the amount of rewind time.", "Con \"Sólo memoria RAM\" habilitada y cuando se alcanza \"Tamaño máximo RAM\", elimine el segmento más antiguo en el búfer en lugar de reemplazarlo por completo. Tenga en cuenta que esto puede reducir la cantidad de tiempo de rebobinado.",
"Wizard", "Asistente",
"Wrapper (capmt_ca.so)", "Envoltura (capmt_ca.so)",
"Write a short message to log identifying the matched parameters. It is useful for debugging your setup or structure of incoming streams.", "Escribe un mensaje corto para registrar la identificación de los parámetros emparejados. Es útil para depurar la configuración o la estructura de las retrasmisiones entrantes.",
"Writes the current in-memory EPG database to disk every x hours, so should a crash/unexpected shutdown occur EPG data is saved periodically to the database (re-read on next startup). Set to 0 to disable.", "Escribe la base de datos EPG actual en la memoria en el disco cada x horas, así podría ocurrir un accidente / apagado inesperado y los datos EPG se guardarían periódicamente en la base de datos (re-leído en el siguiente inicio). Se establece en 0 para desactivar.",
"Writes the current in-memory EPG database to disk shortly after an xmltv import has completed, so should a crash/unexpected shutdown occur EPG data is saved (re-read on next startup).", "Escribir la base de datos de EPG actual en memoria poco tiempo depués de que la importación xmltv se haya completado, por lo que un crasheo/apagado inesperado podría ocurrir y el EPG se encontraría guardado (re-leido al inicio).",
"XMLTV EPG Import", "Importación de EPG XMLTV ",
"Yes", "Sí",
"addpids/delpids supported", "addpids/delpids soportado",
"default: nvenc(h264 / h265)", "Por defecto: nvenc(h264 / h265)",
"fast: h264 / h265 / qsv(h264 / h265)", "Rápido: h264 / h265 / qsv(h264 / h265)",
"faster: h264 / h265 / qsv(h264)", "Más rápido: h264 / h265 / qsv(h264)",
"hrs", "horas",
"invalid option specified [%s]", "opcion inválida especificada [%s]",
"logout", "salir",
"medium: h264 / h265 / qsv(h264 / h265)", "Medio: h264 / h265 / qsv(h264 / h265)",
"min", "min.",
"mins", "mins.",
"option %s requires a value", "la opción %s requiere un valor",
"placebo: h264 / h265", "Placebo: h264 / h265",
"return", "regresar",
"slow: h264 / h265 / qsv(h264 / h265)", "Lento: h264 / h265 / qsv(h264 / h265)",
"slower: h264 / h265 / qsv(h264)", "Más lento: h264 / h265 / qsv(h264)",
"superfast: h264 / h265", "Super Rápido: h264 / h265",
"tsfile input (mux file)", "entrada tsfile (archivo mux)",
"tvhdhomerun client", "Cliente tvhdhomerun",
"ultrafast: h264 / h265", "Ultrá rápido: h264 / h265",
"veryfast: h264 / h265 / qsv(h264)", "Muy rápido: h264 / h265 / qsv(h264)",
"veryslow: h264 / h265 / qsv(h264)", "Muy lento: h264 / h265 / qsv(h264)",
NULL, NULL
};

static const char *tvh_locale_12[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nLisainformatsiooniks külastage Tvheadend'i veebilehte:\nhttps://tvheadend.org\n",
"# Mapped channels", "# Seostatud kanalid",
"# Muxes", "# Muxid",
"# Services", "# Teenused",
"(Default profile)", "(Vaikimisi profiil)",
"(Read-only)", "(Mittemuudetav)",
"1 day", "1 päev",
"1 month", "1 kuu",
"1 week", "1 nädal",
"1 year", "1 aasta",
"2 months", "2 kuud",
"2 weeks", "2 nädalat",
"2 years", "2 aastat",
"2-Port switch (universal LNB)", "2 pordine switch (universaalne LNB)",
"3 days", "3 päeva",
"3 months", "3 kuud",
"3 weeks", "3 nädalat",
"3 years", "3 aastat",
"4-Port switch (universal LNB)", "4 pordine switch (universaalne LNB)",
"5 days", "5 päeva",
"6 months", "6 kuud",
"ACTIVE", "AKTIIVNE",
"ANY", "KÕIK SOBIB",
"ATSC SI Tables", "ATSC SI tabelid",
"ATSC source ID", "ATSC lähteallika ID",
"ATSC-C Network", "ATSC-C võrk",
"ATSC-T Network", "ATSC-T võrk",
"AUTO", "AUTOMAATNE",
"Aborted by user", "Katkestatud kasutaja poolt",
"Accept remote subscription weight", "Luba kauge liitumise kaalukus",
"Accept zero value for TSID", "Nõustu TSID null väärtusega",
"Access", "Ligipääs",
"Access (ACL)", "Ligipääs (ACL)",
"Action", "Põnevus",
"Adapter in use by another subscription", "Adapter on hõivatud mõne muu liitumisega",
"Add file and line numbers to debug", "Lisa fail ja rea number debugimiseks",
"Add the thread ID to debug", "Lisa threedi ID debugimiseks",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "Lisaaeg (sekundites), mille jooksul seatakse tüüner salvestamiseks valmis. See on kasulik selliste tüünerite puhul, millel võtab reageerimine teatud aja ja mis kipuvad sellepärast salvestuse alguses ruudutavat läbu salvestama.",
"Admin password", "Administraatori salasõna",
"Admin username", "Administraatori kasutajanimi",
"Administrator login", "Administraatori sisselogimine",
"Adult movie", "Täiskasvanute film",
"Advanced", "Täppis",
"Adventure", "Seiklus",
"Advertisement / Shopping", "Reklaam / Ostlemine",
"After creating the network scan it for services.", "Pärast võrgu loomist skaneeri teenuseid.",
"All (Streaming plus DVR)", "Kõik (voogedastus + DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Kõik sümbolid, mis võiksid tekitada probleemseid failinimesid, asendatakse allkriipsuga. Vaata täpsemat infot abidokumentatsioonist.",
"All lower-case", "Kõik väiketähed",
"All proceeds are used to support server infrastructure and buy test equipment.", "Kõik tulud kasutatakse serverite taristu ülalpidamiseks ja testseadmete hankimiseks.",
"All: Record if different description", "Kõik: salvesta, kui kirjeldus on erinev",
"All: Record if different episode number", "Kõik: salvesta, kui osa number on erinev",
"All: Record if different subtitle", "Kõik: Salvesta, kui osa nimi on erinev",
"All: Record once per day", "Kõik: salvesta üks kord päevas",
"All: Record once per month", "Kõik: salvesta üks kord kuus",
"All: Record once per week", "Kõik: salvesta üks kord nädalas",
"Allow high bitrate mode (CI+ CAMs only).", "Luba kõrge bitikiiruse režiim (ainult CI+ CAM'id)",
"Allow/Disallow access to the 'Configuration' tab.", "Luba/keela ligipääs \"Seaded\" vahelehele.",
"Allowed network", "Lubatud võrk",
"Allowed networks", "Lubatud võrgud",
"Alternate PID path", "Alternatiivne PID asukohaviide",
"Alternate configuration path", "Alternatiivne seadistuste asukohaviide",
"Altitude (in meters).", "Kõrgus merepinnast (meetrites)",
"Altitude (meters)", "Kõrgus merepinnast (meetrites)",
"Always use the name defined by the network.", "Alati kasuta nime, mis on võrgu poolt määratud.",
"Animals", "Loomad",
"Any", "Kõik sobib",
"Arts", "Kunstid",
"Arts magazines", "Kunstiülevaated",
"Assign predefined muxes to networks", "Määra eeldefineeritud muxid võrkudeks",
"Associate this device with one or more networks.", "Liidesta see seade ühe või rohkema võrguga.",
"Athletics", "Kergejõustik",
"Audio Stream Filter", "Audio voogedastuse filter",
"Audio bitrate (kb/s) (0=auto)", "Audio bitikiirus (kb/s) (0=automaatne)",
"Audio codec", "Audio koodek",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Audio koodek konverteerimiseks. \"Ära kasuta\" lülitab audio väljundi välja.",
"Audio stream", "Audio voog",
"Audio type", "Audio tüüp",
"Audio-only stream", "Ainult audio voog",
"Auto", "Automaatne",
"Auto check disabled", "Automaatne kontroll välja lülitatud",
"Auto check enabled", "Automaatne kontroll sisse lülitatud",
"Auto record", "Automaatselt salvesta",
"Auto record caption", "Automaatse salvestuse tekst",
"Auto recording%s%s", "Automaatne salvestamine%s%s",
"Auto time record", "Automaatse kellaajaga salvestus",
"Auto-Map to channels", "Automaatseosta kanaliteks",
"Automatic checking", "Automaatne kontollimine",
"Automatic recording caption.", "Automaatse salvestuse tekst.",
"Automatic schedule entry not found", "Automaatse ajastuse kirjet ei leitud ",
"Automatic time scheduler entry not found", "Automaatse aja ajastuse kirjet ei leitud ",
"Automatically map EPG source", "Automaatselt seosta EPG allikas",
"Automatically map all available services to channels.", "Automaatselt seosta kõik saadaolevad teenused kanaliteks",
"Automatically map channels defined within the bouquet.", "Automaatselt seosta kanalid, mis kimbu sees on defineeritud.",
"Automatically name from network", "Nimeta automaatselt võrgust",
"Automatically record.", "Automaatselt salvesta.",
"Autorec maximum count (0=unlimited)", "Automaatsalvestuse maksimum arv (0=piiranguteta)",
"Autorec maximum schedules limit (0=unlimited)", "Automaatsalvestuse ajastamiste maksimaalne arv (0=piiranguteta) ",
"Bad", "Halb",
"Bad request", "Halb päring",
"Ballet", "Ballett",
"Bandwidth", "Võrguühendus",
"Base DVB SI Tables (PAT,CAT,PMT,SDT etc.)", "Põhilised DVB SI tabelid (PAT,CAT,PMT,SDT jm)",
"Based on software from", "Kasutajaliidese loomise tarkvara:",
"Basic", "Tavaline",
"Bind to specific local IP address.", "Seostu konkreetse lokaalse IP aadressiga.",
"Blue", "Sinine",
"Bouquet", "Kanalite kimp",
"Bouquet (auto)", "Kanalite kimp (automaatne)",
"Bouquet source.", "Kanalite kimbu algallikas.",
"Bouquets", "Kanalite kimbud.",
"Brand", "Bränd",
"Broadcast", "Levi",
"Broadcast type", "Levi tüüp",
"Broadcast.", "Levi.",
"Broadcasting", "Levitamine",
"Buffering limit (ms)", "Puhverdamise limiit (ms)",
"Build", "Versioon",
"CA (descrambling) Client", "CA (lahtikrüpteerimise) klient",
"CA Stream Filter", "CA voogedastuse filter",
"CA client %i", "CA klient %i",
"CA identification", "CA tuvastamine",
"CA provider", "CA pakkuja",
"CAPMT CA Client", "CAPMT CA klient",
"CAPMT interval (in ms).", "CAPMT intervall (ms).",
"CAPMT interval (ms)", "CAPMT intervall (ms)",
"CAPMT query interval (ms)", "CAPMT päringu intervall (ms)",
"CAPMT query interval (ms).", "CAPMT päringu intervall (ms).",
"CI Module", "CI moodul",
"CPU", "Protsessor",
"CRASH", "KOKKUKUKKUMINE",
"CSA (descrambling)", "CSA (lahtikrüpteerimine)",
"CWC CA Client", "CWC CA klient",
"Cache scheme", "Puhvri mall",
"Cartoons", "Multikad",
"Change parameters", "Muuda valikuid",
"Channel", "Kanal",
"Channel Tags", "Kanali märked",
"Channel does not exist", "Kanal ei eksisteeri",
"Channel icon", "Kanali ikoon",
"Channel icon URL.", "Kanali ikooni URL.",
"Channel icon as defined in EPG data.", "Kanali ikoon nagu elektroonilises telekavas defineeritud.",
"Channel icon name scheme", "Kanali ikooni nime mall",
"Channel icon path", "Asukohaviide kanali ikoonile",
"Channel mapping options", "Kanali seostamise valikud",
"Channel name", "Kanali nimi",
"Channel not found", "Kanalit ei leitud",
"Channel number", "Kanali number",
"Channel number as defined in EPG data.", "Kanali number nagu elektroonilises telekavas defineeritud.",
"Channel number offset", "Kanali numbrite tasakaalustus",
"Channel number range", "Kanalite nummerdamine",
"Channel numbers (heuristic)", "Kanali numbrid (heuristiline)",
"Channel numbers from", "Kanali numbrid alates",
"Channel tag", "Kanali märge",
"Channel tag does not exist", "Kanali märge ei eksisteeri",
"Channel tag reference", "Kanali märke viide",
"Channel tags", "Kanali märked",
"Channel update options", "Kanali uuendamise valikud",
"Channels", "Kanalid",
"Character set", "Tähestik kooditabeli alusel",
"Charset", "Kooditabel",
"Check availability", "Kontrolli saadavust",
"Children's / Youth Programs", "Laste- või noortesaated",
"Children's / Youth programs", "Laste- või noortesaated",
"Cinema", "Kino",
"Class", "Klass",
"Classical", "Klassikaline",
"Classical music", "Klassikaline muusika",
"Clean effects", "Puhasta efektid",
"Client name", "Kliendi nimi",
"Clone scheduled entry on error", "Klooni ajastatud kirje vea juhtumisel ",
"Comedy", "Komöödia",
"Command delay time (ms) (10-200)", "Käskluse viite aeg (ms) (10-200)",
"Command time (ms) (10-100)", "Käskluse aeg (ms) (10-100)",
"Comment", "Kommentaar",
"Commercial break", "Kommertsreklaam",
"Committed", "Rakendatud",
"Completed OK", "Lõpetanud OK",
"Compress EPG database", "Paki kokku EPG andmebaas",
"Compress the EPG database to reduce disk I/O and space.", "Paki kokku EGP andmebaas, et vähendada kõvaketta poole pöördumist ja ruumikasutust.",
"Configuration", "Seaded",
"Configuration - Base", "Konfiguratsioon - Alus",
"Configuration - Image Cache", "Konfiguratsioon - Pildipuhver",
"Configuration - SAT>IP Server", "Konfiguratsioon - SAT>IP server",
"Configuration ID", "Seadete ID",
"Configuration name", "Konfiguratsiooni nimi",
"Configuration version", "Seadete versioon",
"Connection limit type", "Ühenduste piirangu tüüp",
"Connection limits", "Ühenduse limiidid",
"Container", "Konteriner",
"Container to use for the transcoded stream.", "Konteinerformaat skaleeritud voogedastuse jaoks.",
"Content character set", "Sisu tähestik kooditabeli alusel",
"Content type", "Sisu tüüp",
"Content type.", "Sisu tüüp.",
"Contest", "Võistlus",
"Continue recording for x minutes after scheduled stop time", "Jätka salvestamist x minutit peale ajastatud lõpuaega",
"Continue recording for x minutes after scheduled stop time.", "Jätka salvestamist x minutit peale ajastatud lõpuaega.",
"Cookie expiration (days)", "Veebiküpsiste aegumine (päevades)",
"Cooking", "Toiduvalmistamine",
"Copy codec type", "Koopia koodeki tüüp",
"Count of objects", "Objektide loend",
"Create SDTV/HDTV/Radio tags.", "Loo SDTV/HDTV/Raadio märkeid.",
"Create a bouquet from the playlist.", "Loo pleilistist kanalite kimp. ",
"Create a directory per channel when storing recordings. If both this and the 'directory per day' checkbox is enabled, the date-directory will be the parent of the per-channel directory.", "Loo salvestuste hoidmiseks igale kanalile oma eraldi kataloog. Kui nii see kui ka 'kataloog iga päeva kohta' valik on ära märgitud, siis kuupäevapõhine kataloog on peakataloog ja kanali oma alamkataloog on seal all.",
"Create a directory per title when storing recordings. If the day/channel directory checkboxes are also enabled, those directories will be parents of this directory.", "Loo salvestuste hoidmiseks pealkirjapõhine kataloog. Kui päevapõhine või kanalipõhine valik on samuti ära märgitud, siis need kataloogid saavad olema peakataloogid, mille all pealkirjapõhine kataloog saab olema alamkataloog.",
"Create a new directory per day in the recording system path. Folders will only be created when something is recorded. The format of the directory will be ISO standard YYYY-MM-DD.", "Loo salvestuste hoidmiseks uus kataloog iga päeva kohta. Kataloogid luuakse ainult siis, kui midagi on salvestatud. Kataloogide nimed tulevad ISO standardi alusel kujul AAAA-KK-PP  ehk siis aasta-kuu-päev. ",
"Create a provider name tag.", "Loo teenusepakkuja nime märge.",
"Create bouquet", "Loo kanalite kimp",
"Create bouquet tag", "Loo kanalite kimbu märge",
"Create directories using these permissions.", "Loo kataloogid kasutades neid lube.",
"Create files using these permissions.", "Loo failid kasutades neid lube.",
"Create network name tags", "Loo võrgu nime märkeid",
"Create network name tags (set by provider).", "Loo võrgu nime märkeid (määratud teenusepakkuja poolt).",
"Create network tags", "Loo uusi võrgu märkeid",
"Create provider name tags", "Loo teenusepakkujate nimede märkeid",
"Create provider tags", "Loo uusi teenusepakkuja märkeid",
"Create tags", "Loo märkeid",
"Create type-based tags", "Loo tüübile põhinevaid märkeid",
"Created", "Loodud",
"Creator", "Autor",
"Cron multi-line", "Ajastamine cron'iga",
"Culture (without music)", "Kultuur (ilma muusikata)",
"Culture magazines", "Kultuuriülevaated",
"Current RTSP port.", "Praegune RTSP port.",
"Current affairs", "Hetkel juhtumas laias maailmas",
"Current number of objects.", "Objektide praegune koguarv.",
"Current object size.", "Praeguse objekti suurus.",
"Current tuner configuration.", "Praegused tüüneri seaded",
"Custom HTTP headers", "Omaloodud HTTP päised",
"DBus - use the session message bus instead of the system one", "DBus - kasuta sessioonipõhist teavitust eelistades seda süsteemsele teavitusele",
"DES Key.", "DES võti.",
"DES key", "DES võti",
"DSCP/TOS for streaming", "DSCP/TOS striimimiseks",
"DVB CAM Client", "DVB CAM klient",
"DVB CSA (descrambling) Tables", "DVB CSA (lahtikrüpteerimise) tabelid",
"DVB EPG Tables", "DVB EPG tabelid",
"DVB Inputs - Networks", "DVB sisendid - võrgud",
"DVB SI Tables", "DVB SI tabelid",
"DVB Time Tables", "DVB ajatabelid",
"DVB scan files", "DVB failide skaneerimine",
"DVB scan files path", "DVB failide skaneerimise asukohaviide",
"DVB-C Network", "DVB-C võrk",
"DVB-S Network", "DVB-S võrk",
"DVB-S Satellite Configuration", "DVB-S satelliidi seaded",
"DVB-T Network", "DVB-T võrk",
"DVR - Auto-recording (Autorecs)", "DVR - automaatne salvestus (automaatsalvestuse kirjed)",
"DVR - Profiles", "DVR - profiilid",
"DVR behavior", "DVR käitumine",
"DVR configuration", "DVRi seaded",
"DVR configuration profiles", "DVRi seadete profiilid",
"DVR configurations", "Digitaalse videosalvestaja konfiguratsioon",
"DVR entry not found", "DVR kirjet ei leitud",
"DVR file retention period", "DVRi faili alleshoidmise aeg",
"DVR log retention", "DVRi logi alleshoidmise aeg",
"DVR log retention period", "DVRi logi alleshoidmise aeg",
"DVR override: high", "DVR override: kõrge",
"DVR override: important", "DVR override: tähtis",
"DVR override: low", "DVR override: madal",
"DVR override: normal", "DVR override: tavaline",
"DVR override: unimportant", "DVR override: pole tähtis",
"DVR profile to use/used for the recording.", "DVP profiil salvestamise jaoks.",
"DVR schedule does not exist", "DVRi ajakava ei eksisteeri",
"DVR schedule not found", "DVR ajakava ei leitud",
"Dance", "Tants",
"Data errors", "Data vead",
"Data path (if applicable).", "Data asukohaviide (kui on).",
"Days of Week", "Nädalapäevad",
"Days of week", "Nädalapäevad",
"Debate", "Debatt",
"Debug libav log", "Debugimise log libav kohta",
"Debug options", "Debugimise valikud",
"Debug subsystems", "Debugi alamsüsteeme",
"Debug to syslog", "Debugi syslog'i",
"Debug trace (low-level)", "Debugi treiss (low-level)",
"Debugging", "Debugimine",
"Default", "Vaikimisi",
"Default language(s)", "Vaikimisi keel(ed)",
"Default language.", "Vaikimisi keel.",
"Default login", "Vaikimisi sisselogimine",
"Default priority", "Vaikimisi prioriteet",
"Default user interface level.", "Vaikimisi kasutaja kasutajaliidese tase.",
"Define network group to limit network usage.", "Defineeri võrgu grupp piiramaks võrgukasutust.",
"Delivery system", "Kohaleviimise süsteem",
"Demux path", "Demuxi asukohaviide",
"Descramble services (limit per mux)", "Lahtikodeerimisteenused (piirang per mux)",
"Descrambler", "Lahtikodeerija",
"Descrambler EMM", "Lahtikodeerija EMM",
"Descrambler buffer (TS packets)", "Lahtikodeerija puhver (TS paketid)",
"Description", "Kirjeldus",
"Detective", "Detektiiv",
"Device ID", "Seadme ID",
"Device model", "Seadme mudel",
"Device path", "Seadme asukohaviide",
"Differentiated Services Code Point / Type of Service: Set the service class Tvheadend sends with each packet. Depending on the option selected this tells your router the prority in which to give packets sent from Tvheadend, this option does not usually need changing. See https://en.wikipedia.org/wiki/Differentiated_services for more information. ", "Differentiated Services Code Point / Type of Service: märgi paketid vastavate protokollide märgistustega. Oma olemuselt on see viis QoSi tegemiseks, kus paketid vastavalt oma märgistustele saavad eelistatud sihtkohta jõudmisel. Kui Sa ei tea, millest on jutt, siis lase sel olla, Tvheadend töötab ilma selleta ka suurepäraselt, kui võrgukiirused on piisavad :-P Aga huvi korral võid uurida näiteks linki https://en.wikipedia.org/wiki/Differentiated_services",
"Digital Video Recorder", "Digitaalne videosalvestaja",
"Directory", "Kataloog",
"Directory permissions (octal, e.g. 0775)", "Kataloogi loabitid (numbritena, nt 0775)",
"Directory used by the entry.", "Kirje poolt kasutatav kataloog.",
"Disable", "Ära kasuta",
"Disable DVB bouquets", "Lülita välja DVB kanalite kimbud",
"Disable SAT>IP client", "Lülita välja SAT>IP klient",
"Disable X_SATIPM3U tag", "Lülita välja X_SATIPM3U märge",
"Disable all access control checks", "Lülita välja kõik ligipääsu kontrollid",
"Disable debug on stderr", "Lülita välja debugimine standard veavoogu",
"Disable device/firmware-specific workarounds", "Lülita välja seadme/firmware spetsiifilised workaround'id",
"Disable syslog (all messages)", "Lülita välja syslog (kõik teated)",
"Disabled", "Väljalülitatud",
"Discussion", "Vestlus",
"DiseqC repeats", "DiseqC kordused",
"Do not use", "Ära kasuta",
"Documentary", "Dokumentaal",
"Don't backup configuration tree at upgrade", "Ära tee tarkvara uuendamisel kõikidest seadetest tagavarakoopiat  ",
"Don't include the title in the filename.", "Ära kaasa pealkirja faili nimes.",
"Don't include title in filename", "Ära kaasa pealkirja faili nimes",
"Don't keep", "Ära jäta alles",
"Don't re-record", "Ära taassalvesta",
"Don't re-record if recording fails.", "Ära taassalvesta kui salvestus ebaõnnestub.",
"Don't re-schedule if recording fails.", "Ära taasajasta kui salvestus ebaõnnestub.",
"Don't reschedule", "Ära taasajasta",
"Don't use the provider's channel numbers.", "Ära kasuta teenusepakkuja kanalite numeratsiooni.",
"Drama", "Draama",
"Duplicate handling", "Duplikaatide käsitlemine",
"Duplicate recording handling.", "Duplikaatsalvestuste käsitlemine.",
"EIT - skip TSID check", "EIT - jäta vahele TSID kontroll ",
"EIT time offset", "EIT ajanihe",
"EPG Database", "EPG andmebaas",
"EPG Grabber", "EPG hankija",
"EPG Grabber Channel", "EPG hankija kanal",
"EPG Grabber Configuration", "EPG hankija seaded",
"EPG Language (priority order)", "EPG keel (prioriteedijärjekord)",
"EPG name", "EPG nimi",
"EPG scan", "EPG skaneerimine",
"EPG source", "EPG allikas",
"EPG update window", "EPG uuendamise ajaraam",
"Economics", "Majandus",
"Education", "Haridus",
"Educational", "Hariduslik",
"Electronic Program Guide", "Elektrooniline saatekava",
"Elementary Stream Filter", "Elementary Stream filter",
"Elementary stream filter", "Voogedastuse elementaarne filter",
"Enable", "Jah",
"Enable (auto)", "Jah (automaatne)",
"Enable DBus", "Lülita sisse DBus",
"Enable NTP driver", "Lülita sisse NTP draiver",
"Enable coredumps for daemon", "Lülita sisse deemoni coredump funktsionaalsus",
"Enable debug on stderr", "Lülita sisse debugimine stdrerr'i",
"Enable debug subsystems", "Lülita sisse alamsüsteemide debugimine",
"Enable debug to file", "Lülita sisse debugimine faili",
"Enable debug to syslog", "Lülita sisse debugimine syslog'i",
"Enable if the SAT>IP box requests plts=on parameter in the SETUP RTSP command for DVB-S2 muxes.", "Lülita sisse, kui SAT>IP karp vajab SETUP RTSP käsus DVB-S2 muxide jaoks parameetrit plts=on.",
"Enable if the SAT>IP box requires pids=21 parameter in the SETUP RTSP command", "Lülita sisse, kui SAT>IP karp vajab SETUP RTSP käsus parameetrit pids=21.",
"Enable if the SAT>IP box supports the addpids/delpids commands.", "Lülita sisse, kui SAT>IP karp toetab addpids/delpids käske.",
"Enable or disable fast input switching.", "Lülita sisse või välja 'fast input switching'.",
"Enable or disable full mux mode.", "Lülita sisse või välja full mux režiim.",
"Enable or disable ignoring of Event Information Table (EIT) data on this mux.", "Lülita sisse või välja EIT (Event Information Table) infoga arvestamine selles muxis.",
"Enable or disable this configuration.", "Lülita sisse või välja need seaded.",
"Enable syslog", "Lülita sisse syslog",
"Enable system time updates. This will only work if the user running Tvheadend has rights to update the system clock (normally only root).", "Lülita sisse süsteemi kellaaja uuendamine. See töötab ainult juhul, kui Tvheadendi teenuse süsteemsel kasutajanimel on õigused süsteemi kellaaja uuendamiseks (tavaliselt ainult root kasutajanimi)",
"Enable this filter.", "Luba see filter.",
"Enable trace subsystems", "Lülita sisse alamsüsteemide trace",
"Enable web UI debug (non-minified JS)", "Lülita sisse veebiliidese debugimine (mitteminimeeritud javascript)",
"Enable/Disable CA client.", "Lülita CA klient sisse/välja.",
"Enable/Disable service.", "Lülita teenus sisse/välja.",
"Enable/Disable the entry.", "Lülita kirje sisse/välja.",
"Enable/Disable timeshift.", "Lülita ajanihe sisse/välja.",
"Enable/disable LNA.", "Lülita LNA sisse/välja.",
"Enable/disable auto-rec rule.", "Lülita automaatsalvestuse reegel sisse/välja.",
"Enable/disable debugging output to syslog.", "Lülita sisse/välja debugimine syslog'i.",
"Enable/disable interface quick tips.", "Lülita kasutajaliidese vihjed sisse/välja.",
"Enable/disable libav log output.", "Lülita sisse/välja libav logi väljund.",
"Enable/disable logging to syslog.", "Lülita sisse/välja debugimine syslog'i.",
"Enable/disable offering of Entitlement Management Message updates.", "Entitlement Management Message uuenduste sisse/välja lülitamine.",
"Enable/disable profile.", "Lülita profiil sisse/välja.",
"Enable/disable the bouquet.", "Lülita kanalikimp sisse/välja.",
"Enable/disable the channel.", "Lülita kanal sisse/välja.",
"Enable/disable the device.", "Lülita seade sisse/välja.",
"Enable/disable the entry.", "Lülita kirje sisse/välja.",
"Enable/disable the grabber.", "Lülita hankija sisse/välja.",
"Enable/disable the tag.", "Lülita märge sisse/välja.",
"Enable/disable tuner/adapter.", "Lülita tüüner/adapter sisse/välja.",
"Enabled", "Kasutamine",
"Encrypted", "Krüpteeritud",
"Encrypted service", "Krüpteeritud teenus",
"Enter a non-admin user password.", "Sisesta mitteadministraatorist kasutaja salasõna.",
"Enter a non-admin user username.", "Sisesta mitteadministraatorist kasutaja kasutajanimi.",
"Enter an administrator password.", "Sisesta administraatori salasõna.",
"Entertainment programs for 10 to 16", "Meelelahutussaated vanusele 10 kuni 16",
"Entertainment programs for 6 to 14", "Meelelahutussaated vanusele 6 kuni 14",
"Environment", "Keskkond",
"Environment (pipe)", "Keskkonnamuutujad (pipe)",
"Episode", "Osa",
"Episode %d", "Osa %d",
"Episode number/ID.", "Episoodi number/ID.",
"Error code", "Veakood",
"Error code of entry.", "Kirje veakood.",
"Errors", "Veateateid",
"Even key", "Paaris võti",
"Even key.", "Paaris võti",
"Event does not exist", "Juhtum ei eksisteeri",
"Every day", "Iga päev",
"Exclude channel tags", "Välista kanali märked",
"Expeditions", "Ekspeditsioonid",
"Experimental film", "Eksperimentaalfilm",
"Expert", "Ekspert",
"Exported tuners", "Eksporditud tuunerid",
"External", "Väline",
"External EPG grabber", "Väline EPG hankija",
"External IP (NAT)", "Väline IP aadress (NAT)",
"External PyEPG Grabber", "Väline PyEPG hankija",
"External URL", "Väline URL",
"External URL of the bouquet.", "Kanalite kimbu väline URL.",
"External XMLTV EPG Grabber", "Väline XMLTV EPG hankija",
"Extra arguments", "Lisaargumendid",
"Extra start time", "Alustamise ajale lisaks",
"Extra stop time", "Lõpetamise ajale lisaks",
"Extra warming up time (seconds)", "Lisaaeg ülessoojenemiseks (sekundites)",
"FAIL", "EBAÕNNESTUMINE",
"FEC high", "FEC kõrge",
"FEC low", "FEC madal",
"Factual topics", "Faktidel põhinevad teemad",
"Failed", "Ebaõnnestunud",
"Failed to open image", "Pildi avamine ebaõnnestus",
"Failed view", "Ebaõnnestunud salvestuste vaade",
"Fantasy", "Fantaasia",
"Fashion", "Mood",
"Fast input switch", "Sisestamise kiirlülitus",
"Fastscan DVB", "DVB kiirskaneerimine",
"File missing", "Fail puudub",
"File not created", "Faili ei loodud",
"File permissions (octal, e.g. 0664)", "Faili loabitid (numbrites, nt 0664)",
"File removed", "Fail eemaldatud",
"File size", "Faili suurus",
"Filename", "Faili nimi",
"Filename character set", "Faili nime tähestik kooditabeli alusel",
"Filename options", "Faili nime valikud",
"Filename used by the entry.", "Kirje poolt kasutatav faili nimi.",
"Filesystem monitor", "Failisüsteemi monitor",
"Fine arts", "Ilusad kunstid",
"Finished", "Lõppenud",
"First word", "Esimene sõna",
"Fit to RAM (cut rewind)", "Mahuta arvutimällu (lõika maha tagasikerimine)",
"Fitness and health", "Tervis",
"Folkloric", "Folk",
"Football", "Jalgpall",
"Force (auto)", "Sunni peale (automaatne)",
"Force CA ID (e.g. 0x2600)", "Sunni peale CA ID (nt 0x2600)",
"Force old status", "Sunni peale vana olek",
"Force pilot for DVB-S2", "Sunni peale pilootsignaal DVB-S2 jaoks",
"Force priority", "Sunni peale prioriteedi väärtuseks",
"Force service type to 1", "Sunni teenuse tüübiks 1",
"Force teardown delay", "Sunni peale mahavõtmise viide",
"Force usage of entered CA ID on this service.", "Sunni peale sellele teenusele ise sisestatud CA ID.",
"Foreign countries", "Välismaad",
"Forever", "Igavesti",
"Fork and run as daemon", "Jagune ja tööta deemonina",
"Format string", "Format String",
"Found muxes", "Leitud muxe",
"Found services", "Leitud teenuseid",
"Free subscription weight", "Vaba liitumise kaalukus",
"Free-form field, enter whatever you like here.", "Formaadivaba väli, sisesta siia mida aga ise soovid.",
"Free-form text field, enter whatever you like here.", "Formaadivaba tekstiväli, sisesta siia mida aga ise soovid.",
"Free-form text field, enter whatever you like.", "Formaadivaba tekstiväli, sisesta mida aga ise soovid.",
"Free-form text field. You can enter whatever you like here.", "Formaadivaba tekstiväli. Sisesta siia mida aga ise soovid.",
"Free-format text field. Enter whatever you like here.", "Formaadivaba tekstiväli. Sisesta mida aga ise soovid.",
"Frequency", "Sagedus",
"Frequency (Hz)", "Sagedus (Hz)",
"Frequency (MHz)", "Sagedus (MHz)",
"Frequency (in MHz).", "Sagedus (MHz'ides)",
"Frequency (kHz)", "Sagedus (kHz)",
"Fri", "R",
"Friendly name", "Sõbralik nimi",
"Frontend number", "Eesteenuse number",
"Frontend path", "Eesteenuse asukohaviide",
"Full DiseqC", "Täielik DiseqC",
"Full mux RX mode supported", "Täielik muxi RX režiim on toetatud",
"Full pathname specification", "Täispika asukohaviite määramine",
"Full-text", "Täistekst",
"Further education", "Rakendusharidus",
"GOTOX position", "GOTOX positioon",
"Game show", "Mängushow",
"Gardening", "Aiandus",
"General", "Üldine",
"General configuration", "Üldseaded",
"Generic", "Üldine",
"Generic options", "Üldised seaded",
"Global Headers", "Globaalsed päised",
"Global timer", "Globaalne taimer",
"Gray", "Hall",
"HD: high definition", "HD",
"HDHomeRun ATSC-C frontend", "HDHomeRun ATSC-C eesteenus",
"HDHomeRun ATSC-T frontend", "HDHomeRun ATSC-T eesteenus",
"HDHomeRun DVB frontend", "HDHomeRun DVB eesteenus",
"HDHomeRun DVB-C frontend", "HDHomeRun DVB-C eesteenus",
"HDHomeRun DVB-T frontend", "HDHomeRun DVB-T eesteenus",
"HTSP Answer", "HTSP vastus",
"HTSP Default Stream Settings", "HTSP voogedastuse vaikimisi seaded",
"HTSP Request", "HTSP päring",
"HTSP Server", "HTSP server",
"HTSP Stream Profile", "HTSP voogedastuse profiil",
"HTSP Subscription", "HTSP liitumine",
"HTTP CORS origin", "HTTP CORS päritolu",
"HTTP Client", "HTTP klient",
"HTTP Server", "HTTP server",
"Handicraft", "Käsitöö",
"Hearing impaired", "Vähenenud kuulmine",
"Hierarchy", "Hierarhia",
"High", "Kõrge",
"High bitrate mode (CI+ CAMs only)", "Kõrge bitikiiruse režiim (ainult CI+ CAM'id)",
"Highest count of objects.", "Objektide kõrgeim koguarv.",
"Historical movie", "Ajalooline film",
"Horror", "Õudus",
"Hostname (or IP) of the server.", "Serveri nimi (või IP).",
"Hostname/IP", "Hosti nimi / IP",
"How frequently it will re-try fetching an image that has failed to be fetched.", "Kui tihti proovitakse uuesti tõmmata seda pilti, mida ei õnnestunud kätte saada.",
"How frequently the upstream provider is checked for changes.", "Kui tihti teenusepakkujalt uuritakse uuendusi.",
"ID of the network.", "Võrgu ID.",
"IDLE", "PASSIB NIISAMA",
"IGNORE", "IGNOREERI",
"IP address", "IP aadress",
"IP blocking", "IP blokeering",
"IPTV Network", "IPTV võrk",
"IPTV Subcription", "IPTV liitumine",
"IPTV input", "IPTV sisend",
"ISDB-C Network", "ISDB-C võrk",
"ISDB-S Network", "ISDB-S võrk",
"ISDB-T Network", "ISDB-T võrk",
"Icon", "Ikoon",
"Icon (full URL)", "Ikoon (täispikk URL)",
"Icon URL", "Ikooni URL",
"Icon base URL", "Ikooni baasURL",
"Icon base URL.", "Ikooni baasURL.",
"Icon has title", "Ikoonil on pealkiri",
"Icons from", "Ikoonid pärit:",
"Idle", "Passib niisama",
"Idle scan", "Skaneerimine niisama passimise olekus ",
"Idle scan muxes", "Skaneeri muxe niisama passimise olekus",
"If an error occurs clone the scheduled entry and try to record again (if possible).", "Kui juhtub mingi viga, siis klooni saate salvestamiseks tehtud ajastatud kirje ja püüa seda uuesti salvestada (jääb ootama, kuni saade EPG sisus uuesti välja ilmub).",
"If more than x errors occur during a recording schedule a re-record (if possible).", "Kui salvestamise ajal juhtub kvaliteedivigu rohkem kui see määr, siis ajasta uuesti ja püüa seda saadet uuesti salvestada (jääb ootama, kuni saade EPG sisus uuesti välja ilmub).",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Kui ühtegi kasutajat pole, siis loo üks ilma kasutajanime ja ilma paroolita.\nAga kasuta seda varianti ettevaatlikult, sest see lubab Sinu Tvheadend\ninstallatsioonile maailmale täiesti valla oleva administraatori õigustes\nkasutamise. Seda niikauaks kuni Sa lood või korrigeerid kasutajate\nligipääsu Tvheadendi veebiliidese kaudu.",
"If you'd like to support the project, please consider a donation.", "Kui Sa soovid projekti toetada, kaalu selleks annetuse tegemist.",
"Ignore", "Ignoreeri",
"Ignore EPG (EIT)", "Igoreeri EPG'd (EIT)",
"Ignore descriptor 5", "Ignoreeri diskriptorit nr 5",
"Ignore invalid SSL certificate", "Ignoreeri kehtetut SSL sertifikaati",
"Ignore invalid/unverifiable (expired, self-certified, etc.) certificates", "Ignoreeri kehtetuid/kontrollimatuid (aegunud, ise signeeritud jne) sertifikaate",
"Ignore provider's channel numbers", "Ignoreeri pakkujapoolseid kanalite numbreid",
"Image Cache", "Pildi vahemälu",
"Immediately abort", "Katkesta koheselt",
"Important", "Tähtis",
"Include channel name in filename", "Kaasa faili nimes kanali nimi",
"Include date in filename", "Kaasa faili nimes kuupäev ",
"Include episode in filename", "Kaasa faili nimes osa",
"Include subtitle in filename", "Kaasa faili nimes osa nimi",
"Include teletext", "Kaasa teletekst",
"Include time in filename", "Kaasa fail inimes kellaaeg",
"Index", "Indeks",
"Information area", "Informatsiooni ala",
"Informational", "Informatsionaalne",
"Initial scan", "Esialgne skakneerimine",
"Input base", "Baassisend",
"Input buffer (bytes)", "Sisendpuhver (baitides)",
"Input instance", "Sisendinstants",
"Input path", "Sisendi asukohaviide",
"Interface", "Liides",
"Internal", "Sisemine",
"Internal EPG grabber", "Sisemine EPG hankija",
"Internal PyEPG Grabber", "Sisemine PyEPG hankija",
"Internal XMLTV EPG Grabber", "Sisemine XMLTV EPG hankija",
"Internal error", "Sisemine viga",
"Internal grabber", "Sisemine hankija",
"Interview", "Intervjuu",
"Invalid", "Kehtetu",
"Invalid adapter number '%s'\n", "Kehtetu adapteri number '%s'\n",
"Invalid arguments", "Kehtetud argumendid",
"Invalid file", "Kehtetu fail",
"Invalid parameters", "Kehtetud parameetrid",
"Invalid service", "Kehtetu teenus",
"Invalid target", "Kehtetu sihtpunkt",
"Items", "Esemed",
"Keep", "Hoia alles",
"Keep x amount of storage space free.", "Hoia x ühikut salvestusmahtu vabana.",
"Keepalive interval", "Keepalive intervall",
"Keepalive interval in seconds", "Keepalive intervall sekundites.",
"Kill signal (pipe)", "Killi signaal (pipe)",
"Kill timeout (pipe/secs)", "Killimise timeout (pipe/sekundis)",
"LNB type", "LNB tüüp",
"Language", "Keel",
"Language 1", "Keel 1",
"Language 2", "Keel 2",
"Language 3", "Keel 3",
"Language settings", "Keele seaded",
"Languages", "Keeled",
"Last seen", "Viimati nähtud",
"Last updated from", "Viimane uuendus",
"Leisure hobbies", "Vabaaja hobid",
"Limit connections", "Ühenduste limiit",
"Linked input", "Lingitud sisend",
"Linux DVB frontend", "Linux DVB eesteenus",
"LinuxDVB Input", "LinuxDVB sisend",
"LinuxDVB network", "LinuxDVB võrk",
"Listen / Connect port", "Kuula / ühenda porti",
"Listen on IPv6", "Kuula IPv6 võrguliidesel",
"Literature", "Kirjandus",
"Local bind IP address", "Kohalik IP aadress, millega seostada",
"Local channel minor", "Kohalik kanali minor",
"Local channel number", "Kohalik kanali number",
"Local discovery IP address", "Kohaliku tuvastamise IP aadress",
"Location", "Asukoht",
"Location details of the SAT>IP Server.", "SAT>IP serveri lokaliseerimise detailid.",
"Lock preferred CA PID", "Lukusta eelistatud CA PID",
"Locking", "Lukustus",
"Log", "Logi",
"Login password.", "Sisselogimise salasõna.",
"Login username.", "Sisselogimise kasutajanimi.",
"Login/Logout", "Sisse-/väljalogimine",
"Low", "Madal",
"MPEG-TS File", "MPEG-TS fail",
"MPEG-TS Multiplex", "MPEG-TS multiplex",
"MPEG-TS Parser", "MPEG-TS parser",
"MPEG-TS Pass-thru/built-in", "MPEG-TS Pass-thru/sisseehitatud",
"MPEG-TS Service", "MPEG-TS teenus",
"MPEG-TS input", "MPEG-TS sisend",
"MPEG-TS raw service", "MPEG-TS raw teenus",
"Magazines", "Ülevaated",
"Main", "Peamine",
"Maintain free storage space in MiB", "Hoia vabana salvestusmahtu, megabaitides",
"Maintain used storage space in MiB (0=disabled)", "Piira kasutuses olevat haldusmahtu, megabaitides (0=väljalülitatud)",
"Maintained space", "Hallatud salvestusmaht",
"Make subdirectories per channel", "Tee alamkataloogid per kanal",
"Make subdirectories per day", "Tee alamkataloogid per päev",
"Make subdirectories per title", "Tee alamkataloogid per pealkiri",
"Manage all", "Haldamisõigus kõigele",
"Manufacturer", "Tootja",
"Manufacturer URL", "Tootja URL",
"Manufacturer's model number.", "Valmistajapoolne mudeli number.",
"Manufacturer's product description.", "Valmistajapoolne toote kirjeldus.",
"Manufacturer's product information page for the device.", "Valmistajapoolne toote kirjelduse lehekülg.",
"Manufacturer's product name.", "Valmistajapoolne toote nimi.",
"Map all services", "Seosta kõik teenused",
"Map encrypted services", "Seosta krüpteeritud teenused",
"Map radio channels", "Seosta raadiokanalid",
"Map unnamed channels", "Seosta nimetud kanalid",
"Map zero-numbered channels", "Seosta nulliga nummerdatud kanalid",
"Martial sports", "Võitlussport",
"Master tuner", "Peamine tuuner",
"Matroska (mkv)/built-in", "Matroska (mkv)/sisseehitatud",
"Matroska muxer", "Matroska muxija",
"Matroska specific", "Matroska formaati keskne",
"Maximal channel number", "Maksimaalne kanali number",
"Maximum # input streams", "Maksimaalselt # sisendvoogedastust",
"Maximum PIDs", "Maksimaalselt PID'e",
"Maximum RAM size (MB)", "Maksimaalne RAM suurus (MB)",
"Maximum bandwidth (Kbps)", "Maksimum võrgukiirus (Kbps)",
"Maximum count (0=default)", "Maksimum arv (0=vaikimisi)",
"Maximum duration", "Maksimum kestvus",
"Maximum input bandwidth.", "Maksimum sisendvõrgukiirus.",
"Maximum length of PIDs", "Maksimum PID'ide pikkus",
"Maximum period (mins)", "Maksimum periood (minutit)",
"Maximum schedules limit (0=default)", "Ajastamiste maksimaalne arv (0=vaikimisi) ",
"Maximum size (MB)", "Maksimaalne suurus (MB)",
"Maximum timeout (seconds)", "Maksimaalne timeout (sekundites)",
"Medicine", "Meditsiin",
"Melodrama", "Melodraama",
"Memory Information", "Mäluinfo",
"Merge same name", "Liida samanimelised üheks kokku",
"Method not found", "Meetodit ei leitud",
"Minimal channel number", "Minimaalne kanali number",
"Minimum duration", "Minimaalne kestvus",
"Minimum rotor time (seconds)", "Minimaalne rootori aeg (sekundites)",
"Missing In PAT/SDT", "Puudub PAT/SDT's",
"Mode", "Režiim",
"Model description", "Mudeli kirjeldus",
"Model name", "Mudeli nimi",
"Model number", "Mudeli number",
"Modulation", "Modulatsioon",
"Module", "Moodul",
"Module ID", "Mooduli ID",
"Mon", "E",
"More verbose libav log", "Rohkem kirjeldav libav logi",
"Motor sport", "Mootorisport",
"Movie", "Film",
"Movie / drama", "Film / Draama",
"Music", "Muusika",
"Music / Ballet / Dance", "Muusika / Ballett / Tants",
"Musical", "Muusikal",
"Mux Scheduler", "Mux ajastaja",
"Mux handling", "Mux käsitlemine",
"Mux name", "Muxi nimi",
"Mux not enabled", "Muks pole sisse lülitatud",
"Muxer", "Muxija",
"NONE", "SEADISTAMATA",
"Name", "Nimi",
"Name of object.", "Objekti nimi.",
"Name of the bouquet.", "Kanalite kimbu nimi.",
"Name of the client.", "Kliendi nimi.",
"Name of the entry.", "Kirje nimi.",
"Name of the network.", "Võrgu nimi.",
"Name of the profile.", "Profiili nimi.",
"Name of the tag.", "Märke nimi.",
"Name of the tuner.", "Tuuneri nimi.",
"Name of the tuner/adapter.", "Tuuneri/adapteri nimi.",
"Names", "Nimed",
"Natural sciences", "Loodusteadused",
"Nature", "Loodus",
"Network", "Võrk",
"Network ", "Võrk",
"Network A", "Võrk A",
"Network B", "Võrk B",
"Network C", "Võrk C",
"Network D", "Võrk D",
"Network ID (limit scanning)", "Võrgu ID (piira skaneerimist)",
"Network UUID", "Võrgu UUID",
"Network access", "Võrgu ligipääs",
"Network discovery", "Võrgu tuvastamine",
"Network for port A.", "Võrk pordile A.",
"Network for port AA.", "Võrk pordile AA.",
"Network for port AB.", "Võrk pordile AB.",
"Network for port B.", "Võrk pordile B.",
"Network for port BA.", "Võrk pordile BA.",
"Network for port BB.", "Võrk pordile BB.",
"Network for port C.", "Võrk pordile C.",
"Network for port D.", "Võrk pordile D.",
"Network name", "Võrgu nimi",
"Network prefix", "Võrgu prefiks",
"Network settings", "Võrgu seaded",
"Network type", "Võrgu tüüp",
"Networks", "Võrgud",
"New login", "Uus sisselogimine",
"New media", "Uus meedia",
"New muxes + changed muxes", "Uued muxid + muudetud muxid",
"New muxes only", "Ainult uued muxid",
"News", "Uudised",
"News magazine", "Uudiste ülevaade",
"Next tune delay in ms (0-2000)", "Järgmise tuunimise viide millisekundites (0-2000)",
"No", "Ei",
"No PIN", "Pole PIN'i",
"No access", "Pole ligipääsu",
"No adapters specified!\n", "Ühtegi adapterit pole määratud!\n",
"No assigned adapters", "Pole määratletud adapterit",
"No days", "Pole päevi",
"No descrambler", "Pole lahtikrüpteerijat",
"No free adapter", "Pole vaba adapterit",
"No input detected", "Mingit sisendit pole tuvastatud",
"No scheme", "Pole malli",
"No service assigned to channel", "Kanaliga pole seotud ühtegi teenust",
"No service enabled", "Ühtegi teenust pole kasutuses",
"No source available", "Ühtegi allikat pole saadaval",
"None", "Seadistamata",
"Normal", "Tavaline",
"Not all SAT>IP servers use the same signal scaling. Change this setting if the signal level displayed within Tvheadend looks too low.", "Kõik SAT>IP serverid ei kasuta samasugust signaali skaleerimist. Muuda seda seadet kui signaali tase Tvheadendi poolt paistab liiga madal.",
"Not enough disk space", "Andmekandjal pole piisavalt vaba ruumi",
"Not enough memory", "Pole piisavalt mälumahtu",
"Not linked", "Pole lingitud",
"Not set", "Pole seadistatud",
"Not set (none or channel configuration)", "Pole seadistatud (kas pole midagi määratud või kanali enda seadistus)",
"Not set (use channel or DVR configuration)", "Pole seadistatud (kasuta kanali või DVR seadeid)",
"Number of days to keep the file.", "Päevade arv kaua faili alles hoida.",
"Number of days to keep the recorded file.", "Päevade arv kaua salvestatud faili alles hoida.",
"Number of days to keep the recorded files.", "Päevade arv kaua salvestatud faile alles hoida.",
"Number of days to retain information about recordings. Once this period is exceeded, duplicate detection will not be possible anymore.", "Päevade arv kaua hoida alles salvestamise logi. Kui see ajaperiood saab läbi, siis pole enam võimalik kontrollida duplikaatide olemasolu.",
"Number of muxes found.", "Leitud muxide arv.",
"Number of seconds to wait before timing out.", "Mitu sekundit oodata enne loobumist.",
"Number of tsfile tuners", "Tsfile tuunerite arv",
"OK", "Korras",
"OK (partial)", "Korras (osaliselt)",
"OSCam new pc-nodmx (rev >= 10389)", "OSCam'i uus pc-nodmx (rev >= 10389)",
"Odd key", "Paaritu võti",
"Odd key.", "Paaritu võti.",
"Off", "Väljas",
"Older OSCam", "Vanem OSCAM",
"On", "Sees",
"On file removal", "Faili eemaldamisel",
"Only Bulsatcom 39E", "Ainult Bulsatcom 39E",
"Only EIT", "Ainult EIT",
"Only OpenTV Sky Ausat", "Ainult OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Ainult OpenTV Sky Italia",
"Only OpenTV Sky UK", "Ainult OpenTV Sky UK",
"Only PSIP (ATSC)", "Ainult PSIP (ATSC)",
"Only UK Freesat", "Ainult UK Freesat",
"Only UK Freeview", "Ainult UK Freeview",
"Only Viasat Baltic", "Ainult Viasat Baltic",
"Only digits", "Ainult numbrid",
"Only preferred CA PID", "Ainult eelistatud CA PID",
"OpenTV channel number", "OpenTV kanali number",
"Opera", "Ooper",
"Orbital position", "Orbitaalpositsioon",
"Orbital positions", "Orbitaalpositsioonid",
"Orbital positions.", "Orbitaalpositsioonid.",
"Original network ID", "Originaalvõrgu ID",
"Output directory for tsdebug", "Väljundkataloog tsdebug'i jaoks",
"Over-the-air", "Üle-õhu",
"Over-the-air Cron multi-line", "Üle-õhu crontab",
"Over-the-air EPG", "Üle-õhu EPG",
"Over-the-air EPG grabber", "Üle-õhu EPG hankija",
"Over-the-air grabbers", "Üle-õhu hankijad",
"Owner", "Omanik",
"Owner of the rule.", "Reegli omanik.",
"PLS code", "PLS kood",
"PLS mode", "PLS režiim",
"Parent entry", "Ülemkirje",
"Parent entry.", "Ülemkirje.",
"Pass subscription weight", "Anna edasi liitumise kaalukus",
"Password", "Salasõna",
"Password2", "Salasõna2",
"Path", "Asukohaviide",
"Path to a directory (folder) containing your picon collection. See Help for more detailed information.", "Asukohaviide kataloogile, mis sisaldab Su picon kollektsiooni. Vaata lisainformatsiooniks abilehekülgi.",
"Path to an icon for this channel. This can be named however you wish, as either a local (file://) or remote (http://) image. See Help for more infomation.", "Asukohaviide selle kanali ikoonile. See võib olla nimetatud vabal moel, kas kohalik (file://) või kauge (http://) pilt. vaata lisainformatsiooniks abilehekülgi.",
"Performing arts", "Soorituskunstid",
"Periodically save EPG to disk (hours)", "Perioodiliselt salvesta EPG andmekandjale (tundides)",
"Persistent user interface level", "Kasutajaliidese läbivalt püsiv tase",
"Physiology", "Füsioloogia",
"Polarization", "Polarisatsioon",
"Political issues", "Poliitilised hädad",
"Popular culture", "Popkultuur",
"Port to connect to.", "Port, kuhu ühendust võtta.",
"Position", "Positsioon",
"Position of the input.", "Sisendi positsioon.",
"Post-processor command", "Pärast käitlust antav käsklus",
"Post-recording padding", "Lisamine pärast salvestust",
"Post-remove command", "Pärast eemaldamist antav käsklus",
"Power save", "Energiasääst",
"Power-up time (ms) (15-200)", "Sisselülitumise aeg (ms) (15-200)",
"Pre-defined muxes", "Eeldefineeritud muxid",
"Pre-recording padding", "Lisamine enne salvestust",
"Pre-school children's programs", "Eelkooliealiste saated",
"Prefer picons over channel name", "Eelista picon'it kanali nimetusele",
"Preferred CA PID", "Eelistatud CA PID",
"Preferred audio language.", "Eelistatud audio keel.",
"Preferred service video type", "Eelistatud teenuse video tüüp",
"Presentation", "Esitlus",
"Presentation details.", "Esitluse detailid.",
"Priority", "Prioriteet",
"Priority (-10..10)", "Prioriteet (-10..10)",
"Private", "Privaatne",
"Profile name", "Profiili nimi",
"Program identification (PID) number to compare. Zero means any. This comparison is processed only when service comparison is active and for the Conditional Access filter.", "Programmi identifikaatori (PID) number võrdlemiseks. Null tähendab kõiki. See võrdlus tehakse ainult siis, kui teenuse võrdlemine on sisselülitatud ja ainult Conditional Access filtrile.",
"Provider", "Pakkuja",
"Provider ID", "Teenusepakkuja ID",
"Provider network name", "Teenusepakkuja võrgu nimi",
"Psychology", "Psühholoogia",
"Puppets", "Marionetid",
"PyEPG Import", "PyEPG import",
"Quad (4.0)", "Nelik (4.0)",
"Quiz", "Mälumäng",
"RAM only", "AInult RAM",
"RTSP Protocol", "RTSP protokoll",
"RTSP port (554 or 9983, 0 = disable)", "RTSP port (554 või 9983), 0 = lülita välja",
"Radio", "Raadio",
"Re-fetch period (hours)", "Taastõmbamise periood (tundides)",
"Re-fetch period (mins)", "Taastõmbamise periood (minutites)",
"Re-record%s%s", "Taassalvesta%s%s",
"Re-try period (hours)", "Taasüritamise periood (tundides)",
"Read error", "Lugemise viga",
"Record all", "Salvesta kõik",
"Record on these days only.", "Salvesta ainult nendel päevadel.",
"Recording file options", "Salvestamise faili valikud",
"Recording file size.", "Salvestatud faili suurus.",
"Recording system path", "Salvestamise süsteemi asukohaviide",
"Reject", "Keeldu",
"Religion", "Religioon",
"Religious", "Religioosne",
"Remarkable people", "Erilised inimesed",
"Remove HTTP arguments", "Eelmalda HTTP argumendid",
"Remove all unsafe characters from filename", "Eemalda failinimest kõik tõlgendamatud tähemärgid",
"Repeated", "Korratud",
"Replace whitespace in title with '-'", "Asenda tühikud pealkirjas sidekriipsudega",
"Replaces all whitespace in the title with '-'.", "Asendab tühikud pealkirjas sidekriipsudega.",
"Reply to CAM PIN inquiries", "Vasta CAMi PIN küsimistele",
"Reply to PIN inquiries.", "Vastus PIN päringutele.",
"Reports", "Raportid",
"Rerun of", "Taasesitus:",
"Reserved", "Reserveeritud",
"Resolution (height)", "Resolutsioon (kõrgus)",
"Restart", "Taaskäivita",
"Restart on error", "Taaskäivita veasündmuse korral",
"Reuse EPG from", "Taaskasuta EPG'd",
"Rewrite MPEG-TS SI tables", "Ülekirjuta MPEG-TS SI tabelid",
"Rights", "Kasutajaõigused",
"Rock", "Rokkmuusika",
"Romance", "Romantika",
"Rotor initialization time (seconds)", "Rootori käimatõmbamise aeg (sekundites)",
"Rotor type", "Rootori tüüp",
"Run as group", "Käivita grupiga",
"Run as user", "Käivita kasutajana",
"Running", "Töötab",
"SAT->IP frontend number.", "SAT->IP eesteenuse number.",
"SAT>IP ATSC-C Frontend", "SAT>IP ATSC-C eesteenus",
"SAT>IP ATSC-T Frontend", "SAT>IP ATSC-T eesteenus",
"SAT>IP Client", "SAT>IP klient",
"SAT>IP DVB Frontend", "SAT>IP DVB eesteenus",
"SAT>IP DVB-C Frontend", "SAT>IP DVB-C eesteenus",
"SAT>IP DVB-S Frontend", "SAT>IP DVB-S eesteenus",
"SAT>IP DVB-S Slave Frontend", "SAT>IP DVB-S alama eesteenus",
"SAT>IP DVB-T Frontend", "SAT>IP DVB-T eesteenus",
"SAT>IP DVB-T frequency (Hz)", "SAT>IP DVB-T sagedus (Hz)",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "SAT>IP RTSP pordi number serveri jaoks\n(vaikimisi: -1 = välja lülitatud, 0 = veebiseadistus, standardport on 554)",
"SAT>IP Satellite Configuration", "SAT>IP satelliidi seaded",
"SAT>IP Server", "SAT>IP server",
"SAT>IP Server SI Tables", "SAT>IP serveri SI tabelid",
"SAT>IP source number", "SAT>IP lähteallika number",
"SD: standard definition", "SD",
"SSL verify peer", "Kontrolli naabrussuhte SSL",
"STOP", "STOPP",
"Sat", "L",
"Satellite config", "Satelliidi seaded",
"Satellite longitude", "Satelliidi geograafiline pikkus",
"Satellite longitude.", "Satelliidi geograafiline pikkus.",
"Satellite position.", "Satelliidi positsioon.",
"Satellite positions", "Satelliidi positsioonid",
"Scan after creation", "Skaneeri peale loomist",
"Scan progress", "Skaneerimise hetkeseis",
"Scan queue length", "Skaneerimise järjekorra pikkus",
"Scan result", "Skaneerimise tulemus",
"Scan status", "Skaneerimise olek",
"Scanfile", "Skaneerimisfail",
"Schedule a re-recording if more errors than (0=off)", "Ajasta taassalvestamine kui veateateid on rohkem kui (0=välja lülitatud)",
"Schedule status", "Ajastamise olek",
"Schedule status.", "Ajastamise olek.",
"Scheduled Duration", "Salvestamise kestvus",
"Scheduled for recording", "Määratud salvestamisele",
"Scheduled start time", "Algusaeg",
"Scheduled stop time", "Lõpuaeg",
"School programs", "Koolisaated",
"Science", "Teadus",
"Science fiction", "Teaduslik ulme",
"Season", "Hooaeg",
"Season %d", "Hooaeg %d",
"Seek error", "Ülesotsimise viga",
"Select the list of languages (in order of priority) to be used for supplying EPG information to clients that don't provide their own configuration.", "Vali kasutatavate keelte nimekiri (tähtsuse järjekorras), mille alusel jagatakse EPG infot klientidele, kes ei teavita oma kasutusel olevaid seadeid.",
"Send CAPMT OK query before descrambling.", "Saada CAPMT OK päring enne lahtikodeerimist.",
"Send CAPMT query", "Saada CAPMT päring",
"Send full PLAY cmd", "Saada täielik PLAY cmd",
"Serial number", "Serialnumber",
"Series link", "Seriaali link",
"Serious", "Tõsine",
"Serious music", "Tõsine muusika",
"Server UUID", "Serveri UUID",
"Server connectivity", "Serveri ühenduvus",
"Server details.", "Serveri detailid.",
"Service", "Teenus",
"Service ID", "Teenuse ID",
"Service Mapper", "Teenuse seostaja",
"Service Mapping (Map services to channels)", "Teenuse seostamine (seosta teenused kanaliteks)",
"Service configuration", "Teenuse seaded",
"Service count %zi", "Teenuse loend %zi",
"Service mapping", "Teenuse seostamine",
"Service name", "Teenuse nimi",
"Service raw", "Teenus raw kujul",
"Service type", "Teenuse tüüp",
"Services", "Teenused",
"Set the priority of this configuration.", "Määra nende seadete prioriteet.",
"Settings", "Seaded",
"Show", "Šhow",
"Show / Game show", "Show / Mängushow",
"Show this page", "Näita sellel leheküljel",
"Show version information", "Näita versiooni informatsiooni",
"Signal quality too poor", "Signaali kvaliteet liiga halb",
"Signal scale (240 or 100)", "Signaali skaala (240 või 100)",
"Site latitude", "Saidi geograafiline laius",
"Site latitude.", "Saidi geograafiline laius.",
"Site longitude", "Saidi geograafiline pikkus",
"Site longitude.", "Saidi geograafiline pikkus.",
"Size", "Suurus",
"Skip TS packets (0-200)", "Jäta vahele TS paketid (0-200)",
"Skip TSID checking. For when providers use invalid Transport Stream IDs.", "Jäta TSID kontrollimine vahele. Juhuks kui teenusepakkuja kasutab mingeid valesid IDsid.",
"Skip commercials", "Väldi kommertsreklaame",
"Skip initial bytes", "Jäta vahele algbaidid",
"Skip x number of transport packets.", "Jäta vahele x numbrit transpordi pakette.",
"Slave entry", "Alamkirje",
"Slave entry.", "Alamkirje.",
"Slot state", "Ühenduspesa staatus",
"Soap", "Seep",
"Soccer", "Jalgpall",
"Social", "Sotsiaal",
"Social advisory", "Sotsiaalne nõustamine",
"Sort index", "Sortimise indeks",
"Sort index.", "Sortimise indeks.",
"Source", "Allikas",
"Source deleted", "Lähteallikas eemaldatud",
"Source quality is bad", "Lähteallika kvaliteet on halb",
"Source reconfigured", "Lähteallikas taasseadistatud",
"Southern hemisphere (latitude direction)", "Lõunapoolkera (geograafiline laius)",
"Southern hemisphere (latitude direction).", "Lõunapoolkera (geograafiline laius).",
"Special events (Olympic Games, World Cup, etc.)", "Erisündmused (olümpiamängud, maailmameistrivõistlused jne)",
"Specify User-Agent header for the http client", "Määra User-Agent päis HTTP kliendi jaoks",
"Specify alternative htsp port", "Määra altenatiivne HTSP port",
"Specify alternative http port", "Määra altenatiivne HTTP port",
"Specify alternative http webroot", "Määra alternatiivne HTTP juurkataloog",
"Specify bind address", "Määra seostuv IP aadress",
"Specify bind address for SAT>IP server", "Määra seostuv IP aadress SAT>IP serverile",
"Specify extra htsp port", "Määra HTSP lisaport",
"Spiritual sciences", "Esoteerika",
"Sports", "Sport",
"Sports magazines", "Spordiülevaated",
"Start", "Algus",
"Start after", "Alusta pärast",
"Start before", "Alusta ennem",
"Start recording earlier than the EPG/timer-defined start time by x minutes.", "Alusta salvestamist nii mitu mintutit varem kui EPG/taimer seda teeks.",
"Start recording earlier than the defined start time by x minutes.", "Alusta salvestamist nii mitu minutit varem kui see muidu algama peaks.",
"Start time", "Algusaeg",
"Status", "Olek",
"Status period (ms)", "Oleku periood (ms)",
"Stop", "Seiska",
"Stop time", "Seiskamise aeg",
"Storage path", "Andmekandja asukohaviide",
"Storage space", "Andmekandja ruumikasutus",
"Stream ID", "Voogedastuse ID",
"Stream Profile", "Voogedastuse profiil",
"Stream index", "Voogedastuse indeks",
"Stream profile", "Voogedastuse profiil",
"Stream type", "Voogedastuse tüüp",
"Streaming", "Voogedastus",
"Streaming Profile", "Voogedastuse profiil",
"Streaming priority", "Voogedastuse prioriteet",
"Streaming profiles", "Voogedastuse profiilid",
"Subdirectory options", "Alamkataloogi valikud",
"Subscribe to a service permanently", "Liitu teenusega alatiseks",
"Subscription", "Liitumine",
"Subscription does not exist", "Liitumist ei eksisteeri",
"Subscription weight", "Liitumise kaalukus",
"Substitute formatters", "Vaheta formaatijad",
"Subtitle", "Alampealkiri",
"Subtitle Stream Filter", "Subtiitrite voogedastuse filter",
"Subtitle codec", "Subtiitrite koodek",
"Sun", "P",
"Surround (2 front, rear mono)", "Surround (2 ees, taga mono)",
"Switch before rotor", "Lülita enne rootorit",
"Switch type", "Lülituse tüüp",
"Switch: %s", "Lülita: %s",
"Symbol rate (Sym/s)", "Sümbolite määr (Sym/s)",
"Sync", "Sünkimine",
"Sync + Don't keep", "Sünkimine + ära hoia alles",
"System", "Süsteem",
"TCP Protocol", "TCP protokoll",
"TV Adapters - Linux ATSC-C Frontend", "TV adapterid - Linux ATSC-C eesteenus",
"TV Adapters - Linux ATSC-T Frontend", "TV adapterid - Linux ATSC-T eesteenus",
"TV Adapters - Linux DAB Frontend", "TV adapterid - Linux DAB eesteenus",
"TV Adapters - Linux DVB-C Frontend", "TV adapterid - Linux DVB-C eesteenus",
"TV Adapters - Linux DVB-S Frontend (Master)", "TV adapterid - Linux DVB-S eesteenus (Master)",
"TV Adapters - Linux DVB-S Slave Frontend", "TV adapterid - Linux DVB-S alama eesteenus ",
"TV Adapters - Linux DVB-T Frontend", "TV adapterid - Linux DVB-T eesteenus",
"TV Adapters - Linux ISDB-C Frontend", "TV adapterid - Linux ISDB-C eesteenus",
"TV Adapters - Linux ISDB-S Frontend (Master)", "TV adapterid - Linux ISDB-C eesteenus (Master)",
"TV Adapters - Linux ISDB-T Frontend", "TV adapterid - Linux ISDB-T eesteenus",
"Tag files with metadata", "Märgi failid metadataga",
"Tags", "Märked",
"Talk show", "Jutusaade",
"Team sports (excluding football)", "Võistkondlik sport (ei sisalda jalgpalli)",
"Technology", "Tehnoloogia",
"Teletext Stream Filter", "Teleteksti voogedastuse filter",
"Testing", "Testimine",
"Testing options", "Testimise valikud",
"The CAM slot status.", "CAM mooduli olek.",
"The EPG grabber name.", "EPG hankija nimi.",
"The EPG grabber type.", "EPG hankija tüüp.",
"The PIN to use.", "PIN kood.",
"The SAT>IP server's name.", "SAT>IP serveri nimi.",
"The SAT>IP server's universally unique identifier.", "SAT>IP serveri unikaalne identifikaator.",
"The SAT>IP source number.", "SAT>IP lähteallika number.",
"The SAT>IP's discovered IP address.", "SAT>IP tuvastatud IP aadress.",
"The audio stream types the filter should apply to.", "Audio voo tüübid, millele filtrit rakendada.",
"The channel this service is mapped to.", "Kanal, millega teenus seostatud on.",
"The character encoding for this network (e.g. UTF-8).", "Kooditabel selle võrgu jaoks (nt UTF-8).",
"The character encoding for this service (e.g. UTF-8).", "Kooditabel selle teenuse jaoks (nt UTF-8).",
"The current boot ID.", "Pragune buutimise ID.",
"The current configuration ID.", "Praegune seadete ID.",
"The current configuration version.", "Praegune seadete versioon.",
"The default language to use if the user  language isn't set in the Access Entries tab.", "Vaikimisi kasutatav keel kui kasutaja keel pole määratud",
"The delivery system used by your cable provider.", "Teenuse edastamise tüüp, mida Su kaabliteenusepakkuja kasutab.",
"The delivery system used by your provider.", "Teenuse edastamise tüüp, mida Su teenusepakkuja kasutab.",
"The device ID.", "Seadme ID",
"The device's serial number.", "Seadme serialnumber.",
"The forward error correction high value. Most people will not need to change this setting.", "Edastamise veaparanduse ülemine väärtus. Enamus inimesi ei vaja selle väärtuse muutmist.",
"The forward error correction low value. Most people will not need to change this setting.", "Edastamise veaparanduse alumine väärtus. Enamus inimesi ei vaja selle väärtuse muutmist.",
"The forward error correction used on the mux.", "Muxi edastamise veaparandus.",
"The forward error correction. Most people will not need to change this setting.", "Muxi edastamise veaparandus. Enamus inimesi ei vaja selle seade muutmist.",
"The frequency of the mux (in Hertz).", "Muxi sagedus hertsides (Hz).",
"The items have the following functions:", "Väljad ja nendega seotud valikud:",
"The language to which the filter should apply.", "Keel, millele filter peaks rakenduma.",
"The maximum number of recordings that can be scheduled.", "Maksimaalne arv salvestamisi, mida saab tööle rakendada.",
"The modulation used on the mux.", "Muxi modulatsioon",
"The mux the service is on.", "Mux millel see teenus asub.",
"The mux's universally unique identifier.", "Muxi unikaalne identifikaator.",
"The name of the profile.", "Profiili nimi.",
"The name of the the rule.", "Reegli nimi.",
"The network the mux is on.", "Võrk millel mux asub.",
"The network the service is on.", "Võrk millel teenus asub.",
"The networks using this configuration.", "Võrgud, mis kasutavad neid seadeid.",
"The number of ATSC-C (Cable/AnnexB) tuners to export.", "Mitu ATSC-C (Cable/AnnexB) tüünerit eksportida.",
"The number of ATSC-T (Terresterial) tuners to export.", "Mitu ATSC-T (Terresterial) tüünerit eksportida.",
"The number of DVB-C (Cable) tuners to export.", "Mitu DVB-C (Cable) tüünerit eksportida.",
"The number of DVB-C2 (Cable) tuners to export.", "Mitu DVB-C2 (Cable) tüünerit eksportida.",
"The number of DVB-S (Satellite) tuners to export.", "Mitu DVB-S (Satellite) tüünerit eksportida.",
"The number of DVB-S2 (Satellite) tuners to export.", "Mitu DVB-S2 (Satellite) tüünerit eksportida.",
"The number of DVB-T (Terresterial) tuners to export.", "Mitu DVB-T (Terresterial) tüünerit eksportida.",
"The number of DVB-T2 (Terresterial) tuners to export.", "Mitu DVB-T2 (Terresterial) tüünerit eksportida.",
"The number of days cookies set by Tvheadend should expire.", "Mitu päeva on Tvheadendi poolt seatud veebiküpsiste eluiga / aegumine.",
"The number of seconds to wait for a stream to start.", "Mitu sekundit oodata voo algamist.",
"The orbital position of the satellite the mux is on.", "Satelliidi orbitaalpositsioon millel see mux asub.",
"The playlist's character set.", "Esitusnimekirja kooditabel.",
"The polarization used on the mux.", "Muxi polarisatsioon.",
"The provider's ID.", "Teenusepakkuja ID",
"The provider's name.", "Teenusepakkuja nimi.",
"The provider's network ID.", "Teenusepakkuja võrgu ID.",
"The provider's network name.", "Teenusepakkuja võrgu nimi",
"The satellite configuration to use.", "Satelliidi seaded, mida kasutada.",
"The scheduled start time, including any padding.", "Salvestamise algusaeg, arvestades varem salvestamise lisaaega.",
"The scheduled stop time, including any padding.", "Salvestamise lõpuaeg, arvestades salvestamise pikendamise lisaaega.",
"The service ID", "Teenuse ID",
"The service ID as set by the provider.", "Teenusepakkuja poolt määratud teenuse ID.",
"The service ID.", "Teenuse ID.",
"The service name as set by the provider.", "Teenusepakkuja poolt määratud teenuse nimi.",
"The service the filter should apply to. Leave blank to apply the filter to all services.", "Teenus, millele filter rakendada. Jäta tühjaks, kui soovid määrata kõigile teenustele.",
"The service to which the filter should apply. Leave blank to apply the filter to all services.", "Teenus, millele filter rakendada. Jäta tühjaks, kui soovid määrata kõigile teenustele.",
"The service's channel minor as set by the provider.", "Teenuse kanali minor, mis määratud teenusepakkuja poolt.",
"The service's channel number as set by the provider.", "Teenusepakkuja poolt määratud kanali number.",
"The service's encryption status.", "Teenuse krüpteeringu staatus.",
"The start time of the recording.", "Salvestuse algusaeg.",
"The stream ID used for the mux.", "Voogedastuse ID muxi jaoks.",
"The stream profile the DVR profile will use for recordings.", "Voogedastuse profiil, mida salvestamiseks kasutada.",
"The subtitle stream types the filter should apply to.", "Subtiitrite voo tüübid, millele filtrit rakendada.",
"The symbol rate used on the mux.", "Sümbolite määr sellel muxil.",
"The symbol rate.", "Sümbolite määr.",
"The total scheduled duration.", "Ajastamiste kogukestvus.",
"The transponder ID.", "Transponderi ID.",
"The transport stream ID of the mux within the network.", "Muxi transpordi voo ID võrgu sees.",
"The video stream types the filter should apply to.", "Video voo tüübid, millele filtrit rakendada.",
"Theme", "Teema",
"This tuner", "See tuuner",
"This will create an NTP driver (using shmem interface) that you can feed into ntpd. This can be run without root privileges, but generally the performance is not that great.", "See loob NTP draiveri (kasutades shmem liidest) mille saab ette sööta ntpd'le. See saab toimida ka ilma root õigusteta, aga selle variandi jõudlus pole just kõige parem.",
"Thu", "N",
"Time", "Aeg",
"Time missed", "Aeg mööda läinud",
"Time record caption", "Kellaaja alusel salvestamise tekst",
"Time recording%s%s", "Kellaja alusel salvestamine %s%s",
"Time update", "Kellaaja uuendus",
"Timeout (sec) (0=infinite)", "Timeout (sekundites) (0=lõputu)",
"Timeout (seconds)", "Timeout (sekundites)",
"Timeshift", "Ajanihe",
"Title", "Saate nimi",
"Title (regexp)", "Pealkiri (Regexp)",
"Toggle details", "Taustadetailid",
"Too many data errors", "Liiga palju data vigu",
"Total number of mapped channels on this network.", "Selle võrgu seostatud teenuste koguarv. ",
"Total number of muxes found on this network.", "Selles võrgus leitud muxide koguarv.",
"Total number of services found on this network.", "Selles võrgus leitud teenuste koguarv.",
"Total number of services found.", "Leitud teenuste koguarv.",
"Total number of services seen.", "Nähtud teenuste koguarv.",
"Total number of services.", "Teenuste koguarv.",
"Tourism / Travel", "Turism / reisimine",
"Trace subsystems", "Treissi alamsüsteeme",
"Traditional arts", "Traditsioonilised kunstid",
"Traditional music", "Traditsiooniline muusika",
"Transcode", "Skaleeri",
"Transcoding", "Skaleerimine",
"Transponder ID", "Transponderi ID",
"Transport Stream", "Transpordi voogedastus",
"Transport stream ID", "Transpordi voogedastuse ID",
"Tue", "T",
"Tune before DiseqC", "Tuuni ennem DiseqC'd",
"Tuner", "Tuuner",
"Tuner bind IP address", "Tuuneriga seostatud IP aadress",
"Tuner configuration", "Tuuneri seaded",
"Tuner configuration.", "Tuuneri seaded.",
"Tuning failed", "Tuunimine ebaõnnestus",
"Turn off LNB when idle", "Lülita LNB välja kui passib niisama",
"Tvheadend server name", "Tvheadend serveri nimi",
"Type", "Tüüp",
"UDP RTP port number (2 ports)", "UDP RTP pordi number (2 porti)",
"UPnP Protocol", "UPnP protokoll",
"URL of the M3U playlist.", "M3U esitluslisti URL.",
"URL with the SAT>IP server XML location", "URL koos SAT>IP serveri XML asukohaga",
"Unable to get system UTC time", "Ei saa kätte süsteemi UTC aega",
"Unable to get system local time", "Ei saa kätte süsteemi kohalikku aega",
"Unable to get system time", "Ei saa kätte süsteemi aega",
"Unable to open file", "Ei saa faili avada",
"Uncommitted", "Rakendamata",
"Uncommitted first", "Esimesena rakendamata",
"Unicable type", "Unicable tüüp",
"Unimportant", "Tähtsusetu",
"Universal LNB only", "Ainult universaalne LNB",
"Universally unique identifier. Read only.", "Unikaalne identifikaator. Ainult loetav.",
"Unknown", "Tundmatu",
"Unknown file", "Tundmatu fail",
"Unknown reason (%i)", "Tundmatu põhjus (%i)",
"Unlimited size", "Piiramata suurus",
"Unlimited time", "Piiramata aeg",
"Unset (default)", "Seadistamata (vaikimisi)",
"Update Entitlement Management Messages from one mux only.", "Entitlement Management Message uuendamine ainult ühelt muxilt.",
"Update card (EMM)", "Uuenda kaarti (EMM)",
"Update channel icon", "Uuenda kanali ikooni",
"Update channel name", "Uuenda kanali nime",
"Update channel number", "Uuenda kanali numbrit",
"Update disabled", "Uuendamine välja lülitatud",
"Update time", "Uuenda kellaaega",
"Update tolerance (ms)", "Uuendamise tolerants (ms)",
"Updated", "Uuendatud",
"Updates from one mux (EMM)", "Uuendused ühelt muxilt (EMM)",
"Usage: %s [OPTIONS]\n", "Kasutamine: %s [valikud]\n",
"Use EPG running state", "Kasuta EPG tööstaatust",
"Use WEBM format.", "Kasuta WEBM formaati.",
"Use Windows-compatible filenames", "Kasuta Windows'iga ühilduvaid failinimesid",
"Use XSPF playlist instead of M3U", "Kasuta M3U asemel XSPF pleilisti",
"Use a pre-defined list of ATSC-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Kasuta eeldefineeritud ATSC-C muxide nimekirja. Märkus: need nimekirjad võivad olla aegunud ja selle tulemusel võib skaneerimine võtta tavalisest rohkem aega.",
"Use a pre-defined list of ATSC-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Kasuta eeldefineeritud ATSC-T muxide nimekirja. Märkus: need nimekirjad võivad olla aegunud ja selle tulemusel võib skaneerimine võtta tavalisest rohkem aega.",
"Use a pre-defined list of DVB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Kasuta eeldefineeritud DVB-C muxide nimekirja. Märkus: need nimekirjad võivad olla aegunud ja selle tulemusel võib skaneerimine võtta tavalisest rohkem aega.",
"Use a pre-defined list of DVB-S/S2 muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Kasuta eeldefineeritud DVB-S/S2 muxide nimekirja. Märkus: need nimekirjad võivad olla aegunud ja selle tulemusel võib skaneerimine võtta tavalisest rohkem aega.",
"Use a pre-defined list of DVB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Kasuta eeldefineeritud DVB-T muxide nimekirja. Märkus: need nimekirjad võivad olla aegunud ja selle tulemusel võib skaneerimine võtta tavalisest rohkem aega.",
"Use original", "Kasuta algupärast",
"Use service IDs as channel numbers", "Kasuta teenuse ID'sid kanalite numbritena",
"Use tag internally (don't expose to clients).", "Kasuta märget sisemiselt (ära avalda klientidele).",
"Use the provider's service IDs as channel numbers.", "Kasuta teenusepakkuja teenuste ID'sid kanalite numbritena.",
"Use x amount of storage space.", "Kasuta x ühikut salvestusmahtu.",
"User access error", "Kasutaja ligipääsu viga",
"User does not have access", "Kasutajal pole ligipääsu",
"User does not have access to channel", "Kasutajal pole ligipääsu kanalile",
"User icon", "Kasutaja ikoon",
"User interface level", "Kasutajaliidese tase",
"User language", "Kasutaja keel",
"User limit reached", "Kasutaja limiit on täis",
"User login", "Kasutaja sisselogimine",
"User request", "Kasutaja päring",
"Username", "Kasutajanimi",
"Users - Access Entries", "Kasutajad - Kasutajate haldamine",
"Users - IP Blocking", "Kasutajad - IP blokeering",
"Users - Passwords", "Kasutajad - Paroolid",
"Variety show", "Varietee",
"Verify the SSL certificate.", "Kontrolli SSL sertifikaati.",
"Video Stream Filter", "Video voogedastuse filter",
"Video bitrate (kb/s) (0=auto)", "Video bitrate (kb/s) (0=automaatne)",
"Video codec", "Video koodek",
"Video codec preset", "Video koodeki eelseadistus",
"Video codec preset to use for transcoding.", "Video koodeki eelseadistus konverteerimiseks.",
"Video codec to use for the transcode. \"Do not use\" will disable video output.", "Video koodek konverteerimiseks. \"Ära kasuta\" lülitab video väljundi välja.",
"Video recorder", "Video salvestaja",
"View all", "Vaatamisõigus kõigele",
"Visually impaired commentary/audio description", "Vaegnägijatele kommentaarid / audio kirjeldus",
"WEBM/built-in", "WEBM/sisseehitatud",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV profiil H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV profiil H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV profiil VP8/Vorbis/WEBM",
"Waiting for EPG running flag", "Oodates EPG töötamise lipukest",
"Waiting for program start", "Oodates saate algamist",
"Waiting for stream", "Oodates voogedastust",
"War", "Sõda",
"Water sport", "Veesport",
"Weak stream", "Nõrk voogedastus",
"Weather report", "Ilmateade",
"Web User Interface", "Veebiliides",
"Web interface", "Veebiliides",
"Web interface language", "Veebiliidese keel",
"Web interface language.", "Veebiliidese keel.",
"Web interface theme.", "Veebiliidese välimus.",
"Web theme", "Veebiliidese välimus",
"Web user interface", "Veebiliides",
"Wed", "K",
"Welcome - Tvheadend - your TV streaming server and video recorder", "Tervist - Tvheadend - Sinu TV voogedastuse server ja video salvestaja",
"Western", "Vestern",
"Western hemisphere (latitude direction)", "Läänepoolkera (geograafiline laius)",
"Western hemisphere (latitude direction).", "Läänepoolkera (geograafiline laius).",
"When the service was first identified and recorded.", "Millal teenus oli esimest korda ära tuntud ja salvestatud.",
"When the service was last seen during a mux scan.", "Millal viimati teenust nähti muxi skaneerimisel.",
"Winter sports", "Talvesport",
"Wizard", "Juhendav abiprogramm",
"Write a short message to log identifying the matched parameters. It is useful for debugging your setup or structure of incoming streams.", "Kirjuta lühike teade, millega leitud tulemusi logis visuaalselt eraldada. Kasulik seadete, voogude jms veatuvastuseks.",
"XMLTV EPG Import", "XMLTV EPG import",
"Yes", "Jah",
"addpids/delpids supported", "addpids/delpids toetatud",
"default: nvenc(h264 / h265)", "vaikimisi: nvenc(h264 / h265)",
"fast: h264 / h265 / qsv(h264 / h265)", "kiire: h264 / h265 / qsv(h264 / h265)",
"faster: h264 / h265 / qsv(h264)", "kiirem: h264 / h265 / qsv(h264)",
"hrs", "tund(i)",
"invalid option specified [%s]", "vale valik määratletud [%s]",
"medium: h264 / h265 / qsv(h264 / h265)", "keskmine: h264 / h265 / qsv(h264 / h265)",
"mins", "minutit",
"option %s requires a value", "valik %s vajab väärtust",
"slow: h264 / h265 / qsv(h264 / h265)", "aeglane: h264 / h265 / qsv(h264 / h265)",
"slower: h264 / h265 / qsv(h264)", "aeglasem: h264 / h265 / qsv(h264)",
"superfast: h264 / h265", "ülikiire: h264 / h265",
"tsfile input (mux file)", "tsfile sisend (mux fail)",
"tvhdhomerun client", "tvhdhomerun klient",
"ultrafast: h264 / h265", "ultrakiire: h264 / h265",
"veryfast: h264 / h265 / qsv(h264)", "vägakiire: h264 / h265 / qsv(h264)",
"veryslow: h264 / h265 / qsv(h264)", "vägaaeglane: h264 / h265 / qsv(h264)",
NULL, NULL
};

static const char *tvh_locale_13[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nبرای اطلاعات بیشتر به وبسایت Tvheadend مراجعه فرمایید: https://tvheadend.org\n",
"# Muxes", "# تجمیع کننده سیگنال",
"# Services", "# سرویس",
"AC-3 = descriptor 6", "AC-3 = توصیفگر 6",
"ACTIVE", "فعال",
"AUTO", "خودکار",
"Aborted by user", "خاتمه یافته توسط کاربر",
"Access", "دسترسی",
"Action", "فعال",
"Channel", "کانال",
"Created", "ایجاد شده",
"Creator", "ایجاد کننده",
"Default", "پیشفرض",
"No access", "عدم دسترسی",
"None", "هیچکدام",
"Normal", "عادی",
"Number", "شماره",
"OK", "تایید",
"Off", "خاموش",
"On", "روشن",
"Password", "رمز",
"Password2", "رمز 2",
"Port", "پورت",
"Position", "موقعیت",
"Private", "خصوصی",
"Sat", "ماهواره",
"Services", "خدمات",
"Show this page", "این صفحه را نشان بده",
"Show version information", "نمایش اطلاعات ورژن",
"Signal quality too poor", "کیفیت سیگنال بسیار ضعیف است",
"Source", "منبع",
"Source deleted", "منبع پاک شده",
"Source quality is bad", "کیفیت منبع بد است",
"Source reconfigured", "منبع دوباره تنظیم شد",
"Standard", "استاندارد",
"Start", "شروع",
"Status", "وضعیت",
"Stereo", "استریو",
"Stop", "توقف",
"Subtitle", "زیرنویس",
"Sun", "یکشنبه",
"System", "سیستم",
"Thu", "پنج شنبه",
"Tue", "سه شنبه",
"Unimportant", "غیر مهم",
"Universal LNB only", "فقط ال ان بی بین المللی",
"Unknown", "ناشناخته",
"User access error", "خطای دسترسی کاربر",
"User limit reached", "پایان حد دسترسی کاربر",
"Username", "نام کاربری",
"Wed", "چهارشنبه",
"hrs", "ساعت",
"min", "دقیقه",
"mins", "دقایق",
NULL, NULL
};

static const char *tvh_locale_14[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nSaadaksesi enemmän lisätietoja vieraile Tvheadendin sivustolla\nhttps://tvheadend.org\n",
"# Mapped channels", "# Merkatut kanavat",
"# Muxes", "# Niput",
"# Services", "# Palvelut",
"1 day", "1 päivä",
"1 month", "1 kuukausi",
"1 week", "1 viikko",
"1 year", "1 vuosi",
"2 months", "2 kuukautta",
"2 weeks", "2 viikkoa",
"2 years", "2 vuotta",
"3 days", "3 päivää",
"3 months", "3 kuukautta",
"3 weeks", "3 viikkoa",
"3 years", "3 vuotta",
"Aborted by user", "Keskeytys käyttäjältä",
"Admin", "Ylläpitäjä",
"Admin password", "Ylläpidon salasana",
"Admin username", "Ylläpidon käyttäjänimi",
"All proceeds are used to support server infrastructure and buy test equipment.", "Kaikki tuotot käytetään kehitysympäristön ylläpitoon ja testilaitteiston hankintaan.",
"All: Record if different description", "Kaikki: Tallennus, jos poikkeava kuvaus",
"All: Record if different episode number", "Kaikki: Tallennus, jos poikkeava jakson numero",
"All: Record if different subtitle", "Kaikki: Tallennus, jos poikkeava alaotsikko",
"All: Record once per day", "Kaikki: Tallennus kerran päivässä",
"All: Record once per week", "Kaikki: Tallennus kerran viikossa",
"Allow high bitrate mode (CI+ CAMs only).", "Salli korkea bitrate moodi (ainoastaan CI+ CAM)",
"Alternate configuration path", "Vaihtoinen asetuspolku",
"Animals", "Eläimiä",
"Any", "Mikä tahansa",
"Arts", "Taide",
"Athletics", "Yleisurheilu",
"Ballet", "Baletti",
"Bandwidth", "Kaistanleveys",
"Bouquet", "Kimppu",
"Cartoons", "Piirretyt",
"Channel", "Kanava",
"Channels", "Kanavat",
"Cinema", "Elokuva",
"Class", "Luokka",
"Classical", "Klassinen",
"Classical music", "Klassinen musiikki",
"Comedy", "Komedia",
"Comment", "Kommentti",
"Commercial break", "Mainoskatko",
"Configuration", "Asetukset",
"Cooking", "Ruoanlaitto",
"Created", "Luotu",
"Creator", "Tekijä",
"Culture (without music)", "Kulttuuri (ei musiikkia)",
"Current affairs", "Ajankohtaisohjelmat",
"Dance", "Tanssi",
"Days of Week", "Viikonpäivät",
"Debate", "Väittely",
"Default", "Oletus",
"Description", "Kuvaus",
"Detective", "Etsivä",
"Directory", "Hakemisto",
"Discussion", "Keskustelu",
"Do not use", "Älä käytä",
"Documentary", "Dokumentti",
"Drama", "Draama",
"Economics", "Talous",
"Education", "Koulutus",
"Educational", "Opetusohjelmat",
"Environment", "Ympäristö",
"Episode", "Jakso",
"Episode %d", "Jakso %d",
"Errors", "Virheitä",
"Every day", "Joka päivä",
"Fantasy", "Fantasia",
"Filename", "Tiedoston nimi",
"Film", "Filmi",
"Fine arts", "Korkeakulttuuri",
"Fitness and health", "Terveys",
"Folk", "Kansan",
"Folkloric", "Kansanperinne",
"Football", "Jalkapallo",
"Foreign countries", "Vieraat maat",
"Frequency", "Taajuus",
"Frequency (Hz)", "Taajuus (Hz)",
"Frequency (MHz)", "Taajuus (MHz)",
"Frequency (kHz)", "Taajuus (kHz)",
"Fri", "Pe",
"Further education", "Jatkokoulutus",
"Game show", "Visailu",
"Gardening", "Puutarhanhoito",
"General", "Yleinen",
"Generic", "Yleinen",
"Handicraft", "Käsityöt",
"High", "Korkea",
"Historical movie", "Historiallinen elokuva",
"Horror", "Kauhu",
"Icon (full URL)", "Ikoni (koko URL)",
"Icon URL", "Ikonin URL",
"If you'd like to support the project, please consider a donation.", "Jos haluat tukea kehitystyötä, ehkä voisit tehdä lahjoituksen.",
"Important", "Tärkeä",
"Internal", "Sisäinen",
"Interview", "Haastattelu",
"Language", "Kieli",
"Languages", "Kielet",
"Leisure hobbies", "Harrastukset",
"Literature", "Kirjallisuus",
"Location", "Sijainti",
"Log", "Lok",
"Low", "Alhainen",
"Manufacturer", "Valmistaja",
"Medicine", "Lääketiede",
"Melodrama", "Melodraama",
"Mon", "Ma",
"Motor sport", "Moottoriurheilu",
"Motoring", "Autoilu",
"Movie", "Elokuva",
"Music", "Musiikk",
"Musical", "Musikaali",
"Name", "Nimi",
"Natural sciences", "Luonnontieteet",
"Nature", "Luonto",
"Network", "Verkko",
"Networks", "Verkot",
"News", "Uutiset",
"News magazine", "Uutismakasiini",
"No days", "Ei päiviä",
"None", "Ei mitään",
"Not set", "Ei asetettu",
"Number", "Numero",
"Opera", "Ooppera",
"Owner", "Omistaja",
"Password", "Salasana",
"Password2", "Salasana2",
"Performing arts", "Esiintyvä taide",
"Physiology", "Fysiologia",
"Pilot", "Pilotti",
"Political issues", "Politiikka",
"Popular culture", "Populaarikulttuuri",
"Port", "Portti",
"Position", "Sijainti",
"Presentation", "Esitys",
"Priority", "Prioriteetti",
"Priority (-10..10)", "Prioriteetti (-10..10)",
"Private", "Yksityinen",
"Psychology", "Psykologia",
"Puppets", "Nukketeatteri",
"Quiz", "Visailu",
"Religion", "Uskonto",
"Religious", "Uskonnollinen",
"Remarkable people", "Merkkihenkilöt",
"Reports", "Raportit",
"Romance", "Romantiikka",
"Sat", "La",
"Science", "Tiede",
"Science fiction", "Sci-Fi",
"Season", "Tuotantokausi",
"Season %d", "Tuotantokausi %d",
"Serious", "Vakava",
"Server", "Palvelin",
"Service", "Palvelu",
"Services", "Palvelut",
"Settings", "Asetukset",
"Soap", "Saippua",
"Soccer", "Jalkapallo",
"Sports", "Urheilu",
"Sports magazines", "Urheilumakasiinit",
"Subtitle", "Tekstitys",
"Sun", "Su",
"System", "Järjestelmä",
"Team sports (excluding football)", "Joukkueurheilu (ei jalkapallo)",
"Technology", "Teknologia",
"Thriller", "Jännitys",
"Thu", "To",
"Tourism / Travel", "Matkailu",
"Traditional arts", "Perinteiset taiteet",
"Traditional music", "Perinteinen musiikki",
"Tue", "Ti",
"Unknown", "Tuntematon",
"Unknown reason (%i)", "Tuntematon syy (%i)",
"Username", "Käyttäjänimi",
"War", "Sota",
"Water sport", "Vesiurheilu",
"Weather report", "Säätiedotus",
"Wed", "Ke",
"Western", "Lännen elokuva",
NULL, NULL
};

static const char *tvh_locale_15[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nPour plus d'informations, visitez le site de Tvheadend :\nhttps://tvheadend.org\n",
"# Mapped channels", "Chaînes associées",
"# Muxes", "# Transpondeurs",
"# tune repeats", "Répétition du tuner",
"(Default profile)", "(Profil par Défaut)",
"(Read-only)", "(Lecture seule)",
"1 day", "1 jour",
"1 month", "1 mois",
"1 week", "1 semaine",
"1 year", "1 an",
"2 months", "2 mois",
"2 weeks", "2 semaines",
"2 years", "2 ans",
"2-Port switch (universal LNB)", "Switch 2 Ports (universal LNB)",
"3 days", "3 jours",
"3 months", "3 mois",
"3 weeks", "3 semaines",
"3 years", "3 ans",
"4-Port switch (universal LNB)", "Switch 4 Ports (universal LNB)",
"5 days", "5 jours",
"6 months", "6 mois",
"A channel tag (e.g. a group of channels) to which this rule applies.", "Tag de chaine (ou groupe de chaines) auxquelles ces règles s’appliquent",
"AC-3 = descriptor 6", "AC-3 = descripteur 6",
"AC-3 detection", "Détection AC-3",
"ANY", "TOUS",
"ATSC SI Tables", "Tables ATSC SI",
"ATSC source ID", "ID source pour ATSC",
"ATSC-C Network", "Réseau ATSC-C",
"ATSC-T Network", "Réseau ATSC-T",
"Aborted by user", "Annulé par l'utilisateur",
"Accept zero value for TSID", "Accepter la valeur 0 pour TSID",
"Access", "Accès",
"Access (ACL)", "Accès (ACL)",
"Adapter in use by another subscription", "Adaptateur déjà utilisé par un autre service",
"Add file and line numbers to debug", "Ajouter le fichier et le numéro de ligne au debug",
"Add the thread ID to debug", "Ajouter l'ID du processus au debug",
"Additional arguments to pass to the grabber.", "Arguments supplémentaire a passer au grabber",
"Additional service names found in EPG data.", "Nom de services supplémentaires trouvée dans l'EPG",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "Temps supplémentaire (en secondes) a laisser au tuner pour être prêt a enregistrer. C'est utile aux tuner un peu lent et/ou ayant des données sans intérêt a envoyer au début.",
"Admin password", "Mot de passe administrateur",
"Admin username", "Nom de l'admin",
"Administrator login", "Login de l'admin",
"Adult movie", "Films pour adulte",
"Advanced", "Avancé",
"Advanced (non-universal LNBs, rotors, etc.)", "Avancé (LNBs spécifiques, rotors, etc.)",
"Adventure", "Aventure",
"Advertisement / Shopping", "Publicité / Shopping",
"All (Streaming plus DVR)", "Tout (flux + enregistrements)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Tous les caractères pouvant poser problème dans les noms de fichiers seront remplacés par un underscore. Voir l'aide pour les détails.",
"All lower-case", "Tout en minuscule",
"All proceeds are used to support server infrastructure and buy test equipment.", "Toutes les sommes collectées sont utilisées pour supporter l'infrastructure technique et acheter du matériel de test.",
"All: Record if different description", "Tous: Enregistrement si la description est différente",
"All: Record if different episode number", "Tous: Enregistrement si le numéro d’épisode est différent",
"All: Record if different subtitle", "Tous: Enregistrement si les sous-titre sont différents",
"All: Record once per day", "Tous: Enregistrement un par jour",
"All: Record once per week", "Tous: Enregistrement un par semaine",
"Allow high bitrate mode (CI+ CAMs only).", "Autoriser le mode haut débit (seulement pour les CI+ CAMs).",
"Allow idle scan tuning on this device.", "Autoriser les scan inactif sur cet appareil.",
"Allow the combined size of all timeshift buffers to potentially grow unbounded until your storage media runs out of space.", "Autoriser la taille de tous les timeshift a potentiellement dépasser la taille restante sur le disque",
"Allow the timeshift buffer to grow unbounded until your storage media runs out of space. Warning, enabling this option may cause your system to slow down or crash completely!", "Autoriser le buffer du timeshift a dépasser la taille du disque. Attention, activer cette option peut rendre le systeme instable voir le planter définitivement!",
"Allow/Disallow access to the 'Configuration' tab.", "Accepter/Refuser l'accès à l'onglet 'Configuration'",
"Allow/Disallow web interface access (this  includes access to the EPG).", "Accepter/Refuser l'accès à l'interface web (cela inclus l'accès à l'EPG)",
"Allowed DVR profiles. This limits the profiles the user has access to.", "Profiles d'enregistrements autorisés. Cela limite les profiles qu'un utilisateur peut avoir.",
"Allowed network", "Réseau autorisé",
"Allowed networks", "Réseaux autorisés",
"Alternate PID path", "Répertoire de PID alternatif",
"Alternate configuration path", "Répertoire alternatif pour la config",
"Altitude (in meters).", "Altitude (en mètre)",
"Altitude (meters)", "Altitude (mètres)",
"Always send the whole DiseqC sequence including LNB setup (voltage, tone). If this is not checked, only changed settings are sent, which may cause issues with some drivers. If the tuning is not reliable, try activating this option.", "Toujours envoyer la séquence complète du DiseqC incluant le LNB de configuration (voltage, tonalité). Si non coché, seulement les changements sont envoyés, ce qui peut causer des problèmes avec certains drivers. Si le tuning n'est pas correct, essayer d'activer cette option.",
"Always use the name defined by the network.", "Toujours utiliser le nom défini par le réseau.",
"Always use the old ioctls to read the linuxdvb status (signal strength, SNR, error counters). Some drivers are not mature enough to provide the correct values using the new v5 linuxdvb API.", "Toujours utiliser l'ancien ioctls pour lire l’état de linuxdvb (force du signal, SNR, compteurs d'erreur). Certains drivers ne sont pas assez fiable pour fournir les bonnes valeurs avec l'API linuxdvb V5.",
"An event which starts between this \"start after\" and \"start before\" will be matched (including boundary values).", "Un événement démarrant entre \"démarre après\" et \" démarre avant\" doit être vérifié (incluant les valeurs limite).",
"Animals", "Animalier",
"Anonymize HTSP access", "Anonymiser les accès HTSP",
"Any", "Tous",
"Arts / Culture (without music)", "Arts / Culture (sans musique)",
"Arts magazines", "Magazines artistiques",
"Assign predefined muxes to networks", "Assigner les muxes prédéfinis aux réseaux",
"Associate this device with one or more networks.", "Associer cet appareils avec un ou plusieurs réseaux.",
"Athletics", "Athletisme",
"Audio Stream Filter", "Filtre de flux audio",
"Audio bitrate (kb/s) (0=auto)", "Débit Audio (kb/s) (0=Auto)",
"Audio channel layout.", "Disposition audio de la chaine.",
"Audio codec", "Codec audio",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Codec audio a utiliser pour le transcodage. \"Ne pas utiliser\" désactivera la sortie audio.",
"Auto check disabled", "Validation automatique désactivée",
"Auto check enabled", "Vérification automatique activée",
"Auto record", "Enregistrement automatique",
"Auto record caption", "Enregistrement automatique des sous-titres",
"Auto recording%s%s", "Auto enregistrement%s%s",
"Auto time record", "Enregistrement horaire automatique",
"Auto-Map to channels", "Association automatique des chaînes",
"Automatic checking", "Vérification automatique",
"Automatic recording caption.", "Enregistrement automatique des sous-titres.",
"Automatic schedule entry not found", "Entrée de l'enregistrement automatique non trouvée.",
"Automatic time scheduler entry not found", "Entrée de la programmation horaire automatique non trouvée",
"Automatically map EPG source", "Associer automatiquement la source EPG",
"Automatically map all available services to channels.", "Mappe automatiquement tous les services disponibles aux chaines. ",
"Automatically map channels defined within the bouquet.", "Mappe automatiquement les chaines définies dans les bouquets.",
"Automatically name from network", "Réseau automatique pour IPTV",
"Automatically record.", "Enregistrement automatique.",
"Autorec maximum count (0=unlimited)", "Nombre maximum d'enregistrements auto (0=sans limite)",
"Autorec maximum schedules limit (0=unlimited)", "Nombre maximum d'enregistrements programmés (0=sans limite)",
"Bad", "Mauvais",
"Bad request", "Mauvaise requête.",
"Bandwidth", "Bande passante",
"Based on software from", "Basé sur le logiciel de",
"Bind to specific local IP address.", "Attacher a une adresse IP locale spécifique.",
"Blue", "Bleu",
"Boot ID", "ID de démarrage",
"Brand", "Marque",
"Branding information (if available).", "Branding information (si disponible).",
"Broadcast", "Diffusion",
"Broadcast type", "Type de la diffusion.",
"Broadcast.", "Diffusion.",
"Broadcasting", "Diffusion",
"Buffering limit (ms)", "Temps maxi du buffer (ms)",
"Build", "Version",
"By default, linuxdvb's status read period is 1000ms (one second). The accepted range is 250ms to 8000ms. Note that for some hardware or drivers (like USB), the status operations take too much time and CPU. In this case, increase the default value. For fast hardware, this value might be decreased to make the decision of the re-tune algorithm based on the signal status faster.", "Par défaut, linuxdvb lit les informations de statut toutes les 1000 ms (1 seconde). Les valeurs doivent être comprise entre 250 ms et 8000 ms. À noter que pour certains matériels ou pilotes (comme ceux USB), ces lectures prennent trop de temps et de CPU. Dans ce cas, augmenter la valeur par défaut. Pour les matériels performants, cette valeur peut être diminuée pour permettre à l’algorithme de réglage du tuner, selon le statut du signal, d’être plus rapide. ",
"CA ID", "ID du CA ",
"CA client %i", "Client CA %i",
"CA identification", "Identification du CA",
"CA provider", "Fournisseur du CA",
"CAPMT interval (in ms).", "CAPMT intervalle (en ms).",
"CAPMT interval (ms)", "Intervalle CAPMT (ms)",
"CAPMT query interval (ms)", "Intervalle de recherche CAPMT (ms)",
"CAPMT query interval (ms).", "CAPMT intervalle de vérification (ms).",
"CI Module", "Module CI",
"CRID authority", "Autorité CRID",
"CSA (descrambling)", "CSA (décryptage)",
"Cache scheme", "Système de cache",
"Camd.socket filename / IP Address (TCP mode)", "Nom de fichier camd.socket / Adresse IP (mode TCP)",
"Cartoons", "Dessins animés",
"Change parameters", "Changer les paramètres",
"Channel", "Chaine",
"Channel Tags", "Étiquettes de la chaîne",
"Channel does not exist", "La chaine n'existe pas.",
"Channel icon", "Icône de la chaine",
"Channel icon URL.", "URL du logo de la chaine.",
"Channel icon as defined in EPG data.", "Logo de la chaine comme défini dans les données du guide des programmes.",
"Channel icon name scheme", "Plan de nommage du logo de la chaine.",
"Channel icon path", "Chemin de l'icône de la chaîne",
"Channel mapping options", "Options de mappage de la chaine",
"Channel name", "Nom de la chaine",
"Channel not found", "Chaine non trouvée",
"Channel number", "Numéro de la chaîne",
"Channel number as defined in EPG data.", "Numéro de la chaine comme défini dans les données du guide des programmes.",
"Channel number offset", "Décalage des numéros de chaînes",
"Channel numbers (heuristic)", "Numéros de la chaine (heuristique)",
"Channel numbers from", "Numéro de chaîne à partir de",
"Channel tag", "Etiquette de la chaine",
"Channel tag does not exist", "L’étiquette de la chaine n'existe pas",
"Channel tag reference", "Référence de l’étiquette de la chaine",
"Channel tags", "Etiquettes de chaine",
"Channel tags the user is allowed access to/excluded from.", "Etiquettes de la chaine que l'utilisateur est autorisé ou interdit d’accès.",
"Channel to use/used for the recording.", "Chaine a utiliser ou utilisée pour l'enregistrement.",
"Channel update options", "Options de mise à jour de la chaîne",
"Channels", "Chaines",
"Channels EPG data is used by.", "Données du guide des programmes sont utilisées par.",
"Character set", "Codage de caractères",
"Characters not supported in Windows filenames (e.g. for an SMB/CIFS share) will be stripped out or converted.", "Caractères sont supportés dans les noms de fichiers sous Windows (par exemple pour les partages SMB/CIFS) seront supprimés ou convertis.",
"Charset", "Jeu de caractères",
"Check availability", "Vérifier la disponibilité",
"Children's / Youth programs", "Programmes pour enfants",
"Class", "Classe",
"Classical", "Classique",
"Classical music", "Musique classique",
"Client name", "Nom du client",
"Clone scheduled entry on error", "Dupliquer l'entrée programmée en cas d'erreur",
"Comedy", "Comédie",
"Command delay time (ms) (10-200)", "Temps d'attente de commande (ms) (10-200)",
"Command time (ms) (10-100)", "Temps d'attente commande (ms) (10-100)",
"Comment", "Commentaire",
"Commercial break", "Coupure publicitaire",
"Committed", "Validé",
"Completed OK", "Complété avec succès",
"Compress EPG database", "Compresser la base de donnée du guide des programmes",
"Compress the EPG database to reduce disk I/O and space.", "Compresser la base de donnée du guide des programmes pour réduire l'espace disque et les I/O.",
"Configuration - Base", "Configuration - Basic",
"Configuration ID", "ID de Configuration ",
"Configuration name", "Nom de la config",
"Configuration version", "Version de la configuration ",
"Connection limit type", "Type de limite de connexion",
"Container", "Conteneur",
"Container to use for the transcoded stream.", "Conténaire a utiliser pour le flux transcodé.",
"Content character set", "Codage de caractères du contenu",
"Content type", "Type de contenu",
"Content type.", "Contenu type.",
"Contest", "Concours",
"Continue recording for x minutes after scheduled stop time", "Continuer a enregistrer pendant x minutes après l'heure d’arrêt programmée.",
"Continue recording for x minutes after scheduled stop time.", "Continuer a enregistrer pendant x minutes après l'heure d’arrêt programmée.",
"Cookie expiration (days)", "Expiration des cookies (jours)",
"Cooking", "Cuisine",
"Copy codec type", "Copier le type de codec",
"Copy layout", "Copier la disposition",
"Count of objects", "Recensement des objets",
"Create SDTV/HDTV/Radio tags.", "Créer les étiquettes SDTV/HDTV/Radio.",
"Create a bouquet from the playlist.", "Créer un bouquet d'une liste de lecture.",
"Create a directory per channel when storing recordings. If both this and the 'directory per day' checkbox is enabled, the date-directory will be the parent of the per-channel directory.", "Créer un répertoire par chaine pour les enregistrements. Si cette option et 'un répertoire par jour' sont cochées, le répertoire par date sera le parent de celui par chaine.",
"Create a directory per title when storing recordings. If the day/channel directory checkboxes are also enabled, those directories will be parents of this directory.", "Créer un répertoire par titre pour les enregistrements. Si l'option 'un répertoire par jour' ou 'un répertoire par chaine' sont aussi cochés, ils seront parents de ce répertoire.",
"Create a new directory per day in the recording system path. Folders will only be created when something is recorded. The format of the directory will be ISO standard YYYY-MM-DD.", "Créer un nouveau répertoire par jour pour les enregistrements. Les répertoires ne seront créés que s'il existe un enregistrement. Le format sera YYYY-MM-DD.",
"Create and associate a network tag to created channels.", "Créer et associer une étiquette de réseau pour créer les chaines.",
"Create and associate a provider tag to created channels.", "Créer et associer une étiquette de l’émetteur pour créer les chaines.",
"Create bouquet", "Créer un bouquet",
"Create bouquet tag", "Créer une étiquette pour les bouquets",
"Create directories using these permissions.", "Créer des répertoires utilisant ces permissions.",
"Create files using these permissions.", "Créer des fichiers utilisant ces permissions.",
"Create network name tags", "Créer une étiquette pour les noms des réseaux",
"Create network name tags (set by provider).", "Créer une étiquette pour les noms des réseaux (défini par l’émetteur).",
"Create tags", "Créer des étiquettes",
"Created", "Créé",
"Creator", "Auteur",
"Cron multi-line", "programmation multi ligne",
"Culture (without music)", "Culture (sans musique)",
"Culture magazines", "Magazines culturels",
"Current RTSP port.", "Port RTSP actuel.",
"Current affairs", "Affaire en cours",
"Current number of objects.", "Nombre d'objets actuel.",
"Current object size.", "Taille des objets actuels.",
"Current tuner configuration.", "Configuration du tuner actuel.",
"Custom HTTP headers", "En-têtes HTTP personnalisés",
"DBus - use the session message bus instead of the system one", "DBus - utiliser le message de session bus au lieu de celui du système",
"DES Key.", "Clef DES.",
"DES key", "Clef DES",
"DSCP/TOS for streaming", "DSCP/TOS pour les flux",
"DVB", "TNT",
"DVB EPG ID", "ID de l'EPG DVB",
"DVB multiplex", "Multiplexe DVB",
"DVB scan files", "Fichiers des scans DVB",
"DVB scan files path", "Répertoire des fichiers des scan DVB",
"DVB-C multiplex", "Multiplexe DVB-C",
"DVB-S multiplex", "Multiplexe DVB-S",
"DVB-T Network", "Réseau TNT",
"DVB-T multiplex", "Multiplex TNT",
"DVR", "Enregistrements",
"DVR behavior", "Comportement du magnétoscope",
"DVR configuration", "Configuration du magnétoscope",
"DVR configuration profiles", "Profils de configuration de l'enregistrement",
"DVR file retention period", "Période de conservation de l'enregistrement ",
"DVR log retention", "Durée de conservation des logs de l'enregistrement",
"DVR log retention period", "Période de conservation de l'enregistrement ",
"DVR override: high", "Enregistrement passer outre: Haut",
"DVR override: important", "DVR passer outre: Important",
"DVR override: low", "DVR passer outre: Bas",
"DVR override: normal", "DVR passer outre: Normal",
"DVR override: unimportant", "DVR passer outre: sans importance",
"Dance", "Danse",
"Data errors", "Erreurs de données",
"Data path (if applicable).", "Chemin des données (si applicable).",
"Days of Week", "Jour de la semaine",
"Days of the week to which the rule should apply.", "Jours de la semaine auxquelles les règles doivent s'appliquer.",
"Days of week", "Jour de la semaine",
"Debate", "Débat",
"Debug libav log", "Traces de debug libav",
"Debug options", "Options de debug",
"Debug subsystems", "Debug des sous-système",
"Debug to syslog", "Debug vers un syslog",
"Debug trace (low-level)", "Trace de debug (bas niveau)",
"Debugging", "Débogage",
"Default", "Défaut",
"Default language(s)", "Langue par défaut",
"Default language.", "Langue par défaut.",
"Default priority", "Priorité par défaut",
"Default user interface level.", "Niveau de l'interface utilisateur par défaut.",
"Define network group to limit network usage.", "Définir le groupe de réseau pour limiter l'utilisation du réseau.",
"Delivery system", "Système de livraison",
"Demux path", "Chemin pour le demux",
"Descramble services (limit per mux)", "Service de décryptage (max par multiplex)",
"Descrambler buffer (TS packets)", "Taille du Tampon du Débrouilleur (Paquets venant d'un flux de transport MPEG)",
"Detective", "Policier",
"Device ID", "ID d'appareil",
"Device model", "Modèle d'appareil",
"Device path", "Chemin du périphérique",
"Digital Video Recorder", "Caméscope Numérique",
"Directory", "Répertoire",
"Directory permissions (octal, e.g. 0775)", "Permissions du répertoire (octal, par ex. 0775)",
"Directory used by the entry.", "Répertoire utilisé par les entrées.",
"Disable", "Désactivé",
"Disable DVB bouquets", "Désactiver les bouquets DVB",
"Disable SAT>IP client", "Désactiver le client SAT>IP",
"Disable all access control checks", "Désactive tous les contrôles d’accès",
"Disable debug on stderr", "Désactiver le debug vers le stderr",
"Disable device/firmware-specific workarounds", "Désactive les palliatifs spécifiques au périphérique/firmware",
"Disable syslog (all messages)", "Désactive le syslog (Tous les messages)",
"Disabled", "Désactivé",
"Discover more muxes using the Network Information Table (if available).", "Découvrir plus de muxes en utilisant la Table des Information des Réseaux (si disponible).",
"DiseqC repeats", "Répétition du DiseqC",
"Do not send any stream specific information to the HTSP client like signal strength, input source etc.", "Ne pas envoyer d'information spécifique au client HTSP comme la force du signal, source d'entrée, ...",
"Do not use", "Ne pas utiliser",
"Documentary", "Documentaire",
"Don't backup configuration tree at upgrade", "Ne pas sauvegarder la configuration durant la mise à jour",
"Don't include the title in the filename.", "Ne pas inclure le titre dans le nom de fichier.",
"Don't include title in filename", "Ne pas inclure le titre dans le nom de fichier",
"Don't keep", "Ne pas garder",
"Don't re-record", "Ne pas ré-enregistrer",
"Don't re-record if recording fails.", "Ne pas ré-enregistrer si les enregistrements sont en échec.",
"Don't re-schedule if recording fails.", "Ne pas reprogrammer si les enregistrements sont en échec.",
"Don't reschedule", "Ne pas reprogrammer",
"Don't use the provider's channel numbers.", "Ne pas utiliser les numéro de chaine de l’émetteur.",
"Drama", "Drame",
"Duplicate handling", "Gestion des doublons",
"EIT - skip TSID check", "EIT - Ne pas vérifier le TSID",
"EPG Grabber", "Collecteur d'EPG",
"EPG Grabber Channel", "Chaîne du récupérateur d'EPG",
"EPG Grabber Configuration", "Configuration de la récupération d'EPG pour les chaînes",
"EPG Language (priority order)", "Langue de l'EPG (ordre de priorité)",
"EPG name", "Nom du Guide des programmes électronique",
"EPG scan", "Recherche de l'EPG",
"EPG source", "Source de l'EPG",
"EPG update window", "Fenêtre de mise a jour de l'EPG",
"Economics", "Economie",
"Educational", "Education",
"Elementary stream filter", "Filtre élémentaires des flux ",
"Enable (auto)", "Activer (auto)",
"Enable DBus", "Activer DBus",
"Enable NTP driver", "Activer les drivers NTP",
"Enable coredumps for daemon", "Activer les coredumps pour le démon",
"Enable debug on stderr", "Activer le debug vers le stderr",
"Enable debug subsystems", "Activer le debug du sous-système",
"Enable debug to file", "Activer le debug vers un fichier",
"Enable debug to syslog", "Activer le debug vers syslog",
"Enable or disable this configuration.", "Activer ou désactiver cette configuration",
"Enable syslog", "Activer le syslog",
"Enable this filter.", "Activer ce filtre.",
"Enable trace subsystems", "Activer les traces du sous-système",
"Enable web UI debug (non-minified JS)", "Activer le debug webUI (JS non minifié)",
"Enable/disable EPG data for the entry.", "Activer/désactiver les données du guide des programmes pour cette entrée.",
"Enable/disable LNA.", "Activer/désactiver LNA.",
"Enable/disable auto-rec rule.", "Activer/désactiver les règles d'auto enregistrements.",
"Enable/disable debugging output to syslog.", "Activer/désactiver la sortie du débug vers un syslog.",
"Enable/disable libav log output.", "Activer/désactiver la sortie des log de labav.",
"Enable/disable profile.", "Activer/désactiver le profile.",
"Enable/disable the bouquet.", "Activer/désactiver le bouquet.",
"Enable/disable the channel.", "Activer/désactiver la chaine.",
"Enable/disable the device.", "Activer/désactiver l'appareil.",
"Enable/disable the entry.", "Activer/désactiver l'entrée.",
"Enable/disable the tag.", "Activer/désactiver l’étiquette.",
"Enable/disable tuner/adapter.", "Activer/désactiver le tuner/l'adapter.",
"Enabled", "Activé",
"Encrypted", "Encrypté",
"Encrypted service", "Service encrypté",
"Enter a non-admin user password.", "Entrer un mot de passe non admin.",
"Enter a non-admin user username.", "Entrer un utilisateur non admin.",
"Enter allowed network prefix(es). You can enter a comma-seperated list of prefixes here.", "Entrer un préfixe de réseau autorisé. Vous pouvez utiliser une liste séparée par une virgule.",
"Enter an administrator password.", "Entrer un mot de passe admin.",
"Entertainment programs for 10 to 16", "Programmes éducatifs de 10 a 16 ans",
"Entertainment programs for 6 to 14", "Programmes éducatifs de 6 a 14 ans",
"Environment", "Environnement",
"Environment (pipe)", "Environnement (pipe)",
"Episode number/ID.", "Numéro d’épisode/ID.",
"Error code", "Code d'erreur",
"Errors", "Erreurs",
"Every day", "Chaque jours",
"Expeditions", "Expéditions",
"Experimental film", "Films expérimental",
"Exported tuners", "Tuners exportés",
"External", "Externe",
"External EPG grabber", "Récupérateur externe d'EPG",
"External IP (NAT)", "Adresse IP externe (NAT)",
"External URL", "URL externe",
"Extra arguments", "Arguments supplémentaires",
"Extra start time", "Temps supplémentaire avant le début",
"Extra stop time", "Temps supplémentaire après la fin",
"FAIL", "En erreur",
"FEC high", "FEC superieur",
"FEC low", "FEC bas",
"Factual topics", "Sujets d'actualité",
"Failed", "Echoué",
"Fantasy", "Fantastique",
"Fashion", "Mode",
"File missing", "Fichier manquant",
"File not created", "Fichier non créé",
"File permissions (octal, e.g. 0664)", "Permissions du fichier (octal, par ex. 0664)",
"File size", "Taille du fichier",
"Filename", "Nom du fichier",
"Filename character set", "Type de caractères pour les noms de fichiers",
"Filename options", "Options du nom de fichier",
"Fine arts", "Beaux-arts",
"Finished", "Terminé",
"Fitness and health", "Fitness et santé",
"Folkloric", "Folklorique",
"Force CA ID (e.g. 0x2600)", "Forcer l'ID du CA (par ex. 0x2600)",
"Force old status", "Forcer l'état précédent",
"Force pilot for DVB-S2", "Forcer le pilote pour le DVB-S2",
"Force priority", "Forcer la priorité",
"Force teardown delay", "Délais pour forcer la fin de communication",
"Foreign countries", "Pays étrangers",
"Forever", "Pour toujours",
"Fork and run as daemon", "Dupliquer le processus en tant que service",
"Format string", "Format de chaîne de caractères",
"Frequency", "Fréquence",
"Frequency (Hz)", "Fréquence (Hz)",
"Frequency (MHz)", "Fréquence (MHz)",
"Frequency (kHz)", "Fréquence (kHz)",
"Fri", "Ven",
"Friendly name", "Nom commun",
"Frontend number", "Numéro du frontal",
"Frontend path", "Répertoire du frontal",
"Full pathname specification", "Spécification du chemin de fichier complet",
"Full-text", "Mode texte",
"Further education", "Formation continue",
"Game show", "Exposition de jeux",
"Gardening", "Jardinage",
"General", "Général",
"General configuration", "Configuration Générale",
"Generic", "Générique",
"Generic options", "Options générique",
"Guard interval", "Intervalle de protection",
"HD: high definition", "HD: Haute Définition",
"HTSP Default Stream Settings", "Paramètre par défaut du flux  HTSP",
"Handicraft", "Artisanat",
"Hearing impaired", "Malentendant",
"Hierarchy", "Hiérarchie",
"High", "Haut",
"Historical movie", "Film d’époque",
"Horror", "Horreur",
"Hostname/IP", "Nom d’hôte/adresse IP",
"IDLE", "En veille",
"IP address", "Adresse IP",
"IP blocking", "Blocage des IP",
"IPTV Multiplex", "Multiplex de l'IPTV",
"IPTV Network", "Réseau IPTV",
"IPTV Subcription", "Souscription IPTV",
"ISI (Stream ID)", "ISI (ID du flux)",
"Icon", "Icône",
"Icon (full URL)", "Icône (URL complète)",
"Icon URL", "URL de l’icône",
"Icon base URL", "URL de base de l’icône",
"Icon has title", "L'icône possède un titre",
"Icons from", "Icône venant de",
"Idle", "En attente",
"Idle scan", "Scan en veille",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Si aucun compte d'utilisateur n'existe, alors en créer un sans\nnom d'utilisateur ni mot de passe. A utiliser avec précaution\npuisque qu'il donne tous les accès en admin à votre\ninstallation TVHeadEnd tant que vous n'avez pas créé/édité ses\ndroits via la console web.",
"If you'd like to support the project, please consider a donation.", "Si vous voulez aider ce projet, pensez a faire un don.",
"Ignore EPG (EIT)", "Ignorer EPG (EIT)",
"Ignore descriptor 5", "Ignorer le descripteur 5",
"Ignore invalid SSL certificate", "Ignorer le certificat SSL invalide",
"Immediately abort", "Annuler immédiatement",
"Include channel name in filename", "Inclure le nom de la chaine dans le nom du fichier",
"Include date in filename", "Inclure la date dans le nom du fichier",
"Include subtitle in filename", "Inclure les sous-titres dans le nom du fichier",
"Include time in filename", "Inclure l'heure dans le nom du fichier",
"Information area", "Zone d'information",
"Informational", "Information",
"Initial scan", "Scan initial",
"Internal", "Interne",
"Invalid", "Invalide",
"Invalid adapter number '%s'\n", "Numéro d'adaptateur invalide '%s'\n",
"Invalid target", "Destination invalide",
"Keep", "Garder",
"Kill signal (pipe)", "Signal de Fermeture (pipe)",
"Kill timeout (pipe/secs)", "Durée avant Fermeture (pipe/secs)",
"Language", "Langage",
"Language settings", "Réglage de la langue",
"Languages", "Langages",
"Last updated from", "Dernière mise à jour par",
"Leisure hobbies", "Loisir et passe-temps",
"Listen / Connect port", "Port d'écoute/connexion",
"Listen on IPv6", "Utiliser IPv6",
"Literature", "Littérature",
"Local bind IP address", "Liaison locale d'adresse IP",
"Location", "Lieu",
"Low", "Bas",
"MPEG-TS Pass-thru", "MPEG-TS Pass-through",
"MPEG-TS Pass-thru/built-in", "MPEG-TS Pass-through/Interne",
"Make subdirectories per channel", "Créer un sous-répertoire par chaine",
"Make subdirectories per day", "Créer un sous-répertoire par jour",
"Make subdirectories per title", "Créer un sous-répertoire par titre",
"Manufacturer", "Fabriquant",
"Manufacturer URL", "URL du fabriquant",
"Map all services", "Mapper tous les services",
"Matroska (mkv)/built-in", "Matroska (mkv)/par défaut",
"Maximal channel number", "Nombre maximal de chaînes",
"Maximum PIDs", "PID Maximum",
"Maximum RAM size (MB)", "Taille max de la RAM (MB)",
"Maximum bandwidth (Kbps)", "Bande passante max (Kbps)",
"Maximum count (0=default)", "Nombre maximum (0=sans limite)",
"Maximum duration", "Durée maximum",
"Maximum length of PIDs", "Longueur maximum des PIDs",
"Maximum period (mins)", "Période maxi (min)",
"Maximum size (MB)", "Taille max (MB)",
"Maximum timeout (seconds)", "Délais max (secondes)",
"Medicine", "Médecine",
"Melodrama", "Mélodrame",
"Merge same name", "Fusionner les noms identiques",
"Minimal channel number", "Nombre minimal de chaînes",
"Minimum duration", "Durée minimum",
"Missing In PAT/SDT", "Manquant dans PAT/SDT",
"Model description", "Description du modèle",
"Model name", "Nom du modèle",
"Model number", "Numéro du modèle",
"Mon", "Lun",
"More verbose libav log", "Plus de log pour libav",
"Motor sport", "Sports motorisés",
"Motoring", "Automobiles",
"Movie", "Film",
"Music", "Musique",
"Mux UUID", "UUID du Mux",
"Mux not enabled", "Mux non activé",
"NONE", "Aucun",
"Name", "Nom",
"Names", "Noms",
"Natural sciences", "Sciences naturelles ",
"Network", "Réseau",
"Network A", "Réseau A",
"Network B", "Réseau B",
"Network ID (limit scanning)", "ID du réseau (limitation du scan)",
"Network UUID", "UUID du réseau",
"Network access", "Accès réseau",
"Network discovery", "Découverte du réseau",
"Network name", "Nom du réseau",
"Network prefix", "Préfixe du réseau",
"Network settings", "Paramètres du réseau",
"Network type", "Type du réseau",
"Networks", "Réseaux",
"News", "Journal",
"News magazine", "Journal",
"Next tune delay in ms (0-2000)", "Délai de changement de fréquence, en ms (0-2000)",
"No", "Non",
"No PIN", "Pas de PIN",
"No access", "Aucun accès",
"No adapters specified!\n", "Aucun adaptateur spécifié\n",
"No assigned adapters", "Pas d'adaptateurs attribués",
"No days", "Aucun jour",
"No descrambler", "Pas de décrypteur",
"No free adapter", "Pas d'adaptateur disponible",
"No input detected", "Pas d'entrée détectée",
"No service assigned to channel", "Aucun service assigné a cette chaine",
"No service enabled", "Pas de service activé",
"No source available", "Pas de source disponible",
"None", "Aucun",
"Not set", "Non défini",
"Not set (none or channel configuration)", "non défini (rien ou config de chaine)",
"Not set (use channel or DVR configuration)", "Non défini (utiliser la configuration de la chaine ou de l'enregistreur)",
"Number", "Numéro",
"Number of tsfile tuners", "Nombre de tuners tsfile",
"OK (partial)", "OK (partiel)",
"Older OSCam", "Ancien OSCam",
"Only Bulsatcom 39E", "Bulstatcom 39E uniquement",
"Only EIT", "Seulement EIT",
"Only OpenTV Sky Ausat", "OpenTV Sky Ausat uniquement",
"Only OpenTV Sky Italia", "OpenTV Sky Italia uniquement",
"Only OpenTV Sky UK", "OpenTV Sky UK uniquement",
"Only PSIP (ATSC)", "Seulement PSIP (ATSC)",
"Only UK Freesat", "UK Freesat uniquement",
"Only UK Freeview", "UK Freeview uniquement",
"Only Viasat Baltic", "Viasat Baltic uniquement",
"Opera", "Opéra",
"Output directory for tsdebug", "Dossier de sortie pour tsdebug",
"Over-the-air", "hertzien",
"Over-the-air Cron multi-line", "Programmation multi ligne hertzien",
"Over-the-air EPG", "EPG de la diffusion",
"Owner", "Propriétaire",
"PEND", "En attente",
"PIDs 21 in setup", "PIDs 21 dans la configuration",
"PLP ID", "ID du PLP",
"PROXY protocol is an extension for support incoming TCP connections from a remote server (like a firewall) sending the original IP address of the client. The HTTP header 'X-Forwarded-For' do the same with HTTP connections. Both enable tunneled connections.This option should be disabled for standard usage.", "Le protocole PROXY est une extension pour la prise en charge de connexions TCP entrantes en provenance d'un serveur distant (comme un pare-feu) envoyant l'adresse IP originale du client. L'en-tête HTTP 'X-Forwarded-For' fait la même chose avec les connexions HTTP. Tous deux activent les connexions par tunnel. Cette option doit être désactivée en utilisation standard.",
"Parent entry", "Entrée parente",
"Parent entry.", "Entrée parente.",
"Password", "Mot de passe",
"Password2", "Mot de passe 2",
"Path", "Répertoire",
"Path to an icon for this channel. This can be named however you wish, as either a local (file://) or remote (http://) image. See Help for more infomation.", "Chemin vers l’icône pour cette chaine. Peut être nommé comme vous le souhaitez, être une image locale (file://) ou distante (http://). Consulter l'aide pour plus d'information",
"Performing arts", "Arts du spectacle",
"Physiology", "Physiologie",
"Pilot", "Pilote",
"Polarization", "Polarisation",
"Political issues", "Questions politiques",
"Popular culture", "Culture populaire ",
"Power save", "Economie d’énergie",
"Prefer picons over channel name", "Choisir les picons au lieu des noms de chaine",
"Preferred CA PID", "CA favoris du PID",
"Presentation", "Présentation",
"Press", "Presses",
"Priority", "Priorité",
"Priority (-10..10)", "Priorité (-10..10)",
"Private", "Privé",
"Profile name", "Nom du profil",
"Provider ID", "ID du provider",
"Psychology", "Psychologie",
"Puppets", "Marionnette",
"RTSP port", "Port RTSP",
"Re-fetch period (hours)", "Période avant d'aller re-chercher (heures)",
"Re-fetch period (mins)", "Période avant d'aller re-chercher (minutes)",
"Re-record%s%s", "Ré-enregistrer%s%s",
"Re-try period (hours)", "Période pour réessayer (heures)",
"Record all", "Enregistrer tout",
"Reject", "Rejeter",
"Religious", "Religieux",
"Remarkable people", "Célébrités",
"Remove HTTP arguments", "Supprimer les paramètres HTTP",
"Replace whitespace in title with '-'", "Remplacer les espace dans le titre avec '-'",
"Reports", "Rapports",
"Rerun of", "Rediffusion",
"Reserved", "Réservé",
"Resolution (height)", "Résolution (hauteur)",
"Respawn (pipe)", "Recommencer (pipe)",
"Rewrite EIT", "Réécriture EIT",
"Rewrite PAT", "Réécriture PAT",
"Rewrite PMT", "Réécriture PMT",
"Rewrite SDT", "Réécriture SDT",
"Rolloff", "Transfert",
"Run as group", "En tant que groupe",
"Run as user", "En tant qu'utilisateur",
"Running", "En cours",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "Numéro du port RTSP pour le serveur SAT>IP\n(Par défaut: -1 = désactive, 0 = webconfig, le port standard est 554)",
"SSL verify peer", "Vérifications des paires du tunnel SSL ",
"Sat", "Sam",
"Satconf", "Config du satellite",
"Schedule a re-recording if more errors than (0=off)", "Programme un ré-enregistrement si plus de (0=aucun)",
"Scheduled for recording", "Programmé pour l'enregistrement",
"Season", "Saison",
"Season %d", "Saison %d",
"Send full PLAY cmd", "Envois commande complète PLAY",
"Serial number", "Numéro de série",
"Serious", "Grave",
"Serious music", "Musique",
"Server", "Serveur",
"Server UUID", "Serveur UUID",
"Service ID", "ID du service",
"Settings", "Paramètres",
"Show", "Spectacles",
"Show this page", "Montre cette page",
"Show version information", "Information de version",
"Signal quality too poor", "Signal de trop faible qualité",
"Signal scale (240 or 100)", "Echelle du signal (240 ou 100)",
"Soap", "Feuilleton",
"Soccer", "Football",
"Social advisory", "Conseil social",
"Source deleted", "Source effacée",
"Source quality is bad", "La qualité de la source est mauvaise",
"Source reconfigured", "Source reconfigurée",
"Special events (Olympic Games, World Cup, etc.)", "Évènements spéciaux (Jeux Olympiques, Coupe du Monde, etc.)",
"Specify User-Agent header for the http client", "Spécifier un User-Agent pour le client http",
"Specify alternative htsp port", "Spécifier un port htsp alternatif",
"Specify alternative http port", "Spécifier un port http alternatif",
"Specify alternative http webroot", "Spécifier un répertoire web racine alternatif",
"Specify bind address", "Spécifier une adresse de liaison",
"Specify extra htsp port", "Spécifier un port supplémentaire pour le HTSP",
"Spiritual sciences", "Sciences Spirituelle",
"Sports magazines", "Magazines sportifs",
"Start", "Démarrer",
"Start after", "Démarrer après",
"Start before", "Démarrer avant",
"Start time", "Heure de démarrage",
"Status", "Etat",
"Stereo", "Stéréo",
"Stop time", "Heure de fin",
"Storage space", "Espace de stockage",
"Streaming", "Envoi du flux",
"Subscribe to a service permanently", "S'inscrire en permanence à un service",
"Subscription overridden", "Inscription ignorée",
"Subtitle", "Sous-titre",
"Sun", "Dim",
"Sync + Don't keep", "Sync + Ne pas garder",
"System", "Système",
"Tags", "Etiquette",
"Talk show", "Débats",
"Team sports (excluding football)", "Sports collectifs (sauf football)",
"Technology", "Technologique",
"Testing", "En test",
"Testing options", "Options de test",
"Thu", "Jeu",
"Time", "Horaire",
"Time missed", "Horaire manqué",
"Time recording%s%s", "Temps d'enregistrement%s%s",
"Timeout (sec) (0=infinite)", "Délais (sec) (0=infini)",
"Timeout (seconds)", "Délais (secondes)",
"Title", "Titre",
"Toggle details", "inverser les détails",
"Tourism / Travel", "Tourisme / Voyage",
"Trace subsystems", "Traces des sous-systèmes",
"Traditional arts", "Arts traditionnel ",
"Traditional music", "Musique classique",
"Transponder ID", "ID du transpondeur",
"Tue", "Mar",
"Tuner bind IP address", "Liaison tuner d'adresse IP",
"Tuning failed", "Erreur du tuner",
"Turn off LNB when idle", "Arrêter le LNB quand il est au repos",
"Tvheadend server name", "Nom du serveur Tvheadend",
"URL with the SAT>IP server XML location", "URL avec l'emplacement XML du serveur SAT>IP",
"Uncommitted", "Dévalidé",
"Unimportant", "Sans importance",
"Universal LNB only", "Seulement le NLB universel",
"Unknown", "Inconnu",
"Unknown file", "Fichier inconnu",
"Unknown reason (%i)", "Raison inconnue (%i)",
"Unlimited size", "Taille illimitée",
"Unlimited time", "Temps illimité",
"Unset (default)", "Non défini (défaut)",
"Update channel icon", "Mise à jour de l'icônes des chaines",
"Update channel name", "Mise à jour du nom des chaines",
"Update channel number", "Mise à jour du numéro des chaines",
"Update time", "Mise à jour de l'heure",
"Update tolerance (ms)", "Mise à jour de la tolérance (ms)",
"Usage: %s [OPTIONS]\n", "Utilisation : %s [OPTIONS]\n",
"Use Windows-compatible filenames", "Utiliser des nom de fichiers compatible avec Windows",
"Use XSPF playlist instead of M3U", "Utiliser des listes XSPF au lieu de M3U",
"Use original", "Utiliser l'original",
"User access error", "Erreur d’accès pour l'utilisateur",
"User language", "Langue de l'utilisateur",
"User limit reached", "Limite de l'utilisateur atteinte",
"User login", "Identifiant d'utilisateur",
"User request", "Requête utilisateur",
"Username", "Nom de l'utilisateur",
"Variety show", "Spectacle de variété",
"Verify the SSL certificate.", "Vérifier le certificat SSL",
"Video", "Vidéo",
"Video codec", "Codec vidéo",
"View all", "Voir tout",
"Visually impaired commentary/audio description", "Sous-titrage sourds et malentendants",
"WEBM/built-in", "WEBM/Interne",
"Waiting for program start", "En attente du démarrage du programme",
"Waiting for stream", "En attente du flux",
"War", "Guerre",
"Water sport", "Sports aquatique",
"Weather report", "Météo",
"Web interface", "Interface web",
"Web interface language", "Langue de l'interface web",
"Web interface language.", "Langue de l'interface web.",
"Web interface theme.", "Thème de l'interface web.",
"Web theme", "Thème web",
"Wed", "Mer",
"Yes", "Oui",
"addpids/delpids supported", "addpids/delpids supportés",
"invalid option specified [%s]", "Option spécifiée invalide [%s]",
"option %s requires a value", "Une valeur est obligatoire pour l'option %s",
"tsfile input (mux file)", "Entrée tsfile (fichier mux)",
NULL, NULL
};

static const char *tvh_locale_16[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nלמידע נוסף בקר באתר Tvheadend:\nhttps://tvheadend.org\n",
"# Muxes", "# מוקסים",
"# Services", "# שירותים",
"1 day", "יום 1",
"ACTIVE", "פעיל",
"AUTO", "אוטומטי",
"Aborted by user", "נסגר ע\"י המשתמש",
"Access", "גישה",
"Action", "פעולה",
"Adapter in use by another subscription", "המתאם בשימוש ע\"י תהליך אחר",
"Add file and line numbers to debug", "הוסף קובץ ומספרי שורות לניפוי שגיאות",
"Add the thread ID to debug", "הוסף מזהה של הליך משנה לניפוי שגיאות",
"Admin", "מנהל",
"Admin password", "סיסמת מנהל",
"Admin username", "שם משתמש מנהל",
"Adventure", "הרפתקאות",
"Advertisement / Shopping", "פרסומת/ קניות",
"Animals", "בעלי חיים",
"Any", "הכל",
"Arts", "אמנויות",
"Athletics", "אתלטיקה",
"Audio codec", "מקודד אודיו",
"Auto", "אוטומטי",
"Auto record", "הקלטה אוטומטית",
"Ballet", "בלט",
"Bandwidth", "רוחב פס",
"Based on software from", "מבוסס על תוכנה מ",
"Bouquet", "סל ערוצים",
"Bouquet (auto)", "סל ערוצים (אוטומטי)",
"Cartoons", "מצויר",
"Channel", "ערוץ",
"Channel does not exist", "ערוץ לא קיים",
"Channel icon", "סמל ערוץ",
"Channel name", "שם ערוץ",
"Channel not found", "ערוץ לא נמצא",
"Channel number", "מספר ערוץ",
"Channels", "ערוצים",
"Cinema", "קולנוע",
"Comedy", "קומדיה",
"Comment", "הערה",
"Commercial break", "הפסקת פרסומות",
"Completed OK", "הסתיים בהצלחה",
"Configuration", "הגדרות",
"Configuration version", "גירסת הגדרה",
"Cooking", "בישול",
"Copy codec type", "העתק סוג מקודד",
"Created", "נוצר",
"Creator", "יוצר",
"Culture (without music)", "תרבות (ללא מוסיקה)",
"DVR", "הקלטה",
"Days of Week", "ימים בשבוע",
"Debug to syslog", "ניפוי שגיאות ליומן מערכת",
"Debugging", "ניפוי שגיאות",
"Default", "ברירת מחדל",
"Default language.", "שפת ברירת מחדל",
"Description", "תיאור",
"Detective", "בלשי",
"Device ID", "מזהה מכשיר",
"Directory", "תיקיה",
"Disable", "נטרול",
"Disable SAT>IP client", "בטל קליינט SAT>IP",
"Discussion", "דיון",
"Do not use", "אל תשתמש",
"Documentary", "דוקומנטרי",
"Don't backup configuration tree at upgrade", "אל תגבה הגדרות בשדרוג",
"Don't keep", "אל תשמור",
"Drama", "דרמה",
"Economics", "כלכלה",
"Education", "חינוך",
"Educational", "חינוכי",
"Enable (auto)", "פעיל (אוטומטי)",
"Enable debug to file", "הפעל ניפוי שגיאות לקובץ",
"Enabled", "פעיל",
"Encrypted", "מוצפן",
"Encrypted service", "שירות מוצפן",
"Environment", "סביבה",
"Environment (pipe)", "סביבה (צינור)",
"Episode", "פרק",
"Episode %d", "פרק %d",
"Errors", "שגיאות",
"Every day", "כל יום",
"FAIL", "נכשל",
"Fantasy", "פנטזיה",
"Filename", "שם קובץ",
"Film", "סרט",
"Fitness and health", "כושר ובריאות",
"Folk", "עממי",
"Football", "כדורגל",
"Fork and run as daemon", "הרץ כתהליך רקע",
"Frequency", "תדירות",
"Frequency (Hz)", "תדירות (Hz)",
"Frequency (MHz)", "תדירות (MHz)",
"Frequency (kHz)", "תדירות (kHz)",
"Fri", "שישי",
"Full DiseqC", "DiseqC מלא",
"Full-text", "טקסט מלא",
"Game show", "שעשועון",
"Gardening", "גינון",
"General", "כללי",
"Generic", "כללי",
"Hierarchy", "היררכיה",
"High", "גבוה",
"Horror", "אימה",
"Hostname/IP", "שם מארח/IP",
"ID", "מזהה",
"IDLE", "סרק",
"Icon", "סמל",
"Icon (full URL)", "סמל (כתובת מלאה)",
"Icon URL", "כתובת סמל",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "אם לא קיים חשבון משתמש צור אחד\nללא שם משתמש וללא סיסמה. השתמש בתבונה\nכיוון שכל אחד בעולם יקבל גישה \nלהתקנה של Tvheadend  עד שתיצור או תערוך\nאת הגדרות הגישה בממשק הניהול בדפדפן.",
"If you'd like to support the project, please consider a donation.", "אם ברצונך לתמוך בפרויקט זה, אנא שקול לתרום",
"Ignore invalid SSL certificate", "התעלם מתעודת SSL לא תקינה",
"Immediately abort", "סגירה מיידית",
"Important", "חשוב",
"Index", "אינדקס",
"Interface", "ממשק",
"Internal", "פנימי",
"Interview", "ראיון",
"Invalid", "לא תקין",
"Jazz", "ג'אז",
"Language", "שפה",
"Languages", "שפות",
"Last updated from", "עדכון אחרון מ:",
"Literature", "ספרות",
"Location", "מיקום",
"Log", "יומן",
"Login/Logout", "התחבר/התנתק",
"Low", "נמוך",
"Magazines", "מגזינים",
"Manufacturer", "יצרן",
"Manufacturer URL", "כתובת יצרן",
"Matroska (mkv)/built-in", "Matroska (mkv)/מובנה",
"Medicine", "רפואה",
"Mode", "מצב",
"Modulation", "מודולציה",
"Mon", "שני",
"Mono", "מונו",
"Motor sport", "ספורט מוטורי",
"Motoring", "מוטורי",
"Movie", "סרט",
"Music", "מוזיקה",
"Musical", "מוזיקלי",
"Mux", "מוקס",
"Mux not enabled", "מוקס מנוטרל",
"NONE", "ללא",
"Name", "שם",
"Nature", "טבע",
"Network", "רשת",
"Networks", "רשתות",
"News", "חדשות",
"No days", "אין ימים",
"No free adapter", "אין מתאם פנוי",
"No input detected", "לא זוהה מתאם",
"No source available", "אין מקור זמין",
"None", "ללא",
"Normal", "רגיל",
"Not set", "לא מוגדר",
"Number", "מספר",
"OK", "אישור",
"Off", "כבוי",
"Older OSCam", "OSCam ישן",
"On", "פעיל",
"Only EIT", "EIT בלבד",
"Only UK Freesat", "UK Freesat בלבד",
"Only UK Freeview", "UK Freeview בלבד",
"Opera", "אופרה",
"Owner", "בעלים",
"Password", "סיסמה",
"Password2", "סיסמה 2",
"Path", "נתיב",
"Pilot", "פיילוט",
"Pop", "פםפ",
"Port", "פורט",
"Position", "מיקום",
"Priority", "קדימות",
"Priority (-10..10)", "עדיפות (10...10-)",
"Private", "פרטי",
"Provider", "ספק",
"Puppets", "בובות",
"Quiz", "חידון",
"Record all", "הקלט הכל",
"Religion", "דת",
"Religious", "דתי",
"Reports", "דוחות",
"Resolution (height)", "רזולוציה (גובה)",
"Rock", "רוק",
"Romance", "רומנטי",
"Running", "רץ",
"Sat", "שבת",
"Scheduled for recording", "תזמון הקלטה",
"Science", "מדע",
"Science fiction", "מדע בדיוני",
"Season", "עונה",
"Season %d", "עונה %d",
"Server", "שרת",
"Service", "שירות",
"Services", "שירותים",
"Settings", "הגדרות",
"Signal quality too poor", "איכות אות גרועה",
"Soap", "אופרת סבון",
"Soccer", "כדורגל",
"Social", "חברתי",
"Source", "מקור",
"Source deleted", "מקור נמחק",
"Specify alternative htsp port", "הגדר פורט htsp חלופי",
"Specify alternative http port", "הגדר פורט http חלופי",
"Specify extra htsp port", "הגדר פורט htsp נוסף",
"Sports", "ספורט",
"Sports magazines", "מגזין ספורט",
"Standard", "רגיל",
"Start", "התחלה",
"Status", "מצב",
"Stereo", "סטיראו",
"Stop", "עצירה",
"Storage space", "שטח אחסון",
"Streaming", "מזרים",
"Subtitle", "כתובית",
"Sun", "ראשון",
"Sync", "סנכרון",
"System", "מערכת",
"Tags", "תגיות",
"Technology", "טכנולוגיה",
"Tennis", "טניס",
"Testing options", "בודק אפשרויות",
"The name of the profile.", "שם הפרופיל",
"This tuner", "מקלט זה",
"Thriller", "מתח",
"Thu", "חמישי",
"Title", "כותרת",
"Tue", "שלישי",
"Tuner", "מקלט",
"Tuner configuration", "הגדרות מקלט",
"Tuner configuration.", "הגדרות מקלט.",
"Tuning failed", "כוונון נכשל",
"Tvheadend server name", "שם שרת Tvheadend",
"Type", "סוג",
"URL", "כתובת",
"Unable to open file", "לא יכול לפתוח קובץ",
"Unknown", "לא ידוע",
"Unknown file", "קובץ לא ידוע",
"Unlimited size", "גודל לא מוגבל",
"Unlimited time", "זמן לא מוגבל",
"Update channel icon", "עדכן סמל ערוץ",
"Update channel name", "עדכן שם ערוץ",
"Update channel number", "עדכן מספר ערוץ",
"Update time", "עדכן זמן",
"User does not have access", "למשתמש אין גישה",
"User does not have access to channel", "למשתמש אין גישה לערוץ",
"User language", "שפת משתמש",
"Username", "שם משתמש",
"Verify the SSL certificate.", "ודא תעודת SSL",
"Video", "וידאו",
"Video codec", "מקודד וידאו",
"Waiting for program start", "ממתין לתחילת התוכנית",
"War", "מלחמה",
"Water sport", "ספורט מים",
"Weather report", "תחזית מזג אוויר",
"Wed", "רביעי",
"Western", "מערבון",
"hrs", "שע'",
"min", "דקה",
"mins", "דק'",
NULL, NULL
};

static const char *tvh_locale_17[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nZa više informacija posjetite Tvheadend web stranicu:\nhttps://tvheadend.org\n",
"# Mapped channels", "# Mapirani programi",
"# Muxes", "# Multipleksi",
"# Services", "# Usluge",
"(Default profile)", "(Zadani profil)",
"1 day", "1 dan",
"1 month", "1 mjesec",
"1 week", "1 tjedan",
"1 year", "1 godinu",
"2 months", "2 mjeseca",
"2 weeks", "2 tjedna",
"2 years", "2 godine",
"3 days", "3 dana",
"3 months", "3 mjeseca",
"3 weeks", "3 tjedna",
"3 years", "3 godine",
"5 days", "5 dana",
"6 months", "6 mjeseca",
"A channel tag (e.g. a group of channels) to which this rule applies.", "Oznaka programa (npr. grupa programa) na koje se ovo pravilo odnosi.",
"AAC audio", "AAC zvučni zapis",
"AC-3 = descriptor 6", "AC-3 = opisivač 6",
"AC-3 detection", "AC-3 otkrivanje",
"AC3 audio", "AC3 zvučni zapis",
"ACTIVE", "AKTIVNO",
"ANY", "SVEJEDNO",
"ATSC SI Tables", "ATSC SI Tablice",
"ATSC source ID", "ATSC ID izvora",
"ATSC-C Network", "ATSC-C mreža",
"ATSC-C multiplex", "ATSC-C multipleks",
"ATSC-T Network", "ATSC-T mreža",
"ATSC-T multiplex", "ATSC-T multipleks",
"AUTO", "AUTOMATSKI",
"Aborted by user", "Prekinuo korisnik",
"Access", "Pristupačna",
"Access (ACL)", "Pristupačna (ACL)",
"Action", "Radnja",
"Active", "Aktivno",
"Adapter in use by another subscription", "Adapter koristi druga pretplata",
"Add file and line numbers to debug", "Dodaj datoteku i broj redka za otklanjanje grešaka",
"Add the thread ID to debug", "Dodaj ID stavku za otklanjanje grešaka",
"Add this value to PTS for the teletext subtitles. The time value is in milliseconds and may be negative.", "Dodaj ovu vrijednost u PTS za podnaslove teleteksta. Vrijednost vremena je u milisekundama i može biti negativna.",
"Admin", "Administrator",
"Admin password", "Lozinka administratora",
"Admin username", "Korisničko ime administratora",
"Administrator login", "Prijava administratora",
"Adult movie", "Film za odrasle",
"Advanced", "Napredno",
"Advanced (non-universal LNBs, rotors, etc.)", "Napredno (neuniverzalni LNB-ovi, rotori, itd.)",
"Adventure", "Pustolovina",
"Advertisement / Shopping", "Oglašavanje / Kupovina",
"After creating the network scan it for services.", "Nakon stvaranja mreže, pretraži usluge.",
"All (Streaming plus DVR)", "Sve (Strujanje plus DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Svi znakovi koji bi mogli stvarati probleme biti će zamijenjeni podcrtama. Pogledajte pomoć za pojedinsoti.",
"All lower-case", "Sva slova mala",
"All proceeds are used to support server infrastructure and buy test equipment.", "Svi postupci su korišteni za podršku infrastrukture poslužitelja i kupovinu opreme za testiranje.",
"Allowed networks", "Dopuštene mreže",
"Alternate PID path", "Zamjenska PID putanja",
"Alternate configuration path", "Zamjenska putanja podešavanja",
"Altitude (in meters).", "Nadmorska visina (u metrima).",
"Altitude (meters)", "Nadmorska visina (m)",
"Animals", "Životinje",
"Any", "Svejedno",
"Arts", "Umjetnost",
"Arts / Culture (without music)", "Umjetnost / Kultura (bez glazbe)",
"Arts magazines", "Emisije o umjetnosti",
"Athletics", "Atletika",
"Authenticated user", "Ovjereni korisnik",
"Auto", "Automatski",
"Auto check disabled", "Automatska provjera onemogućena",
"Auto check enabled", "Automatska provjera omogućena",
"Auto recording%s%s", "Automatsko snimanje%s%s",
"Auto-Map to channels", "Automatski mapiraj u programe",
"Automatic checking", "Automatska provjera",
"Automatically record.", "Automatski snimi.",
"Ballet", "Balet",
"Bandwidth", "Frekvencijska širina",
"Based on software from", "Temeljeno na softveru",
"Blue", "Plava",
"Bouquet", "Buket",
"Bouquet (auto)", "Buket (automatski)",
"Bouquet source.", "Izvor buketa.",
"Bouquets", "Buketi",
"Broadcast", "Emitiranje",
"Broadcast type", "Vrsta emitiranja",
"Broadcast.", "Emitiranje.",
"Broadcasting", "Emitiranje",
"Buffering limit (ms)", "Ograničenje međuspremanja (ms)",
"Build", "Izgrađeno",
"Cache scheme", "Shema predmemorije",
"Cartoons", "Crtići",
"Change parameters", "Promijeni parametre",
"Channel", "Program",
"Channel Tags", "Oznake programa",
"Channel does not exist", "Program ne postoji",
"Channel icon", "Ikona programa",
"Channel icon URL.", "URL ikone programa.",
"Channel icon as defined in EPG data.", "Ikona programa koja je određena u EPG podacima.",
"Channel icon name scheme", "Naziv sheme ikone programa",
"Channel icon path", "Putanja ikone programa",
"Channel mapping options", "Mogućnosti mapiranja programa",
"Channel name", "Naziv programa",
"Channel not found", "Program nije pronađen",
"Channel number", "Broj programa",
"Channel number as defined in EPG data.", "Broj programa koji je određen u EPG podacima.",
"Channel number offset", "Pmak broja programa",
"Channel number range", "Raspon broja programa",
"Channel numbers (heuristic)", "Broj programa (heuristički)",
"Channel numbers from", "Broj programa iz",
"Channel tag", "Oznaka programa",
"Channel tag does not exist", "Oznaka programa ne postoji",
"Channel tag reference", "Napomena oznake programa",
"Channel tags", "Oznake programa",
"Channel to use/used for the recording.", "Program koji se koristi za snimanje.",
"Channel update options", "Mogućnosti nadopune programa",
"Channels", "Programi",
"Character set", "Kôdnu stranicu",
"Check availability", "Provjeri dostupnost",
"Children's / Youth Programs", "Program za djecu / mlade",
"Children's / Youth programs", "Program za djecu / mlade",
"Cinema", "Kino",
"Class", "Klasika",
"Classical", "Klasično",
"Classical music", "Klasična glazba",
"Clean effects", "Jasni efekti",
"Client name", "Naziv klijenta",
"Clone scheduled entry on error", "Kloniraj zakazano snimanje pri grešci",
"Comedy", "Komedija",
"Command delay time (ms) (10-200)", "Vrijeme odgode nearedbe (ms) (10-200)",
"Command time (ms) (10-100)", "Vrijeme naredbe (ms) (10-100)",
"Comment", "Komentar",
"Commercial break", "Reklame",
"Commercials will be dropped from the recordings. Commercial detection works using EITp/f (EPG running state) and for the Swedish channel TV4 (using teletext info).", "Reklame će biti uklonjene iz snimaka. Otkrivanje reklama se izvodi upotrebom EITp/f (EPG mora biti pokrenut) i za švedski program TV4 (upotrebom informacija iz teleteksta).",
"Completed OK", "Uspješno završeno",
"Compress EPG database", "Sažmi EPG bazu podataka",
"Compress the EPG database to reduce disk I/O and space.", "Sažmi EPG bazu podataka za smanjivanje I/O diska i prostora.",
"Conditional Access", "Uvjetovani pristup",
"Conditional Access Client", "Uvjetovani pristup klijenta",
"Conditional Access Identification.", "Identifikacija uvjetovanog pristupa.",
"Configuration", "Podešavanje",
"Configuration - Base", "Podešavanje - Osnovno",
"Configuration - Image Cache", "Podešavanje - Predmemorija slike",
"Configuration - SAT>IP Server", "Podešavanje - SAT>IP poslužitelj",
"Configuration ID", "ID podešavanja",
"Configuration name", "Naziv podešavanja",
"Configuration version", "Inačica podešavanja",
"Connection limit type", "Vrsta ograničenja povezivanja",
"Connection limits", "Ograničena povezivanja",
"Constellation", "Konstelacija",
"Container", "Spremnik",
"Content character set", "Vrsta znakova",
"Content type", "Vrsta sadržaja",
"Content type.", "Vrsta sadržaja.",
"Contest", "Natjecanje",
"Cookie expiration (days)", "Istek kolačića (u danima)",
"Cooking", "Kuhanje",
"Copy codec type", "Kopiraj vrstu kôdeka",
"Copy layout", "Izlaz kopiranja",
"Count of objects", "Broj objekata",
"Create SDTV/HDTV/Radio tags.", "Stvori SDTV/HDTV/Radio oznake.",
"Create a bouquet from the playlist.", "Stvori buket iz popisa izvođenja.",
"Create a provider name tag.", "Stvori oznaku pružatelja usluge.",
"Create bouquet", "Stvori buket",
"Create bouquet tag", "Stvori oznaku buketa",
"Create tags", "Stvori oznake",
"Created", "Stvoreno",
"Creator", "Stvorio",
"Cron multi-line", "Cron višestruki redak",
"Culture (without music)", "Kultura (bez glazbe)",
"Current affairs", "Aktualni događaji",
"Current tuner configuration.", "Trenutno podešavanje prijemnika.",
"Custom HTTP headers", "Prilagođena HTTP zaglavlja",
"DAB Network", "DAB mreža",
"DAB multiplex", "DAB multipleks",
"DBus - use the session message bus instead of the system one", "DBus - koristi sabirnicu poruka sesije umjesto sustava",
"DES Key.", "DES ključ.",
"DES key", "DES ključ",
"DSCP/TOS for streaming", "DSCP/TOS za strujanje",
"DVB CAM Client", "DVB CAM klijent",
"DVB CSA (descrambling) Tables", "DVB CSA (dekôdiranje) tablice",
"DVB EPG Tables", "DVB EPG tablice",
"DVB Inputs - Networks", "DVB ulazi - Mreže",
"DVB SI Tables", "DVB SI tablice",
"DVB Time Tables", "DVB vremenske tablice",
"DVB multiplex", "DVB multipleks",
"DVB scan files", "DVB datoteke pretraživanja",
"DVB scan files path", "DVB putanje datoteka pretraživanja",
"DVB-C Network", "DVB-C mreža",
"DVB-C multiplex", "DVB-C multipleks",
"DVB-S Network", "DVB-S mreža",
"DVB-S Satellite Configuration", "DVB-S Satelitsko podešavanje",
"DVB-S multiplex", "DVB-S multipleks",
"DVB-T Network", "DVB-T mreža",
"DVB-T multiplex", "DVB-T multipleks",
"DVR - Profiles", "DVR - Profili",
"DVR behavior", "DVR ponašanje",
"DVR configuration", "DVR podešavanje",
"DVR configuration profiles", "DVR profili podešavanja",
"DVR configurations", "DVR podešavanja",
"DVR entry not found", "DVR stavka nije pronađena",
"DVR file retention period", "Razdoblje zadržavanja DVR datoteke",
"DVR log retention", "Zadržavanje DVR zapisa",
"DVR log retention period", "Razdoblje zadržavanja DVR zapisa",
"DVR override: high", "DVR zaobilaženje: visok",
"DVR override: important", "DVR zaobilaženje: značajan",
"DVR override: low", "DVR zaobilaženje: nizak",
"DVR override: normal", "DVR zaobilaženje: normalan",
"DVR override: unimportant", "DVR zaobilaženje: nebitan",
"Dance", "Ples",
"Data errors", "Greške podataka",
"Data path (if applicable).", "Putanja podatka (ako je primjenjivo).",
"Days of Week", "Dani u tjednu",
"Days of the week to which the rule should apply.", "Dani u tjednu na koje se pravilo odnosi.",
"Days of week", "Dani u tjednu",
"Debate", "Rasprava",
"Debug libav log", "libav zapis otklanjanja grešaka",
"Debug options", "Mogućnosti otklanjanja greške",
"Debug subsystems", "Otklanjanje grešaka podsustava",
"Debug to syslog", "Otklanjaj greške u podsustav",
"Debug trace (low-level)", "Praćenje otklanjanja grešaka (niža-razina)",
"Debugging", "Otklanjanje greške",
"Default", "Zadano",
"Default language(s)", "Zadani jezici",
"Default language.", "Zadani jezik.",
"Default login", "Zadana prijava",
"Default priority", "Zadani prioritet",
"Default user interface level.", "Zadana razina korisničkog sučelja.",
"Define network group to limit network usage.", "Odredi mrežnu grupu za ograničenje upotrebe mreže.",
"Delivery system", "Sustav isporuke",
"Demux path", "Putanja demultipleksa",
"Descramble services (limit per mux)", "Dekôdirane usluge (ograničenje po multipleksu)",
"Descrambler", "Dekôder",
"Descrambler EMM", "EMM dekôder",
"Descrambler buffer (TS packets)", "Međuspremnik dekôdiranja (TS paketi)",
"Descrambling timeout (ms)", "Ograničenje dekôdiranja (ms)",
"Description", "Opis",
"Detective", "Dektetivski",
"Device ID", "ID uređaja",
"Device model", "Model uređaja",
"Device path", "Putanja uređaja",
"Digital Video Recorder", "Digitalni video snimač",
"Directory", "Direktorij",
"Disable", "Onemogući",
"Disable DVB bouquets", "Onemogući DVB bukete",
"Disable SAT>IP client", "Onemogući SAT>IP klijenta",
"Disable X_SATIPM3U tag", "Onemogući X_SATIPM3U oznaku",
"Disable all access control checks", "Onemogućei sve provjere upravljanja pristupom",
"Disable debug on stderr", "Onemogući otklanjanje greške na stderr",
"Disable device/firmware-specific workarounds", "Onemogući određeno obilaženje uređaja/frimvera",
"Disable syslog (all messages)", "Onemogući zapis sustava (sve poruke)",
"Disabled", "Onemogućeno",
"Discover more muxes using the Network Information Table (if available).", "Otkrij više multipleksa koristeći tablicu informacije mreže (ako je dostupna).",
"Discussion", "Rasprava",
"DiseqC repeats", "DiseqC ponavljanja",
"Do not use", "Ne koristi",
"Documentary", "Dokumentarac",
"Don't backup configuration tree at upgrade", "Sigurnosno ne kopiraj stablo podešavanja pri nadogradnji",
"Don't include the title in the filename.", "Ne uključuj naslov u naziv datoteke.",
"Don't include title in filename", "Ne uključuj naslov u naziv datoteke",
"Don't keep", "Ne čuvaj",
"Don't re-record", "Ne snimaj ponovno",
"Don't re-record if recording fails.", "Ne snimaj ponovno ako je snimanje neuspjelo",
"EIT - skip TSID check", "EIT - preskoči TSID provjeru",
"EPG Database", "EPG baza podataka",
"EPG Grabber", "EPG sakupljač",
"EPG Grabber Channel", "EPG sakupljač programa",
"EPG Grabber Configuration", "EPG sakupljač podešavanje",
"EPG Language (priority order)", "EPG jezik (poredak prioriteta)",
"EPG data ID.", "EPG ID podatka.",
"EPG name", "EPG naziv",
"EPG overlap cut", "EPG izrezivanje preklapanja",
"EPG scan", "EPG pretraživanje",
"EPG scan time-out in seconds (30-7200)", "Trajanje EPG pretraživanja u sekundama (30-7200)",
"EPG source", "EPG izvor",
"EPG update window", "EPG prozor nadopune",
"Economics", "Ekonomija",
"Education", "Obrazovanje",
"Education / Science / Factual topics", "Obrazovanje / Znanost / Stvarne teme",
"Educational", "Obrazovno",
"Electronic Program Guide", "Elektronski programski vodič",
"Enable (auto)", "Omogući (automatski)",
"Enable DBus", "Omogući DBus",
"Enable NTP driver", "Omogući NTP upravljački program",
"Enable debug on stderr", "Omogući otklanjanje greške na stderr",
"Enable debug subsystems", "Omogući otklanjanje greške podsustava",
"Enable debug to file", "Omogući otklanjanje greške u datoteku",
"Enable debug to syslog", "Omogući otklanjanje greške u zapis sustava",
"Enable trace subsystems", "Omogući praćenje podsustava",
"Enabled", "Omogućeno",
"Encrypted", "Šifrirano",
"Encrypted service", "Šifrirane usluge",
"Environment", "Okoliš",
"Environment (pipe)", "Okoliš (slivnik)",
"Episode", "Epizoda",
"Episode %d", "Epizoda %d",
"Episode number/ID.", "Broj epizode/ID.",
"Error code", "Kôd greške",
"Error code of entry.", "Kôd greške stavke.",
"Errors", "Greške",
"Even key", "Paran ključ",
"Even key.", "Paran ključ.",
"Event does not exist", "Događaj ne postoji",
"Every day", "Svaki dan",
"Exclude channel tags", "Isključi oznake programa",
"Expeditions", "Ekspedicije",
"Experimental film", "Eksperimentalan film",
"Expert", "Stručna",
"External", "Vanjski",
"External EPG grabber", "Vanjski EPG sakupljač",
"External IP (NAT)", "Vanjski IP (NAT)",
"External PyEPG Grabber", "Unutrašnji PyEPG sakupljač",
"External URL", "Vanjski URL",
"External URL of the bouquet.", "Vanjski URL buketa.",
"External XMLTV EPG Grabber", "Vanjski XMLTV EPG sakupljač",
"Extra arguments", "Dodatni argumenti",
"Extra start time", "Dodatno vrijeme početka",
"Extra stop time", "Dodatno vrijeme završetka",
"FAIL", "NEUSPJELO",
"FEC high", "FEC visoki",
"FEC low", "FEC niski",
"Factual topics", "Stvarne teme",
"Failed", "Neuspjelo",
"Failed to open image", "Neuspjelo otvaranje slike",
"Failed view", "Neuspjeli prikaz",
"Fantasy", "Fantazija",
"Fashion", "Moda",
"Fast input switch", "Brzo prebacivanje ulaza",
"Fastscan DVB", "Brzo DVB pretraživanje",
"File missing", "Datoteka nedostaje",
"File not created", "Datoteka nije stvorena",
"File permissions (octal, e.g. 0664)", "Dozvole datoteke (oktalno, npr. 0664)",
"File removed", "Datoteka je uklonjena",
"File size", "Veličina datoteke",
"Filename", "Naziv datoteke",
"Filename (including path)", "Naziv datoteke (uključujući putanju)",
"Filename options", "Mogućnosti naziva datoteke",
"Filename used by the entry.", "Naziv datoteke koju koristi stavka.",
"Filesystem monitor", "Nadgledanje datotečnog sustava",
"Fine arts", "Likovne umjetnosti",
"Finished", "Završeno",
"First word", "Prva riječ",
"Fit to RAM (cut rewind)", "Prilagodi po RAM-u (ukloni premotavanje)",
"Fitness and health", "Fitnes i zdravlje",
"Folkloric", "Folklor",
"Football", "Nogomet",
"Force (auto)", "Prisili (automatski)",
"Force CA ID (e.g. 0x2600)", "Prisili CA ID (npr. 0x2600)",
"Force an initial EPG grab at start-up.", "Prisili pokretanje EPG pretraživanja pri pokretanju",
"Force initial EPG grab at start-up", "Prisili pokretanje EPG pretraživanja pri pokretanju",
"Force old status", "Prisili staro stanje",
"Force pilot for DVB-S2", "Prisili pilota za DVB-S2",
"Force priority", "Prisili prioritet",
"Force profile to use this priority.", "Prisili profil da koristi ovaj prioritet.",
"Force service type to 1", "Prisili vrstu usluge na 1",
"Force teardown delay", "Prisli odgodu rastavljanja",
"Foreign countries", "Strane zemlje",
"Forever", "Zauvijek",
"Fork and run as daemon", "Račvaj i pokreni kao pozadinski program (daemon)",
"Frequency", "Frekvencija",
"Frequency (Hz)", "Frekvencija (Hz)",
"Frequency (MHz)", "Frekvencija (MHz)",
"Frequency (kHz)", "Frekvencija (kHz)",
"Fri", "Pet",
"Full DiseqC", "Puni DiseqC",
"Full-text", "Cijeli tekst",
"Further education", "Stručno usavršavanje",
"Game show", "Kviz",
"Gardening", "Vrtlarstvo",
"General", "Općenito",
"Generic", "Osnovno",
"Gray", "Siva",
"HD: high definition", "HD: Visoka razlučivost",
"HTSP Default Stream Settings", "HTSP zadane postavke strujanja",
"Handicraft", "Radinosti",
"Hearing impaired", "Osobe oštećenog sluha",
"Hierarchy", "Hijerarhija",
"High", "Visok",
"Historical movie", "Povijesni film",
"Horror", "Horor",
"Hostname/IP", "Naziv računala/IP",
"IDLE", "MIROVANJE",
"IPTV Multiplex", "IPTV multipleks",
"ISI (Stream ID)", "ISI (ID strujanja)",
"Icon", "Ikona",
"Icon (full URL)", "Ikona (puni URL)",
"Icon URL", "URL ikone",
"Icon base URL", "Ikona onovni URL",
"Icon has title", "Ikona ima naziv",
"Icons from", "Ikone su s",
"If an error occurs clone the scheduled entry and try to record again (if possible).", "Ako se pojavi greška kloniraj ovo zakazano snimanje i probaj snimiti ponovno (ako je moguće).",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Ako nema korisničkih računa, tada stvorite jedan\nbez korisničkog imena i lozinke. Ovo koristite oprezno\njer će omogućiti administrativan pristup vašoj \nTvheadend instalaciji sve dok ne stvorite ili uredite\npravila pristupa iz Tvheadend web sučelja.",
"If you'd like to support the project, please consider a donation.", "Ako želite podržati projekt, razmislite o donaciji.",
"Ignore EPG (EIT)", "Zanemari EPG (EIT)",
"Ignore descriptor 5", "Zanemari opisivač 5",
"Ignore invalid SSL certificate", "Zanemari neispravne SSL vjerodajnice",
"Immediately abort", "Odmah zaustavi",
"Important", "Značajan",
"Index", "Raspored",
"Information area", "Informativno područje",
"Informational", "Informativni",
"Interface", "Sučelje",
"Internal", "Unutrašnji",
"Internal EPG grabber", "Unutrašnji EPG sakupljač",
"Internal error", "Unutrašnja greška",
"Internal grabber", "Unutrašnji sakupljač",
"Interview", "Intervju",
"Invalid", "Neispravno",
"Invalid adapter number '%s'\n", "Neispravan broj adaptera '%s'\n",
"Invalid target", "Neispravna meta",
"Keep", "Zadrži",
"Kill signal (pipe)", "Ubij signal (cjevovod)",
"Kill timeout (pipe/secs)", "Trajanje ubijanja (cjevovod/sek)",
"Language", "Jezik",
"Language 1", "Jezik 1",
"Language 2", "Jezik 2",
"Language 3", "Jezik 3",
"Language settings", "Postavke jezika",
"Languages", "Jezici",
"Last updated from", "Posljednje nadopuna iz",
"Leisure hobbies", "Hobiji",
"Listen on IPv6", "Osluškuj na IPv6",
"Literature", "Književnost",
"Local bind IP address", "Lokalno povezivanje IP adresa",
"Location", "Lokacija",
"Log", "Zapis",
"Login/Logout", "Prijava/Odjava",
"Low", "Nizak",
"Magazines", "Magazini",
"Maintained space", "Određeno prostorom",
"Manufacturer", "Proizvođač",
"Manufacturer URL", "URL proizvođača",
"Map unnamed channels", "Mapiraj programe bez naziva",
"Matroska (mkv)/built-in", "Matroska (mkv)/ugrađeno",
"Maximum PIDs", "Najviše PID-ova",
"Maximum duration", "Najveće trajanje",
"Maximum length of PIDs", "Najveće trajanje PID-ova",
"Maximum timeout (seconds)", "Najveće razdoblje čekanja (sekunde)",
"Medicine", "Medicina",
"Minimum duration", "Najmanje trajanje",
"Missing In PAT/SDT", "Nedostaje u PAT/SDT",
"Mode", "Način",
"Modulation", "Modulacija",
"Module", "Modul",
"Mon", "Pon",
"More verbose libav log", "Opširniji libav zapis",
"Motor sport", "Motorni sport",
"Motoring", "Nadgledanje",
"Movie", "Film",
"Music", "Glazba",
"Musical", "Mjuzikl",
"Mux", "Multipleks",
"Mux UUID", "Multipleks UUID",
"Mux not enabled", "Multipleks nije omogućen",
"NONE", "NIJEDAN",
"Name", "Naziv",
"Natural sciences", "Prirodne znanosti",
"Nature", "Priroda",
"Network", "Mreža",
"Network A", "Mreža A",
"Network B", "Mreža B",
"Network ID (limit scanning)", "ID mreže (ograniči pretraživanje)",
"Network UUID", "UUID mreže",
"Network prefix", "Predbroj mreže",
"Networks", "Mreže",
"News", "Vijesti",
"News magazine", "Dnevnik",
"Next tune delay in ms (0-2000)", "Odgoda do sljedećeg podešavanja u ms (0-2000)",
"No PIN", "Bez PINA",
"No access", "Bez pristupa",
"No adapters specified!\n", "Nema određenog adapters!\n",
"No days", "Nema dana",
"No descrambler", "Nema dekôdiranja",
"No free adapter", "Nema slobodnog adaptera",
"No input detected", "Nema otkrivenih ulaza",
"No scheme", "Nema sheme",
"No service assigned to channel", "Nema pridruženih usluga programu",
"No service enabled", "Nema omogućenih usluga",
"No source available", "Nema dostupnog izvora",
"None", "Nijedno",
"Normal", "Normalan",
"Not set", "Nije postavljeno",
"Not set (none or channel configuration)", "Nije postavljeno (nepoznato podešavanje programa)",
"Not set (use channel or DVR configuration)", "Nije postavljeno (koristi DVR podešavanje ili podešavanje programa)",
"Number", "Broj",
"Number of tsfile tuners", "Broj ts-datoteka prijemnika",
"OK", "U REDU",
"OK (partial)", "U redu (djelomično)",
"OSCam (rev >= 9095)", "OSCam (izdanje >= 9095)",
"OSCam TCP (rev >= 9574)", "OSCam TCP (izdanje >= 9574)",
"OSCam net protocol (rev >= 10389)", "OSCam mrežni protokol (izdanje >= 10389)",
"OSCam pc-nodmx (rev >= 9756)", "OSCam pc-nodmx (izdanje >= 9756)",
"Off", "Isključeno",
"Older OSCam", "Stariji OSCam",
"On", "Uključeno",
"On file removal", "Pri uklanjanju datoteke",
"Only Bulsatcom 39E", "Samo Bulsatcom 39E",
"Only EIT", "Samo EIT",
"Only OpenTV Sky Ausat", "Samo OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Samo OpenTV Sky Italia",
"Only OpenTV Sky UK", "Samo OpenTV Sky UK",
"Only PSIP (ATSC)", "Samo PSIP (ATSC)",
"Only UK Freesat", "Samo UK Freesat",
"Only UK Freeview", "Samo UK Freeview",
"Only Viasat Baltic", "Samo Viasat Baltic",
"Only digits", "Samo znamenke",
"Output directory for tsdebug", "Izlazni direktorij za ts otklanjnje greške",
"Over-the-air", "Iz signala",
"Over-the-air Cron multi-line", "Iz signala Cron višestruki redak",
"Over-the-air EPG", "EPG iz signala",
"Over-the-air EPG grabber", "EPG sakupljač iz signala",
"Over-the-air grabbers", "Sakupljači iz signala",
"Owner", "Vlasnik",
"PEND", "ČEKANJE",
"PIDs 21 in setup", "PID 21 u postavkama",
"Password", "Lozinka",
"Password2", "Lozinka2",
"Path", "Putanja",
"Performing arts", "Primjenjenja umjetnost",
"Periodically save EPG to disk (hours)", "Povremeno spremi EPG na disk (sati)",
"Physiology", "Filozofija",
"Polarization", "Polarizacija",
"Political issues", "Politička pitanja",
"Popular culture", "Popularna kultura",
"Port", "Ulaz",
"Position", "Položaj",
"Post-recording padding", "Dodatno vrijeme nakon završetka zakazanog snimanja",
"Pre-recording padding", "Dodatno vrijeme prije početka zakazanog snimanja",
"Preferred CA PID", "Preferirani CA PID",
"Preferred service video type", "Željena vrsta video usluge",
"Presentation", "Prezentacija",
"Press", "Tisak",
"Priority", "Prioritet",
"Priority (-10..10)", "Prioritet (-10..10)",
"Private", "Privatno",
"Provider", "Pružatelj",
"Provider ID", "ID pružatelja",
"Psychology", "Psihologija",
"Puppets", "Lutke",
"Quiz", "Kviz",
"Re-fetch period (hours)", "Razdoblje ponovnog preuzimanja (sati)",
"Re-try period (hours)", "Razdoblje ponovnog pokušaja (sati)",
"Record all", "Snimi sve",
"Recording file size.", "Veličina datoteke snimanja.",
"Reject", "Odbij",
"Religion", "Religija",
"Religious", "Religijski",
"Remarkable people", "Neuobičajeni ljudi",
"Remove HTTP arguments", "Ukloni HTTP argumente",
"Resolution (height)", "Razlučivost (visina)",
"Romance", "Romantika",
"Run as group", "Pokreni kao grupu",
"Run as user", "Pokreni kao korisnika",
"Running", "Pokrenuto",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "SAT>IP RTSP broj ulaza poslužitelja\n(zadano: -1 = onemogućeno, 0 = web podešavanje, standard ulaz je 554)",
"SD: standard definition", "SD: Standardna razlučivost",
"Sat", "Sub",
"Satconf", "Sat-podešavanje",
"Schedule a re-recording if more errors than (0=off)", "Zakaži ponovno snimanje ako nastane više greška od (0=isključeno)",
"Scheduled Duration", "Trajanje zakazanog snimanja",
"Scheduled for recording", "Zakazano za snimanje",
"Scheduled start time", "Zakazano vrijeme početka snimanja",
"Scheduled stop time", "Zakazano vrijeme završetka snimanja",
"Science", "Znanost",
"Science fiction", "Znanstvena fantastika",
"Season", "Sezona",
"Season %d", "Sezona %d",
"Send full PLAY cmd", "Pošalji punu naredbu REPRODUKCIJE",
"Serious", "Bitno",
"Serious music", "Ozbiljna glazba",
"Server", "Poslužitelj",
"Server UUID", "UUID poslužitelja",
"Service", "Usluga",
"Service ID", "ID usluge",
"Services", "Usluge",
"Settings", "Postavke",
"Show", "Emisija",
"Show this page", "Prikaži ovu informaciju",
"Show version information", "Prikaži informaciju inačice",
"Signal quality too poor", "Kvaliteta usluge loša",
"Skip TS packets (0-200)", "Preskoči TS pakete (0-200)",
"Soap", "Sapunica",
"Soccer", "Nogomet",
"Social", "Socijalno",
"Social advisory", "Socijalni savjeti",
"Source", "Izvor",
"Source deleted", "Izvor obrisan",
"Source quality is bad", "Kvaliteta izvora je loša",
"Source reconfigured", "Izvor je ponovno podešen",
"Special events (Olympic Games, World Cup, etc.)", "Posebni događaji (Olimpijske igre, Svjetsko prvenstvo, itd.)",
"Specify extra htsp port", "Odredi dodatni htsp ulaz",
"Spiritual sciences", "Duhovne znanosti",
"Sports", "Sportovi",
"Sports magazines", "Sportske emisije",
"Squash", "Skvoš",
"Standard", "Standarno",
"Start", "Pokreni",
"Start recording earlier than the defined Start recording earlier than the defined start time by x minutes: for example, if a program is to start at 13:00 and you set a padding of 5 minutes it will start recording at 12:54:30 (including a warm-up time of 30 seconds). If this isn't specified, any pre-recording padding as set in the channel or DVR entry will be used.", "Pokreni snimanje ranije x minuta nego što je određeno vremenom početka, na primjer ako program počinje u 13:00 i postavite dodatno vrijeme na 5 minuta snimanje će početi u 12:54:30 (uključujući vrijeme pripreme 30 skundi). Ako nije određeno bilo kakvim dodatnim vremenom prije početka snimanja, koristit će se ono postavljeno u DVR stavci ili DVR profilu.",
"Start time", "Vrijeme početka",
"Status", "Stanje",
"Stop", "Zaustavi",
"Stop time", "Vrijeme završetka",
"Storage space", "Prostor pohrane",
"Stream index", "Sadržaj strujanja",
"Stream type", "Vrsta strujanja",
"Streaming", "Strujanje",
"Subscribe to a service permanently", "Pretplati se na uslugu trajno",
"Subtitle", "Podnaslov",
"Sun", "Ned",
"Switch: %s", "Prebaci: %s",
"Sync", "Uskladi",
"Sync + Don't keep", "Uskladi + Ne zadrži",
"System", "Sustav",
"Tags", "Oznake",
"Talk show", "Emisija uživo",
"Team sports (excluding football)", "Ekipni sportovi (osim nogometa)",
"Technology", "Tehnologija",
"Teletext stream type is only available for this filter.", "Vrsta teletekst strujanja je dostupna samo za ovaj filter.",
"Tennis", "Tenis",
"Testing options", "Mogućnost testiranja",
"The CA stream type is only available for this filter.", "Vrsta CA strujanja je dostupna samo za ovaj filter.",
"The MPEGTS stream type is only available for this filter.", "Vrsta MPEGTS strujanja je dostupna samo za ovaj filter.",
"The total scheduled duration.", "Ukupno trajanje zakazanog snimanja.",
"Theme", "Tema",
"Thriller", "Triler",
"Thu", "Čet",
"Time", "Vrijeme",
"Time missed", "Vrijeme propušteno",
"Time recording%s%s", "Vrijeme snimanja%s%s",
"Timeout (sec) (0=infinite)", "Trajanje (sek) (0=brskonačno)",
"Timeout (seconds)", "Trajanje (sekunde)",
"Timeshift", "Premotavanje",
"Title", "Naslov",
"Toggle details", "Prikaži pojedinosti",
"Too many data errors", "Previše grešaka podataka",
"Tourism / Travel", "Turizam / Putovanja",
"Trace subsystems", "Prati podsustave",
"Traditional arts", "Tradicionalana umjetnost",
"Traditional music", "Tradicionalana glazba",
"Tue", "Uto",
"Type", "Vrsta",
"Unimportant", "Nebitan",
"Unknown", "Nepoznato",
"Unset (default)", "Neodabran (zadano)",
"Update channel icon", "Nadopuni ikonu programa",
"Update channel name", "Nadopuni naziv programa",
"Update channel number", "Nadopuni broj programa",
"User language", "Jezik korisnika",
"Username", "Korisničko ime",
"Variety show", "Posebna emisija",
"Waiting for stream", "Čekanje strujanja",
"War", "Rat",
"Water sport", "Vodeni sport",
"Weather report", "Vremenska prognoza",
"Web interface language", "Jezik web sučelja",
"Web interface language.", "Jezik web sučelja.",
"Web interface theme.", "Tema web sučelja.",
"Web theme", "Web tema",
"Wed", "Sri",
"Western", "Vesten",
"Wizard", "Čarobnjak",
"Writes the current in-memory EPG database to disk every x hours, so should a crash/unexpected shutdown occur EPG data is saved periodically to the database (re-read on next startup). Set to 0 to disable.", "Zapisuje trenutnu EPG bazu podataka iz memorije na disk svakih x sati, stoga može nastati rušenje/neočekivano isključivanje kada se s EPG podaci povremeno spremaju u bazu podataka (ponovno pročitaj pri sljedećem pokretanju). Postavite na 0 za onemogućavanje.",
"hrs", "sata",
"invalid option specified [%s]", "određena neispravna mogućnost [%s]",
"mins", "minuta",
NULL, NULL
};

static const char *tvh_locale_18[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nTovábbi információért kérjük látogasson el a Tvheadend honlapjára:\nhttps://tvheadend.org\n",
"# Mapped channels", " #Hozzárendelt csatornák",
"# Muxes", "# Multiplexek",
"# Services", "# Szolgáltatások",
"# tune repeats", "#DiseqC ismétlések száma",
"(Default profile)", "(Alapértelmezett profil)",
"(Read-only)", "(csak olvasható)",
"1 day", "1 nap",
"1 month", "1 hónap",
"1 week", "1 hét",
"1 year", "1 év",
"1316 bytes", "1316 byte",
"16356 bytes", "16356 byte",
"2 months", "2 hónap",
"2 weeks", "2 hét",
"2 years", "2 év",
"2-Port switch (universal LNB)", "2 portos Disecq kapcsoló (Univerzális LNB)",
"2632 bytes", "2632 byte",
"3 days", "3 nap",
"3 months", "3 hónap",
"3 weeks", "3 hónap",
"3 years", "3 év",
"32712 bytes", "32712 byte",
"4-Port switch (universal LNB)", "4 portos Disecq kapcsoló (Univerzális LNB)",
"5 days", "5 nap",
"5264 bytes", "5264 byte",
"6 months", "6 hónap",
"65424 bytes", "65424 byte",
"7896 bytes", "7896 byte",
"A channel tag (e.g. a group of channels) to which this rule applies.", "A csatorna cimke (p.l. egy csatorna csoport) amire ez a szabály érvényes",
"AAC audio", "AAC audió",
"AC-3 detection", "AC-3 felismerés",
"AC3 audio", "AC3 audió",
"ACTIVE", "AKTÍV",
"ATSC SI Tables", "ATSC SI táblák",
"ATSC source ID", "ATSC forrás ID",
"ATSC-C Network", "ATSC-C hálózat",
"ATSC-T Network", "ATSC-T hálózat",
"Aborted by user", "Felhasználó által megszakítva",
"Accept remote subscription weight", "Távoli feliratkozás súlyának elfogadása",
"Accept the remote subscription weight (from the SAT>IP client).", "Távoli feliratkozás súlyának elfogadása ( a SAT>IP klienstől)",
"Accept zero value for TSID", "Nullás érték elfogadása TSID-nek",
"Access", "Hozzáférés",
"Access (ACL)", "Hozzáférés (ACL)",
"Action", "Művelet",
"Active", "Aktív",
"Adapter in use by another subscription", "A kártya használatban van egy másik előfizetésnél",
"Add file and line numbers to debug", "File- és sorszám hozzáadása nyomkövetéshez",
"Add the thread ID to debug", "Folyamat ID hozzáadása a hibakereséshez",
"Add this value to PTS for the teletext subtitles. The time value is in milliseconds and may be negative.", "Adja hozzá ezt az értéket a PTS-hez a teletext feliratokhoz. Az időérték milliszekundumban van és negatív is lehet.",
"Additional service names found in EPG data.", "Az EPG-adatokban található további szolgáltatásnevek.",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "További idő (másodpercben), amellyel a tuner készen áll a felvételre. Ez azok számára hasznos, akik olyan tunerrel rendelkeznek, ahol némi időt vesz igénybe az elején elhelyezett szemét és/vagy hulladék ürítése.",
"Admin password", "Admin jelszó",
"Admin username", "Admin felhasználónév",
"Administrator login", "Adminisztrátori belépés",
"Adult movie", "Felnőtt film",
"Advanced", "Haladó",
"Advanced (non-universal LNBs, rotors, etc.)", "Haladó (nem univerzális LNB-k, forgatók, stb.)",
"Adventure", "Kaland",
"Advertisement / Shopping", "Reklám / Vásárlás",
"After creating the network scan it for services.", "A hálózat létrehozása után vizsgálja meg a szolgáltatásokat.",
"All (Streaming plus DVR)", "Minden (streaming és DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Minden olyan karakter, amely problémákat okozhat a fájlok beillesztésében, aláhúzással lesz helyettesítve. A részleteket lásd a Súgóban.",
"All lower-case", "Minden kisbetűs",
"All proceeds are used to support server infrastructure and buy test equipment.", "Minden adományt a szerver infrastruktúra támogatására és tesztberendezések vásárlására fordítunk.",
"All: Record if different description", "Minden: Felvétel ha eltérő a leírás",
"All: Record if different episode number", "Minden: Felvétel ha különbözik az epizód száma",
"All: Record if different subtitle", "Minden: Felvétel, ha különbözik a feliratnyelv",
"All: Record once per day", "Minden: Felvétel naponta egyszer",
"All: Record once per month", "Minden: Felvétel havonta egyszer",
"All: Record once per week", "Minden: Felvétel hetente egyszer",
"Allow all PIDs", "Minden PID engedélyezése",
"Allow all PIDs (no filter) when the 'Maximum PIDs' limit is reached.", "Minden PID engedélyezése (nincs szűrő) amikor elérte a 'Maximális PID-ek' határértéket.",
"Allow high bitrate mode (CI+ CAMs only).", "Magas bitráta mód engedélyezése (csak CI+ CAM esetén)",
"Allow idle scan tuning on this device.", "Engedélyezze a készenléti szkennelést az eszközön.",
"Allow the combined size of all timeshift buffers to potentially grow unbounded until your storage media runs out of space.", "Engedélyezze ezt, hogy az összes Timeshift puffer kombinált mérete potenciálisan növekedjen, amíg a tárolóeszköz kifogy a helyből.",
"Allow the initial scan tuning on this device (scan when Tvheadend starts or when a new multiplex is added automatically). At least one tuner or input should have this settings turned on. See also 'Skip Startup Scan' in the network settings for further details.", "Engedélyezze a kezdeti szkennelés funkciót az eszközön (szkennelés, amikor a Tvheadend elindul, vagy amikor új multiplexet ad hozzá automatikusan). Legalább egy tunert vagy bemenetet be kell kapcsolni. További részleteket lásd a \"Kezdeti szkennelés kihagyása\" funkciót a hálózati beállításoknál.",
"Allow the timeshift buffer to grow unbounded until your storage media runs out of space. Warning, enabling this option may cause your system to slow down or crash completely!", "Engedélyezze, hogy a timeshift puffer növekedjen addig, amíg a tárolóeszköz kifogy a helyből. Figyelmeztetés! Ha engedélyezi ezt a beállítást, a rendszer lelassulhat vagy teljesen lefagyhat!",
"Allow/Disallow access to the 'Configuration' tab.", "Hozzáférés engedélyezése/tiltása a \"Konfiguráció\" fülhöz.",
"Allow/Disallow web interface access (this  includes access to the EPG).", "A webes felületen való hozzáférés engedélyezése/tiltása (ez magában foglalja az EPG hozzáférését is).",
"Allowed DVR profiles. This limits the profiles the user has access to.", "Engedélyezett DVR profilok. Ez korlátozza azokat a profilokat, amelyekhez a felhasználó hozzáfér.",
"Allowed network", "Lehetséges hálózat",
"Allowed networks", "Engedélyezett hálózat",
"Alternate PID path", "Egyéb PID útvonal",
"Alternate configuration path", "Egyéb beállítási útvonal",
"Altitude (in meters).", "Tengerszinttől mért magasság (m)",
"Altitude (meters)", "Tengerszinttől mért magasság (m)",
"Always send the whole DiseqC sequence including LNB setup (voltage, tone). If this is not checked, only changed settings are sent, which may cause issues with some drivers. If the tuning is not reliable, try activating this option.", "Mindig küldje el az egész DiseqC szekvenciát, beleértve az LNB beállítást (feszültség, hang). Ha ez nincs bejelölve, csak a megváltozott beállítások kerülnek elküldésre, ami egyes illesztőprogramokkal problémákat okozhat. Ha a hangolás nem megbízható, próbálja meg aktiválni ezt a lehetőséget.",
"Always use the name defined by the network.", "Mindig használja a hálózat által definiált nevet.",
"Always use the old ioctls to read the linuxdvb status (signal strength, SNR, error counters). Some drivers are not mature enough to provide the correct values using the new v5 linuxdvb API.", "Mindig használja a régi ioctl-t, hogy kiolvassa a linuxdvb állapotot (jelerősség, SNR, hiba számlálók). Néhány illesztőprogram nem  megfelelő értékeket szolgáltat az új v5 linuxdvb API használatával.",
"Animals", "Állat",
"Anonymize", "Anonimizálás",
"Anonymize HTSP access", "Anonimizált HTSP elérés",
"Any", "Bármi",
"Arts", "Művészetek",
"Arts / Culture (without music)", "Művészet/kultúra (kivéve zene)",
"Arts magazines", "Művészeti magazinok",
"Assign predefined muxes to networks", "Előre definiált muxok hozzárendelése hálózatokhoz",
"Associate this device with one or more networks.", "A készülék egy vagy több hálózathoz csatlakoztatása.",
"Athletics", "Atlétika",
"Audio Stream Filter", "Audió stream szűrő",
"Audio bitrate (kb/s) (0=auto)", "Audió bitráta (kb/s) (0=Auto)",
"Audio bitrate to use for transcoding.", "Audio bitráta transzkódoláshoz",
"Audio channel layout.", "Audio-csatorna elrendezés.",
"Audio codec", "Audio kódec",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Az audió kodek átkódoláshoz való használata. \"Ne használja\" funkció letiltja a hangkimenetet.",
"Audio stream", "Audió stream",
"Audio type", "Audió típus",
"Audio-only stream", "Csak audió stream",
"Audioes muxer", "Hangkeverő",
"Authenticated user", "Azonosított felhasználó",
"Auto check disabled", "Automatikus ellenőrzés kikapcsolva",
"Auto check enabled", "Automatikus ellenőrzés bekapcsolva",
"Auto record", "Automatikus felvétel",
"Auto record caption", "Automatikus felvétel felirat",
"Auto recording%s%s", "Automatikus felvétel%s%s",
"Auto time record", "Automatikus időzített felvétel",
"Auto-Map to channels", "Automatikus csatorna-hozzárendelés",
"Automatic checking", "Automatikus ellenőrzés",
"Automatic recording caption.", "Automatikus felvételi felirat.",
"Automatically map EPG source", "Automatikusan  EPG forrás hozzárendelése",
"Automatically map all available services to channels.", "Az összes elérhető szolgáltatás automatikus mappelése a csatornákhoz.",
"Automatically map channels defined within the bouquet.", "A csokoron belül meghatározott csatornák automatikus megjelenítése.",
"Automatically name from network", "Automatikus névadás a hálozatból",
"Automatically record.", "Automatikus felvétel",
"Autorec maximum count (0=unlimited)", "Automata rögzítés maximális száma (0 - korlátlan)",
"Autorec maximum schedules limit (0=unlimited)", "Automata rögzítés ütemezésének maximális száma (0 - korlátlan)",
"Bad", "Rossz",
"Bad request", "Rossz kérés",
"Ballet", "Balett",
"Bandwidth", "Sávszélesség",
"Base DVB SI Tables (PAT,CAT,PMT,SDT etc.)", "Alap DVB SI táblák (PAT, CAT, PMT, SDT stb.)",
"Based on software from", "Az innen származó szoftver alapján",
"Basic", "Alap",
"Bind to specific local IP address.", "Helyi IP-címhez rendelés.",
"Bitrate to use for the transcode. See Help for details.", "Átkódoláshoz használt bitráta. A részleteket lásd a Súgóban.",
"Blue", "Kék",
"Bouquet", "Csokor",
"Bouquet (auto)", "Csokor (auto)",
"Bouquet source.", "Csokor forrás.",
"Bouquets", "Csokrok",
"Brand", "Márka",
"Branding information (if available).", "Márka-információ (ha elérhető).",
"Broadcast", "Sugárzás",
"Broadcast type", "Adás típus",
"Broadcast.", "Adás.",
"Broadcasting", "Sugárzás",
"Buffering limit (ms)", "Pufferelési limit (ms)",
"CA (descrambling) Client", "CA (visszafejtő) kliens",
"CA Stream Filter", "CA stream szűrő",
"CA client %i", "CA-kliens %i",
"CA identification", "CA-azonosítás",
"CA provider", "CA-szolgáltató",
"CAPMT (Linux Network DVBAPI)", "CAPMT (Linux DVBAPI)",
"CAPMT CA Client", "CAPMT CA kliens",
"CAPMT interval (in ms).", "CAPMT intervallum (ms)",
"CAPMT interval (ms)", "CAPMT intervallum (ms)",
"CAPMT query interval (ms)", "CAPMT lekérdezési intervallum (ms)",
"CAPMT query interval (ms).", "CAPMT lekérdezési intervallum (ms)",
"CI Module", "CI-modul",
"CRASH", "ÜTKÖZÉS",
"CRID authority", "CRID engedély",
"CSA (descrambling)", "CA (visszafejtő)",
"CWC CA Client", "CWC CA kliens",
"Cache scheme", "Gyorsítótár vázlat",
"Camd.socket filename / IP Address (TCP mode)", "Camd.socket file-név / IP-cím (TCP mód)",
"Cartoons", "Rajzfilmek",
"Change parameters", "Paraméterek változtatása",
"Channel", "Csatorna",
"Channel Tags", "Csatorna-címkék",
"Channel does not exist", "A csatorna nem létezik",
"Channel icon", "Csatorna ikon",
"Channel icon URL.", "Csatorna ikon URL.",
"Channel icon as defined in EPG data.", "Csatorna ikon az EPG-adatok szerint.",
"Channel icon name scheme", "Csatorna ikon-név rendszer",
"Channel icon path", "Csatorna ikon útvonal",
"Channel mapping options", "Csatorna-feltérképezési lehetőségek",
"Channel name", "Csatorna-név",
"Channel not found", "A csatorna nem található",
"Channel number", "Csatorna szám",
"Channel number as defined in EPG data.", "Csatornaszám az EPG-adatok szerint.",
"Channel number offset", "Csatorna szám offszet",
"Channel number range", "Csatorna számozás tartomány",
"Channel numbers (heuristic)", "Csatornaszámok (heurisztikus)",
"Channel numbers from", "Csatorna számozás kezdőérték",
"Channel tag", "Csatorna-cimke",
"Channel tag does not exist", "A csatorna-címke nem létezik",
"Channel tag reference", "Csatorna-cimke referencia ",
"Channel tags", "Csatorna-cimkék",
"Channel update options", "Csatornafrissítési opciók",
"Channels", "Csatornák",
"Character set", "Karakterkészlet",
"Charset", "Karakter készlet",
"Check availability", "Elérhetőség ellenőrzése",
"Children's / Youth Programs", "Gyermek / Ifjúsági műsorok",
"Children's / Youth programs", "Gyermek / Ifjúsági műsorok",
"Cinema", "Mozi",
"Class", "Osztály",
"Classical", "Klasszikus",
"Classical music", "Klasszikus zene",
"Clean effects", "Tiszta effektusok",
"Client name", "Kliens név",
"Clone scheduled entry on error", "Hiba esetén klónozza a bejegyzést",
"Code Word Client (newcamd)", "Kódszó kliens (newcamd)",
"Comedy", "Vígjáték",
"Command delay time (ms) (10-200)", "Parancs késleltetési idő (ms) (10-200)",
"Command time (ms) (10-100)", "Parancs idő (ms) (10-100)",
"Comment", "Megjegyzés",
"Commercial break", "Reklámszünet",
"Committed", "Jóváhagyott",
"Completed OK", "Befejezve OK",
"Compress EPG database", "EPG adatbázis tömörítése",
"Compress the EPG database to reduce disk I/O and space.", "EPG adatbázis tömörítése a lemez I/O műveletek és a méret csökkentésének érdekében",
"Conditional Access", "Feltételes hozzáférés",
"Conditional Access Client", "Feltételes hozzáférés ügyfél",
"Conditional Access Identification.", "Feltételes hozzáférés azonosítás",
"Configuration", "Beállítás",
"Configuration - Base", "Beállítás - Alapok",
"Configuration - Image Cache", "Beállítás - Átmeneti képtároló",
"Configuration - SAT>IP Server", "Beállítás - SAT>IP szerver",
"Configuration ID", "Beállítás ID",
"Configuration name", "Konfiguráció neve",
"Configuration version", "Konfiguráció verziója",
"Connection limit type", "Csatlakozás korlátozás típusa",
"Connection limits", "Csatlakozási korlátok",
"Constellation", "Csillagkép",
"Container", "Tároló",
"Content character set", "Tartalom betűtípusa",
"Content reference identifier authority.", "Tartalom referencia azonosító hatóság.",
"Content type", "Tartalom típusa",
"Content type.", "Tartalom típusa.",
"Contest", "Verseny",
"Cookie expiration (days)", "Sütik lejárata (nap)",
"Cooking", "Főző-műsor",
"Copy codec type", "Kodektípus másolása",
"Copy layout", "Másolat kinézete",
"Count of objects", "Objektumok száma",
"Create SDTV/HDTV/Radio tags.", "Hozzon létre SDTV/HDTV/Radio címkéket.",
"Create a provider name tag.", "Szolgáltatói név-címkék létrehozása.",
"Create bouquet", "Csoport létrehozása",
"Create network name tags", "Hálózati név tag-ek létrehozása",
"Create network tags", "Hálózati címkék létrehozása",
"Create provider name tags", "Szolgáltatói név tag-ek létrehozása",
"Create provider tags", "Szolgáltatói címkék létrehozása",
"Create tags", "Címke létrehozása",
"Created", "Létrehozott",
"Creator", "Szerző",
"Cron multi-line", "Cron több-soros",
"Culture (without music)", "Kultúra (kivéve zene)",
"Culture magazines", "Művészeti magazinok",
"Current RTSP port.", "Jelenlegi RTSP port.",
"Current affairs", "Jelenlegi afférok",
"Current number of objects.", "Az aktuális objektumok száma.",
"Current object size.", "Az aktuális objektumok mérete.",
"Current tuner configuration.", "Jelenlegi tuner beállítások.",
"Custom HTTP headers", "Egyedi HTTP fejléc",
"DAB Network", "DAB hálózat",
"DBus - use the session message bus instead of the system one", "DBus - a jelenlegi üzenet használata a rendszer helyett",
"DES Key.", "DES kulcs.",
"DES key", "DES kulcs",
"DSCP/TOS for streaming", "DSCP/TOS a streaming-hez",
"DVB CAM Client", "DVB CAM kliens",
"DVB CSA (descrambling) Tables", "DVB CSA (dekódolás) táblák",
"DVB EPG Tables", "DVB EPG táblák",
"DVB Inputs - Networks", "DVB bemenetek - Hálózatok",
"DVB SI Tables", "DVB SI táblák",
"DVB Time Tables", "DVB Idő táblák",
"DVB scan files", "DVB scan file-ok",
"DVB scan files path", "DVB scan file-ok elérési útvonala",
"DVB-C Network", "DVB-C hálózat",
"DVB-S Network", "DVB-S hálózat",
"DVB-S Satellite Configuration", "DVB-S műhold konfiguráció",
"DVB-T Network", "DVB-T hálózat",
"DVR - Auto-recording (Autorecs)", "DVR - Auto-rögzítés (Autorec)",
"DVR - Profiles", "DVR - profilok",
"DVR - Time-based Recording (Timers)", "DVR - Időalapú felvétel (Időzítők)",
"DVR behavior", "DVR működés",
"DVR configuration", "DVR beállítások",
"DVR configuration profiles", "DVR beállítási profilok",
"DVR configurations", "DVR beállítások",
"DVR entry not found", "DVR bejegyzés nem található",
"DVR file retention period", "DVR file tárolás",
"DVR log retention", "DVR log tárolása",
"DVR log retention period", "DVR log tárolási periódus",
"DVR override: high", "DVR felülbírálása: magas",
"DVR override: important", "DVR felülbírálása: fontos",
"DVR override: low", "DVR felülbírálása: alacsony",
"DVR override: normal", "DVR felülbírálása: normális",
"DVR override: unimportant", "DVR felülbírálása: nem fontos",
"DVR schedule does not exist", "DVR időzítő nem létezik",
"DVR schedule not found", "DVR időzítő nem található",
"Dance", "Tánc",
"Data errors", "Adathibák",
"Data path (if applicable).", "Adat elérési útja (adott esetben).",
"Days of Week", "Hét napjai",
"Days of week", "Hét napjai",
"Debate", "Vitaműsor",
"Debug libav log", "Részletesebb libav naplózás",
"Debug options", "Debug opciók",
"Debug subsystems", "Debug alrendszerek",
"Debug to syslog", "Debug naplózás a syslog-ba",
"Debug trace (low-level)", "Debug trace (alacsony-szintű)",
"Debugging", "Nyomkövetés",
"Default", "Alapértelmezett",
"Default language(s)", "Alapértelmezett nyelv(ek)",
"Default language.", "Alapértelmezett nyelv",
"Default login", "Alapértelmezett belépés",
"Default priority", "Alapértelmezett prioritás",
"Default user interface level.", "Alapértelmezett felhasználói szint",
"Delivery system", "Szállítási rendszer",
"Demux path", "Demux elérési útvonal",
"Descramble services (limit per mux)", "Szolgáltatások dekodolása (mux szerinti korlátozás)",
"Descrambler", "Titkosítás feloldás",
"Descrambler EMM", "Titkosítás feloldás EMM",
"Descrambler buffer (TS packets)", "Dekódolő puffer (TS csomagok)",
"Description", "Leírás",
"Detective", "Detektív",
"Device ID", "Eszköz ID",
"Device model", "Eszköz típus",
"Device path", "Eszköz elérési útvonal",
"Differentiated Services Code Point / Type of Service: Set the service class Tvheadend sends with each packet. Depending on the option selected this tells your router the prority in which to give packets sent from Tvheadend, this option does not usually need changing. See https://en.wikipedia.org/wiki/Differentiated_services for more information. ", "Differentiated Services Code Point / Type of Service: a service class beállítása, amit a Tvheadend küld minden csomaggal. A választástól függően ez közli a routerrel, hogy milyen prioritással küldje tovább a Tvhheadend által továbbított csomagokat, ez a beállítás általában nem igényel változtatást. Lásd: https://en.wikipedia.org/wiki/Differentiated_services további információkért. ",
"Digital Video Recorder", "Digitális Videó Felvevő",
"Directory", "Könyvtár",
"Directory permissions (octal, e.g. 0775)", "Könyvtárengedélyek (oktális, pl. 0775)",
"Disable", "Kikapcsol",
"Disable DVB bouquets", "DVB csokrok kikapcsolása",
"Disable SAT>IP client", "SAT>IP kliens letiltása",
"Disable UPnP", "UPnP letiltása",
"Disable UPnP discovery.", "Letiltja a UPnP felfedezést.",
"Disable all access control checks", "Minden hozzáférési-elérés vizsgálatának kikapcsolása",
"Disable debug on stderr", "Debug stderr interfészen kikapcsolása",
"Disable device/firmware-specific workarounds", "Eszköz / firmware-specifikus megoldások kikapcsolása",
"Disable syslog (all messages)", "Rendszernapló letiltása (minden üzenet)",
"Disabled", "Kikapcsolt",
"Discussion", "Beszélgetés",
"DiseqC repeats", "DiseqC ismétlés",
"Do not use", "Ne használd",
"Documentary", "Dokumentum",
"Don't backup configuration tree at upgrade", "Konfigurációs fa frissítésének mellőzése mentés esetén.",
"Don't include the title in the filename.", "Ne szerepeljen a cím a fájlnévben",
"Don't include title in filename", "Cím használatának mellőzése a fájlnévben",
"Don't keep", "Ne tartsa meg",
"Don't re-record", "Ne rögzítsen újra",
"Don't re-record if recording fails.", "Ne ismételje meg a felvételt, ha első alkalommal sikertelen",
"Don't re-schedule if recording fails.", "Ne időzítse újra, ha a felvétel nem sikerül",
"Don't reschedule", "Ne ütemezze újra",
"Don't use the provider's channel numbers.", "Szolgáltatói csatorna számozás figyelmen kívül hagyása.",
"Drama", "Dráma",
"Duplicate handling", "dupla bejegyzések kezelése",
"Duplicate recording handling.", "Duplikált felvétel kezelése.",
"EIT - skip TSID check", "EIT - TSID ellenőrzésének kihagyása",
"EPG Database", "EPG adatbázis",
"EPG Grabber", "EPG letöltő",
"EPG Grabber Configuration", "EPG letöltés konfiguráció",
"EPG Language (priority order)", "EPG nyelv (prioritás sorrend)",
"EPG name", "EPG név",
"EPG scan", "EPG szkennelés",
"EPG scan time-out in seconds (30-7200)", "EPG szken időtúllépés másodpercben (30-7200)",
"EPG source", "EPG forrása",
"EPG update window", "EPG frissítési ablak",
"Economics", "Gazdaság",
"Education", "Oktatás",
"Education / Science / Factual topics", "Oktatás / tudomány / tényfeltáró",
"Educational", "Oktatási",
"Electronic Program Guide", "Elektronikus műsor újság",
"Elementary stream filter", "Alapvető stream szűrő",
"Enable", "Engedélyezés",
"Enable (auto)", "Engedélyez (automatikus)",
"Enable DBus", "DBUS engedélyezése",
"Enable NTP driver", "NTP meghajtó engedélyezése",
"Enable coredumps for daemon", "Coredump-ok engedélyezése a daemon-hoz",
"Enable debug on stderr", "Debug stderr interfészen  engedelyezése",
"Enable debug subsystems", "Debug alrendszerek engedelyezése",
"Enable debug to file", "Debug file-ba  engedelyezése",
"Enable debug to syslog", "Debug syslog-ba  engedelyezése",
"Enable if the SAT>IP box requests plts=on parameter in the SETUP RTSP command for DVB-S2 muxes.", "Engedélyezze, ha a SAT>IP box a plts=on paramétert kér a SETUP RTSP parancsban DVB-S2 mux esetén.",
"Enable if the SAT>IP box requires pids=21 parameter in the SETUP RTSP command", "Engedélyezze, ha a SAT>IP box a pids=21 paramétert kér a SETUP RTSP parancsban.",
"Enable if the SAT>IP box supports the addpids/delpids commands.", "Engedélyezze, ha a SAT>IP box támogatja az addpids/delpids parancsokat.",
"Enable or disable fast input switching.", "A gyors bemeneti kapcsolás engedélyezése vagy tiltása.",
"Enable or disable full mux mode.", "A teljes mux üzemmód engedélyezése vagy tiltása.",
"Enable or disable ignoring of Event Information Table (EIT) data on this mux.", "Az esemény információs tábla (Event Information Table (EIT)) adatainak figyelmen kívül hagyásának engedélyezése vagy tiltása ezen a mux-on.",
"Enable or disable this configuration.", "Jelenlegi konfiguráció engedélyezése vagy tiltása.",
"Enable syslog", "Syslog engedélyezése",
"Enable system time updates. This will only work if the user running Tvheadend has rights to update the system clock (normally only root).", "Rendszeridő frissítésének engedélyezése. Ez csak akkor működik, ha a Tvheadend-et futtató felhasználónak van jogosultsága a rendszeróra állításához. (Általában ez csak root jogokkal működik).",
"Enable this filter.", "Szűrő engedélyezése",
"Enable trace subsystems", "Trace alrendszerek engedelyezése",
"Enable web UI debug (non-minified JS)", "Webes felületének debug-olásának engedélyezése (nem-minimalizált JS)",
"Enable/Disable CA client.", " CA kliens engedélyezése/tiltása.",
"Enable/Disable service.", "Szolgáltatás engedélyezése/tiltása.",
"Enable/Disable timeshift.", "Időeltolás engedélyezése/tiltása.",
"Enable/disable LNA.", "LNA engedélyezése/tiltása.",
"Enable/disable auto-rec rule.", "Automatikus felvétel szabály engedélyezése/tiltása.",
"Enable/disable profile.", "Profil engedélyezése/tiltása.",
"Enable/disable the channel.", "Csatorna engedélyezése/tiltása.",
"Enabled", "Engedélyezett",
"Encrypted", "Kódolt",
"Encrypted service", "Kódolt csatorna",
"Entertainment programs for 10 to 16", "Szórakoztató programok 10-től 16-ig",
"Entertainment programs for 6 to 14", "Szórakoztató programok 6-tól 14-ig",
"Environment", "Környezet",
"Environment (pipe)", "Környezet (pipe)",
"Episode", "Epizód",
"Episode %d", "Epizód %d",
"Episode number/ID.", "Rész száma/ID.",
"Error code", "Hibakód",
"Errors", "Hibák",
"Even key", "Páros kulcs",
"Every day", "Minden nap",
"Exclude channel tags", "Csatorna címkék mellőzése",
"Expeditions", "Expedíciók",
"Experimental film", "Kísérleti film",
"Expert", "Haladó",
"Exported tuners", "Exportált tunerek",
"External", "Külső",
"External EPG grabber", "Külső EPG letöltő",
"External IP (NAT)", "Külső IP cím (NAT)",
"External URL", "Külső URL",
"Extra arguments", "Plussz paraméterek",
"Extra start time", "Plussz kezdési idő",
"Extra stop time", "Plussz leállítási időpont",
"Extra warming up time (seconds)", "Plussz bemelegítési idő (másodperc)",
"FAIL", "SIKERTELEN",
"FEC high", "FEC magas",
"FEC low", "FEC alacsony",
"Factual topics", "Tényszerű témák",
"Failed", "Sikertelen",
"Fashion", "Divat",
"Fast input switch", "Gyors bemenet váltás",
"File missing", "Hiányzó fájl",
"File not created", "A fájl nem lett létrehozva",
"File permissions (octal, e.g. 0664)", "Fájl jogosultságok (oktális, pl. 0664)",
"File size", "Fájlméret",
"Filename", "Fájlnév",
"Filename character set", "Fájlnév karakterkészlete",
"Filename options", "Fájlnév tulajdonságai",
"Fine arts", "Szépművészet",
"Finished", "Befejezett",
"Fitness and health", "Fitness és egészség",
"Folk", "Népzene",
"Folkloric", "Népi",
"Football", "Labdarúgás",
"Force (auto)", "Kényszerít (automatikus)",
"Force CA ID (e.g. 0x2600)", " CA ID Kényszerítés (pl.. 0x2600)",
"Force all network connections to this tuner to be made over the specified IP address, similar to the setting for the SAT>IP device itself. Setting this overrides the device-specific setting.", "Kényszerítse az összes hálózati kapcsolatot, amely ehhez a tunerhez csatlakozik, hogy ezen a IP-n keresztül jöjjön létre,  hasonlóan a  SAT> IP eszköz beállításához.Ez  a beállítás  felülírja az eszköz-specifikus beállításokat.",
"Force old status", "Régi stilusú státusz lekérés kényszerítése",
"Force pilot for DVB-S2", "Pilot mód kényszerítése  DVB-S2 -re",
"Force priority", "Prioritás kényszerítése",
"Force teardown delay", "Szétszerelés - teardown késleltetés",
"Force the delay between RTSP TEARDOWN and RTSP SETUP command (value from 'Next tune delay in ms' is used). Some devices are not able to handle quick continuous tuning.", "Kényszerítse  a késést  az RTSP TEARDOWN és RTSP SETUP parancsoknál (értékét a next tune delay in ms beallításból kapja). Egyes készülékek nem képesek kezelni a gyors folytonos hangolást.",
"Force the local UDP Port number here. The number should be even (RTP port). The next odd number (+1) will be used as the RTCP port.", "Kényszerítse az itt megadott lokális UDP-port számának használatát. A érték páros kell, hogy legyen (RTP port). A következő páratlan szám (+1) lesz használva mint  RTCP port.",
"Force usage of entered CA ID on this service.", "A beírt CA ID kényszerített használata ennél a szolgáltatásnál.",
"Foreign countries", "Külföldi országok",
"Forever", "végtelen",
"Fork and run as daemon", "Fork-ol és futtat mind daemon",
"Format string", "Karakter formátum",
"Found muxes", "Talált mux-ok",
"Found services", "Talált szolgáltatások",
"Free subscription weight", "Szabad feliratozások súlya",
"Free-format text field. Enter whatever you like here.", "Szabad formátumú szöveges mező. Írjon be bármit ide.",
"Frequency", "Frekvencia",
"Frequency (Hz)", "Frekvencia (Hz)",
"Frequency (MHz)", "Frekvencia (MHz)",
"Frequency (in MHz).", "Frekvencia (MHz-ben).",
"Frequency (kHz)", "Frekvencia (kHz)",
"Fri", "Pé",
"Friendly name", "Barátságos név",
"Frontend number", "Fejegység száma",
"Frontend path", "Fejegység  elérési útvonal",
"Full DiseqC", "Teljes DiseqC",
"Full mux RX mode supported", "Teljes MUX RX mód támogatás",
"Full pathname specification", "Teljes elérési útvonal meghatározása",
"Full-text", "Teljes szöveg",
"Further education", "További oktatás",
"GOTOX position", "GOTOX pozicíó",
"Game show", "Vetélkedő",
"Gardening", "Kertészkedés",
"General", "Általános",
"General configuration", "Általános beállítás",
"Generic", "Általános",
"Generic options", "Általános opciók",
"Gray", "Szürke",
"Guard interval", "Figelési intervallum",
"HD: high definition", "HD: Magas felbontású",
"HDHomeRun DVB frontend", "HDHomeRun DVB fejegység",
"HDHomeRun DVB-C frontend", "HDHomeRun DVB-C fejegység",
"HDHomeRun DVB-T frontend", "HDHomeRun DVB-T fejegység",
"HTSP Answer", "HTSP válasz",
"HTSP Default Stream Settings", "HTSP Alapértelmezett Stream Beállítások",
"HTSP Request", "HTSP kérés",
"HTSP Server", "HTSP szerver",
"HTTP CORS origin", "HTTP CORS kezdeti érték",
"HTTP Server", "HTSP szerver",
"Handicraft", "Kézművesség",
"Hearing impaired", "Hallássérült",
"Hierarchy", "Hierarchia",
"High", "Magas",
"High bitrate mode (CI+ CAMs only)", "Magas bitráta mód (CI+ CAM )",
"Historical movie", "Történelmi film",
"How frequently it will re-try fetching an image that has failed to be fetched.", "Milyen gyakran próbálja újra letölteni a képet, amelyet nem sikerült letöltenie.",
"How frequently the upstream provider is checked for changes.", "Milyen gyakran ellenőrizze a feltöltő szolgáltatót változások miatt.",
"IDLE", "üresjárat",
"IP address", "IP-cím",
"IP blocking", "IP blokkolás",
"IPTV Multiplex", "IPTV-multiplex",
"IPTV Network", "IPTV-hálózat",
"IPTV Subcription", "IPTV előfizetés",
"IPTV input", "IPTV-bemenet",
"ISDB-C Network", "ISDB-C hálózat",
"ISDB-S Network", "ISDB-S hálózat",
"ISDB-T Network", "ISDB-T hálózat",
"Icon", "Ikon",
"Icon (full URL)", "Ikon (teljes URL)",
"Icon URL", "Ikon URL",
"Icon base URL", "Ikon elérési URL ",
"Icon has title", "Ikon címmel rendelkezik",
"Icons from", "Ikonok forrása",
"Idle", "üresjárat",
"Idle scan", "Üresjáratú szkennelés-engedélyezése",
"Idle scan muxes", "Üresjáratú MUX szkennelés-engedélyezése",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Abban az esetben ha nem létezik felhasználói fiók\nHozzon létre egy bejegyzést felhasználónév és jelszó nélkül\nhasználja elővigyázattal, ugyanis korlátlan hozzáférést biztosít egészen addig, míg nem hoz létre/modoítja a létező felhasználót a webes felületen",
"If you'd like to support the project, please consider a donation.", "Ha támogatni szeretnéd a programot, fontold meg a támogatását.",
"Ignore", "Elutasít",
"Ignore EPG (EIT)", "EPG figyelmen kívül hagyása (EIT)",
"Ignore descriptor 5", "Hagyja figyelmen kívül az 5-ös leírót",
"Ignore invalid SSL certificate", "Érvénytelen SSL tanusitvány figyelmen kívül hagyása",
"Ignore invalid/unverifiable (expired, self-certified, etc.) certificates", "Hagyja figyelmen kívül  az érvénytelen / ellenőrizhetetlenek (lejárt, öntanúsított, stb) tanúsítványokat.",
"Ignore provider's channel numbers", "Szolgáltatói csatorna számozás figyelmen kívül hagyása",
"Immediately abort", "Azonnali megszakítás",
"Important", "Fontos",
"Include channel name in filename", "Csatorna nevének használata a fájlnévben",
"Include date in filename", "Dátum használata a fájlnévben",
"Include episode in filename", "Epizód nevének használata a fájlnévben",
"Include subtitle in filename", "Felirat használata a fájlnévben",
"Include time in filename", "Idő használata a fájlnévben",
"Information area", "Információs terület",
"Informational", "Tájékoztató jellegű",
"Initial scan", "Üresjáratú szkennelés-engedélyezése",
"Input base", "Bemenet alapcíme",
"Input buffer (bytes)", "Bemeneti puffer (bájtban)",
"Input instance", "Bemeneti példány",
"Input path", "Bemenet elérési útja",
"Interface", "Felület",
"Internal", "Belső",
"Internal EPG grabber", "Belső EPG letöltő",
"Internal error", "Belső hiba",
"Internal grabber", "Belső letöltő",
"Interview", "Interjú",
"Invalid", "Érvénytelen",
"Invalid adapter number '%s'\n", "Érvénytelen eszköz szám '%s'\n",
"Invalid file", "Érvénytelen file",
"Invalid parameters", "Érvénytelen paraméterek",
"Invalid service", "Érvénytelen szolgáltatás",
"Invalid target", "Érvénytelen cél",
"Keep", "Megtart",
"Kill signal (pipe)", "Kill szignál (pipe)",
"Kill timeout (pipe/secs)", "Kill szignál maximális időkorlát (pipe/másodperc)",
"LNA (low noise amplifier)", "LNA (alacsony zajszintű erőstő)",
"LNB type", "LNB-típus",
"Language", "Nyelv",
"Language 1", "1. nyelv",
"Language 2", "2. nyelv",
"Language 3", "3. nyelv",
"Language settings", "Nyelvi beállítások",
"Languages", "Nyelvek",
"Last seen", "Utoljára látott",
"Last updated from", "Legutoljára frissítve",
"Leisure hobbies", "Szabadidő, hobbi",
"Limit connections", "Kapcsolatok korlátozása",
"Linked input", "Csatolt bemenet",
"Linux DVB frontend", "Linux DVB felület",
"LinuxDVB adapter", "LinuxDVB eszköz",
"LinuxDVB network", "LinuxDVB hálózat",
"Listen / Connect port", "Kapcsolatvárási/csatlakozási port",
"Listen on IPv6", "Szerver kapcsolatot vár IPv6 -on",
"Literature", "Irodalom",
"Local (server) time", "Helyi (szerver) idő",
"Local bind IP address", "Helyileg bekütődő IP-cím",
"Local channel minor", "Helyi csatorna - alsó",
"Local channel number", "Helyi csatornaszám",
"Local discovery IP address", "IP-cím lokális felderítése ",
"Location", "Helyszín",
"Location details of the SAT>IP Server.", "A SAT>IP szerver helyzetének részletei.",
"Lock preferred CA PID", "Preferált CA PID lock-olása",
"Login password.", "Belépési jelszó.",
"Login username.", "Belépési felhasználónév.",
"Login/Logout", "Be/kijelentkezés",
"Logout", "Kilépés",
"Low", "Alacsony",
"MP4 audio", "MP4 audió",
"MPEG-TS File", "MPEG-TS file",
"MPEG-TS Pass-thru", "MPEG-TS áteresztés",
"MPEG-TS Pass-thru/built-in", "MPEG-TS áteresztés/beépített",
"MPEG-TS input", "MPEG-TS bemenet",
"MPEG-TS raw service", "MPEG-TS raw szolgáltatás",
"Magazines", "Magazinok",
"Main", "Fő",
"Maintain free storage space in MiB", "Elérhető tárhely MIB-ben",
"Maintain used storage space in MiB (0=disabled)", "Használt tárhely fenntartása  MB-ban (0=kikapcsolva)",
"Maintained space", "Tárhely fenntartása",
"Make subdirectories per channel", "Alkönyvtárak csatornánkénti létrehozása",
"Make subdirectories per day", "Alkönyvtárak naponkénti létrehozása",
"Make subdirectories per title", "Alkönyvtárak cím szerinti létrehozása",
"Manufacturer", "Gyártó",
"Manufacturer URL", "Gyártói URL",
"Manufacturer's model number.", "Gyártói modell száma.",
"Manufacturer's product description.", "Gyártói termékleírás.",
"Manufacturer's product information page for the device.", "A gyártó termékinformációs oldala.",
"Manufacturer's product name.", "Gyártói terméknév.",
"Map all services", "Minden szolgáltatás hozzárendelése",
"Map encrypted services", "Titkosított szolgáltatások összerendelése.",
"Map radio channels", "Rádio-csatorna szolgáltatások hozzárendelése",
"Map unnamed channels", "Névvel nem rendelkező csatornák hozzárendelése",
"Map zero-numbered channels", "0-s sorszámmal rendelkező csatornák hozzárendelése",
"Martial sports", "Küzdősportok",
"Master tuner", "Fő tuner",
"Matroska (mkv)/built-in", "Matroska (mkv)/beépített",
"Matroska specific", "Matroska specifikus",
"Maximal channel number", "Maximális csatorna szám",
"Maximum # input streams", "Maximális számú bemeneti streamek",
"Maximum PIDs", "Maximum PID",
"Maximum RAM size (MB)", "Maximális RAM méret (MB)",
"Maximum bandwidth (Kbps)", "Maximális sávszélesség (Kbps)",
"Maximum count (0=default)", "Maximális szám (0=alapértelmezett)",
"Maximum duration", "Maximális időtartam",
"Maximum length in characters for the command setting PIDs to the SAT>IP box.", "Maximális karakterhossz a PID parancs beállításoknak a SAT> IP mezőben.",
"Maximum length of PIDs", "PID-ek maximális hoszza",
"Maximum period (mins)", "Maximális periódus (percben)",
"Maximum schedules limit (0=default)", "Maximális betervezésének száma (0=alapértelmezett)",
"Maximum size (MB)", "Maximális méret (MB)",
"Maximum timeout (seconds)", "Maximális időkorlát (másodperben)",
"Medicine", "Gyógyászat",
"Melodrama", "Melodráma",
"Memory Information", "Memoria Információ",
"Merge same name", "Ugyanazon nevek összevonása",
"Minimal channel number", "Minimális csatorna szám",
"Minimum duration", "Minimális időtartam",
"Minimum rotor time (seconds)", "Minimális forgató időtartam (másodperben)",
"Missing In PAT/SDT", "Hiányzik a PAT/SDT -ben",
"Mode", "Üzemmód",
"Model description", "Üzemmód leírás",
"Model name", "Modell név",
"Model number", "Modell szám",
"Modulation", "Moduláció",
"Module", "Modul",
"Module ID", "Modul ID",
"Mon", "Hét",
"More verbose libav log", "Részletesebb libav naplózás",
"Motor rate (milliseconds/deg)", "Forgatási sebesség (milli/fok)",
"Motor sport", "Motorsport",
"Motoring", "Motor",
"Movie", "Film",
"Movie / drama", "Film / dráma",
"Multiple lines of the cron time specification. The default cron triggers the internal grabbers daily at 12:04 and 00:04. See Help on how to define your own.", "Többsoros cron idő meghatározásának. Az alapértelmezett cron kiváltja a Over-the-air EPG gyüjtőt, naponta 02:04 és 14:04 időpontokban . A súgóban több információt megtalál, hogyan határozzon meg a sajátot.",
"Music", "Zene",
"Music / Ballet / Dance", "Zene / balett / tánc",
"Mux handling", "Mux kezelés",
"Mux name", "Mux neve",
"Mux not enabled", "A MUX nem engedélyezett",
"NONE", "EGYIK SEM",
"Name", "Név",
"Name of the entry.", "A bejegyzés neve.",
"Name of the network.", "A hálózat neve.",
"Name of the profile.", "A profil neve.",
"Name of the tag.", "A címke neve.",
"Name of the tuner.", "Tuner neve.",
"Names", "Nevek",
"Natural sciences", "Természettudományok",
"Nature", "Természet",
"Network", "Hálózat",
"Network ", "Hálózat",
"Network A", "A hálózat",
"Network B", "B hálózat",
"Network C", "C hálózat",
"Network D", "D hálózat",
"Network ID (limit scanning)", "Hálózat ID (szkennelés korlátozás)",
"Network UUID", "Hálózati UUID",
"Network access", "Hálózati hozzáférés",
"Network discovery", "Hálózati felderítés",
"Network name", "Hálózat neve",
"Network prefix", "Hálózati előtag",
"Network settings", "Hálózati beállítások",
"Network type", "Hálózat típusa",
"Networks", "Hálózatok",
"New / premiere / unknown", "Újdonság / premier / ismeretlen",
"News", "Hírek",
"News / Current Affairs", "Hírek / Aktuális",
"News magazine", "Hírmagazin",
"Next tune delay in ms (0-2000)", "Hangolási parancs késleltetési idő millisec.-ben (0-2000)",
"No", "Nem",
"No PIN", "Nincs PIN",
"No access", "Nincs hozzáférés",
"No adapters specified!\n", "Nincs meghatározott eszköz\n",
"No assigned adapters", "Nincs hozzárendelt adapter",
"No days", "Nincsenek napok",
"No descrambler", "Nincs kititkosító",
"No free adapter", "Nincs szabad adapter",
"No input detected", "Nincs érzékelt bemenet",
"No service assigned to channel", "Nincs szolgáltatás hozzárendelve a csatornához",
"No service enabled", "Nincs engedélyezett szolgáltatás",
"No source available", "Nincs forrás",
"None", "Egyik sem",
"Normal", "Normál",
"Not all SAT>IP servers use the same signal scaling. Change this setting if the signal level displayed within Tvheadend looks too low.", "Nem minden SAT> IP szerver használja ugyanazt a jel-léptéket. Változtassa meg ezt, ha a Tvheadend-ben megjelenített jelszint túl alacsonynak tűnik.",
"Not enough disk space", "Nincs elég tárhely",
"Not enough memory", "Nincs elég memória",
"Not linked", "Nem linkelt",
"Not set", "Nincs beállítva",
"Not set (none or channel configuration)", "Nem beállított (egy sem vagy csatorna beállítás hiányzik)",
"Not set (use channel or DVR configuration)", "Nem beállított (csatorna használata vagy DVR beállítás)",
"Number", "Szám",
"Number of seconds to wait before timing out.", "Lejárat előtti várakozási idő másodpercekben.",
"Number of tsfile tuners", "Tsfile tőnerek száma",
"OK (partial)", "OK (részleges)",
"OSCam net protocol (rev >= 10389)", "OSCam net protokol (rev >= 10389)",
"OSCam new pc-nodmx (rev >= 10389)", "OSCam új pc-nodmx (rev >= 10389)",
"Odd key", "Páratlan kulcs",
"Off", "Ki",
"Older OSCam", "Régebbi OSCam",
"On", "Be",
"On file removal", "File eltávolításánál",
"Only Bulsatcom 39E", "Csak Bulsatcom 39E",
"Only EIT", "Csak EIT",
"Only OpenTV Sky Ausat", "Csak OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Csak OpenTV Sky Italia",
"Only OpenTV Sky UK", "Csak OpenTV Sky UK",
"Only PSIP (ATSC)", "CSak PSIP (ATSC)",
"Only UK Freesat", "Csak UK Freesat",
"Only UK Freeview", "Csak UK Freeview",
"Only Viasat Baltic", "Csak Viasat Baltic",
"Only preferred CA PID", "Csak preferált CA PID",
"Only update the system clock (doesn't affect NTP driver) if the delta between the system clock and DVB time is greater than this. This can help stop excessive oscillations on the system clock.", "Csak akkor frissítse a rendszer órát (nem befolyásolja az NTP drivert) ha a külömbség a rendszer óra és DVB idő között nagyobb, mint az izz beállított érték. Ez segít a túlzott  rendszer óra oszcillációk ellen.",
"OpenTV channel number", "OpenTV csatorna szám",
"Orbital position", "Műhold-pozíció",
"Orbital positions", "Műhold-pozíciók",
"Orbital positions.", "Műhold-helyzetek.",
"Original network ID", "Eredeti hálózat ID",
"Oscam mode.", "Oscam üzemmód.",
"Output directory for tsdebug", "Kimenő mappa tsdebug-nak",
"Over-the-air", "OTA",
"Over-the-air Cron multi-line", "OTA Cron job több-soros",
"Over-the-air EPG grabber", "Over-the-air EPG letöltő",
"Over-the-air grabbers", "Over-the-air letöltők",
"Override the frontend type.", "Frontend típusának felülbírálata.",
"Owner", "Tulajdonos",
"PEND", "FÜGGŐ",
"PIDs 21 in setup", "PID 21 beállítás alatt",
"PIN inquiry match string", "PIN víszgálatot ellenőrző szó",
"PLS code", "PLS kód",
"PLS mode", "PLS mód",
"Parent entry", "Szülő bejegyzés",
"Pass subscription weight", "Előfizetői súly továbbadása",
"Pass subscription weight to the SAT>IP server (Tvheadend specific extension).", "Előfizetői súly továbbadása a SAT>IP szervereknek (Tvheadend specifikus kiterjesztés).",
"Password", "Jelszó",
"Password2", "Jelszó2",
"Path", "Elérési útvonal",
"Path to a directory (folder) containing your picon collection. See Help for more detailed information.", "A picon gyűjtemény könyvtárára mutató útvonal. (Ld. a súgót további információkért).",
"Path to an icon for this channel. This can be named however you wish, as either a local (file://) or remote (http://) image. See Help for more infomation.", "A csatorna ikonjára mutató útvonal.Ennek elnevezése lehet barmilyen formátumban, helyi fájl (file://) vagy távoli (http://) elérési  útvonalként.  Ld. a súgót további információkért.",
"Performing arts", "Előadóművészet",
"Periodically save EPG to disk (hours)", "EPG lemezre mentése periodikusan(órákban)",
"Persistent user interface level", "Perszisztens felhasználói-felület szint",
"Physiology", "Élettan",
"Picon", "Csatorna ikon",
"Picon path", "Picon útvonal",
"Pilot", "Próba",
"Polarization", "Polarizáció",
"Political issues", "Politikai problémák",
"Popular culture", "Népszerű kultúra",
"Position", "Pozicíó",
"Position of the input.", "Bemenet pozíciója.",
"Post-processor command", "Posztprocesszor parancs",
"Post-recording padding", "Rögzítés utáni ráhagyás",
"Post-remove command", "Utólagos törlés parancs",
"Power save", "Energia takarékos",
"Power-up time (ms) (15-200)", "Bekapcsolás idő (ms) (15-200)",
"Pre-defined muxes", "Előre meghatározott muxok",
"Pre-recording padding", "Rögzítés előtti ráhagyás",
"Pre-school children's programs", "Óvodás programok",
"Prefer picons over channel name", "Csatorna ikonok előnyben részesítése a csatorna nevekkel szemben",
"Preferred CA PID", "Preferált CA PID",
"Preferred service video type", "Preferált szervíz video típus",
"Presentation", "Bemutató",
"Presentation details.", "Megjelenítési részletek.",
"Press", "Sajtó",
"Priority", "Prioritás",
"Priority (-10..10)", "Prioritás (-10..10)",
"Private", "Privát",
"Profile name", "Profilnév",
"Program identification (PID) number to compare. Zero means any. This comparison is processed only when service comparison is active and for the Conditional Access filter.", "(PID) szám összehasonlítására. Nulla jelenése bármennyi. \nEz az összehasonlítás akkor van feldolgozva, amikor a szolgáltatás feldolgozása aktiválva van CAS - korlátozott hozzáférésü rendszer -  filter részére.",
"Provider", "Szolgáltató",
"Provider ID", "Szolgáltató ID",
"Provider network name", "Hálózat neve",
"Psychology", "Pszichológia",
"Puppets", "Bábok",
"Quiz", "Kvíz",
"RAM only", "Csak RAM-ban",
"RTSP port", "RTSP Port",
"RTSP port (554 or 9983, 0 = disable)", "RTSP Port (554 vagy 9983), 0 = letiltás",
"Radio", "Rádió",
"Re-fetch period (hours)", "Újrapróbálkozási idő (órákban)",
"Re-fetch period (mins)", "Újrapróbálkozási idő (percben)",
"Re-record%s%s", "Újrarögzítés %s%s",
"Re-try period (hours)", "Újrapróbálkozási idő (órákban)",
"Read error", "Olvasási hiba",
"Real Time Streaming Protocol (RTSP) port the server should listen on (554 or 9983, 0 = disable).", "A szerver által használt Real Time Streaming Protokol (RTSP) port. (554 vagy 9983, 0=kikapcsolt). ",
"Record all", "Összes rögzítése",
"Recording file options", "Felvett fájl tulajdonságai",
"Recording system path", "Felvétel elérési útvonala",
"Reject", "Tiltás",
"Religion", "Vallás",
"Religious", "Vallási",
"Remarkable people", "Ismert emberek",
"Remove HTTP arguments", "HTTP argumentumok törlése",
"Remove all unsafe characters from filename", "Nem biztonságos karakterek eltávolítása a fájlnévből",
"Repeated", "Ismételt",
"Replace whitespace in title with '-'", "Szóköz '-' re cserélése",
"Reply to CAM PIN inquiries", "Válaszoljon CAM PIN kérésekre",
"Reports", "Jelentések",
"Rerun of", "Újrafuttatása",
"Reserved", "Foglalt",
"Resolution (height)", "Felbontás (magasság)",
"Respawn (pipe)", "Újrakezdése (pipe)",
"Restart", "Újraindítás",
"Restart on error", "Újraindítás hiba esetén",
"Reuse EPG from", "EPG újrafelhasználása",
"Rewrite EIT", "EIT újraírása",
"Rewrite MPEG-TS SI tables", "MPEG-TS SI táblák újraírása",
"Rewrite PAT", "PAT újraírása",
"Rewrite PMT", "PMT újraírása",
"Rewrite SDT", "SDT újraírása",
"Romance", "Romantika",
"Rotor initialization time (seconds)", "Forgató inicializálásának időtartama indítás után (másodperben)",
"Rotor type", "Forgató típusa",
"Run as group", "Futtatás az alábbi felhasználócsportként",
"Run as user", "Futtatás felhasználóként",
"Running", "Futó",
"SAT->IP frontend number.", "SAT->IP frontend szám.",
"SAT>IP Client", "SAT>IP kliens",
"SAT>IP DVB-T frequency (Hz)", "SAT>IP DVB-T frekvencia (Hz)",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "SAT>IP RTSP port szám a szerverhez\n(alapértelmezett: -1 = kikapcsolt, 0 = webconfig, standard port érték 554)",
"SAT>IP source number", "SAT>IP Forrás Száma",
"SD: standard definition", "SD: Normál felbontású",
"SSL verify peer", "SSL pár ellenőrzés",
"Satellite config", "Műhold-beállítás",
"Satellite longitude", "Műhold hosszúsági fok",
"Satellite longitude.", "Műhold hosszúsági fok.",
"Satellite position.", "Műhold-pozíció.",
"Satellite positions", "Műhold-pozíciók",
"Scan after creation", "Szkennelés létrehozás után",
"Scan queue length", "Szkennelési sor hossza",
"Scan result", "Szkennelés eredmény",
"Scan status", "Szkennelés státusz",
"Schedule a re-recording if more errors than (0=off)", "Újrafelvétel ütemezése több mint x észlelt hiba után (0=Ki)",
"Schedule status", "Időzítés státusza",
"Scheduled for recording", "Újravételezésre időzítve",
"Scheduled start time", "Időzítés kezdési időpont",
"Scheduled stop time", "Időzítés leállítási időpont",
"School programs", "Iskolai műsorok",
"Science", "Tudomány",
"Science fiction", "Sci-Fi",
"Season", "Évad",
"Season %d", "%d. Évad",
"Seek error", "Keresési hiba",
"Select the list of languages (in order of priority) to be used for supplying EPG information to clients that don't provide their own configuration.", "Válassza ki az EPG információ szolgáltatásnál használt nyelvet (fontossági sorrendben) ami abban az esetben használt, ahol a kliens nem szolgáltatja a saját konfigurációját",
"Select the master tuner.The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, the polarization and low-high band settings must be equal.if you set other tuner as master, then this tuner will act like a slave one and tvheadend will assure that this tuner will not use incompatible parameters (position, polarization, lo-hi).", "Válassza ki a fő tunert. A jelet az univerzális LNB-től el lehet osztani egy egszerű koaxiális kábellel (nem multiswitch-el) több kimenetre. Ebben az esetben a helyzet, polarizáció valamint alacsony-magas frekvenciatartományok azonosak kell, hogy legyenek. Ha egy másik tunert allítunk be master tunernek, akkor ez a tuner másodlagos tunerként fog viselkedni és a tvheadend biztosíttani fogja, hogy ez a tuner ne hasznájon nem-kompatibilis paramétereket (helyet, polarizáció, frekvenciatartomány)",
"Select the number of satellite positions supported by the SAT>IP hardware and your coaxial cable wiring.", "Válassza ki a SAT>IP hardver által támogatott müholdpoziciők számát és a koaxiális kábelezés tipusát.",
"Send CAPMT query", "Küldjön CAPMT lekérdezést",
"Send full PLAY cmd", "Küldjön tejes LEJÁTSZÁS parancsot",
"Send the full RTSP PLAY command after full RTSP SETUP command. Some devices firmware require this to get an MPEG-TS stream.", "Küldje el a teljes RTSP PLAY parancsot minden teljes RTSP setup parancs után. Egyes eszközök megkövetelik ezt ahhoz, hogy MPEG-TS streamelést biztosítsanak.",
"Serial number", "Gyári szám",
"Series link", "Sorozat hozzáférés",
"Serious", "Komoly",
"Serious music", "Komoly zene",
"Server", "Szerver",
"Server UUID", "Szerver UUID",
"Server connectivity", "Szerverkapcsolat",
"Server details.", "Szerver részletei.",
"Service", "Szolgáltatás",
"Service ID", "Szolgáltatás ID",
"Service Mapper", "Szolgáltatás összerendelő",
"Service configuration", "Szolgáltatás konfiguráció",
"Service count %zi", "Szolgátatások száma %zi",
"Service mapping", "Szolgáltatás összerendelés",
"Service name", "Szolgáltatás neve",
"Service raw", "Szolgáltatás raw",
"Service type", "Szolgáltatás típus",
"Services", "Szolgáltatások",
"Set the display name.", "Név beállítása.",
"Set the maxiumum packet identifiers your SAT>IP server supports.", "A maximális csomagazonosító, amit a SAT>IP szerver támogat.",
"Set the priority of this configuration.", "Ezen beállítás prioritása.",
"Settings", "Beállítások",
"Show this page", "Oldal megjelenítése",
"Show version information", "Verzió információ megjelenítése",
"Show, hide and sort the various details that appear on the interface next to the About tab.", "Mutat, elrejt és rendez különböző részletet, amelyek az interfészen jelennek meg, a Tvheadend-ről fül mellett.",
"Signal quality too poor", "Jelminőség túk gyenge",
"Signal scale (240 or 100)", "Jel slála  (240 vagy 100)",
"Site latitude", "Fekvés szélességi fok",
"Site latitude.", "Helyszín szélességi fok.",
"Site longitude", "Fekvés hosszúsági fok",
"Site longitude.", "Helyszín hosszúsági fok.",
"Size", "Méret",
"Skip TS packets (0-200)", "TS csomagok kihagyása (0-200)",
"Skip TSID checking. For when providers use invalid Transport Stream IDs.", "A TSID ellenőrzés kihagyása. Olyan esetekre, amikor a szolgáltatók érvénytelen Transport Stream ID-t használnak.",
"Skip commercials", "Reklámok kihagyása",
"Skip initial bytes", "Kezdeti x bájt kihagyása",
"Skip x number of transport packets.", "x számú transport csomag elvetése.",
"Slave entry", "Szolga bejegyzés",
"Slot state", "Slot státusz",
"Soap", "Szappanopera",
"Soccer", "Foci",
"Social", "Közösségi",
"Social / Political issues / Economics", "Társadalom / politika / gazdaság",
"Social advisory", "Társadalmi célú hirdetés",
"Sort index", "Indexek rendszerezése",
"Source", "Forrás",
"Source deleted", "Forrás törölve",
"Source quality is bad", "Forrás rossz minőségű",
"Source reconfigured", "Forrás újrabeállítva",
"Southern hemisphere (latitude direction)", "Déli félteke (szélesség irányában)",
"Special events (Olympic Games, World Cup, etc.)", "Különleges események (Olimpia, Világkupák, stb.)",
"Specify User-Agent header for the http client", "Adja meg az User-Agent - felhasználókliens fejlécet a HTTP kliensnek",
"Specify alternative htsp port", "Alternatív htsp port megadása",
"Specify alternative http port", "Alternatív http port megadása",
"Specify alternative http webroot", "Alternatív http webroot megadása",
"Specify bind address", "Bind IP címének megadása",
"Specify extra htsp port", "Alternatív htsp port megadása",
"Spiritual sciences", "Lelki tudományok",
"Sports", "Sportok",
"Sports magazines", "Sportműsorok",
"Squash", "Fallabda",
"Start", "Kezdés",
"Start after", "Kezdés időpont után",
"Start before", "Kezdés időpont előtt",
"Start time", "Kezdés Időpontja",
"Status", "Állapot",
"Status period (ms)", "Állapot periódus (ms-ban)",
"Stereo", "Sztereó",
"Stop", "Leállítás",
"Stop time", "Leállítási idő",
"Storage path", "Tárhely elérési útvonal",
"Storage space", "Tárhely",
"Stream profile", "Stream-profil",
"Stream type", "Stream típus",
"Streaming", "Stream-elés",
"Streaming priority", "Stream-profil",
"Streaming profiles", "Stream-profilok",
"Subdirectory options", "Alkönyvtár opciók",
"Subscribe to a service permanently", "Előfizetés adott szolgáltatásra véglegesen",
"Subscription", "Előfizetés",
"Subscription overridden", "Előfizetés felülbírálva",
"Subscription weight", "Előfizetés súlyzó",
"Substitute formatters", "Helyettesítő formázók",
"Subtitle", "Felirat",
"Subtitle codec", "Felirat kódec",
"Sun", "Vas",
"Surround (2 front, rear mono)", "Térhatás(2 első, monó hát)",
"Switch before rotor", "Switch motor előtt",
"Switch type", "Switch típus",
"Symbol rate (Sym/s)", "Szimbólum sebesség(Sym/s)",
"Sync + Don't keep", "Sync+ne tartsa meg",
"System", "Rendszer",
"TCP Protocol", "TCP protokoll",
"TVHDHomeRun Client", "TVHDHomeRun ügyfél",
"Tag files with metadata", "Fájlok címkézése metaadatokkal",
"Tags", "Címkek",
"Team sports (excluding football)", "Csapat sportok (focit kivéve)",
"Technology", "Technika",
"Teletext stream type is only available for this filter.", "A Teletext stream típus csak ennél a szűrőnél alkalmazható. ",
"Tennis", "Tenisz",
"Testing", "Tesztelés",
"Testing options", "Tesztelési tulajdonságok",
"The ATSC source ID as set by the provider.", "A szolgáltató által beállított ATSC forrás ID.",
"The CA provider to compare. Leave blank to apply to all providers.", "A CA szolgáltató, összehasonlításhoz. Hagyja üresen az összes szolgáltatóra történő alkalmazáshoz.",
"The CA stream type is only available for this filter.", "A CA stream típus csak ennél a szűrőnél alkalmazható. ",
"The CAID to compare. Leave blank to apply to all IDs.", "CAID az összehasonlításhoz. Hagyja üresen az összes ID-ra történő alkalmazáshoz.",
"The Content reference identifier (CRID) authority.", "A Content reference identifier (CRID) hatóság.",
"The MPEGTS stream type is only available for this filter.", "Az MPEGTS stream típus csak ennél a szűrőnél alkalmazható. ",
"The OpenTV channel number as set by the provider.", "Az OpenTV csatorna sorszáma, ahogyan a szolgáltató beállította.",
"The Preferred Conditional Access Packet Identifier. Used for decrypting scrambled streams.", "A Preferred Conditional Access Packet Identifier. Titkosított stream-ek visszafejtéséhez használandó.",
"The SAT>IP server's name.", "A SAT>IP szerver neve.",
"The SAT>IP server's universally unique identifier.", "A SAT>IP szerver általánosan egyedi azonosítója.",
"The SAT>IP's discovered IP address.", "A SAT>IP által megtalált IP-cím.",
"The audio stream types the filter should apply to.", "A szűrő által használt audio stream típusok.",
"The character encoding for this service (e.g. UTF-8).", "Szolgáltatás karakterkódolása (pl. UTF-8).",
"The current boot ID.", "Jelenlegi boot ID.",
"The current configuration ID.", "Jelenlegi beállítási azonosító ID.",
"The current configuration version.", "Jelenlegi beállítás verziója.",
"The default language to use if the user  language isn't set in the Access Entries tab.", "A használandó, alapértelmezett nyelv, amennyiben a felhasználói nyelv nincs beállítva a hozzáférések fülön.",
"The default subscription weight for each subscription.", "Az alapértelmezett előfizetői súlyozás minden egyes előfizetéshez.",
"The device ID.", "Az eszköz ID-je.",
"The device's serial number.", "Az eszköz sorozatszáma.",
"The frequency of the mux (in Hertz).", "A mux frekvenciája (Hertz-ben).",
"The locking mechanism selection for The Preferred Conditional Access Packet Identifier. See Help for more information.", "A használt blokkolási mechanizmus az előnyben részesített CAS csomag azonosítónál. Lásd a Súgót további informácókért.",
"The logical stream index to compare. Note that this index is computed using all filters.Example: If filter is set to AC3 audio type and the language to 'eng' and there are two AC3 'eng' streams in the service, the first stream could be identified using number 1 and the second using number 2.", "Az összehasonlítandó logikai adatfolyam indexe. Ne feledje, hogy ez az index  minden szürő használatával van kiszámítva.\nPélda:  Ha szűrő AC3 audio típusra, a nyelv \"hun\" -ra van beállítva , és  két AC3 \"hun\"  típusú hang adatfolyam van a  szolgáltatásban, akkor az első adatfolyamot 1-es számmal, a másodikat 2-es számmal lehet azonosítani.",
"The manufacturer of the SAT>IP server.", "A SAT>IP szerver gyártója.",
"The maximum number of services to decrypt per mux.", "A mux-onkénti szolgáltatások visszafejtésének száma.",
"The minimum delay before tuning in milliseconds after tuner stop. If the time between the previous and next start is greater than this value then the delay is not applied.", "A minimális késleltetás újrahangolás esetén az előző keresés után.  Ha az idő  az előző és a következő kezdet között meghaladja ezt az értéket, akkor a késleltetés nincs alkalmazva.",
"The modulation used on the mux.", "A mux-on hasznát moduláció.",
"The mux the service is on.", "A szolgáltatás mux-a.",
"The mux's universally unique identifier.", "A mux általánosan egyedi azonosítója.",
"The network the service is on.", "A szolgáltatás hálózata.",
"The networks using this configuration.", "A hálózatok, amelyek ezt a beállítást használják.",
"The number of ATSC-C (Cable/AnnexB) tuners to export.", "Az exportálandó ATSC-C (kábel/AnnexB) tuner-ek száma.",
"The number of ATSC-T (Terresterial) tuners to export.", "Az exportálandó ATSC-T (földfelszíni) tuner-ek száma.",
"The number of DVB-C (Cable) tuners to export.", "Az exportálandó DVB-C (kábel) tuner-ek száma.",
"The number of DVB-C2 (Cable) tuners to export.", "Az exportálandó DVB-C2 (kábel) tuner-ek száma.",
"The number of DVB-S (Satellite) tuners to export.", "Az exportálandó DVB-S (műholdas) tuner-ek száma.",
"The number of DVB-S2 (Satellite) tuners to export.", "Az exportálandó DVB-S2 (műholdas) tuner-ek száma.",
"The number of DVB-T (Terresterial) tuners to export.", "Az exportálandó DVB-T (földfelszíni) tuner-ek száma.",
"The number of DVB-T2 (Terresterial) tuners to export.", "Az exportálandó DVB-T2 (földfelszíni) tuner-ek száma.",
"The number of days cookies set by Tvheadend should expire.", "A Tvheadend által használt sütik lejárati ideje napokban.",
"The provider's name.", "A szolgáltató neve.",
"The provider's network ID.", "A szolgáltató hálózati azonosítója.",
"The provider's network name.", "A szolgáltatói hálózat neve.",
"The rule action defines the operation when all comparisons succeed. See Help for more information on what the various rules do.", "A szabály definiálja a műveletet amennyiben az összehasonlítások sikeresek. További információkért, hogy a különböző szabályok mire használhatóak, nézze meg a súgót.",
"The scan state. New muxes will automatically be changed to the PEND state. You can change this to ACTIVE to queue a scan of this mux.", "A keresés állapota. Az új mux-ok automatikusan PEND állapotba váltanak. Átkapcsolhatja ACTIVE-ba, hogy egy új keresést időzítsen a mux-hoz.",
"The service ID as set by the provider.", "A szolgáltató által beállított service ID.",
"The service name as set by the provider.", "A szolgáltató által beállított szolgáltatás neve.",
"The service the filter should apply to. Leave blank to apply the filter to all services.", "A szolgáltatás amihez a szűrő tartozik.Hagyja üresen az összes szolgáltatásra történő alkalmazáshoz.",
"The service type flag as defined by the DVB specifications (e.g. 0x02 = radio, 0x11 = MPEG2 HD TV, 0x19 = H.264 HD TV)", "A service type flag ahogyan a DVB specifikáció definiálja (pl. 0x02 = rádió, 0x11 = MPEG2 HD TV, 0x19 = H.264 HD TV)",
"The service's channel minor as set by the provider.", "A szolgáltatás csatornájának minor-ja, a szolgáltató beállításai szerint. ",
"The service's channel number as set by the provider.", "A szolgáltatás csatornájának sorszáma, a szolgáltató beállítása alapján.",
"The subtitle stream types the filter should apply to.", "A szűrő által használt felirat stream típusai.",
"The symbol rate.", "Jelsebesség.",
"The transport stream ID of the mux within the network.", "A mux hálózaton belüli transport stream ID-ja.",
"The video stream types the filter should apply to.", "A szűrő által használt videó stream típusai.",
"Theme", "Téma",
"This tuner", "Ez a tuner",
"This will create an NTP driver (using shmem interface) that you can feed into ntpd. This can be run without root privileges, but generally the performance is not that great.", "Ez egy NTP meghajtót hoz létre (az shmem interfész használatával) amit az ntpd-be küldhet. Ez root jogosultságok nélkül futtatható, de a teljesítménye nem túl kiemelkedő.",
"Thu", "Csüt",
"Time", "Idő",
"Time Stamp Fix", "Időbélyegző javítása",
"Time missed", "Kihagyott alkamak",
"Time record caption", "Időrekord vétele",
"Time recording%s%s", "Vételek idelye%s%s",
"Time update", "Idő frissítése",
"Timeout (sec) (0=infinite)", "Időtúllépés (másodperc) (0 = végtelen)",
"Timeout (seconds)", "Időtúllépést (másodperc)",
"Timeout (secs)", "Időtúllépés (másodperc)",
"Timer-based automatic recording.", "Időzítő alapú automata felvétel",
"Timeshift", "Időeltolás",
"Title", "Cím",
"Title (regexp)", "Cím (regexp)",
"Title of the program.", "Műsor címe.",
"Toggle details", "Részletek közötti váltás",
"Total number of services found.", "A megtalált szolgáltatások teljes száma.",
"Total number of services.", "A szolgáltatások teljes száma.",
"Tourism / Travel", "Túrizmus / Utazás",
"Trace subsystems", "Trace alrendszerek ",
"Traditional arts", "Hagyományos művészetek",
"Traditional music", "Hagyományos zene",
"Transcode", "Átkódolás",
"Transcode/av-lib", "Átkódolás/av-lib",
"Transcoding", "Átkódolás",
"Transmission mode", "Átviteli mód",
"Transponder ID", "Transzponder ID",
"Transport stream ID", "Transzport stream ID",
"Tue", "Ke",
"Tune before DiseqC", "Hangolás  DiseqC előtt",
"Tuner bind IP address", "Tuner bind IP-címe",
"Tuner configuration", "Tuner beállítás",
"Tuner configuration.", "Tuner beállítás",
"Tuning failed", "Sikertelen hangolás ",
"Turn off LNB when idle", "LNB lekapcsolása üresjáraton",
"Tvheadend server name", "Tvheadend szerver neve",
"Type", "Típus",
"UDP RTP port number (2 ports)", "UDP RTP port száma (2 port)",
"UPnP Protocol", "UPnP protokoll",
"URL with the SAT>IP server XML location", "SAT> IP szerver  XML eléresének URL-je",
"Unable to open file", "Fájl megnyitása nem lehetséges",
"Uncommitted first", "Uncommitted előbb",
"Unicable I switch (universal LNB)", "Unicable I kapcsoló (univerzális LNB)",
"Unicable type", "Unicable típus",
"Unimportant", "Jelentékelen",
"Universal LNB only", "Csak univerzális LNB ",
"Universally unique identifier. Read only.", "Általánosan egyedi azonosító. Csak olvasható.",
"Unknown", "Ismeretlen",
"Unknown file", "Ismeretlen file",
"Unknown reason (%i)", "Ismeretlen ok (%i)",
"Unlimited size", "Végtelen méret",
"Unlimited time", "Korlátlan idő",
"Unset (default)", "Nem beállított(alapértelmezett)",
"Update card (EMM)", "Kártya (EMM) frissítése",
"Update channel icon", "Csatornaikon frissítése",
"Update channel name", "Csatornanév frissítése",
"Update channel number", "Csatornaszám frissítése",
"Update disabled", "Frissítés kikapcsolva",
"Update time", "Idő frissítése",
"Update tolerance (ms)", "Frissítési tolerancia (ms)",
"Updated", "Frissített",
"Usage: %s [OPTIONS]\n", "Használat:%s [opciók]\n\n",
"Use AC-3 detection.", "AC-3 felismerés használata.",
"Use EPG running state", "Következtessen az EPG állapotából",
"Use WEBM format.", "WEBM formátum használata",
"Use Windows-compatible filenames", "Használjon Windows-kompatibilis fájlneveket",
"Use XSPF playlist instead of M3U", "Használjon XSPF lejátszás listát  M3U helyett",
"Use a pre-defined list of ATSC-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Egy előre definiált ATSC-C mux-lista használata. Megjegyzés: Ezek a listák néha régiek lehetnek és azt eredményezhetik, hogy a keresés hosszabb időt vesz igénybe, mint általában.",
"Use a pre-defined list of ATSC-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Egy előre definiált ATSC-T mux-lista használata. Megjegyzés: Ezek a listák néha régiek lehetnek és azt eredményezhetik, hogy a keresés hosszabb időt vesz igénybe, mint általában.",
"Use a pre-defined list of DVB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Egy előre definiált DVB-C mux-lista használata. Megjegyzés: Ezek a listák néha régiek lehetnek és azt eredményezhetik, hogy a keresés hosszabb időt vesz igénybe, mint általában.",
"Use a pre-defined list of DVB-S/S2 muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Egy előre definiált DVB-S/S2 mux-lista használata. Megjegyzés: Ezek a listák néha régiek lehetnek és azt eredményezhetik, hogy a keresés hosszabb időt vesz igénybe, mint általában.",
"Use a pre-defined list of DVB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Egy előre definiált DVB-T mux-lista használata. Megjegyzés: Ezek a listák néha régiek lehetnek és azt eredményezhetik, hogy a keresés hosszabb időt vesz igénybe, mint általában.",
"Use original", "Eredeti használata",
"Use service IDs as channel numbers", "Használja a szolgáltatás azonosítókat csatornaszámként",
"User access error", "Felhasználó hozzáférési hiba",
"User icon", "Felhasználoi ikon",
"User interface level", "Felhasználófelület szint",
"User language", "Felhasználó nyelve",
"User limit reached", "Felhasználói korlát elérve",
"User login", "Felhasználói belépés",
"User request", "Felhasználói kérés",
"Username", "Felhasználónév",
"Users - Access Entries", "Felhasználók - Hozzáférés szabályozási adatok",
"Users - IP Blocking", "Felhasználók - IP blokkolás",
"Users - Passwords", "Felhasználók - jelszavak",
"Variety show", "Varieté show",
"Verify the SSL certificate.", "SSL tanúsitvány ellenőrzése.",
"Video Stream Filter", "Videofolyam szűrő",
"Video bitrate (kb/s) (0=auto)", "Video bitráta (kb/s) (0=Auto)",
"Video codec", "Video kódec",
"Video codec preset", "Video kodek előbeállítás",
"Video recorder", "Videó-felvevő",
"View all", "Összes megtekintése",
"Visually impaired commentary/audio description", "Kommentár látássérülteknek",
"Vorbis audio", "Vorbis audió",
"WEBM/built-in", "WEBM/beépített",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV profil H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV profil H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV profil VP8/Vorbis/WEBM",
"Waiting for EPG running flag", "Várjon az EPG státusz flag-re",
"Waiting for program start", "Várakozás a műsorkezdésre",
"Waiting for stream", "Stream-re várakozás",
"War", "Háború",
"Water sport", "Vizisport",
"Weak stream", "Gyenge stream",
"Weather report", "Időjárás-jelentés",
"Web User Interface", "WEB-es felhasználói felület",
"Web interface", "Webes felület",
"Web interface language", "Webes felület nyelve",
"Web interface language.", "Webes felület nyelve.",
"Web interface theme.", "Webes felület témája.",
"Web theme", "Web téma",
"Web user interface", "WEB-es felület",
"Wed", "Sze",
"Welcome - Tvheadend - your TV streaming server and video recorder", "Üdvözöljük - Tvheadend - az Ön streaming szervere és videórögzítője",
"Western hemisphere (latitude direction)", "Nyugati félteke (szélesség irányában)",
"When the service was first identified and recorded.", "Amikor a szolgáltatás először került azonosításra és bejegyzésre.",
"When the service was last seen during a mux scan.", "Amikor a szolgáltatás utoljára volt látható egy mux keresésnél.",
"Winter sports", "Téli sportok",
"Wizard", "Varázsló",
"Wrapper (capmt_ca.so)", "Illesztő (capmt_ca.so)",
"Write a short message to log identifying the matched parameters. It is useful for debugging your setup or structure of incoming streams.", "Írjon egy rövid üzenetet a loghoz, a megtalált paramétereket azonosítandó. Hasznos a beállítások hibakeresésénél és a bejövő stream-ek szerkezeténél.",
"Writes the current in-memory EPG database to disk every x hours, so should a crash/unexpected shutdown occur EPG data is saved periodically to the database (re-read on next startup). Set to 0 to disable.", "Írja ki  a jelenleg memóriában tárolt EPG adatbázist lemezre  x óránként, így  egy  váratlan leállás esetében az EPG adatok időszakosan ki lesznek menteve a tárolóba (a következő induláskor újra be lesz töltve). Ha az érték 0, akkor le van tiltva.",
"XMLTV EPG Import", "XMLTV EPG importálás",
"Yes", "Igen",
"addpids/delpids supported", "addpids/delpids támogatva",
"default: nvenc(h264 / h265)", "alapértelmezett: nvenc(h264 / h265)",
"fast: h264 / h265 / qsv(h264 / h265)", "gyors: h264 / h265 / qsv(h264 / h265)",
"faster: h264 / h265 / qsv(h264)", "gyorsabb: h264 / h265 / qsv(h264)",
"hrs", "óra",
"invalid option specified [%s]", "Érvénytelen beállítás lett megadva [%s]",
"logout", "kilépés",
"medium: h264 / h265 / qsv(h264 / h265)", "közepes: h264 / h265 / qsv(h264 / h265)",
"min", "perc",
"mins", "percek",
"option %s requires a value", "%s opció értéket igényel",
"return", "vissza",
"slow: h264 / h265 / qsv(h264 / h265)", "lassú: h264 / h265 / qsv(h264 / h265)",
"slower: h264 / h265 / qsv(h264)", "lassabb: h264 / h265 / qsv(h264)",
"superfast: h264 / h265", "szuper gyors: h264 / h265",
"tsfile input (mux file)", "tsfile bement (mux fájl)",
"tvhdhomerun client", "tvhdhomerun ügyfél",
"ultrafast: h264 / h265", "ultra gyors: h264 / h265",
"veryfast: h264 / h265 / qsv(h264)", "nagyon gyors: h264 / h265 / qsv(h264)",
"veryslow: h264 / h265 / qsv(h264)", "nagyon lassú: h264 / h265 / qsv(h264)",
NULL, NULL
};

static const char *tvh_locale_19[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nPer maggiori informazioni consulta il sito web di Tvheadend:\nhttps://tvheadend.org\n",
"# Mapped channels", "# Canali mappati",
"# Muxes", "# Mux",
"# Services", "# Servizi",
"# tune repeats", "# ripetizioni sintonia",
"(Default profile)", "(Profilo predefinito)",
"1 day", "1 giorno",
"1 month", "1 mese",
"1 week", "1 settimana",
"1 year", "1 anno",
"2 months", "2 mesi",
"2 weeks", "2 settimane",
"2 years", "2 anni",
"2-Port switch (universal LNB)", "Commutatore a 2 porte (LNB universale)",
"3 days", "3 giorni",
"3 months", "3 mesi",
"3 weeks", "3 settimane",
"3 years", "3 anni",
"4-Port switch (universal LNB)", "Commutatore a 4 porte (LNB universale)",
"5 days", "5 giorni",
"6 months", "6 mesi",
"AC-3 = descriptor 6", "AC-3 = descrittore 6",
"AC-3 detection", "Rilevamento AC-3",
"ACTIVE", "ATTIVO",
"ANY", "QUALUNQUE",
"ATSC source ID", "ID sorgente ATSC",
"Aborted by user", "Abortito dall'utente",
"Access", "Accesso",
"Action", "Azione",
"Adapter in use by another subscription", "Adattatore in uso da un'altra sottoscrizione",
"Add file and line numbers to debug", "Aggiungi file e numeri di linea per il debug",
"Add the thread ID to debug", "Aggiungi l'ID del thread per il debug",
"Admin", "Amministrazione",
"Admin password", "Password amministratore",
"Admin username", "Nome utente amministratore",
"Administrator login", "Accesso amministratore",
"Adult movie", "Film per adulti",
"Advanced", "Avanzato",
"Advanced (non-universal LNBs, rotors, etc.)", "Avanzato (LNB non-universali, rotori, ecc.)",
"Adventure", "Avventura",
"Advertisement / Shopping", "Pubblicità / Acquisti",
"All (Streaming plus DVR)", "Tutto (Streaming e DVR)",
"All proceeds are used to support server infrastructure and buy test equipment.", "Tutti i ricavati sono usati per il supporto all'infrastruttura server e per l'acquisto di equipaggiamenti da supportare",
"Alternate PID path", "Percorso PID alternativo",
"Alternate configuration path", "Percorso di configurazione alternativo",
"Altitude (meters)", "Altitudine (metri)",
"Animals", "Animali",
"Anonymize HTSP access", "Anonimizza accesso HTSP",
"Any", "Qualunque",
"Arts", "Arti",
"Athletics", "Atletica",
"Audio bitrate (kb/s) (0=auto)", "Bitrate audio (kb/s) (0=automatico)",
"Audio channel layout.", "Layout canale audio.",
"Audio codec", "Codec audio",
"Auto", "Automatico",
"Auto check disabled", "Controllo automatico disabilitato",
"Auto check enabled", "Controllo automatico abilitato",
"Auto record", "Registrazione automatica",
"Auto record caption", "Didascalia della registrazione automatica",
"Auto recording%s%s", "Auto registrazione%s%s",
"Auto time record", "Registrazione a tempo automatica",
"Auto-Map to channels", "Auto-Mappatura sui canali",
"Automatic checking", "Controllo automatico",
"Automatically map EPG source", "Mappa automaticamente la sorgente EPG",
"Automatically name from network", "Nomina automaticamente dalla rete",
"Automatically record.", "Registra automaticamente.",
"Autorec maximum count (0=unlimited)", "Conteggio massimo per auto-registrazione (0=illimitato)",
"Autorec maximum schedules limit (0=unlimited)", "Limite massimo pianificazioni per auto-registrazione (0=illimitato)",
"Bad", "Cattivo",
"Ballet", "Balletto",
"Bandwidth", "Ampiezza di banda",
"Based on software from", "Basato su software da",
"Basic", "Semplice",
"Boot ID", "ID di avvio",
"Bouquet", "Pacchetto",
"Bouquet (auto)", "Pacchetto (automatico)",
"Brand", "Marchio",
"Broadcast", "Trasmissione",
"Broadcasting", "Telecomunicazioni",
"Build", "Versione",
"CA ID", "ID CA",
"CA client %i", "Cliente CA %i",
"CA identification", "Identificazione CA",
"CA provider", "Fornitore CA",
"CAPMT interval (ms)", "Intervallo CAPMT (ms)",
"CAPMT query interval (ms)", "Intervallo richiesta CAPMT (ms)",
"CRID authority", "Autorità CRID",
"Cache scheme", "Schema della cache",
"Camd.socket filename / IP Address (TCP mode)", "Nome del file camd.socket / indirizzo IP (modalità TCP)",
"Cartoons", "Cartoni",
"Channel", "Canale",
"Channel does not exist", "Canale non esistente",
"Channel icon", "Icona del canale",
"Channel name", "Nome del canale",
"Channel not found", "Canale non trovato",
"Channel number", "Numero del canale",
"Channel number offset", "Scostamento del numero del canale",
"Channel numbers from", "Numeri dei canali da",
"Channel tag", "Etichetta del canale",
"Channel tag reference", "Riferimento dell'etichetta di canale",
"Channel tags", "Etichette del canale",
"Channels", "Canali",
"Character set", "Insieme dei caratteri",
"Check availability", "Verifica disponibilità",
"Children's / Youth programs", "Trasmissione per bambini e per giovani",
"Class", "Classe",
"Classical", "Classico",
"Classical music", "Musica classica",
"Clean effects", "Privo di effetti",
"Client name", "Nome del cliente",
"Clone scheduled entry on error", "Clona la voce programmata su errore",
"Comedy", "Commedia",
"Command delay time (ms) (10-200)", "Tempo di ritardo del comando (ms) (10-200)",
"Command time (ms) (10-100)", "Tempo del comando (ms) (10-100)",
"Comment", "Commento",
"Commercial break", "Interruzione pubblicitaria",
"Committed", "Vincolato",
"Completed OK", "Completato OK",
"Configuration", "Configurazione",
"Configuration ID", "ID di configurazione",
"Configuration name", "Nome della configurazione",
"Configuration version", "Versione della configurazione",
"Connection limit type", "Tipo di limite della connessione",
"Constellation", "Costellazione",
"Container", "Contenitore",
"Content character set", "Insieme dei caratteri del contenuto",
"Content type", "Tipo di contenuto",
"Contest", "Competizione",
"Cookie expiration (days)", "Scadenza cookie (giorni)",
"Cooking", "Cucina",
"Copy codec type", "Copia il tipo di codec",
"Copy layout", "Copia la disposizione",
"Create bouquet", "Crea pacchetto",
"Created", "Creato",
"Creator", "Creatore",
"Cron", "Cronologia",
"Cron multi-line", "Cronologia multi-linea",
"Culture (without music)", "Cultura (senza musica)",
"Current RTSP port.", "Porta RTSP attuale.",
"Current affairs", "Vicende di attualità",
"Custom HTTP headers", "Intestazioni HTTP personalizzate",
"DBus - use the session message bus instead of the system one", "DBus - utilizza il bus dei messaggi di sessione al posto di quello di sistema",
"DES key", "Chiave DES",
"DSCP/TOS for streaming", "DSCP/TOS per lo streaming",
"DVB EPG ID", "ID dell'EPG DVB",
"DVB scan files", "File di scansione DVB",
"DVB scan files path", "Percorso dei file di scansione DVB",
"DVR behavior", "Comportamento DVR",
"DVR configuration", "Configurazione DVR",
"DVR configuration profiles", "Profili di configurazione DVR",
"DVR file retention period", "Tempo di conservazione del file DVR",
"DVR log retention", "Conservazione nel registro DVR",
"DVR log retention period", "Tempo di conservazione nel registro DVR",
"DVR override: high", "Alta con prevalenza DVR",
"DVR override: important", "Prevalenza DVR: importante",
"DVR override: low", "Prevalenza DVR: bassa",
"DVR override: normal", "Prevalenza DVR: normale",
"DVR override: unimportant", "Prevalenza DVR: non importante",
"Dance", "Danza",
"Data errors", "Errori nei dati",
"Days of Week", "Giorni della settimana",
"Days of week", "Giorni della settimana",
"Debate", "Dibattito",
"Debug libav log", "Registro debug libav",
"Debug options", "Opzioni di debug",
"Debug subsystems", "Sottosistemi di debug",
"Debug to syslog", "Debug su syslog",
"Debug trace (low-level)", "Tracciatura di debug (basso livello)",
"Default", "Predefinito",
"Default language(s)", "Lingua/e predefinita/e",
"Default priority", "Priorità predefinita",
"Delivery system", "Sistema di consegna",
"Demux path", "Percorso demux",
"Descramble services (limit per mux)", "Servizi di decodifica (limite per mux)",
"Descrambler buffer (TS packets)", "Buffer di decodifica (pacchetti TS)",
"Description", "Descrizione",
"Detective", "Poliziesco",
"Device ID", "ID del dispositivo",
"Device model", "Modello del dispositivo",
"Device path", "Percorso del dispositivo",
"Directory", "Cartella",
"Directory permissions (octal, e.g. 0775)", "Permessi della cartella (ottale, per es. 0775)",
"Disable", "Disabilita",
"Disable DVB bouquets", "Disabilita pacchetti DVB",
"Disable SAT>IP client", "Disabilita cliente SAT>IP",
"Disable all access control checks", "Disabilita tutte le verifiche sui controlli d'accesso",
"Disable debug on stderr", "Disabilita debug su stderr",
"Disable device/firmware-specific workarounds", "Disabilita tutti gli espedienti specifici a dispositivo/firmware",
"Disable syslog (all messages)", "Disabilita syslog (tutti i messaggi)",
"Disabled", "Disabilitato",
"Discussion", "Dibattito",
"DiseqC repeats", "Ripetizioni DiseqC",
"Do not use", "Non utilizzare",
"Documentary", "Documentario",
"Don't backup configuration tree at upgrade", "Non effettuare copia di riserva dell'albero di configurazione nei passaggi di versione",
"Don't include title in filename", "Non includere il titolo nel nome del file",
"Don't keep", "Non mantenere",
"Don't re-record", "Non ri-registrare",
"Don't reschedule", "Non ripetere",
"Drama", "Drammatico",
"Duplicate handling", "Gestione dei duplicati",
"EIT - skip TSID check", "EIT - salta controllo TSID",
"EPG name", "Nome EPG",
"EPG scan", "Scansione EPG",
"EPG source", "Sorgente EPG",
"EPG update window", "Finestra di aggiornamento EPG",
"Economics", "Economia",
"Education", "Educazione",
"Educational", "Educativo",
"Elementary stream filter", "Filtro dello stream elementare",
"Enable (auto)", "Abilita (automatico)",
"Enable DBus", "Abilita DBus",
"Enable NTP driver", "Abilita driver NTP",
"Enable coredumps for daemon", "Abilita coredump per demone",
"Enable debug on stderr", "Abilita debug su stderr",
"Enable debug subsystems", "Abilita sottosistemi di debug",
"Enable debug to file", "Abilita debug su file",
"Enable debug to syslog", "Abilita debug su syslog",
"Enable syslog", "Abilita syslog",
"Enable trace subsystems", "Abilita sottosistemi di tracciatura",
"Enable web UI debug (non-minified JS)", "Abilita debug della web UI (JS non minimizzato)",
"Enabled", "Abilitato",
"Encrypted", "Cifrato",
"Encrypted service", "Servizio cifrato",
"Entertainment programs for 10 to 16", "Trasmissione di intrattenimento da 10 a 16 anni",
"Entertainment programs for 6 to 14", "Trasmissione di intrattenimento da 6 a 14 anni",
"Environment", "Ambiente",
"Environment (pipe)", "Ambiente (pipe)",
"Episode", "Episodio",
"Episode %d", "Episodio %d",
"Error code", "Codice di errore",
"Errors", "Errori",
"Even key", "Chiave pari",
"Every day", "Ogni giorno",
"Exclude channel tags", "Escludi le etichette dei canali",
"Expeditions", "Spedizioni",
"Experimental film", "Film sperimentali",
"Expert", "Esperto",
"Exported tuners", "Sintonizzatori esportati",
"External", "Esterno",
"External EPG grabber", "Estrattore EPG esterno",
"External IP (NAT)", "IP esterno (NAT)",
"External URL", "URL esterno",
"Extra arguments", "Argomenti supplementari",
"Extra start time", "Tempo di inizio supplementare",
"Extra stop time", "Tempo di termine supplementare",
"Extra warming up time (seconds)", "Tempo supplementare di riscaldamento (secondi)",
"FAIL", "FALLIMENTO",
"FEC high", "FEC alto",
"FEC low", "FEC basso",
"Factual topics", "Tematiche fattuali",
"Fantasy", "Fantasia",
"Fast input switch", "Commutazione veloce dell'ingresso",
"File missing", "File mancante",
"File not created", "File non creato",
"File permissions (octal, e.g. 0664)", "Permessi del file (ottale, per es. 0664)",
"File size", "Dimensione del file",
"Filename", "Nome del file",
"Filename character set", "Insieme dei caratteri per il nome del file",
"Filename options", "Opzioni per il nome del file",
"Fine arts", "Belle arti",
"Finished", "Terminato",
"Fitness and health", "Forma fisica e salute",
"Folk", "Folclore",
"Folkloric", "Folcloristico",
"Football", "Calcio",
"Force (auto)", "Forza (automatico)",
"Force CA ID (e.g. 0x2600)", "Forza ID CA (per es. 0x2600)",
"Force old status", "Forza vecchio stato",
"Force pilot for DVB-S2", "Forza pilot per DVB-S2",
"Force priority", "Forza priorità",
"Force teardown delay", "Ritardo dell'abbattimento forzato",
"Foreign countries", "Paesi stranieri",
"Forever", "Per sempre",
"Fork and run as daemon", "Separa ed esegui come demone",
"Format string", "Stringa di formato",
"Frequency", "Frequenza",
"Frequency (Hz)", "Frequenza (Hz)",
"Frequency (MHz)", "Frequenza (MHz)",
"Frequency (kHz)", "Frequenza (kHz)",
"Fri", "Ven",
"Friendly name", "Nome descrittivo",
"Frontend number", "Numero del frontend",
"Frontend path", "Percorso del frontend",
"Full DiseqC", "DiseqC pieno",
"Full mux RX mode supported", "Modalità RX piena del mux supportata",
"Full pathname specification", "Specifica del percorso completo",
"Full-text", "Testo completo",
"Further education", "Altri tipi educativi",
"GOTOX position", "Posizione GOTOX",
"Game show", "Quiz televisivo",
"Gardening", "Giardinaggio",
"General", "Generale",
"General configuration", "Configurazione generale",
"Generic", "Generico",
"Generic options", "Opzioni generiche",
"Guard interval", "Intervallo di guardia",
"HD: high definition", "HD: alta definizione",
"HDHomeRun DVB frontend", "Frontend HDHomeRun DVB",
"HDHomeRun DVB-C frontend", "Frontend HDHomeRun DVB-C",
"HDHomeRun DVB-T frontend", "Frontend HDHomeRun DVB-T",
"HTSP Default Stream Settings", "Impostazioni dello stream predefinito HTSP",
"HTTP CORS origin", "Origine HTTP CORS",
"Handicraft", "Artigianato",
"Hearing impaired", "Non udenti",
"Hierarchy", "Gerarchia",
"High", "Alto",
"High bitrate mode (CI+ CAMs only)", "Modalità ad alto bitrate (solo CAM CI+)",
"Historical movie", "Film storici",
"Horror", "Orrore",
"Hostname/IP", "Nome della macchina/IP",
"IDLE", "INATTIVO",
"IP address", "Indirizzo IP",
"IP blocking", "Blocco IP",
"IPTV Multiplex", "Multiplex IPTV",
"IPTV input", "Ingresso IPTV",
"ISI (Stream ID)", "ISI (ID Stream)",
"Icon", "Icona",
"Icon (full URL)", "Icona (URL completo)",
"Icon URL", "URL dell'Icona",
"Icon base URL", "URL di base icone",
"Icon has title", "Icona ha un titolo",
"Icons from", "Icone da",
"Idle", "Inattivo",
"Idle scan", "Scansione su inattività",
"Idle scan muxes", "Scansione su inattività dei mux",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Se non esiste ancora un account per utente allora\ncreane uno senza nome utente e senza password.\nUtilizzalo con cautela poiché esso permetterà a\nchiunque di avere un accesso di tipo amministrativo \ndella tua installazione Tvheadend finché non creerai o\nmodificherai il controllo d'accesso dall'interno\ndell'interfaccia web di Tvheadend.",
"If you'd like to support the project, please consider a donation.", "Se volessi supportare il progetto, per piacere considera l'idea di una donazione.",
"Ignore EPG (EIT)", "Ignora EPG (EIT)",
"Ignore descriptor 5", "Ignora descrittore 5",
"Ignore invalid SSL certificate", "Ignora certificati SSL non validi",
"Ignore provider's channel numbers", "Ignora i numeri di canale del fornitore",
"Immediately abort", "Abortisci immediatamente",
"Important", "Importante",
"Include channel name in filename", "Includi il nome del canale del nome del file",
"Include date in filename", "Includi la data nel nome del file",
"Include episode in filename", "Includi l'episodio nel nome del file",
"Include subtitle in filename", "Includi il sottotitolo nel nome del file",
"Include time in filename", "Includi l'orario nel nome del file",
"Index", "Indice",
"Information area", "Area informativa",
"Informational", "Informativo",
"Initial scan", "Scansione iniziale",
"Input base", "Base dell'ingresso",
"Input buffer (bytes)", "Buffer d'ingresso (byte)",
"Input instance", "Istanza d'ingresso",
"Input path", "Percorso d'ingresso",
"Interface", "Interfaccia",
"Internal", "Interno",
"Internal EPG grabber", "Estrattore EPG interno",
"Internal grabber", "Estrattore interno",
"Interview", "Intervista",
"Invalid", "Non valido",
"Invalid adapter number '%s'\n", "Adattatore numero '%s' non valido\n",
"Invalid target", "Obiettivo non valido",
"Keep", "Mantieni",
"Kill signal (pipe)", "Segnale kill (pipe)",
"Kill timeout (pipe/secs)", "Pausa kill (pipe/sec)",
"LNA (low noise amplifier)", "LNA (amplificatore a basso rumore)",
"LNB type", "Tipo LNB",
"Language", "Lingua",
"Language 1", "Lingua 1",
"Language 2", "Lingua 2",
"Language 3", "Lingua 3",
"Language settings", "Impostazioni per la lingua",
"Languages", "Lingue",
"Last seen", "Visto ultima volta",
"Last updated from", "Ultimo aggiornamento da",
"Leisure hobbies", "Passatempi per il tempo libero",
"Limit connections", "Limite delle connessioni",
"Linked input", "Ingresso collegato",
"Linux DVB CA", "CA DVB Linux",
"Linux DVB frontend", "Frontend DVB Linux",
"LinuxDVB adapter", "Adattatore LinuxDVB",
"LinuxDVB network", "Rete LinuxDVB",
"Listen / Connect port", "Porta di ascolto/connessione",
"Listen on IPv6", "Ascolta su IPv6",
"Literature", "Letteratura",
"Local bind IP address", "Indirizzo IP di collegamento locale",
"Local channel minor", "Minore locale del canale",
"Local channel number", "Numero locale del canale",
"Local discovery IP address", "Scoperta locale dell'indirizzo IP",
"Location", "Locazione",
"Lock preferred CA PID", "Blocca PID CA preferito",
"Log", "Registra",
"Login/Logout", "Entrata/Uscita",
"Low", "Basso",
"MPEG-TS Pass-thru", "Pass-thru MPEG-TS",
"MPEG-TS Pass-thru/built-in", "Pass-thru/built-in MPEG-TS",
"MPEG-TS input", "Ingresso MPEG-TS",
"MPEG-TS multiplex PHY", "Multiplex PHY MPEG-TS",
"MPEG-TS raw service", "Servizio grezzo MPEG-TS",
"Magazines", "Riviste",
"Maintained space", "Spazio mantenuto",
"Make subdirectories per channel", "Crea sottocartelle per canale",
"Make subdirectories per day", "Crea sottocartelle per giorno",
"Make subdirectories per title", "Crea sottocartelle per titolo",
"Manufacturer", "Produttore",
"Manufacturer URL", "URL del produttore",
"Map radio channels", "Mappa i canali radio",
"Map unnamed channels", "Mappa canali senza nome",
"Map zero-numbered channels", "Mappa canali con numerazione nulla",
"Master tuner", "Sintonizzatore principale",
"Matroska specific", "Specifiche Matroska",
"Maximal channel number", "Numero massimo del canale",
"Maximum # input streams", "Numero massimo di stream in ingresso",
"Maximum PIDs", "Numero massimo di PID",
"Maximum RAM size (MB)", "Dimensione massima RAM (MB)",
"Maximum bandwidth (Kbps)", "Ampiezza di banda massima (Kbps)",
"Maximum count (0=default)", "Conteggio massimo (0=predefinito)",
"Maximum duration", "Durata massima",
"Maximum length of PIDs", "Massima lunghezza dei PID",
"Maximum period (mins)", "Periodo massimo (minuti)",
"Maximum schedules limit (0=default)", "Limite massimo pianificazioni (0=predefinito)",
"Maximum size (MB)", "Dimensione massima (MB)",
"Maximum timeout (seconds)", "Pausa massima (secondi)",
"Medicine", "Medicina",
"Melodrama", "Melodramma",
"Minimal channel number", "Numero minimo del canale",
"Minimum duration", "Durata minima",
"Minimum rotor time (seconds)", "Tempo minimo del rotore (secondi)",
"Missing In PAT/SDT", "Mancante in PAT/SDT",
"Mode", "Modalità",
"Model description", "Descrizione del modello",
"Model name", "Nome del modello",
"Model number", "Numero del modello",
"Modulation", "Modulazione",
"Module", "Modulo",
"Module ID", "ID del modello",
"Mon", "Lun",
"More verbose libav log", "Registro più dettagliato di libav",
"Motor rate (milliseconds/deg)", "Rapporto del motore (ms/grado)",
"Motor sport", "Sport di motori",
"Motoring", "Motoristico",
"Movie", "Film",
"Music", "Musica",
"Mux UUID", "UUID del mux",
"Mux handling", "Gestione dei mux",
"Mux name", "Nome del mux",
"Mux not enabled", "Mux non abilitato",
"NONE", "NESSUNO",
"Name", "Nome",
"Names", "Nomi",
"Natural sciences", "Scienze naturali",
"Nature", "Natura",
"Network", "Rete",
"Network A", "Rete A",
"Network B", "Rete B",
"Network ID (limit scanning)", "ID di rete (limitare scansione)",
"Network UUID", "UUID della Rete",
"Network discovery", "Scoperta rete",
"Network name", "Nome della rete",
"Network prefix", "Prefisso della rete",
"Network type", "Tipo della rete",
"Networks", "Reti",
"News", "Notizie",
"News magazine", "Rivista di attualità",
"Next tune delay in ms (0-2000)", "Ritardo prossima sintonia in ms (0-2000)",
"No PIN", "Nessun PIN",
"No access", "Nessun accesso",
"No adapters specified!\n", "Nessun adattatore specificato!\n",
"No assigned adapters", "Adattatori non assegnati",
"No days", "Nessun giorno",
"No descrambler", "Nessun decodificatore",
"No free adapter", "Nessun adattatore disponibile",
"No input detected", "Nessun ingresso rilevato",
"No service assigned to channel", "Nessun servizio assegnato al canale",
"No service enabled", "Nessun servizio abilitato",
"No source available", "Nessuna sorgente disponibile",
"None", "Nessuno",
"Normal", "Normale",
"Not linked", "Non collegato",
"Not set", "Non impostato",
"Not set (none or channel configuration)", "Non impostato (nessuno o configurazione di canale)",
"Not set (use channel or DVR configuration)", "Non impostato (usa configurazione di canale o DVR)",
"Number", "Numero",
"Number of tsfile tuners", "Numero di sintonizzatori tsfile",
"OK (partial)", "OK (parziale)",
"OSCam new pc-nodmx (rev >= 10389)", "OSCam nuovo pc-nodmx (rev >= 10389)",
"Odd key", "Chiave dispari",
"Off", "Spento",
"Older OSCam", "OSCam più vecchie",
"On", "Acceso",
"On file removal", "Su rimozione del file",
"Only Bulsatcom 39E", "Solo Bulsatcom 39E",
"Only EIT", "Solo EIT",
"Only OpenTV Sky Ausat", "Solo OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Solo OpenTV Sky Italia",
"Only OpenTV Sky UK", "Solo OpenTV Sky UK",
"Only PSIP (ATSC)", "Solo PSIP (ATSC)",
"Only UK Freesat", "Solo UK Freesat",
"Only UK Freeview", "Solo UK Freeview",
"Only Viasat Baltic", "Solo Viasat Baltic",
"Only preferred CA PID", "Solo il PID CA preferito",
"OpenTV channel number", "Numero OpenTV del canale",
"Orbital position", "Posizione orbitale",
"Orbital positions", "Posizioni orbitali",
"Original network ID", "ID originale della rete",
"Output directory for tsdebug", "Cartella di uscita per tsdebug",
"Over-the-air Cron multi-line", "Cronologia multi-linea over-the-air",
"Over-the-air EPG", "EPG over-the-air",
"Over-the-air EPG grabber", "Estrattore EPG over-the-air",
"Over-the-air grabbers", "Estrattori over-the-air",
"Owner", "Proprietario",
"PEND", "IN ATTESA",
"PIDs 21 in setup", "PID 21 nell'impostazione",
"PIN inquiry match string", "Stringa corrispondente alla richiesta PIN",
"PLP ID", "ID PLP",
"PLS code", "Codice PLS",
"PLS mode", "Modalità PLS",
"Parent entry", "Voce genitore",
"Path", "Percorso",
"Performing arts", "Arti dello spettacolo",
"Periodically save EPG to disk (hours)", "Salvataggio periodico dell'EPG su disco (ore)",
"Persistent user interface level", "Livello dell'interfaccia utente persistente",
"Physiology", "Fisiologia",
"Polarization", "Polarizzazione",
"Political issues", "Questioni politiche",
"Popular culture", "Cultura popolare",
"Port", "Porta",
"Position", "Posizione",
"Post-processor command", "Comando di post-processamento",
"Post-recording padding", "Riempimento post-registrazione",
"Post-remove command", "Comando di post-rimozione",
"Power save", "Risparmio energetico",
"Power-up time (ms) (15-200)", "Tempo di accensione (ms) (15-200)",
"Pre-defined muxes", "Mux predefiniti",
"Pre-recording padding", "Riempimento pre-registrazione",
"Pre-school children's programs", "Trasmissione per bambini pre-scolari",
"Prefer picons over channel name", "Preferisci picon al nome del canale",
"Preferred CA PID", "PID CA preferito",
"Preferred service video type", "Tipo predefinito del servizio video",
"Presentation", "Presentazione",
"Press", "Stampa",
"Priority", "Priorità",
"Priority (-10..10)", "Priorità (-10..10)",
"Private", "Privato",
"Profile name", "Nome del profilo",
"Provider", "Fornitore",
"Provider ID", "ID del fornitore",
"Provider network name", "Nome della rete del fornitore",
"Psychology", "Psicologia",
"Puppets", "Pupazzi",
"Quad (4.0)", "Quadrifonia (4.0)",
"RAM only", "Solo su RAM",
"RTSP port", "Porta RTSP",
"RTSP port (554 or 9983, 0 = disable)", "Porta RTSP (554 o 9983, 0 = disabilita)",
"Re-fetch period (hours)", "Periodo di ri-estrazione (ore)",
"Re-fetch period (mins)", "Periodo di ri-estrazione (minuti)",
"Re-record%s%s", "Ri-registrazione%s%s",
"Re-try period (hours)", "Periodo di riprova (ore)",
"Record all", "Registra tutto",
"Recording file options", "Opzioni per file di registrazione",
"Recording system path", "Percorso di sistema per la registrazione",
"Reject", "Rigetta",
"Religion", "Religione",
"Religious", "Religioso",
"Remarkable people", "Celebrità",
"Remove HTTP arguments", "Rimuovi argomenti HTTP",
"Remove all unsafe characters from filename", "Rimuovi tutti i caratteri non sicuri dal nome del file",
"Replace whitespace in title with '-'", "Rimpiazza gli spazi nel titolo con '-'",
"Reply to CAM PIN inquiries", "Risposta alle richieste PIN CAM",
"Reports", "Inchieste",
"Rerun of", "Riesecuzione di",
"Reserved", "Riservato",
"Resolution (height)", "Risoluzione (altezza)",
"Respawn (pipe)", "Rigenerazione (pipe)",
"Restart on error", "Riavvia su errore",
"Reuse EPG from", "Riutilizza EPG da",
"Rewrite EIT", "Riscrivi EIT",
"Rewrite MPEG-TS SI tables", "Riscrivi tabelle MPEG-TS SI",
"Rewrite PAT", "Riscrivi PAT",
"Rewrite PMT", "Riscrivi PMT",
"Rewrite SDT", "Riscrivi SDT",
"Romance", "Romanzo",
"Rotor initialization time (seconds)", "Tempo di inizializzazione del rotore (secondi)",
"Rotor type", "Tipo di rotore",
"Run as group", "Esegui come gruppo",
"Run as user", "Esegui come utente",
"Running", "In esecuzione",
"SAT>IP DVB-T frequency (Hz)", "Frequenza SAT>IP DVB-T (Hz)",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "Numero di porta SAT>IP RTSP per il server\n(predefinito: -1 = disabilitato, 0 = webconfig, porta standard è 554)",
"SAT>IP source number", "Numero sorgente SAT>IP",
"SD: standard definition", "SD: definizione standard",
"SSL verify peer", "Verifica SSL del pari",
"Sat", "Sab",
"Satellite config", "Configurazione satellite",
"Satellite longitude", "Longitudine del satellite",
"Satellite positions", "Posizioni dei satelliti",
"Scan after creation", "Scansiona dopo la creazione",
"Scan queue length", "Lunghezza della coda di scansione",
"Scan result", "Risultato della scansione",
"Scan status", "Stato della scansione",
"Schedule a re-recording if more errors than (0=off)", "Pianifica una ri-registrazione se più errori di (0=spento)",
"Schedule status", "Stato della programmazione",
"Scheduled for recording", "Programmazione per registrazione",
"Scheduled start time", "Orario di inizio della programmazione",
"Scheduled stop time", "Orario di termine della trasmissione",
"School programs", "Trasmissione per la scuola",
"Science", "Scienza",
"Science fiction", "Fantascienza",
"Season", "Stagione",
"Season %d", "Stagione %d",
"Send CAPMT query", "Invia richiesta CAPMT",
"Send full PLAY cmd", "Invia comando PLAY completo",
"Serial number", "Numero seriale",
"Series link", "Collegamento della serie",
"Serious", "Serio",
"Serious music", "Musica tradizionale",
"Server UUID", "UUID del server",
"Server connectivity", "Connettività del server",
"Server details.", "Dettagli server.",
"Service", "Servizio",
"Service ID", "ID del servizio",
"Service configuration", "Configurazione del servizio",
"Service count %zi", "Conteggio dei servizi %zi",
"Service name", "Nome del servizio",
"Service raw", "Servizio grezzo",
"Service type", "Tipo di servizio",
"Services", "Servizi",
"Settings", "Impostazioni",
"Show", "Spettacolo",
"Show this page", "Mostra questa pagina",
"Show version information", "Mostra informazione di versione",
"Signal quality too poor", "Qualità del segnale troppo bassa",
"Signal scale (240 or 100)", "Scala di segnale (240 o 100)",
"Site latitude", "Latitudine del sito",
"Site longitude", "Longitudine del sito",
"Skip TS packets (0-200)", "Salta pacchetti TS (0-200)",
"Skip commercials", "Salta la pubblicità",
"Skip initial bytes", "Salta byte iniziali",
"Slave entry", "Voce secondaria",
"Slot state", "Stato della posizione",
"Soap", "Telenovela",
"Soccer", "Calcio",
"Social", "Sociale",
"Social advisory", "Avvertenza sociale",
"Sort index", "Indice di ordinamento",
"Source", "Sorgente",
"Source deleted", "Sorgente cancellata",
"Source quality is bad", "La qualità della sorgente è pessima",
"Source reconfigured", "Sorgente riconfigurata",
"Southern hemisphere (latitude direction)", "Emisfero meridionale (direzione della latitudine)",
"Special events (Olympic Games, World Cup, etc.)", "Eventi speciali (Giochi Olimpici, Coppa del Mondo, ecc.)",
"Specify User-Agent header for the http client", "Specifica l'intestazione User-Agent per il cliente http",
"Specify alternative htsp port", "Specifica porta htsp alternativa",
"Specify alternative http port", "Specifica porta http alternativa",
"Specify alternative http webroot", "Specifica radice web http alternativa",
"Specify bind address", "Specifica l'indirizzo di collegamento",
"Specify extra htsp port", "Specifica la porta supplementare htsp",
"Spiritual sciences", "Scienza spirituale",
"Sports", "Sport",
"Sports magazines", "Rivista sportiva",
"Standard", "Predefinito",
"Start", "Inizio",
"Start after", "Inizia dopo",
"Start before", "Inizia prima",
"Start time", "Orario di inizio",
"Status", "Stato",
"Status period (ms)", "Periodo di stato (ms)",
"Stop", "Termine",
"Stop time", "Orario di termine",
"Storage path", "Percorso di memorizzazione",
"Storage space", "Spazio di memorizzazione",
"Stream index", "Indice dello stream",
"Stream profile", "Profilo dello stream",
"Stream type", "Tipo dello stream",
"Streaming priority", "Priorità di streaming",
"Streaming profiles", "Profili di streaming",
"Subdirectory options", "Opzione delle sottocartelle",
"Subscribe to a service permanently", "Sottoscrivi un servizio in modo permanente",
"Subscription overridden", "Sottoscrizione esclusa",
"Subscription weight", "Peso della sottoscrizione",
"Substitute formatters", "Sostituisci i formattatori",
"Subtitle", "Sottotitolo",
"Subtitle codec", "Codec sottotitolo",
"Sun", "Dom",
"Surround (2 front, rear mono)", "Surround (2 frontali, mono posteriore)",
"Switch before rotor", "Commuta prima di ruotare",
"Switch type", "Tipo del commutatore",
"Switch: %s", "Commutatore: %s",
"Sync", "Sincronizza",
"Sync + Don't keep", "Sincronizza + Non Mantenere",
"System", "Sistema",
"Tag files with metadata", "Etichetta i file con i metadati",
"Tags", "Etichette",
"Team sports (excluding football)", "Sport a squadre (escluso il calcio)",
"Technology", "Tecnologia",
"Testing", "Controllo",
"Testing options", "Opzioni di prova",
"The PIN to use.", "Il PIN da utilizzare.",
"This tuner", "Questo sintonizzatore",
"Thu", "Gio",
"Time", "Orario",
"Time missed", "Orario mancato",
"Time record caption", "Didascalia della registrazione a tempo",
"Time recording%s%s", "Orario di registrazione%s%s",
"Time update", "Aggiornamento dell'orario",
"Timeout (sec) (0=infinite)", "Pausa (sec) (0=infinito)",
"Timeout (seconds)", "Pausa (secondi)",
"Title", "Titolo",
"Title (regexp)", "Titolo (regexp)",
"Toggle details", "Attiva dettagli",
"Tone burst", "Impulso del tono",
"Tourism / Travel", "Turismo / Viaggi",
"Trace subsystems", "Sottosistemi di tracciatura",
"Traditional arts", "Arti tradizionali",
"Traditional music", "Musica tradizionale",
"Transcoding", "Transcodifica",
"Transmission mode", "Modalità di trasmissione",
"Transponder ID", "ID del transponder",
"Transport stream ID", "ID dello stream di trasporto",
"Tue", "Mar",
"Tune before DiseqC", "Sintonizza prima di DiseqC",
"Tuner bind IP address", "Indirizzo IP di collegamento del sintonizzatore",
"Tuner configuration", "Configurazione del sintonizzatore",
"Tuning failed", "Sintonia fallita",
"Turn off LNB when idle", "Spegni LNB quando inattivo",
"Tvheadend server name", "Nome del server Tvheadend",
"Type", "Tipo",
"UDP RTP port number (2 ports)", "Numero di porta UDP RTP (2 porte)",
"URL with the SAT>IP server XML location", "URL con locazione XML del server SAT>IP",
"UTC+ 1", "UTC+1",
"UTC+ 2", "UTC+2",
"UTC+ 3", "UTC+3",
"UTC+ 4", "UTC+4",
"UTC+ 4:30", "UTC+4:30",
"UTC+ 5", "UTC+5",
"UTC+ 5:30", "UTC+5:30",
"UTC+ 5:45", "UTC+5:45",
"UTC+ 6", "UTC+6",
"UTC+ 6:30", "UTC+6:30",
"UTC+ 7", "UTC+7",
"UTC+ 8", "UTC+8",
"UTC+ 8:45", "UTC+8:45",
"UTC+ 9", "UTC+9",
"UTC+ 9:30", "UTC+9:30",
"UTC- 1", "UTC-1",
"UTC- 2", "UTC-2",
"UTC- 2:30", "UTC-2:30",
"UTC- 3", "UTC-3",
"UTC- 3:30", "UTC-3:30",
"UTC- 4", "UTC-4",
"UTC- 4:30", "UTC-4:30",
"UTC- 5", "UTC-5",
"UTC- 6", "UTC-6",
"UTC- 7", "UTC-7",
"UTC- 8", "UTC-8",
"UTC- 9", "UTC-9",
"UTC- 9:30", "UTC-9:30",
"Uncommitted", "Non vincolato",
"Uncommitted first", "Prima non vincolato",
"Unicable type", "Tipo unicable",
"Unimportant", "Non importante",
"Universal LNB only", "Solo LNB universale",
"Unknown", "Sconosciuto",
"Unknown file", "File sconosciuto",
"Unknown reason (%i)", "Ragione sconosciuta (%i)",
"Unlimited size", "Dimensione illimitata",
"Unlimited time", "Durata illimitata",
"Unset (default)", "Non impostato (predefinito)",
"Update card (EMM)", "Aggiorna carta (EMM)",
"Update channel icon", "Aggiorna icona del canale",
"Update channel name", "Aggiorna nome del canale",
"Update channel number", "Aggiorna numero del canale",
"Update disabled", "Aggiornamento disabilitato",
"Update time", "Aggiorna orario",
"Update tolerance (ms)", "Tolleranza di aggiornamento (ms)",
"Updated", "Aggiornato",
"Usage: %s [OPTIONS]\n", "Utilizzo: %s [OPZIONI]\n",
"Use EPG running state", "Utilizza lo stato di esecuzione EPG",
"Use Windows-compatible filenames", "Utilizza nomi di file compatibili con Windows",
"Use XSPF playlist instead of M3U", "Utilizza liste XSPF invece di quelle M3U",
"Use original", "Usa l'originale",
"Use service IDs as channel numbers", "Utilizza gli identificativi del servizio come numeri di canale",
"User access error", "Errore di accesso dell'utente",
"User icon", "Icona dell'utente",
"User interface level", "Livello dell'interfaccia utente",
"User language", "Lingua dell'utente",
"User limit reached", "Limite dell'utente raggiunto",
"Username", "Nome dell'utente",
"Variety show", "Varietà",
"Video bitrate (kb/s) (0=auto)", "Bitrate video (kb/s) (0=automatico)",
"Video codec", "Codec video",
"Video recorder", "Registratore video",
"Visually impaired commentary/audio description", "Commento per persone non vedenti",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV profilo H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV profilo H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV profilo VP8/Vorbis/WEBM",
"Waiting for EPG running flag", "Sto aspettando il segnale di esecuzione EPG",
"Waiting for program start", "Sto aspettando l'inizio della trasmissione",
"Waiting for stream", "Sto aspettando lo stream",
"War", "Guerra",
"Water sport", "Sport d'acqua",
"Weak stream", "Flusso debole",
"Weather report", "Situazione meteo",
"Web interface", "Interfaccia web",
"Web interface language", "Lingua dell'interfaccia web",
"Web user interface", "Interfaccia web utente",
"Wed", "Mer",
"Western hemisphere (latitude direction)", "Emisfero settentrionale (direzione della latitudine)",
"Wrapper (capmt_ca.so)", "Adattatore (capmt_ca.so)",
"Yes", "Si",
"addpids/delpids supported", "addpids/delpids supportati",
"hrs", "ore",
"invalid option specified [%s]", "opzione non valida specificata [%s]",
"min", "minuto",
"mins", "minuti",
"option %s requires a value", "l'opzione %s richiede un valore",
"tsfile input (mux file)", "ingresso tsfile (file mux)",
"tvhdhomerun client", "Cliente tvhdhomerun",
NULL, NULL
};

static const char *tvh_locale_20[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\n더 많은 정보를 원하시면 Tvheadend 웹사이트를 방문하세요:\nhttps://tvheadend.org\n",
"# Mapped channels", "# 매핑된 채널",
"# Muxes", "# 먹스",
"# Services", "# 서비스",
"# tune repeats", "# 튠 반복",
"(Default profile)", "(기본 프로필)",
"(Read-only)", "(읽기 전용)",
"1 day", "1일",
"1 month", "1개월",
"1 week", "1주일",
"1 year", "1년",
"2 months", "2개월",
"2 weeks", "2주일",
"2 years", "2년",
"2-Port switch (universal LNB)", "2포트 스위치 (범용 LNB)",
"3 days", "3일",
"3 months", "3개월",
"3 weeks", "3주일",
"3 years", "3년",
"4-Port switch (universal LNB)", "4포트 스위치 (범용 LNB)",
"5 days", "5일",
"6 months", "6개월",
"A channel tag (e.g. a group of channels) to which this rule applies.", "이 규칙이 적용될 채널 태그(예: 채널 그룹)",
"AAC audio", "AAC 오디오",
"AC-3 = descriptor 6", "AC-3 = 디스크립터 6",
"AC-3 detection", "AC-3 자동 판별",
"AC3 audio", "AC3 오디오",
"ACTIVE", "활성",
"AES Constant Code Word", "AES 상수 코드 단어",
"ANY", "모두",
"ATSC SI Tables", "ATSC SI 테이블",
"ATSC source ID", "ATSC 원본 ID",
"ATSC-C Network", "ATSC-C 네트워크",
"ATSC-C multiplex", "ATSC-C 멀티플렉스",
"ATSC-T Network", "ATSC-T 네트워크",
"ATSC-T multiplex", "ATSC-T 멀티플렉스",
"AUTO", "자동",
"Aborted by user", "사용자가 중단함",
"Accept remote subscription weight", "원격 구독 순위를 허용",
"Accept the remote subscription weight (from the SAT>IP client).", "원격 구독 순위를 허용 (위성>IP 클라이언트에서)",
"Accept zero value for TSID", "TSID에 값으로 0 허용",
"Access", "접근",
"Access (ACL)", "접근 (ACL)",
"Action", "액션",
"Active", "활성",
"Adapter in use by another subscription", "다른 구독에 의해 어댑터가 사용중",
"Add file and line numbers to debug", "디버그에 줄 번호와 파일 추가",
"Add the thread ID to debug", "디버그에 글타래 ID 추가",
"Add this value to PTS for the teletext subtitles. The time value is in milliseconds and may be negative.", "이 값을 PTS에 텔레텍스트 자막으로 추가하세요. 이 시간값은 밀리세컨드이며, 음수일 수 있습니다.",
"Additional arguments to pass to the grabber.", "그래버로 보낼 추가 매개변수",
"Additional service names found in EPG data.", "EPG 데이터에서 발견된 추가 서비스 이름",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "튜너가 녹화를 하기 위한 추가 준비 시간(초)입니다. 이것은 튜너 조절과 데이터 정리에 시간이 걸리는 경우 유용합니다.",
"Admin", "관리자",
"Admin password", "관리자 암호",
"Admin username", "관리자 이름",
"Administrator login", "관리자 로그인",
"Adult movie", "성인 영화",
"Advanced", "고급",
"Advanced (non-universal LNBs, rotors, etc.)", "고급 (비-범용 LNB, 로터 등)",
"Adventure", "어드벤쳐",
"Advertisement / Shopping", "광고 / 쇼핑",
"After creating the network scan it for services.", "서비스 네트워크 스캔을 한 다음입니다.",
"All (Streaming plus DVR)", "모두 (스트리밍+DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "문제가 발생할 수 있는 파일 문자는 모두 언더바로 교체될 것입니다. 도움말을 눌러서 자세히 알아보세요.",
"All lower-case", "모두 소문자",
"All proceeds are used to support server infrastructure and buy test equipment.", "모든 수익은 서버 시설을 유지하고, 테스트 장비를 구매하는데 사용됩니다.",
"All: Record if different description", "모두: 설명이 다른 경우 녹화",
"All: Record if different episode number", "모두: 에피소드 번호가 다른 경우 녹화",
"All: Record if different subtitle", "모두: 부제가 다른 경우 녹화",
"All: Record once per day", "모두: 하루 한 번 녹화",
"All: Record once per month", "모두: 한 달에 한 번 녹화",
"All: Record once per week", "모두: 일주일에 한 번 녹화",
"Allow high bitrate mode (CI+ CAMs only).", "높은 비트레이트 모드 허용 (CI+ CAM 전용)",
"Allow idle scan tuning on this device.", "장치에서 유휴 스캔 튜닝 허용",
"Allow the combined size of all timeshift buffers to potentially grow unbounded until your storage media runs out of space.", "미디어 저장 공간이 가득찰 때까지 잠재적으로 버퍼를 계속 늘려서 통합된 크기를 사용할 수 있도록 합니다.",
"Allow the initial scan tuning on this device (scan when Tvheadend starts or when a new multiplex is added automatically). At least one tuner or input should have this settings turned on. See also 'Skip Startup Scan' in the network settings for further details.", "(Tvheadend가 시작하거나, 새로운 멀티플렉스가 자동으로 추가되었을 때) 이 장치에서 초기 스캔 튜닝을 할 수 있도록 합니다. 최소한 하나 이상의 튜너 또는 입력은 이 설정이 켜져 있어야 합니다. 자세한 정보는 네트워크 설정의 '초기 스캔 건너뛰기'를 참조하세요.",
"Allow the timeshift buffer to grow unbounded until your storage media runs out of space. Warning, enabling this option may cause your system to slow down or crash completely!", "타임시프트 버퍼를 늘려서 미디어 저장 공간 전체를 활용하도록 합니다. 주의: 이 옵션을 켜면, 시스템이 느려지거나 고장날 수 있습니다!",
"Allow/Disallow access to the 'Configuration' tab.", "'설정' 탭 접근을 허용하거나 거부합니다.",
"Allow/Disallow web interface access (this  includes access to the EPG).", "(EPG 접근을 포함한) 웹 인터페이스 접근을 허용하거나 거부합니다.",
"Allowed DVR profiles. This limits the profiles the user has access to.", "DVR 프로필 허용. 아는 사용자가 접근할 프로필을 제한합니다.",
"Allowed network", "허용된 네트워크",
"Allowed networks", "허용된 네트워크",
"Alternate PID path", "대체 PID 경로",
"Alternate configuration path", "대체 설정 경로",
"Altitude (in meters).", "고도 (미터)",
"Altitude (meters)", "고도 (미터)",
"Always send the whole DiseqC sequence including LNB setup (voltage, tone). If this is not checked, only changed settings are sent, which may cause issues with some drivers. If the tuning is not reliable, try activating this option.", "LNB 설정 (볼트, 톤)을 포함한 다이섹 신호 전체를 항상 보냅니다. 체크되지 않으면, 바꾼 설정만이 보내지며, 이는 드라이버에 따라 문제를 일으킬 수 있습니다. 튜닝이 올바르지 않다면, 이 옵션을 켜보세요.",
"Always use the name defined by the network.", "자동으로 네트워크에서 이름을 가져옵니다. 항상 네트워크에서 지정된 이름을 사용합니다.",
"Always use the old ioctls to read the linuxdvb status (signal strength, SNR, error counters). Some drivers are not mature enough to provide the correct values using the new v5 linuxdvb API.", "항상 오래된 ioctls를 이용하여 linuxdvb 상태(신호 세기, SNR, 에러 횟수)를 읽습니다. 일부 드라이버는 새로운 v5 linuxdvb API 사용시 아직 올바른 값을 내놓지 못합니다.",
"An event which starts between this \"start after\" and \"start before\" will be matched (including boundary values).", "\"이후 시작\"과  \"이전 시작\" 사이에 있는 이벤트가 해당됩니다 (범위 끝 값도 포함).",
"Animals", "동물",
"Anonymize HTSP access", "HTSP 접근 익명화",
"Any", "모두",
"Arts", "모두",
"Arts / Culture (without music)", "예술 / 문화 (음악 제외)",
"Arts magazines", "예술 매거진",
"Assign predefined muxes to networks", "네트워크에 미리 설정된 먹스 할당",
"Associate this device with one or more networks.", "이 장치를 하나 이상의 네트워크와 연결합니다.",
"Athletics", "운동 경기",
"Audio Stream Filter", "오디오 스트림 필터",
"Audio bitrate (kb/s) (0=auto)", "오디오 비트레이트 (kb/s) (0=자동)",
"Audio bitrate to use for transcoding.", "트랜스코드할 오디오 비트레이트입니다.",
"Audio channel layout.", "오디오 채널 레이아웃입니다.",
"Audio codec", "오디오 코덱",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "트랜스코드할 오디오 코덱입니다. \"사용하지 않음\"은 오디오 출력을 배제합니다.",
"Audio stream", "오디오 스트림",
"Audio type", "오디오 종류",
"Audio-only stream", "오디오만 있는 스트림",
"Audioes muxer", "오디오 먹서",
"Authenticated user", "인증된 사용자",
"Auto", "자동",
"Auto check disabled", "자동 확인 비활성",
"Auto check enabled", "자동 확인 활성",
"Auto record", "자동 녹화",
"Auto record caption", "자동 녹화 캡션",
"Auto recording%s%s", "자동 녹화%s%s",
"Auto time record", "자동 시간 녹화",
"Auto-Map to channels", "채널에 자동 매핑",
"Automatic checking", "자동 확인",
"Automatic recording caption.", "자동 녹화 캡션입니다.",
"Automatic schedule entry not found", "자동 스케쥴 항목을 찾지 못했습니다.",
"Automatic time scheduler entry not found", "자동 타임 스케쥴 항목을 찾지 못했습니다.",
"Automatically link EPG data to the channel (using the channel name for matching). If you turn this option off, only the OTA EPG grabber will be used for this channel unless you've specifically set a different EPG Source.", "자동으로 EPG 원본을 매핑합니다. 자동으로 EPG 데이터를 채널에 연결합니다 (채널 이름을 사용). 이 옵션을 끄면, 다른 EPG 소스를 지정하지 않는한, OTA EPG 그래버만 사용될 것입니다.",
"Automatically map EPG source", "자동으로 EPG 원본 매핑",
"Automatically map all available services to channels.", "자동으로 모든 가용 서비스를 채널에 매핑합니다.",
"Automatically map channels defined within the bouquet.", "자동으로 채널 묶음을 채널에 매핑합니다.",
"Automatically name from network", "자동으로 네트워크에서 가져온 이름",
"Automatically record.", "자동으로 녹화합니다.",
"Automatically update channel icons using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel icons.", "활성화한 EPG 제공사에 의해 제공된 정보를 이용하여 채널 아이콘을 자동으로 갱신합니다. 참고: 기존에 설정된 채널 아이콘이 원치 않게 바뀔 수 있습니다.",
"Automatically update channel names using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel names.", "활성화한 EPG 제공사에 의해 제공된 정보를 이용하여 채널 이름을 자동으로 갱신합니다. 참고: 기존에 설정된 채널 이름이 원치 않게 바뀔 수 있습니다.",
"Automatically update channel numbers using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel numbers.", "활성화한 EPG 제공사에 의해 제공된 정보를 이용하여 채널 번호를 자동으로 갱신합니다. 참고: 기존에 설정된 채널 번호가 원치 않게 바뀔 수 있습니다.",
"Autorec maximum count (0=unlimited)", "자동 녹화 최대 개수 (0=무제한)",
"Autorec maximum schedules limit (0=unlimited)", "자동 녹화 최대 스케쥴 제한 (0=무제한)",
"Bad", "나쁨",
"Bad request", "잘못된 요청",
"Ballet", "불릿",
"Bandwidth", "대역폭",
"Base DVB SI Tables (PAT,CAT,PMT,SDT etc.)", "Base DVB SI Tables (PAT,CAT,PMT,SDT 등)",
"Based on software from", "다음 소프트웨에 기반: ",
"Basic", "기본",
"Bind to specific local IP address.", "로컬 IP 주소에 할당합니다.",
"Bitrate to use for the transcode. See Help for details.", "트랜스코드할 비트레이트입니다. 자세한 정보는 도움말을 참고하세요.",
"Blue", "파랑",
"Boot ID", "부트 ID",
"Bouquet", "채널 묶음",
"Bouquet (auto)", "채널 묶음 (자동)",
"Bouquet source.", "채널 묶음 원본입니다.",
"Bouquets", "채널 묶음",
"Brand", "브랜드",
"Branding information (if available).", "브랜드 정보입니다 (존재할 경우).",
"Broadcast", "브로드캐스트",
"Broadcast type", "브로드캐스트 종류",
"Broadcast.", "브로드캐스트입니다.",
"Broadcasting", "브로드캐스트 중",
"Buffering limit (ms)", "버퍼링 제한 (ms)",
"Build", "빌드",
"By default, linuxdvb's status read period is 1000ms (one second). The accepted range is 250ms to 8000ms. Note that for some hardware or drivers (like USB), the status operations take too much time and CPU. In this case, increase the default value. For fast hardware, this value might be decreased to make the decision of the re-tune algorithm based on the signal status faster.", "기본적으로, linuxdvb의 상태 읽기 간격은 1000ms (1초)입니다. 허용되는 범위는 250ms에서 8000ms입니다. 일부 하드웨어나 드라이버 (USB)에서는 상태 작업은 CPU 부하와 시간 소모를 유발합니다. 이 경우, 기본값을 늘리세요. 빠른 하드웨어에서 이 값은 신호 상태를 빠르게 얻는 튠 알고리즘의 결정에 의해 줄어들 수 있습니다.",
"CA (descrambling) Client", "CA (복호화) 클라이언트",
"CA Stream Filter", "CA 스트림 필터",
"CA client %i", "CA 클라이언트 %i",
"CA identification", "CA 식별",
"CA provider", "CA 제공자",
"CAPMT (Linux Network DVBAPI)", "CAPMT (리눅스 네트워크 DVBAPI)",
"CAPMT CA Client", "CAPMT CA 클라이언트",
"CAPMT interval (in ms).", "CAPMT 간격(ms 단위)입니다.",
"CAPMT interval (ms)", " CAPMT 간격 (ms)",
"CAPMT query interval (ms)", "CAPMT 쿼리 간격 (ms)",
"CAPMT query interval (ms).", "CAPMT 쿼리 간격(ms 단위)입니다.",
"CI Module", "CI 모듈",
"CRASH", "충돌",
"CRID authority", "CRID 인증",
"CSA (descrambling)", "CSA (복호화)",
"CWC CA Client", "CWC CA 클라이언트",
"Cache scheme", "캐시 스키마",
"Camd.socket filename / IP Address (TCP mode)", "Camd.socket 파일 이름 / IP 주소 (TCP 모드)",
"Cartoons", "만화",
"Change parameters", "매개변수 변경",
"Channel", "채널",
"Channel Tags", "채널 태그",
"Channel does not exist", "채널이 존재하지 않음",
"Channel icon", "채널 아이콘",
"Channel icon URL.", "채널 아이콘 URL입니다.",
"Channel icon as defined in EPG data.", "EPG 데이터에 선언된 채널 아이콘입니다.",
"Channel icon name scheme", "채널 아이콘 이름 스키마",
"Channel icon path", "채널 아이콘 경로",
"Channel mapping options", "채널 매핑 옵션",
"Channel name", "채널 이름",
"Channel not found", "채널을 찾을 수 없음",
"Channel number", "채널 번호",
"Channel number as defined in EPG data.", "EPG 데이터에 선언된 채널 번호입니다.",
"Channel number offset", "채널 추가 번호",
"Channel number range", "채널 번호 범위",
"Channel numbers (heuristic)", "채널 번호 (휴리스틱)",
"Channel numbers from", "채널 번호 출처",
"Channel tag", "채널 태그",
"Channel tag does not exist", "채널 태그가 없음",
"Channel tag reference", "채널 태그 레퍼런스",
"Channel tags", "채널 태그",
"Channel tags the user is allowed access to/excluded from.", "사용자가 접근할 수 있는/없는 채널 태그입니다.",
"Channel to use/used for the recording.", "녹화에 사용할/사용된 채널입니다.",
"Channel update options", "채널 갱신 옵션",
"Channels", "채널",
"Channels EPG data is used by.", "사용될 채널 EPG 데이터입니다.",
"Character set", "캐릭터셋",
"Characters not supported in Windows filenames (e.g. for an SMB/CIFS share) will be stripped out or converted.", "Windows 파일 이름(예: SMB/CIFS  공유)에서 지원되지 않는 문자는 제거되거나 변환될 것입니다.",
"Charset", "문자셋",
"Check availability", "가용성 확인",
"Check for the services' presence. If the service is no longer broadcast this field will change to Missing In PAT/SDT.", "서비스의 존재를 확인합니다. 서비스가 방송되지 않는 것으로 확인되면, PAT/SDT에서 미확인으로 이 필드가 변경됩니다.",
"Check services for availability. If enabled, services that are not currently broadcasting (or can't be decrypted) will be ignored. Leave disabled if you want Tvheadend to also map offline services.", "서비스 가용성을 확인합니다. 켜져 있으면, 지금 방송되고 있지 않은 (또는 복호화할 수 없는) 서비스가 무시됩니다. Tvheadend가 오프라인 서비스도 매핑하길 원하면 꺼진 채로 두세요.",
"Check the descrambling status after this timeout.", "이 제한 시간 후 디스크램블 상태를 확인합니다.",
"Children's / Youth Programs", "어린이 / 청소년 프로그램",
"Children's / Youth programs", "어린이 / 청소년 프로그램",
"Cinema", "시네마",
"Class", "클래스",
"Classical", "클래식",
"Classical music", "클래식 뮤직",
"Clean effects", "효과 지우기",
"Client name", "클라이언트 이름",
"Clone scheduled entry on error", "에러시 스케쥴된 항목 복제",
"Code Word Client (newcamd)", "코드 워드 클라이언트 (newcamd)",
"Comedy", "코미디",
"Command delay time (ms) (10-200)", "명령 지연 시간 (ms) (10-200)",
"Command time (ms) (10-100)", "명령 시간 (ms) (10-100)",
"Comment", "코멘트",
"Commercial break", "중간 광고",
"Commercials will be dropped from the recordings. Commercial detection works using EITp/f (EPG running state) and for the Swedish channel TV4 (using teletext info).", "녹화에서 광고가 제거될 것입니다. 광고 확인은 ElTp/f (EPG 구동 상태)와 스웨덴 채널  TV4(텔레텍스트 정보 사용)에서 작동합니다.",
"Committed", "커밋됨",
"Completed OK", "완료됨 OK",
"Compress EPG database", "EPG 데이터베이스 압축",
"Compress the EPG database to reduce disk I/O and space.", "EPG 데이터벡스를 압축하여 디스크 I/O 감소 및 공간을 절약합니다.",
"Concurrent limit per network position (src=) for satellite SAT>IP tuners. The first limit number is for src=1 (AA), second for src=2 (AB) etc.", "위성 SAT>IP 튜너(소스)에 대한 네트워크 동시 제한을 지정합니다. 첫 제한 번호는 src=1 (AA), 두 번째는 src=2 (AB) 순.",
"Conditional Access", "조건부 접근",
"Conditional Access Client", "조건부 클라이언트 접근",
"Conditional Access Identification.", "조건부 접근 식별.",
"Configuration", "설정",
"Configuration - Base", "설정 - 기본",
"Configuration - Image Cache", "설정 - 이미지 캐시",
"Configuration - SAT>IP Server", "설정 - SAT>IP 서버",
"Configuration ID", "설정 ID",
"Configuration name", "설정 이름",
"Configuration version", "설정 버전",
"Connection limit type", "연결 제한 종류",
"Connection limits", "연결 제한",
"Constellation", "배치",
"Container", "컨테이너",
"Container to use for the transcoded stream.", "트랜스코드된 스트림에 사용할 컨테이너입니다.",
"Content character set", "컨텐트 캐릭터셋",
"Content reference identifier authority.", "컨텐트 레퍼런스 식별 인증.",
"Content type", "컨텐트 종류",
"Content type.", "컨텐트 종류.",
"Contest", "대회",
"Continue if descrambling fails", "디스크램블 실패시에도 계속",
"Continue recording for x minutes after scheduled stop time", "예약된 정지 시간 x분 이후까지 녹화 지속",
"Continue recording for x minutes after scheduled stop time.", "예약된 정지 시간 x분 이후까지 녹화를 지속합니다.",
"Cookie expiration (days)", "쿠키 만료 (일)",
"Cooking", "요리",
"Copy codec type", "코덱 종류 복사",
"Copy layout", "레이아웃 복사",
"Count of objects", "오브젝트 수",
"Create SDTV/HDTV/Radio tags.", "SDTV/HDTV/라디오 태그를 추가합니다.",
"Create a bouquet from the playlist.", "재생 목록에서 채널 묶음을 생성합니다.",
"Create a directory per channel when storing recordings. If both this and the 'directory per day' checkbox is enabled, the date-directory will be the parent of the per-channel directory.", "녹화를 저장할 채널당 디렉토리를 만듭니다. 이것과 '일별 디렉토리'가 모두 켜져 있으면, 날짜 디렉토리가 채널 디렉토리 상위에 있게 됩니다.",
"Create a directory per title when storing recordings. If the day/channel directory checkboxes are also enabled, those directories will be parents of this directory.", "녹화를 저장할 때 제목별 디렉토리를 만듭니다. 일별/채널별 디렉토리가 켜져 있으면, 이들이 상위 디렉토리가 됩니다.",
"Create a new directory per day in the recording system path. Folders will only be created when something is recorded. The format of the directory will be ISO standard YYYY-MM-DD.", "시스템 경로에서 매일 새 디렉토리를 만듭니다. 폴더는 녹화가 이루어져야 생성됩니다. 포맷은 ISO 표준인 YYYY-MM-DD입니다.",
"Create a provider name tag.", "제공자 이름 태그를 만듭니다.",
"Create and associate a network tag to created channels.", "생성된 채널에 네트워크 태그를 만들고 연결합니다.",
"Create and associate a provider tag to created channels.", "생성된 채널에 제공자 태그를 만들고 연결합니다.",
"Create and link these tags to channels when mapping.", "매핑할 때 채널에 대한 태그를 만들고 링크합니다.",
"Create bouquet", "채널 묶음 생성",
"Create bouquet tag", "채널 묶음 태그 생성",
"Create directories using these permissions.", "이 권한으로 디렉토리 생성",
"Create files using these permissions.", "이 권한으로 파일 생성",
"Create network name tags", "네트워크 이름 태그 생성",
"Create network name tags (set by provider).", "네트워크 이름 태그 (제공자가 설정) 생성",
"Create network tags", "네트워크 태그 생성",
"Create provider name tags", "제공자 이름 태그 ",
"Create provider tags", "제공자 태그 생성",
"Create tags", "태그 생성",
"Create tags in recordings using media containers that support metadata (if possible).", "(가능하다면) 메타데이터를 지원하는 컨테이너를 사용해서 녹화에 태그를 생성합니다.",
"Create type-based tags", "종류 기반 채그 생성",
"Created", "생성됨",
"Creator", "생성자",
"Cron", "크론",
"Cron multi-line", "크론 다중 행간",
"Culture (without music)", "문화 (음악 제외)",
"Culture magazines", "문화 매거진",
"Current RTSP port.", "현재 RTSP 포트.",
"Current affairs", "시사",
"Current number of objects.", "현재 오브젝트 수.",
"Current object size.", "현재 오브젝트 크기.",
"Current tuner configuration.", "현재 튜너 설정.",
"Custom HTTP headers", "사용자 정의 HTTP 헤더",
"DAB Network", "DAB 네트워크",
"DAB multiplex", "DAB 멀티플렉스",
"DBus - use the session message bus instead of the system one", "DBus - 시스템 대신 세션 메시지 버스 사용",
"DES Constant Code Word", "DES 상수 코드 단어",
"DES Key.", "DES 키.",
"DES key", "DES 키",
"DSCP/TOS for streaming", "스트리밍 DSCP/TOS",
"DVB CAM Client", "DVB CAM 클라이언트",
"DVB CSA (descrambling) Tables", "DVB CSA (복호화) 테이블",
"DVB EPG Tables", "DVB EPG 테이블",
"DVB Inputs - Networks", "DVB 입력 - 네트워크",
"DVB SI Tables", "DVB SI 테이블",
"DVB Time Tables", "DVB 시간 테이블",
"DVB multiplex", "DVB 멀티플렉스",
"DVB scan files", "DVB 스캔 파일",
"DVB scan files path", "DVB 스캔 파일 경로",
"DVB-C Network", "DVB-C 네트워크",
"DVB-C multiplex", "DVB-C 멀티플렉스",
"DVB-S Network", "DVB-S 네트워크",
"DVB-S Satellite Configuration", "DVB-S 위성 설정",
"DVB-S multiplex", "DVB-S 멀티플렉스",
"DVB-T Network", "DVB-T 네트워크",
"DVB-T multiplex", "DVB-T 멀티플렉스",
"DVR - Auto-recording (Autorecs)", "DVR - 자동 녹화 (자동녹화)",
"DVR - Profiles", "DVR - 프로필",
"DVR - Time-based Recording (Timers)", "DVR - 시각 녹화 (타이머)",
"DVR behavior", "DVR 동작",
"DVR configuration", "DVR 설정",
"DVR configuration profiles", "DVR 설정 프로필",
"DVR configurations", "DVR 설정",
"DVR entry not found", "DVR 항목 찾을 수 없음",
"DVR file retention period", "DVR 파일 보유 기간",
"DVR log retention", "DVR 로그 보유",
"DVR log retention period", "DVR 로그 보유 기간",
"DVR override: high", "DVR 오버라이드: 높음",
"DVR override: important", "DVR 오버라이드: ",
"DVR override: low", "DVR 오버라이드: 낮음",
"DVR override: normal", "DVR 오버라이드: 보통",
"DVR override: unimportant", "DVR 오버라이드: 중요하지 않음",
"DVR profile to use/used for the recording.", "녹화에 사용된/사용할 DVR 프로필입니다.",
"DVR schedule does not exist", "DVR 스케쥴이 없음",
"DVR schedule does not have a file yet", "DVR 예약 파일이 아직 없음",
"DVR schedule not found", "DVR 스케쥴 찾을 수 없음",
"Dance", "댄스",
"Data errors", "데이터 에러",
"Data path (if applicable).", "데이터 경로 (적용할 수 있으면).",
"Date the EPG data was last updated (not set for OTA grabbers).", "마지막 업데이트된 EPG 데이터 날짜 (OTA 그래버에선 없음)",
"Days of Week", "요일",
"Days of the week to which the rule should apply.", "규칙이 적용될 요일을 설정합니다.",
"Days of week", "요일",
"Debate", "심의",
"Debug libav log", "libav 로그 디버그",
"Debug options", "디버그 옵션",
"Debug subsystems", "디버그 하부시스템",
"Debug to syslog", "syslog에 디버그",
"Debug trace (low-level)", "디버그 트레이스 (저수준)",
"Debugging", "디버깅",
"Default", "기본값",
"Default language(s)", "기본 언어",
"Default language.", "기본 언어.",
"Default login", "기본 로그인",
"Default priority", "기본 우선순위",
"Default user interface level.", "기본 인터페이스 레벨.",
"Define network group to limit network usage.", "네트워크 사용량을 제한할 네트워크 그룹 지정.",
"Delivery system", "전송 시스템",
"Demux path", "디먹스 경로",
"Descramble services (limit per mux)", "복호화 서비스 (먹스별 제한)",
"Descrambler", "복호화 방법",
"Descrambler EMM", "복호 EMM",
"Descrambler buffer (TS packets)", "복호화 버퍼 (TS 패킷)",
"Descrambling timeout (ms)", "디스크램블 시간 제한 (ms)",
"Description", "설명",
"Detective", "수사",
"Device ID", "장치 ID",
"Device model", "장치 모델",
"Device path", "장치 경로",
"Device path in sysfs", "sysfs에서 장치 경로",
"Differentiated Services Code Point / Type of Service: Set the service class Tvheadend sends with each packet. Depending on the option selected this tells your router the prority in which to give packets sent from Tvheadend, this option does not usually need changing. See https://en.wikipedia.org/wiki/Differentiated_services for more information. ", "서비스 코드 포인트 / 서비스 종류 차이: 서비스 클래스를 Tvheadend가 보내는 각 패킷으로 지정합니다. 이 옵션 선택 여부에 따라 Tvheadend가 보내는 패킷에 우선 순위를 정할 수 있으며, 일반적으로 지정할 필요는 없습니다. https://en.wikipedia.org/wiki/Differentiated_services 항목을 참고하여 더 자세히 알아보세요.",
"Digest access authentication is intended as a security trade-off. It is intended to replace unencrypted HTTP basic access authentication. This option should be enabled for standard usage.", "다이제스트 인증은 안전한 데이터 교환을 위해 만들어졌습니다. HTTP 기본 접근 인증에서 암호화되지 않은 부분을 교체하기 위함입니다. 이 옵션은 보통 활성화되어야 합니다.",
"Digital Video Recorder", "디지털 비디오 녹화",
"Directory", "디렉토리",
"Directory override. Override the subdirectory rules specified by the DVR configuration and put all recordings done by this entry into the specified subdirectory", "디렉토리를 재지정합니다. DVR 설정에 지정된 서브 디렉토리 규칙을 새로 쓰고 모든 녹화를 지정한 서브 디렉토리로 넣습니다.",
"Directory permissions (octal, e.g. 0775)", "디렉토리 권한 (8진수, 예: 0775)",
"Directory used by the entry.", "항목에 사용된 디렉토리.",
"Disable", "해제",
"Disable DVB bouquets", "DVB 채널 묶음 해제",
"Disable SAT>IP client", "SAT>IP 클라이언트 해제",
"Disable X_SATIPM3U tag", "X_SATIPM3U 태그 끄기",
"Disable all access control checks", "모든 접근 제어 확인 해제",
"Disable debug on stderr", "stderr의 디버그 해제",
"Disable device/firmware-specific workarounds", "장치/펌웨어 종속 차선책 해제",
"Disable syslog (all messages)", "syslog (모든 메시지) 해제",
"Disabled", "해제됨",
"Discover more muxes using the Network Information Table (if available).", "(가능하면) 네트워크 정보 표에서 더 많은 먹스 탐색.",
"Discussion", "심의",
"DiseqC", "다이섹",
"DiseqC repeats", "다이섹 반복",
"Do not send X_SATIPM3U information in the XML description to clients.", "클라이언트에 XML 설명으로 X_SATIPM3U 정보를 전달하지 않습니다.",
"Do not send any stream specific information to the HTSP client like signal strength, input source etc.", "HTSP 클라이언트에 신호 세기, 입력 원본 등 상세 정보를 보내지 않음.",
"Do not use", "사용하지 않기",
"Documentary", "다큐멘터리",
"Don't abort streaming when an encrypted stream can't be decrypted by a CA client that normally should be able to decrypt the stream.", "일반적으로 복호화할 수 있는 스트림을, CA 클라이언트가 암호화된 스트림을 복호화할 수 없을 때, 스트리밍을 중단하지 않습니다.",
"Don't backup configuration tree at upgrade", "업그레이드시 설정 트리를 백업하지 않음",
"Don't include the title in the filename.", "파일 이름에 제목을 포함하지 않음",
"Don't include title in filename", "파일 이름을 제목에 포함하지 않음",
"Don't keep", "기억하지 않음",
"Don't re-record", "재기록하지 않음",
"Don't re-record if recording fails.", "녹화에 실패하면 다시 기록하지 않습니다.",
"Don't re-schedule if recording fails.", "녹화에 실패하면 다시 예약하지 않습니다.",
"Don't reschedule", "다시 예약하지 않음",
"Don't use the provider's channel numbers.", "제공자의 채널 번호를 사용하지 않음",
"Drama", "드라마",
"Duplicate handling", "중복 처리",
"Duplicate recording handling.", "중복된 녹화를 처리합니다.",
"EIT - skip TSID check", "EIT - TSID 확인 건너뛰기",
"EIT time offset", "EIT 시각 이동",
"EPG Database", "EPG 데이터베이스",
"EPG Grabber", "EPG 그래버",
"EPG Grabber Channel", "EPG 그래버 채널",
"EPG Grabber Configuration", "EPG 그래버 설정",
"EPG Language (priority order)", "EPG 언어 (우선 순위)",
"EPG data ID.", " EPG 데이터 ID.",
"EPG name", "EPG 이름",
"EPG overlap cut", "EPG 겹침 잘라내기",
"EPG scan", "EPG 스캔",
"EPG scan time-out in seconds (30-7200)", "EPG 스캔 제한 시간 (초, 30-7200)",
"EPG source", "EPG 원본",
"EPG update window", "EPG 갱신 창",
"Economics", "경제",
"Education", "교육",
"Education / Science / Factual topics", "교육 / 과학 / 주제별 주제",
"Educational", "교양",
"Electronic Program Guide", "전자 프로그램 가이드",
"Elementary Stream Filter", "기초 스트림 필터",
"Elementary stream filter", "기초 스트림 필터",
"Enable", "활성화",
"Enable (auto)", "활성 (자동)",
"Enable DBus", "DBus 활성",
"Enable NTP driver", "NTP 드라이버 활성",
"Enable coredumps for daemon", "데몬에 coredumps 활성",
"Enable debug on stderr", "stderr 디버그 활성",
"Enable debug subsystems", "하부시스템 디버그 ",
"Enable debug to file", "파일에 디버그 활성",
"Enable debug to syslog", "syslog에 디버그 활성",
"Enable exclusion of user-config defined channel tags. This will prevent the user from accessing channels associated with the tags selected (below).", "채널 태그가 지정된 사용자 설정의 예외 사항을 켭니다. 아래 선택된 태그가 지정된 채널에 사용자가 접근하는 것을 차단합니다 (아래).",
"Enable if the SAT>IP box requests plts=on parameter in the SETUP RTSP command for DVB-S2 muxes.", "SAT>IP 박스가 DVB-S2 먹스에 대한 SETUP RTSP 명령에서 pids=0 매개변수가 필요하면 표시합니다.",
"Enable if the SAT>IP box requires pids=21 parameter in the SETUP RTSP command", "SAT>IP 박스가 SETUP RTSP 명령에서 pids=21 매개변수를 필요로 하면 표시합니다.",
"Enable if the SAT>IP box supports the addpids/delpids commands.", "SAT>IP 박스가 addpids/delpids 명령을 지원하는 것을 확인합니다.",
"Enable or disable RTP/AVP/TCP transfer mode (embedded data in the RTSP session) support.", "RTP/AVP/TCP 전송 모드 (RTSP 세션에서 포함된 데이터) 지원을 켜거나 끕니다.",
"Enable or disable fast input switching.", "빠른 입력 전환을 켜거나 끕니다.",
"Enable or disable full mux mode.", "전체 먹스 모드를 켜거나 끕니다.",
"Enable or disable ignoring of Event Information Table (EIT) data on this mux.", "먹스에서 이벤트 정보 테이블(EIT) 데이터를 켜거나 끕니다.",
"Enable or disable this configuration.", "설정을 활성 또는 해제합니다.",
"Enable over-the-air program guide (EPG) scanning on this input device.", "이 입력 장치에 대해 OTA 프로그램 가이드(EPG)를 켭니다.",
"Enable syslog", "syslog 활성",
"Enable system time updates. This will only work if the user running Tvheadend has rights to update the system clock (normally only root).", "시스템 시계 업데이트를 켭니다. 이는 Tvheadend가 타당한 권한을 갖고 있는 계정으로 구동되어야 작동합니다 (주로 root 계정).",
"Enable the CAs (conditional accesses) tab in web user interface for the advanced level. By default, this tab is visible only in the expert level.", "고급 레벨에서 CAs (조건부 접근) 탭을 웹 유저 인터페이스에서 활성화합니다. 기본값으로, 이 탭은 전문가 레벨에서만 보입니다.",
"Enable this filter.", "이 필터 활성.",
"Enable trace subsystems", "서브시스템 트레이스 켜기",
"Enable web UI debug (non-minified JS)", "웹 UI 디버그 켜기 (minified되지 않은 JS)",
"Enable, disable or ignore the mux. When the mux is marked as ignore, all discovered services are removed.", "먹스를 켜거나 끄거나 무시합니다. 먹스를 무시하도록 설정하면, 발견된 서비스는 모두 제거됩니다.",
"Enable/Disable CA client.", "CA 클라이언트를 켜거나 끕니다.",
"Enable/Disable service.", "서비스를 켜거나 끕니다.",
"Enable/Disable the entry.", "항목을 켜거나 끕니다.",
"Enable/Disable timeshift.", "타임시프트를 켜거나 끕니다.",
"Enable/disable EPG data for the entry.", "항목에 EPG 데이터를 켜거나 끕니다.",
"Enable/disable LNA.", "LNA를 켜거나 끕니다.",
"Enable/disable auto-rec rule.", "자동 녹화 규칙을 켜거나 끕니다.",
"Enable/disable debugging output to syslog.", "syslog에 디버깅 출력을 켜거나 끕니다.",
"Enable/disable inclusion of low-level debug traces.", "저수준 디버그 트레이스 포함을 켜거나 끕니다.",
"Enable/disable interface quick tips.", "인터페이스 빠른 팁을 켜거나 끕니다.",
"Enable/disable libav log output.", "libav 로그 출력을 켜거나 끕니다.",
"Enable/disable logging to syslog.", "syslog에 로깅을 켜거나 끕니다.",
"Enable/disable offering of Entitlement Management Message updates.", "자격 관리 메시지(EMM) 업데이트 제공을 켜거나 끕니다.",
"Enable/disable pilot tone.", "파일럿 톤을 켜거나 끕니다.",
"Enable/disable power save mode (if supported by the device).", "전원 절약 모드를 켜거나 끕니다 (장치에서 지원하는 경우).",
"Enable/disable profile.", "프로필을 켜거나 끕니다.",
"Enable/disable the bouquet.", "채널 묶음을 켜거나 끕니다.",
"Enable/disable the channel.", "채널을 켜거나 끕니다.",
"Enable/disable the device.", "장치를 켜거나 끕니다.",
"Enable/disable the entry.", "항목을 켜거나 끕니다.",
"Enable/disable the grabber.", "그래버를 켜거나 끕니다.",
"Enable/disable the tag.", "태그를 켜거나 끕니다.",
"Enable/disable tuner/adapter.", "튜너/어댑터를 켜거나 끕니다.",
"Enabled", "켜짐",
"Encrypted", "암호화됨",
"Encrypted service", "암호화된 서비스",
"Enter a non-admin user password.", "관리자가 아닌 암호를 입력하세요.",
"Enter a non-admin user username.", "관리자가 아닌 사용자 이름을 입력하세요.",
"Enter allowed network prefix(es). You can enter a comma-seperated list of prefixes here.", "허용된 네트워크 프리픽스를 입력하세요. 프리픽스를 반점으로 구분하여 입력하세요.",
"Enter an administrator password.", "관리자 암호를 입력하세요.",
"Enter an administrator username. Note: do not use the same username as the superuser backdoor account.", "관리자의 사용자 이름을 입력하세요. 참고: 시스템 관리자 권한 계정과 같은 이름을 입력하지 마세요.",
"Enter comma-separated list of subsystems you want debugging output for (e.g +linuxdvb,+subscriptions,+mpegts).", "디버깅 출력을 할 서브시스템을, 반점으로 구분된 목록으로 입력하세요 (예: +linuxdvb,+subscriptions,+mpegts).",
"Enter comma-separated list of subsystems you want to get traces for (e.g +linuxdvb,+subscriptions,+mpegts).", "트레이스를 얻어낼 서브시스템을, 반점으로 구분된 목록으로 입력하세요 (e.g +linuxdvb,+subscriptions,+mpegts).",
"Enter external IP if behind Network address translation (NAT). Asterisk (*) means accept all IP addresses.", "네트워크 주소 변환(NAT) 하부인 경우 외부 IP를 입력합니다. 별표(*)는 모든 IP 주소 허용을 의미합니다.",
"Enter the filename (including path) where Tvheadend should write the log.", "Tvheadend가 로그를 작성할 파일 이름(경로 포함)입니다.",
"Entertainment programs for 10 to 16", "만 10-16세 예능",
"Entertainment programs for 6 to 14", "만 6-14세 예능",
"Environment", "환경",
"Environment (pipe)", "환경 (파이프)",
"Episode", "에피소드",
"Episode %d", "에피소드 %d",
"Episode number/ID.", "에피소드 번호/ID.",
"Equestrian", "승마",
"Error code", "에러 코드",
"Error code of entry.", "항목 에러 코드.",
"Errors", "에러",
"Even key", "같은 키",
"Even key.", "같은 키.",
"Event does not exist", "이벤트가 없음",
"Every day", "매일",
"Exclude channel tags", "채널 태그 제외",
"Expeditions", "탐험",
"Experimental film", "탐험 영화",
"Expert", "전문가",
"Exported tuners", "전문가 튜너",
"External", "외부",
"External EPG grabber", "외부 EPG 그래버",
"External IP (NAT)", "외부 IP (NAT)",
"External PyEPG Grabber", "외부 PyEPG 그래퍼",
"External URL", "외부 URL",
"External URL of the bouquet.", "채널 묶음의 외부 URL.",
"External XMLTV EPG Grabber", "외부 XMLTV EPG 그래버",
"Extra arguments", "추가 매개변수",
"Extra start time", "추가 시작 시간",
"Extra stop time", "추가 끝 시간",
"Extra warming up time (seconds)", "추가 워밍 업 시간 (초)",
"FAIL", "실패",
"FEC high", "FEC 높음",
"FEC low", "FEC 낮음",
"Factual topics", "논픽션",
"Failed", "실패",
"Failed to open image", "이미지 열기 실패",
"Failed view", "실패한 보기",
"Fantasy", "판타지",
"Fashion", "패션",
"Fast input switch", "빠른 입력 전환",
"Fastscan DVB", "DVB 빠른 스캔",
"File missing", "파일 없음",
"File not created", "파일 생성되지 않음",
"File permissions (octal, e.g. 0664)", "파일 권한 (8진수, 예: 0664)",
"File removed", "파일 제거됨",
"File size", "파일 크기",
"Filename", "파일 이름",
"Filename (including path)", "파일 이름 (경로 포함)",
"Filename character set", "파일 이름 캐릭터셋",
"Filename options", "파일 이름 옵션",
"Filename used by the entry.", "항목에 사용된 파일 이름.",
"Filesystem monitor", "파일 시스템 모니터",
"Film", "영화",
"Fine arts", "예술",
"Finished", "완료됨",
"First word", "첫 단어",
"Fit to RAM (cut rewind)", "램에 맞추기 (되돌리기 없음)",
"Fitness and health", "피트니스 및 헬스",
"Folk", "포크",
"Folkloric", "민속풍",
"Football", "축구",
"Force (auto)", "강제 (자동)",
"Force CA ID (e.g. 0x2600)", "강제 CA ID (예: 0x2600)",
"Force all network connections to this tuner to be made over the specified IP address, similar to the setting for the SAT>IP device itself. Setting this overrides the device-specific setting.", "SAT>IP 장치 자체에 대한 설정과 같이, 이 튜너에 대해 지정된 IP 주소를 사용하도록 합니다. 이 설정은 장치별 설정보다 우선합니다.",
"Force an initial EPG grab at start-up.", "시작시 초기 EPG 그랩을 강제합니다.",
"Force initial EPG grab at start-up", "시작시 EPG 그랩 강제",
"Force old status", "오래된 상태 강제",
"Force pilot for DVB-S2", "DVB-S2에 파일럿 강제",
"Force priority", "우선순위 강제",
"Force profile to use this priority.", "이 우선 순위를 사용할 스트림 프로필을 강제합니다.",
"Force service type to 1", "서비스 타입 1 강제",
"Force teardown delay", "수신 해제 지연 시간 강제",
"Force the delay between RTSP TEARDOWN and RTSP SETUP command (value from 'Next tune delay in ms' is used). Some devices are not able to handle quick continuous tuning.", "RTSP TEARDOWN과 RTSP SETUP 명령 사이 지연 시간 강제합니다 ('다음 튠 지연 시간' 항목의 값). 일부 장치는 연속 튜닝시 빠른 처리가 불가합니다.",
"Force the grace period for which SAT>IP client waits for the data from server. After this grace period, the tuner is handled as dead. The default value is 5 seconds (for DVB-S/S2: 10 seconds).", "서버에서 데이터를 얻기 위해 SAT>IP 클라이언트가 기다릴 기간을 지정합니다. 이 기간 후에 튜너가 제어되거나, 응답 없게 됩니다. 기본값은 5초입니다. (DVB-S/S2의 경우: 10초)",
"Force the local UDP Port number here. The number should be even (RTP port). The next odd number (+1) will be used as the RTCP port.", "로컬 UDP 포트 번호를 강제합니다. 이 번호는 짝수여야 합니다 (RTP 포트). 홀수 번호(+1)는 RTCP 포트로 사용됩니다.",
"Force usage of entered CA ID on this service.", "이 서비스에 입력된 CA ID를 강제합니다.",
"Foreign countries", "외국",
"Forever", "영구",
"Fork and run as daemon", "Fork 후 데몬으로 실행",
"Format string", "문자 포맷",
"Found muxes", "찾은 먹스",
"Found services", "찾은 서비스",
"Free subscription weight", "자유 구독 순위",
"Free-form field, enter whatever you like here.", "자유 폼 필드, 아무거나 입력하세요.",
"Free-form text field, enter whatever you like here.", "자유 폼 문자 필드, 아무거나 입력하세요.",
"Free-form text field, enter whatever you like.", "자유 폼 문자 필드, 아무거나 입력하세요.",
"Free-form text field. You can enter whatever you like here.", "자유 폼 문자 필드. 아무거나 원하시는 걸 입력하세요.",
"Free-format text field. Enter whatever you like here.", "자유 포맷 텍스트 필드. 아무거나 입력하세요.",
"Frequency", "주파수",
"Frequency (Hz)", "주파수 (Hz)",
"Frequency (MHz)", "주파수 (MHz)",
"Frequency (in MHz).", "주파수입니다 (MHz 단위).",
"Frequency (kHz)", "주파수 (kHz)",
"Fri", "금",
"Friendly name", "읽기 이름",
"Frontend number", "프론트앤드 번호",
"Frontend path", "프론트앤드 경로",
"Full DiseqC", "전체 다이섹",
"Full mux RX mode supported", "전체 먹스 RX 모드 지원",
"Full path to an icon used to depict the tag. This can be a TV network logotype, etc.", "태그를 묘사하는 아이콘 전체 경로입니다. TV 네트워크 로고 타입 등이 될 수 있습니다.",
"Full pathname specification", "전체 경로 이름 지정",
"Full-text", "풀텍스트",
"Further education", "중등 교육",
"GOTOX position", "GOTOX 위치",
"Game show", "게임쇼",
"Gardening", "정원",
"General", "일반",
"General configuration", "일반 설정",
"Generic", "공통",
"Generic options", "공통 옵션",
"Global Headers", "글로벌 헤더",
"Global timer", "글로벌 타이머",
"Grabber priority. This option lets you pick which EPG grabber's data get used first. Priority is given to the grabber with the highest value set here. See Help for more info.", "그래버 우선 순위입니다. 이 옵션은 어떤 EPG 데이터를 우선 선택할지 결정합니다. 높은 값의 우선 순위가 지정된 그래버가 설정됩니다. 도움말에서 더 자세히 알아보세요.",
"Grace period", "유예 기간",
"Gray", "회색",
"Guard interval", "가드 주기",
"HD: high definition", "HD: 고선명",
"HDHomeRun ATSC-C frontend", "HDHomeRun ATSC-C 프론트엔드",
"HDHomeRun ATSC-T frontend", "HDHomeRun ATSC-T 프론트엔드",
"HDHomeRun DVB frontend", "HDHomeRun DVB 프론트엔드",
"HDHomeRun DVB-C frontend", "HDHomeRun DVB-C 프론트엔드",
"HDHomeRun DVB-T frontend", "HDHomeRun DVB-T 프론트엔드",
"HTSP Answer", "HTSP 응답",
"HTSP Default Stream Settings", "HTSP 기본 스트림 설정",
"HTSP Request", "HTSP 요청",
"HTSP Server", "HTSP 서버",
"HTSP Stream Profile", "HTSP 스트림 프로필",
"HTSP Subscription", "HTSP 구독",
"HTTP CORS (cross-origin resource sharing) origin. This option is usually set when Tvheadend is behind a proxy. Enter a domain (or IP) to allow cross-domain requests.", "HTTP 접근 제어 (cross-origin resource sharing) 출처입니다. 이 옵션은 Tvheadend가 프록시를 거쳐서 연결될 경우 쓰입니다. 크로스 도메인 요청할 도메인 (또는 IP)를 입력하세요.",
"HTTP CORS origin", "HTTP CORS 출처",
"HTTP Client", "HTTP 클라이언트",
"HTTP Server", "HTTP 서버",
"Handicraft", "수공예",
"Hearing impaired", "청각 장애",
"Hierarchy", "계층 구조",
"High", "높음",
"High bitrate mode (CI+ CAMs only)", "높은 비트레이트 모드 (CI+ CAM 전용)",
"High frequency offset", "고 주파수 오프셋",
"Highest channel number the user can access.", "유저가 접근할 수 있는 가장 높은 채널 번호입니다.",
"Highest count of objects.", "가장 높은 수의 오브젝트입니다.",
"Historical movie", "역사 영화",
"Horror", "공포",
"Hostname (or IP) of the server.", "서버 호스트 이름 (또는 IP)",
"Hostname/IP", "호스트 이름/IP",
"How frequently it will re-try fetching an image that has failed to be fetched.", "이미지를 가져오는데 실패하면 얼마나 자주 다시 가져오려고 시도할지 정합니다.",
"How frequently the upstream provider is checked for changes.", "업스트림 제공자 변경 사항을 얼마나 자주 확인할지 정합니다.",
"ID", "아이디",
"ID of the network.", "네트워크의 ID입니다.",
"IDLE", "유휴",
"IGNORE", "무시",
"IP address", "IP 주소",
"IP blocking", "IP 차단",
"IPTV Automatic Network", "IPTV 자동 네트워크",
"IPTV Multiplex", "IPTV 멀티플렉스",
"IPTV Network", "IPTV 네트워크",
"IPTV Subcription", "IPTV 구독",
"IPTV input", "IPTV 입력",
"ISDB-C Network", "ISDB-C 네트워크",
"ISDB-C multiplex", "ISDB-C 멀티플렉스",
"ISDB-S Network", "ISDB-S 네트워크",
"ISDB-S multiplex", "ISDB-S 멀티플렉스",
"ISDB-T Network", "ISDB-T 네트워크",
"ISDB-T multiplex", "ISDB-T 멀티플렉스",
"ISI (Stream ID)", "ISI (스트림 ID)",
"Icon", "아이콘",
"Icon (full URL)", "아이콘 (전체 URL)",
"Icon URL", "아이콘 URL",
"Icon base URL", "아이콘 기본 URL",
"Icon base URL.", "아이콘 기본 URL",
"Icon has title", "제목을 가진 아이콘",
"Icons from", "아이콘 출처",
"Idle", "유휴",
"Idle scan", "유휴 스캔",
"Idle scan muxes", "먹스 유휴 스캔",
"If an error occurs clone the scheduled entry and try to record again (if possible).", "에러가 발생한다면, 예약 항목을 복제하고 다시 녹화를 시도해보세요 (가능한 경우).",
"If both a picon and a channel-specific (e.g. channelname.jpg) icon are defined, prefer the picon.", "만약 picon과 채널에 지정된 (예: 채널이름.jpg) 아이콘이 동시에 지정되어 있으면, 아이콘을 선호합니다.",
"If more than x errors occur during a recording schedule a re-record (if possible).", "x개 이상 에러가 발생하면, 다시 녹화를 예약합니다 (가능한 경우).",
"If no specific priority was requested. This gives certain users a higher priority by assigning a streaming profile with a higher priority.", "특정 우선 순위가 지정되지 않으면, 높은 우선 순위 스트리밍 프로필을 할당함으로써 특정 사용자가 높은 우선 순위를 부여받게 됩니다.",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "사용자 계정이 없으면, 사용자 이름과 암호 없이\n새로 만듭니다. Tvheadend 웹 인터페이스에서\n새로 접근 컨트롤을 만들지 않는 경우,\n전세계 어디에서나 Tvheadend 서버에\n관리자 접근을 허용하게 됩니다.",
"If set, presentation of the tag icon will not superimpose the tag name on top of the icon.", "설정되어 있으면, 태그 아이콘 표시는 아이콘 위에 태그 이름을 첨가하지 않을 것입니다.",
"If set, the first bytes from the MPEG-TS stream are discarded. It may be required for some drivers or hardware which do not flush the MPEG-TS buffers completely after a frequency/parameter change.", "설정되어 있으면, MPEG-TS 스트림의 첫 바이트는 버려집니다. 이는 주파수나 매개변수 변경 후 MPEG-TS 버퍼를 비우지 않는 일부 드라이버나 하드웨어에 필요합니다.",
"If something fails, try to switch to a different service on another network. Do not try to iterate through all inputs/tuners which are capable to receive the service.", "무언가 오류가 나면, 다른 네트워크의 다른 서비스로 전환을 시도합니다. 서비스를 수신할 수 있는 모든 입력/튜너에 대해 반복하지 마세요.",
"If the DiseqC switch is located before the rotor (i.e. tuner - switch - rotor), enable this.", "다이섹 스위치가 로터 앞에 위치하여 있다면 (예: 튜너 - 스위치 - 로터), 이 옵션을 켜세요.",
"If the subscription weight for the input is below the specified threshold, the tuner is handled as free (according the priority settings). Otherwise, the next tuner (without any subscriptions) is used. Set this value to 10, if you are willing to override scan and epggrab subscriptions.", "지정한 임계점보다 입력에 요청 순위가 낮다면, 튜너는 프리로 처리됩니다 (우선순위 설정에 따라). 아니면, (구독이 없는) 다음 튜너가 사용됩니다. 스캔과 epggrab 요청을 재정의하려면, 이 값을 10으로 지정합니다.",
"If you'd like to support the project, please consider a donation.", "프로젝트를 지원하고 싶으시다면, 기부는 어떠실까요.",
"Ignore", "무시",
"Ignore EPG (EIT)", "EPG (EIT) 무시",
"Ignore descriptor 5", "디스크립터 5 무시",
"Ignore encryption flag, include encrypted services anyway.", "암호 플래그를 무시하고, 암호화된 서비스를 강제로 포함합니다.",
"Ignore invalid SSL certificate", "잘못된 SSL 인증서 무시",
"Ignore invalid/unverifiable (expired, self-certified, etc.) certificates", "잘못되거나 검증할 수 없는 (만료된, 자기 서명된 등) 인증서 무시",
"Ignore provider's channel numbers", "제공자의 채널 번호 무시",
"Image Cache", "이미지 캐시",
"Immediately abort", "즉시 중단",
"Important", "중요",
"Include channel name in filename", "채널 이름을 파일 이름에 포함",
"Include date in filename", "날짜를 파일 이름에 포함",
"Include episode in filename", "에피소드를 파일 이름에 ",
"Include subtitle in filename", "부제를 파일 이름에 포함",
"Include teletext", "텔레텍스트 포함",
"Include teletext in the timeshift buffer. Enabling this may cause issues with some services where the teletext DTS is invalid.", "타임시프트 버퍼에 텔레텍스트를 포함합니다. 텔레텍스트 DTS가 잘못된 일부 서비스에서 문제를 일으킬 수 있습니다. ",
"Include the date for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "녹화 이벤트 제목에서 날짜를 포함합니다. 이는 파일 제목과 파일 이름에 적용됩니다.",
"Include the episode subtitle in the title (if available).", "(가능한 경우) 부제를 제목에 포함합니다.",
"Include the name of the channel in the event title. This applies to both the title stored in the file and to the filename itself.", "이벤트 타이틀에서 채널 이름을 포함합니다. 이는 파일 제목과 파일 이름에 적용됩니다.",
"Include the season and episode in the title (if available).", "(가능한 경우) 시즌과 에피소드를 제목에 포함합니다.",
"Include the time for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "녹화 이벤트 제목에서 시각을 포함합니다. 이는 파일 제목과 파일 이름에 적용됩니다.",
"Include time in filename", "시각을 파일 이름에 포함",
"Index", "인덱스",
"Information area", "정보 영역",
"Informational", "정보 제공",
"Initial scan", "최초 스캔",
"Input base", "입력 베이스",
"Input buffer (bytes)", "입력 버퍼 (바이트)",
"Input instance", "입력 인스턴스",
"Input path", "입력 경로",
"Interface", "인터페이스",
"Internal", "내부",
"Internal EPG grabber", "내부 EPG 그래버",
"Internal PyEPG Grabber", "내부 PyEPG 그래버",
"Internal XMLTV EPG Grabber", "내부 XMLTV EPG 그래버",
"Internal error", "내부 에러",
"Internal grabber", "내부 그래버",
"Interview", "인터뷰",
"Invalid", "잘못됨",
"Invalid EPG object request", "잘못된 EPG 오브젝트 요청",
"Invalid adapter number '%s'\n", "잘못된 어댑터 번호 '%s'\n",
"Invalid arguments", "잘못된 ",
"Invalid file", "잘못된 파일",
"Invalid parameters", "잘못된 매개변수",
"Invalid service", "잘못된 서비스",
"Invalid target", "잘못된 대상",
"Items", "항목",
"Jazz", "재즈",
"Keep", "유지",
"Keep timeshift buffers in RAM only. With this option enabled, the amount of rewind time is limited by how much RAM TVHeadend is allowed.", "RAM에서만 타임시프트 버퍼를 유지합니다. 이 옵션이 켜지면, Tvheadend에 허용된 RAM 용량에 따라 되감기 시간이 제한됩니다.",
"Keep x amount of storage space free.", "저장소 여유 공간을 x 만큼 유지합니다.",
"Keepalive interval", "킵얼라이브 주기",
"Keepalive interval in seconds", "초 단위 킵얼라이브 주기",
"Key and value pairs to remove from the query string in the URL.", "URL 쿼리 문자열에서 제거할 키, 값 쌍입니다.",
"Kill signal (pipe)", "킬 신호 (파이프)",
"Kill timeout (pipe/secs)", "킬 시간 제한 (파이프/초)",
"LNA (low noise amplifier)", "LNA (저잡음 증폭기)",
"LNB type", "LNB 타입",
"Language", "언어",
"Language 1", "언어 1",
"Language 2", "언어 2",
"Language 3", "언어 3",
"Language settings", "언어 설정",
"Languages", "언어",
"Largest size the object has reached.", "오브젝트가 가장 큰 크기에 도달했습니다.",
"Last played position", "최근 재생한 위치",
"Last played position when the recording isn't fully watched yet.", "시청 완료하지 않은 녹화의 최종 재생한 위치입니다.",
"Last seen", "마지막 탐색",
"Last updated from", "마지막 갱신: ",
"Layer A: Constellation", "레이어 A: 배치",
"Layer A: FEC", "레이어 A: FEC",
"Layer A: Segment count", "레이어 A: 세그먼트 수",
"Layer A: Time interleaving", "레이어 A: 시간 인터리빙",
"Layer B: Constellation", "레이어 B: 배치",
"Layer B: FEC", "레이어 B: FEC",
"Layer B: Segment count", "레이어 B: 세그먼트 수",
"Layer B: Time interleaving", "레이어 B: 시간 인터리빙",
"Layer C: Constellation", "레이어 C: 배치",
"Layer C: FEC", "레이어 C: FEC",
"Layer C: Segment count", "레이어 C: 세그먼트 수",
"Layer C: Time interleaving", "레이어 C: 시간 인터리빙",
"Leisure hobbies", "레져 취미",
"Limit connections", "연결 제한",
"Limited/limit scanning to this network ID only.", "이 네트워크 ID만 스캔 제한.",
"Linked input", "연결된 입력",
"Linux DVB CA", "리눅스 DVB CA",
"Linux DVB frontend", "리눅스 DVB 프론트앤드",
"LinuxDVB Input", "LinuxDVB 입력",
"LinuxDVB adapter", "LinuxDVB 어댑터",
"LinuxDVB network", "LinuxDVB 네트워크",
"List of allowed IPv4 or IPv6 hosts or networks (comma-separated).", "허용된 IPv4 또는 IPv6 호스트 또는 네트워크 목록입니다(콤마로 구분).",
"List subsystems", "하부시스템 목록",
"Listen / Connect port", "듣기 / 연결 포트",
"Listen on IPv6", "IPv6에서 듣기",
"Literature", "문학",
"Local (server) time", "로컬 (서버) 시각",
"Local bind IP address", "로컬 IP 주소 할당",
"Local channel minor", "로컬 채널 마이너",
"Local channel number", "로컬 채널 번호",
"Local discovery IP address", "로컬 IP 주소 탐색",
"Local: Record if different description", "로컬: 설명이 다른 경우 녹화",
"Local: Record if different episode number", "로컬: 에피소드 번호가 다른 경우 녹화",
"Local: Record if different subtitle", "로컬: 부제가 다른 경우 녹화",
"Local: Record if different title", "로컬: 제목이 다른 경우 녹화",
"Local: Record once per day", "로컬: 하루 한 번 녹화",
"Local: Record once per month", "로컬: 한 달에 한 번 녹화",
"Local: Record once per week", "로컬: 일주일에 한 번 녹화",
"Location", "지역",
"Location details of the SAT>IP Server.", "SAT>IP 서버 지역 상세 정보.",
"Lock preferred CA PID", "선호하는 CA PID 잠금",
"Locking", "잠금",
"Log", "로그",
"Login password.", "로그인 암호.",
"Login username.", "로그인 사용자 이름.",
"Login/Logout", "로그인/로그아웃",
"Loglevel debug: disabled", "로그레벨 디버그: 비활성",
"Loglevel debug: enabled", "로그레벨 디버그: 활성",
"Logout", "로그아웃",
"Low", "낮음",
"Low frequency offset", "저 주파수 오프셋",
"Lowest channel number the user can access.", "사용자가 접근할 수 있는 가장 낮은 채널 번호.",
"Lowest starting channel number.", "가장 낮은 시작 채널 번호입니다.",
"MP4 audio", "MP4 오디오",
"MPEG-2 audio", "MPEG-2 오디오",
"MPEG-TS File", "MPEG-TS 파일",
"MPEG-TS Input Debug", "MPEG-TS 입력 디버그",
"MPEG-TS Multiplex", "MPEG-TS 멀티플렉스",
"MPEG-TS Parser", "MPEG-TS 파서",
"MPEG-TS Pass-thru", "MPEG-TS 패스스루",
"MPEG-TS Pass-thru/built-in", "MPEG-TS 패스스루/빌트인",
"MPEG-TS Service", "MPEG-TS 서비스",
"MPEG-TS input", "MPEG-TS 입력",
"MPEG-TS multiplex PHY", "MPEG-TS 멀티플렉스 PHY",
"MPEG-TS raw service", "MPEG-TS 그대로 서비스",
"Magazines", "매거진",
"Maintain free storage space in MiB", "저장소 여유 공간 유지 MiB 단위",
"Maintain used storage space in MiB (0=disabled)", "저장소 사용 공간 유지 MiB 단위 (0=해제)",
"Maintained space", "유지된 공간",
"Make subdirectories per channel", "채널별 서브디렉토리 생성",
"Make subdirectories per day", "일별 서브디렉토리 생성",
"Make subdirectories per title", "제목별 서브디렉토리 생성",
"Manage all", "모두 관리",
"Manufacturer", "제조사",
"Manufacturer URL", "제조사 URL",
"Manufacturer's model number.", "제조사 모델 번호입니다.",
"Manufacturer's product description.", "제조사 제품 설명입니다.",
"Manufacturer's product information page for the device.", "장치의 제조사 제품 정보 페이지입니다.",
"Manufacturer's product name.", "제조사 제품 이름입니다.",
"Map all services", "모든 서비스 매핑",
"Map encrypted services", "암호화된 서비스 매핑",
"Map radio channels", "라디오 채널 매핑",
"Map unnamed channels", "원하지 않은 채널 매핑",
"Map zero-numbered channels", "0번 채널 매핑",
"Mapped", "매핑됨",
"Martial sports", "격투",
"Master tuner", "마스터 튜너",
"Matroska muxer", "Matroska 먹서",
"Maximal channel number", "최대 채널 번호",
"Maximum # input streams", "최대 # 입력 스트림",
"Maximum PIDs", "최대 PID",
"Maximum RAM size (MB)", "최대 RAM 크기 (MB)",
"Maximum allowed difference between event start time when the EPG event is changed.", "EPG 이벤트가 변경됐을 이벤트 시작 시간 사이 허용된 최대 차이",
"Maximum bandwidth (Kbps)", "최대 대역폭 (Kbps)",
"Maximum count (0=default)", "최대 개수 (0=기본값)",
"Maximum duration", "최대 길이",
"Maximum input bandwidth.", "최대 입력 대역폭입니다.",
"Maximum length in characters for the command setting PIDs to the SAT>IP box.", "SAT>IP 박스에 대한 PID 명령 설정의 최대 글자 길이",
"Maximum length of PIDs", "PID 최대 길이",
"Maximum period (mins)", "최대 시간 (분)",
"Maximum schedules limit (0=default)", "최대 예약 제한 (0=기본값)",
"Maximum size (MB)", "최대 크기 (MB)",
"Maximum time to wait (in seconds) for a stream before a timeout.", "스트림에서 시간 초과까지 기다릴 최대 시간(초)입니다.",
"Maximum timeout (seconds)", "최대 시간 제한 (초)",
"Medicine", "의료",
"Melodrama", "멜로 드라마",
"Memory Information", "메모리 정보",
"Merge same name", "같은 이름 병합",
"Merge services with the same name into one channel.", "같은 이름의 서비스를 한 채널로 병합",
"Method not found", "기능 찾을 수 없음",
"Minimal channel number", "최소 채널 번호",
"Minimum duration", "최소 길이",
"Minimum rotor time (seconds)", "최소 로터 시간 (초)",
"Missing In PAT/SDT", "PAT/SDT 없음",
"Mode", "모드",
"Model description", "모델 설명",
"Model name", "모델 이름",
"Model number", "모델 번호",
"Modulation", "조정",
"Module", "모듈",
"Module ID", "모듈 ID",
"Module ID used to grab EPG data.", "EPG 데이터를 가져올 모듈 ID입니다.",
"Mon", "월",
"Monitonic timer", "단조함수 타이머",
"Mono", "모노",
"More verbose libav log", "더 자세한 libav 로그",
"Motor rate (milliseconds/deg)", "모터 속도 (밀리초/도)",
"Motor sport", "자동차 스포츠",
"Motoring", "자동차",
"Movie", "영화",
"Movie / drama", "영화 / 드라마",
"Multiple lines of the cron time specification. The default cron triggers the internal grabbers daily at 12:04 and 00:04. See Help on how to define your own.", "크론 타임 정의 문자열입니다. 기본 크론 트리거는 내부 그래버를 매일 12:04와 00:04에 발동시킵니다. 도움말을 참고하여 직접 설정해보세요.",
"Multiple lines of the cron time specification. The default cron triggers the over-the-air grabber daily at 02:04 and 14:04. See Help on how to define your own.", "크론 타임 정의 문자열입니다. 기본 크론 트리거는 매일 02:04와 14:04에 OTA 그래버를 발동시킵니다. 도움말을 참고하여 직접 설정해보세요.",
"Music", "음악",
"Music / Ballet / Dance", "음악 / 발레 / 댄스",
"Musical", "뮤지컬",
"Mux", "먹스",
"Mux Scheduler", "먹스 스케쥴러",
"Mux UUID", "먹스 UUID",
"Mux handling", "먹스 핸들링",
"Mux name", "먹스 이름",
"Mux not enabled", "먹스 활성화되지 않음",
"Muxer", "먹서",
"NONE", "없음",
"Name", "이름",
"Name (or date) of program the entry is a rerun of.", "재방송 프로그램 이름 (또는 날짜)입니다.",
"Name of channel the entry recorded from.", "녹화된 항목의 채널입니다.",
"Name of object.", "오브젝트의 이름입니다.",
"Name of the bouquet.", "채널 묶음의 이름입니다.",
"Name of the client.", "클라이언트의 이름입니다.",
"Name of the entry.", "항목의 이름입니다.",
"Name of the module used to grab EPG data.", "EPG 데이터를 가져오는데 쓰이는 모듈의 이름입니다.",
"Name of the module, grabber or channel that should be used to update this channels EPG info.", "EPG 원본. 이 채널 EPG 정보를 갱신하는데 쓰일 모듈 이름, 그래버 또는 채널.",
"Name of the network.", "네트워크의 이름입니다.",
"Name of the profile.", "프로필의 이름입니다.",
"Name of the tag.", "태그의 이름입니다.",
"Name of the tuner.", "튜너의 이름입니다.",
"Name of the tuner/adapter.", "튜너/어댑터의 이름입니다.",
"Names", "이름",
"Natural sciences", "자연 과학",
"Nature", "자연",
"Network", "네트워크",
"Network ", "네트워크",
"Network A", "네트워크 A",
"Network B", "네트워크 B",
"Network C", "네트워크 C",
"Network D", "네트워크 D",
"Network ID (limit scanning)", "네트워크 ID (스캔 제한)",
"Network UUID", "네트워크 UUID",
"Network access", "네트워크 접근",
"Network discovery", "네트워크 탐색",
"Network for port A.", "포트 A에 대한 네트워크입니다.",
"Network for port AA.", "AA 포트에 대한 네트워크입니다.",
"Network for port AB.", "AB 포트에 대한 네트워크입니다.",
"Network for port B.", "B 포트에 대한 네트워크입니다.",
"Network for port BA.", "BA 포트에 대한 네트워크입니다.",
"Network for port BB.", "BB 포트에 대한 네트워크입니다.",
"Network for port C.", "포트 C에 대한 네트워크입니다.",
"Network for port D.", "포트 D에 대한 네트워크입니다.",
"Network group", "네트워크 그룹",
"Network limit per position", "위치당 네트워크 제한",
"Network name", "네트워크 이름",
"Network prefix", "네트워크 프리픽스",
"Network settings", "네트워크 설정",
"Network type", "네트워크 종류",
"Networks", "네트워크",
"New / premiere / unknown", "신규 / 프리미어 / 알 수 없음",
"New login", "신규 로그인",
"New media", "새 미디어",
"New muxes + changed muxes", "새 먹스 + 바뀐 먹스",
"New muxes only", "새 먹스만",
"News", "뉴스",
"News / Current Affairs", "뉴스 / 시사",
"News magazine", "뉴스 매거진",
"Next tune delay in ms (0-2000)", "ms 단위 다음 튠 딜레이 (0-2000)",
"No", "아니오",
"No PIN", "PIN 없음",
"No access", "접근 없음",
"No adapters specified!\n", "어댑터가 지정되지 않았습니다!\n",
"No assigned adapters", "할당된 어댑터 없음",
"No days", "날 없음",
"No descrambler", "복호화 방법 없음",
"No free adapter", "프리 어댑터 없음",
"No input detected", "입력 탐지되지 않음",
"No scheme", "스키마 없음",
"No service assigned to channel", "채널에 서비스가 할당되지 않음",
"No service enabled", "활성 서비스 없음",
"No source available", "가용 원본 없음",
"Node subsystem", "노드 하부시스템",
"None", "없음",
"Normal", "보통",
"Not all SAT>IP servers use the same signal scaling. Change this setting if the signal level displayed within Tvheadend looks too low.", "모든 SAT>IP 서버가 같은 신호 크기를 사용하진 않습니다. Tvheadend에서 너무 낮게 표시된다면, 이 값을 바꿔보세요.",
"Not enough disk space", "디스크 공간 부족",
"Not enough memory", "메모리 부족",
"Not linked", "링크되지 않음",
"Not set", "설정되지 않음",
"Not set (none or channel configuration)", "설정되지 않음 (없음 또는 채널 설정)",
"Not set (use channel or DVR configuration)", "설정되지 않음 (채널 또는 DVR 설정 사용)",
"Number", "번호",
"Number of days to keep the file.", "파일을 보관할 날 수입니다.",
"Number of days to keep the recorded file.", "녹화된 파일을 보관할 날 수입니다.",
"Number of days to keep the recorded files.", "녹화된 파일을 보관할 날 수입니다.",
"Number of days to retain entry information.", "항목 정보를 가져올 날 수입니다.",
"Number of days to retain information about recording.", "녹화에 대한 정보를 가져올 날 수입니다.",
"Number of days to retain information about recordings. Once this period is exceeded, duplicate detection will not be possible anymore.", "녹화에 대한 정보를 가져올 날 수입니다. 이 기간이 지나면, 더 이상 중복 확인이 불가합니다.",
"Number of errors that occurred during recording (Transport errors).", "녹화중 발생한 오류 수입니다 (전송 에러).",
"Number of errors that occurred during recording.", "녹화중 발생한 오류 수입니다.",
"Number of muxes found.", "발견된 먹스의 수입니다.",
"Number of repeats for the DiseqC commands (default is zero - no DiseqC repeats). Note: this represents the number of repeats, not the number of requests - so 0 means 'send once: don't repeat', 1 means 'send twice: send once, then send one repeat', etc.", "다이섹 명령을 반복할 횟수입니다 (기본값 0 - 다이섹 반복 없음). 참고: 이는 반복 횟수이지, 요청 횟수가 아닙니다. - 따라서 0은 '한 번 보냄; 반복하지 않음', 1은 '두 번 보냄; 한 번 보내고, 한 번 반복함', 등의 의미입니다.",
"Number of repeats for the tune requests (default is zero - no repeats). Note: this represents the number of repeats, not the number of requests - so 0 means 'send once: don't repeat', 1 means 'send twice: send once, then send one repeat', etc.", "튠 요청을 반복할 횟수입니다 (기본값 0 - 반복 없음). 참고: 이는 반복 횟수이지, 요청 횟수가 아닙니다. - 따라서 0은 '한 번 보냄; 반복하지 않음', 1은 '두 번 보냄; 한 번 보내고, 한 번 반복함', 등의 의미입니다.",
"Number of seconds to wait before timing out.", "시간 초과 판정을 낼 때까지 기다릴 초 단위 시간입니다.",
"Number of services.", "서비스 개수입니다.",
"Number of times this recording was played.", "녹화가 재생된 횟수입니다.",
"Number of tsfile tuners", "tsfile 튜너 개수",
"OK", "확인",
"OK (partial)", "(부분적) 확인",
"Odd key", "Odd 키",
"Odd key.", "Odd 키입니다.",
"Off", "끄기",
"Offset the mapped channel numbers by x (value here + channel number).", "매핑된 채널을 x만큼 이동합니다 (여기 입력한 값 + 채널 번호).",
"Older OSCam", "오래된 OSCam",
"On", "켜기",
"On file removal", "파일을 지울 때",
"On-demand (no first rewind)", "수동으로 (첫 되감기는 불가)",
"Once per auto channel", "자동 채널당 한 번",
"One tune request (setup) is sent before the DiseqC sequence (voltage, tone settings). Some linux drivers require this procedure.", "단일 튠 요청(설정)이 다이섹 시퀀스(전압, 톤 설정)보다 먼저 전송됩니다. 일부 리눅스 드라이버는 이 기능이 필요합니다.",
"Only Bulsatcom 39E", "Bulsatcom 39E 전용",
"Only EIT", "EIT 전용",
"Only OpenTV Sky Ausat", "OpenTV Sky Ausat 전용",
"Only OpenTV Sky Italia", "OpenTV Sky Italia 전용",
"Only OpenTV Sky UK", "OpenTV Sky UK 전용",
"Only PSIP (ATSC)", "PSIP (ATSC) 전용",
"Only UK Freesat", "UK Freesat 전용",
"Only UK Freeview", "UK Freeview 전용",
"Only Viasat Baltic", "Viasat Baltic 전용",
"Only activate timeshift when the client makes the first rewind, fast-forward or pause request. Note, because there is no buffer on the first request rewinding is not possible at that point.", "타임시프트 기능은 클라이언트가 처음 되감기, 빨리감기, 일시 정지를 했을 경우에 활성화됩니다. 참고로, 처음에는 버퍼를 쓰지 않으므로 해당 시점에선 되감기가 불가합니다.",
"Only admin can watch the realtime log.", "관리자만 실시간 로그를 확인할 수 있습니다.",
"Only allow users with this tag (or those with no tags at all) set in access configuration to use the tag.", "사용자만이 이 태그 (또는 모든 태그 없이)로 태그를 사용하도록 접근 설정을 합니다.",
"Only digits", "숫자만",
"Only preferred CA PID", "선호하는 CA PID만",
"Only update the system clock (doesn't affect NTP driver) if the delta between the system clock and DVB time is greater than this. This can help stop excessive oscillations on the system clock.", "시스템 시계와 DVB 시각의 차이가 이 값보다 커야 시스템 시계를 갱신합니다 (NTP 드라이버에 영향 없음). 이는 시스템 시계가 지나치게 변동되는 것을 막는데 도움을 줍니다.",
"Only use specified DVB adapters (comma-separated, -1 = none)", "지정한 DVB 어댑터만 사용 (반점으로 구분된 문자열, -1 = 없음)",
"Only use this EPG data once when automatically determining what EPG data to set for a channel.", "채널에 설정된 EPG 데이터를 자동으로 판별할 때 이 EPG 데이터만을 한 번 사용합니다.",
"OpenTV channel number", "OpenTV 채널 번호",
"Opera", "오페라",
"Options to use/used when mapping - see Help for details.", "매핑할 때 사용될 옵션입니다 - 자세한 정보는 도움말을 보세요.",
"Options used when updating channels.", "채널을 업데이트할 때 사용될 옵션입니다.",
"Orbital position", "궤도 위치",
"Orbital positions", "궤도 위치",
"Orbital positions.", "궤도 위치입니다.",
"Original network ID", "원본 네트워크 ID",
"Oscam mode.", " Oscam 모드입니다.",
"Other Stream Filter", "다른 스트림 필터",
"Output directory for tsdebug", "tsdebug할 출력 디렉토리",
"Over-the-air", "OTA",
"Over-the-air Cron multi-line", "OTA 크론 다중 행간",
"Over-the-air EPG", "OTA EPG",
"Over-the-air EPG grabber", "OTA EPG 그래버",
"Over-the-air grabbers", "OTA 그래버",
"Override disabled", "오버라이드 비활성",
"Override the frontend type.", "프론트엔드 종류를 오버라이드합니다.",
"Owner", "소유자",
"Owner of the entry.", "항목의 소유자입니다.",
"Owner of the rule.", "규칙의 소유자입니다.",
"PEND", "진행",
"PIDs 21 in setup", "설정: PIDs 21",
"PIN inquiry match string", "PIN 문자열 일치 조사",
"PIN inquiry match string.", "PIN 문자열이 일치하는지 확인합니다.",
"PIN.", "PIN입니다.",
"PLS code", "PLS 코드",
"PLS mode", "PLS 모드",
"PROXY protocol is an extension for support incoming TCP connections from a remote server (like a firewall) sending the original IP address of the client. The HTTP header 'X-Forwarded-For' do the same with HTTP connections. Both enable tunneled connections.This option should be disabled for standard usage.", "프록시 프로토콜은 클라이언트의 원본 IP 주소를 보내는 원격 서버(방화벽 등)에서 들어오는 TCP 연결을 지원하는 확장 기능입니다. HTTP 헤더 'X-Forwarded-For'는 HTTP 연결과 같은 역할을 합니다. 터널링된 연결에 둘을 허용합니다. 이 옵션은 일반적인 경우 꺼져있어야 합니다.",
"Parent entry", "부모 항목",
"Parent entry.", "부모 항목입니다.",
"Pass subscription weight", "패스 구독 순위",
"Pass subscription weight to the SAT>IP server (Tvheadend specific extension).", "SAT>IP 서버에 대한 구독 순위(Tvheadend 고유 확장 프로그램)입니다.",
"Pass-thru muxer", "패스스루 먹서",
"Passthrough Muxer SI Tables", "Passthrough 먹서 SI 테이블",
"Password", "암호",
"Password for the entry.", "항목에 대한 암호입니다.",
"Password2", "암호2",
"Path", "경로",
"Path to a directory (folder) containing your picon collection. See Help for more detailed information.", "picon 컬렉션이 위치한 디렉토리 (폴더) 경로입니다. 도움말을 참고하여 자세한 정보를 확인하세요.",
"Path to an icon for this channel. This can be named however you wish, as either a local (file://) or remote (http://) image. See Help for more infomation.", "이 채널의 아이콘 경로입니다. 로컬(file://) 또는 원격(http://) 이미지를 기호에 따라 지정할 수 있습니다. 도움말을 참고하여 더 자세히 알아보세요.",
"Path to the frontend used by the device.", "장치에서 사용될 프론트엔드 경로입니다.",
"Path to the grabber executable.", "그래버를 실행 가능한 경로입니다.",
"Path to the socket Tvheadend will read data from.", "Tvheadend가 데이터를 읽어올 소켓 경로입니다.",
"Path to where the timeshift data will be stored. If nothing is specified this will default to CONF_DIR/timeshift/buffer.", "타임시프트 데이터가 저장될 경로입니다. 아무 것도 지정되지 않으면 기본값으로 CONF_DIR/timeshift/buffer 를 사용합니다.",
"Path used by the device.", "장치에서 사용되는 경로입니다.",
"Path where the recordings are stored. If components of the path do not exist, Tvheadend will try to create them.", "녹화 파일이 저장될 경로입니다. 경로가 존재하지 않으면, Tvheadend가 생성할 것입니다.",
"Peak count of objects", "오브젝트 최대 수",
"Peak size", "최대 크기",
"Performing arts", "행위 예술",
"Periodically save EPG to disk (hours)", "EPG를 디스크에 저장할 주기 (시간)",
"Persistent user interface level", "주로 쓰는 유저 인터페이스 레벨",
"Physiology", "생리학",
"Pick the stream with given audio type only.", "주어진 오디오 종류만으로 스트림을 선택합니다.",
"Picon name scheme", "Picon 이름 스키마",
"Picon path", "Picon 경로",
"Pilot", "파일럿",
"Please, follow %s link and cancel the next authorization to correctly clear the cached browser credentals (login and password cache). Then click to the 'Default login' (anonymous access) or 'New login' link in the error page to reauthenticate.", "%s 링크를 따라가서, 브라우저에 캐시된 인증(로그인과 암호 캐시)을 올바르게 제거하기 위해 다음 인증을 취소하세요. 그 다음 재인증을 위해 에러 페이지에서 '기본 로그인' (익명 접근) 또는 '신규 로그인' 링크를 클릭하세요.",
"Polarization", "편파",
"Political issues", "정치 이슈",
"Poll multiplexer", "Poll 멀티플렉서",
"Pop", "팝",
"Popular culture", "대중 문화",
"Port", "포트",
"Port to connect to.", "연결할 포트입니다.",
"Port to listen on or to connect to.", "수신하거나 연결할 포트입니다.",
"Position", "위치",
"Position of the input.", "입력의 위치입니다.",
"Post-processor command", "후처리 명령",
"Post-recording padding", "녹화 후 여유 시간",
"Post-remove command", "삭제 후 명령",
"Power save", "전원 절약",
"Power-up time (ms) (15-200)", "전원 켜기 시간 (ms) (15-200)",
"Pre-defined muxes", "미리 지정된 먹스",
"Pre-processor command", "전처리 명령",
"Pre-recording padding", "녹화 전 여유 시간",
"Pre-school children's programs", "유아 프로그램",
"Prefer picons over channel name", "채널 이름보다 picon 선호",
"Preferred CA PID", "선호하는 CA PID",
"Preferred audio language.", "선호하는 오디오 언어입니다.",
"Preferred service video type", "선호하는 서비스 비디오 종류",
"Presentation", "프리젠테이션",
"Presentation details.", "프리젠테이션 정보입니다.",
"Press", "언론",
"Prevent the user from overriding the default user interface level setting and removes the view level drop-dowm from the interface.", "사용자가 인터페이스 레벨을 바꾸지 못하게 하며, 인터페이스에서 보기 레벨 드롭다운 박스를 제거합니다.",
"Prevent users from overriding the user interface level setting, and remove the view level drop-dowm from the interface.", "사용자가 인터페이스 레벨을 바꾸지 못하게 하며, 인터페이스에서 보기 레벨 드롭다운 박스를 제거합니다.",
"Priority", "우선순위",
"Priority (-10..10)", "우선순위 (-10..10)",
"Priority of the entry, higher-priority entries will take precedence and cancel lower-priority events.", "이 규칙에 의해 정해진 녹화 우선순위는 우선하게 되며, 낮은 순위의 이벤트를 취소할 수 있습니다.",
"Priority of the recording. Higher priority entries will take precedence and cancel lower-priority events.", "이 규칙에 의해 정해진 녹화 우선순위는 우선하게 되며, 낮은 순위의 이벤트를 취소할 수 있습니다.",
"Private", "비공개",
"Profile name", "프로필 이름",
"Program identification (PID) number to compare. Zero means any. This comparison is processed only when service comparison is active and for the Conditional Access filter.", "비교할 프로그램 식별(PID) 번호입니다. 0은 모두를 뜻합니다. 이 비교는 서비스 비교가 활성화되어 있고, CA 필터용일 때 처리됩니다.",
"Program synopsis (display only).", "프로그램 시놉시스입니다 (표시전용).",
"Program synopsis.", "프로그램 시놉시스입니다.",
"Provider", "제공자",
"Provider ID", "제공자 ID",
"Provider network name", "제공자 네트워크 이름",
"Provider's network name.", "제공자의 네트워크 이름입니다.",
"Psychology", "심리학",
"Puppets", "인형극",
"PyEPG Import", "PyEPG 가져오기",
"Quad (4.0)", "쿼드 (4.0)",
"Quiz", "퀴즈",
"RAM only", "램만 사용",
"RTP/AVP/TCP (embedded data)", "RTSP/AVP/TCP (포함된 데이터)",
"RTSP Protocol", "RTSP 프로토콜",
"RTSP port", "RTSP 포트",
"RTSP port (554 or 9983, 0 = disable)", "RTSP 포트 (554 또는 9983, 0 = 비활성)",
"Radio", "라디오",
"Raw Audio Stream", "원시 오디오 스트림",
"Re-fetch period (hours)", "다시 가져오기 간격 (시간)",
"Re-fetch period (mins)", "다시 가져오기 간격 (분)",
"Re-fetch the bouquet every x minutes.", "채널 묶음을 x분마다 다시 가져옵니다.",
"Re-record%s%s", "다시 녹화%s%s",
"Re-try period (hours)", "다시 가져오기 간격 (시간)",
"Read error", "읽기 오류",
"Real Time Streaming Protocol (RTSP) port the server should listen on (554 or 9983, 0 = disable).", "서버의 실시간 스트리밍 프로토콜 (RTSP)는 다음 포트에서 리스닝되어야 합니다 (554 또는 9983, 0 = 비활성).",
"Record all", "모두 녹화",
"Record on these days only.", "이 날만 녹화합니다.",
"Recording file options", "녹화 파일 옵션",
"Recording file size.", "녹화할 파일 크기입니다.",
"Recording play count", "녹화 재생 횟수",
"Recording system path", "녹화할 시스템 경로입니다.",
"Reject", "꺼내기",
"Reject exact match", "일치하는 항목 걸러내기",
"Relative path to the imagecache copy of the icon.", "아이콘 이미지 캐시의 상대 경로입니다.",
"Religion", "종교",
"Religious", "종교적인",
"Remarkable people", "위인",
"Remove HTTP arguments", "HTTP 매개변수 제거",
"Remove all unsafe characters from filename", "파일 이름에서 부적절한 문자열 모두 제거",
"Remove scrambled bits", "암호화된 비트 제거",
"Repeated", "반복됨",
"Replace whitespace in title with '-'", "제목의 공백을 '-'로 치환",
"Replaces all whitespace in the title with '-'.", "모든 공백 포함된 제목을 '-'로 치환합니다.",
"Reply to CAM PIN inquiries", "CAM PIN 질의에 응답",
"Reply to PIN inquiries.", "PIN 질의에 응답합니다.",
"Reports", "보고",
"Rerun of", "재방송: ",
"Reserved", "예약됨",
"Resolution (height)", "해상도 (세로)",
"Respawn (pipe)", "리스폰 (파이프)",
"Restart", "재시작",
"Restart on error", "오류시 다시 시작",
"Restart streaming on error. This is useful for DVR so a recording isn't aborted if an error occurs.", "오류시 스트리밍을 다시 시작합니다. DVR의 경우 오류가 발생해도 녹화를 중단하지 않아 유용합니다.",
"Restart when the subscription is overriden (in then timeout time window).", "(타임아웃 시간 창에서) 구독이 중복되면 재시작합니다.",
"Restrict connections to this type.", "이 종류의 연결을 제한합니다.",
"Restricted log mode (no administrator)", "제한된 로그 모드 (관리자 아님)",
"Reuse EPG from", "EPG 재사용",
"Reuse the EPG from another channel.", "다른 채널에서 EPG를 재사용합니다.",
"Rewrite EIT", "EIT 다시 쓰기",
"Rewrite EIT (Event Information Table) packets to only include information about the currently-streamed service.", "현재 스트림되는 서비스에 대한 정보만을 포함하는 EIT (이벤트 정보 테이블)를 다시 씁니다.",
"Rewrite MPEG-TS SI tables", "MPEG-TS SI 테이블 다시 쓰기",
"Rewrite PAT", "PAT 다시 쓰기",
"Rewrite PAT (Program Association Table) packets to only include information about the currently-streamed service.", "현재 스트림되는 서비스에 대한 정보만을 포함하는 PAT (프로그램 연결 테이블)를 다시 씁니다.",
"Rewrite PMT", "PMT 다시 쓰기",
"Rewrite PMT (Program Map Table) packets to only include information about the currently-streamed service.", "현재 스트림되는 서비스에 대한 정보만을 포함하는 PMT (프로그램 맵 테이블)를 다시 씁니다.",
"Rewrite Program Map Table (PMT) packets to only include information about the currently streamed service.", "현재 스트림되는 서비스에 대한 정보만을 포함하는 프로그램 맵 테이블 (PMT)을 다시 씁니다.",
"Rewrite SDT", "SDT 다시 쓰기",
"Rewrite SDT (Service Description Table) packets to only include information about the currently-streamed service.", "현재 스트림되는 서비스에 대한 정보만을 포함하는 SDT (서비스 설명 테이블)를 다시 씁니다.",
"Rights", "권한",
"Rock", "록",
"Rolloff", "롤 오프",
"Romance", "로맨스",
"Rotor initialization time (seconds)", "로터 초기화 시간 (초)",
"Rotor type", "로터 종류",
"Run as group", "그룹으로 실행",
"Run as user", "유저로 실행",
"Running", "실행 중",
"SAT->IP frontend number.", "SAT->IP 프론트엔드 번호입니다.",
"SAT>IP ATSC-C Frontend", "SAT>IP ATSC-C 프론트엔드",
"SAT>IP ATSC-T Frontend", "SAT>IP ATSC-T 프론트엔드",
"SAT>IP Client", "SAT>IP 클라이언트",
"SAT>IP DVB Frontend", "SAT>IP DVB 프론트엔드",
"SAT>IP DVB-C Frontend", "SAT>IP DVB-C 프론트엔드",
"SAT>IP DVB-S Frontend", "SAT>IP DVB-S 프론트엔드",
"SAT>IP DVB-S Slave Frontend", "SAT>IP DVB-S 슬레이브 프론트엔드",
"SAT>IP DVB-T Frontend", "SAT>IP DVB-T 프론트엔드",
"SAT>IP DVB-T frequency (Hz)", "SAT>IP DVB-T 주파수 (Hz)",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "SAT>IP RTSP 서버 포트 번호\n(기본값: -1 = 끄기, 0 = 웹 설정, 기본 설정은 554)",
"SAT>IP Satellite Configuration", "SAT>IP 위성 설정",
"SAT>IP Server", "SAT>IP 서버",
"SAT>IP Server SI Tables", "SAT>IP 서버 SI 테이블",
"SAT>IP source number", "SAT>IP 소스 번호",
"SCR (Satellite Channel Router) ID.", "SCR (위성 채널 라우터) ID.",
"SD: standard definition", "SD: 표준 화질",
"SSL verify peer", "SSL 피어 인증",
"START", "시작",
"STOP", "정지",
"Sat", "토",
"Satconf", "위성 설정",
"Satellite config", "위성 설정",
"Satellite longitude", "위성 경도",
"Satellite longitude.", "위성 경도입니다.",
"Satellite position.", "위성 위치입니다.",
"Satellite positions", "위성 위치",
"Scan after creation", "생성 후 스캔",
"Scan progress", "스캔 진행률",
"Scan queue length", "스캔 대기열 길이",
"Scan result", "스캔 결과",
"Scan status", "스캔 상태",
"Scanfile", "스캔 파일",
"Schedule a re-recording if more errors than (0=off)", "에러가 다음 임계점을 초과하면 다시 녹화 예약 (0=꺼짐)",
"Schedule frequency (in cron format).", "예약 주기 (크론 포맷)",
"Schedule status", "예약 상태",
"Schedule status.", "예약 상태입니다.",
"Scheduled Duration", "예약 길이",
"Scheduled for recording", "녹화 예약됨",
"Scheduled start time", "예약 시작 시간",
"Scheduled stop time", "예약 끝 시각",
"Scheme to generate the channel icon names (all lower-case, service name picons etc.).", "채널 아이콘 이름을 생성할 스키마입니다 (모두 소문자, 서비스 picon 등).",
"School programs", "학교 프로그램",
"Science", "과학",
"Science fiction", "과학 픽션",
"Script/program to run when a recording completes.", "녹화가 완료되면 실행할 스크립트 또는 프로그램입니다.",
"Script/program to run when a recording gets removed.", "녹화가 제거되면 실행할 스크림트 또는 프로그램입니다.",
"Script/program to run when a recording starts (service is subscribed but no filename available).", "녹화가 시작되면 실행할 스크립트 또는 프로그램입니다 (서비스는 구독되었지만 파일 이름이 없음).",
"Season", "시즌",
"Season %d", "시즌 %d",
"Season information (if available).", "시즌 정보(유효한 경우)입니다.",
"Seek error", "탐색 오류",
"Select an available network type for this tuner.", "이 튜너에 유효한 네트워크 종류를 선택하세요.",
"Select high priority (default) EPG language.", "(기본값) EPG 언어에서 높은 우선 순위를 선택하세요.",
"Select how Tvheadend should handle muxes. See Help for details.", "Tvheadend가 먹스를 어떻게 다뤄야 할지 선택합니다. 도움말을 눌러서 자세히 알아보세요.",
"Select low priority EPG language.", "EPG 언어에서 낮은 우선 순위를 선택하세요.",
"Select medium priority EPG language.", "EPG 언어에서 중간 우선 순위를 선택하세요.",
"Select scheme to generate the picon names (standard, force service type to 1)", "picon 이름을 생성할 스키마를 선택 (보통, 서비스 종류를 1로 강제)",
"Select subtitle codec to use for transcoding.", "트랜스코딩에 쓰일 자막 코덱을 선택하세요.",
"Select the default user interface language. This can be overridden later in \"Access Entries\" on a per-user basis.", "기본 사용자 인터페이스 언어를 선택하세요. 이는 \"접근 항목\"에서 사용자에 따라 나중에 새로 지정할 수 있습니다.",
"Select the list of languages (in order of priority) to be used for supplying EPG information to clients that don't provide their own configuration.", "별도의 설정을 제공하지 않는 클라이언트에 EPG 정보를 제공할 때 쓰일 언어 목록을 선택하세요 (우선 순위 순서로).",
"Select the master tuner.The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, the polarization and low-high band settings must be equal.if you set other tuner as master, then this tuner will act like a slave one and tvheadend will assure that this tuner will not use incompatible parameters (position, polarization, lo-hi).", "마스터 튜너를 선택하세요. 기본 범용 LNB에서 들어오는 신호는 간단한 동축 케이블 분배기(멀티스위치 제외)를 통해 여러 출력으로 나뉠 수 있습니다. 이 경우, 위치, 극성과 하위-상위 밴드 설정은 반드시 똑같아야 합니다. 다른 튜너를 마스터로 지정할 경우, 이 튜너는 슬레이브처럼 동작하며, tvheadend는 이 튜너에 호환되지 않는 매개변수(위치, 극성, 하위-상위)를 사용하지 않게 됩니다.",
"Select the number of satellite positions supported by the SAT>IP hardware and your coaxial cable wiring.", "SAT>IP 하드웨어와 동축 케이블에서 지원하는 위성 위치 번호를 지정하세요.",
"Select the path to use for DVB scan configuration files. Typically dvb-apps stores these in /usr/share/dvb/. Leave blank to use the internal file set.", "DVB 스캔 설정 파일에 사용할 경로를 지정하세요. 일반적으로 dvb 앱은 /usr/share/dvb/에 저장합니다. 내부 파일셋에 저장하려면 빈칸으로 두세요.",
"Select the time offset for EIT events.", "EIT 이벤트 시각 이동값을 선택하세요.",
"Select type of broadcast.", "방송의 종류를 선택하세요.",
"Select whether or not to enable caching. Note, even with this disabled you can still specify local (file://) icons and these will be served by the built-in webserver.", "캐싱 여부를 선택합니다. 참고: 이 기능을 꺼도, 로컬 (file://) 아이콘을 지정하면 내부 웹서버에 저장되어 제공될 것입니다.",
"Select/Selected services to map.", "매핑할 서비스를 선택합니다.",
"Send CAPMT OK query before descrambling.", "복호화하기 전 CAPMT OK 쿼리를 보냅니다.",
"Send CAPMT query", "CAPMT 쿼리 전송",
"Send full PLAY cmd", "전체 PLAY 명령 전송",
"Send previous stream frames to upper layers (before frame start is signalled in the stream). It may cause issues with some clients / players.", "이전 스트림 프레임을 상위 레이어에 전송합니다 (프레임 시작이 스트림에 신호되기 전에). 이는 일부 클라이언트 / 플레이어에서 문제를 일으킬 수 있습니다.",
"Send the full RTSP PLAY command after full RTSP SETUP command. Some devices firmware require this to get an MPEG-TS stream.", "전체 RTSP PLAY 명령을 전체 RTSP SETUP 명령 다음에 보냅니다. 일부 장치 펌웨어는 MPEG-TS 스트림을 얻기 위해 이 과정이 필요합니다.",
"Serial number", "시리얼 번호",
"Series link", "시리즈 연결",
"Series link ID.", "시리즈 연결 ID입니다.",
"Serious", "기능성",
"Serious music", "기능성 음악",
"Server", "서버",
"Server UUID", "서버 UUID",
"Server connectivity", "서버 연결",
"Server details.", "서버 상세 정보입니다.",
"Service", "서비스",
"Service ID", "서비스 ID",
"Service Mapper", "서비스 매퍼",
"Service Mapping (Map services to channels)", "서비스 매핑 (서비스를 채널에 매핑)",
"Service configuration", "서비스 구성",
"Service count %zi", "서비스 개수 %zi",
"Service mapping", "서비스 매핑",
"Service name", "서비스 이름",
"Service name found in EPG data.", "EPG 데이터에서 서비스 이름을 찾았습니다.",
"Service name picons", "picon 서비스 이름",
"Service priority. Enter a value between -10 and 10. A higher value indicates a higher preference. See Help for more info.", "서비스 우선 순위입니다. -10과 10 사이 값을 입력하세요. 높은 값은 높은 선호도를 나타냅니다. 자세한 정보는 도움말을 참조하세요.",
"Service raw", "raw 서비스",
"Service type", "서비스 종류",
"Service type override. This value will override the service type provided by the stream.", "서비스 종류를 다시 설정합니다. 이 값은 스트림에서 제공된 서비스를 다시 설정합니다.",
"Services", "서비스",
"Services associated with the channel.", "채널과 연결된 서비스입니다.",
"Services seen", "발견된 서비스",
"Set as default profile.", "기본 프로필로 설정합니다.",
"Set the display name.", "표시 이름을 설정합니다.",
"Set the maxiumum packet identifiers your SAT>IP server supports.", "SAT>IP 서버가 지원하는 최대 패킷 식별자를 설정합니다.",
"Set the name of the server so you can distinguish multiple instances apart.", "여러 인스턴스를 구분하기 위한 서버 이름을 설정하세요.",
"Set the priority of this configuration.", "이 구성의 우선 순위를 설정합니다.",
"Sets the default interface view level (next to the Help button).", "기본 인터페이스 표시 레벨을 설정합니다 (도움말 버튼 옆).",
"Settings", "설정",
"Shift PTS (ms)", "PTS 이동 (ms)",
"Show", "쇼",
"Show / Game show", "쇼 / 게임 쇼",
"Show this page", "이 페이지 표시",
"Show version information", "버전 정보 표시",
"Show, hide and sort the various details that appear on the interface next to the About tab.", "정보 탭 옆의 인터페이스에 표시할 다양한 상세 정보를 보이거나 숨기거나 정렬합니다.",
"Signal quality too poor", "신호 품질 너무 나쁨",
"Signal scale (240 or 100)", "신호 크기 (240 또는 100)",
"Site latitude", "위치 위도",
"Site latitude.", "위치 위도입니다.",
"Site longitude", "위치 경도",
"Site longitude.", "위치 경도입니다.",
"Size", "크기",
"Skip TS packets (0-200)", "TS 패킷 무시 (0-200)",
"Skip TSID checking. For when providers use invalid Transport Stream IDs.", "TSID 검사를 건너뜁니다. 잘못된 전송 스트림 ID를 사용하는 제공자를 위한 옵션입니다.",
"Skip commercials", "광고 건너뛰기",
"Skip initial bytes", "초기 바이트 무시",
"Skip scanning known muxes when Tvheadend starts. If \"startup scan\" is allowed and new muxes are found then they will still be scanned. See Help for more details.", "Tvheadend가 시작할 때 알려진 먹스의 스캔을 건너뜁니다. 만약 \"시작 스캔\"이 허용되어 있고, 새로운 먹스가 발견되면 스캔될 것입니다. 자세한 정보는 도움말을 보세요.",
"Skip startup scan", "시작 스캔 건너뛰기",
"Skip x number of transport packets.", "전송 패킷을 x번 무시합니다.",
"Slave entry", "슬레이브 항목",
"Slave entry.", "슬레이브 항목입니다.",
"Slot state", "슬롯 상태",
"Soap", "방송 드라마",
"Soccer", "축구",
"Social", "사회",
"Social / Political issues / Economics", "사회 / 정치 / 경제",
"Social advisory", "사회 자문",
"Socket or IP Address (when in TCP mode).", "소켓 또는 IP 주소입니다 (TCP 모드일 때).",
"Sort index", "정렬 색인",
"Sort index.", "정렬 색인입니다.",
"Source", "원본",
"Source deleted", "삭제된 원본",
"Source quality is bad", "원본 품질이 나쁨",
"Source reconfigured", "원본이 다시 설정됨",
"Source video codec", "원본 비디오 코덱",
"Southern hemisphere (latitude direction)", "남반구 (위도 방향)",
"Southern hemisphere (latitude direction).", "남반구(위도 방향)입니다.",
"Spawn", "스폰",
"Special events (Olympic Games, World Cup, etc.)", "특별 이벤트 (올림픽 게임, 월드컵 등)",
"Specifies the incoming buffering limit in milliseconds (PCR based). If the PCR time difference from the system clock is higher than this, the incoming stream is paused.", "(PCR 기반) 입력 버퍼링 제한을 밀리세컨드로 지정합니다. PCR 시간이 시스템 시계와 이 이상 차이가 나면, 입력 스트림은 일시 정지됩니다.",
"Specify User-Agent header for the http client", "http 클라이언트에 대한 User-Agent 헤더 지정",
"Specify alternative htsp port", "대체 htsp 포트 지정",
"Specify alternative http port", "대체 http 포트 지정",
"Specify alternative http webroot", "대체 http webroot 지정",
"Specify bind address", "주소 바인드 지정",
"Specify bind address for SAT>IP server", "SAT>IP 서버에 대한 주소 바인드 지정",
"Specify extra htsp port", "추가 htsp 포트 지정",
"Specify the parameters to be changed. See Help for details.", "변경될 매개변수를 지정합니다. 도움말에서 자세히 알아보세요.",
"Spiritual sciences", "초자연 과학",
"Sports", "스포츠",
"Sports magazines", "스포츠 매거진",
"Squash", "스쿼시",
"Standard", "보통",
"Start", "시작",
"Start after", "다음 시간 이후에 시작",
"Start before", "다음 시간 이전에 시작",
"Start recording earlier than the EPG/timer defined start time by x minutes, for example if a program is to start at 13:00 and you set a padding of 5 minutes it will start recording at 12:54:30 (including a warming-up time of 30 seconds). If this isn't set the pre-recording padding if set in the DVR entry or DVR profile will be used.", "EPG/타이머가 지정한 시각보다 x분 더 빨리 녹화를 시작합니다. 예를 들어, 프로그램이 13:00에 시작하고 5분의 여유 시간을 지정하면, 녹화는 12:54:30 (30초의 워밍업 시간 포함) 부터 시작됩니다. 녹화 전 여유 시간이 지정되지 않으면, DVR 항목이나 DVR 프로필이 사용될 것입니다.",
"Start recording earlier than the EPG/timer-defined start time by x minutes.", "지정되면 EPG/타이머 시작 시각보다 x분 더 빨리 녹화를 시작합니다.",
"Start recording earlier than the defined Start recording earlier than the defined start time by x minutes: for example, if a program is to start at 13:00 and you set a padding of 5 minutes it will start recording at 12:54:30 (including a warm-up time of 30 seconds). If this isn't specified, any pre-recording padding as set in the channel or DVR entry will be used.", "녹화 전 여유 시간입니다. 지정한 시각보다 x분 더 빨리 녹화를 시작합니다. 예를 들어, 프로그램이 13:00에 시작하고 5분의 여유 시간을 지정하면, 녹화는 12:54:30 (30초의 워밍업 시간 포함) 부터 시작됩니다. 녹화 전 여유 시간이 지정되지 않으면, DVR 항목이나 DVR 프로필이 사용될 것입니다.",
"Start recording earlier than the defined start time by x minutes.", "지정되면 시작 시각보다 x분 더 빨리 녹화를 시작합니다.",
"Start time", "시작 시간",
"Status", "상태",
"Status period (ms)", "상태 간격 (ms)",
"Stereo", "스테레오",
"Stop", "정지",
"Stop time", "끝 시각",
"Storage path", "저장소 경로",
"Storage space", "저장소 공간",
"Stream ID", "스트림 ID",
"Stream Profile", "스트림 프로필",
"Stream index", "스트림 인덱스",
"Stream index (starts with zero).", "스트림 인덱스입니다 (0부터 시작).",
"Stream profile", "스트림 프로필",
"Stream setup error", "스트림 설정 오류",
"Stream type", "스트림 종류",
"Streaming", "스트리밍",
"Streaming Profile", "스트리밍 프로필",
"Streaming flags, allow/disallow HTTP streaming, advanced HTTP streaming (e.g, direct service or mux links), HTSP protocol streaming (e.g, Kodi (via pvr.hts) or Movian.", "스트리밍 플래그, HTTP 스트림 허용/비허용, 진보된 HTTP 스트리밍 (예, 직접 서비스 또는 먹스 링크), HTSP 프로토콜 스트리밍 (예, Kodi (pvr.hts 경유) 또는 Movian.",
"Streaming priority", "스트리밍 우선순위",
"Streaming profiles", "스트리밍 프로필",
"Subdirectory options", "하위 디렉토리 옵션",
"Subscribe to a service permanently", "서비스 영구 구독",
"Subscription", "구독",
"Subscription does not exist", "구독이 존재하지 않음",
"Subscription overridden", "구독이 덮어씌워짐",
"Subscription weight", "구독 순위",
"Substitute formatters", "자막 포맷",
"Subtitle", "부제",
"Subtitle Stream Filter", "자막 스트림 필터",
"Subtitle codec", "자막 코덱",
"Subtitle of the program (if any) (display only).", "프로그램 부제입니다 (있다면) (표시 전용).",
"Subtitle of the program (if any).", "프로그램 부제입니다 (있다면).",
"Sun", "일",
"Surround (2 front, rear mono)", "서라운드 (2 프론트, 리어 모노)",
"Switch before rotor", "로터 이전 스위치",
"Switch frequency offset", "주파수 오프셋 전환",
"Switch off the power to the LNB when idle. Note: this may cause interference with other devices when the LNB is powered back up.", "유휴시 LNB를 끕니다. 참고: 이는 LNB 전원이 복귀되었을 때 다른 장비를 방해하게 만들 수 있습니다.",
"Switch to another service", "다른 서비스로 전환",
"Switch type", "스위치 종류",
"Switch: %s", "스위치: %s",
"Symbol rate (Sym/s)", "심볼 속도 (Sym/초)",
"Sync", "싱크",
"Sync + Don't keep", "싱크 + 기억하지 않음",
"System", "시스템",
"TCP Protocol", "TCP 프로토콜",
"TV Adapters - Linux ATSC-C Frontend", "TV 어댑터 - 리눅스 ATSC-C 프론트엔드",
"TV Adapters - Linux ATSC-T Frontend", "TV 어댑터 - 리눅스 ATSC-T 프론트엔드",
"TV Adapters - Linux DAB Frontend", "TV 어댑터 - 리눅스 DAB 프론트엔드",
"TV Adapters - Linux DVB-C Frontend", "TV 어댑터 - 리눅스 DVB-C 프론트엔드",
"TV Adapters - Linux DVB-S Frontend (Master)", "TV 어댑터 - 리눅스 DVB-S 프론트엔드 (마스터)",
"TV Adapters - Linux DVB-S Slave Frontend", "TV 어댑터 - 리눅스 DVB-S 슬레이브 프론트엔드",
"TV Adapters - Linux DVB-T Frontend", "TV 어댑터 - 리눅스 DVB-T 프론트엔드",
"TV Adapters - Linux ISDB-C Frontend", "TV 어댑터 - 리눅스 ISDB-C 프론트엔드",
"TV Adapters - Linux ISDB-S Frontend (Master)", "TV 어댑터 - 리눅스 ISDB-S 프론트엔드 (마스터)",
"TV Adapters - Linux ISDB-T Frontend", "TV 어댑터 - 리눅스 ISDB-T 프론트엔드",
"TV Adapters - SatConfig - 4-Port", "TV 어댑터 - 위성설정 - 4포트",
"TV Adapters - SatConfig - Advanced", "TV 어댑터 - 위성설정 - 고급",
"TV Adapters - SatConfig - DiseqC Rotor", "TV 어댑터 - 위성설정 - 다이섹 로터",
"TV Adapters - SatConfig - DiseqC Switch", "TV 어댑터 - 위성설정 - 다이섹 스위치",
"TV Adapters - SatConfig - EN50494/UniCable I", "TV 어댑터 - 위성설정 - EN50494/UniCable I",
"TV Adapters - SatConfig - EN50607/UniCable II", "TV 어댑터 - 위성설정 - EN50607/UniCable II",
"TV Adapters - SatConfig - GOTOX Rotor", "TV 어댑터 - 위성설정 - GOTOX 로터",
"TV Adapters - SatConfig - Tone Burst/2 Port", "TV 어댑터 - 위성설정 - 톤 버스트/2포트",
"TV Adapters - SatConfig - USALS Rotor", "TV 어댑터 - 위성설정 - USALS 로터",
"TV Adapters - SatConfig - Universal LNB (Simple)", "TV 어댑터 - 위성설정 - 범용 LNB (단순)",
"Tag files with metadata", "메타데이터 태그 파일",
"Tags", "태그",
"Tags linked/to link to the channel.", "채널에 링크된/링크할 태그입니다.",
"Talk show", "토크쇼",
"Team sports (excluding football)", "팀 스포츠 (축구 제외)",
"Technology", "기술",
"Teletext Stream Filter", "텔레텍스트 스트림 필터",
"Teletext stream type is only available for this filter.", "텔레텍스트 스트림 타입은 이 필터에만 유효합니다.",
"Tennis", "테니스",
"Testing", "테스트 중",
"Testing options", "테스트 옵션",
"The ATSC source ID as set by the provider.", "ATSC 소스 ID는 제공자에 의해 설정되었습니다.",
"The CA provider to compare. Leave blank to apply to all providers.", "비교할 CA 제공자입니다. 모든 제공자에 적용하려면 비워두세요.",
"The CA stream type is only available for this filter.", "CA 스트림 종류만이 이 필터에 유효합니다.",
"The CAID to compare. Leave blank to apply to all IDs.", "비교할 CAID입니다. 모든 ID에 적용하려면 비워두세요.",
"The CAM slot status.", "CAM 슬롯 상태입니다.",
"The COFDM modulation used by the mux. If you're not sure of the value leave as AUTO.", "먹스가 사용하는 COFDM 모듈레이션입니다. 값을 확신하지 못한다면 AUTO로 두세요.",
"The Conditional Access ID used for the service.", "서비스에 대한 조건부 접근 ID입니다.",
"The Content reference identifier (CRID) authority.", "컨텐트 레퍼런스 식별 (CRID) 인증입니다.",
"The DVR profile to be used/used by the recording.", "녹화에 사용된/사용할 DVR 프로필입니다.",
"The DVR profile to be used/used by this rule.", "이 규칙에 의해 사용된/사용할 DVR 프로필입니다.",
"The EPG ID used by the entry.", "항목에 사용된 EPG ID입니다.",
"The EPG grabber name.", "EPG 그래버 이름입니다.",
"The EPG grabber to use on the mux. Enable (auto) is the recommended value.", "먹스에 사용하는 EPG 그래버입니다. 활성(자동)이 권장값입니다.",
"The EPG grabber type.", "EPG 그래버 종류입니다.",
"The MPEGTS stream type is only available for this filter.", "MPEGTS 스트림 종류만이 이 필터에 유효합니다.",
"The OpenTV channel number as set by the provider.", "제공자가 설정한 OpenTV 채널 번호입니다.",
"The PIN to use.", "사용할 PIN입니다.",
"The Physical Layer Scrambling (PLS) code used on the mux.", "먹스에 사용된 PLS(물리 계층 스크램블링) 코드입니다.",
"The Physical Layer Scrambling (PLS) mode used on the mux.", "먹스에 사용된 PLS(물리 계층 스크램블링) 모드입니다.",
"The Preferred Conditional Access Packet Identifier. Used for decrypting scrambled streams.", "선호하는 조건부 접근 패킷 식별자입니다. 암호화된 스트림을 복호화하는데 쓰입니다.",
"The SAT>IP server's name.", "SAT>IP 서버 이름입니다.",
"The SAT>IP server's universally unique identifier.", "SAT>IP 서버의 범용 고유 식별자입니다.",
"The SAT>IP source number.", "SAT>IP 소스 번호입니다.",
"The SAT>IP's discovered IP address.", "SAT>IP 발견된 IP 주소입니다.",
"The URL to the playlist.", "플레이리스트 URL입니다.",
"The audio stream types the filter should apply to.", "필터가 적용될 오디오 스트림 종류입니다.",
"The bandwidth the mux uses. If you're not sure of the value leave as AUTO but be aware that tuning may fail as some drivers do not like the AUTO setting.", "먹스 사용자의 대역폭입니다. 값을 확신하지 않는다면 AUTO로 두시고, 일부 장치에선 AUTO 설정으로 튜닝에 실패할 수 있습니다.",
"The bouquet the channel is associated with.", "채널이 연결된 채널 묶음입니다.",
"The cache scheme to use/used to store recordings. Leave as \"system\" unless you have a special use case for one of the others. See Help for details.", "녹화를 저장할 캐시 스키마입니다. 특별한 경우가 아니라면 \"system\"으로 두세요. 자세한 정보는 도움말을 보세요.",
"The channel name the entry will record from.", "녹화할 원본 항목 채널 이름입니다.",
"The channel on which this rule applies, i.e. the channel you're aiming to record. You can leave this field blank to apply the rule to all channels.", "이 규칙이 적용될 채널입니다. 참고: 녹화할 채널입니다. 모든 채널에 적용하고 싶으시면 이 필드를 비워두세요.",
"The channel this service is mapped to.", "이 서비스가 매핑될 채널입니다.",
"The character encoding for this network (e.g. UTF-8).", "이 네트워크의 문자 인코딩입니다 (예: UTF-8).",
"The character encoding for this service (e.g. UTF-8).", "이 서비스의 문자 인코딩입니다 (예: UTF-8).",
"The character set to use/used. You should not have to change this unless channel names and EPG data appear garbled.", "사용할/사용된 캐릭터셋입니다. 채널 이름이나 EPG 정보가 깨지지 않으면 변경하지 말아야 합니다.",
"The content type (Movie/Drama, Sports, etc.) to be used to filter matching events/programs.", "맞는 이벤트/프로그램을 필터하는데 쓸 내용 종류(영화/드라마, 스포츠 등)입니다.",
"The current boot ID.", "현재 부트 ID입니다.",
"The current configuration ID.", "현재 구성 ID입니다.",
"The current configuration version.", "현재 구성 버전입니다.",
"The default language to use if the user  language isn't set in the Access Entries tab.", "사용자 언어가 접근 항목 탭에서 지정되지 않은 경우 사용할 기본 언어입니다.",
"The default subscription weight for each subscription.", "각 구독에 대한 기본 우선순위입니다.",
"The default web interface to use if the user  theme isn't set in the Access Entries tab.", "사용자 테마가 접근 항목 탭에서 설정되지 않은 경우 사용할 기본 웹 인터페이스입니다.",
"The delivery system the mux uses. Make sure that your tuner supports the delivery system selected here.", "먹스가 사용할 전송 시스템입니다. 튜너가 여기서 선택한 전송 시스템을 지원하는지 미리 확인하세요.",
"The delivery system used by the mux.", "먹스가 사용하는 전송 시스템입니다.",
"The delivery system used by your cable provider.", "케이블 제공자가 사용하는 전송 시스템입니다.",
"The delivery system used by your provider.", "제공자가 사용하는 전송 시스템입니다.",
"The demux path used by the device.", "장치가 사용하는 디먹스 경로입니다.",
"The device ID.", "장치 ID입니다.",
"The device path in sysfs filesystem (/sys).", "sysfs 파일 시스템에서 장치 경로입니다 (/sys).",
"The device's serial number.", "장치의 시리얼 번호입니다.",
"The forward error correction high value. Most people will not need to change this setting.", "포워드 오류 수정 최대값입니다. 대부분의 사람들은 이 설정을 바꿀 필요가 없습니다.",
"The forward error correction low value. Most people will not need to change this setting.", "포워드 오류 수정 최저값입니다. 대부분의 사람들은 이 설정을 바꿀 필요가 없습니다.",
"The forward error correction used on the mux.", "먹스에서 사용되는 포워드 오류 수정입니다.",
"The forward error correction. Most people will not need to change this setting.", "포워드 오류 수정입니다. 대부분의 사람들은 이 설정을 바꿀 필요가 없습니다.",
"The frequency of the mux (in Hertz).", "먹스의 주파수입니다 (헤르쯔 단위).",
"The frequency of the mux/transponder in Hertz.", "먹스/트랜스폰더의 주파수를 헤르쯔 단위로 입력하세요.",
"The frontend number given to the device.", "장치에 제공된 프론트엔드 번호입니다.",
"The guard interval used by the mux. If you're not sure of the value leave as AUTO.", "먹스에서 사용된 가드 주기입니다. 값을 확신하지 않는다면 AUTO로 두세요.",
"The hierarchical modulation used by the mux. Most people will not need to change this setting.", "먹스에서 사용된 계층 모듈레이션입니다. 대부분의 사람들은 이 설정을 바꿀 필요가 없습니다.",
"The imagecache path to the icon to use/used for the channel.", "채널에 사용될/사용된 아이콘의 이미지 캐시 경로입니다.",
"The input path used by the device.", "장치에서 사용되는 입력 경로입니다.",
"The items have the following functions:", "이 항목은 다음 기능을 값고 있습니다:",
"The language to which the filter should apply.", "필터가 적용될 언어입니다.",
"The layer A constellation.", "레이어 A 배치입니다.",
"The layer A forward error correction.", "포워드 오류 수정 계층 A입니다.",
"The layer A segment count.", "계층 A 조각 수입니다.",
"The layer A time interleaving.", "계층 A 시간 인터리빙입니다.",
"The layer B constellation.", "레이어 B 배치입니다.",
"The layer B forward error correction.", "계층 B 포워드 오류 수정입니다.",
"The layer B segment count.", "계층 B 조각 수입니다.",
"The layer B time interleaving.", "계층 B 시간 인터리빙입니다.",
"The layer C constellation.", "레이어 C 배치입니다.",
"The layer C forward error correction.", "계층 C 포워드 오류 수정입니다.",
"The layer C segment count.", "계층 C 조각 수입니다.",
"The layer C time interleaving.", "계층 C 시간 인터리빙입니다.",
"The length of time (in seconds) to play the mux (1 hour = 3600).", "먹스를 재생할 시간 길이(초 단위)입니다 (1시간 = 3600).",
"The limit for the PID filter (driver or hardware).", "PID 필터 제한입니다 (드라이버 또는 하드웨어).",
"The locking mechanism selection for The Preferred Conditional Access Packet Identifier. See Help for more information.", "선호하는 조건부 접근 패킷 식별자를 선택하는 잠금 체계입니다. 자세한 정보는 도움말을 보세요.",
"The logical stream index to compare. Note that this index is computed using all filters.Example: If filter is set to AC3 audio type and the language to 'eng' and there are two AC3 'eng' streams in the service, the first stream could be identified using number 1 and the second using number 2.", "비교할 논리 스트림 색인입니다. 이 색인은 모든 필터를 사용하여 계산되는데 주의하세요. 예: 만일 AC3 오디오 종류에 언어가 'eng'고, 서비스에 두 개의 AC3 'eng' 스트림이 있다면, 첫 스트림은 번호 1, 두 번째 스트림은 번호 2가 됩니다.",
"The manufacturer of the SAT>IP server.", "SAT>IP 서버 제조사입니다.",
"The maximal duration of a matching event - in other words, only match programmes that are no longer than this duration.", "이벤트 적중 최대 길이입니다 - 바꿔 말하면, 이 길이보다 긴 프로그램은 해당하지 않습니다.",
"The maximum RAM (system memory) size for timeshift buffers. When free RAM buffers are available they are used for timeshift data in preference to using storage.", "타임시프트 버퍼의 최대 RAM (시스템 메모리) 크기입니다. 여유 RAM 버퍼가 있으면, 저장소 설정을 써서 타임시프트 데이터가 사용됩니다.",
"The maximum amount of time a grabber is allowed scan a mux for data (in seconds).", "그래버에 먹스 데이터 스캔시 허용할 최대 시간입니다. (초 단위).",
"The maximum combined size of all timeshift buffers. If you specify an unlimited period it's highly recommended you specify a value here.", "모든 타임시프트 버퍼의 통합 최대 크기입니다. 무제한 길이를 지정하였다면 여기 특정 값을 지정하길 강력하게 권장합니다.",
"The maximum number of entries that can be matched.", "매치할 수 있는 최대 항목 개수입니다.",
"The maximum number of input streams allowed on this network.", "이 네트워크에 허용된 입력 스트림의 최대 개수입니다.",
"The maximum number of recording entries this rule can create.", "규칙이 만들어질 수 있는 최대 녹화 항목 개수입니다.",
"The maximum number of recordings that can be scheduled.", "예약될 수 있는 최대 녹화 개수입니다.",
"The maximum number of services to decrypt per mux.", "먹스를 복호화할 최대 서비스 개수입니다.",
"The maximum number of times this rule can be triggered.", "발동할 수 있는 최대 규칙 개수입니다.",
"The maximum time period that will be buffered for any given (client) subscription.", "주어진 클라이언트 구독에 주어진 버퍼 최대 (클라이언트) 시간입니다.",
"The minimal duration of a matching event - in other words, only match programs that are no shorter than this duration.", "이벤트 적중 최소 길이입니다 - 바꿔말하면, 이 길이보다 짧은 프로그램은 해당하지 않습니다.",
"The minimum delay after the rotor movement command is sent.", "로터 이동 명령이 보내진 뒤 최소 지연 시간입니다.",
"The minimum delay before tuning in milliseconds after tuner stop. If the time between the previous and next start is greater than this value then the delay is not applied.", "튜너가 멈춘 뒤 튜닝까지 최소 지연 시간(밀리세컨드 단위)입니다. 이전과 다음 시작이 이 값보다 크면, 지연 시간은 적용되지 않습니다.",
"The modulation used on the mux.", "먹스에 사용되는 모듈레이션입니다.",
"The mux the service is on.", "서비스가 켜진 먹스입니다.",
"The mux to play when the entry is triggered.", "항목이 발동되면 재생할 먹스입니다.",
"The mux's universally unique identifier.", "먹스의 범용 고유 식별자입니다.",
"The name (or frequency) the mux is on.", "켜진 먹스 이름 (또는 주파수)입니다.",
"The name given to/of the channel (This is how it'll appear in your EPG.)", "채널에 주어진 이름입니다. (EPG에 표시되는 방법입니다.)",
"The name of the profile.", "프로필 이름입니다.",
"The name of the the rule.", "규칙 이름입니다.",
"The network prefix(es) to block, e.g.192.168.2.0/24 (comma-separated list).", "블록에 대한 네트워크 프리픽스(들)입니다. 예: 192.168.2.0/24 (반점으로 구분된 목록).",
"The network the mux is on.", "네트워크 먹스가 켜졌습니다.",
"The network the service is on.", "네트워크 서비스가 켜졌습니다.",
"The network's priority. The network with the highest priority value will be used out of preference if available. See Help for details.", "네트워크 우선순위입니다. 유효하다면 높은 우선순위 값의 네트워크가 사용될 것입니다. 자세한 정보는 도움말을 보세요.",
"The network's service ID", "네트워크의 서비스 ID",
"The networks assigned to the device.", "네트워크가 이 장치에 할당되었습니다.",
"The networks using this configuration.", "네트워크가 이 구성을 사용합니다.",
"The networks' universally unique identifier (UUID).", "네트워크의 범용 고유 식별자(UUID)입니다.",
"The number of ATSC-C (Cable/AnnexB) tuners to export.", "내보낼 ATSC-C (케이블/AnnexB) 튜너 수입니다.",
"The number of ATSC-T (Terresterial) tuners to export.", "내보낼 ATSC-T (Terresterial) 튜너 수입니다.",
"The number of DVB-C (Cable) tuners to export.", "내보낼 DVB-C (케이블) 튜너 수입니다.",
"The number of DVB-C2 (Cable) tuners to export.", "내보낼 DVB-C2 (케이블) 튜너 수입니다.",
"The number of DVB-S (Satellite) tuners to export.", "내보낼 DVB-S (위성) 튜너 수입니다.",
"The number of DVB-S2 (Satellite) tuners to export.", "내보낼 DVB-S2 (위성) 튜너 수입니다.",
"The number of DVB-T (Terresterial) tuners to export.", "내보낼 DVB-T (Terresterial) 튜너 수입니다.",
"The number of DVB-T2 (Terresterial) tuners to export.", "내보낼 DVB-T2 (Terresterial) 튜너 수입니다.",
"The number of MPEG-TS packets Tvheadend buffers in case there is a delay receiving CA keys. ", "CA 키 수신 지연이 있을 때 Tvheadend 버퍼 MPEG-TS 패킷 수입니다.",
"The number of allowed connections this user can make to the server.", "서버에 허용된 연결 수입니다.",
"The number of bytes to buffer. By default, linuxdvb's input buffer is 18800 bytes long. The accepted range is 18800-1880000 bytes.", "버퍼의 바이트 수입니다. 기본적으로, linuxdvb의 입력 버퍼는 18800 바이트 길이입니다. 허용된 범위는 18800-1880000 바이트입니다.",
"The number of days cookies set by Tvheadend should expire.", "Tvheadend가 만료시킬 쿠키의 일 수입니다.",
"The number of muxes left to scan on this network.", "이 네트워크에서 남은 스캔할 먹스 수입니다.",
"The number of seconds to wait for a stream to start.", "스트림을 시작하기위해 기다리는 초의 수입니다.",
"The number of services currently mapped to channels.", "현재 채널에 매핑된 서비스 수입니다.",
"The orbital position of the satellite the mux is on.", "위성 궤도 위치 먹스가 켜졌습니다.",
"The orbital position of the satellite your dish is pointing at.", "안테나 접시가 가리키는 위성의 궤도 위치입니다.",
"The outcome of the last scan performed.", "실시된 마지막 스캔 출력입니다.",
"The physical layer pipe ID. Most people will not need to change this setting.", "실제 계층 파이프 ID입니다. 대부분의 사람들은 이 설정을 바꿀 필요가 없습니다.",
"The playlist's character set.", "플레이리스트의 캐릭터셋을 설정합니다.",
"The polarization used on the mux.", "먹스에 사용된 극성입니다.",
"The position the channel will appear on your EPG. This is not used by Tvheadend internally, but rather intended to be used by HTSP clients for mapping to remote control buttons, presentation order, etc.", "EPG에서 채널을 나타낼 위치입니다. 이는 Tvheadend에서 내부적으로 사용하지 않지만, HTSP 클라이언트가 원격 제어 버튼, 표시 순서 등에 활용할 용도입니다.",
"The priority of any recordings set because of this rule will take precedence and cancel lower-priority events.", "이 규칙에 의해 정해진 녹화 우선순위는 우선하게 되며, 낮은 순위의 이벤트를 취소할 수 있습니다.",
"The provider's ID.", "제공자 ID입니다.",
"The provider's name.", "제공자의 이름입니다.",
"The provider's network ID.", "제공자의 네트워크 ID입니다.",
"The provider's network name.", "제공자의 네트워크 이름입니다.",
"The quadrature amplitude modulation (QAM) used by the mux. If you're not sure of the value leave as AUTO.", "먹스가 사용하는 직교 진폭 변조(QAM)입니다. 값을 확신하지 못한다면 AUTO로 두세요.",
"The recorded file was removed intentionally", "녹화 파일은 동작에 의해 제거됨",
"The recording/entry status.", "녹화/항목 상태입니다.",
"The rolloff used on the mux.", "먹스에 사용된 롤 오프입니다.",
"The rule action defines the operation when all comparisons succeed. See Help for more information on what the various rules do.", "규칙 액션은 모든 비교가 성공했을 때 동작을 지정합니다. 어떤 규칙이 가능한지 자세한 정보는 도움말을 보세요.",
"The satellite configuration to use.", "사용할 위성 구성입니다.",
"The scan state. New muxes will automatically be changed to the PEND state. You can change this to ACTIVE to queue a scan of this mux.", "스캔 상태입니다. 새로운 먹스는 자동으로 PEND 상태로 바뀝니다. 먹스를 스캔에 추가하려면 ACTIVE로 바꿉니다.",
"The scheduled start time, including any padding.", "여유 시간을 포함한 예약된 시작 시각입니다.",
"The scheduled stop time, including any padding.", "여유 시간을 포함한 예약된 종료 시각입니다.",
"The scrambled bits in MPEG-TS packets are always cleared. It is a workaround for the special streams which are descrambled, but these bits are not touched.", "MPEG-TS 패킷에서 암호화된 비트는 항상 소거됩니다. 복호화된 특정 스트림에서 일부 건드리지 않은 비트가 있는 경우 대처 방법입니다.",
"The selected video type should be preferred when multiple services are available for a channel.", "선택한 비디오 종류는 여러 서비스가 채널에 유효한 경우 선호됩니다.",
"The service ID", "서비스 ID",
"The service ID as set by the provider.", "서비스 ID는 제공자가 설정합니다.",
"The service ID.", "서비스 ID입니다.",
"The service name as set by the provider.", "서비스 이름은 제공자가 설정합니다.",
"The service the filter should apply to. Leave blank to apply the filter to all services.", "필터가 적용될 서비스입니다. 모든 서비스에 필터를 적용하려면 빈칸으로 두세요.",
"The service to which the filter should apply. Leave blank to apply the filter to all services.", "필터가 적용될 대상 서비스입니다. 모든 서비스에 적용하려면 비워두세요.",
"The service type flag as defined by the DVB specifications (e.g. 0x02 = radio, 0x11 = MPEG2 HD TV, 0x19 = H.264 HD TV)", "DVB 스펙에 의해 서비스 종류 플래그가 설정됩니다 (예: 0x02 = radio, 0x11 = MPEG2 HD TV, 0x19 = H.264 HD TV)",
"The service's channel minor as set by the provider.", "제공자에 의해 서비스 채널 마이너가 설정됩니다.",
"The service's channel number as set by the provider.", "제공자에 의해 서비스 채널 번호가 설정됩니다.",
"The service's encryption status.", "서비스의 암호화 상태입니다.",
"The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, the polarization and low-high band settings must be equal. If you set another tuner as master, then this tuner will act as a slave and Tvheadend will assure that this tuner will not use incompatible parameters (position, polarization, lo-hi).", "기본 범용 LNB에서 들어오는 신호는 간단한 동축 케이블 분배기(멀티스위치 제외)를 통해 여러 출력으로 나뉠 수 있습니다. 이 경우, 위치, 극성과 하위-상위 밴드 설정은 반드시 똑같아야 합니다. 다른 튜너를 마스터로 지정할 경우, 이 튜너는 슬레이브처럼 동작하며, tvheadend는 이 튜너에 호환되지 않는 매개변수(위치, 극성, 하위-상위)를 사용하지 않게 됩니다.",
"The signal from the standard universal LNB can be split using a simple coaxial splitter (no multiswitch) to several outputs. In this case, the position, the polarization and low-high band settings must be equal. If you set another tuner as master, then this tuner will act as a slave and tvheadend will assure that this tuner will not use incompatible parameters (position, polarization, lo-hi).", "기본 범용 LNB에서 들어오는 신호는 간단한 동축 케이블 분배기(멀티스위치 제외)를 통해 여러 출력으로 나뉠 수 있습니다. 이 경우, 위치, 극성과 하위-상위 밴드 설정은 반드시 똑같아야 합니다. 다른 튜너를 마스터로 지정할 경우, 이 튜너는 슬레이브처럼 동작하며, tvheadend는 이 튜너에 호환되지 않는 매개변수(위치, 극성, 하위-상위)를 사용하지 않게 됩니다.",
"The start time of the recording.", "녹화 시작 시각입니다.",
"The stream ID used for the mux.", "먹스에 사용된 스트림 ID입니다.",
"The stream profile the DVR profile will use for recordings.", "녹화에 사용될 스트림 프로필, DVR 프로필입니다.",
"The streaming profile to use/used. If not set, the default will be used.", "사용할/사용된 스트림 프로필입니다. 지정되지 않으면, 기본값이 쓰입니다.",
"The string allows you to manually specify the full path generation using predefined modifiers. See Help for full details.", "미리 지정된 수정자를 사용하여 전체 경로 생성이 수동으로 지정될 수 있습니다. 전체 상세 정보는 도움말을 보세요.",
"The subtitle stream types the filter should apply to.", "필터가 적용될 자막 스트림 종류입니다.",
"The symbol rate used on the mux.", "먹스에 사용된 심볼 레이트, 데이터율입니다.",
"The symbol rate used on the mux/transponder.", "먹스/트렌스폰더에 사용된 심볼 레이트, 데이터율입니다.",
"The symbol rate.", "심볼 레이트, 데이터율입니다.",
"The time the entry stops/stopped being recorded.", "녹화가 정지된/정지될 시각입니다.",
"The time window to cut the stop time from the overlapped event in seconds.", "겹친 이벤트에서 타임 윈도우가 정지 시간을 끊도록 하는 초 시간입니다.",
"The title of the program to look for. Note that this accepts case-insensitive regular expressions.", "찾는 프로그램 제목입니다. 대소문자 구분 정규식을 받는데 유의하세요.",
"The total number of services found.", "검색된 서비스의 개수입니다.",
"The total scheduled duration.", "예약된 총 길이입니다.",
"The transmission/OFDM mode used by the mux. If you're not sure of the value leave as AUTO but be aware that tuning may fail as some drivers do not like the AUTO setting.", "먹스에서 사용된 전송/OFDM 모드입니다. 값을 확신하지 못한다면 AUTO로 두시고, 일부 장치에선 AUTO 설정으로 튜닝에 실패할 수 있습니다.",
"The transponder ID.", "트렌스폰더 ID입니다.",
"The transport stream ID of the mux within the network.", "네트워크 내부 먹스의 트랜스포트 스트림 ID입니다.",
"The tuner priority value (a higher value means to use this tuner out of preference). See Help for details.", "튜너 우선순위 값입니다 (높은 값은 높은 우선 순위를 나타냅니다). 자세한 정보는 도움말을 보세요.",
"The tuner priority value for streamed channels through HTTP or HTSP (a higher value means to use this tuner out of preference). If not set (zero), the standard priority value is used. See Help for details.", "HTTP와 HTSP를 통한 스트림된 채널의 튜너 우선 순위입니다 (높은 값은 이 튜너가 선호됨을 나타냄). 설정되지 않으면 (0), 기본 우선 순위가 사용됩니다. 자세한 정보는 도움말을 보세요.",
"The user who created the recording, or the auto-recording source and IP address if scheduled by a matching rule.", "예약된 규칙과 일치하는 녹화 또는 자동 녹화 원본, IP 주소를 만든 사용자입니다.",
"The version of Tvheadend that last updated the config.", "마지막으로 설정을 업데이트한 Tvheadend 버전입니다.",
"The video stream types the filter should apply to.", "필터가 적용될 비디오 스트림 종류입니다.",
"Theme", "테마",
"This tuner", "이 튜너",
"This will create an NTP driver (using shmem interface) that you can feed into ntpd. This can be run without root privileges, but generally the performance is not that great.", "NTP 드라이버를 (shmem 인터페이스를 사용) 만들어서 ntpd로 피드할 수 있도록 합니다. 루트 권한 없이 실행할 수 있지만, 일반적으로 성능이 그리 좋지 않습니다.",
"Thread", "스레드",
"Thriller", "스릴러",
"Thu", "목",
"Time", "시각",
"Time (in milliseconds) for a command to complete.", "명령을 완료할 시간(밀리세컨드 단위)입니다.",
"Time (in milliseconds) for the rotor to power up.", "로터를 켤 시간(밀리세컨드 단위)입니다.",
"Time (in minutes) to re-fetch the playlist.", "플레이리스트를 다시 가져올 시간(분)입니다.",
"Time Stamp Fix", "타임스탬프 조절",
"Time missed", "지각",
"Time record caption", "캡션 녹화 시각",
"Time recording%s%s", "녹화 시각%s%s",
"Time to start the recording/time the recording started.", "녹화 시작 시각/녹화를 시작할 시각입니다.",
"Time to stop recording/time the recording stopped.", "녹화 끝 시각/녹화를 멈출 시각입니다.",
"Time update", "시계 업데이트",
"Timeout (sec) (0=infinite)", "시간 제한 (초) (0=무제한)",
"Timeout (seconds)", "시간 제한 (초)",
"Timeout (secs)", "시간 제한 (초)",
"Timer-based automatic record caption.", "타이머 기반 자동 캡션 녹화입니다.",
"Timer-based automatic recording.", "타이머 기반 자동 녹화입니다.",
"Timeshift", "타임시프트",
"Title", "제목",
"Title (regexp)", "제목 (정규식)",
"Title of the program (display only).", "프로그램 제목입니다 (표시전용).",
"Title of the program.", "프로그램 제목입니다.",
"Title of the recording - this is used to generate the filename.", "녹화 제목 - 파일 이름을 생성하기 위해 사용됩니다.",
"Toggle details", "상세 정보 토글",
"Tone burst", "톤 버스트",
"Too many data errors", "과다한 데이터 에러",
"Total number of mapped channels on this network.", "이 네트워크에서 매핑된 채널의 개수입니다.",
"Total number of muxes found on this network.", "이 네트워크에서 발견된 먹스의 총 개수입니다.",
"Total number of services found on this network.", "이 네트워크에서 검색된 서비스 총 개수입니다.",
"Total number of services found.", "검색된 서비스의 개수입니다.",
"Total number of services seen.", "발견된 서비스의 총 개수입니다.",
"Total number of services.", "서비스의 총 개수입니다.",
"Tourism / Travel", "관광 / 여행",
"Trace subsystems", "서브시스템 트레이스",
"Traditional arts", "전통 예술",
"Traditional music", "전통 음악",
"Transcode", "트랜스코드",
"Transcode video only if source video codec matches. \"Any\" ignores source video codec checking and always transcodes. If no codec match is found, transcoding is done using the \"copy\" codec. if a match is found, transcode with the parameters in this profile. Separate codec names with comma.", "원본 비디오 코덱이 일치할 때만 비디오를 트랜스코드합니다. \"모두\"는 원본 비디오 코덱 확인을 무시하고, 항상 트랜스코드합니다. 일치하는 코덱이 발견되지 않으면, 트랜스코딩은 \"복사\" 코덱을 써서 이뤄집니다. 일치하는 코덱이 발견되면, 이 프로필에서의 매개변수로 트랜스코드됩니다. 코덱 이름은 반점으로 구분합니다.",
"Transcode/av-lib", "트랜스코드/av-lib",
"Transcoding", "트랜스코딩",
"Transmission mode", "전송 모드",
"Transponder ID", "전송자 ID",
"Transport Stream", "트랜스포트 스트림",
"Transport stream ID", "전송 스트림 ID",
"Try to obtain channel numbers from the display-name xml tag. If the first word is number, it is used as the channel number.", "채널 번호를 xml 태그의 표시 이름에서 가져오길 시도합니다. 첫 단어가 숫자이면, 채널 번호로 사용됩니다.",
"Tue", "화",
"Tune before DiseqC", "다이섹 이전 튠",
"Tuner", "튜너",
"Tuner bind IP address", "튜너 할당 IP 주소",
"Tuner configuration", "튜너 설정",
"Tuner configuration.", "튜너 설정입니다.",
"Tuning failed", "튜닝 실패",
"Turn off LNB when idle", "유휴시 LNB 끄기",
"Tvheadend server name", "Tvheadend 서버 이름",
"Type", "종류",
"Type override", "종류 오버라이드",
"UDP RTP port number (2 ports)", "UDP RTP 포트 번호 (2포트)",
"UHD: ultra high definition", "UHD: 초고선명",
"UPnP Protocol", "UPnP 프로토콜",
"URL of the M3U playlist.", "M3U 플레이리스트 URL입니다.",
"URL with the SAT>IP server XML location", "SAT>IP 서버 XML 지역화 URL",
"URL.", "URL입니다.",
"Unable to get system UTC time", "시스템 UTC 시계 획득 실패",
"Unable to get system local time", "시스템 로컬 시계 획득 실패",
"Unable to get system time", "시스템 시계 획득 실패",
"Unable to open file", "파일을 열 수 없음",
"Unable to stat path", "상태 경로가 유효하지 않음",
"Uncommitted", "저장되지 않음",
"Uncommitted first", "저장되지 않은 목록",
"Unicable (EN50494)", "유니케이블 (EN50494)",
"Unicable I (EN50494)", "유니케이블 I (EN50494)",
"Unicable I switch (universal LNB)", "유니케이블 I 스위치 (범용 LNB)",
"Unicable II (EN50607)", "유니케이블 II (EN50607)",
"Unicable II switch (universal LNB)", "유니케이블 II 스위치 (범용 LNB)",
"Unicable type", "유니케이블 종류",
"Unimportant", "덜 중요함",
"Universal LNB only", "범용 LNB 전용",
"Universally unique identifier. Read only.", "범용 고유 식별자입니다. 읽기 전용입니다.",
"Unknown", "알 수 없음",
"Unknown file", "알 수 없는 파일",
"Unknown reason (%i)", "알 수 없는 (%i)",
"Unlimited size", "무제한 크기",
"Unlimited time", "무제한 시간",
"Unset (default)", "설정 해제 (기본값)",
"Update Entitlement Management Messages from one mux only.", "EMM을 한 개 먹스에 대해서만 업데이트합니다.",
"Update card (EMM)", "카드 (EMM) 업데이트",
"Update channel icon", "채널 아이콘 업데이트",
"Update channel name", "채널 이름 업데이트",
"Update channel number", "채널 번호 업데이트",
"Update disabled", "업데이트 ",
"Update time", "업데이트 시간",
"Update tolerance (ms)", "업데이트 대기 (ms)",
"Updated", "업데이트됨",
"Updates from one mux (EMM)", "한 개 먹스에서만 업데이트 (EMM)",
"Upon start, Tvheadend doesn't know the last rotor position. This value defines the initial rotor movement. TVHeadend waits the specified time when the first movement is requested.", "시작시 Tvheadend는 마지막 로터 위치를 알지 못합니다. 이 값은 최초 로터 이동을 지정합니다. TVHeadend는 첫 이동이 요청되면 지정한 시간만큼 기다립니다.",
"Usage: %s [OPTIONS]\n", "사용법: %s [옵션]\n",
"Use AC-3 detection.", "AC-3 탐지를 사용합니다.",
"Use EITp/f to decide event start/stop. This is also known as \"Accurate Recording\". See Help for details.", "EITp/f로 이벤트 시작과 종료를 판별합니다. 이는 \"정확한 녹화\"로도 알려져 있습니다. 자세한 정보는 도움말을 보세요.",
"Use EPG running state", "EPG 방송 상태 사용",
"Use HTTP digest authentication", "HTTP 다이제스트 인증 사용",
"Use PROXY protocol & X-Forwarded-For", "프록시 프로토콜 & X-Forwarded-For 사용",
"Use WEBM format.", "WEBM 포맷을 사용합니다.",
"Use Windows-compatible filenames", "윈도우 호환 파일 이름 사용",
"Use XSPF playlist instead of M3U", "M3U 대신 XSPF 플레이리스트 사용",
"Use a pre-defined list of ATSC-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "ATSC-C 먹스에 미리 지정된 목록을 사용합니다. 참고: 이 목록은 만료될 수 있으며, 평소보다 스캔이 오래 걸릴 수 있습니다.",
"Use a pre-defined list of ATSC-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "ATSC-T 먹스에 미리 지정된 목록을 사용합니다. 참고: 이 목록은 만료될 수 있으며, 평소보다 스캔이 오래 걸릴 수 있습니다.",
"Use a pre-defined list of DAB muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "DAB 먹스에 미리 지정된 목록을 사용합니다. 참고: 이 목록은 만료될 수 있으며, 평소보다 스캔이 오래 걸릴 수 있습니다.",
"Use a pre-defined list of DVB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "DVB-C 먹스에 미리 지정된 목록을 사용합니다. 참고: 이 목록은 만료될 수 있으며, 평소보다 스캔이 오래 걸릴 수 있습니다.",
"Use a pre-defined list of DVB-S/S2 muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "DVB-S/S2 먹스에 미리 지정된 목록을 사용합니다. 참고: 이 목록은 만료될 수 있으며, 평소보다 스캔이 오래 걸릴 수 있습니다.",
"Use a pre-defined list of DVB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "DVB-T 먹스에 미리 지정된 목록을 사용합니다. 참고: 이 목록은 만료될 수 있으며, 평소보다 스캔이 오래 걸릴 수 있습니다.",
"Use a pre-defined list of ISDB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "ISDB-C 먹스에 미리 지정된 목록을 사용합니다. 참고: 이 목록은 만료될 수 있으며, 평소보다 스캔이 오래 걸릴 수 있습니다.",
"Use a pre-defined list of ISDB-S muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "ISDB-S 먹스에 미리 지정된 목록을 사용합니다. 참고: 이 목록은 만료될 수 있으며, 평소보다 스캔이 오래 걸릴 수 있습니다.",
"Use a pre-defined list of ISDB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "ISDB-T 먹스에 미리 지정된 목록을 사용합니다. 참고: 이 목록은 만료될 수 있으며, 평소보다 스캔이 오래 걸릴 수 있습니다.",
"Use only this service ID, filter out others.", "다른 것은 필터링하고, 이 서비스 ID에만 사용합니다.",
"Use original", "원본 사용",
"Use packet backlog", "패킷 백 로그 사용",
"Use service IDs as channel numbers", "채널 번호로 서비스 ID 사용",
"Use tag internally (don't expose to clients).", "태그를 내부적으로 사용(클라이언트에 숨김)합니다.",
"Use the provider's service IDs as channel numbers.", "채널 번호로 제공자 서비스 ID를 사용합니다.",
"Use this character set when setting filenames.", "파일 이름을 설정할 때 이 캐릭터셋을 사용합니다.",
"Use x amount of storage space.", "저장소의 x 사용합니다.",
"User access error", "사용자 접근 오류",
"User does not have access", "접근할 수 없는 사용자",
"User does not have access to channel", "채널에 접근할 수 없는 사용자",
"User icon", "사용자 아이콘",
"User interface level", "사용자 인터페이스 레벨",
"User interface quick tips (tooltips)", "사용자 인터페이스 빠른 팁 (툴팁)",
"User language", "사용자 언어",
"User limit reached", "사용자 제한 도달",
"User login", "사용자 로그인",
"User request", "사용자 요청",
"Username", "사용자 이름",
"Username for the entry (login username).", "항목에 대한 사용자 이름(로그인 사용자 이름)입니다.",
"Username of the entry (this should match a username from within the \"Access Entries\" tab).", "항목의 사용자 이름(\"접근 항목\" 탭에서 사용자 이름과 일치해야 함)입니다.",
"Users - Access Entries", "사용자 - 접근 항목",
"Users - IP Blocking", "사용자 - IP 차단",
"Users - Passwords", "사용자 - 암호",
"Variety show", "버라이어티 쇼",
"Verify the SSL certificate.", "SSL 인증서를 검증합니다.",
"Verify the peer's SSL.", "피어의 SSL을 검증합니다.",
"Vertical resolution (height) of the output video stream. Horizontal resolution is adjusted automatically to preserve aspect ratio. When set to 0, the input resolution is used.", "출력 비디오 스트림의 세로 해상도(높이)입니다. 가로 해상도는 비율에 맞게 자동으로 맞춰집니다. 0으로 지정하면, 입력 해상도가 쓰입니다.",
"Video", "비디오",
"Video Stream Filter", "비디오 스트림 필터",
"Video bitrate (kb/s) (0=auto)", "비디오 비트레이트 (kb/s) (0=자동)",
"Video codec", "비디오 코덱",
"Video codec preset", "비디오 코덱 프리셋",
"Video codec preset to use for transcoding.", "트랜스코딩에 쓰일 비디오 코덱 프리셋입니다.",
"Video codec to use for the transcode. \"Do not use\" will disable video output.", "트랜스코드에 쓰일 비디오 코덱입니다. \"사용하지 않음\"은 비디오 출력을 배제합니다.",
"Video recorder", "비디오 녹화기",
"Video recorder flags, allow/disallow access to video recorder functionality (including Autorecs), allow/disallow users to view other DVR entries, allow/disallow users to work with DVR entries of other users (remove, edit) etc.", "비디오 녹화 플래그, 비디오 녹화 기능 접근 허용/비허용 (자동 녹화 포함), 다른 DVR 항목 보기 허용/비허용, 다른 사용자의 DVR 항목 제어(삭제, 편집) 허용/비허용 등.",
"View all", "모두 보기",
"Visually impaired commentary/audio description", "시각 장애 해설/오디오 설명",
"Vorbis audio", "Vorbis 오디오",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV 프로필 H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV 프로필 H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV 프로필 VP8/Vorbis/WEBM",
"Waiting for EPG running flag", "EPG 활성 플래그 대기중",
"Waiting for program start", "프로그램 시작 대기중",
"Waiting for stream", "스트림 대기중",
"Wake up the linked input whenever this adapter is used. The subscriptions are named as \"keep\". Note that this isn't normally needed, and is here simply as a workaround to driver bugs in certain dual tuner cards that otherwise lock the second tuner.", "어댑터가 사용될 때마다 링크된 입력을 깨웁니다. 구독은 \"보존\"으로 이름지어집니다. 참고로 이는 일반적으로 불필요하며, 특정 듀얼 튜너에서 세컨드 튜너를 잠그는 드라이버 버그에 대한 해결책으로 쓰입니다.",
"War", "전투",
"Water sport", "워터 스포츠",
"Weak stream", "약한 스트림",
"Weather report", "날씨 정보",
"Web User Interface", "웹 유저 인터페이스",
"Web interface", "웹 인터페이스",
"Web interface language", "웹 인터페이스 언어",
"Web interface language.", "웹 인터페이스 언어입니다.",
"Web interface theme.", "웹 인터페이스 테마입니다.",
"Web theme", "웹 테마",
"Web user interface", "웹 유저 인터페이스",
"Wed", "수",
"Welcome - Tvheadend - your TV streaming server and video recorder", "환영합니다 - Tvheadend - 당신만의 TV 스트리밍 서버와 비디오 녹화기",
"Western", "서양",
"Western hemisphere (latitude direction)", "서반구 (위도 방향)",
"Western hemisphere (latitude direction).", "서반구(위도 방향)입니다.",
"When nothing else is happening Tvheadend will continuously rotate among all muxes and tune to them to verify that they are still working when the inputs are not used for streaming. If your adapters have problems with lots of (endless) tuning, disable this. Note that this option should be OFF for the normal operation. This type of mux probing is not required and it may cause issues for SAT>IP (limited number of PID filters).", "유휴 상태일 때 Tvheadend는 지속적으로 모든 먹스를 전환하면서 튠을 하여 입력이 스트리밍에 쓰이지 않을 때도 계속 작동하도록 합니다. 어댑터가 수많은 (무한) 튜닝 문제를 보인다면, 이 기능을 끄세요. 참고로 이 옵션은 일반적인 경우 꺼져 있어야 합니다. 이 종류의 먹스 프로빙은 불필요하며, SAT>IP(제한된 수의 PID 필터)에서 문제를 일으킬 수 있습니다.",
"When specified, this setting overrides the subdirectory rules (except the base directory) defined in the DVR configuration and puts all recordings done by this entry into the subdirectory named here. See Help for more info.", "지정되면, 설정은 DVR 설정에 지정된 서브 디렉토리 규칙 (기본 디렉토리 제외)보다 우선 설정되며, 여기서 이뤄진 모든 녹화는 여기 지정된 서브 디렉토리로 저장됩니다. 도움말을 참고하여 자세한 정보를 알아보세요.",
"When streaming a service (via http or htsp) Tvheadend will use the network with the highest streaming priority set here. See Help for details.", "(http나 htsp를 통해) 서비스를 스트림하는 경우 Tvheadend는 여기서 설정된 가장 높은 스트리밍 우선순위의 네트워크를 사용할 것입니다. 자세한 정보는 도움말을 보세요.",
"When the fulltext is checked, the title pattern is matched against title, subtitle, summary and description.", "전체 텍스트가 체크되면, 패턴은 제목, 부제, 요약과 설명에서 일치 확인이 이루어집니다.",
"When the service was first identified and recorded.", "서비스가 처음 인식되고 녹화된 때입니다.",
"When the service was last seen during a mux scan.", "먹스 스캔시 서비스가 마지막으로 발견된 때입니다.",
"Winter sports", "겨울 스포츠",
"With \"RAM only\" enabled, and when \"Maximum RAM size\" is reached, remove the oldest segment in the buffer instead of replacing it completely. Note, this may reduce the amount of rewind time.", "\"RAM 전용\"이 켜지면, \"최대 RAM 크기\"에 도달했을 때, 버퍼의 오래된 조각부터 제거됩니다. 되감기 시간을 줄일 수 있습니다.",
"Wizard", "마법사",
"Wrapper (capmt_ca.so)", "래퍼 (capmt_ca.so)",
"Write a short message to log identifying the matched parameters. It is useful for debugging your setup or structure of incoming streams.", "일치하는 매개변수를 로그에서 확인할 수 있는 짧은 메시지를 씁니다. 이는 입력 스트림의 설정 또는 구조 디버깅시 유용합니다.",
"Writes the current in-memory EPG database to disk every x hours, so should a crash/unexpected shutdown occur EPG data is saved periodically to the database (re-read on next startup). Set to 0 to disable.", "현재 메모리에 있는 EPG 데이터베이스를 x 시간마다 디스크에 저장하여, 오류 또는 예기치 못한 종료가 발생했을 때, 주기적으로 저장된 EPG 데이터를 살려냅니다 (다음 번 실행 때 다시 읽음). 0을 지정하면 비활성화됩니다.",
"XMLTV EPG Import", "XMLTV EPG 가져오기",
"Yes", "예",
"addpids/delpids supported", "addpids/delpids 지원",
"hrs", "시간",
"invalid option specified [%s]", "잘못된 옵션이 지정됨 [%s]",
"logout", "로그아웃",
"min", "분",
"mins", "분",
"option %s requires a value", "옵션 %s 에 값이 필요함",
"return", "리턴",
"tvhdhomerun client", "tvhdhomerun 클라이언트",
NULL, NULL
};

static const char *tvh_locale_21[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nLai iegūtu vairāk informācijas, lūdzu, apmeklējiet tīmekļa vietni Tvheadend:\nhttps://tvheadend.org\n",
"ACTIVE", "AKTĪVS",
"ANY", "JEBKURŠ",
"Aborted by user", "Lietotāja pārtraukts",
"Access", "Piekļuve",
"Action", "Darbība",
"Any", "Jebkurš",
"Build", "Būvējums",
"Channel", "Kanāls",
"Channels", "Kanāli",
"Created", "Izveidots",
"Creator", "Radītājs",
"Default", "Noklusējums",
"Disable", "Atspējot",
"Enable (auto)", "Iespējot (auto)",
"Enabled", "Iespējots",
"Encrypted", "Šifrēts",
"Errors", "Kļūdas",
"FAIL", "NEIZDEVIES",
"Filename", "Datnes nosaukums",
"Fri", "Pk",
"Immediately abort", "Nekavējoties pārtraukt",
"Important", "Svarīgs",
"Language", "Valoda",
"Languages", "Valodas",
"Mon", "Pirm",
"Movie", "Filma",
"Music", "Mūzika",
"Musical", "Mūzikls",
"NONE", "NAV",
"Name", "Nosaukums",
"Nature", "Daba",
"Network", "Tīkls",
"Network A", "Tīkls A",
"Network B", "Tīkls B",
"Networks", "Tīkli",
"News", "Ziņas",
"None", "Neviens",
"Not set", "Nav iestatīts",
"Number", "Numurs",
"OK", "Labi",
"Off", "Izslēgts",
"On", "Ieslēgts",
"Owner", "Īpašnieks",
"Password", "Parole",
"Password2", "Parole2",
"Run as group", "Palaist kā grupa",
"Run as user", "Palaist kā lietotājs",
"Running", "Darbojas",
"Sat", "Se",
"Season", "Sezona",
"Season %d", "Sezona %d",
"Server", "Serveris",
"Settings", "Iestatījumi",
"Show", "Parādīt",
"Show this page", "Parādīt šo lapu",
"Source", "Avots",
"Start", "Sākt",
"Status", "Statuss",
"Stop", "Apturēt",
"Subtitle", "Subtitri",
"Sun", "Sv",
"Sync", "Sinhronizēt",
"System", "Sistēma",
"Tags", "Birkas",
"Thu", "Cet",
"Title", "Virsraksts",
"Tue", "Otr",
"Type", "Tips",
"Unknown", "Nezināms",
"Unknown reason (%i)", "Nezināms iemesls (%i)",
"Username", "Lietotājvārds",
"War", "Karš",
"Wed", "Tr",
"hrs", "st",
"mins", "min",
NULL, NULL
};

static const char *tvh_locale_22[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nPapildomai informaciją rasite Tvheadend puslapyje: https://tvheadend.org\n",
"# Mapped channels", "# Susieti kanalai",
"# Muxes", "# Multiplekseriai",
"# Services", "# Servisai",
"# tune repeats", "# pakartotiniai kreipimaisi",
"(Default profile)", "(Numatytasis profilis)",
"(Read-only)", "(Tik nuskaitymui)",
"1 day", "1 diena",
"1 month", "1 mėnuo",
"1 week", "1 savaitė",
"1 year", "1 metai",
"2 months", "2 mėnesiai",
"2 weeks", "2 savaitės",
"2 years", "2 metai",
"2-Port switch (universal LNB)", "2-jų portų relė (universali LNB)",
"3 days", "3 dienos",
"3 months", "3 mėnesiai",
"3 weeks", "3 savaitės",
"3 years", "3 mėnesiai",
"4-Port switch (universal LNB)", "4-ių portų relė (universalus LNB)",
"5 days", "5 dienos",
"6 months", "6 mėnesiai",
"A channel tag (e.g. a group of channels) to which this rule applies.", "Kanalų žyma (grupė), kuriai priskiriama ši taisyklė",
"AC-3 = descriptor 6", "AC-3 = 6-as deskritporius",
"AC-3 detection", "AC-3 atradimas",
"ACTIVE", "AKTYVU",
"AES Constant Code Word", "AES Nuolatinis Kodo Žodis",
"ANY", "BETKOKS",
"ATSC SI Tables", "ATSC SI lentelės",
"ATSC source ID", "ATSC šaltinio ID",
"ATSC-C Network", "ATSC-C Tinklas",
"ATSC-C multiplex", "ATSC-C multiplekseris",
"ATSC-T Network", "ATSC-T Tinklas",
"ATSC-T multiplex", "ATSC-T multiplekseris",
"Aborted by user", "Atšaukta vartotojo",
"Accept remote subscription weight", "Sutikti su nuotoliniu svoriu",
"Accept the remote subscription weight (from the SAT>IP client).", "Sutikti su nuotoliniu svoriu (iš SAT>IP kliento)",
"Accept zero value for TSID", "Sutikti su TSID nuline reikšme",
"Access", "Prieiga",
"Access (ACL)", "Prieigos sąrašas (ACL)",
"Action", "Veiksmas",
"Adapter in use by another subscription", "Kita prenumerata naudoją šį adapterį",
"Add file and line numbers to debug", "Pridėti failą ir eilučių skaičių debuginimui",
"Additional arguments to pass to the grabber.", "Papildomi argumentai graberiui",
"Additional service names found in EPG data.", "Papildomi servisų pavadinimai rasti EPG duomenyse",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "Papildomas laikas (sekundėmis) tiuneriui pasiruošti įrašymui. Tai naudinga turintiems tiunerius, kurie delsia ir/ar   teršia pradėdami.",
"Admin", "Administratorius",
"Admin password", "Administratoriaus slaptažodis",
"Admin username", "Administratoriaus vartotojo vardas",
"Administrator login", "Administratoriaus prisijungimas",
"Adult movie", "Suagusiųjų filmas",
"Advanced", "Pažangiau",
"Advanced (non-universal LNBs, rotors, etc.)", "Pažangesnis (ne universalūs LNB, varikliai ir t.t.)",
"Adventure", "Nuotykiai",
"Advertisement / Shopping", "Reklamos / Apsipirkimai",
"After creating the network scan it for services.", "Ieškoti servisų sukūrus tinklą.",
"All (Streaming plus DVR)", "Viskas (Transliacija+DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Visi probleminiai simboliai bus pakeisti pabraukimu. Papildomą informaciją rasite vartotojo gide.",
"All lower-case", "Viskas mažosiomis raidėmis",
"All proceeds are used to support server infrastructure and buy test equipment.", "Visi pavedimai bus naudojami palaikyti serverių infrastruktūrą ir pirkti įrangą bandymams. ",
"All: Record if different description", "Viskas: Įrašyti, jei skiriasi aprašymas",
"All: Record if different episode number", "Viskas: Įrašyti, jei skiriasi epizodo numeris",
"All: Record if different subtitle", "Viskas: Įrašyti, jei skiriasi paantraštė",
"All: Record once per day", "Viskas: Įrašyti kartą per dieną",
"All: Record once per week", "Viskas: Įrašyti kartą per savaitę",
"Allow high bitrate mode (CI+ CAMs only).", "Leisti aukšto bitų dažnumo režimą (tik CI+ CAM'ams)",
"Allow idle scan tuning on this device.", "Leisti skanuoti šiuo įrenginiu, esant tuščiai eigai",
"Allow the combined size of all timeshift buffers to potentially grow unbounded until your storage media runs out of space.", "Leisti visiems laiko-poslinkio buferiams augti tol, kol baigsis talpos atmintis.",
"Allow the timeshift buffer to grow unbounded until your storage media runs out of space. Warning, enabling this option may cause your system to slow down or crash completely!", "Leisti laiko-poslinkio buferiui augti tol, kol baigsis talpos atmintis. Įspėjimas, įgalinus šią parinktį sistema gali sulėtėti ar visiškai nulūžti!",
"Allow/Disallow access to the 'Configuration' tab.", "Leisti/Neleisti prieigą prie \"Konfigūracija\" skirtuko.",
"Allow/Disallow web interface access (this  includes access to the EPG).", "Leisti/Neleisti prieigą prie naršyklės interfeiso (taip pat ir prie EPG).",
"Allowed DVR profiles. This limits the profiles the user has access to.", "Leistini DVR profiliai. Vartotojui bus suaktyvinti tik pasirinkti profiliai.",
"Allowed network", "Leistinas tinklas",
"Allowed networks", "Leistini tinklai",
"Alternate PID path", "Alternatyvus PID kelias",
"Alternate configuration path", "Alternatyvi konfigūracijos vieta/takas",
"Altitude (in meters).", "Aukštis (metrais).",
"Altitude (meters)", "Aukštis (metrai)",
"Always send the whole DiseqC sequence including LNB setup (voltage, tone). If this is not checked, only changed settings are sent, which may cause issues with some drivers. If the tuning is not reliable, try activating this option.", "Visada siųsti pilną DiSEqC komandų seką, įskaitant LNB parametrus (voltažą, toną). Jei ši nuostata nepažymėta, siunčiamos tik pakitusios nuotatos. Tai gali sukelti problemų su kai kuriomis tvarkyklėmis. Jei tiuneris veikia nepatikimai, pabandykite suaktyvinti šią nuostatą.",
"Always use the name defined by the network.", "Visada naudoti tinklo pateiktą pavadinimą.",
"Always use the old ioctls to read the linuxdvb status (signal strength, SNR, error counters). Some drivers are not mature enough to provide the correct values using the new v5 linuxdvb API.", "Visada naudoti seną ioctls linuxdvb būsenai (signalo stiprumo, SNR, klaidų skaitliukų) nuskaityti. Kai kurios tvarkyklės yra nepakankamai ištobulintos naudoti naująjį linuxdvb v5 API.",
"An event which starts between this \"start after\" and \"start before\" will be matched (including boundary values).", "Įrašymas įvyks \"pradėti po ir \"pradėti prieš\" intervale (imtinai)",
"Animals", "Gyvūnai",
"Anonymize HTSP access", "Anonimizuoti HTSP prieigą",
"Any", "Bet koks",
"Arts", "Menai",
"Arts / Culture (without music)", "Menai / Kultūra (be muzikos)",
"Arts magazines", "Menų žurnalai",
"Assign predefined muxes to networks", "Priskirti tinklams iš anksto sugeneruotą multiplekserių sąrašą ",
"Associate this device with one or more networks.", "Susieti šį įrenginį su vienu ar daugiau tinklų.",
"Athletics", "Atletika",
"Audio Stream Filter", "Audio srauto filtras",
"Audio bitrate (kb/s) (0=auto)", "Audio bitų dažnis (kb/s) (0=auto)",
"Audio channel layout.", "Audio kanalo išdėstymas",
"Audio codec", "Audio kodekas",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Perkodavimui naudojamas audio kodekas. \"Nenaudoti\" išjungs garsą.",
"Audio stream", "Audio srautas",
"Audio type", "Audio tipas",
"Audio-only stream", "Tik audio srautas",
"Audioes muxer", "Audio mukseris",
"Auto check disabled", "Automatinis patikrinimas išjungtas",
"Auto check enabled", "Automatinis patikrinimas įjungtas",
"Auto record", "Automatinis įrašymas",
"Auto record caption", "Automatinio įrašymo antraštė",
"Auto recording%s%s", "Automatinis įrašymas%s%s",
"Auto time record", "Automatinis laiko įrašas",
"Auto-Map to channels", "Automatiškai priskirti prie kanalų",
"Automatic checking", "Automatinis patikrinimas",
"Automatic recording caption.", "Automatinė įrašo antraštė",
"Automatic schedule entry not found", "Automatinis tvarkaraščio įrašas nerastas",
"Automatic time scheduler entry not found", "Automatinis tvarkaraščio įrašas nerastas",
"Automatically link EPG data to the channel (using the channel name for matching). If you turn this option off, only the OTA EPG grabber will be used for this channel unless you've specifically set a different EPG Source.", "Automatiškai susieti EPG duomenis su kanalu (naudojant kanalo pavadinimą sutapatinimui). Išjungus šią nuostatą, tik   oru gaunamas EPG bus nadojamas, nebent bus nustatytas kitas konkretus EPG šaltinis.",
"Automatically map EPG source", "Automatiškai susieti EPG šaltinį",
"Automatically map all available services to channels.", "Automatiškai susieti visus prieinamus servisus su kanalais.",
"Automatically map channels defined within the bouquet.", "Automatiškai susieti buketuose nurodytus kanalus.",
"Automatically name from network", "Automatiškai naudoti pavadinimą iš tinklo",
"Automatically record.", "Automatiškai įrašyti.",
"Autorec maximum count (0=unlimited)", "Automatinių įrašų kiekis (0=neribojama)",
"Autorec maximum schedules limit (0=unlimited)", "Maksimalus automatinių įrašų grafikų kiekis (0=neribojama)",
"Bad", "Blogas",
"Bad request", "Bloga užklausa",
"Ballet", "Baletas",
"Bandwidth", "Srautas",
"Base DVB SI Tables (PAT,CAT,PMT,SDT etc.)", "Pagrindinės DVB SI lentelės (PAT,CAT,PMT,SDT ir t.t.)",
"Based on software from", "Pagrįsta programine įranga iš",
"Basic", "Pagrindinis",
"Bind to specific local IP address.", "Susieti su konkrečiu vietiniu IP adresu.",
"Blue", "Mėlynas",
"Bouquet", "Buketas",
"Bouquet (auto)", "Buketas (auto)",
"Bouquet source.", "Buketo šaltinis.",
"Bouquets", "Buketai",
"Brand", "Markė",
"Branding information (if available).", "Prekės ženklo informacija (jei prieinama).",
"Broadcast", "Transliacija",
"Broadcast type", "Transliacijos tipas",
"Broadcast.", "Transliacija.",
"Broadcasting", "Transliavimas",
"Buffering limit (ms)", "Buferio ribojimas (ms)",
"Build", "Versija",
"By default, linuxdvb's status read period is 1000ms (one second). The accepted range is 250ms to 8000ms. Note that for some hardware or drivers (like USB), the status operations take too much time and CPU. In this case, increase the default value. For fast hardware, this value might be decreased to make the decision of the re-tune algorithm based on the signal status faster.", "Pagal nutylėjimą linuxdvb busenos nuskaitymo periodas yra 1000ms (viena sekundė). Priimtinas rėžis yra tarp 250ms ir 8000ms. Pastaba: kai kurioms plokštėms ar jų tvarkyklėms (kaip kad USB) būsenos operacijos užima perdaug laiko ir eikvoja perdaug procesoriaus resursų. Naudojant sparčią \"geležį\" ši reikšmė gali būti sumažinta, siekiant paspartinti signalo būsena paremtą derinimo algoritmo veikimą.",
"CA (descrambling) Client", "CA (iškodavimo) klientas",
"CA Stream Filter", "CA srauto filtras",
"CA client %i", "CA klientas %i",
"CA identification", "CA identifikacija",
"CA provider", "CA tiekėjas",
"CAPMT (Linux Network DVBAPI)", "CAPMT (Linux tinklinis DVBAPI)",
"CAPMT CA Client", "CAPMT CA klientas",
"CAPMT interval (in ms).", "CAPMT intervalas (milisekundėmis).",
"CAPMT interval (ms)", "CAPMT intervalas (ms)",
"CAPMT query interval (ms)", "CAPMT užklausos intervalas (ms)",
"CAPMT query interval (ms).", "CAPMT užklausos intervalas (ms).",
"CI Module", "CI modulis",
"CRASH", "NULŪŽO",
"CSA (descrambling)", "CSA (iškodavimas)",
"CWC CA Client", "CWC CA klientas",
"Cache scheme", "Podėlio schema.",
"Camd.socket filename / IP Address (TCP mode)", "Camd.socket failo vardas / IP adresas (TCP režimas)",
"Cartoons", "Filmukai",
"Change parameters", "Keisti parametrus",
"Channel", "Kanalas",
"Channel Tags", "Kanalų žymos",
"Channel does not exist", "Kanalas neegzistuoja",
"Channel icon", "Kanalo ikona",
"Channel icon URL.", "Kanalo ikonos URL.",
"Channel icon as defined in EPG data.", "Kanalo ikona iš EPG duomenų.",
"Channel icon name scheme", "Kanalo ikonų pavadinimo schema",
"Channel icon path", "Kanalo ikonos kelias",
"Channel mapping options", "Kanalų susiejimo nuostatos",
"Channel name", "Kanalo pavadinimas",
"Channel not found", "Kanalas nerastas",
"Channel number", "Kanalo numeris",
"Channel number as defined in EPG data.", "Kanalo numeris iš EPG duomenų.",
"Channel number offset", "Kanalo numerio pokytis",
"Channel number range", "Kanalo numerių rėžis",
"Channel numbers (heuristic)", "Kanalo numeriai (euristiškai)",
"Channel numbers from", "Kanalų numeriai iš",
"Channel tag", "Kanalo žyma",
"Channel tag does not exist", "Kanalo žyma neegzistuoja",
"Channel tag reference", "Kanalo žymos nuoroda",
"Channel tags", "Kanalo žymos",
"Channel tags the user is allowed access to/excluded from.", "Kanalų žymos, prie kurių vartotojui suteikiama ar uždraudžiama prieiga.",
"Channel to use/used for the recording.", "Įrašymui naudojamas kanalas",
"Channel update options", "Kanalo atnaujinimo nuostatos",
"Channels", "Kanalai",
"Channels EPG data is used by.", "Kanalai, kurie naudoja šį EPG",
"Character set", "Simbolių rinkinys",
"Characters not supported in Windows filenames (e.g. for an SMB/CIFS share) will be stripped out or converted.", "Windows failų varduose (pvz. SMB/CIFS bendrinimuose) nenaudojami simboliai bus iškirpti ar konvertuoti.",
"Charset", "Simbolių rinkinys",
"Check availability", "Patikrinti prieinamumą",
"Check for the services' presence. If the service is no longer broadcast this field will change to Missing In PAT/SDT.", "Tikrinti serviso prieinamumą. Jei servisas nebetransliuoja, šis laukas bus pervadintas į \"Nerandama PAT/SDT\".",
"Children's / Youth Programs", "Vaikų / Jaunimo programos",
"Children's / Youth programs", "Vaikų / Jaunimo programos",
"Cinema", "Kinas",
"Class", "Klasė",
"Classical", "Klasikinis",
"Classical music", "Klasikinė muzika",
"Clean effects", "Švarūs efektai",
"Client name", "Kliento pavadinimas",
"Clone scheduled entry on error", "Esant klaidai klonuoti planuojamąjį įrašą",
"Code Word Client (newcamd)", "Kodinio žodžio klientas (newcamd)",
"Comedy", "Komedija",
"Command delay time (ms) (10-200)", "Komandos uždelsimo laikas (ms) (10-200)",
"Command time (ms) (10-100)", "Komandos laikas (ms) (10-100)",
"Comment", "Komentaras",
"Commercial break", "Komercinė pertraukėlė",
"Commercials will be dropped from the recordings. Commercial detection works using EITp/f (EPG running state) and for the Swedish channel TV4 (using teletext info).", "Reklamos bus iškirptos. Reklamų atrandamimas veikia naudojant EITp/f (EPG veikimo būsena) ir švediškam TV4 kanalui (naudojant teleteksto informaciją).",
"Committed", "Priskirta",
"Completed OK", "Užbaigta gerai",
"Compress EPG database", "Suspausti EPG duomenų bazę",
"Compress the EPG database to reduce disk I/O and space.", "Suspausti EPG duomenų bazę siekiant sumažinti skaitymo/rašymo operacijų skaičių ir taupyti vietą.",
"Conditional Access", "Sąliginė prieiga (CA)",
"Conditional Access Client", "Sąliginės prieigos klientas",
"Conditional Access Identification.", "Sąliginės prieigos identifikacija.",
"Configuration", "Konfigūracija",
"Configuration - Base", "Konfigūracija - Pagrindinė",
"Configuration - Image Cache", "Konfigūracija - Paveiksliukų podėlis",
"Configuration - SAT>IP Server", "Konfigūracija - SAT>IP Serveris",
"Configuration ID", "Konfigūracijos ID",
"Configuration name", "Konfigūracijos pavadinimas",
"Configuration version", "Konfigūracijos versija",
"Connection limit type", "Jungčių apribojimo tipas",
"Connection limits", "Jungčių apribojimai",
"Constellation", "Moduliacija",
"Container", "Konteineris",
"Container to use for the transcoded stream.", "Perkoduotam srautui naudoti šį konteinerį.",
"Content character set", "Turinio simbolių rinkinys",
"Content reference identifier authority.", "Turinio nuorodos identifikatoriaus šaltinis.",
"Content type", "Turinio tipas",
"Content type.", "Turinio tipas.",
"Contest", "Konkursas",
"Continue recording for x minutes after scheduled stop time", "Tęsti įrašymą x minučių po suplanuoto pabaigos laiko",
"Continue recording for x minutes after scheduled stop time.", "Tęsti įrašymą x minučių po suplanuoto pabaigos laiko.",
"Cookie expiration (days)", "Sausainiukų pasibaigimas (dienomis)",
"Cooking", "Maisto gamyba",
"Copy codec type", "Kopijuoti kodeko tipą",
"Copy layout", "Kopijuoti išdėstymą",
"Count of objects", "Objektų kiekis",
"Create SDTV/HDTV/Radio tags.", "Sukurti SDTV/HDTV/Radijo žymes.",
"Create a bouquet from the playlist.", "Sukurti buketą iš grojaraščio.",
"Create a directory per channel when storing recordings. If both this and the 'directory per day' checkbox is enabled, the date-directory will be the parent of the per-channel directory.", "Įrašant sukurti direktoriją pagal kanalo pavadinimą. Jei šis nustatymas pažymėtas kartu su \"direktorija per dieną\", kanalo direktorija bus dienos kategorijos viduje.",
"Create a directory per title when storing recordings. If the day/channel directory checkboxes are also enabled, those directories will be parents of this directory.", "Įrašant sukurti direktoriją pagal įrašo pavadinimą. Jei taip pat pažymeti dienos/kanalo langeliai, pavadinimo katalogas bus jų viduje.",
"Create a new directory per day in the recording system path. Folders will only be created when something is recorded. The format of the directory will be ISO standard YYYY-MM-DD.", "Sukurti direktorijas įrašų sistemos kelyje remiantis dienomis. Katalogai bus sukurti ką nors įrašant. Direktorijų pavadinimai kuriami pagal ISO standartą MMMM-MM-DD.",
"Create a provider name tag.", "Kurti tiekėjo pavadinimo žymę.",
"Create and associate a network tag to created channels.", "Sukurti ir priskirti tinklo pavadinimo žymę sukurtiems kanalams.",
"Create and associate a provider tag to created channels.", "Sukurti ir priskirti tiekėjo pavadinimo žymę sukurtiems kanalams.",
"Create and link these tags to channels when mapping.", "Sukurti ir priskirti šias žymes kanalams susiejimo metu.",
"Create bouquet", "Sukurti buketą",
"Create bouquet tag", "Sukurti buketo žymę",
"Create directories using these permissions.", "Sukurti direktorijas naudojant šias teises.",
"Create files using these permissions.", "Sukurti failus naudojant šias teises.",
"Create network name tags", "Sukurti tinklo vardo žymes",
"Create network name tags (set by provider).", "Sukurti tinklo vardo žymes (nustatytas tiekėjo).",
"Create network tags", "Sukurti tinklo žymes.",
"Create provider name tags", "Sukurti tiekėjo vardo žymes",
"Create provider tags", "Sukurti tiekėjo žymes",
"Create tags", "Sukurti žymes",
"Create tags in recordings using media containers that support metadata (if possible).", "Sukurti žymes įrašuose naudojant media konteinerius, kurie palaiko meta duomenis (jei įmanoma).",
"Create type-based tags", "Sukurti tipu paremtas žymes",
"Created", "Sukurta",
"Creator", "Kūrėjas",
"Cron multi-line", "Cron daugiaeilutinis",
"Culture (without music)", "Kultūra (išskyrus muziką)",
"Culture magazines", "Kultūros žurnalai",
"Current RTSP port.", "Dabartinis RSTP portas.",
"Current affairs", "Aktualijos",
"Current number of objects.", "Dabartinis objektų skaičius.",
"Current object size.", "Dabartinis objekto dydis.",
"Current tuner configuration.", "Dabartinė tiunerio konfigūracija.",
"Custom HTTP headers", "Pasirinktinė HTTP antraštė",
"DAB Network", "DAB tinklas",
"DAB multiplex", "DAB multiplekseris",
"DBus - use the session message bus instead of the system one", "DBus - naudoti session message bus vietoje system",
"DES Constant Code Word", "DES Pastovaus kodo žodis",
"DES Key.", "DES Raktas.",
"DES key", "DES raktas",
"DSCP/TOS for streaming", "DSCP/TOS transliavimui",
"DVB CAM Client", "DVB CAM Klientas",
"DVB CSA (descrambling) Tables", "DVB CSA (iškodavimo) Lentelės",
"DVB EPG Tables", "DVB EPG Lentelės",
"DVB Inputs - Networks", "DVB įvestys - Tinklai",
"DVB SI Tables", "DVB SI Lentelės",
"DVB Time Tables", "DVB Tvarkaraščiai",
"DVB multiplex", "DVB multiplekseris",
"DVB scan files", "DVB skanavimo failai",
"DVB scan files path", "DVB skanavimo failų kelias",
"DVB-C Network", "DVB-C Tinklai",
"DVB-C multiplex", "DVB-C multiplekseris",
"DVB-S Network", "DVB-S Tinklas",
"DVB-S Satellite Configuration", "DVB-S palydovo konfigūracija",
"DVB-S multiplex", "DVB-S multiplekseris",
"DVB-T Network", "DVB-T Tinklas",
"DVB-T multiplex", "DVB-T multiplekseris",
"DVR - Auto-recording (Autorecs)", "DVR - Auto-įrašymas (Autoįrašai)",
"DVR - Profiles", "DVR - Profiliai",
"DVR - Time-based Recording (Timers)", "DVR - Laiku paremtas įrašymas (taimeriai)",
"DVR behavior", "DVR elgesys",
"DVR configuration", "DVR konfigūracija",
"DVR configuration profiles", "DVR konfigūracijos profiliai",
"DVR configurations", "DVR konfigūracijos",
"DVR entry not found", "DVR įrašas nerastas",
"DVR file retention period", "DVR failo išsaugojimo laikotarpis",
"DVR log retention", "DVR žurnalo išsaugojimas",
"DVR log retention period", "DVR žurnalo išsaugojimo laikotarpis",
"DVR override: high", "DVR nepaisymas: aukštas",
"DVR override: important", "DVR nepaisymas: svarbus",
"DVR override: low", "DVR nepaisymas: žemas",
"DVR override: normal", "DVR nepaisymas: normalus",
"DVR override: unimportant", "DVR nepaisymas: nesvarbus",
"DVR profile to use/used for the recording.", "DVR profilis naudoti/naudojamas įrašymams.",
"DVR schedule does not exist", "DVR tvarkaraštis neegzistuoja",
"DVR schedule does not have a file yet", "DVR tvarkarašis kol kas neturi failo",
"DVR schedule not found", "DVR tvarkaraštis nerastas",
"Dance", "Šokiai",
"Data errors", "Duomenų klaidos",
"Data path (if applicable).", "Duomenų kelias (jei taikytina).",
"Date the EPG data was last updated (not set for OTA grabbers).", "Laikas kada paskutinį kartą buvo atnaujinti EPG duomenys (nesiskaito oru gaunamiems duomenims).",
"Days of Week", "Savaitės dienos",
"Days of the week to which the rule should apply.", "Savaitės dienos, kurioms taisyklė taikoma.",
"Days of week", "Savaitės dienos",
"Debate", "Debatai",
"Debug libav log", "Debuginti libav žurnalą",
"Debug options", "Debuginimo nuostatos",
"Debug subsystems", "Debuginimo subsistemos",
"Debug to syslog", "Debuginti į syslog",
"Debug trace (low-level)", "Debuginimo pėdsakas (žemo lygio)",
"Debugging", "Debuginimas",
"Default", "Numatyta",
"Default language(s)", "Numatyta kalba (-os)",
"Default language.", "Numatyta kalba.",
"Default priority", "Numatytas pirmumas",
"Default user interface level.", "Numatytas vartotojo interfeiso lygis.",
"Define network group to limit network usage.", "Nuodykite tinklo grupę apriboti tinklo naudojimui.",
"Delivery system", "Pristatymo sistema",
"Demux path", "Demux takas",
"Descramble services (limit per mux)", "Iškoduoti servisus (apriboti per multiplekserį)",
"Descrambler", "Iškoduotojas",
"Descrambler buffer (TS packets)", "Iškoduotojo buferis (TS paketai)",
"Description", "Aprašymas",
"Detective", "Detektyvas",
"Device ID", "Įrenginio ID",
"Device model", "Įrenginio modelis",
"Device path", "Įrenginio kelias",
"Differentiated Services Code Point / Type of Service: Set the service class Tvheadend sends with each packet. Depending on the option selected this tells your router the prority in which to give packets sent from Tvheadend, this option does not usually need changing. See https://en.wikipedia.org/wiki/Differentiated_services for more information. ", "Differentiated Services Code Point / Type of Service: nustatykite serviso klasę, kurią Tvheadend siunčia su kiekvienu paketu. Priklausomai nuo pasirinimo, tuo maršrutizatoriui nurodoma kokį prioritetą suteikti Tvheadend serveriui. Paprastai šio nustatymo keisti nereikia. Papildomą informaciją rasite: https://en.wikipedia.org/wiki/Differentiated_services",
"Digest access authentication is intended as a security trade-off. It is intended to replace unencrypted HTTP basic access authentication. This option should be enabled for standard usage.", "Digest access authentication skirtas apsaugoti prisijungimo duomenis prisijungimo metu. \nFunkcija skirta pakeisti nešifruotą HTTP basic access authentication. Šis nustatymas turėtų būti įjungtas.",
"Digital Video Recorder", "Skaitmeninis Vaizdo Magnetofonas",
"Directory", "Direktorija",
"Directory override. Override the subdirectory rules specified by the DVR configuration and put all recordings done by this entry into the specified subdirectory", "Direktorijos nepaisymas. Nepaisyti DVR subdirektorijų taisyklių ir talpinti visus įrašus į pateiktą subdirektoriją",
"Directory permissions (octal, e.g. 0775)", "Direktorijų teisės (pvz. 0775)",
"Directory used by the entry.", "Įrašo naudojama direktorija.",
"Disable", "Išjungti",
"Disable DVB bouquets", "Išjungti DVB buketus",
"Disable SAT>IP client", "Išjungti SAT>IP klientą",
"Disable all access control checks", "Išjungti visus prieigos kontrolės patikrinimus",
"Disable debug on stderr", "Išjungti stderr debuginimą ",
"Disable device/firmware-specific workarounds", "Išjungti įrenginio/firmwaro apėjimus",
"Disable syslog (all messages)", "Išjungti syslog (visos žinutės)",
"Disabled", "Išjungta",
"Discover more muxes using the Network Information Table (if available).", "Atrasti daugiau multiplekserių naudojant Tinklo Informacijos Lentelę (jei prieinama).",
"Discussion", "Diskusija",
"DiseqC", "DiSEqC",
"DiseqC repeats", "DiSEqC pakartojimai",
"Do not send any stream specific information to the HTSP client like signal strength, input source etc.", "Nesiųsti srautą apibūdinančios informacijos (signalo stiprumo, įvesties šaltinio ir t.t.) HTSP klientui.",
"Do not use", "Nenaudoti",
"Documentary", "Dokumentika",
"Don't abort streaming when an encrypted stream can't be decrypted by a CA client that normally should be able to decrypt the stream.", "Nenutraukti transliavimo kai užšifruota programa negali būti iššifruota CA kliento, kuris paprastai turėtų galėti iššifruoti konkretų srautą.",
"Don't backup configuration tree at upgrade", "Nedaryti konfigūracijos medžio atsarginės kopijos atnaujinant.",
"Don't include the title in the filename.", "Neįtraukti transliacijos pavadinimo į failo vardą.",
"Don't include title in filename", "Neįtraukti transliacijos pavadinimo į failo vardą",
"Don't keep", "Nelaikyti",
"Don't re-record", "Pakartotinai neįrašinėti",
"Don't re-record if recording fails.", "Pakartotinai neįrašinėti, jei įrašas nepavyko.",
"Don't re-schedule if recording fails.", "Neperplanuoti, jei įrašas nepavyko.",
"Don't reschedule", "Neperplanuoti",
"Don't use the provider's channel numbers.", "Nenaudoti tiekėjo pateikiamų kanalų numerių.",
"Duplicate handling", "Dublikatų tvarkymas",
"Duplicate recording handling.", "Vienodų įrašų tvarkymas.",
"EIT - skip TSID check", "EIT - praleisti TSID patikrą",
"EIT time offset", "EIT laiko kompensavimas",
"EPG Database", "EPG Duombazė",
"EPG Grabber", "EPG Graberis",
"EPG Grabber Channel", "EPG Graberio Kanalas",
"EPG Grabber Configuration", "EPG Graberio Konfigūracija",
"EPG Language (priority order)", "EPG Kalba (pirmumo tvarka)",
"EPG data ID.", "EPG duomenų ID.",
"EPG name", "EPG pavadinimas",
"EPG overlap cut", "EPG persidengimo kirpimas",
"EPG scan", "EPG skanavimas",
"EPG source", "EPG šaltinis",
"EPG update window", "EPG atnaujinimo langas",
"Economics", "Ekonomika",
"Education", "Švietimas",
"Education / Science / Factual topics", "Švietimas / Mokslas / Faktinės temos",
"Educational", "Pedagoginis",
"Electronic Program Guide", "Elektroninis Programų Gidas",
"Elementary Stream Filter", "Elementarus Srauto Filtras",
"Elementary stream filter", "Elementarus srauto filtras",
"Enable", "Įjungti",
"Enable (auto)", "Įjungti (automatiškai)",
"Enable DBus", "Įjungti DBus",
"Enable NTP driver", "Įjungti NTP tvarkyklę",
"Enable coredumps for daemon", "Įjungti coredumps demonui",
"Enable debug on stderr", "Įjungti stderr debuginimą",
"Enable debug subsystems", "Įjungti debuginimą subsistemoms",
"Enable debug to file", "Įjungti debuginimą į failą",
"Enable debug to syslog", "Įjungti debuginimą į syslog",
"Enable exclusion of user-config defined channel tags. This will prevent the user from accessing channels associated with the tags selected (below).", "Įgalinti vartotojo sukonfigūruotų kanalų žymų išimtį. Tai padės išvengti vartotojo priėjimo prie pažymėtų žymų susietų su kanalais (žemiau).",
"Enable if the SAT>IP box requests plts=on parameter in the SETUP RTSP command for DVB-S2 muxes.", "Įjungti, jei SAP>IP dėžė prašo plts=on parametro SETUP RTSP komandoje DVB-S2 multiplekseriams.",
"Enable if the SAT>IP box requires pids=21 parameter in the SETUP RTSP command", "Įjungti, jei SAP>IP dėžė reikalauja pids=21 parametro SETUP RTSP komandoje.",
"Enable if the SAT>IP box supports the addpids/delpids commands.", "Įjungti, jei SAP>IP dėžė palaiko addpids/delpids komandas.",
"Enable or disable RTP/AVP/TCP transfer mode (embedded data in the RTSP session) support.", "Įjungti arba išjungti RTP/AVP/TCP perdavimo režimo (RTSP sesijoje įterpti duomenys) palaikymą.",
"Enable or disable fast input switching.", "Įjungti arba išjungti greitąjį įvesties perjungimą.",
"Enable or disable full mux mode.", "Įjungti arba išjungti pilną multiplekserio režimą.",
"Enable or disable ignoring of Event Information Table (EIT) data on this mux.", "Įjungti arba išjungti Įvykių Informacijos Lentelės (EIT) duomenų ignoravimą šiam multiplekseriui.",
"Enable or disable this configuration.", "Įjungti arba išjungti šią konfigūraciją.",
"Enable over-the-air program guide (EPG) scanning on this input device.", "Įjungti oru gaunamo programų gido (EPG) skanavimą šiam įrenginiui.",
"Enable syslog", "Įjungti syslog",
"Enable system time updates. This will only work if the user running Tvheadend has rights to update the system clock (normally only root).", "Įjungti sistemos laiko atnaujinimus. Nustatymas veiks tik tuo atveju, jei vartotojas paleidęs Tvheadend turi pakankamas teises sistemos laikrodžiui atnaujinti (paprastai tik root).",
"Enable the CAs (conditional accesses) tab in web user interface for the advanced level. By default, this tab is visible only in the expert level.", "Įjungti CAs (conditional accesses) skirtuką vartotojo interfeise pasirinkus pažangų lygį. Pagal nutylėjimą šis skirtukas matomas tik eksperto lygyje.",
"Enable this filter.", "Įjungti šį filtrą.",
"Enable trace subsystems", "Įjungti subsistemų sekimą",
"Enable web UI debug (non-minified JS)", "Įjungti naršyklinį debuginimą (nesumažintas JS)",
"Enable, disable or ignore the mux. When the mux is marked as ignore, all discovered services are removed.", "Įjungti, išjungti arba ignoruoti multiplekserį. Pažymėjus ignoruoti, visi atrasti servisai bus pašalinti.",
"Enable/Disable CA client.", "Įjungti/Išjungti CA klientą.",
"Enable/Disable service.", "Įjungti/Išjungti servisą.",
"Enable/Disable the entry.", "Įjungti/Išjungti įrašą.",
"Enable/disable EPG data for the entry.", "Įjungti/Išjungti įrašui EPG duomenis.",
"Enable/disable LNA.", "Įjungti/Išjungti žematriukšmį stiprintuvą.",
"Enable/disable auto-rec rule.", "Įjungti/Išjungti autoįrašymo taisyklę.",
"Enable/disable debugging output to syslog.", "Įjungti/Išjungti debuginimo išvestį į syslog.",
"Enable/disable inclusion of low-level debug traces.", "Įjungti/Išjungti žemo lygio debuginimo pėdsakų įtraukimą.",
"Enable/disable interface quick tips.", "Įjungti/Išjungti greituosius interfeiso patarimus.",
"Enable/disable libav log output.", "Įjungti/Išjungti libav žurnalo išvestį.",
"Enable/disable logging to syslog.", "Įjungti/Išjungti žurnalizavimą į syslog.",
"Enable/disable offering of Entitlement Management Message updates.", "Įjungti/Išjungti EMM atnaujinimo pasiūlymus.",
"Enable/disable pilot tone.", "Įjungti/Išjungti pilot toną.",
"Enable/disable power save mode (if supported by the device).", "Įjungti/Išjungti energijos taupymo režimą (jei palaiko įrenginys).",
"Enable/disable profile.", "Įjungti/Išjungti profilį.",
"Enable/disable the bouquet.", "Įjungti/Išjungti buketą.",
"Enable/disable the channel.", "Įjungti/Išjungti kanalą.",
"Enable/disable the device.", "Įjungti/Išjungti įrenginį.",
"Enable/disable the entry.", "Įjungti/Išjungti įrašą.",
"Enable/disable the grabber.", "Įjungti/Išjungti graberį.",
"Enable/disable the tag.", "Įjungti/Išjungti žymą.",
"Enable/disable tuner/adapter.", "Įjungti/Išjungti tiunerį/adapterį.",
"Enabled", "Įjungta",
"Encrypted", "Užšifruotas",
"Encrypted service", "Užšifruotas servisas",
"Enter a non-admin user password.", "Įveskite ne administratoriaus vartotojo slaptažodį.",
"Enter a non-admin user username.", "Įveskite ne administratoriaus vartotojo vardą.",
"Enter allowed network prefix(es). You can enter a comma-seperated list of prefixes here.", "Įveskite leistino tinklo prefiksą(-us). Galima nurodyti daugiau prefiksų, juos atskriant kableliais.",
"Enter an administrator password.", "Įveskite administratoriaus slaptažodį.",
"Enter an administrator username. Note: do not use the same username as the superuser backdoor account.", "Įveskite administratoriaus vartotojo vardą. Pastaba: nenaudokite to paties vartotojo vardo kaip ir superuser.",
"Entertainment programs for 10 to 16", "Pramoginės programos 10-ties - 16-kos metų žiūrovams",
"Entertainment programs for 6 to 14", "Pramoginės programos 6-ių - 14-kos metų žiūrovams",
"Environment", "Aplinka",
"Environment (pipe)", "Aplinka (vamzdis)",
"Episode", "Epizodas",
"Episode %d", "%d epizodas",
"Episode number/ID.", "Epizodo numeris/ID.",
"Equestrian", "Jodinėjimas",
"Error code", "Klaidos kodas",
"Error code of entry.", "Įrašo klaidos kodas.",
"Errors", "Klaidos",
"Even key", "Lyginis raktas",
"Even key.", "Lyginis raktas.",
"Event does not exist", "Įvykis neegzistuoja",
"Every day", "Kasdien",
"Exclude channel tags", "Išskirti kanalų žymas",
"Expeditions", "Ekspedicijos",
"Experimental film", "Eksperimentiniai filmai",
"Expert", "Ekspertas",
"Exported tuners", "Eksportuoti tiuneriai",
"External", "Išorinis",
"External EPG grabber", "Išorinis EPG graberis",
"External IP (NAT)", "Išorinis IP (NAT)",
"External PyEPG Grabber", "Išorinis PyEPG Graberis",
"External URL", "Išorinis URL",
"External URL of the bouquet.", "Išorinis buketo URL.",
"External XMLTV EPG Grabber", "Išorinis XMLTV EPG Graberis",
"Extra arguments", "Papildomi argumentai",
"Extra start time", "Papildomas laikas pradėti",
"Extra stop time", "Papildomas laikas sustabdyti",
"Extra warming up time (seconds)", "Papildomas apšilimo laikas (sekundės)",
"FAIL", "NEPAVYKO",
"FEC high", "FEC aukštas",
"FEC low", "FEC žemas",
"Factual topics", "Faktinės temos",
"Failed", "Nepavyko",
"Failed to open image", "Paveikslėlio atverti nepavyko ",
"Failed view", "Nepavykęs rodinys",
"Fantasy", "Fantazija",
"Fashion", "Mada",
"Fast input switch", "Greitas įvesties perjungimas",
"Fastscan DVB", "Spartus DVB skanavimas",
"File missing", "Failas nerastas",
"File not created", "Failas nesukurtas",
"File permissions (octal, e.g. 0664)", "Failo teisės (pvz. 0664)",
"File removed", "Failas pašalintas",
"File size", "Failo dydis",
"Filename", "Failo vardas",
"Filename character set", "Failo simbolių rinkinys",
"Filename options", "Failo vardo nuostatos",
"Filename used by the entry.", "Įrašo naudojama failo vardas.",
"Filesystem monitor", "Failinės sistemos stebėsena",
"Film", "Filmas",
"Fine arts", "Dailė",
"Finished", "Užbaigta",
"First word", "Pirmas žodis",
"Fit to RAM (cut rewind)", "Talpinti operatyviojoje atmintyje RAM (prikirpti atsukimą)",
"Fitness and health", "Fitnesas ir sveikata",
"Folk", "Liaudies",
"Folkloric", "Liaudiška",
"Football", "Futbolas",
"Force (auto)", "Priverstinai (auto)",
"Force CA ID (e.g. 0x2600)", "Priverstinai naudoti CA ID (pvz. 0x2600)",
"Force all network connections to this tuner to be made over the specified IP address, similar to the setting for the SAT>IP device itself. Setting this overrides the device-specific setting.", "Priverstinai nukreipti visus tinklo susijungimus per nurodytą IP adresą. Panašiai kaip ir SAT>IP įrenginio nustatymuose. Nustačius šį nustatymą bus nepaisoma specifinių įrenginių nustatymų.",
"Force an initial EPG grab at start-up.", "Priverstinai pradėti EPG grabinimą paleidus Tvheadend.",
"Force initial EPG grab at start-up", "Priverstinai pradėti EPG grabinimą paleidus Tvheadend",
"Force old status", "Priverstinai naudoti seno tipo būseną",
"Force pilot for DVB-S2", "Priverstinai naudoti pilot'ą DVB-S2",
"Force priority", "Priverstinis pirmumas",
"Force service type to 1", "Priverstinai naudoti 1-netą serviso tipui",
"Force teardown delay", "Priverstinai naudoti teardown uždelsimą",
"Force the delay between RTSP TEARDOWN and RTSP SETUP command (value from 'Next tune delay in ms' is used). Some devices are not able to handle quick continuous tuning.", "Priverstinai uždelsti tarp RTSP TEARDOWN ir RTSP SETUP komandų (naudojama \"Sekančio tune uždelsimas milisekundėmis\" reikšmė). Kai kurie įrenginiai nesusitvarko su staigiu, nuolatiniu perjunginėjimu.",
"Force the local UDP Port number here. The number should be even (RTP port). The next odd number (+1) will be used as the RTCP port.", "Priverstinai naudoti vietinį UDP portą. Skaičius turi būti lyginis (RTP portas), o sekantis nelyginis (+1) bus naudojamas kaip RTCP portas.",
"Force usage of entered CA ID on this service.", "Šiam įrenginiui priverstinai naudoti įvestą CA ID.",
"Foreign countries", "Užsienio valstybės",
"Forever", "Amžinai",
"Format string", "Failo vardas",
"Found muxes", "Rasti multiplekseriai",
"Found services", "Rasti servisai",
"Free subscription weight", "Laisvas prenumeratos svoris",
"Free-form field, enter whatever you like here.", "Laisvos formos laukas. Įveskite betką, kas tik Jums patinka.",
"Free-form text field, enter whatever you like here.", "Laisvos formos tekstinis laukas. Įveskite betką, kas tik Jums patinka.",
"Free-form text field, enter whatever you like.", "Laisvos formos laukas. Įveskite betką, kas tik Jums patinka.",
"Free-form text field. You can enter whatever you like here.", "Laisvos formos laukas. Įveskite betką, kas tik Jums patinka.",
"Free-format text field. Enter whatever you like here.", "Laisvos formos laukas. Įveskite betką, kas tik Jums patinka.",
"Frequency", "Dažnis",
"Frequency (Hz)", "Dažnis (Hz)",
"Frequency (MHz)", "Dažnis (MHz)",
"Frequency (in MHz).", "Dažnis (MHz)",
"Frequency (kHz)", "Dažnis (kHz)",
"Fri", "Pen",
"Friendly name", "Draugiškas vardas",
"Frontend number", "Sąsajos numeris",
"Frontend path", "Sąsajos kelias",
"Full DiseqC", "Pilnas DiSEqC",
"Full mux RX mode supported", "Pilnas multiplekserio RX režimas palaikomas",
"Full path to an icon used to depict the tag. This can be a TV network logotype, etc.", "Pilnas kelias iki ikonos žymei atvaizduoti. Tai gali būti TV tinklo logotipas ir pan.",
"Full pathname specification", "Pilnos tako specifikacijos",
"Full-text", "Pilnas tekstas",
"Further education", "Tolymesnis švietimas",
"GOTOX position", "GOTOX pozicija",
"Game show", "Žaidimų šou",
"Gardening", "Sodininkystė",
"General", "Bendra",
"General configuration", "Bendra konfigūracija",
"Generic", "Bendras",
"Generic options", "Bendros parinktys",
"Global Headers", "Visuotinės Antraštės",
"Global timer", "Visuotinis taimeris",
"Gray", "Pilka",
"HD: high definition", "HD: aukšta raiška",
"HDHomeRun ATSC-C frontend", "HDHomeRun ATSC-C sąsaja",
"HDHomeRun ATSC-T frontend", "HDHomeRun ATSC-T sąsaja",
"HDHomeRun DVB frontend", "HDHomeRun DVB sąsaja",
"HDHomeRun DVB-C frontend", "HDHomeRun DVB-C sąsaja",
"HDHomeRun DVB-T frontend", "HDHomeRun DVB-T sąsaja",
"HTSP Answer", "HTSP Atsakas",
"HTSP Default Stream Settings", "HTSP Numatytieji Srauto Nustatymai",
"HTSP Request", "HTSP Užklausa",
"HTSP Server", "HTSP Serveris",
"HTSP Stream Profile", "HTSP Srauto Profilis",
"HTSP Subscription", "HTSP Prenumerata",
"HTTP CORS origin", "HTTP CORS kilmė",
"HTTP Client", "HTTP Klientas",
"HTTP Server", "HTTP Serveris",
"Handicraft", "Amatai",
"Hearing impaired", "Neprigirdintiems",
"Hierarchy", "Hierarchija",
"High", "Aukštas",
"High bitrate mode (CI+ CAMs only)", "Aukšto bitų dažnumo režimas (tik CI+ CAM'ams)",
"Highest channel number the user can access.", "Aukščiausias kanalo numeris, kurį vartotojas gali pasiekti.",
"Highest count of objects.", "Didžiausias objektų skaičius.",
"Historical movie", "Istorinis filmas",
"Horror", "Siaubo",
"Hostname (or IP) of the server.", "Serverio hostname (arba IP)",
"How frequently it will re-try fetching an image that has failed to be fetched.", "Kaip dažnai bus bandoma pakartotinai parsiųsti paveikslėlį nesėkmės atveju.",
"How frequently the upstream provider is checked for changes.", "Kaip dažnai bus tikrinami pokyčiai tiekėjo pusėje.",
"ID of the network.", "Tinklo ID",
"IDLE", "TUŠČIAEIGIS",
"IGNORE", "IGNORUOTI",
"IP address", "IP adresas",
"IP blocking", "IP blokavimas",
"IPTV Automatic Network", "IPTV Automatinis Tinklas",
"IPTV Multiplex", "IPTV Multiplekseris",
"IPTV Network", "IPTV Tinklas",
"IPTV input", "IPTV įvestis",
"ISDB-C Network", "ISDB-C Tinklas",
"ISDB-C multiplex", "ISDB-C Multiplekseris",
"ISDB-S Network", "ISDB-S Tinklas",
"ISDB-S multiplex", "ISDB-S Multiplekseris",
"ISDB-T Network", "ISDB-T Tinklas",
"ISDB-T multiplex", "ISDB-T Multiplekseris",
"ISI (Stream ID)", "ISI (Srauto ID)",
"Icon", "Ikona",
"Icon (full URL)", "Ikona (pilnas URL)",
"Icon URL", "Ikonos URL",
"Icon base URL", "Ikonų šaltinio URL",
"Icon base URL.", "Ikonų šaltinio URL.",
"Icon has title", "Ikona turi antraštę",
"Icons from", "Ikonos iš",
"Idle", "Tuščiaeigis",
"Idle scan", "Tuščiaeigis skanavimas",
"Idle scan muxes", "Tuščiaeigis multiplekserių skanavimas",
"If an error occurs clone the scheduled entry and try to record again (if possible).", "Klaidos atveju klonuoti planuojamąjį įrašą ir bandyti įrašyti dar kartą (jei įmanoma).",
"If both a picon and a channel-specific (e.g. channelname.jpg) icon are defined, prefer the picon.", "Jei abu, pikona ir konkretaus kanalo ikona (pvz. kanalas.jpg) nurodyti, naudoti pikoną.",
"If more than x errors occur during a recording schedule a re-record (if possible).", "Jei įvyks daugiau nei nurodytas klaidų skaičius, suplanuoti pakartotinį įrašymą (jei įmanoma).",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Jei jokia vartotojo paskyra neegzistuoja, \nsukurkite vieną be jokio vartotojo vardo ir slaptažodžio.\nNaudokite tai su atsarga, nes tokiu būdu suteiksite viešą,\npasauliniu lygiu prieinamą administratoriaus prieigą prie\nsavo Tvheadend, kol sukursite naują ar neparedaguosite\nPrieigos Kontrolė esančio skirtuko naršyklėje.",
"If set, presentation of the tag icon will not superimpose the tag name on top of the icon.", "Jei pažymėta, žymos ikona nebus perkelta virš žymos pavadinimo.",
"If set, the first bytes from the MPEG-TS stream are discarded. It may be required for some drivers or hardware which do not flush the MPEG-TS buffers completely after a frequency/parameter change.", "Jei pažymėta, pirmieji MPEG-TS srauto bitai bus atmesti. Kai kurios tvarkyklės ar \"geležis\" gali to reikalauti, nes nesugeba pilnai nuleisti MPEG-TS buferio po dažnio/parametrų pokyčio.",
"If the DiseqC switch is located before the rotor (i.e. tuner - switch - rotor), enable this.", "Jei DiSEqC relė yra prieš variklį (pvz. tiuneris - relė - variklis), įjungite šią parinktį.",
"If the subscription weight for the input is below the specified threshold, the tuner is handled as free (according the priority settings). Otherwise, the next tuner (without any subscriptions) is used. Set this value to 10, if you are willing to override scan and epggrab subscriptions.", "Jei prenumeratos svoris įvesčiai yra žemesnis nei nurodytas slenkstis, tiuneris laikomas kaip laisvas (atsižvelgiant į pirmumo nustatymus), priešingu atveju sekantis tiuneris (be prenumeratų) bus naudojamas. Nustatykite reikšmę į 10, jei planuojate nepaisyti skanavimo ir EPG grabinimo prenumeratų.",
"If you'd like to support the project, please consider a donation.", "Jei pageidautumėte paremti projektą, prašome apsvarstyti pinigų paaukojimą.",
"Ignore", "Ignoruoti",
"Ignore EPG (EIT)", "Ignoruoti EPG (EIT)",
"Ignore descriptor 5", "Ignoruoti 5 deskriptorių",
"Ignore encryption flag, include encrypted services anyway.", "Ignoruoti šifravimo vėliavas. Betkuriuo atveju įtraukti šifruotus servisus.",
"Ignore invalid SSL certificate", "Ignoruoti negaliojantį SSL sertifikatą",
"Ignore invalid/unverifiable (expired, self-certified, etc.) certificates", "Ignoruoti negaliojančius/nepatikrinamus (pasibaigusius, savo pasirašytus ir t.t.) sertifikatus",
"Ignore provider's channel numbers", "Ignoruoti tiekėjo kanalų numerius",
"Image Cache", "Paveikslėlių podėlis",
"Immediately abort", "Atšaukti tučtuojau",
"Important", "Svarbu",
"Include channel name in filename", "Įtraukti kanalo pavadinimą į failo vardą",
"Include date in filename", "Įtraukti datą į failo vardą",
"Include episode in filename", "Įtraukti epizodą į failo vardą",
"Include subtitle in filename", "Įtraukti subpavadinimą į failo vardą",
"Include the date for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "Įtraukti datą į įvykio pavadinimo įrašą. Tai galioja abiems, faile talpinimam pavadinimui ir pačiam failo vardui.",
"Include the episode subtitle in the title (if available).", "Įtraukti epizodo subpavadinimą į pavadinimą (jei prieinama).",
"Include the name of the channel in the event title. This applies to both the title stored in the file and to the filename itself.", "Įtraukti kanalo pavadinimą į įvykio pavadinimą. Tai galioja abiems, Tai galioja abiems, faile talpinimam pavadinimui ir pačiam failo vardui.",
"Include the season and episode in the title (if available).", "Įtraukti sezono ir epizodo numerį į pavadinimą (jei prieinama).",
"Include the time for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "Įtraukti įrašo laiką į įvykio pavadinimą. Tai galioja abiems, Tai galioja abiems, faile talpinimam pavadinimui ir pačiam failo vardui.",
"Include time in filename", "Įtraukti laiką į failo vardą",
"Index", "Indeksas",
"Information area", "Informacijos sritis",
"Informational", "Informaciniai",
"Initial scan", "Pradinis skanavimas",
"Input base", "Įvesties pagrindas",
"Input buffer (bytes)", "Įvesties buferis (baitai)",
"Input instance", "Įvesties instancija",
"Input path", "Įvesties kelias",
"Interface", "Interfeisas",
"Internal", "Vidinis",
"Internal EPG grabber", "Vidinis EPG graberis",
"Internal PyEPG Grabber", "Vidinis PyEPG graberis",
"Internal XMLTV EPG Grabber", "Vidinis XMLTV EPG Graberis",
"Internal error", "Vidinė klaida",
"Internal grabber", "Vidinis graberis",
"Interview", "Interviu",
"Invalid", "Negaliojantis",
"Invalid EPG object request", "Negaliojanti EPG objekto užklausa",
"Invalid adapter number '%s'\n", "Negaliojantis adapterio numeris '%s'\n",
"Invalid arguments", "Negaliojantys argumentai",
"Invalid file", "Negaliojantis failas",
"Invalid parameters", "Negaliojantys parametrai",
"Invalid service", "Negaliojantis servisas",
"Invalid target", "Negaliojantis taikinys",
"Items", "Daiktai",
"Jazz", "Džiazas",
"Keep", "Laikyti",
"Keep x amount of storage space free.", "Laikyti nurodytą kiekė laisvos vietos.",
"Keepalive interval", "Žadinimo intervalas",
"Keepalive interval in seconds", "Žadinimo intervalas sekundėmis",
"LNA (low noise amplifier)", "LNA (žemo triukšmo stiprintuvas)",
"LNB type", "LNB tipas",
"Language", "Kalba",
"Language 1", "Kalba 1",
"Language 2", "Kalba 2",
"Language 3", "Kalba 3",
"Language settings", "Kalbos nustatymai",
"Languages", "Kalbos",
"Largest size the object has reached.", "Didžiausias objekto pasiektas dydis",
"Last played position", "Paskutinė grota pozicija",
"Last played position when the recording isn't fully watched yet.", "Paskutinė grota pozicija kada įrašas nėra pilnai peržiūrėtas.",
"Last seen", "Paskutinį kartą matyta",
"Last updated from", "Paskutinis atnaujinimas iš",
"Layer A: Constellation", "A sluoksnis: Konsteliacija",
"Layer A: FEC", "A sluoksnis: FEC",
"Layer A: Segment count", "A sluoksnis: Segmentų kiekis",
"Layer A: Time interleaving", "A sluoksnis: Laiko intarpai",
"Layer B: Constellation", "B sluoksnis: Konsteliacija",
"Layer B: FEC", "B sluoksnis: FEC",
"Layer B: Segment count", "B sluoksnis: Segmentų kiekis",
"Layer B: Time interleaving", "B sluoksnis: Laiko intarpai",
"Layer C: Constellation", "C sluoksnis: Konsteliacija",
"Layer C: FEC", "C sluoksnis: FEC",
"Layer C: Segment count", "C sluoksnis: Segmentų kiekis",
"Layer C: Time interleaving", "C sluoksnis: Laiko intarpai",
"Leisure hobbies", "Laisvalaikio hobiai",
"Limit connections", "Apriboti jungtis",
"Limited/limit scanning to this network ID only.", "Apribota/riboti skanavimą tik šiuo ID.",
"Linked input", "Susietą įvestis",
"Linux DVB frontend", "Linux DVB sąsaja",
"LinuxDVB Input", "LinuxDVB įvestis",
"LinuxDVB adapter", "LinuxDVB adapteris",
"LinuxDVB network", "LinuxDVB tinklas",
"List of allowed IPv4 or IPv6 hosts or networks (comma-separated).", "Leistinų IPv4 ar IPv6 įrenginių ar tinklų sąrašas (atskirtas kabutėmis).",
"List subsystems", "Pateikti subsistemas",
"Listen / Connect port", "Klausyti / prisijungti prie porto",
"Listen on IPv6", "Klausyti IPv6",
"Literature", "Literatūra",
"Local (server) time", "Vietinis (serverio) laikas",
"Local bind IP address", "Vietinis priskirtas IP adresas",
"Local channel minor", "Vietinio kanalo mažuma",
"Local channel number", "Vietinio kanalo numeris",
"Local discovery IP address", "Vietinio atradimo IP adresas",
"Local: Record if different description", "Vietinis: Įrašyti, jei skiriasi aprašymas",
"Local: Record if different episode number", "Vietinis: Įrašyti, jei skiriasi epizodo numeris",
"Local: Record if different subtitle", "Vietinis: Įrašyti, jei skiriasi subpavadinimas",
"Local: Record if different title", "Vietinis: Įrašyti, jei skiriasi pavadinimas",
"Local: Record once per day", "Vietinis: Įrašyti kartą per dieną",
"Local: Record once per week", "Vietinis: Įrašyti kartą per savaitę",
"Location", "Vieta",
"Location details of the SAT>IP Server.", "SAT>IP serverio vietos detalės",
"Lock preferred CA PID", "Užrakinti pageidautiną CA PID",
"Locking", "Užrakinimas",
"Log", "Žurnalas",
"Login password.", "Prisijungimo slaptažodis.",
"Login username.", "Prisijungimo vartotojo vardas.",
"Login/Logout", "Prisijungti/Atsijungti",
"Low", "Žemas",
"Lowest channel number the user can access.", "Žemiausias kanalo numeris, kuri vartotojas gali pasiekti",
"Lowest starting channel number.", "Žemiausias prasidedantis kanalo numeris.",
"MPEG-TS File", "MPEG-TS Failas",
"MPEG-TS Input Debug", "MPEG-TS Įvesties Debuginimas",
"MPEG-TS Multiplex", "MPEG-TS Multiplekseris",
"MPEG-TS Parser", "MPEG-TS Analizatorius",
"MPEG-TS Pass-thru/built-in", "MPEG-TS Pass-thru/integruotas",
"MPEG-TS Service", "MPEG-TS Servisas",
"MPEG-TS input", "MPEG-TS Įvestis",
"MPEG-TS multiplex PHY", "MPEG-TS multiplekserio PHY",
"MPEG-TS raw service", "MPEG-TS \"žaliasis\" servisas",
"Magazines", "Žurnalai",
"Main", "Pagrindinis",
"Maintain free storage space in MiB", "Išlaikyti laisvos vietos kiekį megabaitais",
"Maintain used storage space in MiB (0=disabled)", "Išlaikyti laisvos vietos kiekį megabaitais (0=išjungta)",
"Maintained space", "Išlaikyta vieta",
"Make subdirectories per channel", "Daryti subdirektorijas per kanalą",
"Make subdirectories per day", "Daryti subdirektorijas per dieną",
"Make subdirectories per title", "Daryti subdirektorijas per pavadinimą",
"Manage all", "Valdyti visus",
"Manufacturer", "Gamintojas",
"Manufacturer URL", "Gamintojo URL.",
"Manufacturer's model number.", "Gamintojo modelio numeris.",
"Manufacturer's product description.", "Gamintojo produkto aprašymas.",
"Manufacturer's product information page for the device.", "Gamintojo produkto informacijos puslapis įrenginiui.",
"Manufacturer's product name.", "Gamintojo produkto pavadinimas.",
"Map all services", "Susieti visus servisus",
"Map encrypted services", "Susieti užšifruotus servisus",
"Map radio channels", "Susieti radijo kanalus",
"Map unnamed channels", "Susieti neįvardintus kanalus",
"Map zero-numbered channels", "Susieti kanalus su nuliniu numeriu",
"Martial sports", "Kovos sportas",
"Master tuner", "Pagrindinis tiuneris",
"Matroska (mkv)/built-in", "Matroska (mkv)/integruotas",
"Matroska muxer", "Matroska mukseris",
"Matroska specific", "Matroska'ai būdinga",
"Maximal channel number", "Maksimalus kanalo numeris",
"Maximum # input streams", "Maksimalus įvesties srautų skaičius",
"Maximum PIDs", "Maksimalus PID skaičius",
"Maximum RAM size (MB)", "Maksimalus naudojamas RAM dydis (MB)",
"Maximum allowed difference between event start time when the EPG event is changed.", "Maksimalus leistinas skirtumas tarp įvykio ir EPG pradžios pokyčio.",
"Maximum bandwidth (Kbps)", "Maksimalus pralaidumas (Kbps)",
"Maximum count (0=default)", "Maksimalus kiekis (0=default)",
"Maximum duration", "Maksimali trukmė",
"Maximum input bandwidth.", "Maksimalus įvesties pralaidumas",
"Maximum length of PIDs", "Maksimalus PID'ų ilgis",
"Maximum period (mins)", "Maksimalus laikotarpis (minutės)",
"Maximum schedules limit (0=default)", "Maksimalus planuoklių kiekis (0=pagal nutylėjimą)",
"Maximum size (MB)", "Maksimalus dydis (MB)",
"Maximum time to wait (in seconds) for a stream before a timeout.", "Maksimalus laukimo laikas (sekundėmis) prieš srauto laikui pasibaigiant.",
"Maximum timeout (seconds)", "Maksimalus laiko pasibaigimas (sekundės)",
"Medicine", "Medicina",
"Memory Information", "Atminties Informacija",
"Merge same name", "Sujungti vienodu pavadinimu",
"Method not found", "Metodas nerastas",
"Minimal channel number", "Minimalus kanalo numeris",
"Minimum duration", "Minimali trukmė",
"Minimum rotor time (seconds)", "Minimalus variklio laikas (sekundės)",
"Missing In PAT/SDT", "Nerandama PAT/SDT",
"Mode", "Režimas",
"Model description", "Modelio aprašymas",
"Model name", "Modelio pavadinimas",
"Model number", "Modelio numeris",
"Modulation", "Moduliacija",
"Module", "Modulis",
"Module ID", "Mudolio ID",
"Module ID used to grab EPG data.", "Modulio ID naudojamas EPG grabinti.",
"Mon", "Pir",
"More verbose libav log", "Išsamesnis libav žurnalizavimas",
"Motor rate (in milliseconds/deg).", "Variklio kursas (sekundėmis/laipsniais).",
"Motor rate (milliseconds/deg)", "Variklio kursas (sekundėmis/laipsniais)",
"Motor sport", "Motorsportas",
"Motoring", "Automobiliai",
"Movie", "Filmas",
"Movie / drama", "Filmas / drama",
"Multiple lines of the cron time specification. The default cron triggers the internal grabbers daily at 12:04 and 00:04. See Help on how to define your own.", "Daugiaeilutinė Cron laiko specifikacija. Pagal nutylėjimą Cron įjungia oru gaunamų EPG graberius 12:04 ir 00:04 laiku. Žiūrėti \"Help\" pageidaujant sužinoti kaip pakeisti į sau priimtinus nustatymus.",
"Music", "Muzika",
"Music / Ballet / Dance", "Muzika / Baletas / Šokiai",
"Musical", "Miuziklai",
"Mux", "Multiplekseris",
"Mux Scheduler", "Multiplekserių planuoklis",
"Mux UUID", "Multiplekserio UUID",
"Mux handling", "Multiplekserio tvarkymas",
"Mux name", "Multiplekserio pavadinimas",
"Mux not enabled", "Multiplekseris neįjungtas",
"Muxer", "Mukseris",
"NONE", "JOKS",
"Name", "Pavadinimas",
"Name (or date) of program the entry is a rerun of.", "Įrašo gražintas pavadinimas (arba data).",
"Name of channel the entry recorded from.", "Įrašo įrašytas kanalo pavadinimas.",
"Name of object.", "Objekto pavadinimas.",
"Name of the bouquet.", "Buketo pavadinimas.",
"Name of the client.", "Kliento vardas.",
"Name of the entry.", "Įrašo pavadinimas.",
"Name of the module used to grab EPG data.", "Modulio, naudojamo EPG grabinti pavadinimas.",
"Name of the module, grabber or channel that should be used to update this channels EPG info.", "Modulio, graberio ar kanalo pavadinmas, kuris turėtų būti naudojamas atnaujinti šio kanalo EPG informaciją.",
"Name of the network.", "Tinklo pavadinimas.",
"Name of the profile.", "Profilio pavadinimas.",
"Name of the tag.", "Žymės pavadinimas.",
"Name of the tuner.", "Tiunerio vardas.",
"Name of the tuner/adapter.", "Tiunerio/adapterio pavadinimas.",
"Names", "Pavadinimai",
"Natural sciences", "Gamtos mokslai",
"Nature", "Gamta",
"Network", "Tinklas",
"Network ", "Tinklas",
"Network A", "Tinklas A",
"Network B", "Tinklas B",
"Network C", "Tinklas C",
"Network D", "Tinklas D",
"Network ID (limit scanning)", "Tinklo ID (riboti skanavimą)",
"Network UUID", "Tinklo UUID",
"Network access", "Tinklo prieiga",
"Network discovery", "Tinklo atradimas",
"Network for port A.", "Tinklas A portui.",
"Network for port AA.", "Tinklas AA portui.",
"Network for port AB.", "Tinklas AB portui.",
"Network for port B.", "Tinklas B portui.",
"Network for port BA.", "Tinklas BA portui.",
"Network for port BB.", "Tinklas BB portui.",
"Network for port C.", "Tinklas C portui.",
"Network for port D.", "Tinklas D portui.",
"Network group", "Tinklo grupė",
"Network limit per position", "Tinklo apribojimas per poziciją",
"Network name", "Tinklo pavadinimas",
"Network prefix", "Tinklo prefiksas",
"Network settings", "Tinklo nustatymai",
"Network type", "Tinklo tipas",
"Networks", "Tinklai",
"New / premiere / unknown", "Nauja / premjera / nežinoma",
"New media", "Nauja spauda",
"New muxes + changed muxes", "Nauji multiplekseriai + pakitę multiplekseriai",
"New muxes only", "Tik nauji multiplekseriai ",
"News", "Naujienos",
"News / Current Affairs", "Naujienos / Aktualijos",
"News magazine", "Naujienų žurnalas",
"Next tune delay in ms (0-2000)", "Sekančio tune uždelsimas milisekundėmis (0-2000)",
"No", "Ne",
"No PIN", "Jokio PIN",
"No access", "Jokios prieigos",
"No adapters specified!\n", "Nenurodyti jokie adapteriai!\n",
"No assigned adapters", "Nėra priskirtų adapterių",
"No days", "Jokią dieną",
"No descrambler", "Nėra iškoduotojo",
"No free adapter", "Nėra laisvo adapterio",
"No input detected", "Nerasta įvestis",
"No scheme", "Jokios schemos",
"No service assigned to channel", "Joks servisas nepriskirtas prie adapterio",
"No service enabled", "Nėra įjungtų servisų",
"No source available", "Nėra pasiekiamų šaltinių",
"Node subsystem", "Mazgo posistemė",
"None", "Joks",
"Normal", "Normalus",
"Not enough memory", "Nepakanka atminties",
"Not linked", "Nesusieta",
"Not set", "Nenustatyta",
"Not set (none or channel configuration)", "Nenustatyta (joks arba kanalo konfigūracija)",
"Not set (use channel or DVR configuration)", "Nenustatyta (naudokite kanalo arba DVR konfigūracija)",
"Number", "Numeris",
"Number of days to keep the file.", "Dienų skaičius kiek laikyti failą.",
"Number of days to keep the recorded file.", "Dienų skaičius išlaikyti įrašytą failą.",
"Number of days to keep the recorded files.", "Dienų skaičius išlaikyti įrašytus failus.",
"Number of days to retain entry information.", "Dienų skaičius išlaikyti įrašo informaciją.",
"Number of days to retain information about recordings. Once this period is exceeded, duplicate detection will not be possible anymore.", "Dienų skaičius išlaikyti informaciją apie įrašus. Šiam laikotarpiui pasibaigus dublikatų nustatymas nebegalimas.",
"Number of errors that occurred during recording (Transport errors).", "Nutikusių klaidų skaičius įrašymo metu (Transportavimo klaidos).",
"Number of errors that occurred during recording.", "Nutikusių klaidų skaičius įrašymo metu.",
"Number of muxes found.", "Rastų multiplekserių skaičius.",
"Number of services.", "Servisų skaičius.",
"Number of times this recording was played.", "Tiek kartų šis įrašas buvo grotas.",
"Number of tsfile tuners", "tsfile tiunerių skaičius",
"OK (partial)", "OK (dalinai)",
"Odd key", "Nelyginis raktas",
"Odd key.", "Nelyginis raktas.",
"Off", "Išjungta",
"Older OSCam", "Senesnis OSCam",
"On", "Įjungta",
"On file removal", "Failo šalinimo metu",
"On-demand (no first rewind)", "Pagal poreikį (jokio pradinio atsukimo)",
"Once per auto channel", "Kartą per auto kanalą",
"Only Bulsatcom 39E", "Tik Bulsatcom 39E",
"Only EIT", "Tik EIT",
"Only OpenTV Sky Ausat", "Tik OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Tik OpenTV Sky Italia",
"Only OpenTV Sky UK", "Tik OpenTV Sky UK",
"Only PSIP (ATSC)", "Tik PSIP (ATSC)",
"Only UK Freesat", "Tik UK Freesat",
"Only UK Freeview", "Tik UK Freeview",
"Only Viasat Baltic", "Tik Viasat Baltic",
"Only digits", "Tik skaičiai",
"Only preferred CA PID", "Tik pageidaujami CA PID",
"Options to use/used when mapping - see Help for details.", "Nuostatos naudojamos/naudoti susiejimo metu - papildomą informaciją rasite pagalbos skiltyje.",
"Options used when updating channels.", "Nuostatos naudojamos atnaujinant kanalus.",
"Orbital position", "Pozicija orbitoje",
"Orbital positions", "Pozicijos orbitoje",
"Orbital positions.", "Pozicijos orbitoje.",
"Original network ID", "Orginalus tinklo ID",
"Oscam mode.", "Oscam režimas.",
"Other Stream Filter", "Kitas Srauto Filtras",
"Output directory for tsdebug", "Išvesties direktorija tsdebug'inimui",
"Over-the-air", "Oru gaunamas",
"Over-the-air Cron multi-line", "Oru gaunamo Cron daugiaeilutė įvestis",
"Over-the-air EPG", "Oru gaunamas EPG",
"Over-the-air EPG grabber", "Oru gaunamo EPG graberis",
"Over-the-air grabbers", "Oru gaunamų graberiai",
"Override disabled", "Nepaisyti išjungimo",
"Override the frontend type.", "Nepaisyti sąsajos tipo.",
"Owner", "Savininkas",
"Owner of the entry.", "Įrašo savininkas.",
"Owner of the rule.", "Taisyklės savininkas.",
"PEND", "LAUKIAMA",
"PLS code", "PLS kodas",
"PLS mode", "PLS režimas",
"Parent entry", "Pirminis įrašas",
"Parent entry.", "Pirminis įrašas.",
"Pass subscription weight", "Perduoti prenumeratos svorį",
"Pass subscription weight to the SAT>IP server (Tvheadend specific extension).", "Perduoti svorį SAT>IP serveriui (išskirtinai Tvheadend plėtinys).",
"Pass-thru muxer", "Kiaurai praleidžiamas multiplekseris",
"Passthrough Muxer SI Tables", "Kiaurai praleidžiamo multiplekserio SI Lentelės",
"Password", "Slaptažodis",
"Password for the entry.", "Slaptažodis įrašui.",
"Password2", "Slaptažodis2",
"Path", "Kelias",
"Path to a directory (folder) containing your picon collection. See Help for more detailed information.", "Kelias iki direktorijos (katalogo) su Jūsų pikonų kolekcija.\nPapildomą informaciją rasite pagalbos skiltyje.",
"Path to an icon for this channel. This can be named however you wish, as either a local (file://) or remote (http://) image. See Help for more infomation.", "Kelias į ikoną šiam kanalui. Galima naudoti tiek vietinį (file://), tiek nuotolini (http://) paveikslėlį. Daugiau inormacijos rasite pagalbos skiltyje.",
"Path to the frontend used by the device.", "Kelias iki šio įrenginio naudojamos sąsajos.",
"Path to the grabber executable.", "Kelias iki graberio vykdomojo failo.",
"Path to the socket Tvheadend will read data from.", "Kelias iki socket'o, kurio informaciją nuskaitys Tvheadend.",
"Path to where the timeshift data will be stored. If nothing is specified this will default to CONF_DIR/timeshift/buffer.", "Kelias iki vietos, kur laiko poslinkio duomenys bus saugomi. Jei kelias nenurodytas, CONF_DIR/timeshift.buffer bus naudojama.",
"Path used by the device.", "Šio įrenginio naudojamas kelias.",
"Path where the recordings are stored. If components of the path do not exist, Tvheadend will try to create them.", "Kelias iki vietos, kur saugomi visi įrašai. Jei kelio komponentai neegzistuoja, Tvheadend bandys juos sukurti.",
"Peak count of objects", "Didžiausias objektų skaičius",
"Peak size", "Didžiausias dydis",
"Performing arts", "Vaizduojamasis menas",
"Periodically save EPG to disk (hours)", "Periodiškai saugoti EPG į diską (valandos)",
"Persistent user interface level", "Nuolatinis vartotojo interfeiso lygis",
"Physiology", "Psichologija",
"Pick the stream with given audio type only.", "Pasirinkti srautą tik su duotuoju garso tipu.",
"Picon", "Pikona",
"Picon name scheme", "Pikonų įvardinimo schema",
"Picon path", "Pikonų kelias",
"Pilot", "Pilotas",
"Polarization", "Poliarizacija",
"Political issues", "Politinės problemos",
"Poll multiplexer", "Apklausti multiplekserį",
"Popular culture", "Populiarioji kultūra",
"Port", "Portas",
"Port to connect to.", "Portas, prie kurio prisijungti.",
"Position", "Pozicija",
"Position of the input.", "Įvesties pozicija.",
"Post-processor command", "Po veiksmo atliekama komanda",
"Post-recording padding", "Pratęsti įrašą",
"Post-remove command", "Komanda po pašalinimo",
"Power save", "Energijos taupymas",
"Power-up time (ms) (15-200)", "Užsikurimo laikas (ms) (15-200)",
"Pre-defined muxes", "Iš anksto apibrėti multiplekseriai",
"Pre-processor command", "Prieš veiksmą atliekama komanda",
"Pre-recording padding", "Paankstinti įrašą",
"Pre-school children's programs", "Priešmokyklinukų programos",
"Prefer picons over channel name", "Teikti pirmenybę pikonoms vietoje kanalo pavadinimo",
"Preferred CA PID", "Pageidautinas CA ID",
"Preferred audio language.", "Pageidautina audio kalba.",
"Preferred service video type", "Pageidautinas serviso video tipas",
"Presentation", "Prezentacija",
"Presentation details.", "Prezentacijos detalės.",
"Press", "Spauda",
"Prevent the user from overriding the default user interface level setting and removes the view level drop-dowm from the interface.", "Neleisti vartotojui nepaisyti vartotojo interfeiso lygio nustatymų ir pašalina interfeiso lygio pasirinkimo nuostatą.",
"Priority", "Pirmumas",
"Priority (-10..10)", "Prioritetas (-10..10)",
"Private", "Privatus",
"Profile name", "Profilio pavadinimas",
"Program synopsis (display only).", "Trumpa programos apžvalga (tik rodomas).",
"Program synopsis.", "Trumpa programos apžvalga.",
"Provider", "Tiekėjas",
"Provider ID", "Tiekėjo ID",
"Provider network name", "Tiekėjo tinklo pavadinimas",
"Provider's network name.", "Tiekėjo tinklo pavadinimas.",
"Psychology", "Psichologija",
"Puppets", "Lėlės",
"PyEPG Import", "PyEPG Importas",
"Quiz", "Viktorina",
"RAM only", "Tik RAM",
"RTP/AVP/TCP (embedded data)", "RTP/AVP/TCP (įterptiniai duomenys)",
"RTSP Protocol", "RTSP Protokolas",
"RTSP port", "RTSP Portas",
"RTSP port (554 or 9983, 0 = disable)", "RTSP portas (554 arba 9983, 0 = išjungta)",
"Radio", "Radijas",
"Raw Audio Stream", "Nepakeistas Audio Srautas",
"Re-fetch period (hours)", "Pakartotinio parsiuntimo laikotarpis (valandos)",
"Re-fetch period (mins)", "Pakartotinio parsiuntimo laikotarpis (minutės)",
"Re-fetch the bouquet every x minutes.", "Pakartotinai parsiųsti buketą kas x minučių.",
"Re-record%s%s", "Pakartotinai įrašyti%s%s",
"Re-try period (hours)", "Pakartotinio bandymo laikotarpis (valandos)",
"Read error", "Nuskaitymo klaida",
"Real Time Streaming Protocol (RTSP) port the server should listen on (554 or 9983, 0 = disable).", "Realaus Laiko Transliavimo Protokolo (RSTP) portas, kurio serveris turėtų klausytis (554 arba 9983, 0 = išjungta).",
"Record all", "Įrašyti viską",
"Record on these days only.", "Įrašinėti tik šiomis dienomis.",
"Recording file options", "Įrašomo failo nuostatos",
"Recording file size.", "Įrašomo failo dydis.",
"Recording play count", "Įrašo peržiūrų kiekis",
"Recording system path", "Įrašinėjimo sistemos kelias",
"Reject", "Atmesti",
"Reject exact match", "Atmesti visišką atitikmenį",
"Relative path to the imagecache copy of the icon.", "Salyginis kelias iki ikonos kopijos paveikslėlių podėlyje.",
"Religion", "Religijos",
"Religious", "Religinga",
"Remarkable people", "Nepaprasti žmonės",
"Remove HTTP arguments", "Pašalinti HTTP argumentus",
"Remove all unsafe characters from filename", "Pašalinti visus nesaugius simbolius iš failo pavadinimo",
"Repeated", "Pakartota",
"Replace whitespace in title with '-'", "Pakeisti tarpus pavadinime brūkšniais",
"Replaces all whitespace in the title with '-'.", "Pakeisti visus tarpus pavadinime brūkšniais.",
"Reply to CAM PIN inquiries", "Atsakyti į CAM PIN pasiteiravimus",
"Reply to PIN inquiries.", "Atsakyti į PIN pasiteiravimus.",
"Reports", "Ataskaitos",
"Rerun of", "Pakartotas",
"Reserved", "Rezervuota",
"Resolution (height)", "Rezoliucija (aukštis)",
"Restart", "Perkrauti",
"Restart on error", "Perkrauti esant klaidai",
"Restart streaming on error. This is useful for DVR so a recording isn't aborted if an error occurs.", "Perkrauti transliavimą klaidos atveju. Tai naudinga naudojant DVR, nes įrašas nenutraukiamas klaidos atveju.",
"Restrict connections to this type.", "Apriboti šio tipo jungtis.",
"Reuse EPG from", "Panaudoti EPG iš",
"Reuse the EPG from another channel.", "Panaudoti EPG iš kito kanalo.",
"Rewrite EIT", "Perrašyti EIT",
"Rewrite PAT", "Perrašyti PAT",
"Rewrite PMT", "Perrašyti PMT",
"Rewrite SDT", "Perrašyti SDT",
"Rights", "Teisės",
"Rock", "Rokas",
"Romance", "Romantika",
"Rotor initialization time (seconds)", "Rotoriaus pasiruošimo laikas (sekundės)",
"Rotor type", "Rotoriaus tipas",
"Run as group", "Paleisti kaip grupę",
"Run as user", "Paleisti kaip vartotoją",
"Running", "Veikia",
"SAT->IP frontend number.", "SAT>IP sąsajos numeris.",
"SAT>IP ATSC-C Frontend", "SAT>IP ATSC-C Sąsaja",
"SAT>IP ATSC-T Frontend", "SAT>IP ATSC-T Sąsaja",
"SAT>IP Client", "SAT>IP Klientas",
"SAT>IP DVB Frontend", "SAT>IP DVB Sąsaja",
"SAT>IP DVB-C Frontend", "SAT>IP DVB-C Sąsaja",
"SAT>IP DVB-S Frontend", "SAT>IP DVB-S Sąsaja",
"SAT>IP DVB-S Slave Frontend", "SAT>IP DVB-S Dukterinė Sąsaja",
"SAT>IP DVB-T Frontend", "SAT>IP DVB-T Sąsaja",
"SAT>IP DVB-T frequency (Hz)", "SAT>IP DVB-T dažnis (Hz)",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "SAT>IP RTSP porto numeris serveriui\n(numatytasis: -1 = išjungta, 0 = webconfig, standartis portas yra 554)",
"SAT>IP Satellite Configuration", "SAT>IP Palydovų Konfigūracija",
"SAT>IP Server", "SAT>IP Serveris",
"SAT>IP Server SI Tables", "SAT>IP Serverio SI Lentelės",
"SAT>IP source number", "SAT>IP šaltinio numeris",
"SD: standard definition", "SD: standartinė raiška",
"SSL verify peer", "SSL patikrinti prisijungusįjį",
"START", "PRADĖTI",
"STOP", "SUSTABDYTI",
"Sat", "Šeš",
"Satellite config", "Palydovų konfigūracija",
"Satellite longitude", "Palydovo ilguma",
"Satellite longitude.", "Palydovo ilguma.",
"Satellite position.", "Palydovo pozicija.",
"Satellite positions", "Palydovų pozicijos",
"Scan after creation", "Skanuoti sukūrus",
"Scan progress", "Skanavimo eiga",
"Scan queue length", "Skanavimo eilės ilgis",
"Scan result", "Skanavimo rezultatai",
"Scan status", "Skanavimo būsena",
"Scanfile", "Skanavimo failas",
"Schedule frequency (in cron format).", "Planuoklio dažnumas (Cron formatu).",
"Schedule status", "Planuoklio būsena",
"Schedule status.", "Planuoklio būsena.",
"Scheduled Duration", "Suplanuota Trukmė",
"Scheduled for recording", "Suplanuota įrašyti",
"Scheduled start time", "Suplanuotas pradžios laikas",
"Scheduled stop time", "Suplanuotas pabaigos laikas",
"School programs", "Mokyklos programos",
"Science", "Mokslas",
"Science fiction", "Mokslinė fantastika",
"Season", "Sezonas",
"Season %d", "Sezonas %d",
"Season information (if available).", "Sezono informacija (jei prieinama).",
"Seek error", "Ieškoti klaidos",
"Select an available network type for this tuner.", "Pasirinkti prieinamą tinklo tipą šiam tiuneriui.",
"Select high priority (default) EPG language.", "Pasirinkti aukšto prioriteto (numatytą) EPG kalbą.",
"Select how Tvheadend should handle muxes. See Help for details.", "Pasrinkite kaip Tvheadend turėtų tvarkytis su multiplekseriais. Daugiau informacijos rasite pagalbos skiltyje.",
"Select low priority EPG language.", "Pasirinkti žemo prioriteto EPG kalbą.",
"Select medium priority EPG language.", "Pasirinkti vidutinio prioriteto EPG kalbą.",
"Send CAPMT OK query before descrambling.", "Siųsti CAPMT OK prieš iškoduojant",
"Send CAPMT query", "Siųsti CAPMT užklausą",
"Send full PLAY cmd", "Siųsti pilną PLAY komandą",
"Serial number", "Serijinis numeris",
"Series link", "Serijos nuoroda",
"Series link ID.", "Serijos nuorodos ID.",
"Serious", "Rimta",
"Serious music", "Rimta muzika",
"Server", "Serveris",
"Server UUID", "Serverio UUID",
"Server connectivity", "Serverio jungiamumas",
"Server details.", "Serverio detalės.",
"Service", "Servisai",
"Service ID", "Serviso ID",
"Service Mapper", "Servisų Susiejiklis",
"Service Mapping (Map services to channels)", "Servisų Susiejimas (Servisų susiejimas su kanalais)",
"Service configuration", "Servisų konfigūracija",
"Service count %zi", "Servisų kiekis %zi",
"Service mapping", "Servisų susiejimas",
"Service name", "Serviso pavadinimas",
"Service name found in EPG data.", "Serviso pavadinimas rastas EPG duomenyse.",
"Service name picons", "Servisų pavadinimų pikonos",
"Service type", "Serviso tipas",
"Service type override. This value will override the service type provided by the stream.", "Nepaisyti serviso tipo. Ši reikšmė nepaisys srauto pateikto serviso tipo.",
"Services", "Servisai",
"Services associated with the channel.", "Servisai susieti su kanalu.",
"Set the display name.", "Nustatyti atvaizduojamą vardą.",
"Set the maxiumum packet identifiers your SAT>IP server supports.", "Nustatyti maksimalų paketų identifikatorių kiekį, kurį palaiko Jūsų SAT>IP serveris.",
"Set the priority of this configuration.", "Nustatyti šios konfigūracijos prioritetą.",
"Settings", "Nustatymai",
"Show", "Šou",
"Show / Game show", "Šou / Žaidimų šou",
"Show this page", "Rodyti šį puslapį",
"Show version information", "Rodyti versijos informaciją",
"Signal quality too poor", "Signalo kokybė perprasta ",
"Signal scale (240 or 100)", "Signalo skalė (240 arba 100)",
"Site latitude", "Vietos platuma",
"Site latitude.", "Vietos platuma.",
"Site longitude", "Vietos ilguma",
"Site longitude.", "Vietos ilguma.",
"Size", "Dydis",
"Skip TS packets (0-200)", "Praleisti TS paketus (0-200)",
"Skip commercials", "Praleisti reklamas",
"Skip initial bytes", "Praleisti pradinius baitus",
"Slave entry", "Pavaldinio įrašas",
"Slave entry.", "Pavaldinio įrašas.",
"Slot state", "Angos būsena",
"Soap", "Muilo opera",
"Soccer", "Futbolas",
"Social", "Socialinis",
"Social / Political issues / Economics", "Socialinis / Politinės problemos / Ekonomika",
"Sort index", "Rūšiavimo Indeksas",
"Sort index.", "Rūšiavimo Indeksas.",
"Source", "Šaltinis",
"Source deleted", "Šaltinis ištrintas",
"Source quality is bad", "Šaltinio kokybė yra bloga",
"Source reconfigured", "Šaltinis perkonfigūruotas",
"Special events (Olympic Games, World Cup, etc.)", "Ypatingi įvykiai (Olimpinės žaidynės, Pasaulio taurė ir t.t.)",
"Specify alternative htsp port", "Nurodykite alternatyvų HTSP portą",
"Specify alternative http port", "Nurodykite alternatyvų HTTP portą",
"Specify alternative http webroot", "Nurodykite alternatyvią HTTP šaknį",
"Specify bind address", "Nurodykite priskiriamą adresą",
"Specify bind address for SAT>IP server", "Nurodykite SAT>IP serveriui priskiriamą adresą",
"Specify extra htsp port", "Nurodykite papildomą HTSP portą",
"Specify the parameters to be changed. See Help for details.", "Nurodykite parametrus pakeitimui. Daugiau inofrmacijos rasite pagalbos skiltyje.",
"Spiritual sciences", "Dvasiniai mokslai",
"Sports", "Sportas",
"Sports magazines", "Sporto žurnalai",
"Standard", "Standartinis",
"Start", "Pradėti",
"Start after", "Pradėti po",
"Start before", "Pradėti prieš",
"Start time", "Pradžios laikas",
"Status", "Būsena",
"Status period (ms)", "Būsenos laikotarpis (ms)",
"Stop", "Stabdyti",
"Stop time", "Stabdymo laikas",
"Storage path", "Laikemos kelias",
"Storage space", "Laikmenos vieta",
"Stream ID", "Srauto ID",
"Stream Profile", "Srauto Profilis",
"Stream index", "Srauto Indeksas",
"Stream index (starts with zero).", "Srauto Indeksas (prasideda nuliu)",
"Stream profile", "Srauto profilis",
"Stream setup error", "Srauto parametrų klaida",
"Stream type", "Srauto tipas",
"Streaming", "Transliavimas",
"Streaming Profile", "Transliavimo Profilis",
"Streaming flags, allow/disallow HTTP streaming, advanced HTTP streaming (e.g, direct service or mux links), HTSP protocol streaming (e.g, Kodi (via pvr.hts) or Movian.", "Transliavimo vėliavos, leisti/neleisti HTTP transliavimą, pažangus HTTP transliavimas (pvz. tiesioginiai servisai ar multiplekserių siejiniai), transliavimas HTSP protokolu pvz. Kodi (per pvr.hts) ar Movian.",
"Streaming priority", "Srauto pirmumas",
"Streaming profiles", "Srautų profiliai",
"Subdirectory options", "Subdirektorijos nuostatos",
"Subscribe to a service permanently", "Prenumeruoti servisą nuolat",
"Subscription", "Prenumerata",
"Subscription does not exist", "Prenumerata neegzistuoja",
"Subscription overridden", "Nepaisoma prenumeratos",
"Subscription weight", "Prenumeratos svoris",
"Substitute formatters", "Keisti apipavidalintojus",
"Subtitle", "Titrai",
"Subtitle Stream Filter", "Titrų Srauto Filtras",
"Subtitle codec", "Titrų kodekas",
"Sun", "Sek",
"Surround (2 front, rear mono)", "Erdvinis (2 prieky, gale mono)",
"Switch before rotor", "Perjungti prieš rotorių",
"Switch off the power to the LNB when idle. Note: this may cause interference with other devices when the LNB is powered back up.", "Išjungti LNB maitinimą neveiksnumo būsenoje. Pastaba: tai gali sutrikdyti kitų įrenginių darbą kai LNB vėl įjungiamas.",
"Switch type", "Komutatoriaus tipas",
"Switch: %s", "Komutatorius: %s",
"Symbol rate (Sym/s)", "Simbolių rodiklis (Sym/s)",
"Sync", "Sinchronizuoti",
"Sync + Don't keep", "Sinchronizuoti + Nelaikyti",
"System", "Sistema",
"TCP Protocol", "TCP Protokolas",
"TV Adapters - Linux ATSC-C Frontend", "TV Adapteriai - Linux ATSC-C Sąsaja",
"TV Adapters - Linux ATSC-T Frontend", "TV Adapteriai - Linux ATSC-T Sąsaja",
"TV Adapters - Linux DAB Frontend", "TV Adapteriai - Linux DAB Sąsaja",
"TV Adapters - Linux DVB-C Frontend", "TV Adapteriai - Linux DVB-C Sąsaja",
"TV Adapters - Linux DVB-T Frontend", "DVB-C",
"TV Adapters - Linux ISDB-C Frontend", "TV Adapteriai - Linux ISDB-C Sąsaja",
"TV Adapters - Linux ISDB-T Frontend", "TV Adapteriai - Linux  ISDB-T Sąsaja",
"TVHDHomeRun Client", "TVHDHomeRun Klientas",
"Tag files with metadata", "Žymėti failus meta duomenimis",
"Tags", "Žymės",
"Tags linked/to link to the channel.", "Žymės susietos su kanalu.",
"Talk show", "Pokalbių šou",
"Team sports (excluding football)", "Komandinis sportas (išskyrus futbolą)",
"Technology", "Technologijos",
"Teletext Stream Filter", "Teleteksto Srauto Filtras",
"Teletext stream type is only available for this filter.", "Tik teleteksto srauto tipas yra prieinamas šiam filtrui.",
"Tennis", "Tenisas",
"Testing", "Testuojama",
"Testing options", "Testavimo parinktys",
"The ATSC source ID as set by the provider.", "ATSC šaltinio ID kaip nustatyta tiekėjo.",
"The CA provider to compare. Leave blank to apply to all providers.", "CA tiekėjas palyginimui. Palikus tuščią, bus priskiriama visiems tiekėjams.",
"The CA stream type is only available for this filter.", "Šiam filtrui prieinamas tik CA srauto tipas .",
"The CAID to compare. Leave blank to apply to all IDs.", "CAID palyginimui. Pageidaujant pritaikyti visiems ID, palikite tuščią.",
"The CAM slot status.", "CAM lizdo būsena.",
"The COFDM modulation used by the mux. If you're not sure of the value leave as AUTO.", "Multiplekserio naudojama COFDM moduliacija. Jei nesate tikri dėl reikšmės, palikite AUTO.",
"The Conditional Access ID used for the service.", "Serviso naudojamas CS ID.",
"The DVR profile to be used/used by the recording.", "DVR profilį naudoti/naudojmas įrašams.",
"The DVR profile to be used/used by this rule.", "Šios taisyklės naudojmas/naudoti DVR profilis.",
"The EPG ID used by the entry.", "Šio įrašo naudojamas EPG ID.",
"The EPG grabber name.", "EPG graberio pavadinimas.",
"The EPG grabber to use on the mux. Enable (auto) is the recommended value.", "Naudoti EPG graberį multiplekseriui. Auto nustatymas yra rekomenduojama reikšmė.",
"The EPG grabber type.", "EPG graberio tipas",
"The MPEGTS stream type is only available for this filter.", "Tik MPEGTS srauto tipos yra prieinamas šiam filtrui.",
"The OpenTV channel number as set by the provider.", "Kanalo numeris kaip numatyta OpenTV tiekėjo.",
"The PIN to use.", "Naudoti PIN.",
"The Preferred Conditional Access Packet Identifier. Used for decrypting scrambled streams.", "Pageidautinas CA Paketo Identifikatorius. Naudojamas iššifruoti šifruotus srautus.",
"The SAT>IP server's name.", "SAT>IP serverio vardas.",
"The SAT>IP server's universally unique identifier.", "SAT>IP serverio universalus unikalus identifikatorius.",
"The SAT>IP source number.", "SAT>IP šaltinio numeris.",
"The SAT>IP's discovered IP address.", "SAT>IP atrastas IP adresas.",
"The URL to the playlist.", "M3U grojaraščio URL.",
"The audio stream types the filter should apply to.", "Audio sratų tipai, kuriems filtras turėtų būti pritaikytas.",
"The bouquet the channel is associated with.", "Buketas, su kuriuo kanalas yra susietas.",
"The channel this service is mapped to.", "Kanalas, su kuriuo šis servisas yra susietas.",
"The character encoding for this network (e.g. UTF-8).", "Simbolių rinkinys naudojamas šiam tinklui (pvz. UTF-8).",
"The character encoding for this service (e.g. UTF-8).", "Simbolių rinkinys naudojamas šiam servisui (pvz. UTF-8).",
"The current configuration ID.", "Dabartinės konfigūracijos ID.",
"The current configuration version.", "Dabartinės konfigūracijos versija.",
"The device ID.", "Įrenginio ID.",
"The device's serial number.", "Įrenginio serijinis numeris.",
"The name of the profile.", "Profilio pavadinimas",
"The name of the the rule.", "Taisyklės pavadinimas.",
"The network prefix(es) to block, e.g.192.168.2.0/24 (comma-separated list).", "Blokuoti tinklo prefiksą(-us), pvz.: 192.168.2.0/24 (kableliais atskirtas sąrašas).",
"The network the mux is on.", "Tinklo multiplekseris įjungtas.",
"The network the service is on.", "Tinklo servisas yra įjungtas.",
"The network's service ID", "Tinklo serviso ID",
"The networks assigned to the device.", "Šiam įrenginiui priskirti tinklai.",
"The networks using this configuration.", "Tinklai naudojantys šią konfigūraciją.",
"The networks' universally unique identifier (UUID).", "Tinklo universaliai unikalus identifikatorius (UUID).",
"The number of muxes left to scan on this network.", "Šiame tinkle likusių skanuoti multiplekserių skaičius.",
"The provider's ID.", "Tiekėjo ID.",
"The provider's name.", "Tiekėjo pavadinimas.",
"The provider's network ID.", "Tiekėjo tinklo ID.",
"The provider's network name.", "Tiekėjo tinklo pavadinimas.",
"The recording/entry status.", "Įrašymo/įrašo būsena.",
"The service ID", "Serviso ID",
"The service ID as set by the provider.", "Serviso ID kaip nustatyta tiekėjo.",
"The service ID.", "Serviso ID.",
"Thriller", "Trileris",
"Thu", "Ket",
"Time", "Laikas",
"Time (in milliseconds) for a command to complete.", "Laikas (milisekundėmis) komandai užbaigti.",
"Time (in milliseconds) for the rotor to power up.", "Laikas (milisekundėmis) rotoriui įsijungti.",
"Time (in minutes) to re-fetch the playlist.", "Laikas (minutėmis) pakartotinai atsiųsti grojaraštį.",
"Time Stamp Fix", "Laiko Spaudo Pataisa",
"Time missed", "Laikas praleistas",
"Time record caption", "Laiko įrašo antraštė",
"Time recording%s%s", "%s%s laiko įrašinėjama",
"Time to start the recording/time the recording started.", "Laikas pradėti įrašinėti/laikas kada įrašymas pradėtas.",
"Time to stop recording/time the recording stopped.", "Laikas nustoti įrašinėti/laikas kada įrašymas baigėsi.",
"Time update", "Laiko atnaujinimas",
"Timeout (sec) (0=infinite)", "Pasibaigsiantis liakas (timeout) (sekundėmis, 0=amžinas)",
"Timeout (seconds)", "Pasibaigsiantis liakas (sekundės)",
"Timeout (secs)", "Pasibaigsiantis liakas (sekundės)",
"Timeshift", "Laiko poslinkis",
"Title", "Pavadinimas",
"Title (regexp)", "Pavadinimas (regexp)",
"Title of the program (display only).", "Programos pavadinimas (tik rodymui)",
"Title of the program.", "Programos pavadinimas.",
"Title of the recording - this is used to generate the filename.", "Įrašo pavadinimas - tai naudojama failo vardo generavimui.",
"Toggle details", "Perjungti detales",
"Too many data errors", "Perdaug duomenų klaidų",
"Total number of mapped channels on this network.", "Visas su šiuo tinklu susietų kanalų skaičius.",
"Total number of muxes found on this network.", "Visas šiame tinkle surastų multiplekserių skaičius.",
"Total number of services found on this network.", "Visas šiame tinkle surastų servisų skaičius.",
"Total number of services found.", "Visas surastų servisų skaičius.",
"Total number of services seen.", "Visas matytų servisų skaičius.",
"Total number of services.", "Visas servisų skaičius.",
"Tourism / Travel", "Turizmas / Kelionės",
"Traditional arts", "Tradiciniai menai",
"Traditional music", "Tradicinė muzika",
"Transcode", "Perkoduoti",
"Transcode/av-lib", "Perkodavimas/av-lib",
"Transcoding", "Perkodavimas",
"Transmission mode", "Perdavimo režimas",
"Transponder ID", "Transponderio ID",
"Transport Stream", "Transportavimo Srautas",
"Transport stream ID", "Transportavimo srauto ID",
"Tue", "Ant",
"Tuner", "Tiuneris",
"Tuner bind IP address", "Tiuneriui priskirtas IP adresas",
"Tuner configuration", "Tiunerio konfigūracija",
"Tuner configuration.", "Tiunerio konfigūracija.",
"Turn off LNB when idle", "Išjungti LNB neveiksnumo būsenoje",
"Tvheadend server name", "Tvheadend serverio vardas",
"Type", "Tipas",
"Type override", "Nepaisyti tipo",
"UDP RTP port number (2 ports)", "UDP RTP porto numeris (2 portai)",
"UHD: ultra high definition", "UHD: ultra aukšta raiška",
"UPnP Protocol", "UPnP Protokolas",
"URL of the M3U playlist.", "M3U grojaraščio URL.",
"Unable to get system UTC time", "Nepavyksta gauti sistemos UTC laiko",
"Unable to get system local time", "Nepavyksta gauti sistemos vietinio laiko",
"Unable to get system time", "Nepavyksta gauti sistemos laiko",
"Unable to open file", "Nepavyksta atidaryti failo",
"Uncommitted", "Nepriskirta",
"Uncommitted first", "Nepriskirti pirmiausia",
"Unicable I switch (universal LNB)", "Unicable I komutatorius (universalus LNB)",
"Unicable II switch (universal LNB)", "Unicable II komutatorius (universalus LNB)",
"Unicable type", "Unicable tipas",
"Unimportant", "Nesvarbus",
"Universal LNB only", "Tik universalus LNB",
"Unknown", "Nežinoma",
"Unknown file", "Nežinomas failas",
"Unknown reason (%i)", "Nežinoma priežastis (%i)",
"Unlimited size", "Neribotas dydis",
"Unlimited time", "Neribotas laikas",
"Unset (default)", "Nenustatyta (pagal nutylėjimą)",
"Update card (EMM)", "Atnaujinti kortelę (EMM)",
"Update channel icon", "Atnaujinti kanalo ikoną",
"Update channel name", "Atnaujinti kanalo pavadinimą",
"Update channel number", "Atnaujinti kanalo numerį",
"Update disabled", "Atnaujinimas išjungtas",
"Update time", "Atnaujinimo laikas",
"Update tolerance (ms)", "Atnaujinimo tolerancija (ms)",
"Updated", "Atnaujinta",
"Usage: %s [OPTIONS]\n", "Panaudojimas: %s [NUOSTATOS]\n",
"Use HTTP digest authentication", "Naudoti HTTP digest autentifikavimą",
"Use WEBM format.", "Naudoti WEBM formatą.",
"Use Windows-compatible filenames", "Su Windows suderinami failų vardai",
"Use original", "Naudoti orginalą",
"Use service IDs as channel numbers", "Naudoti serviso ID kaip kanalų numerius",
"Use tag internally (don't expose to clients).", "Naudoti žymą tik vidiniam naudojimui (neperteikti klientams).",
"Use the provider's service IDs as channel numbers.", "Naudoti tiekėjo serviso ID kaip kanalų numerius.",
"Use this character set when setting filenames.", "Naudoti šį simbolių rinkinį nustatant failų vardus.",
"Use x amount of storage space.", "Naudoti x kiekį disko vietos.",
"User access error", "Vartotojo prieigos klaida",
"User does not have access", "Vartotojas neturi prieigos",
"User does not have access to channel", "Vartotojas neturi prieigos prie kanalo",
"User icon", "Vartotojo ikona",
"User interface level", "Vartotojo interfeiso lygis",
"User language", "Vartotojo kalba",
"User limit reached", "Pasiektas vartotojo limitas",
"User login", "Vartotojo prisijungimas",
"User request", "Vartotojo užklausa",
"Username", "Vartotojo vardas",
"Username for the entry (login username).", "Vartojo vardas įrašui (prisijungimo vartotojo vardas).",
"Users - Access Entries", "Vartotojai - Prieigos Įrašai",
"Users - IP Blocking", "Vartotojai - IP blokavimas",
"Users - Passwords", "Vartotojai - Slaptažodžiai",
"Variety show", "Estradinės",
"Verify the SSL certificate.", "Patikrinti SSL sertifikatą.",
"Vertical resolution (height) of the output video stream. Horizontal resolution is adjusted automatically to preserve aspect ratio. When set to 0, the input resolution is used.", "Išvesties srauto vertikali rezoliucija (aukštis). Horizontali rezoliucija parenkama automatiškai, siekiant išsaugoti kraštinių santykį. Jei parinkta 0-is, naudojama įvesties rezoliucija.",
"Video Stream Filter", "Video Srauto Filtras",
"Video bitrate (kb/s) (0=auto)", "Video bitų dažnis (kb/s) (0=auto)",
"Video codec", "Video kodekas",
"Video recorder", "Vaizdo magnetofonas",
"View all", "Rodyti viską",
"Visually impaired commentary/audio description", "Vaizdiškai papildyti komentarai/audio apibūdinimai",
"WEBM/built-in", "WEBM/integruotas",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV profilis H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV profilis H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV profilis VP8/Vorbis/WEBM",
"Waiting for EPG running flag", "Laukiama EPG paleisties vėliavos",
"Waiting for program start", "Laukiama programos pradžios",
"Waiting for stream", "Laukiama srauto",
"War", "Karas",
"Water sport", "Vandens sportas",
"Weak stream", "Silpnas srautas",
"Weather report", "Orai",
"Web User Interface", "Naršyklinis Vartotojo Interfeisas",
"Web interface", "Naršyklinis interfeisas",
"Web interface language", "Naršyklinio interfeiso kalba",
"Web interface language.", "Naršyklinio interfeiso kalba",
"Web interface theme.", "Naršyklinio interfeiso tema.",
"Web theme", "Naršyklės tema",
"Web user interface", "Naršyklinis vartotojo interfeisas",
"Wed", "Tre",
"Welcome - Tvheadend - your TV streaming server and video recorder", "Sveiki - Tvheadend - Jūsų TV transliavimo serveris ir vaizdo magnetofonas",
"Western", "Vesternas",
"Winter sports", "Žiemos sportas",
"Wizard", "Vedlys",
"XMLTV EPG Import", "XMLTV EPG Importas",
"Yes", "Taip",
"fast: h264 / h265 / qsv(h264 / h265)", "greitas: h264 / h265 / qsv(h264 / h265)",
"faster: h264 / h265 / qsv(h264)", "greitesnis: h264 / h265 / qsv(h264)",
"hrs", "val",
"medium: h264 / h265 / qsv(h264 / h265)", "vidutinis: h264 / h265 / qsv(h264 / h265)",
"mins", "minuės",
"option %s requires a value", "%s parinkčiai būtina reikšmė",
"slow: h264 / h265 / qsv(h264 / h265)", "lėtas: h264 / h265 / qsv(h264 / h265)",
"slower: h264 / h265 / qsv(h264)", "lėtesnis: h264 / h265 / qsv(h264)",
"superfast: h264 / h265", "super greitas: h264 / h265",
"tvhdhomerun client", "tvhdhomerun klientas",
"ultrafast: h264 / h265", "ultra greitas: h264 / h265",
"veryfast: h264 / h265 / qsv(h264)", "labai greitas: h264 / h265 / qsv(h264)",
"veryslow: h264 / h265 / qsv(h264)", "labai lėtas: h264 / h265 / qsv(h264)",
NULL, NULL
};

static const char *tvh_locale_23[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nVoor meer informatie kunt u de website van Tvheadend bezoeken:\nhttps://tvheadend.org\n",
"# Mapped channels", "# Toegewezen kanalen",
"# Muxes", "# Muxen",
"# Services", "# Diensten",
"# tune repeats", "Tune herhalingen",
"(Default profile)", "(Standaard Profiel)",
"(Read-only)", "(Alleen lezen)",
"1 day", "1 dag",
"1 month", "1 maand",
"1 year", "1 jaar",
"2 months", "2 maanden",
"2 weeks", "2 weken",
"2 years", "2 jaar",
"2-Port switch (universal LNB)", "2 poort switch (universele LNB)",
"3 days", "3 dagen",
"3 months", "3 maanden",
"3 weeks", "3 weken",
"3 years", "3 jaar",
"4-Port switch (universal LNB)", "4 poort switch (universele LNB)",
"5 days", "5 dagen",
"6 months", "6 maanden",
"A channel tag (e.g. a group of channels) to which this rule applies.", "Een kanaalkenmerk (bijv. een groep van kanalen) waarop de regel van toepassing is.",
"AC-3 = descriptor 6", "AC-3 = indicator 6",
"AC-3 detection", "AC-3 detectie",
"ACTIVE", "ACTIEF",
"ANY", "ELKE",
"ATSC source ID", "ATSC bron ID",
"ATSC-C Network", "ATSC-C Netwerk",
"ATSC-T Network", "ATSC-T Netwerk",
"AUTO", "Auto",
"Aborted by user", "Afgebroken door gebruiker",
"Accept remote subscription weight", "Accepteer abonnementsprioriteit",
"Accept the remote subscription weight (from the SAT>IP client).", "Accepteer de abonnementsprioriteit (van SAT>IP client)",
"Accept zero value for TSID", "Nulwaarde aanvaarden voor TSID",
"Access", "Toegang",
"Access (ACL)", "Toegang (ACL)",
"Action", "Actie",
"Active", "Actief",
"Adapter in use by another subscription", "Adapter is in gebruik door andere aanmelding",
"Add file and line numbers to debug", "Voeg bestandsnaam en regelnummer toe aan debug",
"Add the thread ID to debug", "Voeg thread ID toe aan debug",
"Add this value to PTS for the teletext subtitles. The time value is in milliseconds and may be negative.", "Voeg deze waarde voor de teletekst-ondertitels toe aan PTS. De tijdwaarde is in milliseconden en kan negatief zijn.",
"Additional arguments to pass to the grabber.", "Extra attributen om door te geven aan de EPG-grabber",
"Additional service names found in EPG data.", "Extra namen gevonden in EPG data voor dienst.",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "Extra tijd (in seconden) om de tuner klaar te zetten voor opnemen. Dit is nuttig voor diegenen met tuners die tijd nodig hebben om te tunen of die verkeerde gegevens sturen in het begin.",
"Admin", "Beheerder",
"Admin password", "Wachtwoord beheerder",
"Admin username", "Gebruikersnaam beheerder",
"Administrator login", "Beheerder login",
"Adult movie", "Volwassen film",
"Advanced", "Geavanceerd",
"Advanced (non-universal LNBs, rotors, etc.)", "Geavanceerd (niet universele schotels, rotors, etc)",
"Adventure", "Avontuur",
"Advertisement / Shopping", "Reclame / Winkelen",
"After creating the network scan it for services.", "Na het maken van het netwerk, scan het voor diensten.",
"All (Streaming plus DVR)", "Allen (streaming en DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Alle karakters die een problemen in de bestandsnaam kunnen vormen zullen worden vervangen door een underscore. Zie help for meer details.",
"All lower-case", "Alles kleine letters",
"All proceeds are used to support server infrastructure and buy test equipment.", "Alle vooruitgang wordt gebruikt om de server infrastructuur uit te breiden en om test apparatuur te kopen.",
"All: Record if different description", "Alles: Opnemen als omschrijving anders is",
"All: Record if different episode number", "Alles: Opnemen als het afleveringsnummer anders is",
"All: Record if different subtitle", "Alles: Opnemen als ondertiteling anders is",
"All: Record once per day", "Alles: Eenmalig per dag opnemen",
"All: Record once per month", "Alles: één keer per maand opnemen",
"All: Record once per week", "Alles: Eenmalig per week opnemen",
"Allow high bitrate mode (CI+ CAMs only).", "Sta hoge bitrade modus (alleen CI+ CAMSs) toe",
"Allow idle scan tuning on this device.", "Sta rust scan tuning op dit apparaat toe.",
"Allow the combined size of all timeshift buffers to potentially grow unbounded until your storage media runs out of space.", "Laat mogelijkerwijs toe dat de gecombineerde grootte van alle timeshift buffers onbegrensd groeit totdat de opslag media geen ruimte meer heeft.",
"Allow the initial scan tuning on this device (scan when Tvheadend starts or when a new multiplex is added automatically). At least one tuner or input should have this settings turned on. See also 'Skip Startup Scan' in the network settings for further details.", "Sta de eerste afstemming met scan op dit apparaat toe (scan wanneer Tvheadend start of wanneer automatisch een nieuwe multiplex wordt toegevoegd). Op tenminste één tuner of input moeten deze instellingen zijn ingeschakeld. Zie ook 'Opstartscan overslaan' in de netwerkinstellingen voor meer informatie.",
"Allow the timeshift buffer to grow unbounded until your storage media runs out of space. Warning, enabling this option may cause your system to slow down or crash completely!", "Laat toe dat de timeshift buffer onbegrensd groeit totdat de opslag media geen ruimte meer heeft. Let op, het inschakelen van deze optie kan het systeem langzaam maken of compleet laten crashen.",
"Allow/Disallow access to the 'Configuration' tab.", "Toegang tot het tabblad 'Configuratie' toestaan/weigeren.",
"Allow/Disallow web interface access (this  includes access to the EPG).", "Toegang tot webinterface toestaan/weigeren (dit omvat ook toegang tot de EPG).",
"Allowed DVR profiles. This limits the profiles the user has access to.", "Toegestane DVR profielen. Dit beperkt de profielen waar de gebruiker toegang toe heeft.",
"Allowed network", "Toegestaan netwerk",
"Allowed networks", "Toegestane netwerken",
"Alternate PID path", "Alternatief PID pad",
"Alternate configuration path", "Alternatief configuratie pad",
"Altitude (in meters).", "Hoogte (meters)",
"Altitude (meters)", "Hoogte (meters)",
"Always send the whole DiseqC sequence including LNB setup (voltage, tone). If this is not checked, only changed settings are sent, which may cause issues with some drivers. If the tuning is not reliable, try activating this option.", "Stuur altijd de volledige DiseqC reeks, inclusief instellen van LNB (voltage en tone). Als dit niet is aangezet worden alleen de veranderde instellingen doorgestuurd. Dit kan problemen veroorzaken met sommige drivers. Probeer deze optie te activeren als het tunen niet betrouwbaar is.",
"Always use the name defined by the network.", "Gebruik altijd de naam die door het netwerk is gedefinieerd.",
"Always use the old ioctls to read the linuxdvb status (signal strength, SNR, error counters). Some drivers are not mature enough to provide the correct values using the new v5 linuxdvb API.", "Gebruik altijd de oude ioctls manier om de status van linuxdvb uit te lezen (signaal sterkte, SNR, fout tellers). Sommige drivers zijn niet ver genoeg ontwikkeld om de juiste waardes aan te leveren volgens de nieuwe v5 linuxdvb API.",
"An event which starts between this \"start after\" and \"start before\" will be matched (including boundary values).", "Een gebeurtenis die start tussen deze \"start na\" en \"start voor\" wordt vergeleken (inclusief de grenswaarden).",
"Animals", "Dieren",
"Anonymize HTSP access", "Anonimiseer HTSP toegang",
"Any", "Elke",
"Arts", "Kunst",
"Arts / Culture (without music)", "Kunst / cultuur (zonder muziek)",
"Arts magazines", "Kunstmagazines",
"Assign predefined muxes to networks", "Wijs voorgedefinieerde muxen toe aan netwerken",
"Associate this device with one or more networks.", "Verbind dit apparaat met een of meerdere netwerken.",
"Athletics", "Atletiek",
"Audio bitrate (kb/s) (0=auto)", "Audio Bitrate (kb/s) (0=Auto)",
"Audio bitrate to use for transcoding.", "Audio-bitrate om te gebruiken voor transcodering.",
"Audio channel layout.", "Audio kanaal opmaak.",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Te gebruiken audio codec voor transcoderen. \"Niet gebruiken\" zal audio uitvoer uitschakelen.",
"Audio stream", "Audiostream",
"Audio-only stream", "Enkel audiostream",
"Audioes muxer", "Audio muxer",
"Authenticated user", "Geverifieerde gebruiker",
"Auto check disabled", "Automatische check uitgeschakeld",
"Auto check enabled", "Automatische check ingeschakeld",
"Auto record", "Automatisch opnemen",
"Auto record caption", "Onderschrift automatisch opnemen",
"Auto recording%s%s", "Automatisch opnemen%s%s",
"Auto time record", "Automatische tijd opname",
"Auto-Map to channels", "Automatisch mappen naar kanalen",
"Automatic checking", "Automatische controle",
"Automatic recording caption.", "Onderschrift automatisch opnemen",
"Automatic schedule entry not found", "Automatisch gepland item niet gevonden",
"Automatic time scheduler entry not found", "Item van automatische tijd planner niet gevonden",
"Automatically link EPG data to the channel (using the channel name for matching). If you turn this option off, only the OTA EPG grabber will be used for this channel unless you've specifically set a different EPG Source.", "Automatisch EPG-gegevens koppelen aan het kanaal (met behulp van de kanaalnaam voor overeenstemming). Als u deze optie uitschakelt, wordt alleen de OTA EPG-grabber voor dit kanaal gebruikt, tenzij u specifiek een andere EPG-bron hebt ingesteld.",
"Automatically map EPG source", "EPG bronnen automatisch indelen",
"Automatically map all available services to channels.", "Deel alle diensten in kanalen in",
"Automatically map channels defined within the bouquet.", "Deel de kanalen automatisch in zoals in het bouquet gedefinieerd.",
"Automatically name from network", "Naam automatisch vanaf het netwerk",
"Automatically record.", "Automatisch opnemen",
"Automatically update channel icons using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel icons.", "Automatisch kanaalpictogrammen bijwerken met behulp van informatie die wordt verstrekt door de geactiveerde EPG-aanbieders. Let op: dit kan ongewenste wijzigingen veroorzaken in reeds gedefinieerde kanaalpictogrammen.",
"Automatically update channel names using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel names.", "Automatisch kanaalnamen bijwerken met behulp van informatie verstrekt door de ingeschakelde EPG-aanbieders. Let op: dit kan ongewenste wijzigingen aan reeds gedefinieerde kanaalnamen veroorzaken.",
"Automatically update channel numbers using information provided by the enabled EPG providers. Note, this may cause unwanted changes to already defined channel numbers.", "Automatisch kanaalnummers bijwerken met behulp van informatie verstrekt door de ingeschakelde EPG-aanbieders. Let op, dit kan ongewenste wijzigingen aan reeds gedefinieerde kanaalnummers veroorzaken.",
"Autorec maximum count (0=unlimited)", "Maximaal aantal automatische opnames (0=ongelimiteerd)",
"Autorec maximum schedules limit (0=unlimited)", "Maximaal aantal automatische opnames (0=ongelimiteerd)",
"Bad", "Slecht",
"Bad request", "Verkeerd verzoek",
"Bandwidth", "Bandbreedte",
"Base DVB SI Tables (PAT,CAT,PMT,SDT etc.)", "DVB SI-basistabellen (PAT, CAT, PMT, SDT enz.)",
"Based on software from", "Gebaseerd op software van",
"Basic", "Basis",
"Bind to specific local IP address.", "Koppel aan specifiek lokaal ip adres.",
"Bitrate to use for the transcode. See Help for details.", "Bitrate te gebruiken voor de transcode. Zie Help voor details.",
"Blue", "Blauw",
"Bouquet", "Groep",
"Bouquet (auto)", "Groep (automatisch)",
"Bouquet source.", "Bouquet bron.",
"Bouquets", "Boeketten",
"Brand", "Merk",
"Branding information (if available).", "Merk informatie (indien beschikbaar).",
"Broadcast", "Uitzending",
"Broadcast type", "Uitzendingtype",
"Broadcast.", "Uitzenden",
"Broadcasting", "Uitzenden",
"Buffering limit (ms)", "Buffer limit (ms)",
"By default, linuxdvb's status read period is 1000ms (one second). The accepted range is 250ms to 8000ms. Note that for some hardware or drivers (like USB), the status operations take too much time and CPU. In this case, increase the default value. For fast hardware, this value might be decreased to make the decision of the re-tune algorithm based on the signal status faster.", "De status van linuxdvb wordt standaard elke 1000ms (een seconde) ingelezen. Het geaccepteerde instel bereik is van 250ms tot 8000ms. Let op dat voor sommige hardware of drivers (zoals bijvoorbeeld USB), de status operaties teveel tijd en CPU verbruiken. In dit geval moet de standaard waarde verhoogd worden. Voor snelle hardware kan deze waarde verlaagd worden zodat het re-tune algoritme, wat met de signaal status werkt, sneller reageert.",
"CA (descrambling) Client", "CA (ontcijfering) Client",
"CA client %i", "CA Client %i",
"CA identification", "CA Identificatie",
"CA provider", "CA Aanbieder",
"CAPMT interval (in ms).", "CAPMT Interval (ms)",
"CAPMT interval (ms)", "CAPMT Interval (ms)",
"CAPMT query interval (ms).", "CAPMT query Interval (ms)",
"CSA (descrambling)", "CSA (ontcijfering)",
"Cache scheme", "Cache schema",
"Camd.socket filename / IP Address (TCP mode)", "Camd.socket bestandsnaam / IP adres (TCP modus)",
"Change parameters", "Parameters wijzigen",
"Channel", "Kanaal",
"Channel Tags", "Kanaal label",
"Channel does not exist", "Kanaal bestaat niet",
"Channel icon", "Kanaallogo",
"Channel icon URL.", "Kanaal logo URL.",
"Channel icon as defined in EPG data.", "Kanaallogo zoals gedefinieerd in EPG data.",
"Channel icon name scheme", "Kanaallogo naam schema",
"Channel icon path", "Kanaal logo pad",
"Channel mapping options", "Kanaal indeel opties",
"Channel name", "Kanaalnaam",
"Channel not found", "Kanaal niet gevonden",
"Channel number", "Kanaal nummer",
"Channel number as defined in EPG data.", "Kanaalnummer zoals gedefinieerd in EPG data.",
"Channel number offset", "Kanaalnummer offset",
"Channel number range", "Bereik kanaalnummers",
"Channel numbers (heuristic)", "Kanaal nummes (heuristisch)",
"Channel numbers from", "Kanaal nummer vanaf",
"Channel tag", "Kanaallabel",
"Channel tag does not exist", "Kanaallabel bestaat niet",
"Channel tag reference", "Kanaallabel referentie",
"Channel tags", "Kanaallabels",
"Channel tags the user is allowed access to/excluded from.", "Kanaallabels waar de gebruiker (geen) toegang toe heeft.",
"Channel to use/used for the recording.", "Kanaal welke gebruikt/is wordt voor opname.",
"Channel update options", "Kanaal update opties",
"Channels", "Kanalen",
"Channels EPG data is used by.", "Kanalen waar EPG data door gebruikt wordt.",
"Character set", "Karakterset",
"Characters not supported in Windows filenames (e.g. for an SMB/CIFS share) will be stripped out or converted.", "Karakters die niet ondersteunt worden in Windows bestandsnamen (bijvoorbeeld bij een SMB/CIFS share) worden eruit gehaald of aangepast.",
"Charset", "Tekenset",
"Check availability", "Controleer beschikbaarheid",
"Check for the services' presence. If the service is no longer broadcast this field will change to Missing In PAT/SDT.", "Controleer op de aanwezigheid van de dienst. Als de dienst niet langer wordt uitgezonden, verandert dit veld in Ontbreekt In PAT/SDT.",
"Check services for availability. If enabled, services that are not currently broadcasting (or can't be decrypted) will be ignored. Leave disabled if you want Tvheadend to also map offline services.", "Controleer diensten op beschikbaarheid. Indien ingeschakeld, worden diensten die momenteel niet worden uitgezonden (of niet kunnen worden ontsleuteld) genegeerd. Laat uitgeschakeld als je wilt dat Tvheadend ook offline diensten toewijst.",
"Check the descrambling status after this timeout.", "Controleer de ontcijferingsstatus na deze time-out.",
"Children's / Youth Programs", "Kinder- / Jeugdprogramma's",
"Children's / Youth programs", "Kinderen / Jeugd programma's",
"Cinema", "Bioscoop",
"Class", "Klasse",
"Classical", "Klassiek",
"Classical music", "Klassieke muziek",
"Clean effects", "Geen effecten",
"Client name", "Client naam",
"Clone scheduled entry on error", "Gepland item klonen als het fout gaat",
"Comedy", "Komedie",
"Command delay time (ms) (10-200)", "Commando vertraging tijd (ms) (10-200)",
"Command time (ms) (10-100)", "Commando tijd (ms) (10-100)",
"Comment", "Commentaar",
"Commercial break", "Reclame",
"Commercials will be dropped from the recordings. Commercial detection works using EITp/f (EPG running state) and for the Swedish channel TV4 (using teletext info).", "Reclames worden weggelaten in opnames. Reclame detectie werkt alleen wanneer EITp/f (EPG running state) wordt gebruikt en voor het Zweedse kaneel TV4 (gebruikmakend van teletekst informatie).",
"Completed OK", "Succesvol",
"Compress EPG database", "Comprimeer EPG database",
"Compress the EPG database to reduce disk I/O and space.", "Comprimeer de EPG database om disk ruimte te besparen en snelheid te verbeteren.",
"Concurrent limit per network position (src=) for satellite SAT>IP tuners. The first limit number is for src=1 (AA), second for src=2 (AB) etc.", "Gelijktijdige limiet per netwerk positie (src=) voor satelliet SAT>IP tuners. Het eerste limiet nummer is voor src=1 (AA), de tweede voor src=2 (AB), enzovoorts.",
"Conditional Access Identification.", "Conditional Access Identificatie.",
"Configuration", "Configuratie",
"Configuration - Base", "Configuratie - Basis",
"Configuration - Image Cache", "Configuratie - Image Cache",
"Configuration - SAT>IP Server", "Configuratie - SAT>IP Server",
"Configuration ID", "Configuratie ID",
"Configuration name", "Configuratie naam",
"Configuration version", "Configuratie versie",
"Connection limit type", "Connectie limiet type",
"Connection limits", "Verbindingslimieten",
"Constellation", "Stelsel",
"Container to use for the transcoded stream.", "Container die gebruikt moet worden voor de getranscodeerde stream.",
"Content character set", "Tekenset inhoud",
"Content type", "Soort inhoud",
"Content type.", "Soort inhoud.",
"Contest", "Wedstrijd",
"Continue if descrambling fails", "Verdergaan als ontcijferen mislukt",
"Continue recording for x minutes after scheduled stop time", "Opname door laten lopen x minuten na geplande stop tijd",
"Continue recording for x minutes after scheduled stop time.", "Opname door laten lopen x minuten na geplande stop tijd.",
"Cookie expiration (days)", "Cookie verlooptijd (dagen)",
"Cooking", "Koken",
"Copy codec type", "Kopieer codec type",
"Copy layout", "Layout kopieren",
"Create SDTV/HDTV/Radio tags.", "Maak SDTV/HDTV/Radio labels aan.",
"Create a bouquet from the playlist.", "Maak een boeket vanuit de afspeellijst.",
"Create a directory per channel when storing recordings. If both this and the 'directory per day' checkbox is enabled, the date-directory will be the parent of the per-channel directory.", "Maak een directory per kanaal tijdens het opslaan van opnames. Als zowel deze en het 'directory per dag' selectievakje zijn aangezet, komt per kanaal directory onder de datum directory te liggen.",
"Create a directory per title when storing recordings. If the day/channel directory checkboxes are also enabled, those directories will be parents of this directory.", "Maak een map per titel bij het opslaan van opnamen. Als de selectievakjes voor dag/kanaalmap ook zijn ingeschakeld, zijn die mappen de hoofdmappen van deze map.",
"Create a new directory per day in the recording system path. Folders will only be created when something is recorded. The format of the directory will be ISO standard YYYY-MM-DD.", "Maak een nieuwe map per dag in het pad van het opnamesysteem. Mappen worden alleen gemaakt als iets is opgenomen. Het formaat van de map is ISO-standaard JJJJ-MM-DD.",
"Create a provider name tag.", "Maak een provider naam label aan.",
"Create and associate a network tag to created channels.", "Maak een netwerk label aan en wijs deze toe aan de aangemaakte kanelen.",
"Create and associate a provider tag to created channels.", "Maak een provider label aan en wijs deze toe aan de aangemaakte kanalen.",
"Create and link these tags to channels when mapping.", "Maak en koppel deze tags aan kanalen tijdens het toewijzen.",
"Create bouquet", "Bouquet aanmaken",
"Create bouquet tag", "Bouquet label aanmaken",
"Create directories using these permissions.", "Maak directories aan met deze rechten.",
"Create files using these permissions.", "Maak bestanden aan met deze rechten.",
"Create network name tags", "Maak netwerk naam labels aan",
"Create network name tags (set by provider).", "Maak netwerk naam labels aan (ingesteld door provider).",
"Create network tags", "Maak netwerk labels aan",
"Create provider name tags", "Maak provider naam labels aan",
"Create provider tags", "Maak provider labels aan",
"Create tags", "Maak labels aan",
"Create tags in recordings using media containers that support metadata (if possible).", "Maak labels aan in opnames wanneer de media containers dit ondersteunen.",
"Create type-based tags", "Maak type gebaseerde labels aan",
"Created", "Gemaakt",
"Creator", "Gemaakt door",
"Cron multi-line", "Cron meerdere regels",
"Culture (without music)", "Cultuur (zonder muziek)",
"Culture magazines", "Cultuurmagazines",
"Current RTSP port.", "Huidige RTSP poort.",
"Current affairs", "Lopende zaken",
"Current number of objects.", "Huidige aantal objecten.",
"Current object size.", "Huidige object grote.",
"Current tuner configuration.", "Huidige tuner configuratie.",
"Custom HTTP headers", "Aangepaste HTTP headers",
"DAB Network", "DAB Netwerk",
"DBus - use the session message bus instead of the system one", "DBus - gebruik de sessie beheerders bus in plaats van de systeem bus",
"DES key", "DES sleutel",
"DSCP/TOS for streaming", "DSCP/TOS voor streaming",
"DVB CSA (descrambling) Tables", "DVB CSA (ontcijfering) tabellen",
"DVB EPG ID", "DVB Electronische Programma Gids nummer",
"DVB EPG Tables", "DVB EPG-tabellen",
"DVB Inputs - Networks", "DVB Inputs - Netwerken",
"DVB SI Tables", "DVB SI-tabellen",
"DVB Time Tables", "DVB-tijdtabellen",
"DVB scan files", "DVB scan bestanden",
"DVB scan files path", "DVB scan bestanden pad",
"DVB-C Network", "DVB-C Netwerk",
"DVB-S Network", "DVB-S Netwerk",
"DVB-S Satellite Configuration", "DVB-S Satelliet Configuratie",
"DVB-T Network", "DVB-T Netwerk",
"DVR - Auto-recording (Autorecs)", "DVR - Auto-opname (Autorecs)",
"DVR - Profiles", "DVR - Profielen",
"DVR - Time-based Recording (Timers)", "DVR - Op tijd gebaseerde opname (Timers)",
"DVR behavior", "DVR gedrag",
"DVR configuration", "DVR configuratie",
"DVR configuration profiles", "DVR configuratie profielen",
"DVR configurations", "DVR-configuraties",
"DVR entry not found", "DVR item niet gevonden",
"DVR file retention period", "DVR bestand bewaartermijn",
"DVR log retention", "DVR log behoud",
"DVR log retention period", "DVR log bewaartermijn",
"DVR override: high", "DVR voorrang: hoog",
"DVR override: important", "DVR voorrang: belangrijk",
"DVR override: low", "DVR voorrang: laag",
"DVR override: normal", "DVR voorrang: normaal",
"DVR override: unimportant", "DVR voorrang: onbelangrijk",
"DVR profile to use/used for the recording.", "DVR profiel gebruikt of te gebruiken voor opname.",
"DVR schedule does not exist", "DVR planning bestaat niet",
"DVR schedule does not have a file yet", "DVR planning heeft nog geen bestand",
"DVR schedule not found", "DVR planning niet gevonden",
"Dance", "Dans",
"Data errors", "Data fouten",
"Data path (if applicable).", "Data pad (indien van toepassing)",
"Date the EPG data was last updated (not set for OTA grabbers).", "Datum wanneer de EPG data voor het laatst is bijgewerkt (niet ingesteld voor OTA grabbers).",
"Days of Week", "Dagen van de week",
"Days of the week to which the rule should apply.", "Dagen van de week waarop deze regel van toepassing moet zijn",
"Days of week", "Dagen van de week",
"Debate", "Debat",
"Debug libav log", "Debug libav logboek",
"Debug options", "Debug opties",
"Debug subsystems", "Debug subsystemen",
"Debug to syslog", "Debug naar syslog",
"Debug trace (low-level)", "Debug spoor (laag niveau)",
"Default", "Standaard",
"Default language(s)", "Standaard taal/talen",
"Default language.", "Standaard taal",
"Default login", "Standaard aanmelding",
"Default priority", "Standaard prioriteit",
"Default user interface level.", "Standaard gebruikers interface niveau",
"Define network group to limit network usage.", "Definieer netwerk groep om gebruik te limiteren",
"Delivery system", "Ontvangstsysteem",
"Demux path", "Demux pad",
"Descramble services (limit per mux)", "Descramble services (limiet per mux)",
"Descrambler", "Ontcijferaar",
"Descrambler EMM", "EMM-ontcijferaar",
"Descrambler buffer (TS packets)", "Descramble buffer (TS pakketen)",
"Descrambling timeout (ms)", "Time-out ontcijfering (ms)",
"Description", "Omschrijving",
"Device ID", "Apparaat ID",
"Device model", "Apparaat model",
"Device path", "Apparaat pad",
"Device path in sysfs", "Apparaatpad in sysfs",
"Digital Video Recorder", "Digitale Video Recorder",
"Directory override. Override the subdirectory rules specified by the DVR configuration and put all recordings done by this entry into the specified subdirectory", "Map overschrijven. De submap-regels overschrijven die zijn opgegeven door de DVR-configuratie en alle opnamen die door dit item zijn gemaakt in de opgegeven submap plaatsen",
"Directory permissions (octal, e.g. 0775)", "Directorie rechten (octaal, bijvoorbeeld 0775)",
"Directory used by the entry.", "Map gebruikt door het item.",
"Disable", "Uitschakelen",
"Disable DVB bouquets", "Schakel DVB bouquets uit",
"Disable SAT>IP client", "SAT>IP client uitzetten",
"Disable X_SATIPM3U tag", "De tag X_SATIPM3U uitschakelen",
"Disable all access control checks", "Schakel elke vorm van toegangscontrole uit",
"Disable debug on stderr", "Schakel debug uit op stderr (standaard error)",
"Disable device/firmware-specific workarounds", "Schakel apparaat/firmware specifieke oplossingen uit",
"Disable syslog (all messages)", "Schakel syslog uit (alle berichten)",
"Disabled", "Uitgeschakeld",
"Discover more muxes using the Network Information Table (if available).", "Meer muxes ontdekken met behulp van de Netwerk Informatie Tabel (indien beschikbaar).",
"Discussion", "Discussie",
"Do not send X_SATIPM3U information in the XML description to clients.", "Stuur geen X_SATIPM3U-informatie in de XML-beschrijving naar clients.",
"Do not send any stream specific information to the HTSP client like signal strength, input source etc.", "Stuur geen streamspecifieke informatie naar de HTSP-client zoals signaalsterkte, invoerbron, etc.",
"Do not use", "Niet gebruiken",
"Documentary", "Documentaire",
"Don't abort streaming when an encrypted stream can't be decrypted by a CA client that normally should be able to decrypt the stream.", "Het streamen niet onderbreken als een gecodeerde stream niet kan worden ontsleuteld door een CA-client die dit normaal wel moet kunnen.",
"Don't backup configuration tree at upgrade", "Maak geen backup van configuratie bij upgrade",
"Don't include the title in the filename.", "Neem titel niet op in bestandsnaam",
"Don't include title in filename", "Neem titel niet op in bestandsnaam",
"Don't keep", "Niet houden",
"Don't re-record", "Nieuw opnieuw opnemen",
"Don't re-record if recording fails.", "Niet opnieuw opnemen als opname faalt",
"Don't re-schedule if recording fails.", "Niet opnieuw inplannen als opname faalt.",
"Don't reschedule", "Niet opnieuw inplannen",
"Don't use the provider's channel numbers.", "Provider kanaal nummers niet gebruiken",
"Duplicate handling", "Duplicaat afhandeling",
"Duplicate recording handling.", "Opname duplicaat afhandeling",
"EIT - skip TSID check", "EIT - TSID controle overslaan",
"EIT time offset", "EIT tijd uitloop",
"EPG Grabber Channel", "EPG Grabber Kanaal",
"EPG Grabber Configuration", "EPG grabber configuratie",
"EPG Language (priority order)", "EPG taal (prioriteit volgorde)",
"EPG name", "EPG naam",
"EPG scan time-out in seconds (30-7200)", "Time-out EPG-scan in seconden (30-7200)",
"EPG source", "EPG bron",
"EPG update window", "Venster om EPG bij te werken",
"Economics", "Economie",
"Education", "Scholing",
"Education / Science / Factual topics", "Onderwijs / Wetenschap / Feitelijke onderwerpen",
"Educational", " Leerzaam",
"Electronic Program Guide", "Elektronische Programma Gids",
"Elementary Stream Filter", "Elementaire Stream Filter",
"Elementary stream filter", "Elementair stream filter",
"Enable", "Inschakelen",
"Enable (auto)", "Aan (auto)",
"Enable DBus", "Gebruik DBus",
"Enable NTP driver", "NTP driver inschakelen",
"Enable coredumps for daemon", "Gebruik coredumps voor de daemon",
"Enable debug on stderr", "Debug op stderr",
"Enable debug subsystems", "Debug van subsystemen",
"Enable debug to file", "Debug naar bestand",
"Enable debug to syslog", "Debug naar syslog",
"Enable exclusion of user-config defined channel tags. This will prevent the user from accessing channels associated with the tags selected (below).", "Inschakelen van het uitzonderen van gedefinieerde kanaal labels in de gebruikers configuratie. Dit voorkomt dat de gebruiker toegang heeft tot kanalen die verbonden zijn met het geselecteerde label (hieronder). ",
"Enable if the SAT>IP box requests plts=on parameter in the SETUP RTSP command for DVB-S2 muxes.", "Inschakelen als de SAT>IP box de plts=on parameter in het SETUP RTSP commando voor DVB-S2 muxen nodig heeft.",
"Enable if the SAT>IP box requires pids=21 parameter in the SETUP RTSP command", "Inschakelen als de SAT>IP box de pids=21 parameter in het SETUP RTSP commando nodig heeft.",
"Enable if the SAT>IP box supports the addpids/delpids commands.", "Inschakelen als de SAT>IP box de addpids/delpids commando's ondersteunt.",
"Enable or disable RTP/AVP/TCP transfer mode (embedded data in the RTSP session) support.", "Schakel de RTP/AVP/TCP-overdrachtsmodus (ingesloten gegevens in de RTSP-sessie) in of uit.",
"Enable or disable fast input switching.", "In- of uitschakelen van snel omschakelen van input.",
"Enable or disable full mux mode.", "In- of uitschakelen van full mux mode.",
"Enable or disable ignoring of Event Information Table (EIT) data on this mux.", "In- of uitschakelen van het negeren van Event Information Table (EIT) data op deze mux.",
"Enable or disable this configuration.", "Deze configuratie in- of uitschakelen.",
"Enable over-the-air program guide (EPG) scanning on this input device.", "Inschakelen van over-the-air programma gids (EPG) scannen op dit input apparaat.",
"Enable syslog", "Syslog inschakelen.",
"Enable system time updates. This will only work if the user running Tvheadend has rights to update the system clock (normally only root).", "Schakel systeemtijd-updates in. Dit werkt alleen als de gebruiker, die Tvheadend gebruikt, rechten heeft om de systeemklok bij te werken (normaal enkel root).",
"Enable the CAs (conditional accesses) tab in web user interface for the advanced level. By default, this tab is visible only in the expert level.", "Schakel in de web-gebruikersinterface het tabblad CA's (voorwaardelijke toegangen) in voor het geavanceerde niveau. Standaard is dit tabblad alleen zichtbaar op expertniveau.",
"Enable this filter.", "Dit filter inschakelen.",
"Enable trace subsystems", "Trace subsytemen",
"Enable web UI debug (non-minified JS)", "Web UI debug inschakelen (niet verkleinde JS)",
"Enable, disable or ignore the mux. When the mux is marked as ignore, all discovered services are removed.", "Schakel de mux in, uit of negeer deze. Wanneer de mux is gemarkeerd als te negeren, worden alle gevonden diensten verwijderd.",
"Enable/Disable CA client.", "Inschakelen/uitschakelen CA client.",
"Enable/Disable service.", "Dienst in-/uitschakelen.",
"Enable/Disable the entry.", "Item in-/uitschakelen.",
"Enable/Disable timeshift.", "Timeshift in-/uitschakelen.",
"Enable/disable EPG data for the entry.", "In- of uitschakelen van EPG data voor deze aanmelding.",
"Enable/disable LNA.", "LNA in- of uitschakelen.",
"Enable/disable auto-rec rule.", "Inschakelen/uitschakelen auto opname regel",
"Enable/disable debugging output to syslog.", "Inschakelen/uitschakelen debugging output naar syslog",
"Enable/disable inclusion of low-level debug traces.", "Insluiting van debug-sporen op laag niveau in-/uitschakelen.",
"Enable/disable interface quick tips.", "Inschakelen/uitschakelen interface snelle tips.",
"Enable/disable libav log output.", "Inschakelen/uitschakelen libav log output",
"Enable/disable logging to syslog.", "Inschakelen/uitschakelen logging naar syslog",
"Enable/disable offering of Entitlement Management Message updates.", "Aanbieden van Entitlement Management Message updates in-/uitschakelen",
"Enable/disable pilot tone.", "Inschakelen/uitschakelen pilot tone",
"Enable/disable power save mode (if supported by the device).", "Inschakelen/uitschakelen stroom besparing modus (indien het apparaat dit ondersteunt).",
"Enable/disable profile.", "Inschakelen/uitschakelen profiel.",
"Enable/disable the bouquet.", "Inschakelen/uitschakelen dit bouquet.",
"Enable/disable the channel.", "Inschakelen/uitschakelen dit kanaal.",
"Enable/disable the device.", "Inschakelen/uitschakelen dit apparaat.",
"Enable/disable the entry.", "Inschakelen/uitschakelen dit item.",
"Enable/disable the grabber.", "Inschakelen/uitschakelen deze grabber.",
"Enable/disable the tag.", "Inschakelen/uitschakelen dit label.",
"Enable/disable tuner/adapter.", "Inschakelen/uitschakelen tuner/adapter.",
"Enabled", "Aan",
"Encrypted", "Versleuteld",
"Encrypted service", "Versleutelde dienst",
"Enter a non-admin user password.", "Voer een niet-beheerder gebruikers wachtwoord in.",
"Enter a non-admin user username.", "Voer een niet-beheerder gebruikersnaam in.",
"Enter allowed network prefix(es). You can enter a comma-seperated list of prefixes here.", "Voer toegestande netwerk prefix(en) in. U kan een kommagescheiden lijst van prefixen hier invoeren.",
"Enter an administrator password.", "Voer een beheerders wachtwoord in.",
"Enter an administrator username. Note: do not use the same username as the superuser backdoor account.", "Voer een beheerders gebruikersnaam in. Let op: gebruik niet dezelfde gebruikersnaam als de achterdeur hoofd gebruiker.",
"Enter comma-separated list of subsystems you want debugging output for (e.g +linuxdvb,+subscriptions,+mpegts).", "Voer een kommagescheiden lijst van subsystemen in waar je debugging output voor wilt (bijv. +linuxdvb,+subscriptions,+mpegts).",
"Enter comma-separated list of subsystems you want to get traces for (e.g +linuxdvb,+subscriptions,+mpegts).", "Voer een kommagescheiden lijst van subsystemen in waar je nadere informatie van wilt (bijv. +linuxdvb,+subscriptions,+mpegts).",
"Enter external IP if behind Network address translation (NAT). Asterisk (*) means accept all IP addresses.", "Externe IP invoeren indien achter netwerkadresvertaling (NAT). Asterisk (*) betekent dat alle IP-adressen worden geaccepteerd.",
"Enter the filename (including path) where Tvheadend should write the log.", "Voer de bestandsnaam in (inclusief pad) waar Tvheadend het logboek moet schrijven.",
"Entertainment programs for 10 to 16", "Amusements programma's voor 10 tot 16",
"Entertainment programs for 6 to 14", "Amusements programma's voor 6 tot 14",
"Environment", "Omgeving",
"Environment (pipe)", "Omgeving (pipe)",
"Episode", "Aflevering",
"Episode %d", "Aflevering %d",
"Episode number/ID.", "Aflevering nummer/ID.",
"Error code", "Fout code",
"Error code of entry.", "Fout code van item.",
"Errors", "Fouten",
"Even key", "Even sleutel",
"Even key.", "Even sleutel",
"Event does not exist", "Gebeurtenis bestaat niet",
"Every day", "Elke dag",
"Exclude channel tags", "Sluit kanaal labels uit",
"Expeditions", "Expedities",
"Experimental film", "Experimentele film",
"Exported tuners", "Geëxporteerde tuners",
"External", "Extern",
"External EPG grabber", "Externe EPG grabber",
"External IP (NAT)", "Extern IP (NAT)",
"External PyEPG Grabber", "Externe PyEPG-grabber",
"External URL", "Externe URL",
"External URL of the bouquet.", "Externe URL van het bouquet",
"External XMLTV EPG Grabber", "Externe XMLTV EPG-grabber",
"Extra arguments", "Extra argumenten",
"Extra start time", "Extra start tijd",
"Extra stop time", "Extra stop tijd",
"Extra warming up time (seconds)", "Extra opwarm tijd (seconden)",
"FAIL", "FOUT",
"FEC high", "FEC hoog",
"FEC low", "FEC laag",
"Factual topics", "Thema's",
"Failed", "Mislukt",
"Failed to open image", "Plaatje openen mislukt",
"Failed view", "Mislukte weergave",
"Fantasy", "Fantasie",
"Fashion", "Mode",
"Fast input switch", "Snelle input switch",
"Fastscan DVB", "Snelscan DVB",
"File missing", "Bestand mist",
"File not created", "Bestand niet aangemaakt",
"File permissions (octal, e.g. 0664)", "Bestands rechten (octaal, bijv. 0664)",
"File removed", "Bestand verwijderd",
"File size", "Bestands groote",
"Filename", "Bestandsnaam",
"Filename (including path)", "Bestandsnaam (inclusief pad)",
"Filename character set", "Bestandsnaam karakter set",
"Filename options", "Bestandsnaam opties",
"Filename used by the entry.", "Bestandsnaam gebruikt door dit item.",
"Filesystem monitor", "Besturingssysteem-monitor",
"Fine arts", "Beeldende kunst",
"Finished", "Afgelopen",
"First word", "Eerste woord",
"Fit to RAM (cut rewind)", "Passend in RAM (terugspoelen afsnijden)",
"Fitness and health", "Sport en gezondheid",
"Folkloric", "Folkloristisch",
"Football", "Voetbal",
"Force (auto)", "Forceer (automatisch)",
"Force CA ID (e.g. 0x2600)", "Forceer CA ID (bijvoorbeeld 0x2600)",
"Force all network connections to this tuner to be made over the specified IP address, similar to the setting for the SAT>IP device itself. Setting this overrides the device-specific setting.", "Forceer alle netwerk verbindingen naar deze tuner gebruik te maken van het gespecificeerde IP adres,  vergelijkbaar met de instelling van het SAT>IP apparaat zelf. Door dit in te stellen wordt de apparaat specifieke instelling overschreven.",
"Force an initial EPG grab at start-up.", "Forceer een initiele EPG grab tijdens het opstarten.",
"Force initial EPG grab at start-up", "Forceer initiele EPG grab tijdens het opstarten.",
"Force old status", "Forceer de oude status",
"Force pilot for DVB-S2", "Forceer pilot voor DVB-S2",
"Force priority", "Forceer prioriteit",
"Force profile to use this priority.", "Forceer profiel om deze prioriteit te gebruiken.",
"Force service type to 1", "Forceer service type naar 1",
"Force teardown delay", "Forceer afsluit vertraging",
"Force the delay between RTSP TEARDOWN and RTSP SETUP command (value from 'Next tune delay in ms' is used). Some devices are not able to handle quick continuous tuning.", "Forceer de vertraging tussen RTSP TEARDOWN- en RTSP SETUP-opdracht (waarde van 'Volgende afstemmingsvertraging in ms' wordt gebruikt). Sommige apparaten kunnen geen snel doorlopend afstemmen aan.",
"Force the grace period for which SAT>IP client waits for the data from server. After this grace period, the tuner is handled as dead. The default value is 5 seconds (for DVB-S/S2: 10 seconds).", "De respijtperiode forceren waarvoor SAT>IP-client wacht op de gegevens van de server. Na deze respijtperiode wordt de tuner als dood behandeld. De standaardwaarde is 5 seconden (voor DVB-S/S2: 10 seconden).",
"Force the local UDP Port number here. The number should be even (RTP port). The next odd number (+1) will be used as the RTCP port.", "Forceer hier het lokale UDP-poortnummer. Het nummer moet even zijn (RTP-poort). Het volgende oneven nummer (+1) zal als de RTCP-poort gebruikt worden.",
"Force usage of entered CA ID on this service.", "Forceer het gebruik van het ingevoerde CA ID op deze service.",
"Foreign countries", "Buitenland",
"Forever", "Voor altijd",
"Fork and run as daemon", "Fork en start als daemon",
"Format string", "Formateer string",
"Found muxes", "Muxen gevonden",
"Found services", "Services gevonden",
"Free subscription weight", "Gratis abonnement-gewicht",
"Free-form field, enter whatever you like here.", "Vrije vorm veld, voer hier wat dan ook in.",
"Free-form text field, enter whatever you like here.", "Vrije vorm tekst veld, voer hier wat dan ook in.",
"Free-form text field, enter whatever you like.", "Vrije vorm tekst veld, voer wat dan ook in.",
"Free-form text field. You can enter whatever you like here.", "Vrije vorm tekst veld. Je kan hier wat dan ook invoeren.",
"Free-format text field. Enter whatever you like here.", "Vrije vorm veld. Voer hier wat dan ook in.",
"Frequency", "Frequentie",
"Frequency (Hz)", "Frequentie (Hz)",
"Frequency (MHz)", "Frequentie (MHz)",
"Frequency (in MHz).", "Frequentie (in MHz).",
"Frequency (kHz)", "Frequentie (kHz)",
"Fri", "Vrijdag",
"Friendly name", "Vriendelijke naam",
"Frontend number", "Frontend nummer",
"Frontend path", "Frontend pad",
"Full DiseqC", "Volledige DiseqC",
"Full mux RX mode supported", "Volledige mux RX-modus ondersteund",
"Full path to an icon used to depict the tag. This can be a TV network logotype, etc.", "Volledig pad naar een pictogram dat wordt gebruikt om de tag weer te geven. Dit kan een logo van een tv-netwerk zijn, enz.",
"Full pathname specification", "Volledige pad naam specificatie",
"Full-text", "Volledige tekst",
"Further education", "Vervolgopleiding",
"GOTOX position", "GOTOX positie",
"Game show", "Spelshow",
"Gardening", "Tuinieren",
"General", "Algemeen",
"General configuration", "Generieke configuratie",
"Generic", "Algemeen",
"Generic options", "Algemene opties",
"Global Headers", "Algemene headers",
"Global timer", "Algemene timer",
"Grabber priority. This option lets you pick which EPG grabber's data get used first. Priority is given to the grabber with the highest value set here. See Help for more info.", "Grabber-prioriteit. Met deze optie kunt u kiezen welke EPG-grabber's gegevens eerst worden gebruikt. Prioriteit wordt gegeven aan de grabber met de hoogste waarde die hier is ingesteld. Zie Help voor meer informatie.",
"Grace period", "Bedenktijd",
"Gray", "Grijs",
"HTSP Answer", "HTSP antwoord",
"HTSP Default Stream Settings", "HTSP standaard stream instellingen",
"HTSP Request", "HTSP verzoek",
"HTSP Server", "HTSP server",
"HTSP Stream Profile", "HTSP Stream Profiel",
"HTSP Subscription", "HTSP abonnement",
"HTTP CORS (cross-origin resource sharing) origin. This option is usually set when Tvheadend is behind a proxy. Enter a domain (or IP) to allow cross-domain requests.", "HTTP CORS (resource sharing tussen bronnen). Deze optie wordt meestal ingesteld als Tvheadend zich achter een proxy bevindt. Voer een domein (of IP) in om aanvragen voor meerdere domeinen toe te staan.",
"HTTP Client", "HTTP client",
"HTTP Server", "HTTP server",
"Handicraft", "Handwerk",
"Hearing impaired", "Doven",
"Hierarchy", "Hierarchie",
"High", "Hoog",
"High bitrate mode (CI+ CAMs only)", "Hoge bitrade modus (alleen CI+ CAMSs)",
"Highest channel number the user can access.", "Hoogste kanaal nummer waar de gebruiker toegang toe heeft.",
"Highest count of objects.", "Hoogste aantal objecten.",
"Historical movie", "Historische film",
"Hostname (or IP) of the server.", "Hostnaam (of IP) van de server.",
"How frequently it will re-try fetching an image that has failed to be fetched.", "Hoe vaak het opnieuw een afbeelding zal proberen ophalen die eerder niet kon opgehaald worden.",
"How frequently the upstream provider is checked for changes.", "Hoe vaak de upstream-aanbieder wordt gecontroleerd op wijzigingen.",
"ID of the network.", "ID van het netwerk.",
"IDLE", "RUST",
"IGNORE", "NEGEREN",
"IP address", "IP adres",
"IP blocking", "IP blokkeren",
"IPTV Automatic Network", "IPTV automatisch netwerk",
"IPTV Network", "IPTV netwerk",
"IPTV Subcription", "IPTV Abonnement",
"ISDB-C Network", "ISDB-C netwerk",
"ISDB-S Network", "ISDB-S network",
"ISDB-T Network", "ISDB-T-netwerk",
"Icon", "Icoon",
"Icon (full URL)", "Logo (volledige URL)",
"Icon URL", "Logo URL",
"Icon base URL", "Icoon basis URL",
"Icon base URL.", "Pictogram basis-URL.",
"Icon has title", "Logo heeft titel",
"Icons from", "Logos van",
"Idle", "Rust",
"Idle scan", "Rust scan",
"Idle scan muxes", "Rust scan muxes",
"If an error occurs clone the scheduled entry and try to record again (if possible).", "Als er een fout optreedt, de geplande invoer clonen en opnieuw proberen op te nemen (indien mogelijk).",
"If both a picon and a channel-specific (e.g. channelname.jpg) icon are defined, prefer the picon.", "Als zowel een picon als een kanaalspecifiek (bijvoorbeeld kanaalnaam.jpg) pictogram gedefinieerd zijn, de voorkeur geven aan de picon.",
"If more than x errors occur during a recording schedule a re-record (if possible).", "Een nieuwe opname plannen (indien mogelijk) als er tijdens een opname meer dan x fouten voorkomen.",
"If no specific priority was requested. This gives certain users a higher priority by assigning a streaming profile with a higher priority.", "Als er geen specifieke prioriteit gevraagd werd. Dit geeft bepaalde gebruikers een hogere prioriteit door een streamingprofiel met hogere prioriteit toe te wijzen.",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Als er geen gebruiker bestaat, maak er dan een\naan zonder gebruikersnaam en wachtwoord.\nGebruik voorzichtig, omdat het volledige toegang\nbied voor iedereen tot deze Tvheadend installatie\ntotdat de toegangscontrole is bijgewerkt vanuit de\nTvheadend web interface.",
"If set, presentation of the tag icon will not superimpose the tag name on top of the icon.", "Indien ingesteld, zal de voorstelling van het labelpictogram de tagnaam bovenop het pictogram niet overlappen.",
"If set, the first bytes from the MPEG-TS stream are discarded. It may be required for some drivers or hardware which do not flush the MPEG-TS buffers completely after a frequency/parameter change.", "Indien ingesteld, worden de eerste bytes van de MPEG-TS-stream weggegooid.  Dit kan vereist zijn voor sommige stuurprogramma's of hardware die de MPEG-TS-buffers niet volledig leegmaken na een frequentie-/parameterwijziging.",
"If something fails, try to switch to a different service on another network. Do not try to iterate through all inputs/tuners which are capable to receive the service.", "Als iets faalt, probeer dan over te schakelen naar een andere dienst op een ander netwerk. Probeer niet alle invoer/tuners te doorlopen die de dienst kunnen ontvangen.",
"If the DiseqC switch is located before the rotor (i.e. tuner - switch - rotor), enable this.", "De DiseqC-schakelaar inschakelen als deze zich vóór de rotor bevindt (bijv. tuner - schakelaar - rotor).",
"If the subscription weight for the input is below the specified threshold, the tuner is handled as free (according the priority settings). Otherwise, the next tuner (without any subscriptions) is used. Set this value to 10, if you are willing to override scan and epggrab subscriptions.", "Als het abonnementsgewicht voor de invoer onder een bepaalde drempel ligt, wordt de tuner als vrij beschouwd  (volgens de prioriteitsinstellingen).  Anders wordt de volgende tuner (zonder abonnementen) gebruikt. Stel deze waarde in op 10 als u bereid bent om scan- en epggrab-abonnementen op te heffen.",
"If you'd like to support the project, please consider a donation.", "Het project steunen kan door middel van donaties.",
"Ignore", "Negeren",
"Ignore EPG (EIT)", "Negeer EPG (EIT)",
"Ignore descriptor 5", "Negeer descriptor 5",
"Ignore encryption flag, include encrypted services anyway.", "Negeer versleutelingsflag, neem versleutelde diensten sowieso op.",
"Ignore invalid SSL certificate", "Negeer niet geldige SSL certificaten",
"Ignore invalid/unverifiable (expired, self-certified, etc.) certificates", "Negeer ongeldige/niet-verifieerbare (vervallen, zelfgecertificeerde, enz.) certificaten",
"Ignore provider's channel numbers", "Negeer kanaal nummers van provider",
"Image Cache", "Afbeeldingscache",
"Immediately abort", "Direct stoppen",
"Important", "Belangrijk",
"Include channel name in filename", "Kanaal naam in de bestandsnaam",
"Include date in filename", "Datum in de bestandsnaam",
"Include episode in filename", "Aflevering in de bestandsnaam",
"Include subtitle in filename", "Ondertiteling in de bestandsnaam",
"Include teletext", "Teletekst opnemen",
"Include teletext in the timeshift buffer. Enabling this may cause issues with some services where the teletext DTS is invalid.", "Teletekst opnemen in de timeshift-buffer. Dit inschakelen kan problemen veroorzaken bij sommige diensten waarvan de teletekst-DTS ongeldig is.",
"Include the date for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "De opnamedatum opnemen in de titel van het evenement. Dit geldt zowel voor de titel die in het bestand is opgeslagen als voor de bestandsnaam zelf.",
"Include the episode subtitle in the title (if available).", "Aflevering ondertiteling in de titel (indien beschikbaar).",
"Include the name of the channel in the event title. This applies to both the title stored in the file and to the filename itself.", "De kanaalnaam opnemen in de titel van het evenement. Dit geldt zowel voor de titel die in het bestand is opgeslagen als voor de bestandsnaam zelf.",
"Include the season and episode in the title (if available).", "Seizoen en aflevering in de titel (indien beschikbaar).",
"Include the time for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "De opnametijd opnemen in de titel van het evenement. Dit geldt zowel voor de titel die in het bestand is opgeslagen als voor de bestandsnaam zelf.",
"Include time in filename", "Tijd in de bestandsnaam",
"Information area", "Informatie gebied",
"Informational", "Informatief",
"Initial scan", "Initiële scan",
"Input base", "Invoer basis",
"Input buffer (bytes)", "Invoer buffer (bytes)",
"Input instance", "Invoer instantie",
"Input path", "Invoer pad",
"Internal", "Intern",
"Internal EPG grabber", "Interne EPG grabber",
"Internal PyEPG Grabber", "Interne PyEPG-grabber",
"Internal XMLTV EPG Grabber", "Interne XMLTV-EPG-grabber",
"Internal error", "Interne fout",
"Internal grabber", "Interne grabber",
"Invalid", "Ongeldig",
"Invalid EPG object request", "Ongeldig EPG object verzoek",
"Invalid adapter number '%s'\n", "Foutief adapter number '%s'\n",
"Invalid arguments", "Ongeldige argumenten",
"Invalid file", "Ongeldig bestand",
"Invalid parameters", "Ongeldige parameters",
"Invalid service", "Ongeldige service",
"Invalid target", "Ongeldig doel",
"Keep", "Behouden",
"Keep timeshift buffers in RAM only. With this option enabled, the amount of rewind time is limited by how much RAM TVHeadend is allowed.", "Bewaar timeshift-buffers alleen in RAM.  Als deze optie is ingeschakeld, wordt de hoeveelheid terugspoeltijd beperkt door hoeveel RAM aan TVHeadend is toegewezen.",
"Keep x amount of storage space free.", "Houd x hoeveelheid opslag ruimte vrij.",
"Kill signal (pipe)", "Stop signaal (pipe)",
"Kill timeout (pipe/secs)", "Stop timeout (pipe/secs)",
"Language", "Taal",
"Language 1", "Taal 1",
"Language 2", "Taal 2",
"Language 3", "Taal 3",
"Language settings", "Taal instellingen",
"Languages", "Talen",
"Last seen", "Laatst gezien",
"Last updated from", "Laatst bijgewerkt vanaf",
"Leisure hobbies", "Vrije tijd",
"Limit connections", "Beperk verbindingen",
"Limited/limit scanning to this network ID only.", "Beperkt/Beperk scannen tot dit netwerk ID alleen.",
"Linked input", "Gekoppelde input",
"LinuxDVB network", "LinuxDVB netwerk",
"List of allowed IPv4 or IPv6 hosts or networks (comma-separated).", "Lijst van toegestane IPv4 of IPv6 systemen of netwerken (komma gescheiden).",
"Listen / Connect port", "Listen / Connect poort",
"Listen on IPv6", "Gebruik IPv6",
"Literature", "Literatuur",
"Local (server) time", "Lokale (server) tijd",
"Local bind IP address", "Lokaal bind IP adres",
"Local channel minor", "Lokaal kanaal minor ",
"Local channel number", "Lokaal kanaal nummer",
"Local discovery IP address", "Lokaal discovery IP adres",
"Local: Record if different description", "Lokaal: Opnemen als beschrijving anders is",
"Local: Record if different episode number", "Lokaal: Opnemen als afleveringsnummer anders is",
"Local: Record if different subtitle", "Lokaal: Opnamen als ondertiteling anders is",
"Local: Record if different title", "Lokaal: Opnemen als de titel anders is",
"Local: Record once per day", "Lokaal: Neem een keer per dag op",
"Local: Record once per week", "Lokaal: Neem één keer er week op",
"Location", "Locatie",
"Location details of the SAT>IP Server.", "Locatie details van de SAT>IP Server.",
"Log", "Logboek",
"Login password.", "Aanmelden wachtwoord.",
"Login username.", "Aanmelden gebruikersnaam.",
"Login/Logout", "Inloggen/uitloggen",
"Logout", "Afmelden",
"Low", "Laag",
"Lowest channel number the user can access.", "Laagste kanaal nummer wat de gebruiker kan benaderen.",
"MP4 audio", "MP4-audio",
"MPEG-2 audio", "MPEG-2-audio",
"MPEG-TS File", "MPEG-TS-bestand",
"MPEG-TS Input Debug", "MPEG-TS-invoerdebug",
"MPEG-TS Parser", "MPEG-TS-parser",
"Magazines", "Tijdschriften",
"Main", "Hoofd",
"Maintain free storage space in MiB", "Beheer vrije opslag ruimte in MiB",
"Maintain used storage space in MiB (0=disabled)", "Beheer gebruikte opslag ruimte in MiB (0=uitgeschakeld)",
"Maintained space", "Beheerde ruimte",
"Make subdirectories per channel", "Maak subdirectories per kanaal",
"Make subdirectories per day", "Maak subdirectories per dag",
"Make subdirectories per title", "Maak subdirectories per titel",
"Manage all", "Alles beheren",
"Manufacturer", "Fabrikant",
"Manufacturer URL", "Fabrikant URL",
"Manufacturer's model number.", "Modelnummer van fabrikant",
"Manufacturer's product description.", "Productbeschrijving van fabrikant.",
"Manufacturer's product information page for the device.", "Product informatie pagina voor het apparaat van fabrikant.",
"Manufacturer's product name.", "Product naam van fabrikant.",
"Map all services", "Deel alle diensten in",
"Map encrypted services", "Deel versleutelde services in",
"Map radio channels", "Deel radio kanalen in",
"Map unnamed channels", "Deel naamloze kanalen in",
"Map zero-numbered channels", "Deel nul genummerde kanalen in",
"Mapped", "Toegewezen",
"Martial sports", "Vechtsporten",
"Master tuner", "Hoofd tuner",
"Matroska specific", "Matroska specifiek",
"Maximal channel number", "Maximaal kanaal nummer",
"Maximum # input streams", "Maximaal # input streams",
"Maximum PIDs", "Maximaal aantal PID's",
"Maximum RAM size (MB)", "Maximale RAM grootte (MB)",
"Maximum allowed difference between event start time when the EPG event is changed.", "Maximaal toegestaan verschil tussen de start tijd van de gebeurtenis zodra de EPG gebeurtenis is veranderd.",
"Maximum bandwidth (Kbps)", "Maximale bandbreedte (Kbps)",
"Maximum count (0=default)", "Maximum aantal (0=standaard)",
"Maximum duration", "Maximumduur",
"Maximum input bandwidth.", "Maximale invoerbandbreedte.",
"Maximum length in characters for the command setting PIDs to the SAT>IP box.", "Maximale tekenlengte voor het commando om PIDs in te stellen op de SAT>IP box.",
"Maximum length of PIDs", "Maximale lengte van PID's",
"Maximum period (mins)", "Maximumperiode (min)",
"Maximum schedules limit (0=default)", "Maximum schemalimiet (0=standaard)",
"Maximum size (MB)", "Maximale grootte (MB)",
"Maximum time to wait (in seconds) for a stream before a timeout.", "Maximale wachttijd voor een stream (in seconden) alvorens een time-out optreedt.",
"Maximum timeout (seconds)", "Maximale timeout (seconden)",
"Medicine", "Geneeskunde",
"Memory Information", "Geheugen informatie",
"Merge same name", "Voeg dezelfde naam samen",
"Merge services with the same name into one channel.", "Diensten met dezelfde naam samenvoegen in één kanaal.",
"Method not found", "Methode niet gevonden",
"Minimal channel number", "Minimaal kanaal nummer",
"Minimum duration", "Minimumduur",
"Minimum rotor time (seconds)", "Minimale rotor tijd (seconden)",
"Missing In PAT/SDT", "Ontbreekt in PAT/SDT",
"Mode", "Modus",
"Model description", "Modelbeschrijving",
"Model name", "Modelnaam",
"Model number", "Modelnummer",
"Modulation", "Modulatie",
"Module ID used to grab EPG data.", " ",
"Mon", "Maandag",
"More verbose libav log", "Meer gedetailleerde libav log ",
"Motor rate (in milliseconds/deg).", "Motor verhouding (in milliseconden/graden).",
"Motor rate (milliseconds/deg)", "Motor verhouding (milliseconden/graden)",
"Motor sport", "Motorsport",
"Motoring", "Autorijden",
"Movie", "Film",
"Movie / drama", "Film / drama",
"Multiple lines of the cron time specification. The default cron triggers the internal grabbers daily at 12:04 and 00:04. See Help on how to define your own.", "Meerdere regels van de cron-tijdspecificatie. De standaard-cron start dagelijks de interne grabbers om 12:04 en 00:04 uur. Zie Help om uw eigen regel te definiëren.",
"Multiple lines of the cron time specification. The default cron triggers the over-the-air grabber daily at 02:04 and 14:04. See Help on how to define your own.", "Meerdere regels van de cron-tijdsspecificatie. De standaard cron start de OTA-grabber dagelijks om 02:04 en 14:04. Zie Help om uw eigen regel te definiëren.",
"Music", "Muziek",
"Music / Ballet / Dance", "Muziek / Ballet / Dans",
"Mux Scheduler", "Mux-planner",
"Mux handling", "Mux behandeling",
"Mux name", "Muxnaam",
"Mux not enabled", "Mux uitgeschakeld",
"NONE", "GEEN",
"Name", "Naam",
"Name (or date) of program the entry is a rerun of.", "Naam (of datum) van het programma waarvan het item een herhaling is.",
"Name of channel the entry recorded from.", "Naam van het kanaal waarvan het item opgenomen werd.",
"Name of object.", "Objectnaam",
"Name of the bouquet.", "Bouguetnaam",
"Name of the client.", "Klantnaam.",
"Name of the entry.", "Itemnaam.",
"Name of the module used to grab EPG data.", "Naam van de module die wordt gebruikt om EPG-gegevens te grabben.",
"Name of the module, grabber or channel that should be used to update this channels EPG info.", "Naam van de module, de grabber of het kanaal die zou moeten gebruikt worden om de EPG-informatie van dit kanaal bij te werken.",
"Name of the network.", "Netwerknaam",
"Name of the profile.", "Profielnaam",
"Name of the tag.", "Labelnaam",
"Name of the tuner.", "Naam van de tuner.",
"Name of the tuner/adapter.", "Naam van de tuner/adapter.",
"Names", "Namen",
"Natural sciences", "Natuurwetenschappen",
"Nature", "Natuur",
"Network", "Netwerk",
"Network ", "Netwerk",
"Network A", "Netwerk A",
"Network B", "Netwerk B",
"Network C", "Netwerk C",
"Network D", "Netwerk D",
"Network ID (limit scanning)", "Netwerk ID (beperk scannen)",
"Network UUID", "Netwerk UUID",
"Network access", "Netwerk toegang",
"Network discovery", "Netwerk ontdekking",
"Network for port A.", "Netwerk voor poort A.",
"Network for port AA.", "Netwerk voor poort AA.",
"Network for port AB.", "Netwerk voor poort AB.",
"Network for port B.", "Netwerk voor poort B.",
"Network for port BA.", "Netwerk voor poort BA.",
"Network for port BB.", "Netwerk voor poort BB.",
"Network for port C.", "Netwerk voor poort C.",
"Network for port D.", "Netwerk voor poort D.",
"Network group", "Netwerk groep",
"Network limit per position", "Netwerk beperking per positie",
"Network name", "Netwerk naam",
"Network prefix", "Netwerk prefix",
"Network settings", "Netwerk instellingen",
"Network type", "Netwerk type",
"Networks", "Netwerken",
"New / premiere / unknown", "Nieuw / premiere / onbekend",
"New login", "Nieuwe login",
"New media", "Nieuwe media",
"New muxes + changed muxes", "Nieuwe muxen + gewijzigde muxen",
"New muxes only", "Enkel nieuwe muxen",
"News", "Actualiteit",
"News / Current Affairs", "Nieuws / Actueel",
"News magazine", "Actualiteitenrubriek",
"Next tune delay in ms (0-2000)", "Vertraging tussen tunen in ms (0-2000)",
"No", "Nee",
"No PIN", "Geen PIN",
"No access", "Geen toegang",
"No adapters specified!\n", "Geen adapter gespecificeerd!\n",
"No assigned adapters", "Geen toegewezen adapters",
"No days", "Geen dagen",
"No descrambler", "Geen descrambler",
"No free adapter", "Geen vrije adapter",
"No input detected", "Geen input gedetecteerd",
"No scheme", "Geen schema",
"No service assigned to channel", "Geen service toegekend aan kanaal",
"No service enabled", "Geen service geactiveerd",
"No source available", "Geen bron beschikbaar",
"Node subsystem", "Node subsysteem",
"None", "Geen",
"Normal", "Normaal",
"Not all SAT>IP servers use the same signal scaling. Change this setting if the signal level displayed within Tvheadend looks too low.", "Niet alle SAT>IP servers gebruiken de zelfde signaal schaal. Verander deze instelling als het signaal niveau zoals getoond in Tvheadend te laag lijkt.",
"Not enough disk space", "Niet genoeg schijfruimte",
"Not enough memory", "Niet genoeg geheugen",
"Not linked", "Niet gekoppeld",
"Not set", "Niet geconfigureerd",
"Not set (none or channel configuration)", "Niet geconfigureerd (geen of zie kanaal configuratie)",
"Not set (use channel or DVR configuration)", "Niet geconfigureerd (gebruik kanaal of DVR configuratie)",
"Number", "Aantal",
"Number of days to keep the file.", "Aantal dagen om het bestand te behouden.",
"Number of days to keep the recorded file.", "Aantal dagen om het opgenomen bestand te behouden.",
"Number of days to keep the recorded files.", "Aantal dagen om de opgenomen bestanden te behouden.",
"Number of days to retain entry information.", "Aantal dagen om de entry informatie te behouden.",
"Number of days to retain information about recording.", "Aantal dagen om informatie over opnames te bewaren.",
"Number of days to retain information about recordings. Once this period is exceeded, duplicate detection will not be possible anymore.", "Aantal dagen om informatie over opnames te bewaren. Zodra deze periode is overschreden, is het detecteren van dubbels niet meer mogelijk.",
"Number of errors that occurred during recording (Transport errors).", "Aantal fouten dat is opgetreden tijdens de opname (Transport fouten).",
"Number of errors that occurred during recording.", "Aantal fouten dat is opgetreden tijdens opname.",
"Number of muxes found.", "Aantal gevonden muxen.",
"Number of repeats for the DiseqC commands (default is zero - no DiseqC repeats). Note: this represents the number of repeats, not the number of requests - so 0 means 'send once: don't repeat', 1 means 'send twice: send once, then send one repeat', etc.", "Aantal herhalingen voor de DiseqC-opdrachten (standaard is nul - geen DiseqC-herhalingen). Opmerking: dit staat voor het aantal herhalingen, niet het aantal verzoeken - dus 0 betekent 'eenmaal verzenden: niet herhalen', 1 betekent 'tweemaal verzenden: d.i. eenmaal verzenden, vervolgens één herhaling verzenden', enz.",
"Number of repeats for the tune requests (default is zero - no repeats). Note: this represents the number of repeats, not the number of requests - so 0 means 'send once: don't repeat', 1 means 'send twice: send once, then send one repeat', etc.", "Aantal herhalingen voor de afstemmingsverzoeken (standaard is nul - geen herhalingen). Opmerking: dit staat voor het aantal herhalingen, niet het aantal verzoeken - dus 0 betekent 'eenmaal verzenden: niet herhalen', 1 betekent 'tweemaal verzenden: d.i. eenmaal verzenden, vervolgens één herhaling verzenden', enz.",
"Number of seconds to wait before timing out.", "Aantal seconden om te wachten voordat er wordt onderbroken.",
"Number of services.", "Aantal services.",
"Number of times this recording was played.", "Aantal keren dat deze opname werd afgespeeld.",
"Number of tsfile tuners", "Aantal tsfile tuners",
"OK (partial)", "OK (gedeeltelijk)",
"Odd key", "Oneven sleutel",
"Odd key.", "Oneven sleutel.",
"Off", "Uit",
"Offset the mapped channel numbers by x (value here + channel number).", "De toegewezen kanaalnummers met x aanpassen (waarde hier + kanaalnummer).",
"Older OSCam", "Oudere OSCam",
"On", "Aan",
"On file removal", "Tijdens verwijderen bestand",
"On-demand (no first rewind)", "Op aanvraag (geen eerste rewind)",
"Once per auto channel", "Eenmaal per auto kanaal",
"One tune request (setup) is sent before the DiseqC sequence (voltage, tone settings). Some linux drivers require this procedure.", "Eén afstemmingsverzoek (setup) wordt verzonden vóór de DiseqC-reeks (spanning, tooninstellingen). Sommige linux-stuurprogramma's vereisen deze procedure.",
"Only Bulsatcom 39E", "Alleen Bulsatcom 39E",
"Only EIT", "Alleen EIT",
"Only OpenTV Sky Ausat", "Alleen OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Alleen ",
"Only OpenTV Sky UK", "Alleen ",
"Only PSIP (ATSC)", "Alleen PSIP (ATSC)",
"Only UK Freesat", "Alleen UK Freesat",
"Only UK Freeview", "Alleen UK Freeview",
"Only Viasat Baltic", "Alleeen Viasat Baltic",
"Only preferred CA PID", "Alleen voorkeurs CA PID",
"Only use specified DVB adapters (comma-separated, -1 = none)", "Gebruik alleen gespecificeerde DVB adapters (komma gescheiden, -1 = geen)",
"Only use this EPG data once when automatically determining what EPG data to set for a channel.", "Gebruik deze EPG gegevens alleen maar wanneer automatisch wordt vastgesteld welke EPG gegevens in te stellen voor een kanaal",
"OpenTV channel number", "OpenTV kanaal nummer",
"Orbital position", "Orbital positie",
"Orbital positions", "Orbital posities",
"Orbital positions.", "Orbital posities.",
"Original network ID", "Origineel netwerk ID",
"Output directory for tsdebug", "Output directory voor tsdebug",
"Over-the-air", "Via de ether",
"Over-the-air Cron multi-line", "Via de ether Cron multi-line",
"Override disabled", "Voorrang uitgeschakeld",
"Override the frontend type.", "Overschrijf het frontend type.",
"Owner", "Eigenaar",
"PEND", "WACHT",
"Password", "Wachtwoord",
"Password2", "Wachtwoord2",
"Path", "Pad",
"Performing arts", "Uitvoerende kunst",
"Physiology", "Fysiologie",
"Polarization", "Polarisatie",
"Political issues", "Politieke zaken",
"Popular culture", "Populaire cultuur",
"Port", "Poort",
"Position", "Positie",
"Power save", "Energiebesparing",
"Pre-defined muxes", "Vooraf gedefinieerde muxes",
"Pre-school children's programs", "Voorschoolse kinderprogramma's",
"Prefer picons over channel name", "Geef picons voorkeur boven kanaal naam",
"Preferred CA PID", "Voorkeurs CA PID",
"Preferred audio language.", "Voorkeurstaal voor audio.",
"Presentation", "Presentatie",
"Presentation details.", "Presentatiedetails.",
"Press", "Pers",
"Priority", "Prioriteit",
"Priority (-10..10)", "Prioriteit (-10..10)",
"Private", "Prive",
"Program synopsis (display only).", "Synopsis van het programma (enkel scherm).",
"Program synopsis.", "Synopsis van het programma.",
"Psychology", "Psychologie",
"Puppets", "Marionetten",
"RAM only", "enkel RAM",
"Re-fetch period (hours)", "Opnieuw halen periode (uren)",
"Re-fetch period (mins)", "Opnieuw halen periode (mins)",
"Re-record%s%s", "Her-opnemen%s%s",
"Re-try period (hours)", "Opnieuw proberen periode (uren)",
"Record all", "Alles opnemen",
"Reject", "Afwijzen",
"Reject exact match", "Exacte overeenkomst weigeren",
"Religion", "Religie",
"Religious", "Religieus",
"Remarkable people", "Opmerkelijke mensen",
"Remove HTTP arguments", "HTTP argumenten verwijderen",
"Remove all unsafe characters from filename", "Verwijder alle onveilige tekens uit de bestandsnaam",
"Remove scrambled bits", "Versleutelde bits verwijderen",
"Repeated", "Herhaald",
"Replace whitespace in title with '-'", "Witruimte in titel vervangen door '-'",
"Replaces all whitespace in the title with '-'.", "Alle witruimte in de titel vervangen door '-'.",
"Reports", "Rapporten",
"Rerun of", "Herhaling van",
"Reserved", "Gereserveerd",
"Resolution (height)", "Resolutie (hoogte)",
"Respawn (pipe)", "Herstarten (pipe)",
"Restart", "Herstarten",
"Rewrite EIT", "Herschrijf EIT",
"Rewrite PAT", "Herschrijf PAT",
"Rewrite PMT", "Herschrijf PMT",
"Rewrite SDT", "Herschrijf SDT",
"Rights", "Rechten",
"Romance", "Romantiek",
"Run as group", "Start als groep",
"Run as user", "Start als gebruiker",
"Running", "Gestart",
"SAT->IP frontend number.", "SAT->IP frontend nummer.",
"SAT>IP Satellite Configuration", "SAT>IP Satellite Configuratie",
"SAT>IP source number", "SAT>IP source nummer",
"SSL verify peer", "SSL controleer andere zijde",
"START", "STARTEN",
"STOP", "STOPPEN",
"Sat", "Zat",
"Satellite config", "Satelliet config",
"Satellite longitude", "Satelliet longitude",
"Satellite longitude.", "Satelliet longitude.",
"Satellite position.", "Satelliet positie.",
"Satellite positions", "Satelliet posities",
"Scan after creation", "Scannen na creatie",
"Scan progress", "Voortgang scannen",
"Scan queue length", "Lengte wachtrij scannen",
"Scan result", "Resultaat scannen",
"Scan status", "Status scannen",
"Scanfile", "Scanbestand",
"Schedule a re-recording if more errors than (0=off)", "Plan her-opname als er meer dan ... fouten zijn (0=uit)",
"Schedule frequency (in cron format).", "Schema frequentie (in cron formaat).",
"Schedule status", "Status van planning",
"Schedule status.", "Status van planning.",
"Scheduled Duration", "Geplande looptijd",
"Scheduled for recording", "Gepland voor opname",
"Scheduled start time", "Geplande starttijd",
"Scheduled stop time", "Geplande stoptijd",
"Scheme to generate the channel icon names (all lower-case, service name picons etc.).", "Schema voor het aanmaken van de namen van kanaalpictogrammen (alles in kleine letters, servicenaam van picons, enz.).",
"School programs", "School TV",
"Science", "Wetenschap",
"Script/program to run when a recording completes.", "Script/programma dat moet worden uitgevoerd wanneer een opname is voltooid.",
"Script/program to run when a recording gets removed.", "Script/programma dat moet worden uitgevoerd wanneer een opname wordt verwijderd.",
"Script/program to run when a recording starts (service is subscribed but no filename available).", "Script/programma dat moet worden uitgevoerd wanneer een opname start (service is geabonneerd maar er is geen bestandsnaam beschikbaar).",
"Season", "Seizoen",
"Season %d", "Seizoen %d",
"Season information (if available).", "Seizoensinformatie (indien beschikbaar).",
"Seek error", "Fout zoeken",
"Select an available network type for this tuner.", "Selecteer een beschikbaar netwerktype voor deze tuner.",
"Select high priority (default) EPG language.", "Selecteer EPG-taal met hoge prioriteit (standaard).",
"Select low priority EPG language.", "Selecteer EPG-taal met lage prioriteit.",
"Select medium priority EPG language.", "Selecteer EPG-taal met gemiddelde prioriteit.",
"Send full PLAY cmd", "Verstuur volledig PLAY commando",
"Serial number", "Serienummer",
"Series link ID.", "Series link-ID.",
"Serious", "Serieus",
"Serious music", "Serieuze muziek",
"Server connectivity", "Server connectiviteit",
"Server details.", "Serverdetails.",
"Service", "Dienst",
"Service ID", "Dienst ID",
"Service Mapper", "Dienst-toewijzer",
"Service Mapping (Map services to channels)", "Dienst toewijzen (Dienst toewijzen aan kanalen)",
"Service configuration", "Service configuratie",
"Service count %zi", "Aantal diensten %zi",
"Service mapping", "Diensten toewijzen",
"Service name", "Dienstnaam",
"Service name found in EPG data.", "Dienstnaam gevonden in EPG-data.",
"Service name picons", "Dienstnaam picons",
"Service priority. Enter a value between -10 and 10. A higher value indicates a higher preference. See Help for more info.", "Dienstprioriteit.  Voer een waarde in tussen -10 en 10.  Een hogere waarde geeft een hogere voorkeur aan.  Zie Help voor meer informatie.",
"Service raw", "Dienst onbewerkt",
"Service type", "Diensttype",
"Service type override. This value will override the service type provided by the stream.", "Service type overschrijven. Deze waarde zal het service type overschrijven wat in de stream zit.",
"Services", "Diensten",
"Services associated with the channel.", "Diensten die aan het kanaal zijn verbonden.",
"Services seen", "Diensten gezien",
"Set as default profile.", "Instellen als standaardprofiel.",
"Set the display name.", "De schermnaam instellen.",
"Settings", "Instellingen",
"Show / Game show", "Show / Spelshow",
"Show this page", "Laat deze pagina zien",
"Show version information", "Laat versie informatie zien",
"Signal quality too poor", "Signaal kwaliteit te slecht",
"Signal scale (240 or 100)", "Signaal schaal (240 or 100)",
"Skip TS packets (0-200)", "Sla TS packets over (0-200)",
"Soccer", "Voetbal",
"Social", "Sociaal",
"Social advisory", "Sociaal adviseren",
"Source", "Bron",
"Source deleted", "Bron verwijderd",
"Source quality is bad", "Bron kwaliteit is slecht",
"Source reconfigured", "Bron hergeconfigureerd",
"Special events (Olympic Games, World Cup, etc.)", "Speciale gebeurtenisses (Olympische spelen, WK, etc.)",
"Specify User-Agent header for the http client", "Specificeer User-Agent header voor HTTP client",
"Specify alternative htsp port", "Specificeer alternative htsp poort",
"Specify alternative http port", "Specificeer alternative http poort",
"Specify alternative http webroot", "Specificeer alternatieve HTTP webroot",
"Specify bind address", "Specificeer bind adres",
"Specify extra htsp port", "Specificeer extra htsp poort",
"Spiritual sciences", "Spirituele wetenschappen",
"Sports", "Sporten",
"Sports magazines", "Sport tijdschriften",
"Standard", "Standaard",
"Start time", "Starttijd",
"Status period (ms)", "Statusperiode (ms)",
"Stop time", "Stoptijd",
"Storage path", "Opslagpad",
"Storage space", "Opslag ruimte",
"Stream ID", "Stream-ID",
"Stream Profile", "Stream-profiel",
"Stream index", "Stream-index",
"Stream index (starts with zero).", "Stream index (start met nul).",
"Stream profile", "Stream-profiel",
"Stream setup error", "Stream setup-fout",
"Stream type", "Stream-type",
"Streaming Profile", "Streaming-profiel",
"Streaming priority", "Streaming-prioriteit",
"Streaming profiles", "Streaming-profielen",
"Subdirectory options", "Subdirectory-opties",
"Subscribe to a service permanently", "Permanent aanmelden bij een dienst",
"Subscription", "Abonnement",
"Subscription does not exist", "Abonnement bestaat niet",
"Subscription overridden", "Aanmelding overschreven",
"Subscription weight", "Belang abonnement",
"Substitute formatters", "Vervang formatters",
"Subtitle", "Ondertiteling",
"Subtitle Stream Filter", "Ondertitel-streamfilter",
"Subtitle codec", "Ondertitel-codec",
"Subtitle of the program (if any) (display only).", "Ondertitel van het programma (indien aanwezig) (enkel scherm).",
"Subtitle of the program (if any).", "Ondertitel van het programma (indien aanwezig).",
"Sun", "Zon",
"Surround (2 front, rear mono)", "Surround (2 voor, achter mono)",
"Switch off the power to the LNB when idle. Note: this may cause interference with other devices when the LNB is powered back up.", "Zet de stroom naar de LNB uit wanneer in rust. Ket op: dit kan storing veroorzaken op andere apparaten zodra de stroom naar de LNB weer aangezet wordt.",
"Sync + Don't keep", "Sync + niet houden",
"System", "Systeem",
"TCP Protocol", "TCP-protocol",
"TV Adapters - Linux ATSC-C Frontend", "TV-adapters - Linux ATSC-C Frontend",
"TV Adapters - Linux ATSC-T Frontend", "TV-adapters - Linux ATSC-T Frontend",
"TV Adapters - Linux DAB Frontend", "TV-adapters - Linux DAB Frontend",
"TV Adapters - Linux DVB-C Frontend", "TV-adapters - Linux DVB-C Frontend",
"TV Adapters - Linux DVB-S Frontend (Master)", "TV-adapters - Linux DVB-S Frontend (Master)",
"TV Adapters - Linux DVB-S Slave Frontend", "TV-adapters - Linux DVB-S Slave Frontend",
"TV Adapters - Linux DVB-T Frontend", "TV-adapters - Linux DVB-T Frontend",
"TV Adapters - Linux ISDB-C Frontend", "TV-adapters - Linux ISDB-C Frontend",
"TV Adapters - Linux ISDB-S Frontend (Master)", "TV-adapters - Linux ISDB-S Frontend (Master)",
"TV Adapters - Linux ISDB-T Frontend", "TV-adapters - Linux ISDB-T Frontend",
"TV Adapters - SatConfig - 4-Port", "TV-adapters - SatConfig - 4-poorten",
"TV Adapters - SatConfig - Advanced", "TV-adapters - SatConfig - Geavanceerd",
"TV Adapters - SatConfig - DiseqC Rotor", "TV-adapters - SatConfig - DiseqC-rotor",
"TV Adapters - SatConfig - DiseqC Switch", "TV-adapters - SatConfig - DiseqC-schakelaar",
"TV Adapters - SatConfig - EN50494/UniCable I", "TV-adapters - SatConfig - EN50494/UniCable I",
"TV Adapters - SatConfig - EN50607/UniCable II", "TV-adapters - SatConfig - EN50607/UniCable II",
"TV Adapters - SatConfig - GOTOX Rotor", "TV-adapters - SatConfig - GOTOX-rotor",
"TV Adapters - SatConfig - Tone Burst/2 Port", "TV-adapters - SatConfig - Tone Burst/2 poorten",
"TV Adapters - SatConfig - USALS Rotor", "TV-adapters - SatConfig - USALS-rotor",
"TV Adapters - SatConfig - Universal LNB (Simple)", "TV-adapters - SatConfig - Universele LNB (eenvoudig)",
"Tag files with metadata", "Bestanden labelen met metadata",
"Tags", "Labels",
"Tags linked/to link to the channel.", "Labels gekoppeld/om te koppelen aan het kanaal",
"Talk show", "Praat programma",
"Team sports (excluding football)", "Team sporten (excl. voetbal)",
"Technology", "Technologie",
"Teletext Stream Filter", "Teletekst Stream Filter",
"Teletext stream type is only available for this filter.", "Teletekst streamtype is enkel beschikbaar voor deze filter.",
"Testing", "Testen",
"Testing options", "Test opties",
"The ATSC source ID as set by the provider.", "Het ATSC-bron-ID zoals ingesteld door de aanbieder.",
"The CA provider to compare. Leave blank to apply to all providers.", "De CA-aanbieder om te vergelijken. Laat dit leeg om op alle aanbieders toe te passen.",
"The CA stream type is only available for this filter.", "Het CA-streamtype is enkel beschikbaar voor deze filter.",
"The CAID to compare. Leave blank to apply to all IDs.", "De CAID om te vergelijken. Laat dit leeg om op alle ID's toe te passen.",
"The CAM slot status.", "De CAM-slotstatus.",
"The COFDM modulation used by the mux. If you're not sure of the value leave as AUTO.", "De COFDM-modulatie die door de mux wordt gebruikt. Als u niet zeker bent van de waarde, neem dan AUTO.",
"The Conditional Access ID used for the service.", "De Conditional Access-ID die voor de dienst wordt gebruikt.",
"The DVR profile to be used/used by the recording.", "Het DVR-profiel dat moet worden gebruikt/gebruikt werd voor de opname.",
"The DVR profile to be used/used by this rule.", "Het DVR-profiel dat moet worden gebruikt/gebruikt werd voor deze regel.",
"The EPG ID used by the entry.", "De EPG-ID die door de invoer wordt gebruikt.",
"The EPG grabber name.", "De naam voor de EPG-grabber.",
"The EPG grabber to use on the mux. Enable (auto) is the recommended value.", "De EPG-grabber om op de mux te gebruiken. Inschakelen (automatisch) is de aanbevolen waarde.",
"The EPG grabber type.", "Het type EPG-grabber.",
"The MPEGTS stream type is only available for this filter.", "Het MPEGTS-streamtype is enkel beschikbaar voor deze filter.",
"The current boot ID.", "De huidige boot-ID.",
"The current configuration ID.", "Het huidige configuratie-ID.",
"The current configuration version.", "De huidige versie van de configuratie.",
"The default language to use if the user  language isn't set in the Access Entries tab.", "De standaardtaal die moet worden gebruikt als de gebruikerstaal niet is ingesteld op het tabblad Toegangsgegevens.",
"The default subscription weight for each subscription.", "Het standaard gewicht voor elk abonnement.",
"The default web interface to use if the user  theme isn't set in the Access Entries tab.", "De standaard webinterface die moet worden gebruikt als het gebruikersthema niet is ingesteld op het tabblad Toegangsgegevens.",
"The delivery system the mux uses. Make sure that your tuner supports the delivery system selected here.", "Het afleversysteem dat de mux gebruikt. Zorg ervoor dat uw tuner het hier geselecteerde afleversysteem ondersteunt.",
"The delivery system used by the mux.", "Het door de mux gebruikte afleversysteem.",
"The delivery system used by your cable provider.", "Het afleversysteem dat wordt gebruikt door uw kabelaanbieder.",
"The delivery system used by your provider.", "Het afleversysteem dat door uw provider wordt gebruikt.",
"The demux path used by the device.", "Het demux-pad dat door het apparaat wordt gebruikt.",
"The device ID.", "Het apparaat-ID.",
"The device path in sysfs filesystem (/sys).", "Het pad van het apparaat in het sysfs-bestandssysteem (/sys).",
"The device's serial number.", "Het serienummer van het apparaat.",
"The forward error correction high value. Most people will not need to change this setting.", "De hoge waarde van de voorwaartse foutcorrectie.  De meeste mensen hoeven deze instelling niet te wijzigen.",
"The forward error correction low value. Most people will not need to change this setting.", "De lage waarde van de voorwaartse foutcorrectie. De meeste mensen hoeven deze instelling niet te wijzigen.",
"The forward error correction used on the mux.", "De voorwaartse foutcorrectie die gebruikt wordt op de mux.",
"The forward error correction. Most people will not need to change this setting.", "De voorwaartse foutcorrectie.  De meeste mensen hoeven deze instelling niet te wijzigen.",
"The frequency of the mux (in Hertz).", "De frequentie van de mux (in Hertz).",
"The frequency of the mux/transponder in Hertz.", "De frequentie van de mux/transponder in Hertz.",
"The frontend number given to the device.", "Het frontendnummer dat aan het apparaat is gegeven.",
"The guard interval used by the mux. If you're not sure of the value leave as AUTO.", "Het bewakingsinterval dat gebruikt wordt door de mux.  Als u niet zeker bent van de waarde, neem dan AUTO.",
"The hierarchical modulation used by the mux. Most people will not need to change this setting.", "De hiërarchische modulatie die gebruikt wordt door de mux.  De meeste mensen hoeven deze instelling niet te wijzigen.",
"The length of time (in seconds) to play the mux (1 hour = 3600).", "De tijdsduur (in seconden) om de mux te spelen (1 uur = 3600).",
"The limit for the PID filter (driver or hardware).", "De limiet voor de PID-filter (stuurprogramma of hardware).",
"The minimum delay after the rotor movement command is sent.", "De minimale vertraging nadat het commando voor de rotorbeweging is verzonden.",
"The modulation used on the mux.", "De gebruikte modulatie op de mux.",
"The mux the service is on.", "De mux waarop de dienst is.",
"The mux to play when the entry is triggered.", "De mux die moet worden gespeeld, wanneer het item wordt geactiveerd.",
"The mux's universally unique identifier.", "De universeel unieke identifier van de mux.",
"The name (or frequency) the mux is on.", "De naam (of frequentie) waarop de mux is.",
"The name given to/of the channel (This is how it'll appear in your EPG.)", "De naam die wordt gegeven aan/de naam van het kanaal (zo verschijnt het in uw EPG.)",
"The name of the profile.", "De naam van het profiel.",
"The name of the the rule.", "De naam van de regel.",
"The network the mux is on.", "Het netwerk waarop de mux is.",
"The network the service is on.", "Het netwerk waarop de dienst is.",
"The network's priority. The network with the highest priority value will be used out of preference if available. See Help for details.", "De prioriteit van het netwerk. Het netwerk met de hoogste prioriteit zal, indien beschikbaar, buiten de voorkeur worden gebruikt. Zie Help voor details.",
"The network's service ID", "De service-ID van het netwerk",
"The networks assigned to the device.", "De netwerken die aan het apparaat zijn toegewezen.",
"The networks using this configuration.", "De netwerken die deze configuratie gebruiken.",
"The networks' universally unique identifier (UUID).", "De universeel unieke identifier (UUID) van het netwerk.",
"The number of DVB-C (Cable) tuners to export.", "Het aantal DVB-C (kabel) tuners dat moet worden geëxporteerd.",
"The number of DVB-C2 (Cable) tuners to export.", "Het aantal DVB-C2 (kabel) tuners dat moet worden geëxporteerd.",
"The number of DVB-S (Satellite) tuners to export.", "Het aantal DVB-S (satelliet) tuners dat moet worden geëxporteerd.",
"The number of DVB-S2 (Satellite) tuners to export.", "Het aantal DVB-S2 (satelliet) tuners dat moet worden geëxporteerd.",
"The number of DVB-T (Terresterial) tuners to export.", "Het aantal DVB-T (terrestrische) tuners dat moet worden geëxporteerd.",
"The number of DVB-T2 (Terresterial) tuners to export.", "Het aantal DVB-T2 (terrestrische) tuners dat moet worden geëxporteerd.",
"The number of allowed connections this user can make to the server.", "Het aantal toegestane verbindingen dat deze gebruiker kan maken met de server.",
"The number of days cookies set by Tvheadend should expire.", "Het aantal dagen dat cookies, die zijn ingesteld door Tvheadend, moeten verlopen.",
"The number of muxes left to scan on this network.", "Het aantal muxen op dit netwerk dat nog moet gescand worden.",
"The number of seconds to wait for a stream to start.", "Het aantal seconden dat moet worden gewacht alvorens een stream kan starten.",
"The number of services currently mapped to channels.", "Het aantal diensten dat momenteel is toegewezen aan kanalen.",
"The orbital position of the satellite the mux is on.", "De orbitale positie van de satelliet waarop de mux staat.",
"The orbital position of the satellite your dish is pointing at.", "De orbitale positie van de satelliet waarop uw schotel gericht is.",
"The outcome of the last scan performed.", "Het resultaat van de laatste uitgevoerde scan.",
"The playlist's character set.", "De tekenset van de afspeellijst.",
"The polarization used on the mux.", "De gebruikte polarisatie op de mux.",
"The position the channel will appear on your EPG. This is not used by Tvheadend internally, but rather intended to be used by HTSP clients for mapping to remote control buttons, presentation order, etc.", "De positie waarin het kanaal op uw EPG zal verschijnen. Dit wordt intern niet door Tvheadend gebruikt, maar is eerder bedoeld om door HTSP-clients te worden gebruikt voor het toewijzen van afstandsbedieningsknoppen, presentatievolgorde, enz.",
"The provider's ID.", "Het ID van de aanbieder.",
"The provider's name.", "De naam van de aanbieder.",
"The provider's network ID.", "Het netwerk-ID van de aanbieder.",
"The provider's network name.", "De netwerknaam van de aanbieder.",
"The recorded file was removed intentionally", "Het opgenomen bestand werd opzettelijk verwijderd",
"The recording/entry status.", "De status van de opname/invoer.",
"The rolloff used on the mux.", "De gebruikte demping op de mux.",
"The rule action defines the operation when all comparisons succeed. See Help for more information on what the various rules do.", "De regelactie definieert de handeling wanneer alle vergelijkingen slagen. Zie Help voor meer informatie over wat de verschillende regels doen.",
"The satellite configuration to use.", "De satellietconfiguratie die moet worden gebruikt.",
"The scan state. New muxes will automatically be changed to the PEND state. You can change this to ACTIVE to queue a scan of this mux.", "De scanstatus. Nieuwe muxen worden automatisch in de status WACHT gezet. U kunt dit wijzigen in ACTIEF om een scan van deze mux in de wachtrij te plaatsen.",
"The scheduled start time, including any padding.", "De geplande starttijd, inclusief eventuele buffer.",
"The scheduled stop time, including any padding.", "De geplande stoptijd, inclusief eventuele buffer.",
"The service ID", "De dienst-ID",
"The service ID as set by the provider.", "De dienst-ID die is ingesteld door de aanbieder.",
"The service ID.", "De dienst-ID.",
"The service name as set by the provider.", "De dienstnaam die is ingesteld door de aanbieder.",
"The service the filter should apply to. Leave blank to apply the filter to all services.", "De dienst waarop de filter moet worden toegepast. Laat dit leeg om de filter op alle diensten toe te passen.",
"The service to which the filter should apply. Leave blank to apply the filter to all services.", "De dienst waarop de filter moet worden toegepast. Laat dit leeg om de filter op alle diensten toe te passen.",
"The service's channel number as set by the provider.", "Het kanaalnummer van de dienst dat is ingesteld door de aanbieder.",
"The service's encryption status.", "De versleutelingsstatus van de dienst.",
"The symbol rate.", "De symbol rate.",
"The time the entry stops/stopped being recorded.", "De tijd dat de invoer stopt/stopte met opnemen.",
"The time window to cut the stop time from the overlapped event in seconds.", "Het tijdvenster om de stoptijd van de overlappende gebeurtenis in seconden te verminderen.",
"The title of the program to look for. Note that this accepts case-insensitive regular expressions.", "De titel van het programma waarnaar moet worden gezocht. Merk op dat dit niet-hoofdlettergevoelige reguliere expressies accepteert.",
"The total number of services found.", "Het totale aantal gevonden diensten.",
"The total scheduled duration.", "De totale geplande duur.",
"The transmission/OFDM mode used by the mux. If you're not sure of the value leave as AUTO but be aware that tuning may fail as some drivers do not like the AUTO setting.", "De transmissie/OFDM-modus die door de mux wordt gebruikt. Als u niet zeker bent van de waarde, neem dan AUTO, maar houd er rekening mee dat het afstemmen kan mislukken omdat sommige stuurprogramma's de AUTO-instelling niet goed vinden.",
"The transponder ID.", "Het transponder-ID.",
"The transport stream ID of the mux within the network.", "Het transport stream-ID van de mux binnen het netwerk.",
"The tuner priority value (a higher value means to use this tuner out of preference). See Help for details.", "De prioriteitswaarde van de tuner (een hogere waarde betekent om deze tuner buiten voorkeur te gebruiken). Zie Help voor details.",
"The tuner priority value for streamed channels through HTTP or HTSP (a higher value means to use this tuner out of preference). If not set (zero), the standard priority value is used. See Help for details.", "De prioriteitswaarde van de tuner voor gestreamde kanalen via HTTP of HTSP (een hogere waarde betekent om deze tuner buiten de voorkeur te gebruiken). Indien niet-ingesteld (nul), wordt de standaard prioriteitswaarde gebruikt. Zie Help voor details.",
"The user who created the recording, or the auto-recording source and IP address if scheduled by a matching rule.", "De gebruiker die de opname heeft gemaakt, of de auto-opname-bron en het IP-adres indien gepland door een overeenkomende regel.",
"The version of Tvheadend that last updated the config.", "De versie van Tvheadend die de configuratie laatst heeft bijgewerkt.",
"The video stream types the filter should apply to.", "De videostreamtypen waarop de filter moet worden toegepast.",
"Theme", "Thema",
"This tuner", "Deze tuner",
"This will create an NTP driver (using shmem interface) that you can feed into ntpd. This can be run without root privileges, but generally the performance is not that great.", "Hiermee wordt een NTP-stuurprogramma gemaakt (met behulp van de shmem-interface) dat u in ntpd kunt invoeren. Dit kan zonder root-privileges worden uitgevoerd, maar over het algemeen is de uitvoering niet geweldig.",
"Thu", "Do",
"Time", "Tijd",
"Time (in milliseconds) for a command to complete.", "Tijd (in milliseconden) om een opdracht te voltooien.",
"Time (in milliseconds) for the rotor to power up.", "Tijd (in milliseconden) om de rotor aan te zetten.",
"Time (in minutes) to re-fetch the playlist.", "Tijd (in minuten) om de afspeellijst opnieuw op te halen.",
"Time Stamp Fix", "Herstelling van de tijdstempel",
"Time missed", "Tijd gemist",
"Time recording%s%s", "Tijd opname%s%s",
"Time to start the recording/time the recording started.", "Tijd om de opname te starten/tijd waarop de opname startte.",
"Time to stop recording/time the recording stopped.", "Tijd om de opname te stoppen/tijd waarop de opname stopte.",
"Time update", "Tijd update",
"Timeout (sec) (0=infinite)", "Timout (sec) (0=oneindig)",
"Timeout (seconds)", "Timeout (seconden)",
"Timeout (secs)", "Time-out (sec.)",
"Timer-based automatic recording.", "Op timer gebaseerde automatische opname.",
"Title", "Titel",
"Title (regexp)", "Titel (regexp)",
"Title of the program (display only).", "Titel van het programma (enkel scherm).",
"Title of the program.", "Titel van het programma.",
"Title of the recording - this is used to generate the filename.", "Titel van de opname - deze wordt gebruikt om de bestandsnaam aan te maken.",
"Too many data errors", "Te veel gegevensfouten",
"Total number of mapped channels on this network.", "Totaal aantal toegewezen kanalen op dit netwerk.",
"Total number of muxes found on this network.", "Totaal aantal gevonden muxen op dit netwerk.",
"Total number of services found on this network.", "Totaal aantal gevonden diensten op dit netwerk.",
"Total number of services found.", "Totaal aantal gevonden diensten.",
"Total number of services seen.", "Totaal aantal bekeken diensten.",
"Total number of services.", "Totaal aantal diensten.",
"Tourism / Travel", "Reizen",
"Traditional arts", "Traditionele kunsten",
"Traditional music", "Traditionele muziek",
"Transcode", "Transcodeer",
"Transcode video only if source video codec matches. \"Any\" ignores source video codec checking and always transcodes. If no codec match is found, transcoding is done using the \"copy\" codec. if a match is found, transcode with the parameters in this profile. Separate codec names with comma.", "Transcodeer video alleen als de codec van de bronvideo-overeenkomt. \"Elke\" negeert de controle van de codec van de bronvideo en transcodeert altijd. Als er geen overeenkomst in codec wordt gevonden, wordt transcoderen uitgevoerd met behulp van de \"kopieer\"-codec. Als een overeenkomst wordt gevonden, transcodeer met de parameters in dit profiel. Aparte codec-namen met komma.",
"Transcoding", "Transcoderen",
"Transmission mode", "Uitzending mode",
"Try to obtain channel numbers from the display-name xml tag. If the first word is number, it is used as the channel number.", "Probeer kanaalnummers te verkrijgen uit de XML-tag van de display-naam. Als het eerste woord een nummer is, wordt dit gebruikt als het kanaalnummer.",
"Tue", "Di",
"Tune before DiseqC", "Tune voor DiseqC",
"Tuner configuration", "Tuner configuratie",
"Tuner configuration.", "Tuner configuratie.",
"Tuning failed", "Tuning faalde",
"Turn off LNB when idle", "Zet LNB uit wanneer in rust",
"Tvheadend server name", "Tvheadend server naam",
"Type override", "Type overschrijven",
"UDP RTP port number (2 ports)", "UDP RTP-poortnummer (2 poorten)",
"UHD: ultra high definition", "UHD: ultra hoge definitie",
"UPnP Protocol", "UPnP-protocol",
"URL of the M3U playlist.", "URL van de M3U afspeellijst.",
"URL with the SAT>IP server XML location", "URL met de SAT>IP server XML locatie",
"Unable to get system UTC time", "Kan de UTC-tijd niet verkrijgen",
"Unable to get system local time", "Kan de lokale tijd van het systeem niet verkrijgen",
"Unable to get system time", "Kan de systeemtijd niet verkrijgen",
"Unable to open file", "Kan bestand niet openen",
"Unable to stat path", "Kan pad niet volgen",
"Uncommitted", "Niet gecommit",
"Uncommitted first", "Niet-vastgelegde eerst",
"Unicable I switch (universal LNB)", "Unicable I switch (universele LNB)",
"Unicable II switch (universal LNB)", "Unicable II switch (universele LNB)",
"Unimportant", "Niet belangrijk",
"Universal LNB only", "Alleen universele LNB",
"Universally unique identifier. Read only.", "Universeel unieke ID. Alleen lezen.",
"Unknown", "Onbekend",
"Unknown file", "Onbekend bestand",
"Unknown reason (%i)", "Onbekende oorzaak (%i)",
"Unlimited size", "Onbeperkte grootte",
"Unlimited time", "Onbeperkte tijd",
"Unset (default)", "Niet geconfigureerd (standaard)",
"Update card (EMM)", "Kaart bijwerken (EMM)",
"Update channel icon", "Kanaal icoon bijwerken",
"Update channel name", "Kanaal naam bijwerken",
"Update channel number", "Kanaal nummer bijwerken",
"Update disabled", "Bijwerken uitgeschakeld",
"Update time", "Tijd bijwerken",
"Update tolerance (ms)", "Bijwerk tolerantie (ms)",
"Updated", "Bijgewerkt",
"Updates from one mux (EMM)", "Bijwerkingen van één mux (EMM)",
"Upon start, Tvheadend doesn't know the last rotor position. This value defines the initial rotor movement. TVHeadend waits the specified time when the first movement is requested.", "Bij het starten kent Tvheadend de laatste rotorpositie niet. Deze waarde definieert de initiële rotorbeweging. TVheadend wacht de opgegeven tijd wanneer de eerste beweging wordt aangevraagd.",
"Usage: %s [OPTIONS]\n", "Gebruik: %s [OPTIONS]\n",
"Use AC-3 detection.", "Gebruik AC-3 detectie.",
"Use EITp/f to decide event start/stop. This is also known as \"Accurate Recording\". See Help for details.", "Gebruik EITp/f om het starten/stoppen van gebeurtenissen te bepalen. Dit staat ook bekend als \"Nauwkeurig opnemen\". Zie Help voor details.",
"Use EPG running state", "Gebruik EPG bedrijfsstatus",
"Use HTTP digest authentication", "Gebruik HTTP-digest verificatie",
"Use PROXY protocol & X-Forwarded-For", "Gebruik het PROXY-protocol & X-Forwarded-For",
"Use WEBM format.", "Gebruik de WEBM-indeling.",
"Use Windows-compatible filenames", "Gebruik Windows-compatible bestandsnamen",
"Use XSPF playlist instead of M3U", "Gebruik een XSPD playlist in plaats van M3U",
"Use a pre-defined list of ATSC-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Gebruik een vooraf gedefinieerde lijst van ATSC-C-muxen. Opmerking: deze lijsten kunnen soms verouderd zijn en kunnen ertoe leiden dat het scannen langer duurt dan normaal.",
"Use a pre-defined list of ATSC-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Gebruik een vooraf gedefinieerde lijst met ATSC-T-muxen. Opmerking: deze lijsten kunnen soms verouderd zijn en kunnen ertoe leiden dat het scannen langer duurt dan normaal.",
"Use a pre-defined list of DAB muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Gebruik een vooraf gedefinieerde lijst met DAB-muxen. Opmerking: deze lijsten kunnen soms verouderd zijn en kunnen ertoe leiden dat het scannen langer duurt dan normaal.",
"Use a pre-defined list of DVB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Gebruik een vooraf gedefinieerde lijst met DVB-C-muxen. Opmerking: deze lijsten kunnen soms verouderd zijn en kunnen ertoe leiden dat het scannen langer duurt dan normaal.",
"Use a pre-defined list of DVB-S/S2 muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Gebruik een vooraf gedefinieerde lijst met DVB-S/S2-muxen. Opmerking: deze lijsten kunnen soms verouderd zijn en kunnen ertoe leiden dat het scannen langer duurt dan normaal.",
"Use a pre-defined list of DVB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Gebruik een vooraf gedefinieerde lijst met DVB-T-muxen. Opmerking: deze lijsten kunnen soms verouderd zijn en kunnen ertoe leiden dat het scannen langer duurt dan normaal.",
"Use a pre-defined list of ISDB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Gebruik een vooraf gedefinieerde lijst van ISDB-C-muxen. Opmerking: deze lijsten kunnen soms verouderd zijn en kunnen ertoe leiden dat het scannen langer duurt dan normaal.",
"Use a pre-defined list of ISDB-S muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Gebruik een vooraf gedefinieerde lijst met ISDB-S-muxen. Opmerking: deze lijsten kunnen soms verouderd zijn en kunnen ertoe leiden dat het scannen langer duurt dan normaal.",
"Use a pre-defined list of ISDB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Gebruik een vooraf gedefinieerde lijst van ISDB-T-muxen. Opmerking: deze lijsten kunnen soms verouderd zijn en kunnen ertoe leiden dat het scannen langer duurt dan normaal.",
"Use only this service ID, filter out others.", "Gebruik alleen deze dienst-ID, filter andere uit.",
"Use original", "Gebruik origineel",
"Use packet backlog", "Gebruik vertraging van het pakket",
"Use service IDs as channel numbers", "Gebruik dienst-ID's als kanaalnummers",
"Use tag internally (don't expose to clients).", "Gebruik tag intern (niet blootstellen aan klanten).",
"Use the provider's service IDs as channel numbers.", "Gebruik de dienst-ID's van de aanbieders als kanaalnummers.",
"Use this character set when setting filenames.", "Gebruik dit tekenset bij het instellen van bestandsnamen.",
"Use x amount of storage space.", "Gebruik x hoeveelheid opslagruimte.",
"User access error", "Gebruikers toegangsfout",
"User does not have access", "Gebruiker heeft geen toegang",
"User does not have access to channel", "Gebruiker heeft geen toegang tot kanaal",
"User icon", "Gebruikerspictogram",
"User language", "Gebruikers taal",
"User limit reached", "Gebruikerslimiet bereikt",
"Username", "Gebruikersnaam",
"Video codec preset", "Video codec voorinstelling",
"Video codec preset to use for transcoding.", "Video codec voorinstelling om te gebruiken tijdens transcoderen.",
"Visually impaired commentary/audio description", "Slechtzienden commentaar/geluid beschrijving",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV profiel H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV profiel H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV profiel VP8/Vorbis/WEBM",
"Waiting for program start", "Wacht op programma start",
"Waiting for stream", "Wacht op stream",
"War", "Oorlog",
"Water sport", "Watersport",
"Weak stream", "Zwakke stream",
"Weather report", "Weer",
"Web interface language", "Web interface taal",
"Web interface language.", "Web interface taal.",
"Web interface theme.", "Web interface thema.",
"Web theme", "Web thema",
"Web user interface", "Web gebruikers interface",
"Wed", "Woe",
"Welcome - Tvheadend - your TV streaming server and video recorder", "Welkom - Tvheadend - jouw TV streaming server en video recorder",
"Western hemisphere (latitude direction)", "Westelijk halfrond (breedtegraad richting)",
"Western hemisphere (latitude direction).", "Westelijk halfrond (breedtegraad richting).",
"Wizard", "Hulpje",
"Writes the current in-memory EPG database to disk every x hours, so should a crash/unexpected shutdown occur EPG data is saved periodically to the database (re-read on next startup). Set to 0 to disable.", "Schrijft de huidige in het geheugen aanwezige EPG-database elke x uur op de harde schijf, dus als een crash/ onverwacht afsluiten plaatsvindt, worden de EPG-gegevens periodiek bewaard in de database (opnieuw lezen bij het volgende opstarten). Stel in op 0 om uit te schakelen.",
"XMLTV EPG Import", "XMLTV EPG-import",
"Yes", "Ja",
"addpids/delpids supported", "addpids/delpids ondersteund",
"default: nvenc(h264 / h265)", "standaard: nvenc(h264 / h265)",
"fast: h264 / h265 / qsv(h264 / h265)", "snel: h264 / h265 / qsv(h264 / h265)",
"faster: h264 / h265 / qsv(h264)", "sneller: h264 / h265 / qsv(h264)",
"hrs", "uren",
"invalid option specified [%s]", "verkeerde optie opgegeven [%s]",
"logout", "afmelden",
"min", "minuut",
"mins", "minuten",
"option %s requires a value", "optie %s vereist een waarde",
"return", "terugkeren",
"slow: h264 / h265 / qsv(h264 / h265)", "langzaam: h264 / h265 / qsv(h264 / h265)",
"slower: h264 / h265 / qsv(h264)", "langzamer: h264 / h265 / qsv(h264)",
"superfast: h264 / h265", "super snel: h264 / h265",
"tsfile input (mux file)", "tsfile input (mux bestand)",
"ultrafast: h264 / h265", "ultra snel: h264 / h265",
"veryfast: h264 / h265 / qsv(h264)", "zeer snel: h264 / h265 / qsv(h264)",
"veryslow: h264 / h265 / qsv(h264)", "zeer langzaam: h264 / h265 / qsv(h264)",
NULL, NULL
};

static const char *tvh_locale_24[] = {
"# Services", "# Tjenester",
"1 day", "1 dag",
"1 month", "1 måned",
"1 week", "1 uke",
"1 year", "1 år",
"2 months", "2 Måneder",
"2 weeks", "2 uker",
"2 years", "2 år",
"3 days", "3 dager",
"3 months", "3 måneder",
"3 weeks", "3 uker",
"3 years", "3 år",
"5 days", "5 dager",
"6 months", "6 måneder",
"ACTIVE", "AKTIV",
"ATSC source ID", "ATSC kilde ID",
"ATSC-C Network", "ATSC-C Nettverk",
"ATSC-T Network", "ATSC-T Nettverk",
"Access", "Aksess",
"Admin password", "Admin passord",
"Admin username", "Admin brukernavn",
"Advanced", "Avansert",
"Allowed network", "Tillatte nettverk",
"Allowed networks", "Tillate nettverk",
"Altitude (in meters).", "Høye (i meter).",
"Altitude (meters)", "Høyde (meter)",
"Automatically map EPG source", "Automatisk kartlegg EPG kilde",
"Automatically name from network", "Automatisk navn fra nettverk",
"Automatically record.", "Automatisk opptak.",
"Bandwidth", "Båndbredde",
"Based on software from", "Basert på programvare fra",
"Blue", "Blå",
"CA client %i", "CA klient %i",
"CA identification", "CA identifikasjon",
"CA provider", "CA tilbyder",
"CAPMT (Linux Network DVBAPI)", "CAPMT (Linux Nettverk DVBAPI)",
"CRID authority", "CRID Autoritet",
"Cartoons", "Tegnefilm",
"Channel", "Kanal",
"Channel does not exist", "Kanal eksiterer ikke",
"Channel icon", "Kanal ikon",
"Channel icon URL.", "Kanal ikon URL.",
"Channel icon path", "Kanal ikon sti",
"Channel mapping options", "Kanal kartleggings opsjoner",
"Channel name", "Kanal navn",
"Channel not found", "Kanal ikke funnet",
"Channel number", "Kanal nummer",
"Channel numbers from", "Kanal nummer fra",
"Channel update options", "Kanal oppdaterings valg",
"Channels", "Kanaler",
"Character set", "Tegnsett",
"Check availability", "Sjekk tilgjengelighet",
"Children's / Youth programs", "Barne / Ungdoms program",
"Cinema", "Kino",
"Class", "Klasse",
"Classical", "Klassisk",
"Classical music", "Klassisk musikk",
"Client name", "Klient navn",
"Comedy", "Komedie",
"Comment", "Kommentar",
"Commercial break", "Reklamepause",
"Completed OK", "Fullført OK",
"Compress EPG database", "Komprimer EPG database",
"Configuration", "Konfigurasjon",
"Configuration - SAT>IP Server", "Konfigurasjon - SAT>IP Server",
"Configuration ID", "Konfigurasjons ID",
"Configuration name", "Konfigurasjonsnavn",
"Configuration version", "Konfigurasjons versjon",
"Content type", "Innholds type",
"Content type.", "Innholds type.",
"Contest", "Konkuranse",
"Cooking", "Matlaging",
"Copy codec type", "Kopier codec type",
"Created", "Opprettet",
"Current tuner configuration.", "Gjeldende tuner konfigurasjon",
"DAB Network", "DAB nettverk",
"DES Key.", "DES nøkkel.",
"DES key", "DES nøkkel.",
"DVB-C Network", "DVB-C Nettverk",
"DVB-S Network", "DVB-S Nettverk",
"DVB-S Satellite Configuration", "DVB-S Sattelitt konfigurasjon",
"DVB-T Network", "DVB-T Nettverk",
"DVR - Profiles", "DVR - Profiler",
"DVR behavior", "DVR oppførsel",
"DVR configuration", "DVR konfigurasjon",
"DVR configuration profiles", "DVR konfigurasjonsprofiler",
"Dance", "Dans",
"Days of Week", "Ukedager",
"Days of week", "Ukedager",
"Debate", "Debatt",
"Default", "Standard",
"Default language(s)", "Standard språk",
"Default language.", "Standard språk",
"Default priority", "Standard prioritet",
"Demux path", "Demux sti",
"Description", "Beskrivelse",
"Directory", "Katalog",
"Disable", "Deaktivere",
"Disable SAT>IP client", "Deaktivere SAT>IP klient",
"Disable all access control checks", "Deaktivere alle aksess kontroller",
"Disabled", "Deaktivert",
"Discussion", "Diskusjon",
"Do not use", "Ikke bruk",
"Documentary", "Dokumentarer",
"Don't include title in filename", "Ikke inkludert tittel i filnavnet",
"EPG name", "EPG navn",
"EPG scan", "EPG søking",
"EPG source", "EPG kilde",
"Enable (auto)", "Aktivere (auto)",
"Enable DBus", "Aktivere DBus",
"Enable NTP driver", "Aktivere NTP driver",
"Enable or disable full mux mode.", "Aktivere eller deaktivere full mux modus.",
"Enable syslog", "Aktivere syslog",
"Enable/Disable CA client.", "Aktivere/Deaktivere CA klient.",
"Enable/disable LNA.", "Aktivere/Deaktivere LNA.",
"Enable/disable interface quick tips.", "Aktiver/deaktiver interface hurtig tips.",
"Enable/disable libav log output.", "Aktivere/Deaktivere libav logging.",
"Enable/disable logging to syslog.", "Aktivere/Deaktivere logging til Syslog",
"Enable/disable pilot tone.", "Aktiver/deaktiver pilot tone.",
"Enable/disable profile.", "Aktivere/Deaktivere profil.",
"Enable/disable the channel.", "Aktiver/deaktiver kanalen.",
"Enable/disable the device.", "Aktiver/deaktiver enheten.",
"Enable/disable tuner/adapter.", "Aktiver/deaktiver mottaker/adapter.",
"Enabled", "Aktivert",
"Encrypted", "Kryptert",
"Encrypted service", "Kryptert tjeneste",
"Enter an administrator password.", "Skriv inn administrator passordet.",
"Entertainment programs for 10 to 16", "Underholdningsprogram for 10 til 16",
"Entertainment programs for 6 to 14", "Underholdningsprogram for 6 til 14",
"Environment", "Omgivelser",
"Episode number/ID.", "Episode nummer/ID",
"Error code", "Error kode",
"Errors", "Feil",
"Every day", "Hver dag",
"Expert", "Ekspert",
"Exported tuners", "Eksterne mottakere",
"External", "Ekstern",
"External IP (NAT)", "Ekstern IP (NAT)",
"External URL", "Ekstern URL",
"Extra arguments", "Tillegsargumenter",
"Extra start time", "Ekstra starttid",
"FAIL", "FEIL",
"FEC high", "FEC høy",
"FEC low", "FEC lav",
"Failed", "Feilet",
"Fantasy", "Fantasi",
"File missing", "Fil mangler",
"File not created", "Fil ikke opprettet",
"File size", "fil størrelse",
"Filename", "Filnavn",
"Filename options", "Filnavn valg",
"Finished", "Fullført",
"First word", "Første ord",
"Football", "Fotball",
"Force (auto)", "Tving (auto)",
"Force CA ID (e.g. 0x2600)", "Tving CA ID (f.eks 0x2600)",
"Force old status", "Tving gammel status",
"Force priority", "Tving prioritet",
"Found services", "Fant tjenester",
"Frequency", "Frekvens",
"Frequency (Hz)", "Frekvens (Hz)",
"Frequency (MHz)", "Frekvens (MHz)",
"Frequency (in MHz).", "Frekvens (i MHz)",
"Frequency (kHz)", "Frekvens (kHz)",
"Fri", "Fre",
"GOTOX position", "GOTOX posisjon",
"Generic", "Generisk",
"Gray", "Grå",
"High", "Høy",
"Historical movie", "Historisk film",
"IDLE", "TOMGANG",
"IP address", "IP Addresse",
"IPTV Automatic Network", "IPTV automatisk nettverk",
"IPTV Network", "IPTV Nettverk",
"ISDB-C Network", "ISDB-C Nettverk",
"ISDB-S Network", "ISDB-S Nettverk",
"ISDB-T Network", "ISDB-T Nettverk",
"Icon", "Ikon",
"Idle", "Tomgang",
"Idle scan", "Tomgang søk",
"Idle scan muxes", "Tomgang søk muxer",
"Ignore EPG (EIT)", "Ignorer EPG (EIT)",
"Ignore invalid SSL certificate", "Ignorer ugyldig SSL sertifikat",
"Important", "Viktig",
"Include channel name in filename", "Inkluder kanal navn i filnavn",
"Include date in filename", "Inkluder dato i filnavn",
"Include episode in filename", "Inkluder episode i filnavn",
"Include subtitle in filename", "Inkluder undertekst i filnavn",
"Internal", "Intern",
"Internal error", "Intern feil",
"Invalid", "Ugyldig",
"Invalid EPG object request", "Ugyldig EPG objekt forespurt",
"Invalid arguments", "Ugyldige argumenter",
"Invalid file", "Ugyldig fil",
"Invalid service", "Ugyldig tjeneste",
"Language", "Språk",
"Language 1", "Språk 1",
"Language 2", "Språk 2",
"Language 3", "Språk 3",
"Languages", "Språk",
"LinuxDVB network", "LinuxDVB nettverk",
"Location", "Lokasjon",
"Location details of the SAT>IP Server.", "Lokasjonsdetaljer for SAT>IP Server",
"Log", "Logg",
"Low", "Lav",
"MPEG-TS Service", "MPEG-TS Tjeneste",
"MPEG-TS input", "MPEG-TS inngang",
"Manufacturer", "Produsent",
"Manufacturer URL", "Produsent URL",
"Manufacturer's model number.", "Produsentens modellnummer",
"Manufacturer's product description.", "Produsentens produktbeskrivelse",
"Manufacturer's product name.", "Produsentens produktnavn",
"Map all services", "Kartlegg alle tjenester",
"Map encrypted services", "Kartlegg krypterte tjenester",
"Map radio channels", "Kartlegg radiokanaler",
"Map unnamed channels", "Kartlegg kanaler uten navn",
"Master tuner", "Hovedmottaker",
"Names", "Navn",
"Network", "Nettverk",
"Network ", "Nettverk",
"Network A", "Nettverk A",
"Network B", "Nettverk B",
"Network type", "Nettverks type",
"Networks", "Nettverk",
"No", "Nei",
"Number", "Nummer",
"OSCam net protocol (rev >= 10389)", "OSCam net protokoll (rev >= 10389)",
"OSCam new pc-nodmx (rev >= 10389)", "OSCam ny pc-nodmx (rev >= 10389)",
"Off", "Av",
"Older OSCam", "Eldre OSCam",
"On", "På",
"Only Bulsatcom 39E", "Bare Bulsatcom 39E",
"Only EIT", "Bare EIT",
"Only OpenTV Sky Ausat", "Bare OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Bare OpenTV Sky Italia",
"Only OpenTV Sky UK", "Bare OpenTV Sky UK",
"Only PSIP (ATSC)", "Bare PSIP (ATSC)",
"Only UK Freesat", "Bare UK Freesat",
"Only UK Freeview", "Bare UK Freeview",
"Only Viasat Baltic", "Bare Viasat Baltic",
"Only digits", "Bare tall",
"Oscam mode.", "Oscam modus.",
"Password", "Passord",
"Password2", "Passord2",
"Username", "Brukernavn",
NULL, NULL
};

static const char *tvh_locale_25[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nPo więcej informacji odwiedź stronę www:\nhttps://tvheadend.org\n",
"# Mapped channels", "# Zamapowane kanały",
"# Muxes", "# Muxy",
"# Services", "# Usługi",
"# tune repeats", "Powtarzanie strojenia",
"(Default profile)", "(Domyślny profil)",
"(Read-only)", "(Tylko do odczytu)",
"1 day", "1 dzień",
"1 month", "1 miesiąc",
"1 week", "1 tydzień",
"1 year", "1 rok",
"2 months", "2 miesiące",
"2 weeks", "2 tygodnie",
"2 years", "2 lata",
"2-Port switch (universal LNB)", "Przełącznik 2-portowy (universal LNB)",
"3 days", "3 dni",
"3 months", "3 miesiące",
"3 weeks", "3 tygodnie",
"3 years", "3 lata",
"4-Port switch (universal LNB)", "Przełącznik 4-portowy (universal LNB)",
"5 days", "5 dni",
"6 months", "6 miesięcy",
"AC-3 detection", "Wykrywanie AC-3",
"ACTIVE", "AKTYWNY",
"ANY", "DOWOLNY",
"ATSC-C Network", "Sieć ATSC-C",
"ATSC-T Network", "Sieć ATSC-T",
"Aborted by user", "Przerwane przez użytkownika",
"Access (ACL)", "Dostęp (ACL)",
"Action", "Akcja",
"Adapter in use by another subscription", "Adapter w użyciu przez inną subskrypcję",
"Add file and line numbers to debug", "Dodaj plik i numer linii do debugowania",
"Add the thread ID to debug", "Dodaj ID wątku do debugowania",
"Admin password", "Hasło Admin",
"Admin username", "Nazwa użytkownika Admin",
"Administrator login", "Logowanie Administratora",
"Adult movie", "Film dla dorosłych",
"Advanced", "Zaawansowane",
"Advanced (non-universal LNBs, rotors, etc.)", "Zaawansowane (non-universal LNBs, rotors, itp.)",
"Adventure", "Przygodowy",
"Advertisement / Shopping", "Reklamy / Zakupy",
"All (Streaming plus DVR)", "Wszystko (Strumieniowanie + DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Wszystkie znaki, które mogłyby powodować problemy dla nazewnictwa plików zostaną zastąpione znakiem podkreślenia. Zobacz Pomoc dla szczegółów.",
"All proceeds are used to support server infrastructure and buy test equipment.", "Wszystkie wpływy są wykorzystywane do wspierania infrastruktury serwerowej i zakupu sprzętu do badań.",
"Allowed DVR profiles. This limits the profiles the user has access to.", "Dozwolone profile DVR. Do tych ograniczonych profili użytkownik ma dostęp.",
"Allowed network", "Dozwolona sieć",
"Allowed networks", "Dozwolone sieci",
"Alternate PID path", "Alternatywna ścieżka pid",
"Alternate configuration path", "Alternatywna ścieżka konfiguracji",
"Altitude (in meters).", "Wysokość (w metrach).",
"Altitude (meters)", "Wysokość (metry)",
"Animals", "Zwierzęta",
"Anonymize HTSP access", "Dostęp anonimowy HTSP",
"Any", "Dowolny",
"Arts", "Sztuka",
"Assign predefined muxes to networks", "Przypisywanie predefiniowanych muxów do sieci",
"Athletics", "Gimnastyka",
"Audio Stream Filter", "Filtr strumienia audio",
"Audio channel layout.", "Układ kanałów audio.",
"Audio codec", "Audio kodek",
"Auto check disabled", "Auto sprawdzanie wyłączone",
"Auto check enabled", "Auto sprawdzanie włączone",
"Auto record", "Automatyczne nagrywanie",
"Auto record caption", "Auto nagrywanie podpisu",
"Auto recording%s%s", "Auto nagrywanie%s%s",
"Auto time record", "Czas automatycznego nagrywania",
"Auto-Map to channels", "Automatycznie mapuj do kanałów",
"Automatic checking", "Automatyczne sprawdzanie",
"Automatic schedule entry not found", "Automatycznie zaplanowany wpis nie odnaleziony",
"Automatic time scheduler entry not found", "Automatycznie zaplanowany wpis czasu nie odnaleziony",
"Automatically map EPG source", "Automatycznie mapowane źródło EPG",
"Automatically map all available services to channels.", "Automatycznie mapuj wszystkie dostępne serwisy do kanałów.",
"Automatically name from network", "Automatyczna nazwa z sieci",
"Autorec maximum count (0=unlimited)", "Auto-nagrywanie maksymalny licznik (0=bez limitu)",
"Autorec maximum schedules limit (0=unlimited)", "Auto-nagrywanie maksymalny limit harmonogramu (0=bez limitu)",
"Bad", "Zły",
"Ballet", "Balet",
"Bandwidth", "Przepustowość",
"Based on software from", "Bazuje na oprogramowaniu z",
"Basic", "Podstawowy",
"Bind to specific local IP address.", "Powiąż ze specjalnym lokalnym adresem IP.",
"Blue", "Niebieski",
"Bouquet (auto)", "Bouquet (Automatyczny)",
"Brand", "Marka",
"Broadcast", "Emisja",
"Broadcast type", "Typ emisji",
"Broadcast.", "Emisja.",
"Broadcasting", "Transmitowanie",
"Buffering limit (ms)", "Limit buforowania (ms)",
"Build", "Wersja",
"CA client %i", "CA klient %i",
"CA identification", "CA identyfikacja",
"CA provider", "CA dostawca",
"CAPMT interval (in ms).", "CAPMT interwał (w ms).",
"CAPMT interval (ms)", "CAPMT interwał (ms)",
"CAPMT query interval (ms)", "CAPMT interwał zapytań (ms)",
"CAPMT query interval (ms).", "CAPMT interwał zapytań (ms).",
"CRID authority", "CRID organ",
"Cache scheme", "Schemat buforowania",
"Camd.socket filename / IP Address (TCP mode)", "Camd.socket nazwa pliku / adres IP (tryb TCP)",
"Cartoons", "Kreskówki",
"Change parameters", "Zmiana parametrów",
"Channel", "Kanał",
"Channel does not exist", "Kanał nie istnieje",
"Channel icon", "Ikona kanału",
"Channel icon URL.", "URL ikony kanału",
"Channel icon as defined in EPG data.", "Ikona kanału, jak określono w danych EPG.",
"Channel icon name scheme", "Schemat nazwy ikony kanału",
"Channel icon path", "Ścieżka ikon kanałów",
"Channel mapping options", "Opcje mapowania kanału",
"Channel name", "Nazwa kanału",
"Channel not found", "Kanał nie znaleziony",
"Channel number", "Numer kanału",
"Channel number as defined in EPG data.", "Numer kanału, jak określono w danych EPG.",
"Channel number offset", "Numer kanału offset",
"Channel numbers from", "Numer kanału od",
"Channel tag", "Tag kanału",
"Channel tag does not exist", "Tag kanału nie istnieje",
"Channel tag reference", "Referencyjny tag kanału",
"Channel tags", "Tagi kanałów",
"Channel tags the user is allowed access to/excluded from.", "Tagi kanału do których użytkownik ma prawo dostępu/wykluczenia.",
"Channel to use/used for the recording.", "Kanał do użycia/używany do nagrywania.",
"Channels", "Kanały",
"Channels EPG data is used by.", "Dane kanałów EPG są używane przez.",
"Character set", "Ustawienie znaków",
"Check availability", "Sprawdź dostępność",
"Children's / Youth programs", "Programy dla dzieci / młodzieży",
"Cinema", "Kino",
"Class", "Klasa",
"Classical", "Klasyczny",
"Classical music", "Muzyka klasyczna",
"Clean effects", "Czyść efekty",
"Client name", "Nazwa klienta",
"Clone scheduled entry on error", "Klonuj zaplanowane wpisy przy błedzie",
"Comedy", "Komedia",
"Command delay time (ms) (10-200)", "Cmd czas opóźnienia (ms) (10-200)",
"Command time (ms) (10-100)", "Command Time (ms) (10-100)",
"Comment", "Komentarz",
"Commercial break", "Przerwa reklamowa",
"Completed OK", "Zakończone OK",
"Compress EPG database", "Kompresuj bazę danych EPG",
"Compress the EPG database to reduce disk I/O and space.", "Kompresuj bazę danych EPG, aby zmniejszyć przestrzeń dysku I/O.",
"Conditional Access", "Dostęp warunkowy",
"Conditional Access Client", "Dostęp warunkowy klienta",
"Conditional Access Identification.", "Identyfikator dostępu warunkowego",
"Configuration", "Konfiguracja",
"Configuration - Base", "Konifugacja - Baza",
"Configuration - Image Cache", "Konfiguracja - Pamięć podręczna obrazów",
"Configuration ID", "ID konfiguracji",
"Configuration name", "Konfiguracja nazwy",
"Configuration version", "Wersja konfiguracji",
"Connection limit type", "Limit typu połączenia",
"Constellation", "Konstelacja",
"Container", "Kontener",
"Content character set", "Ustawienie znaków zawartości",
"Content type", "Typ zawartości",
"Content type.", "Typ zawartości.",
"Contest", "Zawody",
"Continue recording for x minutes after scheduled stop time", "Kontynuuj nagrywanie przez x minut po planowanym czasie zatrzymania",
"Continue recording for x minutes after scheduled stop time.", "Kontynuuj nagrywanie przez x minut po planowanym czasie zatrzymania.",
"Cookie expiration (days)", "Ważności cookies (dni)",
"Cooking", "Gotowanie",
"Copy codec type", "Kopiuj typ kodeka",
"Copy layout", "Kopiuj układ",
"Create SDTV/HDTV/Radio tags.", "Utwórz tagi SDTV/HDTV/Radio.",
"Create a provider name tag.", "Utwórz tag dostawcy.",
"Create bouquet", "Utwórz bouquet",
"Create directories using these permissions.", "Utwórz katalogi używając tych uprawnień.",
"Create files using these permissions.", "Utwórz pliki używając tych uprawnień.",
"Create network name tags", "Utwórz nazwy tagów sieci",
"Create network name tags (set by provider).", "Utwórz nazwy tagów sieci (ustaw po dostawcy).",
"Create network tags", "Utwórz tagi sieci",
"Create provider name tags", "Utwórz nazwy tagów dostawców",
"Create provider tags", "Utwórz tagi dostawców",
"Created", "Utworzony",
"Creator", "Autor",
"Cron multi-line", "Wielo-liniowy Cron",
"Culture (without music)", "Kultura (bez muzyki)",
"Current RTSP port.", "Obecny port RTSP.",
"Current affairs", "Sprawy bieżące",
"Current number of objects.", "Aktualna liczba obiektów.",
"Current object size.", "Aktualny rozmiar obiektu.",
"Current tuner configuration.", "Obecna konfiguracja tunera.",
"Custom HTTP headers", "Własny nagłówek HTTP",
"DES key", "Klucz DES",
"DSCP/TOS for streaming", "DSCP/TOS dla streamingu",
"DVB scan files", "DVB skan plików",
"DVB scan files path", "DVB ścieżka skanowania plików",
"DVR - Profiles", "DVR - profile",
"DVR behavior", "DVR zachowanie",
"DVR configuration", "DVR konfiguracja ",
"DVR configuration profiles", "DVR profile konfiguracji",
"DVR file retention period", "Okres przechowywania plików DVR",
"DVR log retention", "DVR pamięć logów",
"DVR log retention period", "Okres przechowywania logów DVR",
"DVR override: high", "DVR nadpisywanie: wysoki",
"DVR override: important", "DVR nadpisywanie: nieważny",
"DVR override: low", "DVR nadpisywanie: niski",
"DVR override: normal", "DVR nadpisywanie: normalny",
"DVR override: unimportant", "DVR nadpisywanie: nieważny",
"DVR schedule does not exist", "Harmonogram DVR nie istnieje",
"DVR schedule does not have a file yet", "Harmonogram DVR nie posiada plików",
"DVR schedule not found", "Brak harmonogramu DVR",
"Dance", "Taniec",
"Data errors", "Dane błędów",
"Days of Week", "Dni tygodnia",
"Days of the week to which the rule should apply.", "Dni tygodnia w których dział reguła.",
"Days of week", "Dni tygodnia",
"Debate", "Debata",
"Debug libav log", "Debuguj log libav",
"Debug options", "Opcje debugowania",
"Debug subsystems", "Debuguj podsystemy",
"Debug to syslog", "Debuguj do syslog",
"Debug trace (low-level)", "Wyszukiwanie błędów (niski poziom)",
"Debugging", "Debugowanie",
"Default", "Domyślny",
"Default language(s)", "Domyślny(e) język(i)",
"Default language.", "Domyślny język.",
"Default priority", "Domyślny priorytet",
"Default user interface level.", "Domyślny poziom interfejsu użytkownika.",
"Delivery system", "System przeznaczenia",
"Demux path", "Ścieżka Demux",
"Descramble services (limit per mux)", "Usługi dekodujące (limit na muxa)",
"Descrambler buffer (TS packets)", "Bufor deszyfratora (pakiety TS)",
"Description", "Opis",
"Detective", "Detektyw",
"Device ID", "ID Urządzenia",
"Device model", "Model urządzenia",
"Device path", "Ścieżka urządzenia",
"Directory", "Katalog",
"Directory permissions (octal, e.g. 0775)", "Uprawnienia katalogu (ósemkowe, np. 0775)",
"Disable", "Wyłącz",
"Disable DVB bouquets", "Wyłącz DVB bouquets",
"Disable SAT>IP client", "Wyłącz SAT>IP klient",
"Disable all access control checks", "Wyłącz wszystkie sprawdzania kontroli dostępu",
"Disable debug on stderr", "Wyłącz debugowanie na stderr",
"Disable syslog (all messages)", "Wyłącz syslog (wszystkie wiadomości)",
"Disabled", "Wyłączony",
"Discussion", "Dyskusja",
"Do not send any stream specific information to the HTSP client like signal strength, input source etc.", "Nie wysyłaj do klienta HTSP informacji o strumieniu takich jak: siła sygnału, żródło itp.",
"Do not use", "Nie używaj",
"Documentary", "Dokumentalny",
"Don't backup configuration tree at upgrade", "Nie wykonuj backupu drzewa konfiguracji przy aktualizacji",
"Don't include title in filename", "Nie dołączaj tytułu do nazwy pliku",
"Don't keep", "Nie trzymaj",
"Don't re-record", "Nie nagrywaj ponownie",
"Don't reschedule", "Nie planuj ponownie",
"Drama", "Dramat",
"Duplicate handling", "Duplicate Handling",
"EPG Language (priority order)", "Język EPG (priorytet zamówienia)",
"EPG data ID.", "ID danych EPG.",
"EPG name", "Nazwa EPG",
"EPG scan", "EPG skan",
"EPG source", "Źródło EPG",
"EPG update window", "Okno aktualizacji EPG",
"Economics", "Ekonomika",
"Education", "Edukacja",
"Educational", "Edukacyjny",
"Elementary stream filter", "Filtr strumienia podstawowego",
"Enable (auto)", "Włączony (automatyczny)",
"Enable DBus", "Włącz DBus",
"Enable NTP driver", "Włącz sterownik NTP",
"Enable coredumps for daemon", "Włącz coredumps dla demona",
"Enable debug on stderr", "Włącz debugowanie na stderr",
"Enable debug subsystems", "Włącz debugowanie podsystemów",
"Enable debug to file", "Włącz debugowanie do pliku",
"Enable debug to syslog", "Włącz debugowanie do sysloga",
"Enable or disable fast input switching.", "Włącz lub wyłącz szybkie przełączanie wejścia.",
"Enable or disable full mux mode.", "Włącz lub wyłącz tryb full mux.",
"Enable or disable this configuration.", "Włącz lub wyłącz tą konfigurację.",
"Enable syslog", "Włącz syslog",
"Enable this filter.", "Włącz ten filtr.",
"Enable trace subsystems", "Włącz śledzenie podsystemów",
"Enable web UI debug (non-minified JS)", "Włącz debugowanie web UI (non-minified JS)",
"Enable/disable LNA.", "Włącz/wyłącz LNA.",
"Enable/disable auto-rec rule.", "Włącz/wyłącz zasady automatycznego zapisu.",
"Enable/disable interface quick tips.", "Włącz/wyłącz szybkie wskazówki interfejsu.",
"Enable/disable profile.", "Włącz/wyłącz profil.",
"Enable/disable the channel.", "Włącz/wyłącz kanał.",
"Enable/disable the device.", "Włącz/wyłącz urządzenie.",
"Enable/disable the entry.", "Włącz/wyłącz wpis.",
"Enable/disable the grabber.", "Włącz/wyłącz grabber.",
"Enable/disable tuner/adapter.", "Włącz/wyłącz tuner/adapter.",
"Enabled", "Włączony",
"Encrypted", "Zaszyfrowany",
"Encrypted service", "Usługa kodowana",
"Entertainment programs for 10 to 16", "Programy rozrywkowe od 10 do 16",
"Entertainment programs for 6 to 14", "Programy rozrywkowe od 6 do 14",
"Environment", "Środowisko",
"Environment (pipe)", "Środowisko (pipe)",
"Episode", "Odcinek",
"Episode %d", "Odcinek %d",
"Error code", "Kod błędu",
"Errors", "Błędy",
"Even key", "Parzysty klucz",
"Every day", "Codziennie",
"Exclude channel tags", "Wyklucz tagi kanałów",
"Expeditions", "Wyprawy",
"Experimental film", "Eksperymentalny film",
"Expert", "Ekspert",
"Exported tuners", "Eksportowane tunery",
"External", "Zewnętrzny",
"External EPG grabber", "Zewnętrzny Grabber EPG",
"External IP (NAT)", "Zewnętrzne IP (NAT)",
"External URL", "Zewnętrzny URL",
"Extra arguments", "Ekstra czas rozpoczęcia",
"Extra start time", "Ekstra czas rozpoczęcia",
"Extra stop time", "Ekstra czas zakończenia",
"Extra warming up time (seconds)", "Ekstra czas rozgrzania (sekundy)",
"FAIL", "Błąd",
"FEC high", "FEC High",
"FEC low", "FEC Low",
"Factual topics", "Faktyczne tematy",
"Failed", "Błędne",
"Fast input switch", "Przełącz szybkie wprowadzanie",
"File missing", "Plik utracony",
"File not created", "Plik nie utworzony",
"File permissions (octal, e.g. 0664)", "Uprawnienia pliku (ósemkowe, np. 0664)",
"File removed", "Plik usunięty",
"File size", "Rozmiar pliku",
"Filename", "Nazwa pliku",
"Filename character set", "Ustawienie znaków nazwy plików",
"Filename options", "Opcje nazwy plików",
"Fine arts", "Piękne sztuki",
"Finished", "Zakończone",
"Fit to RAM (cut rewind)", "Dopasuj do RAM (cut rewind)",
"Fitness and health", "Fitness i zdrowie",
"Folkloric", "Folklorystyczny",
"Football", "Piłka nożna",
"Force (auto)", "Wymuś (automatycznie)",
"Force CA ID (e.g. 0x2600)", "Wymuś CA ID (np. 0x2600)",
"Force old status", "Wymuś stary status",
"Force pilot for DVB-S2", "Wymuś pilota dla DVB-S2",
"Force priority", "Wymuś priorytet",
"Force teardown delay", "Wymuś czas zwłoki porzucenia",
"Foreign countries", "Obce kraje",
"Fork and run as daemon", "Fork i uruchom jako demon",
"Format string", "Format składni",
"Frequency", "Częstotliwość",
"Frequency (Hz)", "Częstotliwość (Hz)",
"Frequency (MHz)", "Częstotliwość (MHz)",
"Frequency (kHz)", "Częstotliwość (kHz)",
"Fri", "Pt.",
"Friendly name", "Przyjazna nazwa",
"Frontend number", "Numer nakładki",
"Frontend path", "Ścieżka nakładki",
"Full mux RX mode supported", "Tryb obsługi pełnego mux Rx",
"Full pathname specification", "Pełna specyfikacja nazwy ścieżki",
"Full-text", "Pełny tekst",
"Further education", "Dalsza edukacja",
"GOTOX position", "Pozycja GOTOX",
"Game show", "Teleturniej",
"Gardening", "Ogrodnictwo",
"General", "Ogólne",
"General configuration", "Ogólna konfiguracja",
"Generic", "Ogólne",
"Generic options", "Opcje ogólne",
"Gray", "Szary",
"Guard interval", "Straż interwału",
"HDHomeRun DVB frontend", "Nakładka HDHomeRun DVB",
"HDHomeRun DVB-C frontend", "Nakładka HDHomeRun DVB-C",
"HDHomeRun DVB-T frontend", "Nakładka HDHomeRun DVB-T",
"HTSP Default Stream Settings", "HTSP domyślne ustawienia strumienia",
"HTTP Client", "Klient HTTP",
"Handicraft", "Rękodzieła",
"Hearing impaired", "Niedosłyszący",
"Hierarchy", "Hierarchia",
"High", "Wysoki",
"Historical movie", "Film historyczny",
"Hostname/IP", "Nazwa hosta/IP",
"IDLE", "BEZCZYNNY",
"IP address", "Adres IP",
"IP blocking", "Blokowanie IP",
"IPTV Multiplex", "IPTV multiplex",
"IPTV input", "IPTV wejście",
"ISI (Stream ID)", "ISI (Strumieniuj ID)",
"Icon", "Ikona",
"Icon (full URL)", "Ikona (pełny URL)",
"Icon URL", "URL ikony",
"Icon base URL", "Podstawowy URL ikony",
"Icon has title", "Ikona ma tytuł",
"Icons from", "Ikony z",
"Idle", "Bezczynny",
"Idle scan", "Bezczynne skanowanie",
"Idle scan muxes", "Bezczynne skanowanie muxów",
"If you'd like to support the project, please consider a donation.", "Jeśli chcesz wesprzeć projekt, proszę rozważ darowiznę.",
"Ignore EPG (EIT)", "Ignoruj EPG (EIT)",
"Ignore invalid SSL certificate", "Ignoruj nieprawidłowy certyfikat SSL",
"Ignore provider's channel numbers", "Ignoruj numery kanałów dostwców",
"Immediately abort", "Natychmiast przerwać",
"Important", "Ważny",
"Include channel name in filename", "Zawieraj nazwę kanału w nazwie pliku",
"Include date in filename", "Zawieraj datę w nazwie pliku",
"Include episode in filename", "Zawieraj epizod w nazwie pliku",
"Include subtitle in filename", "Zawieraj podtytuł w nazwie pliku",
"Include time in filename", "Zawieraj godzinę w nazwie pliku",
"Index", "Indeks",
"Information area", "Informacje o przestrzeni",
"Informational", "Informacyjny",
"Initial scan", "Wstępne skanowanie",
"Input base", "Podstawowe wejście",
"Input buffer (bytes)", "Bufor wejścia (bajty)",
"Input instance", "Wejście instancji",
"Input path", "Ścieżka wejścia",
"Interface", "Interfejs",
"Internal", "Wewnętrzny",
"Internal EPG grabber", "Wbudowany grabber EPG",
"Internal grabber", "Wbudowany grabber",
"Interview", "Wywiad",
"Invalid", "Nieprawidłowy",
"Invalid adapter number '%s'\n", "Błędny numer adaptera '%s'\n",
"Invalid target", "Nieprawidłowy cel",
"Keep", "Zachowaj",
"Kill signal (pipe)", "Zabij sygnał (pipe)",
"Kill timeout (pipe/secs)", "Zabij timeout (pipe/secs)",
"LNA (low noise amplifier)", "LNA (Low noise amplifier)",
"LNB type", "LNB typ",
"Language", "Język",
"Language settings", "Ustawienia języka",
"Languages", "Języki",
"Last seen", "Ostatnio widziany",
"Last updated from", "Ostatnia aktualizacja",
"Leisure hobbies", "Hobby czas wolny",
"Limit connections", "Limit połączeń",
"Linked input", "Powiązane wejścia",
"LinuxDVB network", "LinuxDVB sieć",
"Listen / Connect port", "Słuchaj/Połącz port",
"Listen on IPv6", "Nasłuchuj na IPv6",
"Literature", "Literatura",
"Local bind IP address", "Powiąż lokalny adres IP",
"Local channel number", "Lokalny numer kanału",
"Local discovery IP address", "Odkryj lokalny adres IP",
"Location", "Lokalizacja",
"Lock preferred CA PID", "Blokada preferowanych CA PID",
"Login/Logout", "Zaloguj/Wyloguj",
"Low", "Niski",
"Magazines", "Magazyny",
"Maintained space", "Utrzymywanie miejsca",
"Make subdirectories per channel", "Utwórz podkatalog dla kanału",
"Make subdirectories per day", "Utwórz podkatalog dla dnia",
"Make subdirectories per title", "Utwórz podkatalog dla tytułu",
"Manufacturer", "Producent",
"Manufacturer URL", "URL producenta",
"Map all services", "Mapuj wszystkie serwisy",
"Map radio channels", "Mapuj kanały radiowe",
"Map unnamed channels", "Mapuj nie nazwane kanały",
"Master tuner", "Główny tuner",
"Matroska specific", "Matroska specyfikacja",
"Maximal channel number", "Maksymalny numer kanału",
"Maximum # input streams", "Maksymalnie # wejścia strumieni",
"Maximum PIDs", "Maksymalny PIDs",
"Maximum RAM size (MB)", "Max. rozmiar pamięci RAM (MB)",
"Maximum bandwidth (Kbps)", "Maksymalna przepustowość (Kbps)",
"Maximum count (0=default)", "Maksymalny licznik (0=bez limitu)",
"Maximum duration", "Maksymalny czas trwania",
"Maximum length of PIDs", "Maksymalna długość PIDs",
"Maximum period (mins)", "Max. okres (minuty)",
"Maximum schedules limit (0=default)", "Maksymalny limit harmonogramu (0=bez limitu)",
"Maximum size (MB)", "Max. rozmiar (MB)",
"Maximum timeout (seconds)", "Maksymalny timout (sekundy)",
"Medicine", "Medycyna",
"Melodrama", "Melodramat",
"Minimal channel number", "Minimalny numer kanału",
"Minimum duration", "Minimalny czas trwania",
"Minimum rotor time (seconds)", "Minimum rotor time (sekundy)",
"Mode", "Tryb",
"Model description", "Opis modelu",
"Model name", "Nazwa modelu",
"Model number", "Numer modelu",
"Modulation", "Modulacja",
"Module", "Moduł",
"Module ID", "ID Modułu",
"Mon", "Pon.",
"More verbose libav log", "Więcej komunikatów logu libav",
"Motor sport", "Sporty motorowe",
"Motoring", "Jazda samochodem",
"Movie", "Film",
"Music", "Muzyka",
"Musical", "Muzyczny",
"Mux handling", "Obsługa muxa",
"Mux name", "Nazwa muxa",
"Mux not enabled", "Mux nie włączony",
"NONE", "ŻADEN",
"Name", "Nazwa",
"Names", "Nazwy",
"Natural sciences", "Nauki przyrodnicze",
"Nature", "Natura",
"Network", "Sieć",
"Network A", "Sieć A",
"Network B", "Sieć B",
"Network ID (limit scanning)", "ID sieci (limit skanowania)",
"Network UUID", "UUID sieci",
"Network discovery", "Odkryj sieć",
"Network name", "Nazwa sieci",
"Network prefix", "Prefix sieci",
"Network type", "Typ sieci",
"Networks", "Sieci",
"New / premiere / unknown", "Nowość / premiera / nieznany",
"News", "Informacje",
"News magazine", "Magazyn informacyjny",
"Next tune delay in ms (0-2000)", "Opóźnienie następnego strojenia w ms (0-2000)",
"No", "Nie",
"No PIN", "Brak PINu",
"No access", "Brak dostępu",
"No adapters specified!\n", "Brak sprecyzowanych adapterów!\n",
"No assigned adapters", "Brak przypisanych adapterów",
"No days", "Brak dni",
"No descrambler", "Brak deszyfratora",
"No free adapter", "Brak wolnego adaptera",
"No input detected", "Brak wykrytego wejścia",
"No scheme", "Brak schematu",
"No service assigned to channel", "Brak usług przypisanych do kanału",
"No service enabled", "Brak włączonych usług",
"No source available", "Brak dostępnego źródła",
"None", "Żaden",
"Normal", "Normalny",
"Not linked", "Nie powiązane",
"Not set", "Nie ustawiony",
"Not set (none or channel configuration)", "Nie ustawiony (brak lub konfiguracja kanału)",
"Not set (use channel or DVR configuration)", "Nie ustawiony (użyj kanału lub konfiguracji DVR)",
"Number", "Numer",
"Number of tsfile tuners", "Numer tunera tsfile",
"OK (partial)", "OK (częściowe)",
"Odd key", "Nieparzysty klucz",
"Off", "Wył.",
"Older OSCam", "Starszy OSCam",
"On", "Wł.",
"On-demand (no first rewind)", "Na żądanie (no first rewind)",
"Only Bulsatcom 39E", "Tylko Bulsatcom 39E",
"Only EIT", "Tylko EIT",
"Only OpenTV Sky Ausat", "Tylko OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Tylko OpenTV Sky Italia",
"Only OpenTV Sky UK", "Tylko OpenTV Sky UK",
"Only PSIP (ATSC)", "Tylko PSIP (ATSC)",
"Only UK Freesat", "Tylko UK Freesat",
"Only UK Freeview", "Tylko UK Freeview",
"Only Viasat Baltic", "Tylko Viasat Baltic",
"OpenTV channel number", "Numer Kanału OpenTV",
"Orbital position", "Pozycja orbitalne",
"Orbital positions", "Pozycje orbitalne",
"Original network ID", "Oryginalne ID sieci",
"Output directory for tsdebug", "Katalog wyjściowy dla tsdebug",
"Over-the-air Cron multi-line", "Wielo-liniowy Cron Over-the-air",
"Owner", "Właściciel",
"PEND", "WSTRZYMANY",
"PIDs 21 in setup", "PIDs 21 w ustawieniach",
"PLS code", "Kod PLS",
"PLS mode", "Tryb PLS",
"Parent entry", "Wpis rodzica",
"Password", "Hasło",
"Password2", "Hasło 2",
"Path", "Ścieżka",
"Performing arts", "Koncerty i spektakle",
"Periodically save EPG to disk (hours)", "Okresowy zapis EPG na dysku (godziny)",
"Persistent user interface level", "Stały poziom intefejsu użytkownika",
"Physiology", "Fizjologia",
"Picon name scheme", "Nazwa schematu Picon",
"Picon path", "Ścieżka Picon",
"Polarization", "Polaryzacja",
"Political issues", "Kwestie polityczne",
"Popular culture", "Popularna kultura",
"Position", "Pozycja",
"Post-processor command", "Komenda po procesie",
"Post-recording padding", "Nagrywanie po",
"Post-remove command", "Komenda po usunięciu",
"Power save", "Oszczędzanie energii",
"Power-up time (ms) (15-200)", "Czas pierwszego uruchomienia (ms) (15-200)",
"Pre-defined muxes", "Predefiniowane muxy",
"Pre-recording padding", "Nagrywanie przed",
"Prefer picons over channel name", "Preferuj Picony zamiast nazw kanałów",
"Preferred CA PID", "Preferowany CA PID",
"Preferred service video type", "Preferowany rodzaj usługi wideo",
"Presentation", "Prezentacja",
"Press", "Prasa",
"Priority", "Priorytet",
"Priority (-10..10)", "Priorytet (-10..10)",
"Private", "Prywatny",
"Profile name", "Nazwa profilu",
"Provider", "Dostawca",
"Provider ID", "ID Dostawcy",
"Provider network name", "Nazwa dostawcy sieci",
"Psychology", "Psychologia",
"Puppets", "Kukiełki",
"RAM only", "Tylko RAM",
"RTSP Protocol", "Protokół RTSP ",
"RTSP port", "Port RTSP",
"RTSP port (554 or 9983, 0 = disable)", "RTSP port (554 lub 9983), 0 = wyłaczony",
"Re-fetch period (hours)", "Ponowne pobieranie (godziny)",
"Re-fetch period (mins)", "Okres ponownego pobierania (minuty)",
"Re-record%s%s", "Nagraj ponownie%s%s",
"Re-try period (hours)", "Ponowny okres prób (godziny)",
"Record all", "Nagraj wszystko",
"Recording file options", "Opcje nagrywania plików",
"Recording system path", "Systemowa ścieżka nagrywania",
"Reject", "Odrzuć",
"Religion", "Religia",
"Religious", "Religijny",
"Remarkable people", "Niezwykli ludzie",
"Remove HTTP arguments", "Usuń argument HTTP",
"Remove all unsafe characters from filename", "Usuń wszystkie niedozwolone znaki z nazwy pliku",
"Replace whitespace in title with '-'", "Zastąp odstępy w tytułach znakiem '-'",
"Reports", "Raporty",
"Rerun of", "Powtórka",
"Reserved", "Zarezerwowany",
"Resolution (height)", "Rozdzielczość (wysokość)",
"Restart on error", "Restartuj przy błędzie",
"Reuse EPG from", "Ponownie użyj EPG od",
"Rewrite EIT", "Przepisz EIT",
"Rewrite PAT", "Przepisz PAT",
"Rewrite PMT", "Przepisz PMT",
"Rewrite SDT", "Przepisz SDT",
"Romance", "Romans",
"Rotor initialization time (seconds)", "Czas inicjalizacji rotora (sekundy)",
"Rotor type", "Typ rotora",
"Run as group", "Uruchom jako grupa",
"Run as user", "Uruchom jako użytkownik",
"Running", "Uruchomiony",
"SAT>IP DVB-T frequency (Hz)", "SAT>IP DVB-T częstotliwość (Hz)",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "SAT>IP RTSP numer portu dla serwera\n(default: -1 = wyłączony, 0 = webconfig, standardowy port 554)",
"SAT>IP source number", "SAT>IP źródłowy numer",
"SSL verify peer", "Weryfikacja SLL",
"Sat", "Sob.",
"Satellite longitude", "Długość satelity",
"Satellite positions", "Pozycje satelity",
"Scan after creation", "Skanuj po utworzeniu",
"Scan queue length", "Długość skanowania kolejki",
"Scan result", "Wynik skanowania",
"Scan status", "Status skanowania",
"Schedule a re-recording if more errors than (0=off)", "Harmonogram ponownego nagrywania jeśli błędów więcej niż (0=wył.)",
"Schedule status", "Status harmonogramu",
"Scheduled for recording", "Zaplanowane do nagrywania",
"Scheduled start time", "Zaplanowany czas rozpoczęcia",
"Scheduled stop time", "Zaplanowany czas zakończenia",
"School programs", "Programy szkolne",
"Science", "Nauka",
"Season", "Sezon",
"Season %d", "Sezon %d",
"Send CAPMT query", "Wyślij zapytanie CAPMT",
"Send full PLAY cmd", "Wyślij pełny cmd PLAY",
"Serial number", "Numer seryjny",
"Serious", "Poważna",
"Serious music", "Muzyka poważna",
"Server", "Serwer",
"Server UUID", "Serwer UUID",
"Server connectivity", "Łączność serwera",
"Service", "Usługa",
"Service ID", "ID Usługi",
"Service configuration", "Konfiguracja usługi",
"Service count %zi", "Licznik usług %zi",
"Service name", "Nazwa usługi",
"Service name picons", "Nazwa serwisu picons",
"Service raw", "Usługa raw",
"Service type", "Typ usługi",
"Services", "Usługi",
"Settings", "Ustawienia",
"Show this page", "Pokaż tą stronę",
"Show version information", "Pokaż informacje wersji",
"Signal quality too poor", "Zbyt słaba jakość sygnału",
"Signal scale (240 or 100)", "Skala Sygnału (240 lub 100)",
"Site latitude", "Strona szerokości",
"Site longitude", "Strona długości",
"Size", "Rozmiar",
"Skip TS packets (0-200)", "Pomiń pakiety TS (0-200)",
"Skip commercials", "Pomiń reklamy",
"Skip initial bytes", "Pomiń wstępne bajty",
"Slave entry", "Slave wpis",
"Slot state", "Stan gniazda",
"Soap", "Mydło",
"Soccer", "Piłka nożna",
"Social", "Społeczny",
"Social advisory", "Doradztwo społeczne",
"Sort index", "Sortuj index",
"Source", "Źródło",
"Source deleted", "Źródło usunięte",
"Source quality is bad", "Źródło jest złej jakości",
"Source reconfigured", "Źródło przekonfigurowane",
"Special events (Olympic Games, World Cup, etc.)", "Wydarzenia specjalne (Igrzyska Olimpijskie, Mistrzostwa Świata, itp)",
"Specify User-Agent header for the http client", "Określ nagłówek User-Agent dla klienta http",
"Specify alternative htsp port", "Określ alternatywny port htsp",
"Specify alternative http port", "Określ alternatywny port http",
"Specify alternative http webroot", "Określ alternatywny http webroot",
"Specify bind address", "Określ powiązania adresu",
"Specify extra htsp port", "Podaj dodatkowy port htsp",
"Spiritual sciences", "Nauki duchowe",
"Sports", "Sporty",
"Sports magazines", "Magazyn sportowy",
"Start", "Początek",
"Start after", "Rozpocznij po",
"Start before", "Rozpocznij przed",
"Start time", "Czas rozpoczęcia",
"Status period (ms)", "Status okresu (ms)",
"Stop", "Zatrzymaj",
"Stop time", "Czas zakończenia",
"Storage path", "Ścieżka przechowywania",
"Storage space", "Przestrzeń przechowywania",
"Stream index", "Indeks strumienia",
"Stream profile", "Profil strumienia",
"Stream type", "Typ strumienia",
"Streaming", "Strumieniowanie",
"Streaming priority", "Priorytet strumieniowania",
"Streaming profiles", "Profile strumieniowania",
"Subdirectory options", "Opcje podkatalogów",
"Subscribe to a service permanently", "Zapisz się do usługi na stałe",
"Subscription overridden", "Subskrypcja nadpisana",
"Subscription weight", "Waga subskrypcji",
"Substitute formatters", "Zastępowanie formatek",
"Subtitle", "Podtytuł",
"Subtitle codec", "Kodek napisów",
"Sun", "Nd.",
"Surround (2 front, rear mono)", "Surround (2 front, tył mono)",
"Switch before rotor", "Przełącz przed rotorem",
"Switch off the power to the LNB when idle. Note: this may cause interference with other devices when the LNB is powered back up.", "Wyłącz zasilanie LNB w trybie bezczynności. Uwaga: może to powodować zakłócenia w pracy innych urządzeń w trakcie włączania LNB.",
"Switch type", "Typ przełącznika",
"Switch: %s", "Przełącznik: %s",
"Tag files with metadata", "Tagi pików z metadanymi",
"Tags", "Tagi",
"Team sports (excluding football)", "Drużyny sportowe (z wyłączeniem piłki nożnej)",
"Technology", "Technologia",
"Tennis", "Tenis",
"Testing", "Testowanie",
"Testing options", "Opcje testowania",
"The number of muxes left to scan on this network.", "Liczba muxów pozostałych do przeskanowania w tej sieci.",
"This tuner", "Ten tuner",
"Thu", "Czw.",
"Time", "Czas",
"Time missed", "Czas minął",
"Time record caption", "Czas nagrywania podpisu",
"Time recording%s%s", "Czas nagrywania%s%s",
"Time update", "Aktualizacja czasu",
"Timeout (sec) (0=infinite)", "Timeout (sek) (0=nieskończony)",
"Timeout (seconds)", "Timeout (sekundy)",
"Title", "Tytuł",
"Title (regexp)", "Tytuł (regexp)",
"Toggle details", "Pokaż szczegóły",
"Total number of mapped channels on this network.", "Całkowita liczba zmapowanych kanałów w tej sieci.",
"Total number of muxes found on this network.", "Całkowita liczba muxów znaleziona w tej sieci.",
"Total number of services found on this network.", "Całkowita liczba usług znaleziona w tej sieci.",
"Total number of services found.", "Całkowita liczba znalezionych usług.",
"Total number of services seen.", "Całkowita liczba widzianych usług.",
"Total number of services.", "Całkowita liczba usług.",
"Tourism / Travel", "Turystyka / Podróże",
"Trace subsystems", "Śledź podsystemy",
"Traditional arts", "Sztuka tradycyjna",
"Traditional music", "Muzyka tradycyjna",
"Transcoding", "Transkodowanie",
"Transmission mode", "Tryb transmisji",
"Transponder ID", "ID transpordera",
"Transport stream ID", "ID transportu strumienia",
"Tue", "Wt.",
"Tuner bind IP address", "Powiąż tuner z adresem IP",
"Tuner configuration", "Konfiguracja tunera",
"Tuning failed", "Nie udane strojenie",
"Turn off LNB when idle", "Wyłącz LNB kiedy bezczynny",
"Tvheadend server name", "Nazwa serwera Tvheadend",
"Type", "Typ",
"UDP RTP port number (2 ports)", "UDP RTP numer portu (2 porty)",
"URL with the SAT>IP server XML location", "URL z SAT> IP serwera lokalizacji XML",
"Unicable type", "Typ unicable",
"Unimportant", "Nieważny",
"Universal LNB only", "Tylko uniwersalny LNB",
"Unknown", "Nieznany",
"Unknown file", "Nieznany plik",
"Unknown reason (%i)", "Nieznany powód (%i)",
"Unlimited size", "Nieograniczony rozmiar",
"Unlimited time", "Nieograniczony czas",
"Unset (default)", "Nie ustawiono (domyślne)",
"Update card (EMM)", "Aktualizuj kartę (EMM)",
"Update channel icon", "Aktualizuj ikonę kanału",
"Update channel name", "Aktualizuj nazwę kanału",
"Update channel number", "Aktualizuj numer kanału",
"Update disabled", "Aktualizacja wyłączona",
"Update time", "Aktualizuj czas",
"Update tolerance (ms)", "Tolerancja aktualizacji (ms)",
"Updated", "Zaktualizowane",
"Usage: %s [OPTIONS]\n", "Użycie: %s [OPCJONALNIE]\n",
"Use EPG running state", "Użyj stanu uruchomienia EPG",
"Use HTTP digest authentication", "Użyj szyfrowanego protokołu uwierzytelniania HTTP",
"Use Windows-compatible filenames", "Użyj kompatybilności nazw plików Windows",
"Use XSPF playlist instead of M3U", "Użyj zamiast XSPF listy odtwarzania M3U",
"Use original", "Użyj oryginalnego",
"Use packet backlog", "Użyj zaległych pakietów",
"Use service IDs as channel numbers", "Użyj ID usług jako numery kanałów",
"User access error", "Błąd dostępu użytkownika",
"User icon", "Ikona użytkownika",
"User interface level", "Poziom interfejsu użytkownika",
"User interface quick tips (tooltips)", "Szybkie wskazówki interfejsu użytkownika (podpowiedzi)",
"User language", "Język użytkownika",
"User limit reached", "Osiągnięto limit użytkowników",
"User login", "Login uzytkownika",
"Username", "Nazwa użytkownika",
"Variety show", "Różnorodne pokazy",
"Video", "Wideo",
"Video codec", "Kodek video",
"Video recorder", "Nagrywarka wideo",
"Visually impaired commentary/audio description", "Niedowidzący komentarz / opis dźwięku",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV profil H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV profil H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV profil VP8/Vorbis/WEBM",
"Waiting for EPG running flag", "Oczekiwanie na flagę działania EPG",
"Waiting for program start", "Czekam na start programu",
"Waiting for stream", "Czekam na Strumień",
"War", "Wojenny",
"Water sport", "Wodny sport",
"Weak stream", "Słaby strumień",
"Weather report", "Pogoda",
"Web interface", "Web interfejs",
"Web interface language", "Język interfejsu Web",
"Web interface language.", "Język interfejsu Web.",
"Web interface theme.", "Temat interfejsu Web.",
"Web theme", "Temat Web",
"Web user interface", "Interfejs Web użytkownika",
"Wed", "Śr.",
"Wizard", "Kreator",
"Writes the current in-memory EPG database to disk every x hours, so should a crash/unexpected shutdown occur EPG data is saved periodically to the database (re-read on next startup). Set to 0 to disable.", "Zapisuje aktualną bazę danych EPG w pamięci na dysku co x godzin, więc w przypadku wystąpienia awarii/nieoczekiwanego zamknięcia dane EPG są okresowo zapisane w bazie danych (ponowne odczytanie następuje przy następnym uruchomieniu). Ustaw na 0, aby wyłączyć.",
"Yes", "Tak",
"default: nvenc(h264 / h265)", "domyślny: nvenc(h264 / h265)",
"fast: h264 / h265 / qsv(h264 / h265)", "szybki: h264 / h265 / qsv(h264 / h265)",
"faster: h264 / h265 / qsv(h264)", "szybszy: h264 / h265 / qsv(h264)",
"hrs", "godz",
"invalid option specified [%s]", "Podano nieprawidłową opcję [%s]",
"medium: h264 / h265 / qsv(h264 / h265)", "średni: h264 / h265 / qsv(h264 / h265)",
"mins", "minut",
"option %s requires a value", "opcja %s wymaga wartości",
"slow: h264 / h265 / qsv(h264 / h265)", "wolny: h264 / h265 / qsv(h264 / h265)",
"slower: h264 / h265 / qsv(h264)", "wolniejszy: h264 / h265 / qsv(h264)",
"superfast: h264 / h265", "super szybki: h264 / h265",
"tvhdhomerun client", "klient tvhdhomerun",
"ultrafast: h264 / h265", "ultra szybki: h264 / h265",
"veryfast: h264 / h265 / qsv(h264)", "bardzo szybki: h264 / h265 / qsv(h264)",
"veryslow: h264 / h265 / qsv(h264)", "bardzo wolny: h264 / h265 / qsv(h264)",
NULL, NULL
};

static const char *tvh_locale_26[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nPara mais informações visite o site do Tvheadend:\nhttps://tvheadend.org\n",
"# Mapped channels", "Canais mapeados",
"# Muxes", "Muxes",
"# Services", "# Serviços ",
"# tune repeats", "Repetição de sintonização",
"(Default profile)", "(Perfil padrão)",
"(Read-only)", "(Apenas leitura)",
"1 day", "1 dia",
"1 month", "1 mês",
"1 week", "1 semana",
"1 year", "1 ano",
"2 months", "2 meses",
"2 weeks", "2 semanas",
"2 years", "2 anos",
"2-Port switch (universal LNB)", "2-Porta variável (LNB universal)",
"3 days", "3 dias",
"3 months", "3 meses",
"3 weeks", "3 semanas",
"3 years", "3 anos",
"4-Port switch (universal LNB)", "4-Porta variável (LNB universal)",
"5 days", "5 dias",
"6 months", "6 meses",
"A channel tag (e.g. a group of channels) to which this rule applies.", "Etiqueta (ex. um grupo de canais) para o qual esta regra se aplica",
"AC-3 = descriptor 6", "AC-3 = descritor 6",
"AC-3 detection", "Detecção AC-3",
"ACTIVE", "ACTIVO",
"ANY", "QUALQUER UM",
"ATSC source ID", "ID fonte ATSC",
"ATSC-C Network", "Rede ATSC-C",
"ATSC-T Network", "Rede ATSC-T",
"Aborted by user", "Abortado pelo utilizador",
"Accept remote subscription weight", "Aceitar subscrição remota",
"Accept the remote subscription weight (from the SAT>IP client).", "Aceitar subscrição remota (do cliente SAT>IP)",
"Accept zero value for TSID", "Aceite zero para TSID",
"Access", "Acesso",
"Action", "Acção",
"Active", "Activo",
"Adapter in use by another subscription", "Adaptador ocupado por outra subscrição",
"Add file and line numbers to debug", "Adicionar ficheiro e número de linhas para depuração",
"Add the thread ID to debug", "Adicionar o ID do processo para depuração",
"Additional arguments to pass to the grabber.", "Argumentos adicionais do capturador",
"Additional service names found in EPG data.", "Nomes adicionais dos serviços encontrados nos dados EPG",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "Tempo adicional (em segundos) para o qual o sintonizador deve estar pronto. Isto é útil para sintonizadores que demoram algum tempo a sintonizar. ",
"Admin", "Administrador",
"Admin password", "Palavra-passe de administrador",
"Admin username", "Nome do administrador",
"Administrator login", "Login como administrador",
"Adult movie", "Filme Pornográfico",
"Advanced", "Avançado",
"Advanced (non-universal LNBs, rotors, etc.)", "Avançado (LNBs não-universais, motores, etc.)",
"Adventure", "Aventura",
"Advertisement / Shopping", "Publicidade",
"After creating the network scan it for services.", "Depois de criar a rede, pesquisar os serviços.",
"All (Streaming plus DVR)", "Todos (Transmissão mais DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Todos os caracteres que possam causar problemas no nome do ficheiro  serão substituídos por um underscore.  Ver ajuda para mais detalhes.",
"All lower-case", "Letra minúscula ",
"All proceeds are used to support server infrastructure and buy test equipment.", "Todos os lucros são utilizados para apoiar a infra-estrutura do servidor e comprar equipamento de teste",
"All: Record if different description", "Todos: Gravar com descrição diferente",
"All: Record if different episode number", "Todos: Gravar  se o número do episódio for diferente ",
"All: Record if different subtitle", "Todos: Gravar se tiver uma legenda diferente",
"All: Record once per day", "Todos: Gravar uma vez por dia",
"All: Record once per month", "Todos: Registo uma vez por mês",
"All: Record once per week", "Todos: Gravar uma vez por semana",
"Allow all PIDs", "Permite todos PIDs",
"Allow all PIDs (no filter) when the 'Maximum PIDs' limit is reached.", "Permite todos PIDs (sem filtro) quando o 'Máximo de PIDs' ter atingido o limite.",
"Allow high bitrate mode (CI+ CAMs only).", "Permitir modo de taxa de bit elevado (apenas CI+ CAM).",
"Allow idle scan tuning on this device.", "Permitir  análise em espera para este dispositivo.",
"Allow the combined size of all timeshift buffers to potentially grow unbounded until your storage media runs out of space.", "Permitir que o tamanho conjunto dos dados armazenados das faixas de video em tempo real  aumente, até não existir mais espaço de armazenamento no disco.  ",
"Allow the timeshift buffer to grow unbounded until your storage media runs out of space. Warning, enabling this option may cause your system to slow down or crash completely!", "Permitir o armazenamento de video em tempo real até atingir o limite de espaço disponível. Cuidado, ao activar esta opção o sistema pode abrandar ou mesmo ir abaixo!",
"Allow/Disallow access to the 'Configuration' tab.", "Permitir/Não Permitir acesso à aba de 'Configuração'.",
"Allow/Disallow web interface access (this  includes access to the EPG).", "Permitir/Não permitir acesso ao interface web (isto inclui o acesso ao EPG).",
"Allowed DVR profiles. This limits the profiles the user has access to.", "Perfis com permissão ao DVR. Limita as funcionalidades do utilizador.",
"Allowed network", "Rede permitida",
"Allowed networks", "Redes permitidas",
"Alternate PID path", "Caminho do pid alternativo",
"Alternate configuration path", "Caminho da configuração alternativa",
"Altitude (in meters).", "Altitude (metros).",
"Altitude (meters)", "Altitude (metros)",
"Always send the whole DiseqC sequence including LNB setup (voltage, tone). If this is not checked, only changed settings are sent, which may cause issues with some drivers. If the tuning is not reliable, try activating this option.", "Enviar sempre a sequência DiseqC, incluindo a informação do LNB (voltagem, tone). Se esta opção não estiver activa, apenas será enviada a informação das alterações realizadas, o que poderá causar problemas nalguns controladores. Se a sintonização não for a correcta, tente activar esta opção.",
"Always use the name defined by the network.", "Apenas usa a definição do nome pela rede.",
"Always use the old ioctls to read the linuxdvb status (signal strength, SNR, error counters). Some drivers are not mature enough to provide the correct values using the new v5 linuxdvb API.", "Utilizar sempre o modo ioctls antigo de modo a ler o estado do linuxdvb ( força do sinal, SNR, número de erros). Alguns controladores não suportam a visualização correcta desta informação, utilizando a nova API do linuxdvb v5. ",
"An event which starts between this \"start after\" and \"start before\" will be matched (including boundary values).", "Um evento será inicializado entre \"start after [começar depois]\" e \"start before [começar antes]\" (incluindo nos limites definidos). ",
"Animals", "Animais",
"Anonymize", "Anonimizar",
"Anonymize HTSP access", "Acesso HTSP anónimo",
"Any", "Qualquer um",
"Arts", "Artes",
"Arts / Culture (without music)", "Arte / Cultura (sem música)",
"Arts magazines", "Magazine de arte",
"Assign predefined muxes to networks", "Atribuir muxes predefinidos das redes.",
"Associate this device with one or more networks.", "Associar este dispositivo a uma ou mais redes.",
"Athletics", "Desporto",
"Audio Stream Filter", "Filtro de transmissão de áudio",
"Audio bitrate (kb/s) (0=auto)", "Taxa de áudio (kb/s) (0=auto)",
"Audio channel layout.", "Modo do canal de áudio",
"Audio codec", "Codec de áudio",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Codec de áudio a utilizar na conversão. \"Do not use (não utilizar)\" inactiva a saída de áudio.",
"Audio stream", "Stream audio",
"Audio-only stream", "Stream só audio",
"Authenticated user", "Utilizador autenticado",
"Auto check disabled", "Verificação automática desactivada",
"Auto check enabled", "Verificação automática activada",
"Auto record", "Gravar automaticamente",
"Auto record caption", "Gravar legendas automaticamente",
"Auto recording%s%s", "Gravação automática%s%s",
"Auto time record", "Gravar tempo automaticamente",
"Auto-Map to channels", "Mapeamento automático dos canais",
"Automatic checking", "Análise automática",
"Automatic recording caption.", "Gravação automática de legendas",
"Automatic schedule entry not found", "Item não encontrado",
"Automatic time scheduler entry not found", "Item não encontrado",
"Automatically map EPG source", "Mapear as fontes EPG automaticamente",
"Automatically map all available services to channels.", "Mapear todos os serviços disponíveis para canais.",
"Automatically map channels defined within the bouquet.", "Mapear todos os canais definidos no bouquet.",
"Automatically name from network", "Renomear com a informação obtida na rede",
"Automatically record.", "Gravar automaticamente",
"Autorec maximum count (0=unlimited)", "Gravação Auto. limite máx. (0=ilimitado)",
"Autorec maximum schedules limit (0=unlimited)", "Limite de gravações automáticas agendadas (0=ilimitado)",
"Bad", "Mau",
"Bad request", "Erro no pedido",
"Bandwidth", "Largura de Banda",
"Based on software from", "Baseado no software ",
"Basic", "Básico",
"Bind to specific local IP address.", "Vincular a um endereçamento IP local, específico.",
"Blue", "Azul",
"Bonjour", "Bom dia",
"Boot ID", "ID de inicialização",
"Bouquet source.", "Fonte do Bouquet.",
"Brand", "Marca",
"Branding information (if available).", "Informação da marca (se disponível).",
"Broadcast", "Transmissão",
"Broadcast type", "Tipo de transmissão",
"Broadcast.", "Transmissão.",
"Broadcasting", "Transmissão",
"Buffering limit (ms)", "Limites do buffer (ms)",
"Build", "Distribuição",
"By default, linuxdvb's status read period is 1000ms (one second). The accepted range is 250ms to 8000ms. Note that for some hardware or drivers (like USB), the status operations take too much time and CPU. In this case, increase the default value. For fast hardware, this value might be decreased to make the decision of the re-tune algorithm based on the signal status faster.", "Por omissão, o período de leitura do linuxdvb's é de 1000 ms (1 segundo). O intervalo aceite varia entre 250 ms até 8000 ms. Nota: o estado nalguns controladores e hardware (ex. USB), podem levar ao consumo intensivo de CPU e tempo. Se tal, aumente o valor definido por omissão. No caso do hardware ser rápido, este valor deve ser diminuído de modo a melhorar o algoritmo de sintonização.",
"CA Stream Filter", "Filtro de Transmissão CA",
"CA client %i", "Cliente CA %i",
"CA identification", "Identificação CA",
"CA provider", "Fornecedor CA",
"CAPMT (Linux Network DVBAPI)", "CAPMT (DVBAPI rede Linux)",
"CAPMT interval (in ms).", "Intervalo CAPMT (em ms).",
"CAPMT interval (ms)", "Intervalo CAPMT (ms)",
"CAPMT query interval (ms)", "Intervalo de pesquisa CAPMT  (ms)",
"CAPMT query interval (ms).", "Intervalo de pesquisa CAPMT  (ms)",
"CRID authority", "Autoridade CRID",
"Cache scheme", "Esquema de cache",
"Camd.socket filename / IP Address (TCP mode)", "Nome do Camd.socket / Endereço IP (modo TCP)",
"Cartoons", "Desenhos Animados",
"Channel", "Canal",
"Channel Tags", "Etiquetas dos Canais",
"Channel does not exist", "O canal não existe",
"Channel icon", "ícone do canal",
"Channel icon URL.", "URL do ícone do canal",
"Channel icon as defined in EPG data.", "Ícone do canal definido nos dados EPG ",
"Channel icon name scheme", "Nome do esquema do ícone do canal",
"Channel icon path", "Caminho do ícone do canal",
"Channel mapping options", "Opções de mapeamento dos canais",
"Channel name", "Nome do Canal",
"Channel not found", "Canal não encontrado",
"Channel number", "Nome do Canal",
"Channel number as defined in EPG data.", "Número do canal como definido nos dados EPG",
"Channel number offset", "Tolerância no número do canal ",
"Channel number range", "Intervalo de posições de canais",
"Channel numbers (heuristic)", "Numeração dos canais (heurístico)",
"Channel numbers from", "Canais de",
"Channel tag", "Etiqueta do canal",
"Channel tag does not exist", "Não existe etiqueta para o canal",
"Channel tag reference", "Referência da etiqueta do canal",
"Channel tags", "Etiquetas dos Canais",
"Channel tags the user is allowed access to/excluded from.", "Etiquetas dos canais permitidos/excluidos ao utilizador",
"Channel to use/used for the recording.", "Canal a utilizar/utilizado na gravação ",
"Channel update options", "Opções de actualização dos canais",
"Channels", "Canais",
"Channels EPG data is used by.", "Os dados EPG dos canais são utilizados por:",
"Character set", "Conjunto de caracteres",
"Characters not supported in Windows filenames (e.g. for an SMB/CIFS share) will be stripped out or converted.", "Os caracteres não suportados pelo Windows (ex. partilhas SMB/CIFS) serão eliminados ou convertidos.",
"Check availability", "Verificar disponibilidade ",
"Children's / Youth programs", "Programa para crianças",
"Class", "Classe",
"Classical", "Clássico",
"Classical music", "Música Clássica",
"Clean effects", "Limpar efeitos",
"Client name", "Nome do cliente",
"Clone scheduled entry on error", "Replicar o item agendado em caso de erro",
"Comedy", "Comédia",
"Command delay time (ms) (10-200)", "Tempo de atraso do comando (ms) (10-200)",
"Command time (ms) (10-100)", "Tempo do comando (ms) (10-100)",
"Comment", "Comentário",
"Commercial break", "Anúncios publicitários",
"Committed", "Confirmado",
"Completed OK", "Concluído OK",
"Compress EPG database", "Comprimir a base de dados EPG",
"Compress the EPG database to reduce disk I/O and space.", "Comprimir a base de dados EPG para reduzir espaço em disco.",
"Concurrent limit per network position (src=) for satellite SAT>IP tuners. The first limit number is for src=1 (AA), second for src=2 (AB) etc.", "O limite na concorrência por posição da rede satélite SAT>IP (src=). O primeiro limite é para a src=1 (AA) a segunda para a src=2 (AB), etc...  ",
"Conditional Access", "Acesso condicional",
"Conditional Access Client", "Cliente de acesso condicional",
"Conditional Access Identification.", "Identificação do acesso condicional",
"Configuration", "Configuração",
"Configuration - Base", "Configuração - Base",
"Configuration - Image Cache", "Configuração - Armazenamento de imagens",
"Configuration - SAT>IP Server", "Configuração - Servidor SAT>IP",
"Configuration ID", "ID da Configuração",
"Configuration name", "Nome da configuração",
"Configuration version", "Versão da configuração",
"Connection limit type", "Tempo limite da conexão ",
"Constellation", "Constelação",
"Container", "Recipiente",
"Container to use for the transcoded stream.", "Formato a utilizar na conversão.",
"Content character set", "Conjunto de caracteres do conteúdo",
"Content reference identifier authority.", "Identificação da autoridade de referência de conteúdo",
"Content type", "Tipo de conteúdo",
"Content type.", "Tipo de Conteúdo",
"Contest", "Concurso",
"Continue recording for x minutes after scheduled stop time", "Continuar a gravar x minutos após o tempo agendado na gravação",
"Continue recording for x minutes after scheduled stop time.", "Continuar a gravar x minutos após o tempo agendado na gravação",
"Cookie expiration (days)", "Os Cookies expiram em (dias)",
"Cooking", "Cozinha",
"Copy codec type", "Codec Igual ao Original ",
"Copy layout", "Copiar layout",
"Count of objects", "Contagem de objectos",
"Create SDTV/HDTV/Radio tags.", "Criar etiquetas SDTV/HDTV/Radio.",
"Create a directory per channel when storing recordings. If both this and the 'directory per day' checkbox is enabled, the date-directory will be the parent of the per-channel directory.", "Criar um directório por canal ao gravar. Se esta opção e o  'directory per day [directório por dia] ' estiverem ambos activos, o segundo vai ser implementado em primeiro lugar ficando as pasta referentes à gravação de cada canal no interior da pasta 'directório por dia' .",
"Create a directory per title when storing recordings. If the day/channel directory checkboxes are also enabled, those directories will be parents of this directory.", "Criar um directório por título ao gravar. Se a opção 'gravar directório por dia/canal' estiver activa esses directórios serão criados primeiro e só depois serão criados os directórios por título dentro destes.    ",
"Create a new directory per day in the recording system path. Folders will only be created when something is recorded. The format of the directory will be ISO standard YYYY-MM-DD.", "Criar um directório por dia nas pastas de gravação. As pastas serão criadas apenas quando algum programa for gravado. Será utilizado o formato ISO YYY-MM-DD para nomear o directório. ",
"Create a provider name tag.", "Criar uma etiqueta para o provedor.",
"Create and associate a network tag to created channels.", "Criar e associar uma etiqueta da rede para os canais",
"Create and associate a provider tag to created channels.", "Criar e associar uma etiqueta do provedor para os canais",
"Create bouquet", "Criar bouquet",
"Create bouquet tag", "Criar uma etiqueta para os canais",
"Create directories using these permissions.", "Criar directórios utilizando estas permissões.",
"Create files using these permissions.", "Criar ficheiros utilizando estas permissões.",
"Create network name tags", "Criar etiquetas com o nome das redes",
"Create network name tags (set by provider).", "Criar etiquetas com o nome das redes (pelo provedor)",
"Create network tags", "Criar etiquetas para as redes",
"Create provider name tags", "Criar etiquetas dos nomes dos provedores",
"Create provider tags", "Criar etiquetas dos provedores",
"Create tags", "Criar etiquetas",
"Create tags in recordings using media containers that support metadata (if possible).", "Criar etiquetas nos ficheiros de gravações, quando são utilizados formatos que suportam metadados (se possível).",
"Created", "Criado",
"Creator", "Criar",
"Cron multi-line", "Linhas do Cron",
"Culture (without music)", "Cultura (sem música)",
"Current RTSP port.", "Porta RTSP actual",
"Current affairs", "Assuntos actuais",
"Current number of objects.", "Número actual de objectos.",
"Current object size.", "Tamanho actual do objecto.",
"Current tuner configuration.", "Configuração actual do sintonizador",
"Custom HTTP headers", "Cabeçalhos HTTP personalizados ",
"DAB Network", "Rede DAB",
"DBus - use the session message bus instead of the system one", "DBus -Utilizar mensagem da sessão do bus em alternativa ao bus do sistema",
"DES Key.", "Chave DES",
"DES key", "Chave DES",
"DSCP/TOS for streaming", "DSCP/TOS na transmissão",
"DVB Inputs - Networks", "Redes - Entradas DVB",
"DVB scan files", "Análise de ficheiros DVB",
"DVB scan files path", "Verificação dos caminhos de ficheiros DVB ",
"DVB-C Network", "Rede DVB-C",
"DVB-S Network", "Rede DVB-S",
"DVB-S Satellite Configuration", "Configuração satélite DVB-S ",
"DVB-T Network", "Rede DVB-T",
"DVR - Auto-recording (Autorecs)", "DVR - Gravação Automática (Autorecs)",
"DVR - Profiles", "DVR - Perfis",
"DVR - Time-based Recording (Timers)", "DVR - Gravação baseada no agendamento (Temporizadores)",
"DVR behavior", "Comportamento DVR",
"DVR configuration", "Configuração DVR",
"DVR configuration profiles", "Configuração dos perfis de DVR",
"DVR entry not found", "Item do DVR não encontrado",
"DVR file retention period", "Período de retenção de um item no DVR",
"DVR log retention", "Retenção do log do DVR",
"DVR log retention period", "Período de retenção do log do DVR",
"DVR override: high", "Sobrepor DVR: elevado",
"DVR override: important", "Sobrepor DVR: importante",
"DVR override: low", "Sobrepor DVR: baixo",
"DVR override: normal", "Sobrepor DVR: normal",
"DVR override: unimportant", "Sobrepor DVR: sem importância",
"DVR profile to use/used for the recording.", "Perfil do DVR a utilizar/utilizado na gravação.",
"DVR schedule does not exist", "Não existe o item agendado no DVR ",
"DVR schedule does not have a file yet", "A agenda do DVR ainda não contém itens. ",
"DVR schedule not found", "Item agendado não encontrado no DVR",
"Dance", "Dança",
"Data errors", "Erro nos dados",
"Data path (if applicable).", "Caminho para os dados (se aplicável).",
"Date the EPG data was last updated (not set for OTA grabbers).", "Mostrar a data da última actualização efectuada ao EPG (não pode ser utilizado em capturadores OTA). ",
"Days of Week", "Dias da Semana",
"Days of the week to which the rule should apply.", "Dias da semana em que esta regra deve ser aplicada.",
"Days of week", "Dias da semana",
"Debate", "Debates",
"Debug libav log", "Log de depuração do libav",
"Debug options", "Opções de depuração",
"Debug subsystems", "Depuração dos subsistemas",
"Debug to syslog", "Depuração para syslog",
"Debug trace (low-level)", "Rastreamento da depuração (baixo nível)",
"Debugging", "Depuração",
"Default", "Por omissão",
"Default language(s)", "Linguagem(ns) padrão",
"Default language.", "Linguagem por omissão",
"Default priority", "Prioridade por omissão",
"Default user interface level.", "Nível da interface do utilizador por omissão",
"Define network group to limit network usage.", "Definir um grupo da rede para limitar a utilização da rede.",
"Delivery system", "Sistema de entrega",
"Demux path", "Caminho de Demux",
"Descramble services (limit per mux)", "Serviços a descodificar (limitar por mux)",
"Descrambler buffer (TS packets)", "Descodificar buffer (pacotes TS)",
"Description", "Descrição",
"Detective", "Policial ",
"Device ID", "ID do Dispositivo",
"Device model", "Modelo do dispositivo ",
"Device path", "Caminho do dispositivo ",
"Digest access authentication is intended as a security trade-off. It is intended to replace unencrypted HTTP basic access authentication. This option should be enabled for standard usage.", "Autenticação Digest permite um compromisso na segurança. O seu objectivo passa por substituir a autenticação HTTP básica não encriptada. Esta opção deve estar activa para numa utilização normal.  ",
"Directory", "Directório",
"Directory permissions (octal, e.g. 0775)", "Permissões do directório (octal, ex. 0775)",
"Directory used by the entry.", "Directório utilizado na configuração",
"Disable", "Desactivar",
"Disable DVB bouquets", "Desactivar bouquets DVB",
"Disable SAT>IP client", "Desactivar cliente SAT>IP",
"Disable all access control checks", "Desactivar todos os controlos de acesso",
"Disable debug on stderr", "Desactivar a depuração no stderr",
"Disable device/firmware-specific workarounds", "Desactivar dispositivo/soluções alternativas específicas do firmware",
"Disable syslog (all messages)", "Desactivar syslog (todas msgs)",
"Disabled", "Desactivado",
"Discover more muxes using the Network Information Table (if available).", "Procurar outros muxes utilizando a informação obtida na rede (se disponível)",
"Discussion", "Discussão",
"DiseqC repeats", "Repetição DiseqC",
"Do not send any stream specific information to the HTSP client like signal strength, input source etc.", "Não enviar nenhuma informação especifica acerca da transmissão para o cliente HTSP, como a força do sinal, a fonte do canal, etc...",
"Do not use", "Não Utilizar",
"Documentary", "Documentário",
"Don't abort streaming when an encrypted stream can't be decrypted by a CA client that normally should be able to decrypt the stream.", "Não abortar a transmissão quando um canal encriptado não puder ser desencriptado pelo cliente CA.  ",
"Don't backup configuration tree at upgrade", "Não realizar o backup da árvore de configuração na actualização",
"Don't include the title in the filename.", "Não incluir o título no nome do ficheiro.",
"Don't include title in filename", "Não incluir título no ficheiro",
"Don't keep", "Não manter",
"Don't re-record", "Não regravar",
"Don't re-record if recording fails.", "Não regravar caso houver falha na gravação ",
"Don't re-schedule if recording fails.", "Não reagendar caso houver falha na gravação ",
"Don't reschedule", "Não reagendar",
"Don't use the provider's channel numbers.", "Não utilizar a numeração dos canais dos provedores.",
"Duplicate handling", "Duplicar item",
"Duplicate recording handling.", "Duplicar gravação",
"EIT - skip TSID check", "EIT - saltar verificação TSID",
"EIT time offset", "Tolerância do tempo EIT",
"EPG Grabber", "Capturador EPG",
"EPG Grabber Channel", "Capturador EPG do canal",
"EPG Grabber Configuration", "Configuração da captura EPG",
"EPG Language (priority order)", "Língua utilizada no EPG (ordem prioritária)",
"EPG data ID.", "ID dos dados EPG",
"EPG name", "Nome EPG",
"EPG scan", "Análise EPG",
"EPG source", "Fonte EPG",
"EPG update window", "Actualização do EPG ",
"Economics", "Economia",
"Education", "Educação",
"Educational", "Educacional ",
"Elementary stream filter", "Filtro de Transmissão",
"Enable (auto)", "Activar (auto)",
"Enable DBus", "Activar DBus",
"Enable NTP driver", "Activar o driver NTP",
"Enable coredumps for daemon", "Activar coredumps do serviço",
"Enable debug on stderr", "Activar depuração no stderr",
"Enable debug subsystems", "Activar depuração dos subsistemas",
"Enable debug to file", "Activar depuração para ficheiro",
"Enable debug to syslog", "Activar depuração para syslog",
"Enable exclusion of user-config defined channel tags. This will prevent the user from accessing channels associated with the tags selected (below).", "Activar a exclusão de etiquetas do canal definidas na configuração do utilizador. Previne que o utilizador consiga aceder a canais com as seguintes etiquetas (abaixo).",
"Enable if the SAT>IP box requires pids=21 parameter in the SETUP RTSP command", "Activar se o SAT>IP requerer o parâmetro pids=21 no comando RTSP",
"Enable if the SAT>IP box supports the addpids/delpids commands.", "Activar de o SAT>IP suportar comandos addpids/delpids ",
"Enable or disable fast input switching.", "Activar ou desactivar mudança rápida de entrada",
"Enable or disable full mux mode.", "Activar ou desactivar o modo full mux",
"Enable or disable this configuration.", "Activar ou desactivar esta configuração",
"Enable over-the-air program guide (EPG) scanning on this input device.", "Activar a captura da programação televisiva (EPG) over-the-air neste dispositivo.",
"Enable syslog", "Activar syslog",
"Enable this filter.", "Activar este filtro",
"Enable trace subsystems", "Activar rastreamento dos subsistemas",
"Enable web UI debug (non-minified JS)", "Activar visualização da depuração através da web UI (non-minified JS)",
"Enable/Disable CA client.", "Activar/desactivar cliente CA.",
"Enable/disable EPG data for the entry.", "Activar/desactivar dados EPG para este item.",
"Enable/disable LNA.", "Activar/desactivar LNA.",
"Enable/disable auto-rec rule.", "Activar/desactivar regra de gravação automática.",
"Enable/disable debugging output to syslog.", "Activar/desactivar escrita da depuração para o syslog",
"Enable/disable inclusion of low-level debug traces.", "Activar/desactivar a inclusão de rastos de depuração de baixo-nível",
"Enable/disable interface quick tips.", "Activar/desactivar ajudas rápidas ao utilizador.",
"Enable/disable libav log output.", "Activar/desactivar escrita da depuração do libav.",
"Enable/disable logging to syslog.", "Activar/desactivar depuração para o syslog.",
"Enable/disable offering of Entitlement Management Message updates.", "Activar/desactivar mensagens de actualização.",
"Enable/disable pilot tone.", "Activar/desactivar pilot tone.",
"Enable/disable power save mode (if supported by the device).", "Activar/desactivar poupança de energia (se o dispositivo suportar). ",
"Enable/disable profile.", "Activar/desactivar perfil.",
"Enable/disable the bouquet.", "Activar/desactivar bouquet.",
"Enable/disable the channel.", "Activar/desactivar canal.",
"Enable/disable the device.", "Activar/desactivar dispositivo.",
"Enable/disable the entry.", "Activar/desactivar item.",
"Enable/disable the grabber.", "Activar/desactivar capturador.",
"Enable/disable the tag.", "Activar/desactivar etiqueta.",
"Enable/disable tuner/adapter.", "Activar/desactivar sintonizador/adaptador.",
"Enabled", "Activo",
"Encrypted", "Encriptado",
"Encrypted service", "Serviço encriptado",
"Enter a non-admin user password.", "Preencher uma palavra-passe para utilizador sem permissões de administrador.",
"Enter a non-admin user username.", "Preencher com o nome de utilizador sem permissões de administrador.",
"Enter an administrator password.", "Defina uma palavra-passe de administrador",
"Enter an administrator username. Note: do not use the same username as the superuser backdoor account.", "Definir um nome de administrador. Nota: não utilizar um nome igual ao que foi definido na conta de super-utilizador. ",
"Enter comma-separated list of subsystems you want debugging output for (e.g +linuxdvb,+subscriptions,+mpegts).", "Definir uma lista dos sub-sistemas que pretende depurar, separados por virgula (ex. +linuxdvb,+subscriptions,+mpegts).",
"Enter comma-separated list of subsystems you want to get traces for (e.g +linuxdvb,+subscriptions,+mpegts).", "Definir uma lista dos sub-sistemas que pretende depurar, separados por virgula (ex. +linuxdvb,+subscriptions,+mpegts).",
"Entertainment programs for 10 to 16", "Programas de entretenimento dos 10 aos 16",
"Entertainment programs for 6 to 14", "Programas de Entretenimento dos 6 aos 14",
"Environment", "Ambiente",
"Environment (pipe)", "Ambiente (pipe)",
"Episode", "Episódio",
"Episode %d", "Episódio %d",
"Episode number/ID.", "Número/ID do episódio.",
"Error code", "Código do Erro",
"Error code of entry.", "Código de erro do item",
"Errors", "Erros",
"Even key", "Chave do Evento",
"Even key.", "Chave do Evento",
"Event does not exist", "O evento não existe",
"Every day", "Todos os dias",
"Exclude channel tags", "Excluir etiquetas dos canais",
"Expeditions", "Expedições",
"Experimental film", "Filme alternativo",
"Expert", "Experiente ",
"Exported tuners", "Sintonizadores exportados",
"External", "Externo",
"External EPG grabber", "Capturador EPG externo",
"External IP (NAT)", "IP externo (NAT)",
"External PyEPG Grabber", "Capturador PyEPG externo",
"External URL", "URL externa",
"External URL of the bouquet.", "URL externo do bouquet.",
"External XMLTV EPG Grabber", "Capturador XMLTV externo do EPG",
"Extra arguments", "Argumentos extra",
"Extra start time", "Tempo extra no ínicio",
"Extra stop time", "Tempo extra no final ",
"Extra warming up time (seconds)", "Tempo extra de inicialização (segundos)",
"FAIL", "FALHOU",
"FEC high", "FEC Elevado",
"FEC low", "FEC Baixo",
"Factual topics", "Temas factuais",
"Failed", "Falhou/falhado",
"Failed to open image", "Falhou ao abrir a imagem",
"Fantasy", "Fantasia",
"Fast input switch", "Alteração rápida de entrada",
"File missing", "Falta o ficheiro",
"File not created", "Ficheiro não foi criado",
"File permissions (octal, e.g. 0664)", "Permissões do Ficheiro (octal, e.g. 0664)",
"File size", "Tamanho do Ficheiro",
"Filename", "Nome do Ficheiro",
"Filename character set", "Caracteres do Ficheiro",
"Filename options", "Opções do Ficheiro",
"Filename used by the entry.", "Nome do ficheiro utilizado na configuração",
"Film", "Filme",
"Fine arts", "Belas-Artes",
"Finished", "Terminado",
"First word", "Primeira palavra",
"Fit to RAM (cut rewind)", "Adequar à RAM (apagar antigo)",
"Fitness and health", "Saúde e Fitness",
"Folkloric", "Folclore",
"Football", "Futebol",
"Force (auto)", "Forçar (auto)",
"Force CA ID (e.g. 0x2600)", "Forçar CA ID (e.g. 0x2600)",
"Force all network connections to this tuner to be made over the specified IP address, similar to the setting for the SAT>IP device itself. Setting this overrides the device-specific setting.", "Forçar a que todas as conexões a este sintonizador sejam realizadas através de um IP   específico. Ao utilizar esta opção irá ignorar as definições especificas do dispositivo.   ",
"Force an initial EPG grab at start-up.", "Forçar a procura EPG ao iniciar.",
"Force initial EPG grab at start-up", "Forçar a procura EPG ao iniciar.",
"Force old status", "Forçar  o estado anterior",
"Force pilot for DVB-S2", "Forçar DVB-S2",
"Force priority", "Forçar Prioridade",
"Force service type to 1", "Forçar o tipo do serviço para 1",
"Force teardown delay", "Forçar atraso no teardown",
"Force usage of entered CA ID on this service.", "Forçar a utilização deste CA ID neste serviço.",
"Foreign countries", "Estrangeiro",
"Forever", "Para sempre",
"Fork and run as daemon", "Bifurcar e executar como serviço",
"Format string", "Formato da String",
"Found muxes", "Muxes encontrados",
"Found services", "Muxes encontrados",
"Free subscription weight", "Peso da subscrição ",
"Free-form field, enter whatever you like here.", "Coloque o que pretender neste campo.",
"Free-form text field, enter whatever you like here.", "Coloque o que pretender neste campo.",
"Free-form text field, enter whatever you like.", "Coloque o que pretender neste campo.",
"Free-form text field. You can enter whatever you like here.", "Coloque o que pretender neste campo.",
"Free-format text field. Enter whatever you like here.", "Coloque o que pretender neste campo.",
"Frequency", "Frequência",
"Frequency (Hz)", "Frequência (Hz)",
"Frequency (MHz)", "Frequência (MHz)",
"Frequency (in MHz).", "Frequência (MHz)",
"Frequency (kHz)", "Frequência (kHz)",
"Fri", "Sexta-Feira",
"Friendly name", "Apelido",
"Frontend number", "Número da Interface",
"Frontend path", "Caminho para a Interface",
"Full DiseqC", "DiseqC Total",
"Full mux RX mode supported", "Suporte completo com Mux Rx",
"Full pathname specification", "Especificação do caminho",
"Full-text", "Texto Completo",
"Further education", "Educação avançada",
"GOTOX position", "Posição GOTOX",
"Gardening", "Jardinagem",
"General", "Geral",
"General configuration", "Configuração Geral",
"Generic", "Genérico",
"Generic options", "Opções Genéricas",
"Gray", "Cinzento",
"Guard interval", "Intervalo de Guarda",
"HD: high definition", "HD: Alta-definição",
"HTSP Default Stream Settings", "Configurações de transmissão HTSP",
"HTSP Stream Profile", "Perfil de Transmissão HTSP",
"HTTP CORS origin", "Origem HTTP CORS ",
"Handicraft", "Artesanato",
"Hearing impaired", "Deficiência auditiva",
"Hierarchy", "Hierarquia ",
"High", "Elevado",
"High bitrate mode (CI+ CAMs only)", "Modo de alta taxa de bits (CI+ CAMs Only)",
"Highest channel number the user can access.", "Número máximo do canal que o utilizador pode ter acesso. ",
"Highest count of objects.", "Contagem mais elevada de objectos",
"Historical movie", "História",
"Horror", "Terror",
"Hostname (or IP) of the server.", "Hostname (ou IP) do servidor.",
"Hostname/IP", "Nome do Servidor / IP",
"How frequently the upstream provider is checked for changes.", "A frequência para a análise de mudanças do provedor ",
"ID of the network.", "ID da rede.",
"IP address", "Endereço IP",
"IP blocking", "Bloquear IP",
"IPTV Automatic Network", "Rede IPTV automática ",
"IPTV Network", "Rede IPTV",
"IPTV input", "Entrada IPTV",
"ISDB-C Network", "Rede ISDB-C",
"ISDB-S Network", "Rede ISDB-S",
"ISDB-T Network", "Rede ISDB-T",
"ISI (Stream ID)", "ISI (ID Fluxo)",
"Icon", "Ícone",
"Icon (full URL)", "ícone (URL completo)",
"Icon URL", "URL do ícone",
"Icon base URL", "URL base do ícone",
"Icon has title", "Título do ícone",
"Icons from", "Ícones de",
"Idle", "Inactivo",
"Idle scan", "Verificação em espera",
"Idle scan muxes", "Verificação dos Muxes em espera",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Se não existir uma conta de utilizador, criar então\numa sem username e password. Utilizar com cuidado\npois pode permitir acessos indevidos de administrador \nà sua conta do Tvheadend, até editar ou criar um controlo de acesso\natravés da interface Web do TVheadend",
"If you'd like to support the project, please consider a donation.", "Se quiser ajudar a suportar o nosso projecto, dê-nos o seu donativo, obrigado. ",
"Ignore EPG (EIT)", "Ignorar EPG (EIT)",
"Ignore descriptor 5", "Ignorar o descritor 5 ",
"Ignore invalid SSL certificate", "Ignorar certificados SSL inválidos",
"Ignore provider's channel numbers", "Ignorar o número do canal dos provedores",
"Immediately abort", "Abortar imediatamente",
"Important", "Importante",
"Include channel name in filename", "Incluir o nome do canal no ficheiro",
"Include date in filename", "Incluir data no ficheiro",
"Include episode in filename", "Incluir o nome do episódio no ficheiro",
"Include subtitle in filename", "Incluir legenda no ficheiro",
"Include time in filename", "Incluir a duração no ficheiro ",
"Index", "Índice",
"Information area", "Área de informação",
"Informational", "Informação",
"Initial scan", "Procura inicial",
"Input base", "Entrada base",
"Input buffer (bytes)", "Buffer de entrada (Bytes)",
"Input instance", "Instância de entrada",
"Input path", "Caminho",
"Internal", "Interno",
"Internal EPG grabber", "Capturador EPG interno",
"Internal PyEPG Grabber", "Capturador PyEPG interno",
"Internal XMLTV EPG Grabber", "Capturador XMLTV interno do EPG",
"Internal error", "Erro interno",
"Internal grabber", "Capturador interno",
"Interview", "Entrevista",
"Invalid", "Inválido",
"Invalid EPG object request", "Pedido inválido do objecto EPG",
"Invalid adapter number '%s'\n", "Número do adaptador inválido '%s'\n",
"Invalid arguments", "Argumentos inválidos",
"Invalid file", "Ficheiro inválido",
"Invalid parameters", "Parâmetros inválidos",
"Invalid service", "Serviço inválido",
"Invalid target", "Destino inválido",
"Items", "Itens",
"Keep", "Manter",
"Keep x amount of storage space free.", "Manter x quantidade de espaço livre em disco",
"Kill signal (pipe)", "Terminar processo (pipe)",
"Kill timeout (pipe/secs)", "Tempo limite para terminar (pipe/s) ",
"LNB type", "Tipo de LNB",
"Language", "Idioma",
"Language 1", "Idioma 1",
"Language 2", "Idioma 2",
"Language 3", "Idioma 3",
"Language settings", "Configurações do idioma",
"Languages", "Idiomas",
"Last seen", "Visto pela última vez",
"Last updated from", "Última actualização",
"Layer A: Constellation", "Layer A: Constelação",
"Layer A: Segment count", "Layer A: Contagem de segmentos",
"Layer A: Time interleaving", "Layer A: Intervalo de tempo",
"Layer B: Constellation", "Layer B: Constelação",
"Layer B: Segment count", "Layer B: Contagem de segmentos",
"Layer B: Time interleaving", "Layer B: Intervalo de tempo",
"Layer C: Constellation", "Layer C: Constelação",
"Layer C: Segment count", "Layer C: Contagem de segmentos",
"Layer C: Time interleaving", "Layer C: Intervalo de tempo",
"Leisure hobbies", "Lazer",
"Limit connections", "Limitar conexões ",
"Limited/limit scanning to this network ID only.", "Limitar a análise apenas a esta rede.",
"Linked input", "Entrada Ligada",
"LinuxDVB adapter", "Adaptador LinuxDVB",
"LinuxDVB network", "Rede LinuxDVB",
"List of allowed IPv4 or IPv6 hosts or networks (comma-separated).", "Lista de endereços IP permitidos (IPv4 e IPv6, separados por virgula). ",
"Listen / Connect port", "Porta Escuta/Conexão ",
"Listen on IPv6", "Escutar ligação IPv6",
"Literature", "Literatura",
"Local (server) time", "Tempo local (servidor)",
"Local bind IP address", "Endereço IP de ligação local",
"Local channel minor", "Canal local ",
"Local channel number", "Número do canal",
"Local discovery IP address", "Procura do endereço de IP local",
"Local: Record if different description", "Local: Gravar se tiver uma descrição diferente ",
"Local: Record if different episode number", "Gravar se o número do episódio for diferente",
"Local: Record if different subtitle", "Gravar se a legenda for diferente",
"Local: Record if different title", "Gravar se o titulo for diferente",
"Local: Record once per day", "Grava apenas uma vez por dia",
"Local: Record once per week", "Gravar apenas uma vez por semana",
"Location", "Localização",
"Lock preferred CA PID", "Bloquear CA PID preferido",
"Login password.", "Palavra passe de Login",
"Login username.", "Nome de Utilizador",
"Low", "Baixo",
"Lowest channel number the user can access.", "Número mínimo do canal que o utilizador pode ter acesso. ",
"MPEG-TS Pass-thru", "MPEG-TS Pass-through",
"MPEG-TS Pass-thru/built-in", "MPEG-TS Pass-through/built-in",
"MPEG-TS Service", "Formato MPEG-TS",
"MPEG-TS input", "Entrada MPEG-TS",
"MPEG-TS multiplex PHY", "MPEG-TS Multiplex PHY",
"MPEG-TS raw service", "Formato puramente MPEG-TS",
"Magazines", "Revistas",
"Maintain free storage space in MiB", "Manter espaço livre em disco em MB",
"Maintain used storage space in MiB (0=disabled)", "Manter espaço livre em disco em MB (0=desactivado)",
"Maintained space", "Espaço de Armazenamento",
"Make subdirectories per channel", "Criar Subdirectórios por Canal",
"Make subdirectories per day", "Criar Subdirectórios por Dia",
"Make subdirectories per title", "Criar Subdirectórios por Título",
"Manufacturer", "Fabricante",
"Manufacturer URL", "URL do Fabricante",
"Manufacturer's model number.", "Número do modelo do fabricante.",
"Manufacturer's product description.", "Descrição do fabricante.",
"Manufacturer's product information page for the device.", "Descrição do fabricante para o dispositivo. ",
"Manufacturer's product name.", "Nome do fabricante do produto.",
"Map all services", "Mapear todos os serviços",
"Map encrypted services", "Mapear os serviços encriptados",
"Map radio channels", "Mapear  canais de rádio",
"Map unnamed channels", "Mapear  canais sem nome",
"Map zero-numbered channels", "Mapear os canais numerados a zero. ",
"Master tuner", "Sintonizador Principal ",
"Matroska specific", "Matroska específico ",
"Maximal channel number", "Número máximo de canais",
"Maximum # input streams", "Máximo # Faixas de entrada",
"Maximum PIDs", "PIDs Máximo",
"Maximum RAM size (MB)", "Tamanho Máx. de RAM (MB)",
"Maximum bandwidth (Kbps)", "Máxima Largura de Banda (Kbps)",
"Maximum count (0=default)", "Contagem máxima (0=valor por omissão)",
"Maximum duration", "Duração Máxima",
"Maximum length of PIDs", "Tamanho Máx. dos PIDs",
"Maximum period (mins)", "Período máximo (mins)",
"Maximum schedules limit (0=default)", "Limite de gravações automáticas agendadas (0=valor por omissão)",
"Maximum size (MB)", "Tamanho Máximo (MB)",
"Maximum timeout (seconds)", "Tempo limite Máx. (segundos)",
"Medicine", "Medicina",
"Memory Information", "Informação da memória",
"Merge same name", "Juntar o mesmo nome",
"Method not found", "Método não encontrado",
"Minimal channel number", "Número mínimo do canal",
"Minimum duration", "Duração Mínima",
"Minimum rotor time (seconds)", "Tempo Min. do Motor (segundos)",
"Missing In PAT/SDT", "Falta o PAT/SDT",
"Mode", "Modo",
"Model description", "Descrição do Modelo",
"Model name", "Nome do Modelo",
"Model number", "Número do Modelo",
"Modulation", "Modulação",
"Module", "Modulo",
"Module ID", "ID do módulo",
"Mon", "Segunda-Feira",
"More verbose libav log", "Log libav mais detalhado",
"Motor rate (in milliseconds/deg).", "Velocidade do Motor (ms/graus)",
"Motor rate (milliseconds/deg)", "Velocidade do Motor (ms/graus)",
"Motor sport", "Desporto Motorizado",
"Motoring", "Automobilismo",
"Movie", "Filme",
"Music", "Música",
"Musical", "Musical ",
"Mux", "Mistura",
"Mux Scheduler", "Agendador do Mux",
"Mux UUID", "UUID da Mistura",
"Mux handling", "Manuseamento do Mux",
"Mux name", "Nome do Mux",
"Mux not enabled", "Mistura inactiva",
"NONE", "NENHUM",
"Name", "Nome",
"Natural sciences", "Ciências Naturais",
"Nature", "Natureza",
"Network", "Rede",
"Network ", "Rede",
"Network A", "Rede A",
"Network B", "Rede B",
"Network ID (limit scanning)", "ID da Rede (Limitar procura)",
"Network UUID", "UUID da Rede",
"Network discovery", "Procurar  Rede",
"Network name", "Nome da Rede",
"Network prefix", "Prefixo da Rede",
"Network type", "Tipo de Rede",
"Networks", "Redes",
"News", "Notícias ",
"News magazine", "Revistas de Notícias",
"Next tune delay in ms (0-2000)", "Atraso na próxima sintonização em ms (0-2000)",
"No", "Não",
"No PIN", "Sem PIN",
"No access", "Sem acesso",
"No adapters specified!\n", "Sem adaptadores especificados\n",
"No days", "Sem Dias",
"No descrambler", "Sem descodificador",
"No free adapter", "Nenhum adaptador livre",
"No input detected", "Não foi detectado nenhuma entrada",
"No service assigned to channel", "Não existem serviços atribuídos ao canal",
"No service enabled", "Serviço inactivo",
"No source available", "Nenhuma fonte disponível",
"None", "Nenhum",
"Not linked", "Sem conexões",
"Not set", "Não definido",
"Not set (none or channel configuration)", "Não definido (nenhum ou configuração do canal )",
"Not set (use channel or DVR configuration)", "Não definido (utilizar configuração do canal ou DVR)",
"Number", "Número",
"Number of tsfile tuners", "Número de Sintonizadores tsfile",
"OK (partial)", "OK (parcial)",
"OSCam net protocol (rev >= 10389)", "protocolo de rede OSCam (rev >= 10389)",
"Off", "Desligado",
"Older OSCam", "Versão antiga OSCam ",
"On", "Ligado",
"Only Bulsatcom 39E", "Apenas Bulsatcom 39E",
"Only EIT", "Apenas EIT",
"Only OpenTV Sky Ausat", "Apenas OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Apenas OpenTV Sky Italia",
"Only OpenTV Sky UK", "Apenas OpenTV Sky UK",
"Only PSIP (ATSC)", "Apenas PSIP (ATSC)",
"Only UK Freesat", "Apenas UK Freesat",
"Only UK Freeview", "Apenas UK Freeview",
"Only Viasat Baltic", "Apenas Viasat Baltic",
"Only preferred CA PID", "Apenas CA PID preferidos",
"Opera", "Ópera",
"Orbital position", "Posição Orbital",
"Orbital positions", "Posições Orbitais",
"Orbital positions.", "Posições Orbitais.",
"Output directory for tsdebug", "Directório de saída do tsdebug",
"Over-the-air Cron multi-line", "Linhas do Cron para Over-the-air",
"Owner", "Proprietário",
"PEND", "PENDENTE",
"PIDs 21 in setup", "PIDs 21 na configuração",
"Password", "Palavra passe",
"Password2", "Palavra passe 2",
"Path", "Caminho",
"Performing arts", "Artes cénicas",
"Physiology", "Fisiologia",
"Picon path", "Caminho do Picon",
"Pilot", "Piloto",
"Polarization", "Polarização",
"Political issues", "Politica",
"Popular culture", "Cultura",
"Port", "Porta",
"Position", "Posição",
"Power save", "Poupança de Energia",
"Power-up time (ms) (15-200)", "Tempo de inicialização (ms)(15-200)",
"Pre-defined muxes", "Misturadores Pré-definidos",
"Pre-processor command", "Comando de Pré-processamento",
"Prefer picons over channel name", "Preferir picons em alternativa ao nome do canal",
"Preferred CA PID", "CA PID preferido",
"Preferred audio language.", "Idioma do audio preferido.  ",
"Presentation", "Apresentação",
"Presentation details.", "Detalhes de apresentação.",
"Press", "Notícias",
"Priority", "Prioridade",
"Priority (-10..10)", "Prioridade (-10..10)",
"Private", "Privado",
"Provider", "Provedor",
"Provider ID", "ID do Provedor",
"Psychology", "Psicologia",
"Puppets", "Animação",
"Re-fetch period (hours)", "Período até nova busca (horas)",
"Re-fetch period (mins)", "Período até nova busca (mins)",
"Re-record%s%s", "Gravações%s%s",
"Re-try period (hours)", "Período até nova tentativa (horas)",
"Record all", "Gravar Tudo",
"Reject", "Rejeitar",
"Religion", "Religião",
"Religious", "Religioso ",
"Remarkable people", "Pessoas Importantes",
"Remove HTTP arguments", "Remover argumentos HTTP",
"Reports", "Reportagens",
"Rerun of", "reexecução do ",
"Reserved", "Reservado",
"Resolution (height)", "Resolução (altura)",
"Rewrite EIT", "Reescrever EIT",
"Rewrite PAT", "Reescrever PAT",
"Rewrite PMT", "Reescrever PMT",
"Rewrite SDT", "Reescrever SDT",
"Run as group", "Executar como Grupo",
"Run as user", "Executar como Utilizador",
"Running", "Em Execução",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "SAT>IP número da porta RTSP do servidor\n(padrão: -1 = inactivo, 0 = webconfig, porta padrão é a 554)",
"SSL verify peer", "SSL para verificação do utilizador",
"Sat", "Sábado",
"Satellite longitude", "Longitude do Satélite",
"Satellite longitude.", "Longitude do Satélite",
"Satellite positions", "Posições dos Satélites",
"Scan progress", "Progresso da procura",
"Scan status", "Estado da Procura",
"Schedule a re-recording if more errors than (0=off)", "Agendar nova gravação se houver erros superior a (0=desligado)",
"Scheduled for recording", "Gravação Programada",
"Scheduled stop time", "Tempo de Finalização Programado",
"School programs", "Programas escolares",
"Science", "Ciência ",
"Science fiction", "Ficção científica ",
"Season", "Temporada",
"Season %d", "Temporada %d",
"Send CAPMT query", "Enviar consulta CAPMT",
"Send full PLAY cmd", "Enviar comando PLAY",
"Serial number", "Número de Série",
"Series link", "Ligação de Séries",
"Serious", "Real",
"Serious music", "Música ao vivo ",
"Server", "Servidor",
"Server UUID", "UUID do Servidor",
"Server connectivity", "Conexão do Servidor ",
"Service", "Serviço",
"Service ID", "ID do Serviço",
"Service configuration", "Configuração do Serviço",
"Service name", "Nome do Serviço",
"Service raw", "Serviço original",
"Service type", "Tipo de Serviço",
"Services", "Serviços",
"Settings", "Configurações",
"Show", "Espectáculo",
"Show this page", "Mostrar esta página",
"Show version information", "Mostrar informação acerca da versão",
"Signal quality too poor", "Qualidade do Sinal Muito Fraca",
"Signal scale (240 or 100)", "Escala do Sinal (240 ou 100)",
"Site latitude", "Latitude",
"Site latitude.", "Latitude",
"Site longitude", "Longitude",
"Site longitude.", "Longitude",
"Size", "Tamanho",
"Skip TS packets (0-200)", "Ignorar pacotes TS (0-200)",
"Skip commercials", "Ignorar Publicidade",
"Skip initial bytes", "Ignorar Bytes Iniciais ",
"Slave entry", "Entrada secundária",
"Slave entry.", "Entrada secundária",
"Slot state", "Estado do Slot",
"Soccer", "Futebol",
"Social advisory", "Aconselhamento social",
"Sort index", "Ordenar Índice",
"Sort index.", "Ordenar Índice",
"Source", "Origem",
"Source deleted", "Origem Apagada",
"Source quality is bad", "Qualidade da Origem é má",
"Source reconfigured", "Origem reconfigurada",
"Special events (Olympic Games, World Cup, etc.)", "Eventos especiais (Jogos Olímpicos, Campeonato do Mundo, etc.) ",
"Specify User-Agent header for the http client", "Especifique o cabeçalho User-Agent  para o cliente http",
"Specify alternative htsp port", "Especifique uma porta alternativa para htsp",
"Specify alternative http port", "Especifique uma porta alternativa para http",
"Specify alternative http webroot", "Especifique uma porta alternativa para http webroot",
"Specify bind address", "Especifique o endereço de ligação",
"Specify extra htsp port", "Especifique uma porta extra para htsp",
"Spiritual sciences", "Ciências Espirituais",
"Sports", "Desportos",
"Sports magazines", "Notícias de Desporto",
"Standard", "Padrão",
"Start", "Iniciar",
"Start after", "Iniciar Depois",
"Start before", "Iniciar Antes",
"Start time", "Início",
"Status", "Estado",
"Status period (ms)", "Estado do Período (ms) ",
"Stop", "Parar",
"Stop time", "Termina",
"Storage path", "Caminho para o armazenamento",
"Storage space", "Espaço de Armazenamento",
"Stream ID", "ID do perfil",
"Stream Profile", "Perfil de Transmissão",
"Stream index", "Índice de Transmissão",
"Stream profile", "Perfil de Transmissão",
"Stream type", "Tipo de Transmissão",
"Streaming", "Transmissão",
"Streaming priority", "Prioridade da Transmissão",
"Streaming profiles", "Perfis de Transmissão",
"Subdirectory options", "Opções do Sub-directório",
"Subscribe to a service permanently", "Subscrever o serviço permanentemente",
"Subscription overridden", "Sobrescrever Subscrição",
"Subscription weight", "Peso da subscrição ",
"Substitute formatters", "Substituir formatadores",
"Subtitle", "Legenda",
"Subtitle Stream Filter", "Filtro de Transmissão da Legenda",
"Subtitle codec", "Codec da Legenda",
"Sun", "Domingo",
"Switch: %s", "Mudança: %s",
"Sync", "Sincronizar",
"Sync + Don't keep", "Sincronizar + Não manter",
"System", "Sistema",
"Tags", "Etiquetas",
"Team sports (excluding football)", "Desporto de equipas (excluindo o Futebol)",
"Technology", "Tecnologia ",
"Tennis", "Ténis",
"Testing", "Teste",
"Testing options", "Opções de Teste",
"The device ID.", "ID do Dispositivo",
"The layer A constellation.", "Layer A: Constelação",
"The layer A segment count.", "Layer A: Contagem de segmentos",
"The layer A time interleaving.", "Layer A: Intervalo de tempo",
"The layer B constellation.", "Layer B: Constelação",
"The layer B segment count.", "Layer B: Contagem de segmentos",
"The layer B time interleaving.", "Layer B: Intervalo de tempo",
"The layer C constellation.", "Layer C: Constelação",
"The layer C segment count.", "Layer C: Contagem de segmentos",
"The layer C time interleaving.", "Layer C: Intervalo de tempo",
"The service ID", "ID do Serviço",
"The service ID.", "ID do Serviço",
"The transponder ID.", "ID do Transponder",
"This tuner", "Este sintonizador",
"Thu", "Quinta-Feira",
"Time", "Tempo",
"Time missed", "Tempo perdido",
"Time record caption", "Tempo de Gravação das Legendas",
"Time recording%s%s", "Tempo de Gravação%s%s",
"Time update", "Actualização do tempo",
"Timeout (sec) (0=infinite)", "Tempo limite (sec) (0=infinito)",
"Timeout (seconds)", "Tempo limite (segundos)",
"Timeshift", "Gravação em tempo real",
"Title", "Título",
"Title (regexp)", "Título (Regexp)",
"Toggle details", "Mostrar detalhes",
"Tone burst", "Tone Burst",
"Too many data errors", "Muitos erros nos dados",
"Total number of mapped channels on this network.", "Número total de canais mapeados desta rede.",
"Total number of muxes found on this network.", "Número total de muxes encontrados nesta rede.",
"Total number of services found on this network.", "Número total de serviços encontrados nesta rede.",
"Total number of services found.", "Número total de serviços encontrados.",
"Total number of services seen.", "Número total de serviços analisados.",
"Total number of services.", "Número total de serviços.",
"Tourism / Travel", "Viagens / Turismo",
"Trace subsystems", "Rastreamento dos subsistemas",
"Traditional arts", "Artes Tradicionais",
"Traditional music", "Música Tradicional ",
"Transcode/av-lib", "Transcodificar/av-lib",
"Transcoding", "Converter",
"Transmission mode", "Modo Transmissão",
"Transponder ID", "ID do Transponder",
"Transport stream ID", "ID da Transmissão",
"Tue", "Terça-Feira",
"Tune before DiseqC", "Sintonizar antes do DiseqC",
"Tuner", "Sintonizador",
"Tuner bind IP address", "Sintonizador vinculado ao endereço de IP",
"Tuner configuration", "Configuração do Sintonizador",
"Tuner configuration.", "Configuração do Sintonizador",
"Tuning failed", "Falha na Sintonização",
"Turn off LNB when idle", "Desligar LNB quando em espera",
"Tvheadend server name", "Nome do servidor Tvheadend",
"Type", "Tipo",
"UDP RTP port number (2 ports)", "UDP RTP Número das Portas (2 portas)",
"URL of the M3U playlist.", "URL da lista M3U.",
"URL with the SAT>IP server XML location", "URL com a Localização XML do Servidor SAT>IP",
"Unable to get system UTC time", "Incapaz de adquirir a hora UTC",
"Unable to get system local time", "Incapaz de adquirir a hora local",
"Unable to get system time", "Incapaz de adquirir as horas do sistema",
"Unable to open file", "Incapaz de abrir o ficheiro",
"Unable to stat path", "Incapaz de definir o caminho",
"Uncommitted", "Não autorizado",
"Uncommitted first", "Em primeiro os não autorizados",
"Unicable type", "Tipo Unicable",
"Unimportant", "Não Importante",
"Universal LNB only", "Apenas LNB Universal",
"Universally unique identifier. Read only.", "Identificador único. Apenas leitura.",
"Unknown", "Desconhecido",
"Unknown file", "Ficheiro desconhecido",
"Unknown reason (%i)", "Razão desconhecida (%i)",
"Unlimited size", "Tamanho ilimitado",
"Unlimited time", "Tempo ilimitado",
"Unset (default)", "Desactivado (padrão)",
"Update card (EMM)", "Actualizar cartão (EMM)",
"Update channel icon", "Actualizar ícone do canal",
"Update channel name", "Actualizar o nome do canal",
"Update channel number", "Actualizar o número do canal",
"Update disabled", "Actualização desactivada",
"Update time", "Actualizar tempo",
"Update tolerance (ms)", "Actualizar tolerância (ms)",
"Updated", "Actualizado",
"Updates from one mux (EMM)", "Actualizar a partir de um mux (EMM)",
"Usage: %s [OPTIONS]\n", "Utilização: %s [OPCÕES]\n",
"Use EPG running state", "Utilizar o estado do EPG",
"Use HTTP digest authentication", "Utilizar autenticação HTTP digest",
"Use WEBM format.", "Utilizar o formato WEBM.",
"Use Windows-compatible filenames", "Utilizar Nomes nos Ficheiros Compatíveis com Windows",
"Use XSPF playlist instead of M3U", "Utilizar lista XSPF em vez de M3U",
"Use original", "Utilizar o Original",
"Use packet backlog", "Utilizar aplicação backlog",
"Use service IDs as channel numbers", "Utilizar ID dos serviços como número dos canais",
"Use tag internally (don't expose to clients).", "Utilizar a etiqueta internamente (não mostrar aos utilizadores)",
"Use the provider's service IDs as channel numbers.", "Utilizar a numeração dos canais dos provedores.",
"Use this character set when setting filenames.", "Utilizar este conjunto de caracteres nos nomes dos ficheiros",
"Use x amount of storage space.", "Utilizar x quantidade de espaço em disco",
"User access error", "Erro de acesso do Utilizador",
"User does not have access", "O utilizador não tem acesso",
"User does not have access to channel", "O utilizador não tem acesso a este canal",
"User icon", "Ícone do Utilizador",
"User interface level", "Nível da interface do utilizador",
"User interface quick tips (tooltips)", "Ajudas rápidas ao utilizador (tooltips)",
"User language", "Linguagem do utilizador",
"User limit reached", "Tempo Limite do Utilizador Atingido",
"User login", "Login do utilizador",
"User request", "Pedido do utilizador",
"Username", "Nome de Utilizador",
"Username for the entry (login username).", "Nome do utilizador (login)",
"Users - IP Blocking", "Utilizadores - Bloqueador IP",
"Users - Passwords", "Utilizadores - Palavras-passes",
"Variety show", "Espectáculo de Variedades",
"Verify the SSL certificate.", "Verificar certificado SSL",
"Video", "Filme",
"Video Stream Filter", "Filtro de Transmissão de Vídeo",
"Video bitrate (kb/s) (0=auto)", "Taxa de bits do Vídeo (kb/s) (0=Auto)",
"Video codec", "Codec de vídeo",
"Video codec preset", "Codec de vídeo pré-definido",
"Video codec preset to use for transcoding.", "Codec de vídeo pré-definido para utilizar na conversão",
"Video recorder", "Gravador de Vídeo",
"View all", "Ver todos",
"Visually impaired commentary/audio description", "Comentário de deficiência visual/ descrição do som",
"Vorbis audio", "Audio Vorbis",
"WEBTV profile H264/AAC/MPEG-TS", "Perfil WEBTV H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "Perfil WEBTV H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "Perfil WEBTV VP8/Vorbis/WEBM",
"Waiting for EPG running flag", "Aguardando o início do EPG",
"Waiting for program start", "Aguardando o início do programa",
"Waiting for stream", "À espera da Transmissão",
"War", "Guerra",
"Water sport", "Desportos aquáticos ",
"Weak stream", "Transmissão fraca",
"Weather report", "Meteorologia ",
"Web User Interface", "Utilizador Interface Web",
"Web interface", "Interface Web",
"Web interface language", "Idioma da Interface Web",
"Web interface language.", "Idioma da Interface Web",
"Web interface theme.", "Tema da interface Web",
"Web theme", "Tema  Web",
"Web user interface", "Interface Web do Utilizador",
"Wed", "Quarta-Feira",
"Winter sports", "Desportos de inverno",
"Wizard", "Assistente de configuração",
"XMLTV EPG Import", "Importar XMLTV EPG",
"Yes", "Sim",
"addpids/delpids supported", "Suporte addpids/delpids",
"default: nvenc(h264 / h265)", "por omissão: nvenc(h264 / h265)",
"fast: h264 / h265 / qsv(h264 / h265)", "rápido: h264 / h265 / qsv(h264 / h265)",
"faster: h264 / h265 / qsv(h264)", "mais rápido: h264 / h265 / qsv(h264)",
"invalid option specified [%s]", "Opção especificada inválida [%s]",
"logout", "Terminar sessão",
"medium: h264 / h265 / qsv(h264 / h265)", "médio: h264 / h265 / qsv(h264 / h265)",
"option %s requires a value", "Opção %s requer um valor",
"return", "voltar",
"slow: h264 / h265 / qsv(h264 / h265)", "lento: h264 / h265 / qsv(h264 / h265)",
"slower: h264 / h265 / qsv(h264)", "mais lento: h264 / h265 / qsv(h264)",
"superfast: h264 / h265", "super rápido: h264 / h265",
"tsfile input (mux file)", "entrada tsfile (misturar ficheiro)",
"tvhdhomerun client", "cliente tvhdhomerun",
"ultrafast: h264 / h265", "ultra rápido: h264 / h265",
"veryfast: h264 / h265 / qsv(h264)", "muito rápido: h264 / h265 / qsv(h264)",
"veryslow: h264 / h265 / qsv(h264)", "muito lento: h264 / h265 / qsv(h264)",
NULL, NULL
};

static const char *tvh_locale_27[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nPentru mai multe informatii vizitati Tvheadend pe:\nhttps://tvheadend.org\n",
"# Services", "# Servicii",
NULL, NULL
};

static const char *tvh_locale_28[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nДля получения более подробной информации, пожалуйста, посетите веб-сайт Tvheadend по адресу:\nhttps://tvheadend.org\n",
"# Mapped channels", "# Сопоставленные каналы",
"# Muxes", "# Мультиплексы",
"# Services", "# Сервисы",
"# tune repeats", "Настроить повторы",
"(Default profile)", "(Профиль по умолчанию)",
"(Read-only)", "(Только чтение)",
"1 day", "1 день",
"1 month", "1 месяц",
"1 week", "1 неделя",
"1 year", "1 год",
"2 months", "2 месяца",
"2 weeks", "2 недели",
"2 years", "2 года",
"3 days", "3 дня",
"3 months", "3 месяца",
"3 weeks", "3 недели",
"3 years", "3 года",
"5 days", "5 дней",
"6 months", "6 месяцев",
"A", "А",
"A channel tag (e.g. a group of channels) to which this rule applies.", "Тег каналов (например, группа каналов), к которым это правило применяется.",
"AA", "АА",
"ACTIVE", "АКТИВНЫЙ",
"ANY", "ЛЮБОЙ",
"AUTO", "АВТО",
"Aborted by user", "Отменено пользователем",
"Accept remote subscription weight", "Принять вес удаленной подписки",
"Accept the remote subscription weight (from the SAT>IP client).", "Принять вес удаленной подписки (от SAT>IP клиента).",
"Access", "Доступ",
"Access (ACL)", "Доступ (ACL)",
"Action", "Экшн",
"Adapter in use by another subscription", "Адаптер занят другой подпиской",
"Add file and line numbers to debug", "Добавить файл и число строк к дебагу",
"Add the thread ID to debug", "Добавить тред ID в дебаг",
"Additional service names found in EPG data.", "Доп. названия служб, найденные в данных EPG.",
"Admin", "Админ",
"Admin password", "Пароль администратора",
"Admin username", "Имя администратора",
"Administrator login", "Логин администратора",
"Adult movie", "Кино для взрослых 18+",
"Advanced", "Расширенный",
"Advanced (non-universal LNBs, rotors, etc.)", " Расширенный (Non-Universal LNBs, Rotors, etc.)",
"Adventure", "Приключения",
"Advertisement / Shopping", "Реклама / Магазины",
"All (Streaming plus DVR)", "Всё (Потоковое плюс DVR)",
"All lower-case", "Всё в нижнем регистре",
"Allowed network", "Доступные сети",
"Allowed networks", "Доступные сети",
"Alternate PID path", "Альтернативный путь ID",
"Alternate configuration path", "Альтернативный путь к конфигурации",
"Altitude (meters)", "Altitude (метры)",
"Animals", "Про животных",
"Any", "Любой",
"Arts", "Искусство",
"Athletics", "Атлетика",
"Audio bitrate (kb/s) (0=auto)", "Аудио битрейт (кбит/с) (0=Авто)",
"Audio codec", "Аудио кодек",
"Auto", "Авто",
"Auto check disabled", "Авто проверка отключена",
"Auto check enabled", "Авто проверка включена",
"Auto record", "Авто Запись",
"Auto record caption", "Автоматическая метка записи",
"Auto recording%s%s", "Время Записи%s%s",
"Auto time record", "Время Автомат. Записи",
"Auto-Map to channels", "Авто-сопоставление каналов",
"Automatic checking", "Автоматическая проверка",
"Automatically map EPG source", "Автоматически сопоставлять источник EPG",
"Automatically name from network", "Автоматическое имя из сети",
"Autorec maximum count (0=unlimited)", "Максимальное количество автозаписей (0=не ограничено)",
"Autorec maximum schedules limit (0=unlimited)", "Ограничение количества расписаний автозаписи (0=неограничено)",
"Bad", "Плохой",
"Ballet", "Балет",
"Bandwidth", "Пропускная способность",
"Based on software from", "Основано на софте от",
"Basic", "Основной",
"Bind to specific local IP address.", "Привязать к определенному IP адресу",
"Boot ID", "Загрузочный ID",
"Bouquet", "Букет",
"Bouquet (auto)", "Букет (Авто)",
"Brand", "Марка",
"Broadcast", "Трансляция",
"Broadcast type", "Тип Трансляции",
"Broadcast.", "Трансляция.",
"Broadcasting", "Транслирование",
"Build", "Билд",
"CA client %i", "CA Клиент %i",
"CA identification", "CA Идентификация",
"CA provider", "CA Провайдер",
"CAPMT interval (ms)", "CAPMT интервал (мсек.)",
"CAPMT query interval (ms)", "CAPMT Интервал запроса (мсек.)",
"CRID authority", "CRID Полномочия",
"Cache scheme", "Схема Кэша",
"Camd.socket filename / IP Address (TCP mode)", "Camd.socket Имя файла / IP Адрес (TCP режим)",
"Cartoons", "Мультфильмы",
"Channel", "Канал",
"Channel Tags", "Теги канала",
"Channel does not exist", "Канал не существует",
"Channel icon", "Иконка Канала",
"Channel icon URL.", "URL Иконки Канала",
"Channel icon path", "Путь к иконкам каналов",
"Channel name", "Имя канала",
"Channel not found", "Канал не найден",
"Channel number", "Номер канала",
"Channel number offset", "Смещение Номера Канала",
"Channel numbers from", "Номера каналов от",
"Channel tag", "Тег канала",
"Channel tag does not exist", "Тег канала не существует",
"Channel tag reference", "Упоминаемые метки канала",
"Channel tags", "Теги канала",
"Channels", "Каналы",
"Character set", "Кодировка",
"Charset", "Кодировка",
"Check availability", "Проверить доступность",
"Children's / Youth programs", "Детские / Молодежные программы",
"Cinema", "Кинематография",
"Class", "Класс",
"Classical", "Классика",
"Classical music", "Классическая музыка",
"Clean effects", "Вычистить эффекты",
"Client name", "Имя Клиента",
"Clone scheduled entry on error", "Клонирование Запланированной записи при Ошибке",
"Comedy", "Комедия",
"Command delay time (ms) (10-200)", "Команда задержки времени (милисек) (10-200)",
"Command time (ms) (10-100)", "Время комманды (милисек) (10-100)",
"Comment", "Коментарий",
"Commercial break", "Рекламная пауза",
"Completed OK", "Завершено успешно",
"Compress EPG database", "Сжать базу EPG",
"Conditional Access", "Условный доступ",
"Conditional Access Client", "Клиент условного доступа",
"Conditional Access Identification.", "CA Идентификация",
"Configuration", "Конфигурация",
"Configuration ID", "ID Конфигурации",
"Configuration name", "Название конфигурации",
"Configuration version", "Версия конфигурации",
"Connection limit type", "Тип ограничения соединения",
"Constellation", "Созвездие",
"Container", "Контейнер",
"Content character set", "Кодировка Контента",
"Content type", "Тип контента",
"Content type.", "Тип контента.",
"Contest", "Конкурс",
"Cookie expiration (days)", "Время жизни Cookie (Дней)",
"Cooking", "Кухня",
"Copy codec type", "Копировать тип кодека",
"Copy layout", "Копировать разметку",
"Count of objects", "Число объектов",
"Create bouquet", "Создать букет",
"Create bouquet tag", "Создать метку букета",
"Create network name tags", "Создать метку имени сети",
"Create network tags", "Создать метку сети",
"Create provider name tags", "Создать метку имени провайдера",
"Create provider tags", "Создать метки провайдеров",
"Create tags", "Создать метки",
"Create type-based tags", "Создать метки на основе типа",
"Created", "Создан",
"Creator", "Создатель",
"Cron", "Крон",
"Culture (without music)", "Культура (исключая музыку)",
"Current RTSP port.", "Текущий RTSP порт",
"Current affairs", "Текущие события",
"Current number of objects.", "Текущее число объектов",
"Current object size.", "Текущий размер объекта.",
"Current tuner configuration.", "текущая конфигурация тюнера",
"DBus - use the session message bus instead of the system one", "DBus - использовать сообщения шины вместо системных",
"DES key", "Ключ DES",
"DVB EPG ID", "DVB ЕПГ ID",
"DVB scan files", "DVB Scan Files",
"DVB-C Network", "DVB-C Сеть",
"DVB-S Network", "DVB-S Сеть",
"DVB-T Network", "DVB-T Сеть",
"DVR - Profiles", "DVR - Профили",
"DVR behavior", "DVR режим",
"DVR configuration", "DVR Конфигурация",
"DVR configuration profiles", "Профили Конфигурации DVR",
"DVR entry not found", "DVR Запись не найдена",
"Dance", "Танцы",
"Data errors", "Ошибки Данных",
"Days of Week", "Дни недели",
"Days of week", "Дни недели",
"Debate", "Дебаты",
"Debug options", "Настройки отладки",
"Debugging", "Отладка",
"Default", "По умолчанию",
"Default language(s)", "Язык(и) по умолчанию",
"Default language.", "Язык по умолчанию.",
"Default priority", "Приоритет по умолчанию",
"Default user interface level.", "Уровень пользовательского интерфейса по умолчанию",
"Delivery system", "Система доставки",
"Demux path", "Путь к Демультиплексу",
"Descramble services (limit per mux)", "Descramble Services (Limit Per Mux)",
"Description", "Описание",
"Detective", "Детектив",
"Device ID", "ID Устройства",
"Device model", "Модель устройства",
"Device path", "Путь к устройству",
"Directory", "Каталог",
"Directory permissions (octal, e.g. 0775)", "Права Директории (восьм., пример - 0775)",
"Disable", "Отключить",
"Disable DVB bouquets", "Отключить DVB Букеты",
"Disable SAT>IP client", "Заблоктровать SAT>IP клиента",
"Disable all access control checks", "Отключить проверку контроля доступа",
"Disable debug on stderr", "Заблокировать дебаг на stderr",
"Disable device/firmware-specific workarounds", "Отключить устройство/firmware-specific обходной путь",
"Disable syslog (all messages)", "Отключить системный лог (все сообщения)",
"Disabled", "Отключено",
"Discussion", "Обсуждения",
"DiseqC repeats", "DiseqC повторы",
"Do not use", "Не использовать",
"Documentary", "Документальные",
"Don't backup configuration tree at upgrade", "Не резервируется дерево конфигурации при обновлении",
"Don't include title in filename", "Не включая Титл в имени Файла",
"Don't keep", "Не держать",
"Don't re-record", "Не перезаписывать",
"Don't reschedule", "Не перепланировать",
"Drama", "Драма",
"EPG Grabber", "ЕПГ Граббер",
"EPG name", "Название EPG",
"EPG scan", "Сканирование ЕПГ",
"EPG source", "ЕПГ источник",
"EPG update window", "Окно обновления ЕПГ",
"Economics", "Экономика",
"Education", "Образование",
"Educational", "Образовательные",
"Elementary stream filter", "Простой фильтр потока",
"Enable (auto)", "Включение (авто)",
"Enable DBus", "Включение DBus",
"Enable NTP driver", "Включить драйвер NTP",
"Enable coredumps for daemon", "Включить дамп ядра для демона",
"Enable debug on stderr", "Включение дебага на stderr",
"Enable debug subsystems", "Включение подсистемы дебага",
"Enable debug to file", "Включение дебага в файл",
"Enable debug to syslog", "Включить дебаг в syslog",
"Enable or disable fast input switching.", "Включить или отключить быстрое переключение ввода. ",
"Enable or disable full mux mode.", "Включить или отключить полный мукс режим. ",
"Enable or disable this configuration.", "Включить или отключить эту конфигурацию. ",
"Enable syslog", "Включить syslog",
"Enable trace subsystems", "Включение подсистемы трассировки",
"Enable web UI debug (non-minified JS)", "Enable webUI debug (non-minified JS)",
"Enabled", "Включено",
"Encrypted", "Зашифрованные",
"Encrypted service", "Зашифрованный сервис",
"Enter a non-admin user password.", "Введите пароль не администратора",
"Enter a non-admin user username.", "Введите имя не администратора",
"Entertainment programs for 10 to 16", "Развлекательные программы от 10 до 16",
"Entertainment programs for 6 to 14", "Развлекательные программы от 6 до 14",
"Environment", "Окружение",
"Environment (pipe)", "Окружение (pipe)",
"Episode", "Эпизод",
"Episode %d", "Эпизод %d",
"Error code", "Код Ошибки",
"Errors", "Ошибки",
"Every day", "Каждый день",
"Exclude channel tags", "Исключая метки канала",
"Expeditions", "Экспедиции",
"Experimental film", "Эксперементальное кино",
"Expert", "Экспертный",
"Exported tuners", "Экспортированные Тюнеры",
"External", "Внешний",
"External EPG grabber", "Внешний EPG граббер",
"External IP (NAT)", "Внешний IP (NAT)",
"External URL", "Внешний URL",
"Extra arguments", "Эктра аргументы",
"Extra start time", "Дополнительное Время старта",
"Extra stop time", "Дополнительное Время остановки",
"FAIL", "ПРОВАЛ",
"FEC high", "FEC High",
"FEC low", "FEC Low",
"Factual topics", "Фактические темы",
"Failed", "Неудачно",
"Fantasy", "Фентези",
"Fast input switch", "Быстрый переключатель ввода",
"File missing", "Файл Отсутствует",
"File not created", "Файл не создан",
"File permissions (octal, e.g. 0664)", "Права на Файл (восьм. например 0664)",
"File size", "Размер файла",
"Filename", "Имя файла",
"Filename character set", "Набор символов в имени файла",
"Filename options", "Параметры имени файла",
"Film", "Кино",
"Fine arts", "Изящные искусства",
"Finished", "Окончено",
"Fitness and health", "Фитнес и здоровье",
"Folk", "Народное",
"Folkloric", "Фольклёр",
"Football", "Футбол",
"Force (auto)", "Форсировать (авто)",
"Force priority", "Форсировать Приоритетное",
"Foreign countries", "Зарубежные страны",
"Forever", "Навсегда",
"Fork and run as daemon", "Перезапустить как Демон",
"Format string", "Формат строки",
"Found muxes", "Найдено мультиплексов",
"Found services", "Найдено сервисов",
"Free subscription weight", "Свободный вес подписки",
"Frequency", "Частота",
"Frequency (Hz)", "Частота (Hz)",
"Frequency (MHz)", "Частота (MHz)",
"Frequency (in MHz).", "Частота (MHz)",
"Frequency (kHz)", "Частота (kHz)",
"Fri", "Пятн.",
"Friendly name", "Дружественное Имя",
"Frontend number", "Номер Внешнего Интерфейса",
"Frontend path", "Путь к фронтенду",
"Full DiseqC", "Полный DiseqC",
"Full pathname specification", "Спецификация полного пути",
"Further education", "Дальнейшее образование",
"GOTOX", "Переход на Х",
"GOTOX position", "Переход на Х Позицию",
"Game show", "Игровое Шоу",
"Gardening", "Сад и огород",
"General", "Главная",
"General configuration", "Главная Конфигурация",
"Generic", "Общие",
"Generic options", "Общие Настройки",
"Guard interval", "Защитный интервал",
"HD: high definition", "HD: Высокое Разрешение",
"HDHomeRun DVB frontend", "HDHomeRun DVB Frontend",
"HDHomeRun DVB-C frontend", "HDHomeRun DVB-C Frontend",
"HDHomeRun DVB-T frontend", "HDHomeRun DVB-T Frontend",
"HTSP Default Stream Settings", "Настройки потока HTSP по умолчанию",
"Handicraft", "Ремёсла",
"Hearing impaired", "Для глухих",
"Hierarchy", "Иерархия",
"High", "Высокий",
"High bitrate mode (CI+ CAMs only)", "Режим с Высоким Битрейтом (CI+ CAMы только)",
"Historical movie", "Исторические фильмы",
"Horror", "Ужасы",
"Hostname/IP", "Имя Хоста/IP",
"IDLE", "ПРОСТОЙ",
"IP address", "IP адрес",
"IP blocking", "Блокировка IP",
"IPTV input", "IPTV Ввод",
"Icon", "иконка",
"Icon (full URL)", "Иконка (полный URL)",
"Icon URL", "URL иконки",
"Icon base URL", "URL иконки",
"Icon has title", "Иконка имеет название",
"Icons from", "иконки от",
"Idle", "простой",
"Idle scan", "Сканирование при простое",
"Idle scan muxes", "Сканирование Мультиплексов в простое",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "если аккаунт пользователя не существует, создайте профиль\nбез имени и пароля. Используйте с осторожностью\nэто открывает полный административный доступ\nк вашей Tvheadend системе с полным доступом\nи контролю через Tvheadend web интерфейс.",
"If you'd like to support the project, please consider a donation.", "Если вы хотите помочь проекту, пожалуйста рассмотрите возможность пожертвования.",
"Ignore EPG (EIT)", "Игнорировать ЕПГ (EIT)",
"Ignore descriptor 5", "игнорировать дескриптор 5",
"Ignore invalid SSL certificate", "игнорировать неправильные SSL сертификаты",
"Ignore provider's channel numbers", "Игнорировать Провайдерские Номера Каналов",
"Immediately abort", "Сразу прервать",
"Important", "Важный",
"Include channel name in filename", "Включая Имя Канала в имени Файла",
"Include date in filename", "Вставить дату в имя файла",
"Include episode in filename", "Включить эпизод в имени файла",
"Include subtitle in filename", "Включить субтитры в имени файла",
"Include time in filename", "Вставить время в имя файла",
"Index", "Индекс",
"Information area", "Информационный раздел",
"Informational", "Информационный",
"Initial scan", "Первоначальное сканирование",
"Input base", "Базовый ввод",
"Input buffer (bytes)", "Входной Буфер (Байты)",
"Input instance", "Ввод экземпляра",
"Input path", "Входной путь",
"Interface", "Интерфейс",
"Internal", "Внутренний",
"Internal EPG grabber", "Внутренний EPG граббер",
"Internal grabber", "Внутренний граббер",
"Interview", "интервью",
"Invalid", "Неправильный",
"Invalid adapter number '%s'\n", "Неправильный номер адаптера '%s'\n",
"Invalid target", "Неправильная цель",
"Jazz", "Джаз",
"Keep", "Оставить",
"LNB type", "Тип LNB",
"Language", "Язык",
"Language settings", "Настройки Языковые",
"Languages", "Языки",
"Last seen", "Последнее просмотренное",
"Last updated from", "Последнее обновление из",
"Leisure hobbies", "Досуг и отдых",
"Limit connections", "Лимит Соединений",
"Linked input", "Соединенный вход",
"LinuxDVB adapter", "LinuxDVB Адаптер",
"LinuxDVB network", "LinuxDVB Сеть",
"Listen / Connect port", "Прослушивать/Порт соединения",
"Listen on IPv6", "Прослушивать в IPv6",
"Literature", "Литература",
"Local bind IP address", "Локальная привязка IP адреса",
"Local channel minor", "Локальный минор Канала",
"Local channel number", "Локальный Номер Канала",
"Local discovery IP address", "Локальное определение IP адреса",
"Location", "Расположение",
"Lock preferred CA PID", "Виксированный Предпочтительный CA PID",
"Log", "Лог",
"Login/Logout", "Вход/Выход",
"Low", "Низкий",
"MPEG-TS input", "MPEG-TS Вход",
"Magazines", "Журналы",
"Maintain free storage space in MiB", "Удерживать свободное место на носителе в MiB",
"Maintain used storage space in MiB (0=disabled)", "Удерживать занятое место на носителе в MiB (0=отключено)",
"Maintained space", "Поддерживаемое пространство",
"Make subdirectories per channel", "Сделать Поддиректории по Каналу",
"Make subdirectories per day", "Сделать Поддиректории по Дням",
"Make subdirectories per title", "Сделать Поддиректории по Титл",
"Manufacturer", "Производитель",
"Manufacturer URL", "URL Производителя",
"Manufacturer's model number.", "Номер модели производителя.",
"Manufacturer's product description.", "Описание продукта производителя.",
"Manufacturer's product information page for the device.", "Информация производителя об устройстве.",
"Manufacturer's product name.", "Название продукта производителя.",
"Map radio channels", "Отображение радио каналов",
"Map unnamed channels", "Отобразить не именованные каналы",
"Map zero-numbered channels", "Отобразить нулевые каналы",
"Master tuner", "Мастер Тюнер",
"Matroska", "Матрёшка",
"Matroska (mkv)/av-lib", "Матрёшка (mkv)/av-lib",
"Matroska (mkv)/built-in", "Матрёшка (mkv)/built-in",
"Maximal channel number", "Максимальный номер канала",
"Maximum # input streams", "Максимум # входных Потоков",
"Maximum PIDs", "Максимум PIDs",
"Maximum RAM size (MB)", "Максимальный размер RAM (MB)",
"Maximum bandwidth (Kbps)", "Максимальная Пропускная (Кб в сек.) ",
"Maximum duration", "Максимальная Продолжительность",
"Maximum length of PIDs", "Максимальная длинна PIDов",
"Maximum period (mins)", "Макисмальный Период (мин.)",
"Maximum size (MB)", "Максимальный Размер (MB)",
"Maximum timeout (seconds)", "Максимум таймаут (сек.)",
"Medicine", "Медицина",
"Melodrama", "Мелодрама",
"Minimal channel number", "Минимальный номер канала",
"Minimum duration", "Минимальная Продолжительность",
"Minimum rotor time (seconds)", "Мин Ротор Время (сек.)",
"Missing In PAT/SDT", "Отсутствует в PAT/SDT",
"Mode", "Режим",
"Model description", "Описание Модели",
"Model name", "Название модели",
"Model number", "Номер модели",
"Modulation", "Модуляция",
"Module", "Модуль",
"Module ID", "ID Модуля",
"Mon", "Пнд.",
"Mono", "Моно",
"More verbose libav log", "Более читабельный libav лог",
"Motor sport", "Мото спорт",
"Movie", "Кино",
"Music", "Музыка",
"Musical", "Музыкальный",
"Mux", "Мультиплекс",
"Mux UUID", "UUID Мультиплекса",
"Mux name", "Название Мультиплекса",
"Mux not enabled", "Мультиплекс не включен",
"NONE", "НЕТ",
"Name", "Имя",
"Names", "Имена",
"Natural sciences", "Естественные Нуки",
"Nature", "Природа",
"Network", "Сеть",
"Network A", "Сеть A",
"Network B", "Сеть B",
"Network ID (limit scanning)", "Сетевой ID (ограничение сканирования)",
"Network UUID", "UUID Сети",
"Network access", "Сетевой доступ",
"Network discovery", "Обнаружение Сети",
"Network name", "Название Сети",
"Network prefix", "префикс Сети",
"Network settings", "Настройки сети",
"Network type", "Тип Сети",
"Networks", "Сети",
"New muxes + changed muxes", "Новые муксы + измененные муксы",
"New muxes only", "Только новые муксы",
"News", "Новости",
"News magazine", "Журналы новостей",
"Next tune delay in ms (0-2000)", "Задержка следующей настройки в мсек. (0-2000)",
"No", "Нет",
"No PIN", "Нет PIN",
"No access", "Нет доступа",
"No adapters specified!\n", "Адаптеры не указаны!\n",
"No assigned adapters", "Нет назначенного адаптера.",
"No days", "Нет дней",
"No descrambler", "Нет дешифратора",
"No free adapter", "Нет свободного адаптера",
"No input detected", "Вход  не обнаружен",
"No service assigned to channel", "Не назначен сервис каналу",
"No service enabled", "Сервис не включен",
"No source available", "Источник не доступен",
"None", "Нет",
"Normal", "Нормальный",
"Not set", "Не задано",
"Not set (none or channel configuration)", "Не задан (нет или канал  не сконфигурирован)",
"Not set (use channel or DVR configuration)", "Не задан (используйте канал или конфигурацию DVR)",
"Number", "Число",
"Number of seconds to wait before timing out.", "Сколько секунд ждать перед таймаут.",
"Number of tsfile tuners", "Число tsfile тюнеров ",
"OK", "ОК",
"OK (partial)", "Ок (частично)",
"OSCam net protocol (rev >= 10389)", "OSCam сетевой протокол (rev >= 10389)",
"Off", "Выкл.",
"Older OSCam", "Старый OSCam",
"On", "Вкл.",
"On file removal", "При удалении файла",
"Only Bulsatcom 39E", "Только Bulsatcom 39E",
"Only EIT", "Только EIT",
"Only OpenTV Sky Ausat", "Только OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Только OpenTV Sky Italia",
"Only OpenTV Sky UK", "Только OpenTV Sky UK",
"Only PSIP (ATSC)", "Только PSIP (ATSC)",
"Only UK Freesat", "Только UK Freesat",
"Only UK Freeview", "Только UK Freeview",
"Only Viasat Baltic", "Только Viasat Baltic",
"Only preferred CA PID", "Только Предпочтительный CA PID",
"OpenTV channel number", "Номер Канала OpenTV",
"Opera", "Опера",
"Orbital position", "Орбитальная позиция",
"Orbital positions", "Орбитальные Позиции",
"Original network ID", "Настоящий ID Сети",
"Output directory for tsdebug", "Директория вывода tsdebug",
"Over-the-air", "По воздуху",
"Over-the-air EPG", "ЕПГ по воздуху",
"Over-the-air EPG grabber", "Граббер ЕПГ по воздуху",
"Over-the-air grabbers", "Грабберы по воздуху",
"Override the frontend type.", "Заместит тип фронтэнда.",
"Owner", "Владелец",
"PEND", "ОЖИДАНИЕ",
"PLS code", "PLS Код",
"PLS mode", "PLS режим",
"Parent entry", "Родительская запись",
"Password", "Пароль",
"Password2", "Пароль2",
"Path", "Путь",
"Path to a directory (folder) containing your picon collection. See Help for more detailed information.", "Путь к директории с коллекцией пиконов. Смотри помощь для детальной информации.",
"Path to an icon for this channel. This can be named however you wish, as either a local (file://) or remote (http://) image. See Help for more infomation.", "Путь к иконке этого канала. This can be named however you wish, as either a local (file://) or remote (http://) image. Смотри помощь для более детальной информации.",
"Performing arts", "Исполнительские виды искусства",
"Periodically save EPG to disk (hours)", "Периодичность сохранения EPG на диск (часы)",
"Physiology", "Физиология",
"Picon", "Пикон",
"Pilot", "Пилот",
"Polarization", "Поляризация",
"Political issues", "Политические вопросы",
"Pop", "Поп",
"Popular culture", "Популярная культура",
"Port", "Порт",
"Position", "Позиция",
"Position of the input.", "Позиция ввода.",
"Post-processor command", "Команда Пост-Процессора",
"Post-remove command", "Команда после удаления",
"Power save", "Энергосбережение",
"Power-up time (ms) (15-200)", "Время включения (милисек.) (15-200)",
"Pre-defined muxes", "Предопределённые Мультиплексы",
"Pre-school children's programs", "Дошкольные детские программы",
"Prefer picons over channel name", "Предпочтительные пиконы  на имя канала",
"Preferred CA PID", "Предпочтительные CA PID",
"Preferred service video type", "Предпочтительные типы Видео Сервисов",
"Presentation", "Презентация",
"Presentation details.", "Детали презентации.",
"Press", "Пресса",
"Priority", "Приоритет",
"Priority (-10..10)", "Приоритет (-10..10)",
"Private", "Частный",
"Profile name", "Имя Профиля",
"Provider", "Провайдер",
"Provider ID", "ID Провайдера",
"Provider network name", "Имя сети провайдера",
"Psychology", "Психология",
"Puppets", "Куклы",
"Quiz", "Викторины",
"RAM only", "RAM Только",
"RTSP port", "RTSP Порт",
"RTSP port (554 or 9983, 0 = disable)", "порт RTSP (554 или 9983, 0 = откл)",
"Re-fetch period (hours)", "Период перезапроса (час.)",
"Re-fetch period (mins)", "Период перезапроса (мин.)",
"Re-record%s%s", "Время Записи%s%s",
"Re-try period (hours)", "Период повтора (час.)",
"Record all", "Записать всё",
"Recording file options", "Свойства записываемого файла",
"Recording system path", "Системный Путь Записи",
"Reject", "Извлечь",
"Religion", "Религия",
"Religious", "Религиозный",
"Remarkable people", "Замечательные люди",
"Remove HTTP arguments", "Удалить HTTP аргументы",
"Remove all unsafe characters from filename", "Удалить все не сохраняемые символы из имени файла",
"Replace whitespace in title with '-'", "Заменить Пробелы в Титлах '-'",
"Reports", "Репортажи",
"Rerun of", "Перезапуск из",
"Reserved", "Зарезервировано",
"Resolution (height)", "Разрешение (высота)",
"Respawn (pipe)", "Респаун (pipe)",
"Restart on error", "Рестарт при ошибке",
"Rewrite EIT", "Перезаписать EIT",
"Rewrite PAT", "Перезаписать PAT",
"Rewrite PMT", "Перезаписать PMT",
"Rewrite SDT", "Перезаписать SDT",
"Rock", "Рок",
"Rolloff", "Откатить",
"Romance", "Романтика",
"Rotor initialization time (seconds)", "Время Инициализации Ротора (сек.)",
"Rotor type", "Тип Ротора",
"Run as group", "Запустить от группы",
"Run as user", "Запустить от пользователя",
"Running", "Запуск",
"SAT>IP DVB-T frequency (Hz)", "Частота SAT>IP DVB-T (Гц)",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "SAT>IP RTSP номер порта сервера\n(по умолчанию: -1=откл., 0=webconfig, стандартный порт 554)",
"SD: standard definition", "SD: Стандартное Разрешение",
"SSL verify peer", "SSL проверка пира",
"Sat", "Спутник",
"Satellite longitude", "Долгота спутника",
"Satellite positions", "Позиция Спутника",
"Scan after creation", "Сканировать после создания",
"Scan queue length", "Сканирование длины очереди",
"Scan result", "Результат Сканирования",
"Scan status", "Статус Сканирования",
"Schedule a re-recording if more errors than (0=off)", "Журнал презаписи если ошибок больше чем (0=откл.)",
"Schedule status", "Статус расписания",
"Scheduled for recording", "Запланированные записи",
"Scheduled start time", "Запланированное время старта",
"Scheduled stop time", "Запланированное время остановки",
"School programs", "Школьные программы",
"Science", "Наука",
"Science fiction", "Научная фантастика",
"Season", "Сезон",
"Season %d", "Сезон %d",
"Serial number", "Серийный Номер",
"Series link", "Ссылка на серии",
"Serious", "Серьёзное",
"Serious music", "Серьёзная музыка",
"Server", "Сервер",
"Server UUID", "Сервер UUID",
"Server details.", "О сервере.",
"Service", "Сервис",
"Service ID", "ID Сервиса",
"Service configuration", "Конфигурация Сервиса",
"Service count %zi", "Кол-во Сервисов %zi",
"Service name", "Имя сервиса",
"Service type", "Тип Сервиса",
"Services", "Сервисы",
"Set the display name.", "Задать имя дисплея.",
"Set the priority of this configuration.", "Установить приоритет этой конфигурации.",
"Settings", "Настройки",
"Show", "Шоу",
"Show this page", "Показать страницу",
"Show version information", "Показать информацию о версии",
"Signal quality too poor", "Качество сигнала очень плохое",
"Signal scale (240 or 100)", "Шкала Сигнала (240 или 100)",
"Site latitude", "Широта сайта",
"Site longitude", "Долгота сайта",
"Skip commercials", "Пропуская Комерческие",
"Skip initial bytes", "Пропустить начальные Байты",
"Skip x number of transport packets.", "Пропускать х транспортных пакетов.",
"Slave entry", "Вспомогательная запись",
"Slot state", "Состояние Слота",
"Soap", "Мыло",
"Soccer", "Бильярд",
"Social", "Социальное",
"Social advisory", "социальный консультативный",
"Sort index", "Сортировка по индексу",
"Source", "Источник",
"Source deleted", "Источник удален",
"Source quality is bad", "Качество источника плохое",
"Source reconfigured", "Источник переконфигурирован",
"Specify User-Agent header for the http client", "Указать User-Agent заголовок для http клиента",
"Specify alternative htsp port", "Задайте альтернативный htsp порт",
"Specify alternative http port", "Задайте альтернативный http порт",
"Specify alternative http webroot", "Задайте альтернативный http webroot",
"Specify bind address", "Определить адрес привязки",
"Spiritual sciences", "Спиритические науки",
"Sports", "Спортивные",
"Sports magazines", "Спортивные журналы",
"Squash", "Сквош",
"Standard", "Стандарт",
"Start", "Старт",
"Start after", "Старт после",
"Start before", "Старт перед",
"Start time", "Время старта",
"Status", "Статус",
"Status period (ms)", "Периодичность статуса (мсек.)",
"Stereo", "Стерео",
"Stop", "Стоп",
"Stop time", "Время Остановки",
"Storage path", "Путь к носителю",
"Storage space", "Ёмкость накопителя",
"Stream index", "Индекс Потока",
"Stream profile", "Профиль Потока",
"Stream type", "Тип Потока",
"Subdirectory options", "Опции поддиректории",
"Subscribe to a service permanently", "Подписаться на сервис постоянно",
"Subscription overridden", "Подписка Переопределена",
"Subscription weight", "Вес Подписки",
"Subtitle", "Субтитры",
"Subtitle codec", "Codec Субтитров",
"Sun", "Воскр.",
"Switch before rotor", "Переключатель перед Ротором",
"Switch type", "Тип переключателя",
"Switch: %s", "Переключиться: %s",
"Symbol rate (Sym/s)", "Символьная Скорость (симв./сек.)",
"Sync", "Синхронизировать",
"System", "Система",
"Tags", "Метки",
"Talk show", "Разговорное Шоу",
"Team sports (excluding football)", "Командный спорт (исключая футбол)",
"Technology", "Технологии",
"Tennis", "Теннис",
"Testing", "Тестирование",
"Testing options", "Опции тестирования",
"The character encoding for this service (e.g. UTF-8).", "Кодировка сервиса. Например (UTF-8).",
"The current boot ID.", "Текущий boot ID.",
"The current configuration ID.", "ID текущей конфигурации",
"The current configuration version.", "Версия текущей конфигурации.",
"The default subscription weight for each subscription.", "Вес для каждой подписки по умолчанию.",
"The device ID.", "ID устройства.",
"The device's serial number.", "Серийный номер устройства.",
"The frequency of the mux (in Hertz).", "Частота мультиплксора в (Hertz).",
"The manufacturer of the SAT>IP server.", "Производитель SAT>IP сервера.",
"The networks using this configuration.", "Сети использующие эту конфигурацию.",
"The number of ATSC-C (Cable/AnnexB) tuners to export.", "Номера ATSC-C (Cable/AnnexB) тюнеров для экспорта.",
"The number of ATSC-T (Terresterial) tuners to export.", "Номера ATSC-T (Terresterial) тюнеров для экспорта.",
"The number of DVB-C (Cable) tuners to export.", "Номера DVB-C (Cable) тюнеров для экспорта.",
"The number of DVB-C2 (Cable) tuners to export.", "Номера DVB-C2 (Cable) тюнеров для экспорта.",
"The number of DVB-S (Satellite) tuners to export.", "Номера DVB-S (Satellite) тюнеров для экспорта.",
"The number of DVB-S2 (Satellite) tuners to export.", "Номера DVB-S2 (Satellite) тюнеров для экспорта.",
"The number of DVB-T (Terresterial) tuners to export.", "Номера DVB-T (Terresterial) тюнеров для экспорта.",
"The number of DVB-T2 (Terresterial) tuners to export.", "Номера DVB-T2 (Terresterial) тюнеров для экспорта.",
"The provider's name.", "Имя провайдера.",
"The provider's network ID.", "ID сети провайдера.",
"The provider's network name.", "Имя сети провайдера.",
"The service ID as set by the provider.", "ID сервиса установлен провайдером.",
"The service name as set by the provider.", "Имена сервисов установленные провайдером.",
"The service's channel number as set by the provider.", "Номера каналов установленные провайдером.",
"The symbol rate.", "Символьная скорость.",
"Theme", "Тема",
"This tuner", "Этот Тюнер",
"Thriller", "Триллер",
"Thu", "Четв.",
"Time", "Время",
"Time missed", "Время не указано",
"Time record caption", "Метка времени записи",
"Time recording%s%s", "Время Записи%s%s",
"Time update", "Время Обновления",
"Timeout (sec) (0=infinite)", "ТаймАут (сек.) (0=бесконечн.)",
"Timeout (seconds)", "Таймаут (сек.)",
"Timeshift", "Тайм шифт",
"Title", "Название",
"Title (regexp)", "Титлы (regexp)",
"Tourism / Travel", "Туризм / Путешествия",
"Traditional arts", "Традиционные искуства",
"Traditional music", "Традиционная музыка",
"Transmission mode", "Режим Передачи",
"Transponder ID", "ID Траспондера",
"Transport stream ID", "Транспортный поток ID",
"Tue", "Вт.",
"Tune before DiseqC", "Настроить перед DiseqC",
"Tuner bind IP address", "Привязать тюнер к IP адресу",
"Tuner configuration", "Конфигурация тюнера",
"Tuning failed", "Настройка провалилась",
"Turn off LNB when idle", "Выкл. LNB в простое",
"Tvheadend server name", "Название Tvheadend сервера",
"Type", "Тип",
"UDP RTP port number (2 ports)", "UDP RTP Номер порта (2 порта)",
"Unimportant", "Неважный",
"Universal LNB only", "Только Универсальные LNB",
"Unknown", "Неизвестно",
"Unknown reason (%i)", "Причина неизвестна (%i)",
"Unlimited size", "Не ограниченный Размер",
"Unlimited time", "Не ограниченное Время",
"Unset (default)", "Отменить (по умолчанию)",
"Update channel icon", "Обновить иконки канала",
"Update channel name", "Обновить имя канала",
"Update channel number", "Обновить номер канала",
"Update disabled", "Обновление отключено",
"Update time", "Обновить время",
"Updated", "Обновлённый",
"Use Windows-compatible filenames", "Использовать Windows-совместимые имена файлов",
"Use XSPF playlist instead of M3U", "Использовать XSPF плейлист вместо M3U",
"Use a pre-defined list of ATSC-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Использовать предопределенный список ATSC-C мультиплексоров. Внимание: эти списки иногда могут быть устаревшими и могут вызвать сканирование которое занимает больше времени, чем обычно.",
"Use a pre-defined list of ATSC-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Использовать предопределенный список ATSC-T мультиплексоров. Внимание: эти списки иногда могут быть устаревшими и могут вызвать сканирование которое занимает больше времени, чем обычно.",
"Use a pre-defined list of DVB-C muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Использовать предопределенный список DVB-C мультиплексоров. Внимание: эти списки иногда могут быть устаревшими и могут вызвать сканирование которое занимает больше времени, чем обычно.",
"Use a pre-defined list of DVB-S/S2 muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Использовать предопределенный список DVB-S/S2 мультиплексоров. Внимание: эти списки иногда могут быть устаревшими и могут вызвать сканирование которое занимает больше времени, чем обычно.",
"Use a pre-defined list of DVB-T muxes. Note: these lists can sometimes be outdated and may cause scanning to take longer than usual.", "Использовать предопределенный список DVB-T мультиплексоров. Внимание: эти списки иногда могут быть устаревшими и могут вызвать сканирование которое занимает больше времени, чем обычно.",
"Use original", "Используйте оригинал",
"Use service IDs as channel numbers", "Использовать ID сервиса как номера каналов",
"User access error", "Ошибка доступа пользователя",
"User icon", "Пользовательские иконки",
"User interface level", "Уровень пользовательского интерфейса",
"User language", "Пользовательский язык",
"User login", "Логин пользователя",
"Username", "Имя пользователя",
"Variety show", "Варьете",
"Video", "Видео",
"Video bitrate (kb/s) (0=auto)", "Видео Битрейт (кб/сек.) (0=Авто)",
"Video codec", "Видео кодек",
"Video recorder", "Видео рекордер",
"Visually impaired commentary/audio description", "Комментарий/описание аудио",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV профиль H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV профиль H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV профиль VP8/Vorbis/WEBM",
"Waiting for program start", "В ожидании запуска программы",
"Waiting for stream", "В ожидании потока",
"War", "Война",
"Water sport", "Водный спорт",
"Weak stream", "Слабый поток",
"Weather report", "Прогноз погоды",
"Web interface", "Web Интерфейс",
"Web interface language", "Язык WEB интерфейса",
"Web user interface", "Web Пользовательский интерфейс",
"Wed", "Ср.",
"Western", "Вестерн",
"Wizard", "Мастер",
"Wrapper (capmt_ca.so)", "Враппер (capmt_ca.so)",
"Yes", "Да",
"addpids/delpids supported", "addpids/delpids поддерживается",
"hrs", "час.",
"invalid option specified [%s]", "Указанный параметр недействителен [%s]",
"min", "мин.",
"mins", "минут",
"option %s requires a value", "%s необходимо значение",
"tvhdhomerun client", "tvhdhomerun Клиент",
NULL, NULL
};

static const char *tvh_locale_29[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nZa več informacij obiščite spletno stran Tvheadend:\nhttps: //tvheadend.org\n",
"# Mapped channels", "Preslikani kanali",
"# Muxes", "Multipleksi",
"# Services", "Storitve",
"# tune repeats", "Nastavitve ponavljanj",
"(Default profile)", "(Privzeti profil)",
"(Read-only)", "(Samo za branje)",
"1 day", "1 dan",
"1 month", "1 mesec",
"1 week", "1 teden",
"1 year", "1 leto",
"2 months", "2 meseca",
"2 weeks", "2 tedna",
"2 years", "2 leti",
"2-Port switch (universal LNB)", "stikalo 2-Port (univerzalni LNB)",
"3 days", "3 dni",
"3 months", "3 meseci",
"3 weeks", "3 tedni",
"3 years", "3 leta",
"4-Port switch (universal LNB)", "stikalo 4-Port (univerzalni LNB)",
"5 days", "5 dni",
"6 months", "6 mesecev",
"A channel tag (e.g. a group of channels) to which this rule applies.", "Oznaka Kanala (ali skupino kanalov), za katerega velja to pravilo.",
"AAC audio", "AAC zvok",
"AC-3 = descriptor 6", "AC-3 = deskriptor 6",
"AC-3 detection", "Zaznavanje AC-3",
"AC3 audio", "AC3 zvok",
"ACTIVE", "AKTIVNO",
"ANY", "KARKOLI",
"ATSC SI Tables", "ATSC SI Tabela",
"ATSC source ID", "ATSC vir ID",
"ATSC-C Network", "ATSC-C omrežje",
"ATSC-C multiplex", "ATSC-C multipleks",
"ATSC-T Network", "ATSC-T Omrežje",
"ATSC-T multiplex", "ATSC-T multipleks",
"Aborted by user", "Prekinil uporabnik",
"Accept zero value for TSID", "Sprejmi ničelno vrednost za TSID",
"Access", "Dostop",
"Access (ACL)", "Dostop (ACL)",
"Action", "Akcija",
"Active", "Aktivno",
"Adapter in use by another subscription", "Adapter uporablja druge naročnine",
"Add file and line numbers to debug", "Dodajanje številk v datoteke in vrstice za razhroščevanje",
"Add the thread ID to debug", "Dodaj ID nit za razhroščevanje",
"Additional service names found in EPG data.", "Dodatna imena storitev najdena v podatkih EPG.",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "Dodaten čas (v sekundah), v kateri bo sprejemnik pripravljen za snemanje. To je uporabno za tiste tunerje, ki rabijo nekaj časa, da se pripravijo in / ali pošljejo podatke pred začetkom.",
"Admin password", "Admin geslo",
"Admin username", "Admin uporabniško ime",
"Administrator login", "Administrator prijava",
"Adult movie", "Film za odrasle",
"Advanced", "Napredno",
"Advanced (non-universal LNBs, rotors, etc.)", "Napredno (non-universal LNBs, rotors, etc.)",
"Adventure", "Avanture",
"Advertisement / Shopping", "Oglaševanje / Nakup",
"After creating the network scan it for services.", "Ko ste ustvarili omrežje, skeniraj storitve.",
"All (Streaming plus DVR)", "Vse (Streaming plus DVR)",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "Vsi znaki, ki bi lahko povzročili težave v imenu bodo nadomeščeni z podčrtajem. Glej pomoč za podrobnosti.",
"All lower-case", "Vse z malo pisavo",
"All proceeds are used to support server infrastructure and buy test equipment.", "Vsi prihodki se uporabljajo za podporo strežniške infrastrukture in nakup preskusne opreme.",
"All: Record if different description", "Vse: Snemaj, če je drugačen opis",
"All: Record if different episode number", "Vse: Snemaj če je drugačna številka epizode",
"All: Record if different subtitle", "Vse: Snemaj če se razlikujejo podnapisi",
"All: Record once per day", "Vse: Snemaj enkrat na dan",
"All: Record once per week", "Vse: Snemaj enkrat na teden",
"Allow high bitrate mode (CI+ CAMs only).", "Dovoli visoki način bitrate (samo CI + CAM).",
"Allow idle scan tuning on this device.", "Dovoli iskanje postaj na tej napravi, ko je nedejavna.",
"Allow the combined size of all timeshift buffers to potentially grow unbounded until your storage media runs out of space.", "Dovoli kombinirano velikost buffer-ja, dokler mediju za shranjevanje ne zmanjka prostora.",
"Allow the timeshift buffer to grow unbounded until your storage media runs out of space. Warning, enabling this option may cause your system to slow down or crash completely!", "Pusti, da se buffer nemoteno polni, dokler mediju za shranjevanje ne zmanjka prostora. Opozorilo, ta možnost lahko upočasni ali povsem zruši vaš sistem !",
"Allow/Disallow access to the 'Configuration' tab.", "Dovoli / Ne dovoli dostop do zavihka »konfiguracija«.",
"Allow/Disallow web interface access (this  includes access to the EPG).", "Dovoli / Ne dovoli dostop do spletnega vmesnika (to vključuje dostop do EPG).",
"Allowed DVR profiles. This limits the profiles the user has access to.", "Dovoljeni DVR profili. To omejuje dostop do profilov uporabnika.",
"Allowed network", "Dovoljeno omrežje",
"Allowed networks", "Dovoljena omrežja",
"Alternate PID path", "PID pot po izbiri",
"Alternate configuration path", "Nadomestna konfiguracija pot",
"Altitude (in meters).", "Nadmorska višina (v metrih)",
"Altitude (meters)", "Nadmorska višina (metri)",
"Always send the whole DiseqC sequence including LNB setup (voltage, tone). If this is not checked, only changed settings are sent, which may cause issues with some drivers. If the tuning is not reliable, try activating this option.", "Vedno pošli celotno DiSEqC zaporedje vključno LNB nastavitev (napetost, tona). Če se to ne preverja, se pošlje le spremebe nastavitev, ki lahko povzročijo težave z nekaterimi gonilniki. Če nastavitev ni zanesljiva, poskusite aktivirate to možnost.",
"Always use the name defined by the network.", "Vedno uporabi ime definirano z omrežjem.",
"Always use the old ioctls to read the linuxdvb status (signal strength, SNR, error counters). Some drivers are not mature enough to provide the correct values using the new v5 linuxdvb API.", "Vedno uporabljaj stari ioctls prebere status linuxdvb (števci moč signala, SNR, napake). Nekateri gonilniki niso dovolj zreli za zagotavljanje pravilne vrednosti z uporabo novega v5 linuxdvb API.",
"An event which starts between this \"start after\" and \"start before\" will be matched (including boundary values).", "Dogodek, ki se začne med \"začetek po\" in \"start pred\" bo ustrezal (vključno z mejnimi vrednostmi).",
"Animals", "Živali",
"Anonymize HTSP access", "Anonimni HTSP dostop",
"Any", "Katerikoli",
"Arts", "Umetnost",
"Arts / Culture (without music)", "Umetnost / Kultura (brez glasbe)",
"Arts magazines", "Umetniške revije",
"Assign predefined muxes to networks", "Določi vnaprej določene mukse na omrežjih",
"Associate this device with one or more networks.", "Poveži te naprave z enim ali več omrežji.",
"Athletics", "Atletika",
"Audio channel layout.", "Postavitev avdio kanala.",
"Audio codec", "Audio Kodek",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Avdio kodek uporabljen za transkodiranje. \"Ne uporabljajte\" onemogoči zvok.",
"Audio stream", "Avdio tok",
"Audio type", "Audio Tip",
"Audio-only stream", "Audio-samo stream",
"Audioes muxer", "Audio mikser",
"Authenticated user", "Preverjen uporabnik",
"Auto check disabled", "Samodejna prijava onemogočena",
"Auto check enabled", "Samodejna prijava omogočena",
"Auto record", "Auto snemanje",
"Auto record caption", "Auto zapis - napis",
"Auto recording%s%s", "Samodejno snemanje%s%s",
"Auto time record", "Auto čas snemanja",
"Auto-Map to channels", "Auto-Map kanalov",
"Automatic checking", "Samodejno preverjanje",
"Automatic recording caption.", "Samodejno snemanje napisa.",
"Automatic schedule entry not found", "Avtomatski vnos urnika ni bilo mogoče najti",
"Automatic time scheduler entry not found", "Samodejni čas vnosa urnika ni bilo mogoče najti",
"Automatically link EPG data to the channel (using the channel name for matching). If you turn this option off, only the OTA EPG grabber will be used for this channel unless you've specifically set a different EPG Source.", "Samodejno poveže podatke EPG na kanal (z uporabo imena kanala za usklajevanje). Če to možnost izključite, bo samo OTA EPG graberji ki se uporabljajo za ta kanal, razen če jih izrecno določi drugačen EPG Vir.",
"Automatically map EPG source", "Samodejni map vir EPG",
"Automatically map all available services to channels.", "Samodejni map vseh razpoložljivih storitev kanalov.",
"Automatically map channels defined within the bouquet.", "Samodejno map načinov, določenih v šopku.",
"Automatically name from network", "Samodejno ime iz omrežja",
"Automatically record.", "Samodejno snemanje.",
"Autorec maximum count (0=unlimited)", "Autorec Največje število (0 = neomejeno)",
"Autorec maximum schedules limit (0=unlimited)", "Autorec zgornja meja urnika (0 = neomejeno)",
"Bad", "Slabo",
"Bad request", "Slaba prošnja",
"Ballet", "Balet",
"Bandwidth", "pasovna širina",
"Base DVB SI Tables (PAT,CAT,PMT,SDT etc.)", "Osnovne DVB SI tabele (PAT, CAT, PMT, SDT itd)",
"Based on software from", "Na podlagi programske opreme iz",
"Basic", "Osnovno",
"Bind to specific local IP address.", "Vezano na določene lokalne IP naslove.",
"Blue", "Modro",
"Bouquet", "Šopek",
"Bouquet (auto)", "Šopek (samodejno)",
"Bouquet source.", "Vir šopka",
"Bouquets", "Šopki",
"Brand", "Znamka",
"Branding information (if available).", "Informacije Znamke (če je na voljo)",
"Broadcast", "Oddaja",
"Broadcast type", "Tip Oddaje",
"Broadcast.", "Oddaja",
"Broadcasting", "Oddajanje",
"Buffering limit (ms)", "Omejitev buffering (ms)",
"Build", "Verzija",
"By default, linuxdvb's status read period is 1000ms (one second). The accepted range is 250ms to 8000ms. Note that for some hardware or drivers (like USB), the status operations take too much time and CPU. In this case, increase the default value. For fast hardware, this value might be decreased to make the decision of the re-tune algorithm based on the signal status faster.", "Privzeto je obdobje linuxdvb 1000ms (ena sekunda). Sprejeto območje je 250ms do 8000ms. Upoštevajte, da za nekatere strojne opreme ali gonilnike (kot USB), operacije stanje traja preveč časa in CPU. V tem primeru se poveča privzeto vrednost. Za hitro strojno opremo, lahko to vrednost znižate, da bi pohitrili algoritem.",
"CA (descrambling) Client", "CA (dekodirni) Client",
"CA identification", "identifikacija CA",
"CA provider", "ponudnik CA",
"CAPMT (Linux Network DVBAPI)", "CAPMT (Linux omrežje DVBAPI)",
"CAPMT interval (in ms).", "CAPMT interval (v ms).",
"CAPMT query interval (ms)", "CAPMT izraz intervala (ms)",
"CAPMT query interval (ms).", "CAPMT izraz intervala (ms).",
"CRASH", "ZRUŠITEV",
"CRID authority", "CRID organ",
"CSA (descrambling)", "CSA (dekodirni)",
"Cache scheme", "shema predpomnilnika",
"Camd.socket filename / IP Address (TCP mode)", "Camd.socket naziv / IP naslov (način TCP)",
"Cartoons", "Risanke",
"Change parameters", "Spremeni parametre",
"Channel", "Kanal",
"Channel Tags", "Oznaka Kanala",
"Channel does not exist", "Kanal ne obstaja",
"Channel icon", "Ikona Kanala",
"Channel icon URL.", "URL ikone Kanala",
"Channel icon as defined in EPG data.", "Ikona Kanala kot je definirano v EPG podatkih",
"Channel icon name scheme", "Kanal shema ime ikon",
"Channel icon path", "Pot do ikone Kanala",
"Channel mapping options", "Možnosti kanala za preslikavo",
"Channel name", "Ime Kanala",
"Channel not found", "Kanal ni najden",
"Channel number", "Številka Kanala",
"Channel number as defined in EPG data.", "Številka kanala, kot je opredeljeno v podatkih EPG.",
"Channel number offset", "Številka kanala zamik",
"Channel number range", "Kanal številka razpona ",
"Channel numbers (heuristic)", "Številke kanalov (hevristično)",
"Channel numbers from", "Številke Kanalov iz",
"Channel tag", "Oznaka Kanala",
"Channel tag does not exist", "Oznaka Kanala ne obstaja",
"Channel tag reference", "Referenčna oznaka kanala",
"Channel tags", "Oznake Kanala",
"Channel tags the user is allowed access to/excluded from.", "Oznake Kanala uporabnik omogočen dostop / izvzeto.",
"Channel to use/used for the recording.", "Kanal uporabi / uporabljen za snemanje.",
"Channel update options", "Možnosti posodobitve kanala",
"Channels", "Kanali",
"Channels EPG data is used by.", "Uporablja podatke kanalov EPG",
"Character set", "nabor znakov",
"Characters not supported in Windows filenames (e.g. for an SMB/CIFS share) will be stripped out or converted.", "Znaki niso podprti v imenih datotek Windows (na primer za SMB / CIFS delnico) bodo odstranjeni ali zamenjani.",
"Charset", "nabor",
"Check availability", "Preveri razpoložljivost",
"Check for the services' presence. If the service is no longer broadcast this field will change to Missing In PAT/SDT.", "Preverite prisotnost služb. Če storitev več ne oddaja, na tem področju spremeniti, da manjka v PAT / SDT.",
"Children's / Youth Programs", "Otroci / mladostniki",
"Children's / Youth programs", "Otroški / mladostnik programi",
"Cinema", "Kino",
"Class", "razred",
"Classical", "Klasika",
"Classical music", "Klasična glasba",
"Clean effects", "čisti učinki",
"Client name", "ime odjemalca",
"Comedy", "Komedija",
"Command delay time (ms) (10-200)", "Ukaz časovni zamik (ms) (10-200)",
"Command time (ms) (10-100)", "Čas ukaz (ms) (10-100)",
"Comment", "Komentar",
"Commercial break", "Reklamni premor",
"Commercials will be dropped from the recordings. Commercial detection works using EITp/f (EPG running state) and for the Swedish channel TV4 (using teletext info).", "Reklame bodo izrezane iz posnetkov. Reklame odkriva z uporabo EITP / f (EPG teče državno) in za švedsko kanala TV4 (z uporabo informacij teleteksta).",
"Committed", "zavezano",
"Completed OK", "Končan OK",
"Compress EPG database", "Stisni EPG baze podatkov",
"Compress the EPG database to reduce disk I/O and space.", "Stisni bazo podatkov EPG za zmanjšanje disk operacij in prostora.",
"Concurrent limit per network position (src=) for satellite SAT>IP tuners. The first limit number is for src=1 (AA), second for src=2 (AB) etc.", "Sočasno omejitev na položaju omrežja (src =) za satelitsko SAT> IP-sprejemnik. Prva mejna številka za src = 1 (AA), drugi za src = 2 (AB), itd",
"Conditional Access", "pogojni dostop",
"Conditional Access Client", "Pogojni dostop odjemalca",
"Conditional Access Identification.", "Pogojni dostop Identifikacija.",
"Configuration", "konfiguracija",
"Configuration - Base", "Konfiguracija - Osnovna",
"Configuration - Image Cache", "Konfiguracija - Slika Cache",
"Configuration - SAT>IP Server", "Konfiguracija - SAT> IP Server",
"Configuration ID", "Konfiguracija ID",
"Configuration name", "ime konfiguracije",
"Configuration version", "različica konfiguracije",
"Connection limit type", "Omejitev Vrste povezave",
"Connection limits", "omejitve povezave",
"Constellation", "ozvezdje",
"Container to use for the transcoded stream.", "Conteiner za uporabo za prekodiran tok.",
"Content character set", "Vsebina nabor znakov",
"Content reference identifier authority.", "Vsebina organa - referenčna oznaka.",
"Content type", "Vrsta vsebine",
"Content type.", "Tip vsebine.",
"Contest", "tekmovanje",
"Continue recording for x minutes after scheduled stop time", "Snemanje še za x minut po načrtovanem času zaustavitve",
"Continue recording for x minutes after scheduled stop time.", "Snemanje še za x minut po načrtovanem času zaustavitve.",
"Cookie expiration (days)", "Pišketek preneha po (dni)",
"Cooking", "Kuhanje",
"Copy codec type", "Kopiraj tip kodeka",
"Copy layout", "Kopiraj postavitev",
"Count of objects", "Število predmetov",
"Create SDTV/HDTV/Radio tags.", "Ustvari SDTV / HDTV / Radio oznake.",
"Create a bouquet from the playlist.", "Ustvari šopek iz seznama.",
"Create a directory per channel when storing recordings. If both this and the 'directory per day' checkbox is enabled, the date-directory will be the parent of the per-channel directory.", "Ustvari mapo za kanal pri shranjevanju posnetkov. Če je to omogočeno in \"mapa na dan\", bo datum mape pred mapo kanala.",
"Create a directory per title when storing recordings. If the day/channel directory checkboxes are also enabled, those directories will be parents of this directory.", "Ustvarite mapo po naslovu pri shranjevanju posnetkov. Če je omogočena tudi mapa okenca na dan/kanal, bodo te mapo v teh.",
"Create a new directory per day in the recording system path. Folders will only be created when something is recorded. The format of the directory will be ISO standard YYYY-MM-DD.", "Ustvari novo mapo na dan za snemalno sistemsko pot. Mape bodo nastala šele, ko se nekaj zabeleži. Oblika imenik bo standard ISO LLLL-MM-DD.",
"Create a provider name tag.", "Ustvarite oznako za ime ponudnika .",
"Create and associate a network tag to created channels.", "Ustvarjanje in povezava omrežne oznake ustvarjenih kanalov.",
"Create and associate a provider tag to created channels.", "Ustvarjanje in povezava oznak ponudnika za ustvarjenei kanale.",
"Create and link these tags to channels when mapping.", "Ustvarjanje in povezovanje teh oznak na poti pri kartiranju.",
"Create bouquet", "Ustvari šopek",
"Create bouquet tag", "Ustvari šopek z oznako",
"Create directories using these permissions.", "Ustvarjanje mape z uporabo teh dovoljenj.",
"Create files using these permissions.", "Ustvarjanje datotek z uporabo teh dovoljenj.",
"Create network name tags", "Ustvari imenske oznake omrežja",
"Create network name tags (set by provider).", "Ustvari imenske oznake omrežja (določi ponudnik).",
"Create network tags", "Ustvari oznake omrežja",
"Create provider name tags", "Ustvari oznake imen ponudnika",
"Create provider tags", "Ustvari oznake ponudnika",
"Create tags", "ustvarjanje oznake",
"Create tags in recordings using media containers that support metadata (if possible).", "Ustvarjanje oznake na posnetkih s pomočjo medijev, kontejnerjev, ki podpirajo metapodatke (če je možno).",
"Create type-based tags", "Ustvarjanje oznake, ki temelji na tipu",
"Created", "Ustvarjeno",
"Creator", "Kreator",
"Cron", "Cron delovna mesta",
"Culture (without music)", "Kultura (brez glasbe)",
"Culture magazines", "revije Kulture",
"Current RTSP port.", "Trenutna vrata RTSP.",
"Current affairs", "Trenutne zadeve",
"Current number of objects.", "Sedanje število predmetov.",
"Current object size.", "Trenutna velikost objekta.",
"Current tuner configuration.", "Sedanja konfiguracija sprejemnika.",
"Custom HTTP headers", "Poljubne HTTP glave",
"DAB Network", "DAB omrežja",
"DAB multiplex", "DAB multipleks",
"DES Key.", "DES ključ.",
"DES key", "DES ključ",
"DSCP/TOS for streaming", "DSCP / TOS za pretok",
"DVB EPG Tables", "DVB EPG Tabele",
"DVB Inputs - Networks", "DVB Vhodi - Omrežja",
"DVB SI Tables", "DVB SI Tabele",
"DVB Time Tables", "DVB Time Tabele",
"DVB multiplex", "DVB multipleks",
"DVB scan files", "DVB datoteke skeniranja",
"DVB scan files path", "DVB skeniranje poti do datotek",
"DVB-C Network", "DVB-C Omrežje",
"DVB-C multiplex", "DVB-C multipleks",
"DVB-S Network", "DVB-S Omrežje",
"DVB-S Satellite Configuration", "DVB-S Satelit Konfiguracija",
"DVB-S multiplex", "DVB-S multipleks",
"DVB-T Network", "DVB-T Omrežje",
"DVB-T multiplex", "DVB-T multipleks",
"DVR - Auto-recording (Autorecs)", "DVR - Auto-snemanje (Autorecs)",
"DVR - Profiles", "DVR - Profil",
"DVR - Time-based Recording (Timers)", "DVR - Čas, ki temelji na snemanju (merilniki časa)",
"DVR behavior", "vedenje DVR",
"DVR configuration", "konfiguracija DVR",
"DVR configuration profiles", "konfiguracija profilov DVR",
"DVR configurations", "konfiguracije DVR",
"DVR entry not found", "vnos DVR ni bilo mogoče najti",
"DVR file retention period", "DVR obdobje hranjenja datotek",
"DVR log retention", "DVR ohranjanje dnevnika",
"DVR log retention period", "DVR obdobje hranjenja dnevnika",
"DVR override: high", "DVR preglasi: visoko",
"DVR override: important", "DVR preglasi: pomembno",
"DVR override: low", "DVR preglasi: nizko",
"DVR override: normal", "DVR preglasi: normalno",
"DVR override: unimportant", "DVR preglasi: nepomembno",
"DVR profile to use/used for the recording.", "DVR profil uporabiti / uporabljen za snemanje.",
"DVR schedule does not exist", "DVR urnik ne obstaja",
"DVR schedule does not have a file yet", "DVR razpored še nima datoteke",
"DVR schedule not found", "DVR urnika ni bilo mogoče najti",
"Dance", "Ples",
"Data errors", "napake podatkov",
"Data path (if applicable).", "Pot podatkov (če pride v poštev).",
"Date the EPG data was last updated (not set for OTA grabbers).", "Datum podatkov EPG je bil nazadnje posodobljen (ni nastavljen za OTA karticah).",
"Days of Week", "Dnevi tedna",
"Days of the week to which the rule should apply.", "Dnevi v tednu kdaj bi se moralo pravilo uporabljati.",
"Days of week", "Dnevi v tednu",
"Debate", "razprava",
"Debug libav log", "Debug libav dnevnik",
"Debug options", "možnosti za odpravljanje napak",
"Debug subsystems", "odpravljanje napak podsistemi",
"Debug to syslog", "Odpravljanje napak v syslog",
"Debug trace (low-level)", "Odpravljanje napak v sledovih (nizka raven)",
"Debugging", "Odpravljanje napak",
"Default", "privzeto",
"Default language(s)", "privzeti Jezik(i)",
"Default language.", "privzeti Jezik.",
"Default login", "Privzeta prijava",
"Default priority", "privzeta prednost",
"Default user interface level.", "Privzeta raven uporabniškega vmesnika.",
"Define network group to limit network usage.", "Določite omrežno skupino za omejitev uporabe omrežja.",
"Delivery system", "sistem dostave ",
"Demux path", "Demux pot",
"Descramble services (limit per mux)", "Vdiranje v storitve (omejitev na MUX)",
"Description", "Opis",
"Detective", "Detektiv",
"Device ID", "ID naprave",
"Device model", "model naprave",
"Device path", "pot naprave",
"Differentiated Services Code Point / Type of Service: Set the service class Tvheadend sends with each packet. Depending on the option selected this tells your router the prority in which to give packets sent from Tvheadend, this option does not usually need changing. See https://en.wikipedia.org/wiki/Differentiated_services for more information. ", "Diferencirana Services Code Point/Vrsta storitve: Nastavite razred storitve, ki jo Tvheadend pošilja z vsakega paketa. Glede na izbrano možnost, ta pove vašemu usmerjevalniku za PREDNOSTNE pakete poslane iz Tvheadend, ta možnost se običajno ne spreminja. Glej https://en.wikipedia.org/wiki/Differentiated_services za več informacij.",
"Digest access authentication is intended as a security trade-off. It is intended to replace unencrypted HTTP basic access authentication. This option should be enabled for standard usage.", "Digest dostop pristnosti je namenjena kot varnostno kompromisu. To naj bi nadomestil nešifrirano HTTP avtentikacijo za osnovni dostop. Ta možnost bi morala biti omogočena za splošno rabo.",
"Directory", "Imenik",
"Directory permissions (octal, e.g. 0775)", "Imenik dovoljenja (octal, npr 0775)",
"Directory used by the entry.", "Imenik uporabljen z vpisom .",
"Disable", "Onemogoči",
"Disable DVB bouquets", "Onemogoči DVB šopke",
"Disable SAT>IP client", "Onemogoči SAT> IP odjemalca",
"Disable all access control checks", "Onemogoči vse preglede za nadzor dostopa",
"Disable debug on stderr", "Onemogoči razhroščevanje na stderr",
"Disable syslog (all messages)", "Onemogoči syslog (vsa sporočila)",
"Disabled", "Onemogočeno",
"Discover more muxes using the Network Information Table (if available).", "Odkrijte več \"muxes\" z uporabo informacijske tabele omrežja (če je na voljo).",
"Discussion", "Diskusija",
"DiseqC repeats", "DiseqC ponavlja",
"Do not send any stream specific information to the HTSP client like signal strength, input source etc.", "Ne pošiljajte toka posebnih informacij HTSP za stranko , kot so moč signala, vhodni vir itd",
"Do not use", "Ne uporabljajte",
"Documentary", "dokumenti",
"Don't abort streaming when an encrypted stream can't be decrypted by a CA client that normally should be able to decrypt the stream.", "Ne prekini pretakanja, ko šifriran tok ni mogoče dešifrirati s clinetom CA, ki je ponavadi  sposoben dešifrirati tok.",
"Don't include the title in the filename.", "Ne vsebuje naslova v imenu datoteke.",
"Don't include title in filename", "Ne vsebuje naslova v imenu datoteke",
"Don't keep", "Ne obdrži",
"Don't re-record", "Ponovno ne snemaj",
"Don't re-record if recording fails.", "Ponovno ne snemaj, če snemanje ne uspe.",
"Don't re-schedule if recording fails.", "Ne dodaj ponovno na urnik, če snemanje ne uspe.",
"Don't reschedule", "Ne prestavi na urniku",
"Don't use the provider's channel numbers.", "Ne uporabljajte številk kanalov ponudnika.",
"Duplicate handling", "ravnanje dvojnika",
"Duplicate recording handling.", "Ravnanje s podvojenim snemanjem.",
"EIT - skip TSID check", "EIT - preskoči ID pošiljatelja",
"EIT time offset", "Čas ETI zamika",
"EPG Grabber Channel", "EPG Grabber kanal",
"EPG Grabber Configuration", "EPG Grabber Konfiguracija",
"EPG Language (priority order)", "EPG Jezik (prednostni red)",
"EPG data ID.", "EPG podatki ID.",
"EPG name", "EPG ime",
"EPG overlap cut", "EPG prekrivanje rez",
"EPG scan", "EPG skeniranje",
"EPG source", "EPG Vir",
"EPG update window", "EPG okno nadgradnje",
"Economics", "ekonomija",
"Education", "Izobrazba",
"Education / Science / Factual topics", "Izobraževanje / znanost / resničnostne teme",
"Educational", "izobraževalno",
"Electronic Program Guide", "Elektronski programski vodič",
"Elementary Stream Filter", "Osnovni Stream Filter",
"Elementary stream filter", "Osnovna Stream Filter",
"Enable", "Omogoči",
"Enable (auto)", "Omogočite (auto)",
"Enable DBus", "Omogoči DBus",
"Enable NTP driver", "Omogoči NTP gonilnik",
"Enable debug on stderr", "Omogočite razhroščevanje na stderr",
"Enable debug subsystems", "Omogoči Debuger za odpravljanje napak podsistema",
"Enable debug to file", "Omogočite razhroščevanje v datoteko",
"Enable debug to syslog", "Omogočite razhroščevanje v dnevnik",
"Enable exclusion of user-config defined channel tags. This will prevent the user from accessing channels associated with the tags selected (below).", "Omogočajo izključitev uporabniških nastavitvenih opredeljenih oznak kanalov. To bo preprečilo, da bi uporabnik dostopal do kanalov, povezanih z izbranimi oznakami (spodaj).",
"Enable if the SAT>IP box requests plts=on parameter in the SETUP RTSP command for DVB-S2 muxes.", "Omogočite, če IP so polja > zahteva plts = na parameter v nastavitvah RTSP ukaz za DVB-S2 muxes.",
"Enable if the SAT>IP box requires pids=21 parameter in the SETUP RTSP command", "Omogočite, če> IP polje je potrebno PID = 21 parametrov v nastavitvah RTSP ukaz",
"Enable if the SAT>IP box supports the addpids/delpids commands.", "Omogočite, če IP polje > podpira addpids / delpids ukaze.",
"Enable or disable RTP/AVP/TCP transfer mode (embedded data in the RTSP session) support.", "Omogočanje ali onemogočanje RTP / AVP načina / TCP prenosa (vgrajenih podatkov v seji RTSP) podporo.",
"Enable or disable fast input switching.", "Omogoči ali onemogoči hitri vnos preklapljanja.",
"Enable or disable full mux mode.", "Omogoči ali onemogoči celoten način MUX.",
"Enable or disable ignoring of Event Information Table (EIT) data on this mux.", "Omogoči ali onemogoči ignoriranje dogodkov v informacijski tabeli (EIT), podatke o tem MUXu.",
"Enable or disable this configuration.", "Omogočite ali onemogočite to konfiguracijo.",
"Enable over-the-air program guide (EPG) scanning on this input device.", "Omogoči over-the-air vodič programov (EPG) skeniranje na tej vhodni napravi.",
"Enable syslog", "Omogoči syslog",
"Enable system time updates. This will only work if the user running Tvheadend has rights to update the system clock (normally only root).", "Omogočite posodobitev sistemskega časa. To bo delovalo le, če ima uporabnik Tvheadend pravice za posodobitev sistemske ure (običajno samo Admin).",
"Enable the CAs (conditional accesses) tab in web user interface for the advanced level. By default, this tab is visible only in the expert level.", "Omogoči CAs (pogojni dostopi) zavihek v spletni uporabniški vmesnik na visoki ravni. Privzeto je ta zavihek viden le na strokovni ravni.",
"Enable this filter.", "Omogoči filter.",
"Enable trace subsystems", "Omogoči sledi v podsistemu",
"Enable web UI debug (non-minified JS)", "Omogoči web UI razhroščevanje (ne-stisnjen JS)",
"Enable, disable or ignore the mux. When the mux is marked as ignore, all discovered services are removed.", "Omogočite, onemogočite ali ignorirate MUX. Ko je mux označen kot prezrti, so vse odkrite storitve odstranjene.",
"Enable/Disable CA client.", "Omogoči / Onemogoči CA stranko.",
"Enable/Disable service.", "Omogoči / onemogoči storitev.",
"Enable/Disable the entry.", "Omogoči / onemogoči vnos.",
"Enable/Disable timeshift.", "Omogoči/onemogoči časovni zamik",
"Enable/disable EPG data for the entry.", "Vklop / izklop podatkov EPG za vpis.",
"Enable/disable LNA.", "Omogoči / onemogoči LNA.",
"Enable/disable auto-rec rule.", "Vklop / izklop pravilo samodejnega snemanja.",
"Enable/disable debugging output to syslog.", "Omogoči / onemogoči odpravljanje izhoda napak v dnevnik.",
"Enable/disable inclusion of low-level debug traces.", "Omogoči / onemogoči vključitev debug sledi na nizki ravni.",
"Enable/disable interface quick tips.", "Omogoči / onemogoči vmesnik hitrih nasvetov.",
"Enable/disable libav log output.", "Omogoči / onemogoči libav dnevnik izhod.",
"Enable/disable logging to syslog.", "Omogočiti / onemogočiti beleženje v dnevnik.",
"Enable/disable offering of Entitlement Management Message updates.", "Omogoči / onemogoči ponudbo posodobitev - Pravica upravljanje sporočila.",
"Enable/disable pilot tone.", "Omogoči/onemogoči pilotni signal.",
"Enable/disable power save mode (if supported by the device).", "Omogoči / onemogoči način varčevanja z energijo (če jo naprava podpira).",
"Enable/disable profile.", "Omogoči / onemogoči profil.",
"Enable/disable the bouquet.", "Omogočanje / onemogočanje šopek.",
"Enable/disable the channel.", "Vklop / izklop kanal.",
"Enable/disable the device.", "Vklop / izklop naprave.",
"Enable/disable the entry.", "Omogoči / onemogoči vnos.",
"Enable/disable the grabber.", "Omogoči/onemogoči graber.",
"Enable/disable the tag.", "Omogoči / onemogoči oznake.",
"Enable/disable tuner/adapter.", "Omogoči / onemogoči sprejemnik / adapter.",
"Enabled", "omogočeno",
"Encrypted", "šifrirano",
"Encrypted service", "šifrirana storitev",
"Enter a non-admin user password.", "Vnesite uporabniško geslo non-admin.",
"Enter a non-admin user username.", "Vnesite uporabniško ime non-admin.",
"Enter allowed network prefix(es). You can enter a comma-seperated list of prefixes here.", "Vnesite dovoljeno omrežno predpono (e). Lahko vnesete z vejicami ločen seznam predpon tukaj.",
"Enter an administrator password.", "Vnesite skrbniško geslo.",
"Enter an administrator username. Note: do not use the same username as the superuser backdoor account.", "Vnesite skrbniško uporabniško ime. Opomba: ne uporabljajo isto ime kot račun skrbnik backdoor.",
"Enter comma-separated list of subsystems you want debugging output for (e.g +linuxdvb,+subscriptions,+mpegts).", "Vnesite z vejicami ločen seznam podsistemov, ki jih želite za odpravljanje izhodnih napak (npr + linuxdvb, + naročnine + mpegts).",
"Enter comma-separated list of subsystems you want to get traces for (e.g +linuxdvb,+subscriptions,+mpegts).", "Vnesite vejicami ločen seznam podsistemov za željene sledove (npr + linuxdvb, + naročnine + mpegts).",
"Entertainment programs for 10 to 16", "Zabavni programi za 10. do 16.",
"Entertainment programs for 6 to 14", "Zabavni programi za 6. do 14.",
"Environment", "okolje",
"Environment (pipe)", "Okolje (pipe)",
"Episode", "Epizoda",
"Episode %d", "Epizoda %d",
"Episode number/ID.", "Številka Epizode number/ID.",
"Equestrian", "Konjeniški",
"Error code", "Napačna koda",
"Error code of entry.", "Koda napake vstopa.",
"Errors", "napake",
"Even key", "enak ključ",
"Even key.", "Enak ključ",
"Event does not exist", "Dogodek ne obstaja",
"Every day", "Vsak dan",
"Exclude channel tags", "Izključi oznake kanalov ",
"Expeditions", "Odprave",
"Experimental film", "eksperimentalni film",
"Expert", "Strokovnjak",
"Exported tuners", "Izvozni tunerji",
"External", "Zunanji",
"External EPG grabber", "Zunanji EPG grabber",
"External IP (NAT)", "Zunanja IP (NAT)",
"External PyEPG Grabber", "Zunanji PyEPG Grabber",
"External URL", "Zunanji URL",
"External URL of the bouquet.", "Zunanji URL šopek.",
"External XMLTV EPG Grabber", "Zunanji XMLTV EPG Grabber",
"Extra arguments", "dodatni argumenti",
"Extra start time", "Poseben čas začetka",
"Extra stop time", "Poseben čas Konca",
"Extra warming up time (seconds)", "Dodatno čas ogrevanja (v sekundah)",
"FAIL", "NAPAKA",
"FEC high", "FEC Visoko",
"FEC low", "FEC nisko",
"Factual topics", "dejanske teme",
"Failed", "ni uspelo",
"Failed to open image", "Napaka pri odpiranju slike",
"Failed view", "ni pogleda",
"Fantasy", "Fantazija",
"Fashion", "Moda",
"Fast input switch", "stikalo hitri vhod ",
"File missing", "datoteka manjka",
"File not created", "Datoteka ni ustvarjena",
"File permissions (octal, e.g. 0664)", "Dovoljenja datotek (osmiško, npr 0664)",
"File removed", "datoteke odstranjenje",
"File size", "Velikost datoteke",
"Filename", "Ime datoteke",
"Filename (including path)", "Ime datoteke (vključno s potjo)",
"Filename character set", "Ime datoteke - nabor znakov",
"Filename options", "Ime datoteke - možnosti",
"Filename used by the entry.", "Ime datoteke - uporaba z vpisom.",
"Filesystem monitor", "Pregled datotečnega sistema",
"Fine arts", "Likovno",
"Finished", "Končano",
"First word", "prva beseda",
"Fit to RAM (cut rewind)", "Prilagodi RAM (rez nazaj)",
"Fitness and health", "Fitnes in zdravje",
"Folk", "Narodno",
"Folkloric", "Ljudska",
"Football", "Nogomet",
"Force all network connections to this tuner to be made over the specified IP address, similar to the setting for the SAT>IP device itself. Setting this overrides the device-specific setting.", "Uporabi vse omrežne povezave s tem sprejemnikom, ki so določene z IP naslovom, podobno kot nastavitev za SAT>IP naprave. Nastavitev bo preglasila normalno nastavitev naprave.",
"Force an initial EPG grab at start-up.", "Prisili začetni EPG ob zagonu.",
"Force initial EPG grab at start-up", "Prisili začetni EPG ob zagonu",
"Force old status", "Prisili staro stanje",
"Force the delay between RTSP TEARDOWN and RTSP SETUP command (value from 'Next tune delay in ms' is used). Some devices are not able to handle quick continuous tuning.", "Zamik med RTSP nastavitvijo in RTSP SETUP ukazom (vrednost iz \"Naprej nastavi zamude v ms). Nekatere naprave niso sposobne izpeljati hitrega neprekinjenega strema.",
"Force the local UDP Port number here. The number should be even (RTP port). The next odd number (+1) will be used as the RTCP port.", "Lokalno številko UDP vrat. Številka mora biti tudi (RTP port). Naslednje liho število (1), se uporablja kot port RTCP.",
"Force usage of entered CA ID on this service.", "Prisili uporabo ID CA za to storitev.",
"Foreign countries", "Tuje države",
"Forever", "Za vedno",
"Format string", "format niza",
"Found muxes", "Najdeni muxi",
"Found services", "Najdene storitve",
"Free subscription weight", "Brezplačna naročnina",
"Free-form field, enter whatever you like here.", "Prosto polje, vnesi kar ti je všeč.",
"Free-form text field, enter whatever you like here.", "Polje z prostim besedilom, vpiši kar ti je všeč.",
"Free-form text field, enter whatever you like.", "polje s prostim besedilom, vnesite karkoli želite.",
"Free-form text field. You can enter whatever you like here.", "Prosto besedilno polje. Lahko vnesete karkoli želite.",
"Free-format text field. Enter whatever you like here.", "Prosti format - besedilno polje. Vpiši, kar ti je všeč.",
"Frequency", "frekvenca",
"Frequency (Hz)", "Frekvenca (Hz)",
"Frequency (MHz)", "Frekvenca (MHz)",
"Frequency (in MHz).", "Frekvenca (v MHz).",
"Frequency (kHz)", "Frekvenca (kHz)",
"Fri", "Pet",
"Friendly name", "prijazno ime",
"Frontend number", "številka začelja",
"Frontend path", "pot začelja",
"Full mux RX mode supported", "Polni mux RX način podprt",
"Full path to an icon used to depict the tag. This can be a TV network logotype, etc.", "Celotna pot do ikone. To je lahko omrežni logotip TV, itd",
"Full pathname specification", "Kompletna pot",
"Full-text", "Polno besedilo",
"Further education", "Nadaljnje izobraževanje",
"GOTOX position", "GOTOX pozicija",
"Game show", "oddaja igre",
"Gardening", "Vrtnarjenje",
"General", "splošno",
"General configuration", "Splošna konfiguracija",
"Generic", "generična",
"Generic options", "generične možnosti",
"Global Headers", "Globalne Glave",
"Global timer", "globalni čas",
"Gray", "Siva",
"HD: high definition", "HD: visoka ločljivost",
"HDHomeRun ATSC-C frontend", "HDHomeRun ATSC-C začelje",
"HDHomeRun ATSC-T frontend", "HDHomeRun ATSC-T začelje",
"HDHomeRun DVB frontend", "HDHomeRun DVB začelje",
"HDHomeRun DVB-C frontend", "HDHomeRun DVB-C začelje",
"HDHomeRun DVB-T frontend", "HDHomeRun DVB-T začelje",
"HTSP Answer", "HTSP Odgovor",
"HTSP Default Stream Settings", "Nastavitve HTSP toka (Privzeto)",
"HTSP Request", "HTSP Zahteva",
"HTSP Stream Profile", "HTSP Stream profila",
"HTSP Subscription", "HTSP Naročnina",
"HTTP CORS (cross-origin resource sharing) origin. This option is usually set when Tvheadend is behind a proxy. Enter a domain (or IP) to allow cross-domain requests.", "HTTP CORS (delitev cross-izvor virov) poreklo. Ta možnost je ponavadi nastavljena, ko je Tvheadend za Proxy strežnikom. Vnesite domeno (ali IP), da se omogoči meddomenske zahteve.",
"Handicraft", "obrti",
"Hearing impaired", "gluhi in naglušni",
"Hierarchy", "hierarhija",
"High", "visoko",
"High bitrate mode (CI+ CAMs only)", "Način visoke bitne hitrosti (samo CI + CAM)",
"Highest channel number the user can access.", "Največje število kanala do katerega lahko uporabnik dostopa.",
"Highest count of objects.", "Najvišja število predmetov.",
"Historical movie", "Zgodovinski film",
"Horror", "Grozljivka",
"Hostname (or IP) of the server.", "Ime gostitelja (ali IP) strežnika.",
"Hostname/IP", "Ime gostitelja / IP",
"How frequently it will re-try fetching an image that has failed to be fetched.", "Kako pogosto bo ponovno poskusil dobiti sliko, ki je ni bilo mogoče pridobiti.",
"How frequently the upstream provider is checked for changes.", "Kako pogosto ponudnik preveri spremembe.",
"ID of the network.", "ID omrežja.",
"IGNORE", "IGNORIRAJ",
"IP address", "IP naslov",
"IP blocking", "IP blokiranje",
"IPTV Automatic Network", "IPTV Samodejno omrežje",
"IPTV Multiplex", "IPTV Multipleks",
"IPTV Network", "IPTV omrežje",
"IPTV Subcription", "Naročnina IPTV",
"IPTV input", "IPTV vhod",
"ISDB-C Network", "ISDB-C omrežje",
"ISDB-C multiplex", "ISDB-C multipleks",
"ISDB-S Network", "ISDB-S Omrežje",
"ISDB-S multiplex", "ISDB-S multipleks",
"ISDB-T Network", "ISDB-T Omrežje",
"ISDB-T multiplex", "ISDB-T multipleks",
"Icon", "Ikona",
"Icon (full URL)", "Icon (polni URL)",
"Icon URL", "URL icone",
"Icon base URL", "Ikona baze URL",
"Icon base URL.", "baza URL ikona.",
"Icon has title", "Ikona ima naslov",
"Icons from", "ikone iz",
"Idle scan", "idle skeniranje",
"Idle scan muxes", "Idle muxes skeniranje",
"If an error occurs clone the scheduled entry and try to record again (if possible).", "Če pride do napake, kloniraj redni vpis in poskusi ponovno snemanje (če je možno).",
"If both a picon and a channel-specific (e.g. channelname.jpg) icon are defined, prefer the picon.", "Če sta oba Picon in kanal (npr channelname.jpg) ikona definirani, raje izberi Picon.",
"If more than x errors occur during a recording schedule a re-record (if possible).", "Če je več kot x napak v snemalnem urniku, ponovno snemaj (če je možno).",
"If no specific priority was requested. This gives certain users a higher priority by assigning a streaming profile with a higher priority.", "Če ni bila zahtevana nobena posebna prednost. To daje nekaterim uporabnikom večjo prednost z dodelitvijo profila pretakanja z višjo prioriteto.",
"If no user account exists then create one with\nno username and no password. Use with care as\nit will allow world-wide administrative access\nto your Tvheadend installation until you create or edit\nthe access control from within the Tvheadend web interface.", "Če ni uporabniškega računa, ustvari novega\nni uporabniškega imena in gesla. PREVIDNA Uporaba\nUpravljanje bo možno od povsod po svetu\nna svoj Tvheadend namestite, dokler ne boste ustvarili ali uredili\nnadzor dostopa znotraj spletnega vmesnika Tvheadend.",
"If set, presentation of the tag icon will not superimpose the tag name on top of the icon.", "Če je nastavljeno, postavitev ikone oznake ne prikaže imena oznake na vrhu ikone.",
"If set, the first bytes from the MPEG-TS stream are discarded. It may be required for some drivers or hardware which do not flush the MPEG-TS buffers completely after a frequency/parameter change.", "Če je določeno, se prvi bajt iz toka MPEG-TS zavrže. Lahko zahteva za nekatere gonilnike ali strojno opremo, ki ne podpira buferja MPEG-TS popolnoma po spremembi frekvence / parametrov.",
"If the DiseqC switch is located before the rotor (i.e. tuner - switch - rotor), enable this.", "Če se DiSEqC stikalo nahaja pred rotorjem (tj sprejemnik - stikalo - rotor), ga omogoči.",
"If the subscription weight for the input is below the specified threshold, the tuner is handled as free (according the priority settings). Otherwise, the next tuner (without any subscriptions) is used. Set this value to 10, if you are willing to override scan and epggrab subscriptions.", "Če je naročnina za vnos pod določenim pragom, se sprejemnik obravnava kot prost (v skladu s prednostnimi nastavitvami). V nasprotnem primeru se uporabi naslednji sprejemnik (brez naročnine). To vrednost nastavite na 10, če ste pripravljeni, da preskočite skeniranje in epggrab naročnine.",
"If you'd like to support the project, please consider a donation.", "Če želite podpreti projekt, donirajte.",
"Ignore", "Ignoriraj",
"Ignore EPG (EIT)", "Ignoriraj EPG (EIT)",
"Ignore descriptor 5", "Ignoriraj deskriptor 5",
"Ignore invalid SSL certificate", "Ignoriraj neveljavno potrdilo SSL",
"Ignore invalid/unverifiable (expired, self-certified, etc.) certificates", "Ignoriraj neveljavnih / nepreverljive (potekla, self-certifikatom, itd) potrdila",
"Ignore provider's channel numbers", "Ignoriraj številke kanalov ponudnika",
"Immediately abort", "takoj prekini",
"Important", "pomembno",
"Include channel name in filename", "Vključi ime kanala v datoteko",
"Include date in filename", "Vključi datum v ime datoteke",
"Include episode in filename", "Vključi epizodo v ime datoteke",
"Include subtitle in filename", "Vključi podnapis v ime datoteke",
"Include teletext", "Vključi teletekst",
"Include the date for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "Vključuje datum za vpis v naslovu dogodka. To velja tako za naslov, shranjen v datoteki in ime same datoteke.",
"Include the episode subtitle in the title (if available).", "Vključi epizodo podnapis v naslovu (če je na voljo).",
"Include the name of the channel in the event title. This applies to both the title stored in the file and to the filename itself.", "Vsebuje ime kanala v naslovu dogodka. To velja tako za naslov, shranjene v datoteki in ime same datoteke.",
"Include the season and episode in the title (if available).", "Vključi sezono in epizodo v naslovu (če je na voljo).",
"Include the time for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "Vključuje čas za vpis v naslovu dogodka. To velja tako za naslov, shranjene v datoteki in ime same datoteke.",
"Include time in filename", "Vključi čas v ime datoteke",
"Index", "Indeks",
"Information area", "območje z informacijami",
"Informational", "Informacijsko",
"Initial scan", "začetno skeniranje",
"Input base", "vhodna baza",
"Input buffer (bytes)", "Input buffer (bajti)",
"Input instance", "vhod primer",
"Input path", "vhodna pot",
"Interface", "vmesnik",
"Internal", "notranja",
"Internal EPG grabber", "Notranji EPG grabber",
"Internal PyEPG Grabber", "Notranji PyEPG Grabber",
"Internal XMLTV EPG Grabber", "Notranji XMLTV EPG Grabber",
"Internal error", "Notranja napaka",
"Internal grabber", "notranji grabber",
"Interview", "Intervju",
"Invalid", "Neveljavno",
"Invalid EPG object request", "Neveljavna zahteva za EPG objekt",
"Invalid adapter number '%s'\n", "Neveljavna številka adapter-ja '%s'\n",
"Invalid arguments", "Neveljavni argumenti",
"Invalid file", "Neveljavna datoteka",
"Invalid parameters", "Neveljavni parametri",
"Invalid service", "Neveljavna storitev",
"Invalid target", "Neveljavni cilj",
"Items", "predmeti",
"Keep", "Obdrži",
"Keep x amount of storage space free.", "Naj bo x količina prostora prazno.",
"Keepalive interval in seconds", "Keepalive interval v sekundah",
"Key and value pairs to remove from the query string in the URL.", "Ključne in vrednost parov odstranjenje iz poizvedbe v URL.",
"LNA (low noise amplifier)", "LNA (nizki ojačevalnik hrupa)",
"LNB type", "LNB tip",
"Language", "Jezik",
"Language 1", "Jezik 1",
"Language 2", "Jezik 2",
"Language 3", "Jezik 3",
"Language settings", "jezikovne nastavitve",
"Languages", "jeziki",
"Largest size the object has reached.", "Največja velikost, ki jo je dosegel cilj.",
"Last played position", "Zadnji predvajani položaj ",
"Last played position when the recording isn't fully watched yet.", "Zadnja predvajani položaj, ko posneto še ni ogledano v celoti.",
"Last seen", "Nazadnje viden",
"Last updated from", "Zadnja posodobitev iz",
"Layer A: Constellation", "Sloj A: Constellation",
"Layer A: FEC", "Sloj A: FEC",
"Layer A: Segment count", "Sloj A: Število segmentov",
"Layer A: Time interleaving", "Sloj A: Čas prepletanja",
"Layer B: Constellation", "Plast B: Constellation",
"Layer B: FEC", "Sloj B: FEC",
"Layer B: Segment count", "Plast B: Število segmentov",
"Layer B: Time interleaving", "Sloj B: Čas prepletanja",
"Layer C: Constellation", "Sloj C: Constellation",
"Layer C: FEC", "Plast C: FEC",
"Layer C: Segment count", "Sloj C: Število segmentov",
"Layer C: Time interleaving", "Plast C: Čas prepletanja",
"Leisure hobbies", "Prosti čas",
"Limit connections", "mejne povezave",
"Limited/limit scanning to this network ID only.", "Omejitev skeniranja v omrežju ID.",
"Linked input", "vhod povezave ",
"Linux DVB frontend", "Linux DVB začelje",
"LinuxDVB Input", "LinuxDVB vnosa",
"LinuxDVB network", "LinuxDVB omrežje",
"List of allowed IPv4 or IPv6 hosts or networks (comma-separated).", "Seznam dovoljenih IPv4 ali IPv6 gostiteljev ali omrežij (ločenih z vejico).",
"List subsystems", "Seznam podsistema",
"Listen / Connect port", "Poslušaj / Poveži vrata",
"Listen on IPv6", "Poslušaj na IPv6",
"Literature", "Literatura",
"Local bind IP address", "Lokalni IP naslov",
"Local channel minor", "Lokalni kanal minor",
"Local channel number", "Lokalna številka kanala",
"Local discovery IP address", "Lokalno odkritje IP naslov",
"Local: Record if different description", "Lokalno: Record, če je drugačen opis",
"Local: Record if different episode number", "Lokalno: Record če je drugačna številka epizode",
"Local: Record if different subtitle", "Lokalno: Record, če je drugačen podnaslov",
"Local: Record if different title", "Lokalno: Record, če je drugačen podnaslov",
"Local: Record once per day", "Lokalno: Snemanje enkrat na dan",
"Local: Record once per week", "Lokalno: Snemanje enkrat na teden",
"Location", "lokacija",
"Location details of the SAT>IP Server.", "Podrobnosti lokacije za SAT> IP Server.",
"Locking", "zaklepanje",
"Log", "dnevnik",
"Login password.", "Prijavno geslo.",
"Login username.", "Prijavno uporabniško ime.",
"Login/Logout", "Prijava / Odjava",
"Logout", "Odjava",
"Low", "nizka",
"Lowest channel number the user can access.", "Najnižja številka kanala, do katere lahko uporabnik dostopa.",
"Lowest starting channel number.", "Najnižja začetna številka kanala.",
"MPEG-TS File", "MPEG-TS Datoteka",
"MPEG-TS Input Debug", "MPEG-TS vhod za odpravljanje napak",
"MPEG-TS Multiplex", "MPEG-TS Multipleks",
"MPEG-TS Pass-thru/built-in", "MPEG-TS Pass-through / vgrajeni",
"MPEG-TS input", "MPEG-TS vhod",
"MPEG-TS multiplex PHY", "MPEG-TS multipleks PHY",
"MPEG-TS raw service", "MPEG-TS raw storitev",
"Magazines", "revije",
"Main", "glavna",
"Maintain free storage space in MiB", "Ohrani prosti prostor za shranjevanje v MIB",
"Maintain used storage space in MiB (0=disabled)", "Ohrani rabljeni prostor v MIB (0 = onemogočeno)",
"Maintained space", "vzdrževani prostor",
"Make subdirectories per channel", "Doda podimenike za kanal",
"Make subdirectories per day", "Doda podimenike za dan",
"Make subdirectories per title", "Dodaj podimenike za naslov",
"Manage all", "Upravljanje vseh",
"Manufacturer", "Proizvajalec",
"Manufacturer URL", "URL Proizvajalec",
"Manufacturer's model number.", "Številka modela proizvajalca.",
"Manufacturer's product description.", "Opis izdelka proizvajalca.",
"Manufacturer's product information page for the device.", "Stran z informacijami izdelka proizvajalca naprave.",
"Manufacturer's product name.", "Ime izdelka proizvajalca.",
"Map all services", "Mapiraj vse storitve",
"Map encrypted services", "Mapiraj šifrirane storitve",
"Map radio channels", "Mapiraj radio kanale",
"Map unnamed channels", "Mapiraj neimenovane kanale",
"Map zero-numbered channels", "Mapiraj kanale brez številk",
"Martial sports", "borilni športi",
"Matroska (mkv)/built-in", "Matroska (mkv)/vgrajen",
"Matroska muxer", "Matroska muks",
"Matroska specific", "Matroska specifično",
"Maximal channel number", "Maksimalno število kanalov",
"Maximum # input streams", "Največje # vhodnih tokov ",
"Maximum PIDs", "Največ PID",
"Maximum RAM size (MB)", "Največja velikost RAM (MB)",
"Maximum allowed difference between event start time when the EPG event is changed.", "Največja dovoljena razlika med začetnim časom dogodka , ko se dogodek EPG spremeni.",
"Maximum bandwidth (Kbps)", "Maksimalna pasovna širina (Kbps)",
"Maximum count (0=default)", "Največje število (0 = privzeto)",
"Maximum duration", "Najdaljše trajanje",
"Maximum input bandwidth.", "Največja vhodna pasovna širina.",
"Maximum length in characters for the command setting PIDs to the SAT>IP box.", "Največja dolžina znakov za ukaz nastavitev PID-ov na SAT>IP polju.",
"Maximum length of PIDs", "Največja dolžina PID",
"Maximum period (mins)", "Najdaljše obdobje (min)",
"Maximum schedules limit (0=default)", "Zgornja meja urnika (0 = privzeto)",
"Maximum size (MB)", "Največja velikost (MB)",
"Maximum time to wait (in seconds) for a stream before a timeout.", "Najdaljši čas čakanja (v sekundah) za stream pred prekinitvijo.",
"Maximum timeout (seconds)", "Največja Zakasnitev (v sekundah)",
"Medicine", "Medicina",
"Memory Information", "Informacije o pomnilniku",
"Merge same name", "Spoji isto ime",
"Method not found", "Metode ni bilo mogoče najti",
"Minimal channel number", "Minimalno število kanalov",
"Minimum duration", "Minimalno trajanje",
"Minimum rotor time (seconds)", "Minimalni čas rotorja (sekund)",
"Missing In PAT/SDT", "Manjka V PAT / SDT",
"Mode", "način",
"Model description", "opis modela",
"Model name", "ime modela",
"Model number", "Številka modela",
"Modulation", "modulacija",
"Module", "modul",
"Module ID", "modul ID",
"Module ID used to grab EPG data.", "Modul ID se uporablja, da zgrabi podatke EPG.",
"Mon", "Pon",
"Motor rate (in milliseconds/deg).", "Stopnja motorja (v milisekundah/deg).",
"Motor rate (milliseconds/deg)", "Stopnja motorja (ms/deg)",
"Motor sport", "motorni športi",
"Motoring", "avtomobilizem",
"Movie", "Film",
"Movie / drama", "Film / drama",
"Music", "Glasba",
"Music / Ballet / Dance", "Glasba / balet / Ples",
"Musical", "Glasbeno",
"Mux Scheduler", "Mux Urnik",
"Mux handling", "ravnanje MUX",
"Mux name", "Mux Ime",
"Mux not enabled", "MUX ni omogočen",
"NONE", "NIČ",
"Name", "Ime",
"Name of object.", "Ime objekta.",
"Name of the bouquet.", "Ime šopka.",
"Name of the client.", "Ime clienta.",
"Name of the entry.", "Ime vpisa.",
"Name of the module used to grab EPG data.", "Ime modula, ki se uporablja za pridobitev podatkov EPG.",
"Name of the module, grabber or channel that should be used to update this channels EPG info.", "Ime modula, Grabber ali kanal, ki ga je treba uporabljati za posodobitev kanalov EPG.",
"Name of the network.", "Ime omrežja.",
"Name of the profile.", "Ime profila.",
"Name of the tag.", "Ime oznake.",
"Name of the tuner.", "Ime sprejemnika.",
"Name of the tuner/adapter.", "Ime tuner / adapter.",
"Names", "Imena",
"Natural sciences", "Naravoslovno",
"Nature", "Narava",
"Network", "Omrežje",
"Network ", "Omrežje",
"Network A", "Omrežje A",
"Network B", "Omrežje B",
"Network C", "Omrežje C",
"Network D", "Omrežje D",
"Network ID (limit scanning)", "Omrežje ID (Omejitev skeniranja)",
"Network UUID", "Omrežje UUID",
"Network access", "dostop do omrežja",
"Network discovery", "odkrivanje omrežja",
"Network for port A.", "Mreža za port A.",
"Network for port AA.", "Mreža za port AA.",
"Network for port AB.", "Mreža za port AB.",
"Network for port B.", "Mreža za port B.",
"Network for port BA.", "Mreža za port BA.",
"Network for port BB.", "Mreža za port BB.",
"Network for port C.", "Mreža za port C.",
"Network for port D.", "Mreža za port D.",
"Network group", "omrežna skupina",
"Network limit per position", "Omejitev omrežja na pozicijo",
"Network name", "ime omrežja",
"Network prefix", "Predpona mreže",
"Network settings", "omrežne nastavitve",
"Network type", "Vrsta omrežja",
"Networks", "omrežja",
"New / premiere / unknown", "Novo / premiera / neznano",
"New login", "Nova prijava",
"New media", "Novi mediji",
"New muxes + changed muxes", "Novi muxi + spremenjeni muxi",
"New muxes only", "Samo novi muxi",
"News", "Novice",
"News / Current Affairs", "Novice / Aktualno",
"News magazine", "revija novice",
"Next tune delay in ms (0-2000)", "Naslednji glas zamudi v ms (0-2000)",
"No", "Ne",
"No PIN", "Brez PIN",
"No access", "ni dostopa",
"No adapters specified!\n", "Adapterji niso določeni\n",
"No assigned adapters", "Brez dodeljenega adapterja",
"No days", "ni dni",
"No descrambler", "Ni deskremblerja",
"No free adapter", "Ni prostega adapterja",
"No input detected", "Vhod ni zaznan",
"No scheme", "Ni sheme",
"No service assigned to channel", "Kanalu ni dodeljena storitev",
"No service enabled", "Storitev ni omogočena",
"No source available", "Vir ni na voljo",
"Node subsystem", "Node podsistem",
"None", "Noben",
"Not all SAT>IP servers use the same signal scaling. Change this setting if the signal level displayed within Tvheadend looks too low.", "SAT> IP strežniki ne uporabljajo enakega razmerja signala. Spremenite te nastavitve, če je nivo signala, ki je prikazan v Tvheadend, prenizka.",
"Not enough disk space", "Ni dovolj prostora na disku",
"Not enough memory", "Ni dovolj pomnilnika",
"Not linked", "ni povezan",
"Not set", "ni nastavljeno",
"Not set (none or channel configuration)", "Ni nastavljeno (nobena konfiguracija kanalov)",
"Not set (use channel or DVR configuration)", "Ni nastavljeno (uporaba kanalov ali razporeditve DVR)",
"Number", "Število",
"Number of days to keep the file.", "Datoteka se hrani (število dni).",
"Number of days to keep the recorded file.", "Število dni, hrambe posnetka.",
"Number of days to keep the recorded files.", "Število dni, hrambe posnetih datotek.",
"Number of days to retain entry information.", "Število dni hranjenja podatkov vnosa.",
"Number of days to retain information about recordings. Once this period is exceeded, duplicate detection will not be possible anymore.", "Število dni hranjenja podatkov o posnetku. Ko je ta rok prekoračen, dvojnik zaznave ne bo več mogoč.",
"Number of errors that occurred during recording (Transport errors).", "Število napak, ki so nastale med snemanjem (napake v pretoku).",
"Number of errors that occurred during recording.", "Število napak, ki so nastale med snemanjem.",
"Number of muxes found.", "Število najdenih muxov.",
"Number of services.", "Število storitev.",
"Number of times this recording was played.", "Koliko krat se je posneto predvajalo.",
"Number of tsfile tuners", "Število tsfile tunerjev",
"OK (partial)", "OK (delno)",
"Odd key", "Odd ključ",
"Odd key.", "Odd ključ.",
"Offset the mapped channel numbers by x (value here + channel number).", "Offset preslikano število kanalov s x (ta vrednost + številka kanala).",
"Older OSCam", "starejši OSCam",
"On", "Vključeno",
"On file removal", "Po odstranitvi datoteke",
"On-demand (no first rewind)", "Na zahtevo",
"Once per auto channel", "Enkrat na auto kanal",
"One tune request (setup) is sent before the DiseqC sequence (voltage, tone settings). Some linux drivers require this procedure.", "Ena zahteva tune (nastavitev) se pošlje pred DiSEqC zaporedju (napetost, nastavitve tonov). Nekateri gonilniki za linux zahtevajo ta postopek.",
"Only Bulsatcom 39E", "Samo Bulsatcom 39E",
"Only EIT", "samo EIT",
"Only OpenTV Sky Ausat", "Samo OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Samo OpenTV Sky Italia",
"Only OpenTV Sky UK", "Samo OpenTV Sky UK",
"Only PSIP (ATSC)", "Samo PSIP (ATSC)",
"Only UK Freesat", "Samo UK Freesat",
"Only UK Freeview", "SamoUK Freeview",
"Only Viasat Baltic", "Samo Viasat Baltic",
"Only digits", "Samo številke",
"Orbital position", "Orbitalni položaj",
"Orbital positions", "Orbitalni položaji",
"Orbital positions.", "Orbitalni položaji.",
"Oscam mode.", "Oscam način.",
"Owner", "Lastnik",
"Password", "Geslo",
"Web interface", "Spletni vmesnik",
"Web interface language", "Jezik spletnega vmesnika",
"Web interface language.", "Jezik spletnega vmesnika.",
"Winter sports", "Zimski športi",
"Wizard", "Čarovnik",
"Yes", "Da",
"ultrafast: h264 / h265", "ultra hitro: h264 / h265",
"veryfast: h264 / h265 / qsv(h264)", "zelo hitro: h264 / h265 / qsv(h264)",
"veryslow: h264 / h265 / qsv(h264)", "zelo počasi: h264 / h265 / qsv(h264)",
NULL, NULL
};

static const char *tvh_locale_30[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nPre viac informacií, prosim navštívte stránky Tvheadend:\nhttps://tvheadend.org\n",
"# Mapped channels", "# Namapované kanály",
"# Muxes", "# Multiplexy",
"# Services", "# Služby",
"# tune repeats", "# opakované ladenie",
"(Default profile)", "(Predvolený profil)",
"(Read-only)", "(Iba na čítanie)",
"1 day", "1 deň",
"1 month", "1 mesiac",
"1 week", "1 týždeň",
"1 year", "1 rok",
"1316 bytes", "1316 bajtov",
"16356 bytes", "16356 bajtov",
"2 months", "2 mesiace",
"2 weeks", "2 týždne",
"2 years", "2 roky",
"2-Port switch (universal LNB)", "2-Portový prepínač (univerzálne LNB)",
"2632 bytes", "2632 bajtov",
"3 days", "3 dni",
"3 months", "3 mesiace",
"3 weeks", "3 týždne",
"3 years", "3 roky",
"32712 bytes", "32712 bajtov",
"4-Port switch (universal LNB)", "4-Portový prepínač (univerzálne LNB)",
"5 days", "5 dní",
"5264 bytes", "5264 bajtov",
"6 months", "6 mesiacov",
"65424 bytes", "65424 bajtov",
"7896 bytes", "7896 bajtov",
"A channel tag (e.g. a group of channels) to which this rule applies.", "Tag kanálu (napr. skupina kanálov) pre ktoré je pravidlo platné.",
"AC-3 = descriptor 6", "AC-3 = deskriptor 6",
"AC-3 detection", "detekcia AC-3",
"ACTIVE", "AKTÍVNE",
"ANY", "Akékoľvek",
"ATSC SI Tables", "ATSC SI Tabuľky",
"ATSC source ID", "ATSC ID zdroja",
"ATSC-C Network", "ATSC-C Sieť",
"ATSC-T Network", "ATSC-T Sieť",
"AUTO", "AUTOMATICKY",
"Aborted by user", "Prerušené uživateľom",
"Accept remote subscription weight", "Povoliť vzdialené predplatné",
"Accept the remote subscription weight (from the SAT>IP client).", "Povoliť vzdialené predplatné (z SAT>IP klienta).",
"Accept zero value for TSID", "Povoliť nulovú hodnotu pre TSID",
"Access", "Prístup",
"Access (ACL)", "Prístup (ACL)",
"Action", "Akcia",
"Active", "Aktívne",
"Adapter in use by another subscription", "Adaptér je používaný iným predplatným",
"Add file and line numbers to debug", "Pridať súbor a čísla riadkov pre ladenie",
"Add the thread ID to debug", "Pridať ID vlákna pre ladenie",
"Add this value to PTS for the teletext subtitles. The time value is in milliseconds and may be negative.", "Pridať túto hodnotu do PTS pre teletextové titulky. Čas je v milisekundách a može byť záporný.",
"Additional arguments to pass to the grabber.", "Ďaľšie argumenty predávané do grabberu.",
"Additional service names found in EPG data.", "Dodatočné názvy služieb nájdené v EPG dátach.",
"Admin", "Administrátor",
"Admin password", "Heslo administrátora",
"Admin username", "Uživateľské meno administrátora",
"Administrator login", "Prihlásenie administrátora",
"Adult movie", "Film pre dospelých",
"Advanced", "Rozšírené",
"Advanced (non-universal LNBs, rotors, etc.)", "Rozšírené (neuniverzálne LNB, motory, atď.)",
"Adventure", "Dobrodružné",
"Advertisement / Shopping", "Reklama / Nakupovanie",
"After creating the network scan it for services.", "Po vytvorení siete, skenovať služby",
"All (Streaming plus DVR)", "Všetko (Streaming plus DVR)",
"All lower-case", "Všetko malými písmenami",
"All proceeds are used to support server infrastructure and buy test equipment.", "Výnos je používaný pre podporu serverovej infraštruktúry a nákup testovacích nástrojov.",
"All: Record if different description", "Všetko: Nahrávať ak sa líší popis",
"All: Record if different episode number", "Všetko: Nahrávať ak sa líši číslo epizódy",
"All: Record if different subtitle", "Všetko: Nahrávať ak sa líšia titulky",
"All: Record once per day", "Všetky: Nahrávať raz denne",
"All: Record once per month", "Všetko: Nahrávať mesačne",
"All: Record once per week", "Všetko: Nahrávať týždenne",
"Allow all PIDs", "Povoliť všetky PIDy",
"Allow all PIDs (no filter) when the 'Maximum PIDs' limit is reached.", "Povoliť všetky PIDy (bez filtra), keď je dosiahnutý limit 'Maximálny PID'.",
"Allow high bitrate mode (CI+ CAMs only).", "Povoliť mód vyššieho bitrate (iba CI+ CAMy).",
"Allow idle scan tuning on this device.", "Povoliť ladenie na tomto zariadení.",
"Allow/Disallow access to the 'Configuration' tab.", "Povoliť/Zakázať prístup k záložke \"Konfigurácia\"",
"Allow/Disallow web interface access (this  includes access to the EPG).", "Povoliť/Zakázať  prístup k webovému rozhraniu (zahŕňa aj prístup k EPG).",
"Allowed DVR profiles. This limits the profiles the user has access to.", "Povolené profily DVR. Obmedzuje profily ku ktorým ma uživateľ prístup.",
"Allowed network", "Povolená sieť",
"Allowed networks", "Povolené siete",
"Alternate PID path", "Alternatívna cesta k PID",
"Alternate configuration path", "Alternatívna cesta ku konfigurácii",
"Altitude (in meters).", "Nadmorská výška (v metroch).",
"Altitude (meters)", "Nadmorská výška (v metroch)",
"Always use the name defined by the network.", "Vždy používať názov poskytnutý sieťou.",
"Animals", "Zvieratá",
"Anonymize", "Anonymizovať",
"Anonymize HTSP access", "Anonymizovať HTSP prístup",
"Any", "Akékoľvek",
"Arts", "Umenie",
"Arts / Culture (without music)", "Umenie/Kultúra (bez hudby)",
"Arts magazines", "Umelecké magazíny",
"Assign predefined muxes to networks", "Priradiť predefinované multiplexy k sieťam",
"Associate this device with one or more networks.", "Priradiť zariadenie k jednej alebo k viacerým sieťam.",
"Athletics", "Atletika",
"Audio Stream Filter", "Filter audio streamu",
"Audio bitrate (kb/s) (0=auto)", "Audio bitrate (kb/s) (0=automaticky)",
"Audio bitrate to use for transcoding.", "Audio bitrate používaný pre prekódovanie.",
"Audio channel layout.", "Rozloženie zvukových kanálov.",
"Audio codec", "Audio kodek",
"Audio codec to use for the transcode. \"Do not use\" will disable audio output.", "Audio kodek používaný pre prekódovanie. \"Nepoužívať\" vypne audio výstup.",
"Audio type", "Typ audia",
"Audio-only stream", "Iba audio",
"Audioes muxer", "Audio multiplexy",
"Authenticated user", "Prihlásený uživateľ",
"Auto", "Automaticky",
"Auto check disabled", "Automatická kontrola vypnutá",
"Auto check enabled", "Automatická kontrola zapnutá",
"Auto record", "Automatický záznam",
"Auto record caption", "Filmový titulok automatického záznamu",
"Auto recording%s%s", "Automatický záznam%s%s",
"Auto time record", "Automatický časový záznam",
"Auto-Map to channels", "Automatické namapovanie na kanály",
"Automatic checking", "Automatická kontrola",
"Automatic recording caption.", "Automatický názov nahrávky.",
"Automatic schedule entry not found", "Nenájdený automatický plán",
"Automatic time scheduler entry not found", "Nenájdený záznam automatického časového rozvrhu",
"Automatically link EPG data to the channel (using the channel name for matching). If you turn this option off, only the OTA EPG grabber will be used for this channel unless you've specifically set a different EPG Source.", "Automaticky priradiť EPG dáta ku kanálu (používa meno kanálu pre namapovanie). Ak túto možnosť vypnete, bude sa používať OTA EPG grabber, pokiaľ nevyberiete iný zdroj EPG.",
"Automatically map EPG source", "Automaticky priradiť zdroj EPG",
"Automatically map all available services to channels.", "Automaticky priradiť všetky dostupné služby ku kanálom.",
"Automatically map channels defined within the bouquet.", "Automaticky priradiť kanály definované v buketu.",
"Automatically name from network", "Automatický názov zo siete",
"Automatically record.", "Automaticky nahrávať.",
"Bad", "Chyba",
"Bad request", "Chybná požiadavka",
"Ballet", "Balet",
"Bandwidth", "Šírka pásma",
"Base DVB SI Tables (PAT,CAT,PMT,SDT etc.)", "Základné DVB SI Tabuľky (PAT,CAT,PMT,SDT atď.)",
"Based on software from", "Založené na software",
"Basic", "Základné",
"Bind to specific local IP address.", "Počúvať na špecifickej IP adrese.",
"Blue", "Modré",
"Bouquet (auto)", "Bouquet  (automaticky)",
"Bouquet source.", "Zdroj pre Bouquet",
"Bouquets", "Bouquet-y",
"Brand", "Výrobca",
"Branding information (if available).", "Informácie o výrobcovi (ak sú prístupné)",
"Broadcast", "Vysielanie",
"Broadcast type", "Typ vysielania",
"Broadcast.", "Vysielanie.",
"Broadcasting", "Vysielanie",
"Buffering limit (ms)", "Limit vyrovnávacej pamäte (ms)",
"Build", "Zostavenie",
"CA (descrambling) Client", "CA (dekódovací) Klient",
"CA Stream Filter", "CA Filter Stream-u",
"CA client %i", "CA klient%i",
"CA identification", "CA identifikácia",
"CAPMT CA Client", "CAPMT CA Klient",
"CAPMT interval (in ms).", "CAPMT interval (v ms).",
"CAPMT query interval (ms)", "CAPMT interval požiadavky (ms)",
"CAPMT query interval (ms).", "CAPMT interval dotazu (ms).",
"CI Module", "CI Modul",
"CRID authority", "CRID autorita",
"CSA (descrambling)", "CSA (dekódovanie)",
"CWC CA Client", "CWC CA Klient",
"Cache scheme", "Schémy cache",
"Camd.socket filename / IP Address (TCP mode)", "Súbor Camd.socket / IP Adresa (TCP mód)",
"Cartoons", "Rozprávky",
"Change parameters", "Zmena parametrov",
"Channel", "Kanál",
"Channel Tags", "Tagy kanálu",
"Channel does not exist", "Kanál neexistuje",
"Channel icon", "Ikona kanálu",
"Channel icon URL.", "URL ikony kanálu",
"Channel icon as defined in EPG data.", "Ikona kanálu z EPG dát.",
"Channel icon name scheme", "Schéma názvu ikony kanálu",
"Channel icon path", "Cesta k ikone kanálu",
"Channel mapping options", "Možnosti mapovania kanálu",
"Channel name", "Názov kanálu",
"Channel not found", "Kanál neboj nájdený",
"Channel number", "Číslo kanálu",
"Channel number as defined in EPG data.", "Číslo kanálu podľa EPG dát.",
"Channel number offset", "Ofset čísla kanálu",
"Channel number range", "Rozsah čísla kanálu",
"Channel numbers (heuristic)", "Čísla kanálov (heuristika)",
"Channel numbers from", "Čísla kanálov z",
"Channel tag", "Tag kanálu",
"Channel tag does not exist", "Tag kanálu neexistuje",
"Channel tags", "Tagy kanálov",
"Channel tags the user is allowed access to/excluded from.", "Tagy kanálov ku ktorým má/nemá uživateľ prístup",
"Channel to use/used for the recording.", "Kanál používaný pre nahrávanie.",
"Channel update options", "Možnosti aktualizácie kanálu",
"Channels", "Kanály",
"Channels EPG data is used by.", "Použitie EPG dát kanálov.",
"Character set", "Znaková sada",
"Charset", "Znaková sada",
"Check availability", "Skontrolovať dostupnosť",
"Check for the services' presence. If the service is no longer broadcast this field will change to Missing In PAT/SDT.", "Skontrolovať či služba existuje. Ak nie je služba naďalej vysielaná, bude zmenená na Chýbajúcu v PAT/SDT.",
"Children's / Youth Programs", "Programy pre deti",
"Children's / Youth programs", "Programy pre deti",
"Cinema", "Kino",
"Class", "Trieda",
"Classical", "Klasika",
"Classical music", "Klasická hudba",
"Client name", "Meno klienta",
"Clone scheduled entry on error", "Klonovať naplánovanú položku pri chybe",
"Comedy", "Komédie",
"Command delay time (ms) (10-200)", "Čas oneskorenia príkazu (ms) (10-200)",
"Command time (ms) (10-100)", "Čas príkazu (ms) (10-100)",
"Comment", "Komentár",
"Commercial break", "Komerčná prestávka",
"Committed", "Použité",
"Completed OK", "Ukončené OK",
"Compress EPG database", "Komprimovať EPG databázu",
"Compress the EPG database to reduce disk I/O and space.", "Komprimovať databázu EPG pre zníženie I/O záťaže na disk a ušetrenie priestoru.",
"Conditional Access", "Podmienený prístup (CA)",
"Conditional Access Client", "Klient podmieneného prístupu",
"Conditional Access Identification.", "Identifikácia podmieneného prístupu",
"Configuration", "Konfigurácia",
"Configuration - Base", "Konfigurácia - Základné",
"Configuration - Image Cache", "Konfigurácia - Cache Obrázku",
"Configuration - SAT>IP Server", "Konfigurácia - SAT>IP Server",
"Configuration ID", "Konfigurácia ID",
"Configuration name", "Názov konfigurácie",
"Configuration version", "Verzia konfigurácie",
"Connection limit type", "Typ obmedzenia pripojenia",
"Connection limits", "Limit pripojení",
"Constellation", "Konštalácia",
"Container", "Kontajner",
"Container to use for the transcoded stream.", "Kontajner pre transkódované streamy.",
"Content character set", "Znaková sada obsahu",
"Content type", "Typ obsahu",
"Content type.", "Typ obsahu.",
"Contest", "Súťaž",
"Continue if descrambling fails", "Pokračovať ak dekódovanie zlyhá",
"Continue recording for x minutes after scheduled stop time", "Pokračovať v nahrávaní po x minút po nastavenom čase ukončenia",
"Continue recording for x minutes after scheduled stop time.", "Pokračovať v nahrávaní po x minút po nastavenom čase ukončenia.",
"Cookie expiration (days)", "Expirácia cookie (v dňoch)",
"Cooking", "Varenie",
"Copy codec type", "Typ kopírovacieho kodeku",
"Copy layout", "Kopírovať usporiadanie",
"Count of objects", "Počet objektov",
"Create SDTV/HDTV/Radio tags.", "Vytvoriť SDTV/HDTV/Rádio tagy.",
"Create a bouquet from the playlist.", "Vytvoriť bouquet z playlistu.",
"Create a provider name tag.", "Vytvoriť tag mena poskytovateľa.",
"Create and associate a network tag to created channels.", "Vyvoriť a priradiť sieťový tag k vytvoreným kanálom.",
"Create and associate a provider tag to created channels.", "Vytvoriť a priradiť tag poskytovateľa k vytvoreným kanálom.",
"Create and link these tags to channels when mapping.", "Vytvoriť a spojiť tieto tagy s kanálmi pri mapovaní.",
"Create bouquet", "Vytvoriť bouquet",
"Create bouquet tag", "Vytvoriť bouquet tag",
"Create directories using these permissions.", "Vytvárať adresáre s týmito právami.",
"Create files using these permissions.", "Vytvoriť súbory s použitím týchto práv.",
"Create network name tags", "Vytvoriť tagy sieťových mien",
"Create network name tags (set by provider).", "Vytvoriť tagy mien sietí (nastavené poskytovateľom)",
"Create network tags", "Vytvoriť tagy siete",
"Create provider name tags", "Vytvoriť tagy sieťových poskytovateľov",
"Create provider tags", "Vytvoriť tagy poskytovateľa",
"Create tags", "Vyvoriť tagy",
"Create type-based tags", "Vytvoriť tagy na základe typu",
"Created", "Vytvorené",
"Creator", "Vytvoril",
"Cron multi-line", "Viacriadkový Cron",
"Culture (without music)", "Kultúra (bez hudby)",
"Culture magazines", "Magazíny o kultúre",
"Current RTSP port.", "Aktuálny RTSP port.",
"Current affairs", "Aktuality",
"Current number of objects.", "Aktuálny počet objektov.",
"Current object size.", "Aktuálna veľkosť objektu",
"Current tuner configuration.", "Aktuálna konfigurácia tuneru.",
"Custom HTTP headers", "Uživateľské HTTP hlavičky",
"DAB Network", "Sieť DAB",
"DBus - use the session message bus instead of the system one", "DBus - používať zbernicu správ namiesto systémovej",
"DES Key.", "Kľúč DES.",
"DES key", "Kľúč DES",
"DSCP/TOS for streaming", "DSCP/TOS pre stremovanie",
"DVB CAM Client", "DVB CAM Klient",
"DVB CSA (descrambling) Tables", "DVB CSA (dekódovanie) Tabuľky",
"DVB EPG Tables", "DVB EPG Tabuľky",
"DVB Inputs - Networks", "DVB Vstupy - Siete",
"DVB SI Tables", "DVB SI Tabuľky",
"DVB Time Tables", "DVB Časové Tabuľky",
"DVB scan files", "Súbory skenu DVB",
"DVB scan files path", "Cesta k súborom skenu DVB",
"DVB-C Network", "DVB-C Sieť",
"DVB-S Network", "Sieť DVB-S",
"DVB-S Satellite Configuration", "Konfigurácia satelitu DVB-S",
"DVB-T Network", "Sieť DVB-T",
"DVR - Auto-recording (Autorecs)", "DVR - Auto-nahrávanie (Autorec)",
"DVR - Profiles", "DVR - Profily",
"DVR - Time-based Recording (Timers)", "DVR - Časové Nahrávanie (Časovače)",
"DVR behavior", "DVR správanie",
"DVR configuration", "DVR konfigurácia",
"DVR configuration profiles", "DVR konfiguračné profily",
"DVR configurations", "DVR konfigurácie",
"DVR entry not found", "DVR položka nenájdená",
"DVR log retention", "DVR uchovávanie logu",
"DVR log retention period", "DVR perióda uchovávania logu",
"DVR profile to use/used for the recording.", "DVR profil používaný/použitý pre nahrávanie.",
"DVR schedule does not exist", "DVR plán neexistuje",
"DVR schedule does not have a file yet", "DVR plán zatiaľ nemá súbor",
"DVR schedule not found", "DVR program nenájdený",
"Dance", "Tanečné",
"Data errors", "Chyby dát",
"Data path (if applicable).", "Cesta k dátam (ak je použiteľné)",
"Days of Week", "Dni v Týždni",
"Days of the week to which the rule should apply.", "Dni v týždni pre ktoré je pravidlo aktívne.",
"Days of week", "Dni v týždni",
"Debate", "Debata",
"Debug libav log", "Ladiť libav log",
"Debug options", "Možnosti ladenia",
"Debug subsystems", "Ladenie subsystémov",
"Debug to syslog", "Ladenie do syslog-u",
"Debugging", "Ladenie",
"Default", "Predvolené",
"Default language(s)", "Predvolený jazyk(y)",
"Default language.", "Predvolený jazyk",
"Default login", "Predvolené uživateľské meno",
"Default priority", "Predvolená priorita",
"Default user interface level.", "Predvolené uživateľské rozhranie.",
"Define network group to limit network usage.", "Definovať skupinu siete pre zníženie sieťovej záťaže.",
"Delivery system", "Systém doručovania",
"Demux path", "Cesta k demultiplexu",
"Descramble services (limit per mux)", "Dekódovať služby (limit na 1 multiplex)",
"Descrambler", "Dekodér",
"Descrambler EMM", "Dekodér EMM",
"Descrambler buffer (TS packets)", "Dekódovací bufer (TS pakety)",
"Descrambling timeout (ms)", "Dekodér timeout (ms)",
"Description", "Popis",
"Detective", "Dedektívka",
"Device ID", "ID zariadenia",
"Device model", "Model zariadenia",
"Device path", "Cesta k zariadeniu",
"Device path in sysfs", "Cesta k zariadeniu v sysfs",
"Digital Video Recorder", "Digitálny Video Recordér",
"Directory", "Adresár",
"Directory permissions (octal, e.g. 0775)", "Práva adresára (v osmičkovej sústave, napr. 0775)",
"Directory used by the entry.", "Adresár použitý položkou.",
"Disable", "Zakázať",
"Disable DVB bouquets", "Zakázať DVB bouquet-y",
"Disable SAT>IP client", "Zakázať SAT>IP klienta",
"Disable X_SATIPM3U tag", "Zakázať X_SATIPM3U tagy",
"Disable all access control checks", "Vypnúť všetky kontroly prístupových práv",
"Disable debug on stderr", "Zakázať ladenie na stderr",
"Disable device/firmware-specific workarounds", "Vypnúť obchádzanie špecifík zariadenia/firmware",
"Disable syslog (all messages)", "Vypnúť syslog (všetky správy)",
"Disabled", "Zakázané",
"Discussion", "Diskusie",
"DiseqC repeats", "DiseqC opakovanie",
"Do not use", "Nepoužívať",
"Documentary", "Dokumentárne",
"Don't include title in filename", "Nezahŕňať titulok v názve súboru",
"Don't keep", "Nedržať",
"Don't re-record", "Nenahrávať znova",
"Don't re-record if recording fails.", "Nenahrávať znova ak nahrávanie zlyhalo.",
"Don't re-schedule if recording fails.", "Nenaplánovať znova ak nahrávanie zlyhalo.",
"Don't reschedule", "Neopakovať plánovanie",
"Don't use the provider's channel numbers.", "Nepoužívať čísla kanálov od poskytovateľa.",
"Drama", "Dráma",
"Duplicate handling", "Riešenie duplicít",
"Duplicate recording handling.", "Riešenie duplicít nahrávok.",
"EIT - skip TSID check", "EIT - preskočiť TSID kontrolu",
"EPG Database", "EPG Databáza",
"EPG Language (priority order)", "EPG Jazyk (zoradené podľa priority)",
"EPG data ID.", "ID dát EPG.",
"EPG name", "EPG názov",
"EPG overlap cut", "Prekrytie EPG",
"EPG scan", "EPG sken",
"EPG source", "EPG zdroj",
"EPG update window", "EPG aktualizačné okno",
"Economics", "Ekonomika",
"Education", "Vzdelávanie",
"Educational", "Vzdelávacie",
"Electronic Program Guide", "Elektronický Sprievodca Programom",
"Enable", "Povoliť",
"Enable (auto)", "Zapnúť (automaticky)",
"Enable DBus", "Povoliť DBus",
"Enable NTP driver", "Povoliť NTP ovládač.",
"Enable coredumps for daemon", "Zapnúť coredump-y pre démona",
"Enable debug on stderr", "Zapnúť ladenie na stderr",
"Enable debug subsystems", "Zapnúť ladenie subsystémov",
"Enable debug to file", "Povoliť ladienie do súboru",
"Enable debug to syslog", "Povoliť ladienie do syslog-u",
"Enable if the SAT>IP box requests plts=on parameter in the SETUP RTSP command for DVB-S2 muxes.", "Povoľte pokiaľ SAT>IP box vyžaduje parameter plts=on v príkaze SETUP RTSP pre DVB-S2 multiplexy.",
"Enable or disable fast input switching.", "Povoliť alebo zakázať rýchle prepínanie vstupu.",
"Enable or disable full mux mode.", "Povoliť alebo zakázať plný mód multiplexu.",
"Enable or disable this configuration.", "Povoliť alebo zakázať túto konfiguráciu.",
"Enable syslog", "Zapnúť syslog",
"Enable this filter.", "Povoliť tento filter.",
"Enable trace subsystems", "Zapnúť trasovanie subsystémov",
"Enable web UI debug (non-minified JS)", "Povoliť ladenie webového rozhrania (neminimalizovaný JS)",
"Enable, disable or ignore the mux. When the mux is marked as ignore, all discovered services are removed.", "Povolí, zakáže alebo ignoruje multiplex. Keď je multiplex označený ako ignorovaný, všetky objavené služby su odstránené.",
"Enable/Disable CA client.", "Povoliť/Zakázať CA klienta.",
"Enable/Disable service.", "Povoliť/Zakázať službu.",
"Enable/Disable the entry.", "Povoliť/Zakázať položku.",
"Enable/Disable timeshift.", "Povoliť/Zakázať timeshift.",
"Enable/disable EPG data for the entry.", "Povoliť/Zakázať EPG dáta pre položku.",
"Enable/disable LNA.", "Povoliť/Zakázať LNA.",
"Enable/disable auto-rec rule.", "Zapnúť/vypnúť pravidlo automatického nahrávania.",
"Enable/disable debugging output to syslog.", "Povoliť/Zakázať výstup ladenia do syslog-u.",
"Enable/disable inclusion of low-level debug traces.", "Povoliť/Zakázať nízkoúrovňové ladiace krokovanie.",
"Enable/disable interface quick tips.", "Povoliť/Zakázať rýchle tipy v rozhraní.",
"Enable/disable libav log output.", "Povoliť/Zakázať libav logovací výstup.",
"Enable/disable logging to syslog.", "Povoliť/Zakázať logovanie do syslog-u.",
"Enable/disable power save mode (if supported by the device).", "Povoliť/Zakázať šetriaci režim (ak je podporované zariadením).",
"Enable/disable profile.", "Povoliť/Zakázať profil.",
"Enable/disable the bouquet.", "Povoliť/Zakázať bouquet-y.",
"Enable/disable the channel.", "Povoliť/Zakázať kanál.",
"Enable/disable the device.", "Povoliť/Zakázať zariadenie.",
"Enable/disable the entry.", "Povoliť/Zakázať položku.",
"Enable/disable the grabber.", "Povoliť/Zakázať graber.",
"Enable/disable the tag.", "Povoliť/Zakázať tag.",
"Enable/disable tuner/adapter.", "Povoliť/Zakázať tuner/adaptér.",
"Enabled", "Povolené",
"Encrypted", "Zakódované",
"Encrypted service", "Zakódovaná služba",
"Enter a non-admin user password.", "Vložte heslo uživateľa ktorý nie je administrátor.",
"Enter a non-admin user username.", "Vložte meno uživateľa ktorý nie je adminstrátor.",
"Enter allowed network prefix(es). You can enter a comma-seperated list of prefixes here.", "Vložte povolené prefix(i) siete. Môžete vložiť čiarkou oddelený zoznam prefixov.",
"Enter an administrator password.", "Vložte heslo administrátrora.",
"Enter an administrator username. Note: do not use the same username as the superuser backdoor account.", "Vložte uživateľské meno administrátora. Poznámka: nepoužívajte rovnaké uživateľské meno ako pre super-uživateľa.",
"Enter the filename (including path) where Tvheadend should write the log.", "Vložte názov súboru (vrátane cesty) kde môže Tvheadend zapisovať log.",
"Entertainment programs for 10 to 16", "Zábavné programy pre vekovú kategóriu od 10 do 16",
"Entertainment programs for 6 to 14", "Zábavné programy pre vekovú kategóriu od 6 do 14",
"Environment", "Prostredie",
"Environment (pipe)", "Prostredie (pipe)",
"Episode", "Epizóda",
"Episode %d", "Epizóda %d",
"Episode number/ID.", "Epizóda číslo/ID.",
"Equestrian", "Jazdectvo",
"Error code", "Chybový kód",
"Error code of entry.", "Chybový kód položky.",
"Errors", "Chyby",
"Even key", "Párny kľúč",
"Even key.", "Párny kľúč.",
"Event does not exist", "Udalosť neexistuje",
"Every day", "Každý deň",
"Exclude channel tags", "Vynechať tagy kanálov",
"Expeditions", "Expedície",
"Experimental film", "Experimentálny film",
"Exported tuners", "Exportované tunery",
"External", "Externý",
"External EPG grabber", "Externý EPG graber",
"External IP (NAT)", "Externá IP (NAT)",
"External PyEPG Grabber", "Externý PyEPG Graber",
"External URL", "Externá URL",
"External XMLTV EPG Grabber", "Externý XMLTV EPG Graber",
"Extra arguments", "Extra argumenty",
"FAIL", "CHYBA",
"Factual topics", "Faktické témy",
"Failed", "Zlyhalo",
"Failed to open image", "Chyba pri otváraní obrázku",
"Fantasy", "Fantastické",
"Fashion", "Móda",
"Fast input switch", "Rýchle prepínanie vstupu",
"File missing", "Súbor neexistuje",
"File not created", "Súbor nebol vytvorený",
"File permissions (octal, e.g. 0664)", "Práva súborov (v osmičkovej sústave, napr. 0664)",
"File removed", "Súbor vymazaný",
"File size", "Veľkosť súboru",
"Filename", "Meno súboru",
"Filename (including path)", "Meno súboru (s cestou)",
"Filename character set", "Znaková sada mena súboru",
"Filename options", "Možnosti mena súboru",
"Filename used by the entry.", "Meno súboru použité položkou.",
"Filesystem monitor", "Monitorovanie súborového systému",
"Fine arts", "Moderne umenie",
"Finished", "Dokončené",
"First word", "Prvé slovo",
"Fitness and health", "Fitnes a zdravie",
"Folkloric", "Folklór",
"Football", "Futbal",
"Force (auto)", "Vynútiť (automaticky)",
"Force CA ID (e.g. 0x2600)", "Vynútiť CA ID (e.g. 0x2600)",
"Force an initial EPG grab at start-up.", "Vynútené počiatočné EPG grabovanie pri štarte.",
"Force initial EPG grab at start-up", "Vynútiť počiatočné EPG grabovanie pri štarte.",
"Force old status", "Vynútiť starý stav",
"Force pilot for DVB-S2", "Vynútiť pilot pre DVB-S2",
"Force priority", "Vynútiť prioritu",
"Force profile to use this priority.", "Vynútiť použitie tejto priority pre profil.",
"Force service type to 1", "Vynútiť typ služby 1",
"Force teardown delay", "Vynútiť oneskorenie štartu",
"Force usage of entered CA ID on this service.", "Vynútiť používanie vloženého CA ID na tejto službe.",
"Foreign countries", "Vynútiť krajiny",
"Forever", "Navždy",
"Fork and run as daemon", "Fork-núť a spustiť ako démona",
"Format string", "Formát reťazca",
"Found muxes", "Nájdené multiplexy",
"Found services", "Vyhľadané služby",
"Frequency", "Frekvencia",
"Frequency (Hz)", "Frekvencia (Hz)",
"Frequency (MHz)", "Frekvencia (MHz)",
"Frequency (in MHz).", "Frekvencia (v MHz).",
"Frequency (kHz)", "Frekvencia (kHz)",
"Fri", "Pi",
"Friendly name", "Čitateľný názov",
"Frontend number", "Číslo frontendu",
"Frontend path", "Cesta k frontendu",
"Full DiseqC", "Plný DiseqC",
"Full mux RX mode supported", "Plný RX mód multiplexu je podporovaný",
"Further education", "Ďalšie vzdelávanie",
"GOTOX position", "GOTOX pozícia",
"Game show", "Hry",
"Gardening", "Záhradkárstvo",
"General", "Hlavné",
"General configuration", "Hlavná configurácia",
"Generic", "Základné",
"Generic options", "Základné možnosti",
"Global Headers", "Globálne Nadpisy",
"Global timer", "Globálny časovač",
"Grace period", "Odklad",
"Gray", "Sivý",
"Guard interval", "Ochranný interval",
"HD: high definition", "HD: vysoké rozlíšenie",
"HTSP Answer", "HTSP Odpoveď",
"HTSP Default Stream Settings", "HTSP Predvolené Nastavenia Streamu",
"HTSP Request", "HTSP Požiadavka",
"HTSP Stream Profile", "HTSP Streamovací Profil",
"HTSP Subscription", "HTSP Subskripcia",
"HTTP Client", "HTTP Klient",
"Handicraft", "Ručné práce",
"Hearing impaired", "Sluchovo postihnutý",
"Hierarchy", "Hierarchia",
"Historical movie", "Historický film",
"Horror", "Horor",
"Hostname (or IP) of the server.", "Hostname (alebo IP) servera.",
"ID of the network.", "ID siete.",
"IDLE", "NEČINNÝ",
"IGNORE", "IGNOROVAŤ",
"IP address", "IP adresa",
"IP blocking", "IP blokovanie",
"IPTV Automatic Network", "IPTV Automatická Sieť",
"IPTV Network", "IPTV Sieť",
"IPTV Subcription", "IPTV Subskripcia",
"IPTV input", "IPTV vstup",
"ISDB-C Network", "ISDB-C Sieť",
"ISDB-S Network", "ISDB-S Sieť",
"ISDB-T Network", "ISDB-T Sieť",
"ISI (Stream ID)", "ISI (ID Streamu)",
"Icon", "Ikona",
"Icon (full URL)", "Ikona (plná URL)",
"Icon URL", "Ikona URL",
"Icons from", "Ikony z",
"Idle", "Nečinný",
"Idle scan", "Nečinný sken",
"Idle scan muxes", "Nečinné skenovanie multiplexov",
"If you'd like to support the project, please consider a donation.", "Ak by ste radi podporili tento projekt, prosíme, zvážte príspevok.",
"Ignore", "Ignorovať",
"Ignore EPG (EIT)", "Ignorovať EPG (EIT)",
"Ignore invalid SSL certificate", "Ignorovať neplatný SSL certifikát",
"Ignore provider's channel numbers", "Ignorovať čísla kanálov od poskytovateľa",
"Image Cache", "Cache obrázkov",
"Immediately abort", "Okamžite prerušiť",
"Important", "Dôležité",
"Include channel name in filename", "Zahrnúť meno kanálu do mena súboru",
"Include date in filename", "Zahrnúť dátum do mena súboru",
"Include episode in filename", "Zahrnúť názov epizódy do mena súboru",
"Include subtitle in filename", "Zahrnúť titulky do mena súboru",
"Include teletext", "Zahrnúť teletext",
"Include time in filename", "Zahnúť čas do názvu súboru",
"Information area", "Informačné pole",
"Informational", "Medzinárodné",
"Initial scan", "Počiatočný sken",
"Input buffer (bytes)", "Vstupný buffer (bajty)",
"Input path", "Cesta k vstupu",
"Interface", "Rozhranie",
"Internal", "Interné",
"Internal EPG grabber", "Interný EPG graber",
"Internal PyEPG Grabber", "Interný PyEPG Graber",
"Internal XMLTV EPG Grabber", "Interný XMLTV EPG Graber",
"Internal error", "Interná chyba",
"Internal grabber", "Interný graber",
"Invalid", "Neplatné",
"Invalid adapter number '%s'\n", "Neplatné číslo adaptéru '%s'\n",
"Invalid arguments", "Neplatné argumenty",
"Invalid file", "Neplatný súbor",
"Invalid parameters", "Neplatné parametre",
"Invalid service", "Neplatná služba",
"Invalid target", "Neplatný cieľ",
"Items", "Položky",
"Keep", "Zachovať",
"Keep x amount of storage space free.", "Zachovať x miesta na disku volného.",
"Keepalive interval", "Interval udržiavania",
"Keepalive interval in seconds", "Interval udržiavania v sekunách",
"LNA (low noise amplifier)", "LNA (nízkošumový zosilovač)",
"LNB type", "Typ LNB",
"Language", "Jazyk",
"Language 1", "Jazyk 1",
"Language 2", "Jazyk 2",
"Language 3", "Jazyk 3",
"Language settings", "Nastavenie jazyka",
"Languages", "Jazyky",
"Last played position", "Posledná prehrávaná pozícia",
"Last played position when the recording isn't fully watched yet.", "Pokračovať od poslednej pozície ak nahrávka nebola prehraná celá.",
"Last seen", "Naposledy videné",
"Last updated from", "Naposledy obnovené z",
"Leisure hobbies", "Voľná zábava",
"Limit connections", "Obmedzenia pripojení",
"Limited/limit scanning to this network ID only.", "Limitovať skenovanie iba na ID tejto siete.",
"Linked input", "Spojený vstup",
"LinuxDVB Input", "LinuxDVB Vstup",
"LinuxDVB adapter", "LinuxDVB adaptér",
"LinuxDVB network", "LinuxDVB sieť",
"List of allowed IPv4 or IPv6 hosts or networks (comma-separated).", "Zoznam povolených IPv4 alebo IPv6 hostiteľov alebo sietí (oddelených čiarkou)",
"List subsystems", "Zoznam subsystémov",
"Listen on IPv6", "Počúvať na IPv6",
"Literature", "Literatúra",
"Local (server) time", "Lokálny čas (zo serveru)",
"Local channel number", "Lokálne číslo kanálu",
"Location", "Lokácia",
"Locking", "Zamykanie",
"Login password.", "Prihlasovacie heslo.",
"Login username.", "Prihlasovacie meno.",
"Login/Logout", "Prihlásiť/Odhlásiť",
"Logout", "Odhlásiť",
"MPEG-TS File", "MPEG-TS Súbor",
"MPEG-TS Input Debug", "MPEG-TS Ladenie Vstupu",
"MPEG-TS Pass-thru/built-in", "MPEG-TS Pass-thru/vstavané",
"MPEG-TS Service", "MPEG-TS Služba",
"MPEG-TS input", "MPEG-TS vstup",
"MPEG-TS raw service", "MPEG-TS surové dáta",
"Magazines", "Magazíny",
"Main", "Hlavné",
"Maintain free storage space in MiB", "Zachovávať voľné miesto na úložisku v MiB",
"Maintain used storage space in MiB (0=disabled)", "Zachovávať použité miesto na úložisku v MiB (0=vypnuté)",
"Maintained space", "Dostupné miesto",
"Make subdirectories per channel", "Vytvárať podadresáre podľa kanálu",
"Make subdirectories per day", "Vytvárať podadresáre denne",
"Make subdirectories per title", "Vytvárať podadresáre podľa názvu",
"Manage all", "Spravovať všetky",
"Manufacturer", "Výrobca",
"Manufacturer URL", "URL výrobcu",
"Manufacturer's model number.", "Číslo modelu výrobcu",
"Manufacturer's product description.", "Popis produktu od výrobcu.",
"Manufacturer's product information page for the device.", "Produktová stránka výrobcu zariadenia.",
"Manufacturer's product name.", "Názov produktu od výrobcu.",
"Map all services", "Namapovať všetky služby",
"Map encrypted services", "Namapovať zakódované služby",
"Map radio channels", "Namapovať rádio kanály",
"Map unnamed channels", "Namapovať nepomenované kanály",
"Map zero-numbered channels", "Namapovať kanály bez čísla kanálu",
"Mapped", "Namapované",
"Martial sports", "Bojové športy",
"Master tuner", "Hlavný tuner",
"Matroska (mkv)/built-in", "Matroska (mkv)/vstavaný",
"Matroska specific", "Matroska špecifiké",
"Maximal channel number", "Maximálne číslo kanálu",
"Maximum # input streams", "Maximálne # vstupných streamov",
"Maximum PIDs", "Maximálne PID",
"Maximum RAM size (MB)", "Maximálna veľkosť RAM  (MB)",
"Maximum bandwidth (Kbps)", "Maximálna šírka pásma (Kbps)",
"Maximum count (0=default)", "Maximálny počet (0=predvolené)",
"Maximum duration", "Maximálna dĺžka",
"Maximum input bandwidth.", "Maximálna šírka pásma vstupu.",
"Maximum length of PIDs", "Maximálna dĺžka PID-ov",
"Maximum period (mins)", "Maximálna doba (minuty)",
"Maximum size (MB)", "Maximálna veľkosť (MB)",
"Maximum time to wait (in seconds) for a stream before a timeout.", "Maximálna doba čakania (v sekundách) predtým ako vyprší timeout streamu.",
"Maximum timeout (seconds)", "Maximálny timeout (v sekundách)",
"Medicine", "Medicínske",
"Melodrama", "Melodramatické",
"Memory Information", "Informácie o pamäti",
"Merge same name", "Spojiť rovnaké mená",
"Merge services with the same name into one channel.", "Spojiť služby s rovnakým menom do jedného kanálu.",
"Method not found", "Metóda nenájdená",
"Minimal channel number", "Minimálne číslo kanálu",
"Minimum duration", "Minimálna dĺžka",
"Minimum rotor time (seconds)", "Minimálny čas motora (v sekundách)",
"Missing In PAT/SDT", "Chýbajúce v PAT/SDT",
"Mode", "Mód",
"Model description", "Popis modelu",
"Model name", "Meno modelu",
"Model number", "Číslo modelu",
"Modulation", "Modulácia",
"Module", "Modul",
"Module ID", "ID Modulu",
"Module ID used to grab EPG data.", "ID Modulu pre grabovanie EPG dát",
"Mon", "Po",
"More verbose libav log", "Podrobnejšie logovanie libav",
"Motor rate (in milliseconds/deg).", "Rýchlosť motora (v milisekundách/stupeň).",
"Motor rate (milliseconds/deg)", "Rýchlosť motora (v milisekundách/stupeň)",
"Motor sport", "Motoristický šport",
"Motoring", "Motor",
"Movie", "Film",
"Movie / drama", "Film / Dráma",
"Music", "Hudba",
"Music / Ballet / Dance", "Hudba / Balet / Tanec",
"Musical", "Hudobné",
"Mux", "Multiplex",
"Mux Scheduler", "Plánovač multiplexu",
"Mux UUID", "Multiplex UUID",
"Mux handling", "Ovládanie multiplexu",
"Mux name", "Názov multiplexu",
"Mux not enabled", "Nepovolený multiplex",
"NONE", "ŽIADNE",
"Name", "Názov",
"Name of channel the entry recorded from.", "Názov kanálu z ktorého bolo nahrávané.",
"Name of object.", "Názov objektu.",
"Name of the bouquet.", "Názov bouquet-u.",
"Name of the client.", "Názov klienta.",
"Name of the entry.", "Názov položky.",
"Name of the module used to grab EPG data.", "Názov modulu pre grabovanie EPG dát.",
"Name of the network.", "Názov siete.",
"Name of the profile.", "Názov profilu.",
"Name of the tag.", "Názov tagu.",
"Name of the tuner.", "Názov tuneru.",
"Name of the tuner/adapter.", "Názov tuneru/adaptéru.",
"Names", "Názvy",
"Natural sciences", "Prírodné vedy",
"Nature", "Príroda",
"Network", "Sieť",
"Network ", "Sieť",
"Network A", "Sieť A",
"Network B", "Sieť B",
"Network C", "Sieť C",
"Network D", "Sieť D",
"Network ID (limit scanning)", "ID siete (limit skenovania)",
"Network UUID", "UUID Siete",
"Network access", "Prístup k sieti",
"Network discovery", "Vyhľadanie siete",
"Network for port A.", "Sieť pre port A.",
"Network for port AA.", "Sieť pre port AA.",
"Network for port AB.", "Sieť pre port AB.",
"Network for port B.", "Sieť pre port B.",
"Network for port BA.", "Sieť pre port BA.",
"Network for port BB.", "Sieť pre port BB.",
"Network for port C.", "Sieť pre port C.",
"Network for port D.", "Sieť pre port B.",
"Network group", "Skupina siete",
"Network limit per position", "Obmedziť siet na danú pozíciu",
"Network name", "Názov siete",
"Network prefix", "Prefix siete",
"Network settings", "Nastavenie siete",
"Network type", "Typ siete",
"Networks", "Siete",
"New / premiere / unknown", "Nové / premiéra / neznáme",
"New login", "Nový účet",
"New media", "Nové médium",
"New muxes + changed muxes", "Nové + zmenené multiplexy",
"New muxes only", "Iba nové multiplexy",
"News", "Správy",
"News / Current Affairs", "Novinky / Aktuality",
"News magazine", "Správy",
"No", "Číslo",
"No PIN", "Žiadny PIN",
"No access", "Žiadny prístup",
"No adapters specified!\n", "Žiadne adaptéry neboli zadané!\n",
"No assigned adapters", "Nepriradené adaptéry",
"No days", "Žiadne dni",
"No free adapter", "Žiadny voľný adaptér",
"No input detected", "Žiadny vstup nebol detekovaný",
"No scheme", "Žiadna schéma",
"No service assigned to channel", "Žiadna služba nie je priradená ku kanálu",
"No service enabled", "Žiadna služba nie je aktivovaná",
"No source available", "Žiadny zdroj",
"None", "Žiadne",
"Normal", "Normálne",
"Not enough disk space", "Nedostatok miesta na disku",
"Not enough memory", "Nedostatok pamäte",
"Not linked", "Nezlúčené",
"Not set", "Nenastavené",
"Not set (none or channel configuration)", "Nenastavené (žiadne alebo konfigurované kanálom)",
"Not set (use channel or DVR configuration)", "Nenastavené (použiť kanál alebo konfiguráciu DVR)",
"Number", "Číslo",
"Number of days to keep the file.", "Počet dní pre zachovanie súboru.",
"Number of errors that occurred during recording (Transport errors).", "Počet chýb počas nahrávania (Chyby prenosu).",
"Number of errors that occurred during recording.", "Počet chýb počas nahrávania.",
"Number of muxes found.", "Počet nájdených multiplexov.",
"Number of services.", "Počet služieb.",
"Number of tsfile tuners", "Počet tsfile tunerov",
"OK (partial)", "OK (čiastočne)",
"OSCam new pc-nodmx (rev >= 10389)", "OSCam nový pc-nodmx (rev >= 10389)",
"Odd key", "Nepárny kľúč",
"Odd key.", "Nepárny kľúč.",
"Off", "Vypnuté",
"Older OSCam", "Starší OSCam",
"On", "Zapnuté",
"On file removal", "Pri vymazaní súboru",
"Only Bulsatcom 39E", "Iba Bulsatcom 39E",
"Only EIT", "Iba EIT",
"Only OpenTV Sky Ausat", "Iba OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Iba OpenTV Sky Italia",
"Only OpenTV Sky UK", "Iba OpenTV Sky UK",
"Only PSIP (ATSC)", "Iba PSIP (ATSC)",
"Only UK Freesat", "Iba UK Freesat",
"Only UK Freeview", "Iba UK Freeview",
"Only Viasat Baltic", "Iba Viasat Baltic",
"Only digits", "Iba čísla",
"Only preferred CA PID", "Iba preferované CA PID",
"Only use specified DVB adapters (comma-separated, -1 = none)", "Používať iba tieto DVB adaptéry (čiarkou oddelené, -1 = žiadne)",
"OpenTV channel number", "Číslo kanálu OpenTV",
"Orbital position", "Orbitálna pozícia",
"Orbital positions", "Orbitálne pozície",
"Orbital positions.", "Orbitálne pozície.",
"Original network ID", "Originálne ID siete",
"Oscam mode.", "Mód Oscamu.",
"Over-the-air EPG grabber", "Over-the-air EPG graber",
"Over-the-air grabbers", "Over-the-air grabery",
"Override disabled", "Prepísanie zakázané",
"Override the frontend type.", "Prepísať typ frontendu.",
"Owner", "Vlastník",
"Owner of the entry.", "Vlastník položky.",
"Owner of the rule.", "Vlastník pravidla.",
"PLS code", "PLS kód",
"PLS mode", "PLS mód",
"Parent entry", "Nadradená položka",
"Parent entry.", "Nadradená položka.",
"Passthrough Muxer SI Tables", "SI tabuľky Passthrough Muxeru",
"Password", "Heslo",
"Password for the entry.", "Heslo pre položku.",
"Password2", "Heslo2",
"Path", "Cesta",
"Path used by the device.", "Cesta používaná zariadením.",
"Periodically save EPG to disk (hours)", "Periodicky ukladať EPG na disk (v hodinách)",
"Physiology", "Fyziológia",
"Polarization", "Polarizácia",
"Popular culture", "Populárna kultúra",
"Port to connect to.", "Port pre pripojenie.",
"Port to listen on or to connect to.", "Port kde sa má počúvať alebo ku ktorému sa pripojiť.",
"Position", "Pozícia",
"Position of the input.", "Pozícia vstupu.",
"Power save", "Šetrenie energiou",
"Pre-defined muxes", "Predefinované multiplexy",
"Pre-school children's programs", "Programy pre predškolákov",
"Preferred CA PID", "Preferované CA PID",
"Preferred audio language.", "Preferovaný jazyk audia.",
"Presentation", "Prezentácia",
"Press", "Tlač",
"Priority", "Priorita",
"Priority (-10..10)", "Priorita (-10..10)",
"Private", "Súkromné",
"Profile name", "Názov profilu",
"Provider", "Poskytovateľ",
"Provider ID", "ID poskytovateľa",
"Provider network name", "Názov siete poskytovateľa",
"Provider's network name.", "Názvy sietí poskytovateľov",
"Psychology", "Psychlógia",
"Quiz", "Kvíz",
"RAM only", "iba RAM",
"RTP/AVP/TCP (embedded data)", "RTP/AVP/TCP (vložené dáta)",
"RTSP Protocol", "RTSP Protokol",
"RTSP port (554 or 9983, 0 = disable)", "RTSP port (554 or 9983, 0 = vypnuté)",
"Radio", "Rádio",
"Raw Audio Stream", "Surový Audio Stream",
"Re-record%s%s", "Znova nahrať%s%s",
"Read error", "Chyba čítania",
"Record all", "Nahrať všetko",
"Record on these days only.", "Nahrávať iba v týchto dňoch.",
"Recording file options", "Možnosti súboru nahrávky",
"Recording file size.", "Veľkosť súboru nahrávky.",
"Recording play count", "Počet prehraní nahrávky",
"Reject", "Odmietnuť",
"Religion", "Náboženstvo",
"Religious", "Náboženské",
"Remove HTTP arguments", "Odstrániť HTTP argumenty",
"Remove all unsafe characters from filename", "Vynechať nebezpečné znaky z názvu súboru",
"Repeated", "Repríza",
"Replace whitespace in title with '-'", "Nahradiť medzeri v názve znakom '-'.",
"Replaces all whitespace in the title with '-'.", "Nahrádzať všetky medzeri v názve znakom  '-'.",
"Reports", "Správy",
"Reserved", "Rezervované",
"Resolution (height)", "Rozlíšenie (výška)",
"Restart", "Reštart",
"Restart on error", "Reštart pri chybe",
"Restrict connections to this type.", "Obmedziť pripojenia na tento typ.",
"Reuse EPG from", "Znovupoužiť EPG z",
"Reuse the EPG from another channel.", "Použiť EPG z iného kanála.",
"Rewrite EIT", "Prepísať EIT",
"Rewrite MPEG-TS SI tables", "Prepísať MPEG-TS SI tabuľky",
"Rewrite PAT", "Prepísať PAT",
"Rewrite PMT", "Prepísať PMT",
"Rewrite SDT", "Prepísať SDT",
"Rights", "Práva",
"Romance", "Romantika",
"Rotor type", "Typ motora",
"Run as group", "Spustiť pod skupinou",
"Run as user", "Spustiť pod uživateľom",
"Running", "Beží",
"SAT->IP frontend number.", "SAT->IP číslo frontend-u.",
"SAT>IP Client", "SAT>IP Klient",
"SAT>IP DVB-S Slave Frontend", "SAT>IP DVB-S Sekundárny Frontend",
"SAT>IP DVB-T frequency (Hz)", "SAT>IP DVB-T frekvencia (Hz)",
"SAT>IP Satellite Configuration", "SAT>IP Konfigurácia Satelitu",
"SAT>IP Server SI Tables", "SAT>IP Serverové SI Tabuľky",
"SAT>IP source number", "SAT>IP číslo zdroja",
"SD: standard definition", "SD: norma SD",
"SSL verify peer", "SSL verifikácia klienta",
"Sat", "Satelit",
"Satellite config", "Konfigurácia satelitu",
"Satellite longitude", "Zemepisná dĺžka satelitu",
"Satellite longitude.", "Zemepisná dĺžka satelitu.",
"Satellite position.", "Pozícia satelitu.",
"Satellite positions", "Pozície satelitov",
"Scan after creation", "Skenovať po vytvorení",
"Scan progress", "Pribeh skenovania",
"Scan result", "Výsledok skenovania",
"Scan status", "Stav skenovania",
"Scanfile", "Skenovanie zo súboru",
"Schedule a re-recording if more errors than (0=off)", "Naplánovať znova nahrávanie ak je chýb viac ako (0=vypnuté)",
"Schedule status", "Stav plánovania",
"Schedule status.", "Stav plánovania.",
"Scheduled Duration", "Nastavená Doba",
"Scheduled for recording", "Naplánované nahrávanie",
"School programs", "Programy pre školy",
"Science", "Veda",
"Script/program to run when a recording completes.", "Spustiť skript/program po dokončení nahrávania.",
"Script/program to run when a recording gets removed.", "Spustiť skript/program po odstránení nahrávky.",
"Season", "Príbeh",
"Season %d", "Príbeh č. %d",
"Season information (if available).", "Informácie o príbehu (ak sú dostupné).",
"Seek error", "Chyba pri posúvaní",
"Select an available network type for this tuner.", "Vybrať možný typ siete pre tento tuner.",
"Select high priority (default) EPG language.", "Vybrať jazyky EPG s najvyššou prioritou (predvolené).",
"Select how Tvheadend should handle muxes. See Help for details.", "Vyberte, akým spôsobom bude Tvheadend narábať s multiplexami. Podrobnejšie detaily nájdete v Nápovede.",
"Select low priority EPG language.", "Vybrať EPG jazyk s najnižšou prioritou.",
"Select medium priority EPG language.", "Vybrať jazyk so strednou prioritou.",
"Select subtitle codec to use for transcoding.", "Vybrať kodér titulkov pre prekódovanie.",
"Select the number of satellite positions supported by the SAT>IP hardware and your coaxial cable wiring.", "Vyberte počet podporovaných satelitných pozícií vašim SAT>IP hardvérom a vašim koaxiálnym káblom.",
"Select the path to use for DVB scan configuration files. Typically dvb-apps stores these in /usr/share/dvb/. Leave blank to use the internal file set.", "Vyberte cestu pre konfiguračné súbory DVB skenovania. Typicky dvb-apps toto ukladajú do /usr/share/dvb/. Nechajte prázdne pre použitie interných súborov.",
"Select the time offset for EIT events.", "Vybrať časový ofset pre udalosti EIT.",
"Select type of broadcast.", "Vyberte typ vysielania.",
"Select/Selected services to map.", "Vybrať/Vybrané služby pre mapovanie.",
"Send CAPMT OK query before descrambling.", "Poslať CAPMT OK požiadavku pred dekódovaním.",
"Send CAPMT query", "Zaslať požiadavku CAPMT",
"Send full PLAY cmd", "Zaslať plný príkaz PLAY",
"Serial number", "Sériové číslo",
"Series link", "Séria",
"Series link ID.", "ID Série",
"Serious", "Vážne",
"Serious music", "Vážna hudba",
"Server UUID", "UUID Servera",
"Server connectivity", "Pripojenie serveru",
"Server details.", "Detaily serveru.",
"Service", "Služba",
"Service ID", "ID služby",
"Service Mapper", "Mapovač služieb",
"Service Mapping (Map services to channels)", "Mapovanie služieb (Mapovanie služieb na kanály)",
"Service configuration", "Konfigurácia služby",
"Service count %zi", "Počet služieb %zi",
"Service mapping", "Mapovanie služby",
"Service name", "Názov služby",
"Service name found in EPG data.", "Názov služby nájdený v dátach EPG",
"Service type", "Typ služby",
"Services", "Služby",
"Services associated with the channel.", "Služby priradené s kanálom.",
"Services seen", "Videné služby",
"Set as default profile.", "Nastaviť ako predvolený profil.",
"Set the display name.", "Nastaviť zobrazenie mena.",
"Set the priority of this configuration.", "Nastaviť prioritu konfigurácie.",
"Settings", "Nastavenia",
"Show", "Zobraziť",
"Show / Game show", "Zábava / Zábavná hra",
"Show this page", "Zobraz túto stránku",
"Show version information", "Zobraziť informácie o verzii",
"Signal quality too poor", "Kvalita signálu je veľmi zlá",
"Signal scale (240 or 100)", "Škála signálu (240 alebo 100)",
"Site latitude", "Zemepisná šírka inštalácie",
"Site latitude.", "Zemepisná šírka inštalácie.",
"Site longitude", "Zemepisná dĺžka inštalácie",
"Site longitude.", "Zemepisná dĺžka inštalácie.",
"Size", "Veľkosť",
"Skip TS packets (0-200)", "Preskočiť TS pakety (0-200)",
"Skip commercials", "Preskočiť reklamy",
"Skip initial bytes", "Preskočit počiatočné bajty",
"Skip startup scan", "Preskočit počiatočný sken",
"Skip x number of transport packets.", "Preskočiť x transportných paketov.",
"Slave entry", "Podradená položka",
"Slave entry.", "Podradená položka.",
"Slot state", "Stav slotu",
"Soccer", "Futbal",
"Social", "Spoločnosť",
"Social / Political issues / Economics", "Spoločnosť / Politika / Ekonomika",
"Social advisory", "Sociálne poradenstvo",
"Socket or IP Address (when in TCP mode).", "Socket alebo IP adresa  (ak je v TCP móde).",
"Sort index", "Zoradiť index",
"Sort index.", "Zoradiť index.",
"Source", "Zdroj",
"Source deleted", "Zdroj vymazaný",
"Source quality is bad", "Kvalita príjmu je zlá",
"Source reconfigured", "Zdroj prekonfigurovaný",
"Source video codec", "Zdrojový video kodek",
"Southern hemisphere (latitude direction)", "Južná pologuľa (smer zemepisnej šírky)",
"Southern hemisphere (latitude direction).", "Južná pologuľa (smer zemepisnej šírky).",
"Special events (Olympic Games, World Cup, etc.)", "Špeciálne udalosti (Olympíjske Hry, Svetový pohár, atď.)",
"Specify alternative htsp port", "Určiť alternatívny htsp port",
"Specify alternative http port", "Určiť alternatívny http port",
"Specify alternative http webroot", "Určiť alternatívny http webroot",
"Specify extra htsp port", "Určiť extra htsp port",
"Specify the parameters to be changed. See Help for details.", "Určiť parametre ktoré budú zmenené. Viac informácií v Pomoci.",
"Spiritual sciences", "Duchovné vedy",
"Sports", "Šport",
"Sports magazines", "Športové programy",
"Standard", "Štandard",
"Start", "Začiatok",
"Start after", "Začať po",
"Start before", "Začať pred",
"Start time", "Čas začiatku",
"Status", "Stav",
"Status period (ms)", "Perióda stavu (ms)",
"Stop", "Zastaviť",
"Stop time", "Čas ukončenia",
"Storage path", "Cesta k úložisku",
"Storage space", "Miesto v úložisku",
"Stream ID", "ID Stream-u",
"Stream Profile", "Profil Streamu",
"Stream index", "Index streamu",
"Stream index (starts with zero).", "Index streamu (začína nulou)",
"Stream profile", "Profil streamu",
"Stream setup error", "Chyba nastavenia streamu",
"Stream type", "Typ stream-u",
"Streaming", "Streamovanie",
"Streaming Profile", "Profil streamovania",
"Streaming priority", "Priorita streamovania",
"Streaming profiles", "Profily streamu",
"Subdirectory options", "Možnosti podadresára",
"Subscribe to a service permanently", "Prihlásiť sa k trvalému odberu služby",
"Subscription", "Odber",
"Subscription does not exist", "Odber neexistuje",
"Subscription overridden", "Nahradený odber",
"Subscription weight", "Váha odberu",
"Subtitle", "Titulky",
"Subtitle codec", "Kodek tituliek",
"Subtitle of the program (if any) (display only).", "Titulky programu (ak existujú) (iba pre zobrazenie).",
"Sun", "Ne",
"Surround (2 front, rear mono)", "Surround (2 predné, zadné mono)",
"Switch before rotor", "Prepnúť pred nastavením motora",
"Switch frequency offset", "Prepnúť frekvenčný ofset",
"Switch to another service", "Prepnúť na inú službu",
"Switch type", "Typ prepínača",
"Switch: %s", "Prepínač: %s",
"Symbol rate (Sym/s)", "Symbolová rýchlosť (Sym/s)",
"Sync", "Synchronizácia",
"Sync + Don't keep", "Synchronizovať + Neudržiavať",
"System", "Systém",
"TCP Protocol", "TCP Protokol",
"TV Adapters - Linux ATSC-C Frontend", "TV Adaptéry - Linux ATSC-C Frontend",
"TV Adapters - Linux ATSC-T Frontend", "TV Adaptéry - Linux ATSC-T Frontend",
"TV Adapters - Linux DAB Frontend", "TV Adaptéry - Linux DAB Frontend",
"TV Adapters - Linux DVB-C Frontend", "TV Adaptéry - Linux DVB-C Frontend",
"TV Adapters - Linux DVB-S Frontend (Master)", "TV Adaptéry - Linux DVB-S Frontend (Hlavný)",
"TV Adapters - Linux DVB-S Slave Frontend", "TV Adaptéry - Linux DVB-S Vedľajší Frontend",
"TV Adapters - Linux DVB-T Frontend", "TV Adaptéry - Linux DVB-T Frontend",
"TV Adapters - Linux ISDB-C Frontend", "TV Adaptéry - Linux ISDB-C Frontend",
"TV Adapters - Linux ISDB-S Frontend (Master)", "TV Adaptéry - Linux ISDB-S Frontend (Hlavný)",
"TV Adapters - Linux ISDB-T Frontend", "TV Adaptéry - Linux ISDB-T Frontend",
"TV Adapters - SatConfig - 4-Port", "TV Adaptéry - SatConfig - 4-Port",
"TV Adapters - SatConfig - Advanced", "TV Adaptéry - SatConfig - Advanced",
"TV Adapters - SatConfig - DiseqC Rotor", "TV Adaptéry - SatConfig - DiseqC Rotor",
"TV Adapters - SatConfig - DiseqC Switch", "TV Adaptéry - SatConfig - DiseqC Switch",
"TV Adapters - SatConfig - EN50494/UniCable I", "TV Adaptéry - SatConfig - EN50494/UniCable I",
"TV Adapters - SatConfig - EN50607/UniCable II", "TV Adaptéry - SatConfig - EN50607/UniCable II",
"TV Adapters - SatConfig - GOTOX Rotor", "TV Adaptéry - SatConfig - GOTOX Rotor",
"TV Adapters - SatConfig - Tone Burst/2 Port", "TV Adaptéry - SatConfig - Tone Burst/2 Port",
"TV Adapters - SatConfig - USALS Rotor", "TV Adaptéry - SatConfig - USALS Rotor",
"TV Adapters - SatConfig - Universal LNB (Simple)", "TV Adaptéry - SatConfig - Universal LNB (Jednoduchý)",
"TVHDHomeRun Client", "TVHDHomeRun Klient",
"Tag files with metadata", "Tagovanie súborov metadatami",
"Tags", "Tagy",
"Talk show", "Diskusná relácia",
"Team sports (excluding football)", "Tímové športy (okrem futbalu)",
"Technology", "Technológia",
"Tennis", "Tenis",
"Testing", "Testovanie",
"Testing options", "Možnosti testovania",
"The ATSC source ID as set by the provider.", "ID zdroja pre ATSC ako bolo nastavené poskytovateľom.",
"The CA provider to compare. Leave blank to apply to all providers.", "CA poskytovateľ pre porovnanie. Nechajte prázdne pre všetkých poskytovateľov.",
"The CA stream type is only available for this filter.", "Typ CA stream je prípustný iba pre tento filter.",
"The CAID to compare. Leave blank to apply to all IDs.", "CAID pre porovnanie.  Nechajte prázdne pre všetky ID.",
"The CAM slot status.", "Stav CAM slotu.",
"The Conditional Access ID used for the service.", "ID podmieneného prístupu použité pre službu.",
"The Content reference identifier (CRID) authority.", "Content reference identifier (CRID) autorita.",
"The DVR profile to be used/used by the recording.", "DVR profil použitý/používaný pri nahrávaní.",
"The DVR profile to be used/used by this rule.", "DVR profil použitý/používaný týmto pravidlom.",
"The EPG ID used by the entry.", "EPG ID použité položkou.",
"The EPG grabber name.", "Názov EPG graberu.",
"The EPG grabber type.", "Typ EPG graberu.",
"The MPEGTS stream type is only available for this filter.", "Typ MPEGTS stream je prípustný iba pre tento filter.",
"The PIN to use.", "Použiť PIN.",
"The SAT>IP server's name.", "The SAT>IP názvy serverov.",
"The SAT>IP server's universally unique identifier.", "SAT>IP univerzálny unikátny identifikátor serverov.",
"The SAT>IP source number.", "SAT>IP číslo zdroja.",
"The SAT>IP's discovered IP address.", "SAT>IP's vyhľadaná IP adresa.",
"The URL to the playlist.", "URL k playlistu.",
"The channel this service is mapped to.", "Kanál ku ktorému je služba priradená.",
"The character encoding for this network (e.g. UTF-8).", "Kódovanie znakov pre túto sieť (napr. UTF-8).",
"The character encoding for this service (e.g. UTF-8).", "Kódovanie znakov pre túto službu (napr. UTF-8).",
"The current boot ID.", "Aktuálne boot ID.",
"The current configuration ID.", "Aktuálne ID konfigurácie.",
"The current configuration version.", "Verzia súčasnej konfigurácie.",
"The delivery system used by your cable provider.", "Systém doručovania používaný vašim káblovým poskytovateľom.",
"The delivery system used by your provider.", "Systém doručovania používaný vašim poskytovateľom.",
"The demux path used by the device.", "Cesta k demultiplexeru používaného zariadením.",
"The device ID.", "ID zariadenia.",
"The device path in sysfs filesystem (/sys).", "Cesta k zariadeniam v sysfs súborovom systéme (/sys).",
"The device's serial number.", "Sériové číslo zariadenia.",
"The frequency of the mux (in Hertz).", "Frequencia multiplexu (v Hertzoch).",
"The frequency of the mux/transponder in Hertz.", "Frequencia multiplexu/transpondéru v Hertzoch.",
"The frontend number given to the device.", "Číslo frontendu priradené k zariadeniu.",
"The input path used by the device.", "Vstupná cesta použitá zariadením.",
"The items have the following functions:", "Položky majú nasledovné funkcie:",
"The language to which the filter should apply.", "Jazyk pre ktorý bude aplikovaný filter.",
"The manufacturer of the SAT>IP server.", "Výrobca SAT>IP servera.",
"The maximum number of entries that can be matched.", "Maximálny počet položiek ktoré zodpovedajú pravidlu.",
"The maximum number of input streams allowed on this network.", "Maximálny počet vstupných streamov povolených na tejto sieti.",
"The maximum number of recording entries this rule can create.", "Maximálny počet nahrávok ktoré môže pravidlo vytvoriť.",
"The maximum number of recordings that can be scheduled.", "Maximálny počet nahrávok ktoré môžu byť naplánované.",
"The modulation used on the mux.", "Modulácia použitá pre multiplex.",
"The mux the service is on.", "Multiplex, ku ktorému je služba priradená.",
"The mux's universally unique identifier.", "Univerzálny unikátny identifikátor multiplexu.",
"The name (or frequency) the mux is on.", "Názov (alebo frekvencia) na ktorej je multiplex.",
"The name of the profile.", "Názov profilu.",
"The name of the the rule.", "Názov pravidla.",
"The network prefix(es) to block, e.g.192.168.2.0/24 (comma-separated list).", "Blokované siete, napr. 192.168.2.0/24 (list oddelený čiarkami).",
"The networks assigned to the device.", "Siete priradené k zariadeniu.",
"The networks using this configuration.", "Siete používajúce túto konfiguráciu.",
"The networks' universally unique identifier (UUID).", "Jednoznačný univerzálny identifikátor siete (UUID).",
"The number of muxes left to scan on this network.", "Počet multiplexov zostávajúcich na skenovanie na tejto sieti.",
"The orbital position of the satellite the mux is on.", "Orbitálna pozícia satelitu na ktorom je multiplex.",
"The playlist's character set.", "Znaková sada playlistu.",
"The polarization used on the mux.", "Polarizácia použitá multiplexom.",
"The provider's ID.", "ID poskytovateľov.",
"The provider's name.", "Názvy poksytovateľov.",
"The provider's network ID.", "ID sietí poskytovateľov.",
"The provider's network name.", "Názvy sietí poskytovateľov.",
"The satellite configuration to use.", "Použiť konfiguráciu satelitu.",
"The service ID", "ID služby",
"The service ID as set by the provider.", "ID služby ako ho nastavil poskytovateľ.",
"The service ID.", "ID služby.",
"The service name as set by the provider.", "Názov služby ako ju nastavil poskytovateľ.",
"The service's encryption status.", "Stav šifrovania služieb.",
"The start time of the recording.", "Počiatočný čas nahrávania.",
"The symbol rate used on the mux.", "Symbolová rýchlosť použitá multiplexom.",
"The symbol rate used on the mux/transponder.", "Symbolová rýchlosť použitá na multiplexe/transpodére.",
"The symbol rate.", "Symbolová rýchlosť.",
"The total number of services found.", "Celkový počet nájdených služieb.",
"The total scheduled duration.", "Celkový čas naplánovaného nahrávania.",
"The transponder ID.", "ID transpondéra.",
"Theme", "Téma",
"This tuner", "Tento tuner",
"Thread", "Vlákno",
"Thriller", "Triller",
"Thu", "Št",
"Time", "Čas",
"Time recording%s%s", "Časové nahrávanie%s%s",
"Time update", "Obnova času",
"Timeout (sec) (0=infinite)", "Timeout (sek) (0=nekonečne)",
"Timeout (seconds)", "Timeout (sekundy)",
"Timeout (secs)", "Timeout (sek.)",
"Title", "Názov",
"Title (regexp)", "Názov (regexp)",
"Title of the program (display only).", "Názov programu (len pre zobrazenie).",
"Title of the program.", "Názov programu.",
"Toggle details", "Prepnúť detaily",
"Too many data errors", "Príliš veľa chýb dát",
"Total number of services.", "Celkový počet služieb.",
"Tourism / Travel", "Turizmus / Cestovanie",
"Trace subsystems", "Subsystémy pre ladenie",
"Traditional arts", "Tradičné umenie",
"Traditional music", "Tradičná hudba",
"Transcode", "Prekódovať",
"Transcode/av-lib", "Prekódovať/av-lib",
"Transcoding", "Prekódovanie",
"Transponder ID", "ID transpondéra",
"Tue", "Ut",
"Tune before DiseqC", "Ladiť pred DiseqC",
"Tuner configuration", "Konfigurácia Tuneru",
"Tuner configuration.", "Konfigurácia Tuneru.",
"Tuning failed", "Ladenie neúspešné",
"Turn off LNB when idle", "Vypnúť LNB pri nečinnosti",
"Tvheadend server name", "Meno hostiteľa Tvheadend-u.",
"Type", "Typ",
"Type override", "Prepísanie typu",
"UDP RTP port number (2 ports)", "UDP RTP číslo portu (2 porty)",
"UPnP Protocol", "UPnP Protokol",
"URL of the M3U playlist.", "URL M3U playlistu.",
"Unable to open file", "Nemôžem otvoriť súbor",
"Unable to stat path", "Nemožem pristúpiť k ceste",
"Unicable I switch (universal LNB)", "Unicable I prepínač (univerzálne LNB)",
"Unicable type", "Typ Unicable",
"Universal LNB only", "Iba univerzálne LNB",
"Unknown", "Neznáme",
"Unknown file", "Neznámy súbor",
"Unknown reason (%i)", "Neznámy dôvod (%i)",
"Unlimited size", "Neobmedzená veľkosť",
"Unlimited time", "Neobmedzený čas",
"Unset (default)", "Nenastavené (predvolené)",
"Update card (EMM)", "Aktualizovať kartu (EMM)",
"Update channel icon", "Aktualizovať ikonu kanálu",
"Update channel name", "Aktualizovať názov kanálu",
"Update channel number", "Aktualizovať číslo kanálu",
"Update disabled", "Aktualizácia vypnutá",
"Update time", "Čas akutalizácie",
"Updated", "Aktualizované",
"Usage: %s [OPTIONS]\n", "Použitie: %s [PREPÍNAČE]\n",
"Use AC-3 detection.", "Použiť detekciu AC-3.",
"Use WEBM format.", "Použiť formát WEBM.",
"Use XSPF playlist instead of M3U", "Použiť XSPF playlist namiesto M3U",
"Use original", "Použiť originál",
"Use the provider's service IDs as channel numbers.", "Použiť čísla kanálov od poskytovateľa služby.",
"Use this character set when setting filenames.", "Použiť túto znakovú sadu pri nastavovaní mien súborov.",
"User access error", "Chyba prístupu uživateľa",
"User does not have access", "Užívateľ nemá prístup",
"User does not have access to channel", "Uživateľ nemá prístup ku kanálu",
"User language", "Jazyk uživateľa",
"User limit reached", "Dosiahnutý limit uživateľa",
"User login", "Prihlásenie uživateľa",
"Username", "Uživateľské meno",
"Users - Access Entries", "Uživatelia - Prístupové práva",
"Users - IP Blocking", "Uživatelia - Blokovanie IP",
"Users - Passwords", "Uživatelia - Heslá",
"Verify the SSL certificate.", "Overiť SSL certifikát.",
"Verify the peer's SSL.", "Overiť SSL klienta.",
"Video Stream Filter", "Filter Video Streamu",
"Video codec", "Video kodek",
"Video codec preset", "Prednastavený video kodek",
"Video codec preset to use for transcoding.", "Prednastavený video kodek pre prekódovanie.",
"Video recorder", "Nahrávanie videa",
"View all", "Zobraziť všetko",
"WEBM/built-in", "WEBM/vstavaný",
"WEBTV profile H264/AAC/MPEG-TS", "WEBTV profil H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "WEBTV profil H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "WEBTV profil VP8/Vorbis/WEBM",
"Waiting for program start", "Počkať na začiatok programu",
"Waiting for stream", "Čakanie na stream",
"War", "Vojna",
"Water sport", "Vodné športy",
"Weather report", "Informácie o počasí",
"Web User Interface", "Webové Uživateľské Rozhranie",
"Web interface", "Webové rozhranie",
"Web interface language", "Jazyk webového rozhrania",
"Web interface language.", "Jazyk webového rozhrania.",
"Web interface theme.", "Téma webového rozhrania.",
"Web theme", "Webová téma",
"Web user interface", "Webové uživateľské rozhranie",
"Wed", "St",
"Western hemisphere (latitude direction).", "Západná pologuľa (smer šírky).",
"Winter sports", "Zimné športy",
"Wizard", "Sprievodca",
"Yes", "Áno",
"addpids/delpids supported", "addpids/delpids podporované",
"default: nvenc(h264 / h265)", "predvolené: nvenc(h264 / h265)",
"fast: h264 / h265 / qsv(h264 / h265)", "rýchly: h264 / h265 / qsv(h264 / h265)",
"faster: h264 / h265 / qsv(h264)", "rýchlejší: h264 / h265 / qsv(h264)",
"hrs", "hod.",
"invalid option specified [%s]", "neplatná položka [%s]",
"logout", "odhlásiť",
"medium: h264 / h265 / qsv(h264 / h265)", "stredná: h264 / h265 / qsv(h264 / h265)",
"min", "minúta",
"mins", "minúty",
"option %s requires a value", "položka %s vyžaduje hodnotu",
"return", "návrat",
"slow: h264 / h265 / qsv(h264 / h265)", "pomalý: h264 / h265 / qsv(h264 / h265)",
"slower: h264 / h265 / qsv(h264)", "pomalší: h264 / h265 / qsv(h264)",
"superfast: h264 / h265", "super rýchly: h264 / h265",
"tvhdhomerun client", "tvhdhomerun klient",
"ultrafast: h264 / h265", "ultra rýchly: h264 / h265",
"veryfast: h264 / h265 / qsv(h264)", "veľmi rýchly: h264 / h265 / qsv(h264)",
"veryslow: h264 / h265 / qsv(h264)", "veľmi pomalý: h264 / h265 / qsv(h264)",
NULL, NULL
};

static const char *tvh_locale_31[] = {
NULL, NULL
};

static const char *tvh_locale_32[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nFör mer information var vänlig och besök Tvheadends webbplats:\nhttps://tvheadend.org\n",
"# Mapped channels", "# Mappade kanaler",
"# Muxes", "# Muxar",
"# Services", "# Servicar",
"(Default profile)", "(Standard profil)",
"(Read-only)", "(Skrivskyddat)",
"1 day", "1-dag",
"1 month", "1-månad",
"1 week", "1-vecka",
"1 year", "1-år",
"2 months", "2-månader",
"2 weeks", "2-veckor",
"2 years", "2-år",
"2-Port switch (universal LNB)", "2-port DiSEqC-switch (Universal-LNB Huvud)",
"3 days", "3-dagar",
"3 months", "3-månader",
"3 weeks", "3-veckor",
"3 years", "3-år",
"4-Port switch (universal LNB)", "4-port DiSEqC-switch (Universal-LNB Huvud)",
"5 days", "5-dagar",
"6 months", "6-månader",
"AAC audio", "AAC ljud",
"AC3 audio", "AC3 ljud",
"ACTIVE", "AKTIV",
"ANY", "NÅGON",
"Aborted by user", "Avbruten av användaren",
"Adapter in use by another subscription", "Adapter används av en annan subskription",
"Add file and line numbers to debug", "Lägg till fil och radnummer till fellog",
"Add the thread ID to debug", "Lägg till tråd-id till fellogg",
"Admin", "Administratör",
"Admin password", "Admin lösenord",
"Admin username", "Admin användare",
"Administrator login", "Administratörsinloggning",
"Adult movie", "Vuxenfilm",
"Advanced", "Avancerat",
"Adventure", "Äventyr",
"Advertisement / Shopping", "Reklam / Shopping",
"All (Streaming plus DVR)", "Alla (strömning samt DVR)",
"All lower-case", "Små bokstäver",
"All proceeds are used to support server infrastructure and buy test equipment.", "Alla bidrag används till drift av servrar och köp av testutrustning.",
"Alternate PID path", "Alternativ PID-sökväg",
"Alternate configuration path", "Alternativ konfigurationssökväg",
"Animals", "Djur",
"Any", "Någon",
"Arts", "Konst",
"Athletics", "Idrott",
"Avahi", "Avahi (Zeroconf)",
"Ballet", "Balett",
"Bandwidth", "Bandbredd",
"Based on software from", "Baserad på mjukvara från",
"Blue", "Blå",
"Brand", "Varumärke",
"Broadcast", "Utsändning",
"Broadcasting", "Utsändning",
"Build", "Bygge",
"CRASH", "KRASCH",
"Cache scheme", "Cacha schema",
"Cartoons", "Tecknat",
"Channel", "Kanal",
"Channel tag", "Kanal tagg",
"Channels", "Kanaler",
"Cinema", "Film",
"Class", "Klass",
"Classical", "Klassiskt",
"Classical music", "Klassisk musik",
"Clean effects", "Rensa effekter",
"Comedy", "Komedi",
"Comment", "Kommentar",
"Commercial break", "Reklampaus",
"Completed OK", "Klar OK",
"Configuration", "Konfiguration",
"Configuration ID", "Konfigurations-id",
"Configuration version", "Konfigurationsversion",
"Constellation", "Konstellation",
"Contest", "Tävling",
"Cooking", "Matlagning",
"Copy codec type", "Kopiera kodek-typ",
"Copy layout", "Kopiera layout",
"Created", "Skapad",
"Creator", "Skapare",
"Culture (without music)", "Kultur (utan musik)",
"Current affairs", "Aktuella frågor",
"DBus - use the session message bus instead of the system one", "D-Bus - använd sessionens meddelandebuss istället för systemets",
"DVB scan files path", "DVB sökväg skanna filer ",
"Dance", "Dans",
"Days of Week", "Veckodagar",
"Debate", "Debatt",
"Debug libav log", "Debugga libav log",
"Debug subsystems", "Debug undersystem",
"Debug to syslog", "Debug till syslog",
"Debug trace (low-level)", "Debug trace (lågnivå)",
"Default", "Standardläge",
"Default language(s)", "Standard språk",
"Description", "Beskrivning",
"Detective", "Detektiv",
"Device ID", "Enhets-id",
"Directory", "Katalog",
"Disable", "Inaktivera",
"Disable DVB bouquets", "Inaktivera DVB bouquets",
"Disable all access control checks", "Inaktivera alla accesskontrollkontroller",
"Disable device/firmware-specific workarounds", "Inaktivera enhet/firmware-specifika fixar",
"Disable syslog (all messages)", "Inaktivera syslog (alla meddelanden)",
"Discussion", "Diskussion",
"DiseqC repeats", "DiseqC upprepningar",
"Do not use", "Använd inte",
"Documentary", "Dokumentär",
"Don't backup configuration tree at upgrade", "Ta inte säkerhetskopia av konfigurationen vid uppgradering",
"Don't keep", "Spara inte",
"Economics", "Ekonomi",
"Education", "Utbildning",
"Educational", "Pedagogisk",
"Enable (auto)", "Aktivera (automatisk)",
"Enable DBus", "Aktivera D-Bus",
"Enable coredumps for daemon", "Aktivera minnesdumpar för daemon",
"Enable debug on stderr", "Aktivera felsökning till stderr",
"Enable debug subsystems", "Aktivera felsökningsundersystem",
"Enable debug to file", "Aktivera felsökning till fil",
"Enable debug to syslog", "Aktivera felsökning till syslog",
"Enable trace subsystems", "Aktivera spår-undersystem",
"Enabled", "Aktiverad",
"Encrypted", "Krypterad",
"Environment", "Miljö",
"Environment (pipe)", "Miljö (pipa)",
"Episode", "Avsnitt",
"Episode %d", "Avsnitt %d",
"Errors", "Fel",
"Every day", "Varje dag",
"Reserved", "Reserverad",
"Rewrite EIT", "Skriv om EIT",
"Rewrite PAT", "Skriv om PAT",
"Rewrite PMT", "Skriv om PMT",
"Rewrite SDT", "Skriv om SDT",
"Scheduled for recording", "Schemalagd för inspelning",
"Season", "Säsong",
"Season %d", "Säsong %d",
"Send full PLAY cmd", "Skicka fullt PLAY-kommando",
"Show version information", "Visa versions-information",
"Signal quality too poor", "För dålig signal-kvalité",
"Soap", "Såpa",
"Soccer", "Fotboll",
"Source", "Källa",
"Source deleted", "Källa raderad",
"Source quality is bad", "Källkvalité är dålig",
"Source reconfigured", "Källa omkonfigurerad",
"Specify User-Agent header for the http client", "Specificera User-Agent header för http-klienten",
"Specify alternative htsp port", "Specificera alternativ htsp-port",
"Specify alternative http port", "Specificera alternativ http-port",
"Specify alternative http webroot", "Specificera alternativ http webroot",
"Specify bind address", "Specificera adress att binda",
"Specify extra htsp port", "Specificera extra htsp-port",
"Stop", "Stoppa",
"Streaming", "Strömning",
"Subtitle", "Undertext",
"Sun", "Sön",
"Sync", "Synk",
"Sync + Don't keep", "Synk + spara inte",
"Tags", "Taggar",
"Testing options", "Testval",
"Thu", "Tors",
"Title", "Titel",
"Tue", "Tis",
"Unknown", "Okänd",
"Unknown reason (%i)", "Anledning okänd (%i)",
"Use Windows-compatible filenames", "Använd Windows-kompatibla filnamn",
"Username", "Användarnamn",
"Waiting for stream", "Väntar på ström",
"Wed", "Ons",
"invalid option specified [%s]", "ogiltigt val specificerat [%s]",
"mins", "minuter",
"option %s requires a value", "alternativ %s kräver ett värde",
NULL, NULL
};

static const char *tvh_locale_33[] = {
"1 day", "1 gün",
"1 month", "1 ay",
"1 week", "1 hafta",
"1 year", "1 yıl",
"Aborted by user", "Kullanıcı tarafından iptal edildi",
"Admin", "Yönetici",
"Admin password", "Yönetici parolası",
"Admin username", "Yönetici kullanıcı adı",
"Administrator login", "Yönetici girişi",
"Audio bitrate (kb/s) (0=auto)", "Ses bit hızı (kb/sn) (0=Otomatik)",
"Audio channel layout.", "Ses kanalı düzeni.",
"Auto check disabled", "Otomatik kontrol devre dışı",
"Auto check enabled", "Otomatik kontrol etkin",
"Bandwidth", "Bant genişliği",
"Basic", "Temel",
"Channel", "Kanal",
"Channel Tags", "Kanal Etiketleri",
"Channel does not exist", "Kanal yok",
"Channel icon", "Kanal Simgesi",
"Channel icon URL.", "Kanal simgesi URL'si.",
"Channel icon name scheme", "Kanal simgesi adı şeması",
"Channel name", "Kanal Adı",
"Channel not found", "Kanal bulunamadı",
"Channel number", "Kanal Numarası",
"Channel tag", "Kanal etiketi",
"Channel tag does not exist", "Kanal etiketi mevcut değil",
"Channel tags", "kanal etiketleri",
"Channels", "Kanallar",
"Character set", "Karakter takımı",
"Check availability", "Kullanılabilirliği kontrol et",
"Children's / Youth programs", "Çocuk / Gençlik Programları",
"Cinema", "Sinema",
"Class", "Sınıf",
"Classical", "Klasik",
"Classical music", "Klasik Müzik",
"Client name", "İstemci adı",
"Comedy", "Komedi",
"Commercial break", "Reklam arası",
"Configuration", "Yapılandırma",
"Configuration version", "Yapılandırma sürümü",
"Container", "Konteyner",
"Copy codec type", "Codec türünü kopyala",
"Copy layout", "Düzen kopyala",
"Count of objects", "Nesnelerin sayısı",
"Create SDTV/HDTV/Radio tags.", "SDTV/HDTV/Radyo etiketleri oluşturun.",
"Create a provider name tag.", "Bir sağlayıcı adı etiketi oluşturun.",
"Create tags", "Etiketler oluştur",
"Dance", "Dans",
"Default", "Varsayılan",
"Default language(s)", "Varsayılan dil(ler)",
"Default language.", "Varsayılan dil.",
"Default priority", "Varsayılan öncelik",
"Default user interface level.", "Varsayılan kullanıcı arabirimi düzeyi.",
"Digital Video Recorder", "Dijital Video Kaydedici",
"Education", "Eğitim",
"Educational", "Eğitimsel",
"Electronic Program Guide", "Elektronik Program Rehberi",
"Encrypted", "Şifreli",
"Encrypted service", "Şifreli hizmet",
"Enter an administrator password.", "Bir yönetici şifresi girin.",
"Failed", "Başarısız oldu",
"Fantasy", "Fantastik",
"File size", "Dosya boyutu",
"Filename", "Dosya adı",
"Folkloric", "Folklorik",
"Football", "Futbol",
"Frequency", "Frekans",
"Frequency (Hz)", "Frekans (Hz)",
"Frequency (MHz)", "Frekans (MHz)",
"Frequency (kHz)", "Frekans (kHz)",
"HD: high definition", "HD: yüksek çözünürlüklü",
"Horror", "Korku",
"Icon", "Simge",
"Ignore", "Görmezlikten gel",
"Information area", "Bilgi alanı",
"Informational", "Bilgilendirici",
"Initial scan", "İlk tarama",
"Interface", "Arayüz",
"Internal", "Dahili",
"Internal error", "İç hata",
"Interview", "Röportaj",
"Invalid", "Geçersiz",
"Invalid arguments", "Geçersiz argümanlar",
"Invalid file", "Geçersiz dosya",
"Invalid parameters", "Geçersiz parametreler",
"Invalid service", "Geçersiz hizmet",
"Invalid target", "Geçersiz hedef",
"Items", "Öğeler",
"Jazz", "Caz",
"LNB type", "LNB tipi",
"Language", "Dil",
"Language 1", "Dil 1",
"Language 2", "Dil 2",
"Language 3", "Dil 3",
"Language settings", "Dil ayarları",
"Languages", "Diller",
"Location", "Konum",
"Log", "Günlük",
"Login/Logout", "Giriş / Çıkış Yap",
"Low", "Düşük",
"Minimal channel number", "Minimum kanal numarası",
"Module", "Modül",
"Music", "Müzik",
"News", "Haber",
"Physiology", "Fizyoloji",
"Power save", "Güç tasarrufu",
"Romance", "Romantizm",
"Scan status", "Tarama durumu",
"Science", "Bilim",
"Science fiction", "Bilim kurgu",
"Server details.", "Sunucu ayrıntıları.",
"Wizard", "Sihirbaz",
NULL, NULL
};

static const char *tvh_locale_34[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\nДля більшої інформації зазирніть на веб-сторінку Tvheadend:\nhttps://tvheadend.org\n",
"# Mapped channels", "# Відображені Канали",
"# Muxes", "# Мультиплекси",
"# Services", "# Сервіси",
"# tune repeats", "# повтори тюнера",
"(Default profile)", "(Типовий Профіль)",
"(Read-only)", "(Тільки для читання)",
"1 day", "1 день",
"1 month", "1 місяць",
"1 week", "1 тиждень",
"1 year", "1 рік",
"2 months", "2 місяці",
"2 weeks", "2 тижні",
"2 years", "2 роки",
"2-Port switch (universal LNB)", "2-портовий перемикач (Універсальний LNB)",
"3 days", "3 дні",
"3 months", "3 місяці",
"3 weeks", "3 тижні",
"3 years", "3 роки",
"4-Port switch (universal LNB)", "4-портовий перемикач (універсальний LNB)",
"5 days", "5 днів",
"6 months", "6 місяців",
"A", "А",
"AA", "АА",
"AB", "АБ",
"AC-3 = descriptor 6", "AC-3 = дескриптор 6",
"AC-3 detection", "Виявлення AC-3",
"ACTIVE", "АКТИВНИЙ",
"ANY", "БУДЬ-ЯКИЙ",
"ATSC source ID", "ID джерела ATSC",
"ATSC-C Network", "Мережа ATSC-C",
"ATSC-C multiplex", "Мультиплекс ATSC-C",
"ATSC-T Network", "Мережа ATSC-T",
"ATSC-T multiplex", "Мультиплекс ATSC-T",
"AUTO", "АВТО",
"Aborted by user", "Перервано користувачем",
"Access", "Доступ",
"Action", "Дія",
"Adapter in use by another subscription", "Адаптер вже у використанні",
"Add file and line numbers to debug", "Додати файл і номери стрічок до відладки",
"Admin", "Адміністратор",
"Adult movie", "Фільм для дорослих",
"Advanced", "Просунутий",
"Advanced (non-universal LNBs, rotors, etc.)", "Просунутий (не універсальні LNB, ротори, етс.)",
"Adventure", "Пригоди",
"Advertisement / Shopping", "Реклама / Купівля",
"All (Streaming plus DVR)", "Все (передача потоку плюс DVR)",
"All proceeds are used to support server infrastructure and buy test equipment.", "Всі кошти використовуються для підтримки інфраструктури серверу і покупки тестового обладнання.",
"Alternate PID path", "Альтернативний шлях PID",
"Alternate configuration path", "Альтернативний шлях до конфігурації",
"Altitude (meters)", "Висота (метри)",
"Animals", "Тварини",
"Any", "Будь-що",
"Arts", "Мистецтво",
"Athletics", "Атлетика",
"Audio bitrate (kb/s) (0=auto)", "Шв. потоку аудіо (кб/с) (0=Авто)",
"Audio codec", "Кодек аудіо",
"Auto", "Авто",
"Auto check disabled", "Авто-перевірка вимкнена",
"Auto check enabled", "Авто-перевірка ввімкнута",
"Auto record", "Автозапис",
"Auto recording%s%s", "Авто запис%s%s",
"Auto time record", "Авто запис по часу ",
"Auto-Map to channels", "Авто-відображення до каналів",
"Automatic checking", "Автоматична перевірка",
"Automatically map EPG source", "Автоматично відображати джерело телепрограми",
"Automatically name from network", "Автоматична назва з мережі",
"Autorec maximum count (0=unlimited)", "Максимальна кількість автозаписів (0=необмежено)",
"B", "Б",
"BB", "ББ",
"Ballet", "Балет",
"Bandwidth", "Ширина каналу",
"Based on software from", "Основане на програмному забезпеченні від",
"Basic", "Базовий",
"Boot ID", "ID завантаження",
"Bouquet", "Букет каналів",
"Bouquet (auto)", "Букет каналів (авто)",
"Brand", "Марка",
"Broadcast", "Трансляція",
"Broadcasting", "Трансляція",
"Build", "Збірка",
"CA client %i", "CA Клієнт %i",
"CA identification", "CA Ідентифікація",
"CA provider", "CA Постачальник",
"CAPMT interval (ms)", "CAPMT Інтервал (мс)",
"CAPMT query interval (ms)", "CAPMT Інтервал запиту (мс)",
"Cache scheme", "Схема кешу",
"Camd.socket filename / IP Address (TCP mode)", "Назва файлу Camd.socket  / IP Адреса (режим TCP)",
"Cartoons", "Мультфільми",
"Channel", "Канал",
"Channel icon", "Іконка каналу",
"Channel name", "Назва каналу",
"Channel number", "Номер каналу",
"Channel number offset", "Зсув номеру каналу",
"Channel numbers from", "Номер каналу з",
"Channel tag", "Мітка каналу",
"Channel tag reference", "Довідка про мітку каналу",
"Channel tags", "Мітки каналу",
"Channels", "Канали",
"Character set", "Налаштування символів",
"Children's / Youth programs", "Дитячі / Молодіжні програми",
"Cinema", "Кіно",
"Class", "Клас",
"Classical", "Класика",
"Classical music", "Класична музика",
"Clean effects", "Ефекти очищення",
"Client name", "Назва клієнта",
"Comedy", "Комедія",
"Command delay time (ms) (10-200)", "Час затримки команди (мс) (10-200)",
"Command time (ms) (10-100)", "Час команди (мс) (10-100)",
"Comment", "Коментар",
"Commercial break", "Реклама",
"Committed", "Зроблено",
"Completed OK", "Завершено добре",
"Configuration", "Конфігурація",
"Configuration ID", "ID Конфігурації",
"Configuration name", "Назва конфігурації",
"Configuration version", "Версія конфігурації",
"Connection limit type", "Тип обмеження з’єднання",
"Constellation", "Сузір’я",
"Container", "Контейнер",
"Content character set", "Налаштування символів вмісту",
"Content type", "Тип вмісту",
"Contest", "Зміст",
"Cookie expiration (days)", "Тривалість життя тістечок (дні)",
"Cooking", "Куховарство",
"Copy codec type", "Тип кодека копіювання",
"Copy layout", "Скопіювати розкладку",
"Create bouquet", "Створити букет",
"Created", "Створено",
"Creator", "Творець",
"Cron multi-line", "Стрічки для cron ",
"Culture (without music)", "Культура (без музики)",
"Current affairs", "Поточні справи",
"DBus - use the session message bus instead of the system one", "DBus - використання буферу повідомлень сесії замість системного",
"DES key", "Ключ DES",
"DSCP/TOS for streaming", "DSCP/TOS для потоку",
"DVB scan files", "Файли сканування DVB",
"DVB scan files path", "Шлях до файлів сканування DVB",
"DVR behavior", "Поведінка DVR",
"DVR configuration", "Конфігурація DVR",
"DVR configuration profiles", "Профілі конфігурації DVR",
"Dance", "Танці",
"Data errors", "Помилка даних",
"Days of Week", "День тижня",
"Days of week", "День тижня",
"Debate", "Дебати",
"Debug options", "Опції відладки",
"Debug subsystems", "Відладка підсистем",
"Debug to syslog", "Передача журналу відладки до системного журналу",
"Debugging", "Відладка",
"Default", "Типово",
"Default language(s)", "Типова мова(и)",
"Default priority", "Типовий пріоритет",
"Delivery system", "Система доставки",
"Demux path", "Шлях до мультиплекса",
"Description", "Опис",
"Detective", "Детектив",
"Device ID", "ID пристрою",
"Device model", "Модель пристрою",
"Device path", "Шлях до пристрою",
"Directory", "Тека",
"Directory permissions (octal, e.g. 0775)", "Права доступу до теки (вісімкові, напр. 0775)",
"Disable", "Виключити",
"Disable DVB bouquets", "Виключити групи каналів DVB",
"Disable SAT>IP client", "Виключити клієнт SAT>IP",
"Disable syslog (all messages)", "Вимкнути syslog (всі повідомлення)",
"Disabled", "Вимкнено",
"Discussion", "Дискусія",
"DiseqC repeats", "DiseqC повтори",
"Do not use", "Не використовувати",
"Documentary", "Документальне",
"Don't backup configuration tree at upgrade", "Не створювати резервну копію дерева конфігурації при оновленні",
"Don't include title in filename", "Не додавати назву до назви файлу",
"Don't keep", "Не зберігати",
"Don't re-record", "Не перезаписувати",
"Don't reschedule", "Не переплановувати",
"Drama", "Драма",
"Duplicate handling", "Поводження з дублікатами",
"EIT - skip TSID check", "EIT - пропустити перевірку TSID",
"EPG name", "Назва ТБ-програми",
"EPG scan", "Сканування телепрограми",
"EPG source", "Джерело телепрограми",
"EPG update window", "Вікно оновлення телепрограми",
"Economics", "Економіка",
"Education", "Освіта",
"Educational", "Освітнє",
"Elementary stream filter", "Фільтр елементарного потоку",
"Enable (auto)", "Ввімкнути (авто)",
"Enable DBus", "Ввімкнути DBus",
"Enable NTP driver", "Ввімкнути драйвер NTP",
"Enable coredumps for daemon", "Включити створення дампів для демону",
"Enable debug subsystems", "Ввімкнути підсистеми відладки",
"Enable debug to file", "Ввімкнути відладку до файлу",
"Enable debug to syslog", "Ввімкнути відладку до системного журналу",
"Enable syslog", "Включити системний журнал",
"Enable trace subsystems", "Ввімкнути підсистеми відслідковування",
"Enabled", "Ввімкнено",
"Encrypted", "Зашифровано",
"Encrypted service", "Закритий сервіс",
"Entertainment programs for 10 to 16", "Розважальні програми для 10 до 16",
"Entertainment programs for 6 to 14", "Розважальні програми для 6 до 14",
"Environment", "Середовище",
"Environment (pipe)", "Середовище (pipe)",
"Episode", "Епізод",
"Episode %d", "Епізод %d",
"Error code", "Код помилки",
"Errors", "Помилки",
"Even key", "Парний ключ",
"Every day", "Кожен день",
"Exclude channel tags", "Виключити мітки каналу",
"Expeditions", "Експедиції",
"Experimental film", "Експериментальний філм",
"Expert", "Експертний",
"Exported tuners", "Експортовані тюнери",
"External", "Зовнішні",
"External EPG grabber", "Зовнішній захоплювач телепрограми",
"External IP (NAT)", "Зовнішній IP (NAT)",
"External URL", "Зовнішній URL",
"Extra arguments", "Додаткові аргументи",
"Extra start time", "Додатковий час початку",
"Extra stop time", "Додатковий час зупинки",
"Extra warming up time (seconds)", "Додатковий час ініціалізації (секунди)",
"FAIL", "НЕВДАЧА",
"FEC high", "Найвищий FEC",
"FEC low", "Найнижчий FEC",
"Fantasy", "Фантастика",
"File missing", "Пропущено файл",
"File not created", "Файл не створено",
"File permissions (octal, e.g. 0664)", "Права доступу до файлу (вісімкові, напр. 0664)",
"File size", "Розмір файлу",
"Filename", "Назва файлу",
"Filename character set", "Налаштування символів назви файлу",
"Filename options", "Опції назви файлу",
"Film", "Фільм",
"Fine arts", "Образотворче мистецтво",
"Finished", "Закінчено",
"Fitness and health", "Фітнес і здоров’я",
"Folk", "Народне",
"Folkloric", "Фольклор",
"Football", "Футбол",
"Force (auto)", "Вимусити (авто)",
"Force CA ID (e.g. 0x2600)", "Вимусити CA ID (e.g. 0x2600)",
"Force old status", "Вимусити старий статус",
"Force priority", "Вимусити пріоритет",
"Foreign countries", "Іноземні країни",
"Forever", "Завжди",
"Format string", "Формат стрічки",
"Frequency", "Частота",
"Frequency (Hz)", "Частота (Гц)",
"Frequency (MHz)", "Частота (МГц)",
"Frequency (kHz)", "Частота (кГц)",
"Fri", "П’ят",
"Friendly name", "Скорочена назва",
"Full DiseqC", "Повний DiseqC",
"Full-text", "Повний текст",
"Further education", "Додаткова освіта",
"GOTOX position", "GOTOX Позиція",
"Game show", "Ігрове шоу",
"Gardening", "Садівництво",
"General", "Загальне",
"General configuration", "Загальна конфігурація",
"Generic", "Загальне",
"Generic options", "Загальні опції",
"HD: high definition", "HD: Висока роздільна здатність",
"HTSP Default Stream Settings", "Типові налаштування потоку HTSP",
"Hierarchy", "Ієрархія",
"High", "Високий",
"Historical movie", "Історичне кіно",
"Horror", "Жахи",
"Hostname/IP", "Назва серверу/IP",
"IP address", "Адреса IP",
"IP blocking", "Блокування IP",
"IPTV Multiplex", "Мультиплекс IPTV",
"IPTV input", "Ввід IPTV",
"ISI (Stream ID)", "ISI (ID потоку)",
"Icon", "Іконка",
"Icon (full URL)", "Іконка (повна адреса URL)",
"Icon URL", "URL іконки",
"Icon base URL", "Головний URL іконки",
"Icon has title", "Іконка має назву",
"Icons from", "Іконка з",
"Idle", "\"Ліниве\" сканування",
"Idle scan", "\"Ліниве\" сканування",
"Idle scan muxes", "\"Ліниве\" сканування мультиплексів",
"If you'd like to support the project, please consider a donation.", "Якщо Ви бажаєте підтримати проект, то зробіть, будь ласка, пожертву.",
"Ignore EPG (EIT)", "Ігнорувати Телепрограму (EIT)",
"Ignore descriptor 5", "Ігнорувати дескриптор 5",
"Ignore invalid SSL certificate", "Ігнорувати неправильні сертифікати SSL",
"Ignore provider's channel numbers", "Ігнорувати номери каналів, наданих постачальником",
"Immediately abort", "Негайно перервати",
"Important", "Важливо",
"Include channel name in filename", "Додавати назву каналу в назву файлу",
"Include date in filename", "Додавати дату до назви файлу",
"Include episode in filename", "Додавати епізод до назви файлу",
"Include subtitle in filename", "Додавати субтитри в назву файлу",
"Include time in filename", "Додавати час до назви файлу",
"Index", "Індекс",
"Information area", "Інформаційне поле",
"Informational", "Інформаційно",
"Initial scan", "Початкове сканування",
"Input buffer (bytes)", "Вхідний буфер (байти)",
"Input path", "Шлях вводу",
"Interface", "Інтерфейс",
"Internal", "Внутрішнє",
"Internal EPG grabber", "Внутрішній захоплювач телепрограми",
"Internal grabber", "Внутрішній захоплювач",
"Interview", "Інтерв’ю",
"Invalid", "Неправильно",
"Invalid adapter number '%s'\n", "Неправильний номер адаптера '%s'\n",
"Invalid target", "Неправильна ціль",
"Jazz", "Джаз",
"Keep", "Зберегти",
"Kill signal (pipe)", "Вбити сигнал (трубка, pipe)",
"Kill timeout (pipe/secs)", "Затримка  вбивання сигналу (pipe/сек)",
"LNA (low noise amplifier)", "LNA (низько-шумний підсилювач)",
"LNB type", "Тип LNB",
"Language", "Мова",
"Language settings", "Налаштування мови",
"Languages", "Мови",
"Last seen", "Востаннє помічено",
"Last updated from", "Востаннє оновлювалося з",
"Limit connections", "Обмеження з'єднань",
"LinuxDVB network", "мережа LinuxDVB",
"Listen / Connect port", "Слухати / Порт з'єднання",
"Listen on IPv6", "Слухати на IPv6",
"Literature", "Література",
"Local bind IP address", "Локальна прив'язка адреси IP",
"Local channel number", "Локальний номер каналу",
"Local discovery IP address", "Локальний пошук адреси IP",
"Location", "Місцезнаходження",
"Log", "Журнал",
"Login/Logout", "Вхід/Вихід",
"Low", "Низький",
"MPEG-TS input", "Ввід MPEG-TS",
"Magazines", "Журнали",
"Maintained space", "Зайняте місце",
"Make subdirectories per channel", "Робити підтеки для кожного каналу",
"Make subdirectories per day", "Робити підтеки для днів",
"Make subdirectories per title", "Робити підтеки для кожної назви",
"Manufacturer", "Виробник",
"Manufacturer URL", "URL виробника",
"Map encrypted services", "Відображати закриті сервіси",
"Map radio channels", "Відображення радіо-каналів",
"Map unnamed channels", "Відображення неназваних каналів",
"Map zero-numbered channels", "Відображення 0-нумерованих каналів",
"Master tuner", "Головний тюнер",
"Matroska (mkv)/built-in", "Matroska (mkv)/вбудований",
"Matroska specific", "Специфічне для Matroska",
"Maximal channel number", "Максимальний номер каналу",
"Maximum duration", "Максимальна тривалість",
"Maximum length of PIDs", "Максимальна довжина PID'ів",
"Maximum period (mins)", "Максимальний період (хв)",
"Medicine", "Медицина",
"Melodrama", "Мелодрама",
"Minimal channel number", "Мінімальний номер каналу",
"Minimum duration", "Мінімальна тривалість",
"Minimum rotor time (seconds)", "Мінімальний час ротору (секунди)",
"Mode", "Режим",
"Model description", "Опис моделі",
"Model name", "Назва моделі",
"Model number", "Номер моделі",
"Modulation", "Модуляція",
"Module", "Модуль",
"Module ID", "ID модулю",
"Mon", "Пон",
"Mono", "Моно",
"More verbose libav log", "Більш розширений журнал libav",
"Movie", "Фільми",
"Music", "Музика",
"Musical", "Музичні",
"Mux", "Мультиплекс",
"NONE", "НІЧОГО",
"Name", "Назва",
"Names", "Назви",
"Natural sciences", "Природничі науки",
"Nature", "Природа",
"Network", "Мережа",
"Network A", "Мережа A",
"Network B", "Мережа B",
"Network ID (limit scanning)", "ID мережі (обмеження сканування)",
"Network UUID", "UUID мережі",
"Network discovery", "Пошук мережі",
"Network name", "Назва мережі",
"Network prefix", "Префікс мережі",
"Network type", "Тип мережі",
"Networks", "Мережі",
"News", "Новини",
"News magazine", "Журнали новин",
"Next tune delay in ms (0-2000)", "Наступна затримка настройки в мс (0-2000)",
"No", "Ні",
"No PIN", "Немає PIN",
"No access", "Немає доступу",
"No adapters specified!\n", "Не вказано адаптера!\n",
"No assigned adapters", "Немає асоційованого адаптера",
"No days", "Немає днів",
"No descrambler", "Немає розкодовувача",
"No free adapter", "Немає вільного адаптера",
"No input detected", "Не знайдено вхідного сигналу",
"No service assigned to channel", "Немає сервісу, асоційованого з каналом",
"No service enabled", "Немає ввімкнутих сервісів",
"No source available", "Немає доступних джерел",
"None", "Нічого",
"Normal", "Нормальне",
"Not set", "Не встановлено",
"Not set (none or channel configuration)", "Не встановлено (немає або конфігурація каналу)",
"Not set (use channel or DVR configuration)", "Не встановлено (використати конфігурацію каналу або DVR)",
"Number", "Номер",
"OK", "Добре",
"OK (partial)", "Добре (частково)",
"OSCam (rev >= 9095)", "OSCam (рев. >= 9095)",
"OSCam TCP (rev >= 9574)", "OSCam TCP (рев >= 9574)",
"OSCam net protocol (rev >= 10389)", "Протокол мережі OSCam (рев >= 10389)",
"OSCam new pc-nodmx (rev >= 10389)", "OSCam новий pc-nodmx (rev >= 10389)",
"OSCam pc-nodmx (rev >= 9756)", "OSCam pc-nodmx (рев >= 9756)",
"Odd key", "Непарний ключ",
"Off", "Викл.",
"Older OSCam", "Старіший OSCam",
"On", "Вкл.",
"On file removal", "При вилученні файлу",
"Only Bulsatcom 39E", "Тільки Bulsatcom 39E",
"Only EIT", "Тільки EIT",
"Only OpenTV Sky Ausat", "Тільки OpenTV Sky Ausat",
"Only OpenTV Sky Italia", "Тільки OpenTV Sky Italia",
"Only OpenTV Sky UK", "Тільки OpenTV Sky UK",
"Only UK Freesat", "Тільки UK Freesat",
"Only UK Freeview", "Тільки UK Freeview",
"Only Viasat Baltic", "Тільки Viasat Baltic",
"Only preferred CA PID", "Тільки улюблені CA PID",
"OpenTV channel number", "Номер каналу OpenTV",
"Opera", "Опера",
"Orbital position", "Орбітальна позиція",
"Orbital positions", "Орбітальні позиції",
"Original network ID", "Оригінальний ID мережі",
"Owner", "Власник",
"PLS code", "Код PLS",
"PLS mode", "Режим PLS",
"Password", "Пароль",
"Password2", "Пароль2",
"Path", "Шлях",
"Periodically save EPG to disk (hours)", "Періодично записувати телепрограму на диск (години)",
"Persistent user interface level", "Постійний рівень інтерфейсу користувача",
"Physiology", "Психологія",
"Polarization", "Поляризація",
"Political issues", "Політичні події",
"Pop", "Поп",
"Popular culture", "Популярна культура",
"Port", "Порт",
"Position", "Позиція",
"Post-processor command", "Команда після процесу",
"Post-remove command", "Команда після вилучення",
"Power save", "Збереження енергії",
"Preferred CA PID", "Улюблені PIDи CA",
"Preferred service video type", "Улюблений тип відео сервісу",
"Presentation", "Презентація",
"Press", "Преса",
"Priority", "Пріоритет",
"Priority (-10..10)", "Пріорітет (-10..10)",
"Private", "Приватне",
"Profile name", "Назва профілю",
"Provider", "Постачальник",
"Provider ID", "ID постачальника",
"Provider network name", "Назва мережі постачальника",
"Psychology", "Психологія",
"Puppets", "Ляльки",
"Quiz", "Вікторина",
"RAM only", "Тільки RAM",
"RTSP port", "Порт RTSP",
"RTSP port (554 or 9983, 0 = disable)", "Порт RTSP (554 або 9983, 0 = відключити)",
"Re-fetch period (hours)", "Період отримання даних (години)",
"Re-fetch period (mins)", "Період отримання даних (хв)",
"Re-record%s%s", "Перезапис%s%s",
"Re-try period (hours)", "Період поновної спроби (години)",
"Record all", "Записати все",
"Recording file options", "Опції файлу запису",
"Recording system path", "Системний шлях запису ",
"Reject", "Вийняти",
"Religion", "Релігія",
"Religious", "Рілігійні",
"Remarkable people", "Відомі люди",
"Remove HTTP arguments", "Вилучити аргументи HTTP",
"Reports", "Звіти",
"Rerun of", "Перезапустити",
"Reserved", "Зарезервовано",
"Resolution (height)", "Роздільна здатність (висота)",
"Restart on error", "Перезапустити при помилці",
"Reuse EPG from", "Використати телепрограму з",
"Rewrite EIT", "Перезаписати EIT",
"Rock", "Рок",
"Rolloff", "Крен",
"Romance", "Романс",
"Rotor initialization time (seconds)", "Час ініціалізації ротора (секунди)",
"Rotor type", "Тип ротору",
"Run as group", "Запустити як група",
"Run as user", "Запустити як користувач",
"Running", "Запущено",
"SD: standard definition", "HD: стандартна роздільна здатність",
"Sat", "Суб",
"Satellite longitude", "Довгота супутника",
"Satellite positions", "Позиція супутника",
"Scan after creation", "Сканування після створення",
"Scan result", "Результат сканування",
"Scan status", "Статус сканування",
"Schedule status", "Статус розкладу",
"Scheduled for recording", "Заплановано для запису",
"Scheduled start time", "Запланований час початку",
"Scheduled stop time", "Запланований час закінчення",
"School programs", "Шкільні програми",
"Science", "Наука",
"Science fiction", "Фантастика",
"Season", "Сезон",
"Season %d", "Сезон %d",
"Send CAPMT query", "Вислати запит CAPMT",
"Serial number", "Серійний номер",
"Serious", "Серйозне",
"Serious music", "Серйозна музика",
"Server", "Сервер",
"Server UUID", "Сервер UUID",
"Service", "Сервіс",
"Service ID", "ID сервісу",
"Service configuration", "Конфігурація сервісу",
"Service mapping", "Відображення сервісу",
"Service name", "Назва сервісу",
"Service type", "Тип сервісу",
"Services", "Сервіси",
"Settings", "Налаштування",
"Show", "Показати",
"Show this page", "Показати цю сторінку",
"Show version information", "Показати інформацію про версію",
"Signal quality too poor", "Якість сигналу дуже погана",
"Signal scale (240 or 100)", "Шкала сигналу (240 або 100)",
"Skip TS packets (0-200)", "Пропустити пакети TS (0-200)",
"Skip commercials", "Пропустити реклами",
"Skip initial bytes", "Пропустити початкові байти",
"Soap", "Мильні опери",
"Soccer", "Футбол",
"Social", "Соціальні",
"Social advisory", "Соціальні поради",
"Sort index", "Індекс сортування",
"Source", "Джерело",
"Source deleted", "Джерело вилучено",
"Source quality is bad", "Якість джерела погана",
"Source reconfigured", "Джерело переконфігуроване",
"Special events (Olympic Games, World Cup, etc.)", "Спеціальні події (Олімпійські ігри, Кубок світу, етс.)",
"Specify alternative htsp port", "Вказати альтернативний порт htsp",
"Specify alternative http port", "Вказати альтернативний порт http",
"Sports", "Спорт",
"Sports magazines", "Спортивні журнали",
"Squash", "Сквош",
"Standard", "Стандартний",
"Start", "Почати",
"Start after", "Почати після",
"Start before", "Почати перед",
"Start time", "Час початку",
"Status", "Статус",
"Status period (ms)", "Період перевірки статусу (мс)",
"Stereo", "Стерео",
"Stop", "Зупинити",
"Stop time", "Час зупинки",
"Storage path", "Шлях до місця зберігання",
"Storage space", "Розмір місця зберігання",
"Stream index", "Індекс потоку",
"Stream profile", "Профіль потоку",
"Stream type", "Тип потоку",
"Streaming", "Висилання потоку",
"Streaming priority", "Пріоритет потоку",
"Streaming profiles", "Профілі потоку",
"Subdirectory options", "Опції підтеки",
"Subscribe to a service permanently", "Підписатися до сервісу на постійно",
"Subscription overridden", "Підписка перезаписана",
"Subscription weight", "Вага підписки",
"Subtitle", "Субтитри",
"Subtitle codec", "Кодек субтитрів",
"Sun", "Нед",
"Switch before rotor", "Перемкнути перед ротору",
"Switch type", "Тип переключення",
"Switch: %s", "Переключення: %s",
"Symbol rate (Sym/s)", "Швидкість передачі (Символів/сек)",
"Sync", "Синхронізація",
"Sync + Don't keep", "Синхронізувати + Не зберігати",
"System", "Система",
"Tag files with metadata", "Файли міток з метаданими",
"Tags", "Мітки",
"Talk show", "Ток-шоу",
"Team sports (excluding football)", "Командний спорт (за виключенням футболу)",
"Technology", "Технологія",
"Tennis", "Теніс",
"Testing", "Тестування",
"Testing options", "Опції тестування",
"This tuner", "Цей тюнер",
"Thriller", "Трилер",
"Thu", "Чет",
"Time", "Час",
"Time missed", "Час пропущено",
"Time recording%s%s", "Час запису%s%s",
"Time update", "Оновлення часу",
"Timeout (sec) (0=infinite)", "Затримка (сек) (0=нескінченість)",
"Timeout (seconds)", "Затримка (секунди)",
"Timeshift", "Зсув часу",
"Title", "Назва",
"Title (regexp)", "Назва (регулярний вираз)",
"Tourism / Travel", "Туризм / Подорожі",
"Traditional arts", "Традиційне мистецтво",
"Traditional music", "Традиційна музика",
"Transcode/av-lib", "Транскодування/av-lib",
"Transcoding", "Транскодування",
"Transmission mode", "Режим передачі",
"Transponder ID", "ID транспондера",
"Transport stream ID", "ID транспортного потоку",
"Tue", "Вів",
"Tuner configuration", "Налаштування тюнера",
"Turn off LNB when idle", "Виключити LNB, якщо не використовується",
"Tvheadend server name", "Назва серверу Tvheadend",
"Type", "Тип",
"UDP RTP port number (2 ports)", "Номер порту UDP RTP (2 порти)",
"Universal LNB only", "Тільки універсальний LNB",
"Unknown", "Невідомо",
"Unknown reason (%i)", "Невідома причина (%i)",
"Unlimited size", "Необмежений розмір",
"Unlimited time", "Необмежений час",
"Unset (default)", "Не встановлено (типово)",
"Update card (EMM)", "Оновити карту (EMM)",
"Update channel icon", "Оновити іконки каналу",
"Update channel name", "Оновити назву каналу",
"Update channel number", "Оновити номер каналу",
"Update disabled", "Оновлення вимкнено",
"Update time", "Оновити час",
"Updated", "Оновлено",
"Usage: %s [OPTIONS]\n", "Використання: %s [ОПЦІЇ]\n",
"Use original", "Використовувати оригінал",
"Use service IDs as channel numbers", "Використовувати ID сервісу як номер каналу",
"User access error", "Помилка доступу користувача",
"User interface level", "Рівень інтерфейсу користувача",
"User language", "Мова користувача",
"User limit reached", "Досягнуто межу кількості користувачів",
"Username", "Назва користувача",
"Variety show", "Шоу варьєте",
"Video", "Відео",
"Video bitrate (kb/s) (0=auto)", "Шв. потоку відео (кб/с) (0=авто)",
"Video codec", "Кодек відео",
"Video recorder", "Записувач відео",
"WEBM/built-in", "WEBM/вбудоване",
"WEBTV profile H264/AAC/MPEG-TS", "Профіль WEBTV для H264/AAC/MPEG-TS",
"WEBTV profile H264/AAC/Matroska", "Профіль WEBTV для H264/AAC/Matroska",
"WEBTV profile VP8/Vorbis/WEBM", "Профіль WEBTV для VP8/Vorbis/WEBM",
"Waiting for program start", "Очікування на початок програми",
"Waiting for stream", "Очікування на потік",
"War", "Війна",
"Water sport", "Водні види спорту",
"Weak stream", "Слабий потік",
"Weather report", "Прогноз погоди",
"Web interface", "Веб-інтерфейс",
"Web interface language", "Мова веб-інтерфейсу",
"Web user interface", "Веб-інтерфейс користувача",
"Wed", "Сер",
"Western", "Вестерн",
"Yes", "Так",
"addpids/delpids supported", "addpids/delpids підтримується",
"hrs", "год",
"invalid option specified [%s]", "вказано неправильну опцію [%s]",
"min", "хв",
"mins", "хвилини",
"option %s requires a value", "опція %s вимагає значення",
"tvhdhomerun client", "клієнт tvhdhomerun",
NULL, NULL
};

static const char *tvh_locale_35[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\n\n欲想了解更多資訊，請瀏覽 Tvheadend 網站:\nhttps://tvheadend.org\n",
"# Mapped channels", "# 對應頻道",
"# Muxes", "# 多工器",
"# Services", "# 服務",
"(Default profile)", "(預設設定檔)",
"(Read-only)", "(唯讀)",
"1 day", "1日",
"1 month", "1個月",
"1 week", "1週",
"1 year", "1年",
"2 months", "2個月",
"2 weeks", "2週",
"2 years", "2年",
"2-Port switch (universal LNB)", "2埠交換器(universal LNB)",
"3 days", "3日",
"3 months", "3個月",
"3 weeks", "3週",
"3 years", "3年",
"4-Port switch (universal LNB)", "4埠交換器(universal LNB)",
"5 days", "5日",
"6 months", "6個月",
"AC-3 detection", "AC-3檢查",
"AC3 audio", "AC3音軌",
"ACTIVE", "啟動",
"ANY", "任意",
"ATSC", "ATSC制式",
"ATSC PSIP EPG", "ATSC制式封裝EPG",
"ATSC SI Tables", "ATSC制式 DVB SI表",
"ATSC source ID", "ATSC制式源ID",
"ATSC-C", "ATSC-C制式",
"ATSC-C Network", "ATSC-C制式網路",
"ATSC-C multiplex", "ATSC-C制式多工",
"ATSC-T", "ATSC-T制式",
"ATSC-T Network", "ATSC-T網路",
"ATSC-T multiplex", "ATSC-T制式多工",
"AUTO", "自動",
"Aborted by user", "使用者取消",
"Access", "控制",
"Access (ACL)", "ACL控制清單",
"Action", "啟動",
"Active", "啟動",
"Adapter in use by another subscription", "適配器正在被使用",
"Additional service names found in EPG data.", "從EPG資料中追加服務名",
"Admin", "管理員",
"Admin password", "管理員密碼",
"Admin username", "管理員使用者名稱",
"Administrator login", "管理員登入",
"Adult movie", "成人節目",
"Advanced", "進階",
"Advanced (non-universal LNBs, rotors, etc.)", "進階(非一般的LNBs設定、極軸等",
"After creating the network scan it for services.", "在掃描服務前建立網路",
NULL, NULL
};

static const char *tvh_locale_36[] = {
"\nFor more information please visit the Tvheadend website:\nhttps://tvheadend.org\n", "\n请访问Tvheadend网站以获取更多信息。\nhttps://tvheadend.org\n",
"# Mapped channels", "已匹配频道",
"# Muxes", "复用器",
"# Services", "服务",
"# tune repeats", "调谐重试次数",
"(Default profile)", "(默认策略)",
"(Read-only)", "(只读)",
"1 day", "1天",
"1 month", "1月",
"1 week", "1周",
"1 year", "1年",
"1316 bytes", "1316字节",
"2 months", "2个月",
"2 weeks", "2周",
"2 years", "2年",
"2-Port switch (universal LNB)", "2端口切换器（高频头）",
"3 days", "3天",
"3 months", "3个月",
"3 weeks", "3周",
"3 years", "3年",
"4-Port switch (universal LNB)", "4端口切换器(高频头)",
"5 days", "5天",
"6 months", "6个月",
"A channel tag (e.g. a group of channels) to which this rule applies.", "运用此规则的频道标签（频道组）",
"AAC audio", "AAC音轨",
"AC-3 = descriptor 6", "AC-3 = 描述符6",
"AC-3 detection", "AC-3检查",
"AC3 audio", "AC3音轨",
"ACTIVE", "激活",
"ANY", "任意",
"API", "接口",
"ATSC", "ATSC制式",
"ATSC PSIP EPG", "ATSC制式封装EPG",
"ATSC SI Tables", "ATSC制式 DVB SI表",
"ATSC source ID", "ATSC制式源ID",
"ATSC-C", "ATSC-C制式",
"ATSC-C Network", "ATSC-C制式网络",
"ATSC-C multiplex", "ATSC-C制式复用",
"ATSC-T", "ATSC-T制式",
"ATSC-T Network", "ATSC-T网络",
"ATSC-T multiplex", "ATSC-T制式复用",
"AUTO", "自动",
"Aborted by user", "用户取消",
"Accept remote subscription weight", "接受远程订阅权重",
"Accept the remote subscription weight (from the SAT>IP client).", "接受远程订阅权重（从SAT>IP客户端）",
"Accept zero value for TSID", "接受0值TSID",
"Access", "控制",
"Access (ACL)", "ACL控制列表",
"Action", "启动",
"Active", "激活",
"Adapter in use by another subscription", "适配器正在被使用",
"Add file and line numbers to debug", "添加文件及行号以调试",
"Add the thread ID to debug", "添加线程ID以调试",
"Add this value to PTS for the teletext subtitles. The time value is in milliseconds and may be negative.", "将此值为teletext字幕添加到PTS，时间数值单位是毫秒，并可以为负数。",
"Additional arguments to pass to the grabber.", "传递给抓取器的额外变量",
"Additional service names found in EPG data.", "从EPG数据中追加服务名",
"Additional time (in seconds) in which to get the tuner ready for recording. This is useful for those with tuners that take some time to tune and/or send garbage data at the beginning. ", "使调谐器准备好录制所需的额外的时间（秒）。这对某些调谐器在调谐时需要一些时间，或在开始时发送一些垃圾数据时有用。",
"Admin", "管理员",
"Admin password", "管理员密码",
"Admin username", "管理员用户名",
"Administrator login", "管理员登录",
"Adult movie", "成人节目",
"Advanced", "高级",
"Advanced (non-universal LNBs, rotors, etc.)", "高级(非常规的LNBs设置、极轴等)",
"Adventure", "冒险",
"Advertisement / Shopping", "广告/购物",
"After creating the network scan it for services.", "在扫描服务前创建网络",
"All characters that could possibly cause problems for filenaming will be replaced with an underscore. See Help for details.", "可能在文件命名时造成问题的所有字符将会被替换为下划线。查看帮助以获取详情。",
"All lower-case", "所有小写",
"All: Record if different description", "全部：描述不同则录制",
"All: Record if different episode number", "全部：集数不同则录制",
"All: Record if different subtitle", "全部：副标题不同则录制",
"All: Record once per day", "全部：一天录制一次",
"All: Record once per month", "全部：一月录制一次",
"All: Record once per week", "全部：一周录制一次",
"Allow all PIDs", "允许所有PIDs",
"Allow all PIDs (no filter) when the 'Maximum PIDs' limit is reached.", "当达到“最大PIDs”限制时允许所有PIDs（无过滤）。",
"Allow high bitrate mode (CI+ CAMs only).", "允许高比特率模式（仅CI+CAMs）。",
"Allow idle scan tuning on this device.", "允许此设备空闲时扫描。",
"Allow/Disallow access to the 'Configuration' tab.", "允许或禁止访问配置功能",
"Allow/Disallow web interface access (this  includes access to the EPG).", "允许或禁止web访问，(该选项包含访问EPG数据)",
"Allowed network", "允许的网络",
"Allowed networks", "允许的网络",
"An event which starts between this \"start after\" and \"start before\" will be matched (including boundary values).", "所有节目在“最晚开始”和“最早开始”之间的将会匹配(包含边界值)。",
"Animals", "动物",
"Any", "任意",
"Arts", "艺术",
"Arts / Culture (without music)", "艺术/文化(不含音乐)",
"Arts magazines", "艺术杂志",
"Athletics", "运动员",
"Authenticated user", "认证用户",
"Auto", "自动",
"Auto check disabled", "关闭自动检查",
"Auto check enabled", "启动自动检查",
"Auto record", "自动录制",
"Auto record caption", "自动录制说明",
"Auto recording%s%s", "自动录制%s%s",
"Auto time record", "自动时间录制",
"Automatic recording caption.", "自动录制说明",
"Automatically map EPG source", "自动匹配EPG源",
"Automatically record.", "自动录制",
"Autorec maximum count (0=unlimited)", "自动录制最大计数(0为无限)",
"Autorec maximum schedules limit (0=unlimited)", "自动录制最大计划限制(0=无限)",
"Bad", "错误",
"Bad request", "错误请求",
"Ballet", "芭蕾",
"Bandwidth", "带宽",
"Base DVB SI Tables (PAT,CAT,PMT,SDT etc.)", "DVB业务信息表(PAT,CAT,PMT,SDT等)",
"Bind to specific local IP address.", "绑定到本地IP地址",
"Brand", "商标",
"Branding information (if available).", "商标信息(如果有)。",
"Broadcast", "广播",
"Broadcast type", "广播类型",
"Broadcast.", "广播。",
"Broadcasting", "正在广播",
"Build", "编译",
"CA (descrambling) Client", "CA（解扰）客户端",
"CA Stream Filter", "CA流过滤",
"CAPMT (Linux Network DVBAPI)", "CAPMT（DVB-API）方式",
"CAPMT CA Client", "CAPMT CA 客户端",
"CI Module", "CI 模块",
"CRASH", "崩溃",
"CRID authority", "权威CRID",
"CSA (descrambling)", "CSA（解扰）",
"CWC CA Client", "CWC CA 客户端",
"Cache scheme", "缓存方案",
"Cartoons", "动画",
"Channel", "频道",
"Channel Tags", "频道标记",
"Channel does not exist", "频道不存在",
"Channel icon", "频道图标",
"Channel icon URL.", "频道图标URL地址",
"Channel icon as defined in EPG data.", "采用EPG数据作为频道默认图标",
"Channel icon name scheme", "频道图标名称方案",
"Channel icon path", "频道图标路径",
"Channel mapping options", "频道匹配操作",
"Channel name", "频道名",
"Channel not found", "未匹配频道",
"Channel number", "频道号",
"Channel number as defined in EPG data.", "采用EPG数据作为默认频道号",
"Channel number range", "频道号排列",
"Channel tag", "频道标记",
"Channel tag does not exist", "频道标记不存在",
"Channel tags", "频道标记",
"Channel to use/used for the recording.", "录像将/已使用的频道",
"Channels", "频道",
"Character set", "字符集设置",
"Characters not supported in Windows filenames (e.g. for an SMB/CIFS share) will be stripped out or converted.", "在Windows中不支持的字符(例如SMB/CIFS共享)将被移除或转换。",
"Charset", "字符集",
"Children's / Youth Programs", "儿童/青年节目",
"Children's / Youth programs", "儿童/青年节目",
"Cinema", "电影院",
"Classical", "古典",
"Classical music", "古典音乐",
"Client name", "客户端名",
"Clone scheduled entry on error", "出错时克隆计划",
"Comedy", "喜剧",
"Comment", "注释",
"Commercial break", "广告休息",
"Commercials will be dropped from the recordings. Commercial detection works using EITp/f (EPG running state) and for the Swedish channel TV4 (using teletext info).", "广告将被从录像中丢弃。广告检测使用EITp/f(EPG运行状态)工作，对于瑞典TV4频道(使用teletext信息)。",
"Completed OK", "完成OK",
"Configuration name", "设置名称",
"Content type", "内容类型",
"Content type.", "内容类型",
"Contest", "比赛",
"Continue recording for x minutes after scheduled stop time", "在计划停止时间后再录制x分钟",
"Continue recording for x minutes after scheduled stop time.", "在计划停止时间后再录制x分钟。",
"Cooking", "厨艺",
"Create a directory per channel when storing recordings. If both this and the 'directory per day' checkbox is enabled, the date-directory will be the parent of the per-channel directory.", "在储存录像时为每个频道创建目录。如果此选项与“每天新建目录”都被选中，日期目录将为频道目录的父目录。",
"Create a directory per title when storing recordings. If the day/channel directory checkboxes are also enabled, those directories will be parents of this directory.", "在储存录像时为每个节目标题创建目录。如果日期、频道目录都被选中，这些目录将为此目录的父目录。",
"Create a new directory per day in the recording system path. Folders will only be created when something is recorded. The format of the directory will be ISO standard YYYY-MM-DD.", "在录像系统中每天新建目录。只有在有节目被录制时创建。目录名称是ISO标准的YYYY-MM-DD。",
"Create directories using these permissions.", "使用这些权限创建目录。",
"Create files using these permissions.", "使用这些权限创建文件。",
"Create tags in recordings using media containers that support metadata (if possible).", "为录像创建标签，使用支持元数据的封装器(如果可能)。",
"Creator", "创建者",
"Culture (without music)", "文化(不含音乐)",
"Culture magazines", "文化杂志",
"Current affairs", "时事",
"DVR - Auto-recording (Autorecs)", "DVR - 自动录像 (Autorecs)",
"DVR - Profiles", "DVR - 策略",
"DVR - Time-based Recording (Timers)", "DVR - 基于时间的自动录像 (Timers)",
"DVR behavior", "DVR动作",
"DVR configuration", "DVR设置",
"DVR file retention period", "DVR文件保留时间",
"DVR log retention", "DVR日志保留",
"DVR log retention period", "DVR日志保留时间",
"DVR profile to use/used for the recording.", "录像已/将使用的DVR策略",
"Dance", "舞蹈",
"Data errors", "日期错误",
"Days of Week", "周几",
"Days of the week to which the rule should apply.", "规则将应用到周几。",
"Days of week", "周几",
"Debate", "辩论",
"Description", "描述",
"Detective", "侦探",
"Digital Video Recorder", "录像机",
"Directory", "目录",
"Directory override. Override the subdirectory rules specified by the DVR configuration and put all recordings done by this entry into the specified subdirectory", "目录覆盖。覆盖DVR设置中的子目录规则，并使用此条目将所有完成的录像放入指定的子目录。",
"Directory permissions (octal, e.g. 0775)", "目录权限(八进制，例：0775)",
"Directory used by the entry.", "此条目使用的目录",
"Discussion", "讨论",
"Documentary", "纪录",
"Don't include the title in the filename.", "在文件名中不包含节目标题。",
"Don't include title in filename", "在文件名中不包含节目标题。",
"Don't keep", "不保持",
"Don't re-record", "不录制",
"Don't re-record if recording fails.", "录制失败时不重录。",
"Don't re-schedule if recording fails.", "录制失败时不重新计划。",
"Don't reschedule", "不重新计划",
"Drama", "戏剧",
"Duplicate handling", "复制操作",
"Duplicate recording handling.", "复制录像操作",
"EPG update window", "EPG更新窗口",
"Economics", "经济",
"Education", "教育",
"Education / Science / Factual topics", "教育/科学/事实主题",
"Educational", "教育",
"Enable/disable auto-rec rule.", "启用/禁用自动录制规则。",
"Enable/disable profile.", "启用/禁用策略。",
"Enable/disable the entry.", "启用/禁用记录。",
"Enabled", "启用",
"Entertainment programs for 10 to 16", "10-16岁教育节目",
"Entertainment programs for 6 to 14", "6-14岁教育节目",
"Environment", "环境",
"Episode", "集",
"Episode %d", "集 %d",
"Episode number/ID.", "集数/集ID",
"Equestrian", "马术",
"Error code", "错误代码",
"Error code of entry.", "记录错误代码",
"Errors", "错误",
"Every day", "每天",
"Extra start time", "额外启动时间",
"Extra stop time", "额外停止时间",
"Extra warming up time (seconds)", "额外暖机时间(秒)",
"File missing", "文件丢失",
"File not created", "文件未创建",
"File permissions (octal, e.g. 0664)", "文件权限(八进制，例 0664)",
"File removed", "文件已删除",
"File size", "文件大小",
"Filename", "文件名",
"Filename character set", "文件名字符集",
"Filename options", "文件名选项",
"Filename used by the entry.", "此记录使用的文件名。",
"Finished", "已完成",
"Forever", "永远",
"Format string", "格式字符串",
"Free-form field, enter whatever you like here.", "无格式区，写你所想。",
"Free-form text field, enter whatever you like here.", "无格式文本区，写你所想。",
"Fri", "周五",
"Full pathname specification", "定义完成路径",
"Full-text", "完整文本",
"High", "高",
"If an error occurs clone the scheduled entry and try to record again (if possible).", "如果出现错误，则克隆已计划的记录并尝试重新录制(如果可能)。",
"If more than x errors occur during a recording schedule a re-record (if possible).", "如果在录制中出现多于x个错误，则计划一个重新录制(如果可能)。",
"Important", "重要",
"Include channel name in filename", "文件名中包含频道名",
"Include date in filename", "文件名中包含日期",
"Include episode in filename", "文件名中包含集数",
"Include subtitle in filename", "在文件名中不包含节目副标题",
"Include the date for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "在录像的记录标题中添加日期。应用于文件中保存的标题和文件名。",
"Include the episode subtitle in the title (if available).", "在标题中包含集副标题(如果有)。",
"Include the name of the channel in the event title. This applies to both the title stored in the file and to the filename itself.", "在记录标题中包含频道名。应用于文件中保存的标题和文件名。",
"Include the season and episode in the title (if available).", "在标题中包含季和集(如果有)。",
"Include the time for the recording in the event title. This applies to both the title stored in the file and to the filename itself.", "在记录中包含录像时间。应用于文件中保存的标题和文件名。",
"Include time in filename", "在文件名中包含时间",
"Invalid", "非法",
"Keep x amount of storage space free.", "保证磁盘中有x的剩余空间。",
"Last played position", "上次播放位置",
"Last played position when the recording isn't fully watched yet.", "录像未完全观看时的上次播放位置。",
"Local: Record if different description", "本地：描述不同则录制",
"Local: Record if different episode number", "本地：集数不同则录制",
"Local: Record if different subtitle", "本地：副标题不同则录制",
"Local: Record if different title", "本地：标题不同则录制",
"Local: Record once per day", "本地：一天录制一次",
"Local: Record once per month", "本地：一月录制一次",
"Local: Record once per week", "本地：一周录制一次",
"Low", "低",
"Maintain free storage space in MiB", "保持剩余空间，MB",
"Maintain used storage space in MiB (0=disabled)", "保持已用磁盘空间(MB，0=禁用)",
"Maintained space", "保持的空间",
"Make subdirectories per channel", "为每个频道新建子目录",
"Make subdirectories per day", "每天新建子目录",
"Make subdirectories per title", "为每个标题新建子目录",
"Maximum allowed difference between event start time when the EPG event is changed.", "在EPG记录变更时，最多允许的开始时间差。",
"Maximum count (0=default)", "最大计数(0=默认)",
"Maximum duration", "最大持续时间",
"Maximum schedules limit (0=default)", "最大计划限制 (0=默认)",
"Minimum duration", "最小持续时间",
"Mon", "周一",
"Name", "名称",
"Name of channel the entry recorded from.", "此记录从哪个频道录制。",
"Name of the entry.", "记录名。",
"Name of the profile.", "记录策略。",
"New / premiere / unknown", "新/首映/未知",
"No days", "无日期",
"Normal", "正常",
"Not set", "未设置",
"Not set (none or channel configuration)", "未设置(无，或频道设置)",
"Not set (use channel or DVR configuration)", "未设置(使用频道或DVR设置)",
"Number of days to keep the file.", "保存文件的天数。",
"Number of days to keep the recorded file.", "保存已录制文件的天数。",
"Number of days to keep the recorded files.", "保存已录制文件的天数。",
"Number of days to retain entry information.", "保留记录信息的天数。",
"Number of days to retain information about recording.", "保留录像信息的天数。",
"Number of days to retain information about recordings. Once this period is exceeded, duplicate detection will not be possible anymore.", "保留录像信息的天数。当到限时，复制检测将无效。",
"Number of errors that occurred during recording (Transport errors).", "录制中出现的错误(传输错误)。",
"Number of errors that occurred during recording.", "录制中出现的错误。",
"Number of times this recording was played.", "此录像播放的次数。",
"On file removal", "在文件删除时",
"Owner", "所有者",
"Owner of the entry.", "记录的所有者。",
"Owner of the rule.", "规则的所有者。",
"Parent entry", "父记录",
"Parent entry.", "父记录。",
"Path where the recordings are stored. If components of the path do not exist, Tvheadend will try to create them.", "录像文件存储路径。如果路径中有不存在的要素，TVheadend将创建它们。",
"Post-processor command", "后置处理器命令",
"Post-recording padding", "延长录制",
"Post-remove command", "删除后的命令",
"Pre-processor command", "前置处理器命令",
"Pre-recording padding", "提前录制",
"Priority", "优先级",
"Priority of the entry, higher-priority entries will take precedence and cancel lower-priority events.", "记录的优先级，高优先级记录将优先于并取消低优先级记录。",
"Priority of the recording. Higher priority entries will take precedence and cancel lower-priority events.", "记录的优先级，高优先级记录将优先于并取消低优先级记录。",
"Program synopsis (display only).", "节目摘要(仅显示)。",
"Program synopsis.", "节目摘要",
"Re-record%s%s", "重录制%s%s",
"Record all", "全部录制",
"Record on these days only.", "只在这些日期录制。",
"Recording file options", "录像文件选项",
"Recording file size.", "录像文件大小。",
"Recording play count", "录像播放计数",
"Recording system path", "录像系统路径",
"Remove all unsafe characters from filename", "从文件名中移除所有不安全字符",
"Repeated", "重复",
"Replace whitespace in title with '-'", "在标题中以'-'代替空格",
"Replaces all whitespace in the title with '-'.", "在标题中将所有'-'以空格代替。",
"Rerun of", "重播",
"Running", "正在运行",
"SAT>IP DVB-T Frontend", "SAT转IP DVB-T Frontend前端",
"SAT>IP DVB-T frequency (Hz)", "SAT转IP DVB-T 频率(单位Hz)",
"SAT>IP RTSP port number for server\n(default: -1 = disable, 0 = webconfig, standard port is 554)", "SAT转IP 服务端口号\n（默认：-1是disable，0是webconfig，常规端口是554）",
"SAT>IP Satellite Configuration", "SAT转IP 卫星配置",
"SAT>IP Server", "SAT转IP 服务器",
"SAT>IP Server SI Tables", "SAT转IP 服务器 SI表",
"SAT>IP source number", "SAT转IP 源节目号",
"Sat", "卫星",
"Schedule a re-recording if more errors than (0=off)", "在出错大于多少时计划一个重录制(0=关闭)",
"Schedule status", "计划状态",
"Schedule status.", "计划状态。",
"Scheduled Duration", "计划的持续时间",
"Scheduled for recording", "已计划录制",
"Scheduled start time", "计划的开始时间",
"Scheduled stop time", "计划的结束时间",
"Script/program to run when a recording completes.", "当录制结束时运行的脚本/程序。",
"Script/program to run when a recording gets removed.", "当录像被删除时运行的脚本/程序。",
"Script/program to run when a recording starts (service is subscribed but no filename available).", "当录制开始时运行的脚本/程序(服务已请阅，但无文件名可用)。",
"Season", "季",
"Season %d", "季%d",
"Season information (if available).", "季信息(如果有)。",
"Select type of broadcast.", "选择广播类型。",
"Series link", "系列链接",
"Series link ID.", "系列链接ID",
"Skip commercials", "跳过广告",
"Slave entry", "从记录",
"Slave entry.", "从记录。",
"Start", "开始",
"Start after", "在此之后开始",
"Start before", "在此之前开始",
"Start recording earlier than the EPG/timer-defined start time by x minutes.", "在EPG/计时器定义的开始时间之前X分钟开始录制。",
"Start recording earlier than the defined Start recording earlier than the defined start time by x minutes: for example, if a program is to start at 13:00 and you set a padding of 5 minutes it will start recording at 12:54:30 (including a warm-up time of 30 seconds). If this isn't specified, any pre-recording padding as set in the channel or DVR entry will be used.", "在定义的开始时间之前定义的开始录制时间之前X分钟开始录制：例，如果一节目在13:00开始，并且你设置了一个5分钟的提前量，那么将在12:54:30开始录制(包括30秒的热机时间)。如果此项设有设置，将使用任何频道或DVR条目设置的提前录制量。",
"Start recording earlier than the defined start time by x minutes.", "在定义的开始时间之前X分钟开始录制。",
"Start time", "开始时间",
"Status", "状态",
"Stop", "停止",
"Stop time", "停止时间",
"Stream profile", "流策略",
"Subdirectory options", "子目录选项",
"Subtitle", "字幕",
"Subtitle of the program (if any) (display only).", "节目字幕(如果有)(仅显示)。",
"Subtitle of the program (if any).", "节目字幕(如果有)。",
"Sun", "周日",
"Sync", "同步",
"Sync + Don't keep", "同步+不保持",
"System", "系统",
"Tag files with metadata", "带元数据的标签文件",
"The DVR profile to be used/used by the recording.", "录像已/将使用的DVR策略。",
"The DVR profile to be used/used by this rule.", "此规则已/将使用的DVR策略。",
"The EPG ID used by the entry.", "此条目使用的EPG ID。",
"The cache scheme to use/used to store recordings. Leave as \"system\" unless you have a special use case for one of the others. See Help for details.", "将/已使用的存储录像文件的缓存策略。将它设为“系统，”除非你有特殊的用途。查看帮助以获取详情。",
"The channel name the entry will record from.", "此记录将要录制的频道名。",
"The channel on which this rule applies, i.e. the channel you're aiming to record. You can leave this field blank to apply the rule to all channels.", "此规则将应用到哪些频道。例：你正在计划要录制的频道。你可将此项留白以将此规则应用到所有频道。",
"Use service IDs as channel numbers", "使用服务ID作为频道号",
"Use tag internally (don't expose to clients).", "使用内部标记（不会在客户端显示）",
"Use the provider's service IDs as channel numbers.", "采用私有服务ID作为频道号",
"Use this character set when setting filenames.", "设置文件名时采用这个字符集",
"User access error", "用户访问错误",
"User does not have access", "用户无法访问",
"User does not have access to channel", "用户无法访问频道",
"User icon", "用户图标",
"User interface level", "用户权限级别",
"User language", "用户语言",
"User login", "用户登录",
"User request", "用户请求",
"Username", "用户名",
"Username for the entry (login username).", "输入用户名(登录用户名)",
"Username of the entry (this should match a username from within the \"Access Entries\" tab).", "输入用户名（与用户-入口处信息保持一致）",
"Users - Access Entries", "用户-入口",
"Users - IP Blocking", "用户-IP封锁",
"Users - Passwords", "用户-密码",
"Variety show", "显示验证信息",
"Verify the SSL certificate.", "验证SSL证书",
"Verify the peer's SSL.", "验证对等的SSL证书",
"Video", "视频",
"Video Stream Filter", "视频流过滤",
"Video bitrate (kb/s) (0=auto)", "视频码率(单位kb/s)(0代表自动)",
"Video codec", "视频解码器",
"View all", "访问全部",
"Wizard", "配置向导",
NULL, NULL
};

static struct tvh_locale tvh_locales[] = {
{ "ach", tvh_locale_1 },
{ "ady", tvh_locale_2 },
{ "ar", tvh_locale_3 },
{ "az", tvh_locale_4 },
{ "bg", tvh_locale_5 },
{ "cs", tvh_locale_6 },
{ "da", tvh_locale_7 },
{ "de", tvh_locale_8 },
{ "en_US", tvh_locale_9 },
{ "en_GB", tvh_locale_10 },
{ "es", tvh_locale_11 },
{ "et", tvh_locale_12 },
{ "fa", tvh_locale_13 },
{ "fi", tvh_locale_14 },
{ "fr", tvh_locale_15 },
{ "he", tvh_locale_16 },
{ "hr", tvh_locale_17 },
{ "hu", tvh_locale_18 },
{ "it", tvh_locale_19 },
{ "ko", tvh_locale_20 },
{ "lv", tvh_locale_21 },
{ "lt", tvh_locale_22 },
{ "nl", tvh_locale_23 },
{ "no", tvh_locale_24 },
{ "pl", tvh_locale_25 },
{ "pt", tvh_locale_26 },
{ "ro", tvh_locale_27 },
{ "ru", tvh_locale_28 },
{ "sl", tvh_locale_29 },
{ "sk", tvh_locale_30 },
{ "sq", tvh_locale_31 },
{ "sv", tvh_locale_32 },
{ "tr", tvh_locale_33 },
{ "uk", tvh_locale_34 },
{ "zh", tvh_locale_35 },
{ "zh-Hans", tvh_locale_36 }
};
